Package: ggcube
Type: Package
Title: 3D Plotting with 'ggplot2'
Version: 0.2.0
Authors@R: 
    person("Matthew", "Kling", , "mattkling@berkeley.edu", 
    role = c("aut", "cre", "cph"))
Description: A 'ggplot2' extension for creating 3D figures.
    Provides 3D geoms, stats, and a coord_3d() coordinate
    system supporting rotation, perspective, and lighting.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
URL: https://github.com/matthewkling/ggcube,
        https://matthewkling.github.io/ggcube/
BugReports: https://github.com/matthewkling/ggcube/issues
Depends: R (>= 3.5), ggplot2
Imports: dplyr, tidyr, stringr, scales, rlang, grid, lifecycle,
        labeling, purrr, magrittr, polyclip, isoband, systemfonts,
        htmlwidgets
Suggests: alphashape3d, av, base64enc, geometry, gifski, gtable, knitr,
        MASS, mgcv, patchwork, progress, ragg, rmarkdown, rstudioapi,
        shiny, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/Needs/website: rmarkdown
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-21 13:50:13 UTC; matthewkling
Author: Matthew Kling [aut, cre, cph]
Maintainer: Matthew Kling <mattkling@berkeley.edu>
Repository: CRAN
Date/Publication: 2026-07-21 14:30:02 UTC
Built: R 4.6.1; ; 2026-07-21 17:06:38 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(knitr::knit_print,gif_3d)
S3method(knitr::knit_print,video_3d)
S3method(print,gif_3d)
S3method(print,light)
S3method(print,video_3d)
export(aes)
export(anim_save_3d)
export(animate_3d)
export(annotate_3d)
export(av_renderer_3d)
export(camera_facing)
export(coord_3d)
export(element_rect)
export(file_renderer_3d)
export(geom_bar_3d)
export(geom_col_3d)
export(geom_contour_3d)
export(geom_density_3d)
export(geom_distributions_3d)
export(geom_function_3d)
export(geom_hull_3d)
export(geom_path_3d)
export(geom_point_3d)
export(geom_polygon_3d)
export(geom_ridgeline_3d)
export(geom_segment_3d)
export(geom_smooth_3d)
export(geom_surface_3d)
export(geom_text_3d)
export(geom_voxel_3d)
export(gifski_renderer_3d)
export(guide_colorbar_3d)
export(guide_legend_3d)
export(light)
export(orbit3dOutput)
export(orbit_3d)
export(position_on_face)
export(renderOrbit3d)
export(scale_z_continuous)
export(scale_z_discrete)
export(stat_bar_3d)
export(stat_col_3d)
export(stat_contour_3d)
export(stat_density_3d)
export(stat_distributions_3d)
export(stat_function_3d)
export(stat_hull_3d)
export(stat_identity_3d)
export(stat_path_3d)
export(stat_point_3d)
export(stat_segment_3d)
export(stat_smooth_3d)
export(stat_surface_3d)
export(stat_text_3d)
export(stat_voxel_3d)
export(text_outlines)
export(zlim)
import(dplyr)
import(ggplot2)
importFrom(ggplot2,resolution)
importFrom(grDevices,as.raster)
importFrom(grDevices,chull)
importFrom(grDevices,col2rgb)
importFrom(grDevices,hsv)
importFrom(grDevices,rgb)
importFrom(grDevices,rgb2hsv)
importFrom(grid,convertHeight)
importFrom(grid,convertWidth)
importFrom(grid,gpar)
importFrom(grid,grobHeight)
importFrom(grid,grobTree)
importFrom(grid,grobWidth)
importFrom(grid,nullGrob)
importFrom(grid,polygonGrob)
importFrom(grid,rectGrob)
importFrom(grid,segmentsGrob)
importFrom(grid,textGrob)
importFrom(isoband,isobands)
importFrom(labeling,extended)
importFrom(lifecycle,deprecate_soft)
importFrom(lifecycle,is_present)
importFrom(magrittr,"%>%")
importFrom(polyclip,polyclip)
importFrom(polyclip,polysimplify)
importFrom(purrr,map_dfr)
importFrom(rlang,"%||%")
importFrom(scales,extended_breaks)
importFrom(scales,rescale)
importFrom(stats,ave)
importFrom(stats,density)
importFrom(stats,dnorm)
importFrom(stats,glm)
importFrom(stats,lm)
importFrom(stats,loess)
importFrom(stats,na.omit)
importFrom(stats,predict)
importFrom(stats,setNames)
importFrom(stats,weighted.mean)
importFrom(stringr,str_remove)
importFrom(systemfonts,glyph_outline)
importFrom(systemfonts,match_fonts)
importFrom(systemfonts,shape_string)
importFrom(tidyr,expand_grid)
importFrom(tidyr,gather)
importFrom(tidyr,unite)
importFrom(utils,modifyList)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# ggcube <a href="https://matthewkling.github.io/ggcube/"><img src="man/figures/logo.png" align="right" height="139" alt="ggcube website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/matthewkling/ggcube/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/matthewkling/ggcube/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/ggcube)](https://CRAN.R-project.org/package=ggcube)
<!-- badges: end -->

**ggcube** lets you build 3D figures using ggplot2. Use it to create 3D
surfaces, volumes, scatter plots, and complex layered visualizations
using familiar ggplot2 syntax with `aes(x, y, z)` and `coord_3d()`.

The package provides a variety of 3D-specific `geoms` to render
surfaces, prisms, points, paths, and text in 3D; it also works with some
standard ggplot2 layer functions. You can control plot geometry with 3D
projection parameters, apply a range of 3D lighting models, and mix 3D
layers with 2D layers rendered on cube faces. Standard ggplot2 features
like faceting, themes, scales, and legends work as expected.

3D plots are wonderful for exploration, storytelling, and data art. But
note that for precise quantitative communication, where occlusion and
perspective distortion can be problematic, 2D is usually the better
choice.

## Installation

``` r
# You can install the package from CRAN:
install.packages("ggcube")

# Or get the development version from GitHub:
devtools::install_github("matthewkling/ggcube")
```

## Related packages

R has several other tools for 3D visualization. ggcube is designed for
users who want to stay within the ggplot2 ecosystem. Other packages
offer different tradeoffs:

- [**plotly**](https://plotly.com/r/) and
  [**rgl**](https://dmurdoch.github.io/rgl/) provide interactive 3D
  viewers with rotation and zoom via WebGL/OpenGL, using their own APIs.
- [**rayshader**](https://www.rayshader.com/) produces ray-traced
  renderings of terrain data with photorealistic lighting and shadows;
  it also makes 3D popups of ggplot2 figures, but not 3D ggplots per se.
- [**gg3D**](https://github.com/AckerDWM/gg3D) was another ggplot2
  extension for 3D plotting, but is no longer actively maintained.

## Quick start

The essential ingredient of a ggcube plot is `coord_3d()`. Adding this
to a standard ggplot, and providing a `z` aesthetic variable, creates a
3D plot:

``` r
library(ggplot2)
library(ggcube)

# Basic 3D scatter plot
ggplot(mpg, aes(x = displ, y = hwy, z = drv, color = class)) +
      geom_point() +
      coord_3d()
```

<img src="man/figures/README-quickstart-1.png" alt="" width="100%" />

You can control plot rotation, perspective, and dimensions, as well as
axis label placement and panel selection, via parameters to
`coord_3d()`. See the [3D
view](https://matthewkling.github.io/ggcube/articles/coord_3d.html)
article for a comprehensive guide.

``` r
ggplot(mpg, aes(displ, hwy, drv, color = class)) +
      geom_point() +
      coord_3d(pitch = 0, roll = 60, yaw = 0, dist = 1.4, 
               ratio = c(2, 1, 1), panels = "all") +
      theme(panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .1))
```

<img src="man/figures/README-rotation-1.png" alt="" width="100%" />

## 3D surfaces

- `geom_surface_3d()` renders surfaces based on existing grid data such
  as terrain data
- `geom_ridgeline_3d()` renders surfaces as a series of cross-sections
- `geom_contour_3d()` renders surfaces as layer cakes of stacked
  contours
- `stat_function()` visualizes mathematical functions
- `stat_smooth_3d()` fits statistical models with two predictors and
  visualizes fitted surfaces with confidence intervals
- `stat_density_3d()` creates perspective visualizations of 2D kernel
  density estimates
- `stat_hull_3d()` plots triangulated volumes based on convex or alpha
  hulls of 3D points

See the
[surfaces](https://matthewkling.github.io/ggcube/articles/surfaces.html)
article for a full guide to surface options.

Example: a terrain surface using `geom_surface_3d()`:

``` r
ggplot(mountain, aes(x, y, z)) +
      geom_surface_3d(aes(fill = z, color = z)) +
      scale_fill_viridis_c() + scale_color_viridis_c() +
      coord_3d(ratio = c(1.5, 2, 1), expand = FALSE, panels = "zmin",
               light = light(direction = c(1, 0, 0))) +
      guides(fill = guide_colorbar_3d()) +
      theme_light()
```

<img src="man/figures/README-surfaces-1.png" alt="" width="100%" />

Example: a terrain surface using `geom_contour_3d()`:

``` r
ggplot(mountain, aes(x, y, z)) +
      geom_contour_3d(fill = "black", color = "white", linewidth = .5) +
      coord_3d(yaw = 60, ratio = c(1.5, 2, 1), light = "none") +
      theme_void()
```

<img src="man/figures/README-contour-1.png" alt="" width="100%" />

Example: a mathematical surface using `geom_function_3d()`:

``` r
ggplot() +
      geom_function_3d(fun = function(x1, x2) cos(x1) * sin(x2),
                       xlim = c(-pi, pi), ylim = c(-2*pi, 2*pi),
                       fill = "#7a2100", color = "#b3725b", 
                       grid = "right1", linewidth = .2) +
      coord_3d(yaw = 160, roll = -70, 
               scales = "fixed", ratio = c(1, 1, 2)) +
      labs(x = expression(x[1]),
           y = expression(x[2]),
           z = expression(cos(x[1]) %*% sin(x[2]))) +
      theme_minimal()
```

<img src="man/figures/README-functions-1.png" alt="" width="100%" />

Example: a fitted model surface using `geom_smooth_3d()`:

``` r
# Generate scattered 3D data
set.seed(123)
d <- data.frame(x = rnorm(50),
                y = rnorm(50))
d$z <- d$x + d$x^2 - d$y^2 + rnorm(50)

# Plot GAM fit with uncertainty layers
ggplot(d, aes(x, y, z)) + 
      geom_smooth_3d(aes(fill = after_stat(level)),
                     method = "gam", formula = z ~ te(x, y),
                     se = TRUE, level = 0.99,
                     color = "black", grid = "equilateral") +
      scale_fill_manual(values = c("red", "darkorchid4", "steelblue")) +
      coord_3d(light = NULL)
```

<img src="man/figures/README-smooth-1.png" alt="" width="100%" />

## 3D paths

`geom_path_3d()` renders paths in 3D space with depth-based sorting and
scaling:

``` r
butterfly <- ggcube:::lorenz_attractor(n_points = 8000, dt = .01)
ggplot(butterfly, aes(x, y, z, color = time)) +
      geom_path_3d(linewidth = 0.1, color = "black",
                   position = position_on_face(c("xmax", "ymax", "zmin"))) +
      geom_path_3d(linewidth = 0.3) +
      scale_color_gradientn(colors = c("blue", "purple", "red", "orange")) +
      coord_3d() +
      theme_light()
```

<img src="man/figures/README-paths-1.png" alt="" width="100%" />

## 3D points

While `ggplot2::geom_point()` works with ggcube as demonstrated above,
`geom_point_3d()` creates 3D-aware scatter plots with proper point
ordering, depth-scaled point sizes, and options to include reference
lines and reference points projecting 3D points onto 2D face panels:

``` r
ggplot(mpg, aes(x = displ, y = hwy, z = drv, fill = class)) +
      geom_point_3d(size = 3, shape = 21, color = "black", stroke = .1,
                    ref_lines = TRUE, ref_points = TRUE,
                    ref_faces = c("ymax", "xmax")) +
      coord_3d()
```

<img src="man/figures/README-points-1.png" alt="" width="100%" />

## 3D prisms

- `geom_col_3d()` produces 3D column charts
- `geom_bar_3d()` creates 3D histograms of 2D discrete or continuous
  variables
- `geom_voxel_3d()` renders sparse 3D pixel data as arrays of cubes

Example: a 3D histogram using `geom_bar_3d()`:

``` r
ggplot(faithful, aes(waiting, eruptions)) +
      geom_bar_3d(fill = "steelblue", color = "steelblue", 
                  bins = 15, width = .9) +
      coord_3d(yaw = 60) +
      scale_z_continuous(expand = c(0, 0))
```

<img src="man/figures/README-bar-1.png" alt="" width="100%" />

## 3D text

`geom_text_3d()` creates 3D-aware text, rendered either as “billboard”
text that faces the viewing plane or as 3D polygons that can face any
direction:

``` r
df <- expand.grid(x = c("B", "H", "N"), y = c("a", "o", "u"), z = c("g", "t"))
df$label <- paste0(df$x, df$y, df$z)
ggplot(df, aes(x, y, z, label = label, fill = x)) +
      geom_text_3d(method = "polygon", facing = "zmax", 
                   size = 5, weight = "bold") +
      coord_3d(scales = "fixed", rotate_labels = FALSE) +
      theme(axis.title = element_blank())
```

<img src="man/figures/README-text-1.png" alt="" width="100%" />

## Lighting effects

Lighting of 3D polygon layers is controlled by adding a `light()`
specification to the plot, or providing to the `light` parameter of a
layer function or `coord_3d()`. See the [lighting and
shading](https://matthewkling.github.io/ggcube/articles/lighting.html)
article for a comprehensive guide.

``` r
ggplot(sphere_points, aes(x, y, z)) +
      coord_3d(scales = "fixed") +
      scale_fill_viridis_c() +
      scale_color_viridis_c() +
      theme_dark() +
      theme(legend.position = "none") +
      
      # apply shading to solid color/fill
      geom_hull_3d(fill = "#8a2900", color = "#8a2900",
                   light = light(method = "direct", mode = "hsl", 
                                 direction = c(0, 0, 1))) +
      
      # apply shading to aesthetic color/fill
      geom_hull_3d(aes(x = x + 2.5, fill = x, color = x),
                   light = light(method = "diffuse", mode = "hsv", 
                                 direction = c(0, 0, 1), contrast = 2)) +
      
      # map surface orientation to 3D RGB color channels
      geom_hull_3d(aes(x = x + 5),
                   light = light(method = "rgb", direction = c(1, 0, -1)))
```

<img src="man/figures/README-lighting-1.png" alt="" width="100%" />

## Animation and interaction

A major limitation of 3D figures is that you can’t get a full view of
the data from any single angle. One way to mitigate this is by rotating
the figure to view the data from different directions. ggcube offers
animated rotation via `animate_3d()`, and interactive drag-to-rotate
plots via `orbit_3d()`. See the [animation and
interaction](https://matthewkling.github.io/ggcube/articles/animation.html)
article for details. Here’s an example of a rotating gif:

``` r
mammoth <- data.frame(do.call(rbind, rjson::fromJSON(
  file = paste0("https://raw.githubusercontent.com/PAIR-code/",
                "understanding-umap/master/raw_data/mammoth_3d_50k.json")
)))
colnames(mammoth) <- c("Y", "X", "Z")

p <- ggplot(mammoth, aes(X, Y, Z)) +
      geom_hull_3d(method = "alpha", radius = 8, 
                   fill = "darkred", color = "darkred") +
      light(mode = "hsl", direction = c(-1, 1, 0), anchor = "camera") +
      coord_3d(roll = -90, scales = "fixed", 
               panels = "zmin", expand = FALSE) +
      theme(axis.text = element_blank(),
            axis.ticks = element_blank(),
            axis.title = element_blank())

animate_3d(p, yaw = c(0, 360), nframes = 72, fps = 8, width = 700, cores = 8)
```

<img src="man/figures/README-anim-1.gif" alt="" width="100%" />

## Face projection

3D and 2D layers can be mixed by using `position_on_face()` to project
data onto 2D cube faces. We saw this in the `geom_path_3d()` example
above, but here’s another example that mixes different geoms, including
natively-2D layers like `ggplot2::stat_density_2d()`:

``` r

ggplot(iris, aes(Sepal.Length, Sepal.Width, Petal.Length, 
                 color = Species, fill = Species)) +
      coord_3d() + xlim(4, 8) +
      
      # place 2D density plot on zmin face
      stat_density_2d(position = position_on_face(faces = "zmin", axes = c("x", "y")),
                      geom = "polygon", alpha = .1, linewidth = .25) +
      
      # flatten 3D hull layer onto ymax face
      geom_hull_3d(position = position_on_face("ymax"), alpha = .5) +
      
      # flatten 3D voxels onto xmax face to create 2D bins
      geom_voxel_3d(aes(round(Sepal.Length), round(Sepal.Width), round(Petal.Length)),
                    position = position_on_face("xmax"), alpha = .15, light = NULL) +
      
      # 3D scatter plot (added last so it renders in front)
      geom_point_3d( shape = 21, color = "black", stroke = .25)
```

<img src="man/figures/README-position-1.png" alt="" width="100%" />

## Learn more

- The [getting
  started](https://matthewkling.github.io/ggcube/articles/ggcube.html)
  vignette gives a tour of the main features.
- The [3D
  view](https://matthewkling.github.io/ggcube/articles/coord_3d.html),
  [surfaces](https://matthewkling.github.io/ggcube/articles/surfaces.html),
  and [lighting and
  shading](https://matthewkling.github.io/ggcube/articles/lighting.html)
  articles go deeper on those topics.
- The [function
  reference](https://matthewkling.github.io/ggcube/reference/index.html)
  documents every exported function.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi = 150,
  warning = FALSE,
  message = FALSE
)

## ----setup, message = FALSE---------------------------------------------------
library(ggcube)

## ----basics-------------------------------------------------------------------
ggplot(mpg, aes(x = displ, y = hwy, z = drv, color = class)) +
      geom_point() +
      coord_3d()

## ----rotation, fig.height = 4-------------------------------------------------
ggplot(mpg, aes(displ, hwy, drv, color = class)) +
      geom_point() +
      coord_3d(pitch = 0, roll = 60, yaw = 0, dist = 1.4,
               ratio = c(2, 1, 1), panels = "all") +
      theme(panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .1))

## ----surface------------------------------------------------------------------
ggplot(mountain, aes(x, y, z)) +
      geom_surface_3d(aes(fill = z, color = z)) +
      scale_fill_viridis_c() + scale_color_viridis_c() +
      coord_3d(ratio = c(1.5, 2, 1), expand = FALSE, panels = "zmin",
               light = light(direction = c(1, 0, 0))) +
      guides(fill = guide_colorbar_3d()) +
      theme_light()

## ----points-------------------------------------------------------------------
ggplot(mpg, aes(x = displ, y = hwy, z = drv, fill = class)) +
      geom_point_3d(size = 3, shape = 21, color = "black", stroke = .1,
                    ref_lines = TRUE, ref_faces = c("ymax", "xmax")) +
      coord_3d()

## ----paths--------------------------------------------------------------------
x <- seq(0, 20*pi, pi/16)
spiral <- data.frame(x = x, y = sin(x), z = cos(x), time = 1:length(x))
ggplot(spiral, aes(x, y, z, color = time)) +
      geom_path_3d() +
      scale_color_gradientn(colors = c("blue", "purple", "red", "orange")) +
      coord_3d() +
      theme_light()

## ----bar----------------------------------------------------------------------
ggplot(iris, aes(Species, Sepal.Length, fill = Species)) +
      geom_bar_3d(bins = 20, width = c(.5, 1)) +
      coord_3d(scales = "fixed", ratio = c(1, 3, .25), yaw = 60) +
      scale_z_continuous(expand = c(0, 0)) +
      theme(legend.position = "none")

## ----hull---------------------------------------------------------------------
ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d(method = "convex", fill = "#9e2602", color = "#5e1600") +
      coord_3d()

## ----distributions------------------------------------------------------------
ggplot(iris, aes(y = Sepal.Length, x = Species, fill = Species)) +
      stat_distributions_3d() +
      scale_z_continuous(expand = expansion(mult = c(0, NA))) +
      coord_3d() +
      theme(legend.position = "none")

## ----text---------------------------------------------------------------------
df <- expand.grid(x = c("H", "B"), y = c("a", "o", "u"), z = c("g", "t"))
df$label <- paste0(df$x, df$y, df$z)
ggplot(df, aes(x, y, z, label = label, fill = x)) +
      geom_text_3d(method = "polygon", facing = "zmax",
                   size = 5, weight = "bold") +
      coord_3d(scales = "fixed", light = NULL)

## ----lighting-----------------------------------------------------------------
p <- ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d(fill = "#9e2602", color = "#5e1600")

p + coord_3d(light = light(method = "direct", mode = "hsl",
                           direction = c(0, 0, 1)))

## ----zlim, fig.show = "hide"--------------------------------------------------
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
      geom_point() +
      zlim(15, 20) +
      coord_3d()

## ----guide, fig.show = "hide"-------------------------------------------------
ggplot(mountain, aes(x, y, z, fill = z)) +
      stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
      guides(fill = guide_colorbar_3d()) +
      scale_fill_gradientn(colors = c("tomato", "dodgerblue")) +
      coord_3d()

## ----theme--------------------------------------------------------------------
ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d() +
      coord_3d(panels = "all") +
      theme(panel.background = element_rect(color = "black"),
            panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .3),
            panel.grid.foreground = element_line(color = "gray", linewidth = .25),
            axis.text = element_text(color = "darkblue"),
            axis.text.z = element_text(color = "darkred"),
            axis.title = element_text(margin = margin(t = 30)),
            axis.title.x = element_text(color = "magenta"))

## ----annotate, eval = FALSE---------------------------------------------------
# summit <- filter(mountain, z == max(z))
# ggplot(mountain, aes(x, y, z)) +
#       geom_contour_3d(
#             annotate = list(
#                   annotate_3d("point", x = summit$x, y = summit$y, z = summit$z, color = "red"),
#                   annotate_3d("text", x = summit$x, y = summit$y, z = summit$z, color = "red",
#                               label = "Summit", fontface = "bold", vjust = -1)
#             ), fill = "black"
#       ) +
#       coord_3d(ratio = c(2, 3, 1.5), light = "none")

## ----face_projection----------------------------------------------------------
ggplot(iris, aes(Sepal.Length, Sepal.Width, Petal.Length,
                 color = Species, fill = Species)) +
      coord_3d() + xlim(4, 8) +
      stat_density_2d(position = position_on_face(faces = "zmin", axes = c("x", "y")),
                      geom = "polygon", alpha = .1, linewidth = .25) +
      geom_hull_3d(position = position_on_face("ymax"), alpha = .5) +
      geom_point_3d(shape = 21, color = "black", stroke = .25)

## ----animation, eval = FALSE--------------------------------------------------
# p <- ggplot(mountain, aes(x, y, z)) +
#       geom_contour_3d(fill = "black", color = "white", linewidth = .5) +
#       coord_3d(ratio = c(1.5, 2, 1), light = "none", zoom = 1.25) +
#       theme_void()
# 
# orbit_3d(p, yaw = c(360, 0), roll = c(-90, 0), n = c(24, 8),
#          start = c(yaw = 300, roll = -60))

StatCol3D <- ggproto("StatCol3D", Stat,
                     required_aes = c("x", "y", "z"),
                     optional_aes = c("zmin"),
                     default_aes = aes(group = after_stat(group)),

                     compute_panel = function(data, scales, na.rm = FALSE,
                                              width = 1.0, faces = "all",
                                              light = NULL, zmin = NULL,
                                              cull_backfaces = NULL) {

                           # Remove missing values if requested
                           if (na.rm) {
                                 data <- data[complete.cases(data[c("x", "y", "z")]), ]
                           }

                           # Check we have enough data
                           if (nrow(data) < 1) {
                                 stop("stat_col_3d requires at least 1 point")
                           }

                           # Handle zmin with proper precedence:
                           # 1. Parameter (if provided and not NULL) - overrides aesthetic
                           # 2. Aesthetic (if mapped) - used if parameter is NULL
                           # 3. Default (0) - if neither provided

                           if (!is.null(zmin)) {
                                 # Explicit parameter provided - use for all columns (overrides aesthetic)
                                 data$zmin <- zmin
                           } else if (!"zmin" %in% names(data)) {
                                 # No aesthetic mapped and no parameter - use default of 0
                                 data$zmin <- 0
                           }
                           # If aesthetic exists and parameter is NULL, keep the aesthetic values

                           # Ensure z is never less than zmin (handle negative bars correctly)
                           zmin_vals <- pmin(data$zmin, data$z)
                           zmax_vals <- pmax(data$zmin, data$z)
                           data$zmin <- zmin_vals
                           data$z <- zmax_vals

                           # Generate numeric positions before calculating spacing
                           data <- convert_to_numeric(data)

                           # Calculate grid spacing using resolution (works for both regular and sparse grids)
                           x_spacing <- resolution(data$x, zero = FALSE)
                           y_spacing <- resolution(data$y, zero = FALSE)

                           # Validate and process faces parameter
                           selected_faces <- select_faces(faces)

                           if (length(selected_faces) == 0) {
                                 # Return empty data frame with required columns for consistency
                                 return(data.frame(
                                       x = numeric(0), y = numeric(0), z = numeric(0),
                                       group = character(0), col_id = integer(0), face_type = character(0),
                                       light = numeric(0),
                                       normal_x = numeric(0), normal_y = numeric(0), normal_z = numeric(0)
                                 ))
                           }

                           # Create columns
                           col_faces <- create_cols(data, x_spacing, y_spacing, width, selected_faces)

                           if (nrow(col_faces) == 0) {
                                 # Return empty data frame with required columns for consistency
                                 return(data.frame(
                                       x = numeric(0), y = numeric(0), z = numeric(0),
                                       group = character(0), col_id = integer(0), face_type = character(0),
                                       light = numeric(0),
                                       normal_x = numeric(0), normal_y = numeric(0), normal_z = numeric(0)
                                 ))
                           }

                           col_faces %>%
                                 average_aesthetics() %>%
                                 mutate(cull_backfaces = cull_backfaces) %>%
                                 attach_light(light) %>%
                                 return()
                     }
)

#' Create column faces from grid data
#'
#' @param data Data frame with x, y, z, zmin columns
#' @param x_spacing Grid spacing in x direction
#' @param y_spacing Grid spacing in y direction
#' @param width Width factor (1.0 = full grid spacing)
#' @param selected_faces Character vector of face names to render
#' @return Data frame with column face vertices
#' @keywords internal
#' @noRd
create_cols <- function(data, x_spacing, y_spacing, width, selected_faces) {

      # Calculate actual column dimensions
      col_width_x <- x_spacing * width
      col_width_y <- y_spacing * width
      half_x <- col_width_x / 2
      half_y <- col_width_y / 2

      all_faces <- list()

      # Debug: check inputs
      if (nrow(data) == 0) {
            warning("create_cols: No data points provided")
            return(data.frame())
      }

      if (length(selected_faces) == 0) {
            warning("create_cols: No faces selected")
            return(data.frame())
      }

      for (i in 1:nrow(data)) {
            point <- data[i, ]
            cx <- point$x  # Center x
            cy <- point$y  # Center y

            # Define bounds
            # xmin_bound = left, xmax_bound = right
            # ymin_bound = front (smaller y), ymax_bound = back (larger y)
            # zmin_bound = bottom, zmax_bound = top
            xmin_bound <- cx - half_x
            xmax_bound <- cx + half_x
            ymin_bound <- cy - half_y
            ymax_bound <- cy + half_y
            zmin_bound <- point$zmin
            zmax_bound <- point$z

            # Define the 8 corners of the column
            # Using consistent naming: corners indexed by (x, y, z) position
            corners <- list(
                  c(xmin_bound, ymin_bound, zmin_bound),  # 1: left-front-bottom
                  c(xmax_bound, ymin_bound, zmin_bound),  # 2: right-front-bottom
                  c(xmax_bound, ymax_bound, zmin_bound),  # 3: right-back-bottom
                  c(xmin_bound, ymax_bound, zmin_bound),  # 4: left-back-bottom
                  c(xmin_bound, ymin_bound, zmax_bound),  # 5: left-front-top
                  c(xmax_bound, ymin_bound, zmax_bound),  # 6: right-front-top
                  c(xmax_bound, ymax_bound, zmax_bound),  # 7: right-back-top
                  c(xmin_bound, ymax_bound, zmax_bound)   # 8: left-back-top
            )

            # Define faces using corner indices (order: ccw when viewed from outside)
            # Each face uses corners that share the same value on that axis
            face_definitions <- list(
                  zmin = c(1, 4, 3, 2),  # Bottom face (all corners with z = zmin_bound), ccw from below
                  zmax = c(5, 6, 7, 8),  # Top face (all corners with z = zmax_bound), ccw from above
                  xmin = c(1, 5, 8, 4),  # Left face (all corners with x = xmin_bound), ccw from left
                  xmax = c(2, 3, 7, 6),  # Right face (all corners with x = xmax_bound), ccw from right
                  ymin = c(1, 2, 6, 5),  # Front face (all corners with y = ymin_bound), ccw from front
                  ymax = c(4, 8, 7, 3)   # Back face (all corners with y = ymax_bound), ccw from back
            )

            # Create requested faces
            for (face_name in selected_faces) {
                  if (face_name %in% names(face_definitions)) {
                        corner_indices <- face_definitions[[face_name]]

                        # Create 2-level hierarchical group ID: col_id__face_type
                        hierarchical_group <- paste0("col", i, "__", face_name)

                        # Create face vertices with hierarchical grouping
                        face_vertices <- data.frame(
                              x = sapply(corner_indices, function(idx) corners[[idx]][1]),
                              y = sapply(corner_indices, function(idx) corners[[idx]][2]),
                              z = sapply(corner_indices, function(idx) corners[[idx]][3]),
                              z_raw = point$z_raw,
                              group = hierarchical_group,
                              col_id = i,
                              face_type = face_name
                        )

                        # Preserve all non-coordinate columns
                        non_coord_cols <- setdiff(names(point), names(face_vertices))
                        for (col_name in non_coord_cols) {
                              face_vertices[[col_name]] <- rep(point[[col_name]], 4)
                        }

                        all_faces[[length(all_faces) + 1]] <- face_vertices
                  }
            }
      }

      if (length(all_faces) == 0) {
            return(data.frame())
      }

      # Combine all faces
      result <- do.call(rbind, all_faces)
      rownames(result) <- NULL

      return(result)
}


#' 3D columns from grid data
#'
#' Creates 3D columns (rectangular prisms) from grid data in which x and y fall on a
#' regular grid. Works with both complete and sparse grid data. Each data point becomes
#' a rectangular 3D column extending from a base level to the data value.
#'
#' This is analogous to [ggplot2::geom_col()] for 3D plots. For automatic counting or
#' binning, see [geom_bar_3d()].
#'
#' Note that column geometries often require pairwise depth sorting for correct rendering.
#' This is the default for smaller data sets, but not for larger data sets due to compute speed;
#' in those cases you may wish to manually specify `sort_method = "pairwise"` for good results.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use on the data. Defaults to `StatCol3D`.
#' @param geom The geometric object used to display the data. Defaults to `GeomPolygon3D`.
#'
#' @param zmin Base level for all columns. When provided as a parameter, overrides any
#'   \code{zmin} aesthetic mapping. If \code{NULL} (the default), uses the \code{zmin}
#'   aesthetic if mapped, otherwise defaults to 0.
#'
#' @inheritParams col_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' `stat_col_3d()` requires the following aesthetics:
#' - **x**: X coordinate (grid position)
#' - **y**: Y coordinate (grid position)
#' - **z**: Z coordinate (column top height)
#'
#' And optionally understands:
#' - **zmin**: Base level for each column (can be overridden by the \code{zmin} parameter)
#'
#' @section Computed variables:
#' - `normal_x`, `normal_y`, `normal_z`: Face normal components
#' - `col_id`: Sequential column number
#' - `face_type`: Face name ("zmax", "xmin", etc.)
#'
#' @examples
#' # Basic 3D bar chart from regular grid
#' # (columns extend from z=0 by default)
#' d <- expand.grid(x = 1:5, y = 1:5)
#' d$z <- d$x + d$y + rnorm(25, 0, 0.5)
#' ggplot(d, aes(x, y, z)) +
#'   geom_col_3d() +
#'   coord_3d()
#'
#' # Set uniform base level using `zmin` parameter
#' ggplot(d, aes(x, y, z)) +
#'   geom_col_3d(aes(fill = z), color = "white",
#'               zmin = 5) +
#'   coord_3d()
#'
#' # Set variable base levels using `zmin` aesthetic
#' d$base_level <- runif(nrow(d), -5, 1)
#' ggplot(d, aes(x, y, z = z, zmin = base_level)) +
#'   geom_col_3d(color = "black") +
#'   coord_3d()
#'
#' # Show only a subset of column faces
#' ggplot(d, aes(x, y, z)) +
#'   geom_col_3d(faces = c("zmax", "ymin"),
#'     cull_backfaces = FALSE,
#'     fill = "steelblue", color = "black") +
#'   coord_3d()
#'
#' # With gaps between columns
#' ggplot(d, aes(x, y, z)) +
#'   geom_col_3d(color = "black", width = 0.6) +
#'   coord_3d()
#'
#' @seealso [geom_bar_3d()] for automatic counting/binning, [stat_surface_3d()] for smooth
#'   surface rendering, [coord_3d()] for 3D coordinate systems, [light()] for lighting
#'   specifications.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_col_3d
#' @export
geom_col_3d <- function(mapping = NULL, data = NULL,
                        stat = StatCol3D,
                        position = "identity",
                        ...,
                        width = 1.0, faces = "all", zmin = NULL,
                        light = NULL,
                        cull_backfaces = TRUE, sort_method = NULL,
                        scale_depth = TRUE, force_convex = FALSE,
                        na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomPolygon3D,
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, width = width, faces = faces, light = light, zmin = zmin,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}


#' @rdname geom_col_3d
#' @export
stat_col_3d <- function(mapping = NULL, data = NULL,
                        geom = GeomPolygon3D,
                        position = "identity",
                        ...,
                        width = 1.0, faces = "all", zmin = NULL,
                        light = NULL,
                        cull_backfaces = TRUE, sort_method = NULL,
                        scale_depth = TRUE, force_convex = FALSE,
                        na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = StatCol3D, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, width = width, faces = faces, light = light, zmin = zmin,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}

test_that("continuous and discrete scales run without error for various stats", {
      cont <- expand.grid(a = 1:3, b = 1:3, c = 1:3)
      disc <- expand.grid(a = letters[1:3], b = letters[1:3], c = letters[1:3])
      mix <- expand.grid(a = 1:3, b = letters[1:3], c = letters[1:3])

      test <- function(d, layer, scl = "fixed"){
            cont %>%
                  ggplot(aes(a, b, c)) +
                  layer +
                  coord_3d(scales = scl)
      }

      expect_no_error({
            test(cont, geom_point())
            test(disc, geom_point())
            test(mix, geom_point())

            test(cont, geom_point_3d())
            test(disc, geom_point_3d())
            test(mix, geom_point_3d())

            test(cont, stat_voxel_3d())
            test(disc, stat_voxel_3d())
            test(mix, stat_voxel_3d())

            test(cont, stat_col_3d())
            test(disc, stat_col_3d())
            test(mix, stat_col_3d())
      })
})

test_that("each plot uses its own z scale, regardless of construction order", {

      d <- expand.grid(x = 1:5, y = 1:5)

      make_plot <- function(upper) {
            dd <- d
            dd$z <- seq(0, upper, length.out = nrow(dd))
            ggplot(dd, aes(x, y, z = z)) +
                  geom_point() +
                  scale_z_continuous(limits = c(0, upper), expand = expansion(0)) +
                  coord_3d()
      }

      z_limits <- function(p) {
            ggplot_build(p)$layout$panel_params[[1]]$scale_info$z$limits
      }

      p_small <- make_plot(0.001)
      p_big <- make_plot(1000)

      # p_small's scale_z_continuous() call is no longer the most recent one
      expect_equal(z_limits(p_small), c(0, 0.001))
      expect_equal(z_limits(p_big), c(0, 1000))

      # And rebuilding in the other order gives the same answer
      expect_equal(z_limits(p_big), c(0, 1000))
      expect_equal(z_limits(p_small), c(0, 0.001))
})

test_that("repeated builds don't accumulate z scale range", {

      d <- expand.grid(x = 1:5, y = 1:5)
      d$z <- d$x * d$y

      p <- ggplot(d, aes(x, y, z = z)) +
            geom_point() +
            coord_3d()

      z_limits <- function(p) {
            ggplot_build(p)$layout$panel_params[[1]]$scale_info$z$limits
      }

      expect_equal(z_limits(p), z_limits(p))
})

StatDensity3D <- ggproto("StatDensity3D", Stat,
                         required_aes = c("x", "y"),
                         default_aes = aes(fill = after_stat(density), z = after_stat(z)),

                         compute_group = function(data, scales,
                                                  n = NULL, grid = NULL, direction = NULL, trim = NULL,
                                                  h = NULL, adjust = 1, pad = 0.1, min_ndensity = 0,
                                                  cull_backfaces = NULL,
                                                  light = NULL, na.rm = FALSE) {

                               # Remove missing values if requested
                               if (na.rm) {
                                     data <- data[complete.cases(data[c("x", "y")]), ]
                               }

                               # Error if grid is incompatible with min_ndensity
                               if (min_ndensity > 0 && identical(grid, "equilateral")) {
                                     stop("`min_ndensity` is not supported with `grid = \"equilateral\"`; ",
                                             "ignoring. Use `grid = \"rectangle\"`, `\"right1\"`, or `\"right2\"`.")
                               }

                               # Check we have enough data
                               if (nrow(data) < 3) {
                                     stop("stat_density_3d requires at least 3 points per group")
                               }

                               # Load MASS for kde2d
                               if (!requireNamespace("MASS", quietly = TRUE)) {
                                     stop("stat_density_3d requires the 'MASS' package. Install with: install.packages('MASS')")
                               }

                               # Calculate bandwidth if not provided
                               if (is.null(h)) {
                                     # Use same approach as MASS::kde2d default
                                     h <- c(MASS::bandwidth.nrd(data$x), MASS::bandwidth.nrd(data$y))
                               } else if (length(h) == 1) {
                                     h <- c(h, h)
                               } else if (length(h) != 2) {
                                     stop("h must be a single number or a vector of length 2")
                               }

                               # Apply bandwidth adjustment
                               h <- h * adjust

                               # Determine grid limits from data range
                               x_range <- range(data$x, na.rm = TRUE)
                               y_range <- range(data$y, na.rm = TRUE)

                               # Add some padding (similar to what kde2d does by default)
                               x_extend <- diff(x_range) * pad
                               y_extend <- diff(y_range) * pad
                               xlim <- c(x_range[1] - x_extend, x_range[2] + x_extend)
                               ylim <- c(y_range[1] - y_extend, y_range[2] + y_extend)

                               # # Generate grid
                               d <- make_point_grid(grid, n, direction, xlim, ylim, trim)

                               # Compute kernel density
                               d$z <- kde2d(data$x, data$y, d$x, d$y, h)

                               # Preserve group-level aesthetics (everything except coordinates)
                               non_coord_cols <- setdiff(names(data), names(d))
                               if (length(non_coord_cols) > 0) {
                                     for (col_name in non_coord_cols) {
                                           d[[col_name]] <- rep(data[[col_name]][1], nrow(d))
                                     }
                               }

                               # Compute additional variables to match stat_density_2d
                               d$density <- d$z
                               n_obs <- nrow(data)
                               max_density <- max(d$density, na.rm = TRUE)
                               d$ndensity <- if (max_density > 0) d$density / max_density else 0
                               d$count <- d$density * n_obs
                               d$n <- n_obs

                               d <- compute_point_gradients(d)

                               # Remove data below ndensity threshold. Row/column indices are
                               # retained so the geom can tile the surviving sparse grid without
                               # falling back to Delaunay triangulation.
                               if(min_ndensity > 0){
                                     d <- filter(d, ndensity >= min_ndensity)
                               }

                               # Add computed variables and light info
                               d <- d %>%
                                     # compute_surface_vars() %>%
                                     # average_aesthetics() %>%
                                     mutate(cull_backfaces = cull_backfaces) %>%
                                     attach_light(light)

                               return(d)
                         }
)

# Adapted from MASS::kde2d in order to support non-rectangular grids
kde2d <- function(x, y, eval_x, eval_y, h) {
      # x, y: data points
      # eval_x, eval_y: points where we want to evaluate density
      # h: bandwidth (optional)

      nx <- length(x)
      if (length(y) != nx)
            stop("data vectors must be the same length")
      if (length(eval_x) != length(eval_y))
            stop("evaluation point vectors must be the same length")
      if (any(!is.finite(x)) || any(!is.finite(y)))
            stop("missing or infinite values in the data are not allowed")
      if (any(!is.finite(eval_x)) || any(!is.finite(eval_y)))
            stop("missing or infinite values in evaluation points are not allowed")

      # Estimate bandwidth if not provided
      h <- if (missing(h))
            c(MASS::bandwidth.nrd(x), MASS::bandwidth.nrd(y))
      else rep(h, length.out = 2L)
      if (any(h <= 0))
            stop("bandwidths must be strictly positive")
      h <- h/4  # Same scaling as original kde2d

      n_eval <- length(eval_x)

      # For each evaluation point, compute density
      # Vectorized approach: compute all distances at once

      # Create matrices for vectorized computation
      # eval_x[i] - x[j] for all i,j combinations
      dx_matrix <- outer(eval_x, x, "-") / h[1L]  # n_eval x nx
      dy_matrix <- outer(eval_y, y, "-") / h[2L]  # n_eval x nx

      # Evaluate normal densities
      kernel_x <- dnorm(dx_matrix)  # n_eval x nx
      kernel_y <- dnorm(dy_matrix)  # n_eval x nx

      # For each evaluation point, sum over all data points
      # kernel_x[i,j] * kernel_y[i,j] gives kernel contribution from data point j to eval point i
      density <- rowSums(kernel_x * kernel_y) / (nx * h[1L] * h[2L])

      return(density)
}


#' 3D surface from 2D density estimate
#'
#' A 3D version of `ggplot2::stat_density_2d()`.
#' Creates surfaces from 2D point data using kernel density estimation.
#' The density values become the z-coordinates of the surface, allowing
#' visualization of data concentration as peaks and valleys in 3D space.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. This stat
#'   requires `x` and `y` aesthetics. By default, `fill` is mapped to
#'   `after_stat(density)` and `z` is mapped to `after_stat(density)`.
#' @param data The data to be displayed in this layer. Must contain x and y columns
#'   with point coordinates.
#' @param stat The statistical transformation to use on the data. Defaults to `StatDensity3D`.
#' @param geom The geometric object used to display the data. Defaults to `GeomPolygon3D`.
#'
#' @param h Bandwidth vector. If `NULL` (default), uses automatic bandwidth selection
#'   via `MASS::bandwidth.nrd()`. Can be a single number (used for both dimensions)
#'   or a vector of length 2 for different bandwidths in x and y directions.
#' @param adjust Multiplicative bandwidth adjustment factor. Values greater than 1
#'   produce smoother surfaces; values less than 1 produce more detailed surfaces.
#'   Default is 1.
#' @param pad Proportional range expansion factor. The computed density grid extends
#'   this proportion of the raw data range beyond each data limit. Default is 0.1.
#' @param min_ndensity Lower cutoff for normalized density (computed variable `ndensity`
#'   described below), below which to filter out results. This is particularly useful for
#'   removing low-density corners of rectangular density grids when density surfaces are
#'   shown for multiple groups, as in the example below. Default is 0 (no filtering).
#'
#' @inheritParams grid_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' `stat_density_3d()` requires the following aesthetics from input data:
#' - **x**: X coordinate of data points
#' - **y**: Y coordinate of data points
#'
#' And optionally understands:
#' - **group**: Grouping variable for computing separate density surfaces
#' - Additional aesthetics are passed through for surface styling
#'
#' @inheritSection surface_computed_vars Computed variables
#'
#' @section Computed variables specific to StatDensity3D:
#' - `density`: The kernel density estimate at each grid point
#' - `ndensity`: Density estimate scaled to maximum of 1 within each group
#' - `count`: Density estimate × number of observations in group (expected count)
#' - `n`: Number of observations in each group
#'
#' @section Grouping:
#' When aesthetics like `colour` or `fill` are mapped to categorical variables,
#' `stat_density_3d()` computes separate density surfaces for each group, just
#' like `stat_density_2d()`. Each group gets its own density calculation with
#' proper `count` and `n` values.
#'
#' @examples
#' # Basic density surface from scattered points
#' p <- ggplot(faithful, aes(eruptions, waiting)) +
#'   coord_3d() +
#'   scale_fill_viridis_c()
#' p + geom_density_3d() + guides(fill = guide_colorbar_3d())
#'
#' \donttest{
#'
#' # Specify alternative grid geometry and light model
#' p + geom_density_3d(grid = "equilateral", n = 30, direction = "y",
#'                     light = light("direct"),
#'                     color = "white", linewidth = .1) +
#'   guides(fill = guide_colorbar_3d())
#'
#' # Color by alternative density metric
#' p + geom_density_3d(aes(fill = after_stat(count)))
#'
#' # Adjust bandwidth for smoother or more detailed surfaces
#' p + geom_density_3d(adjust = 0.5, n = 100, color = "white")  # More detail
#' p + geom_density_3d(adjust = 2, color = "white")   # Smoother
#'
#' # As contour plot instead of default surface plot
#' p + stat_density_3d(geom = "contour_3d", light = "none",
#'                     color = "black", bins = 25,
#'                     sort_method = "pairwise")
#'
#' # Multiple density surfaces by group,
#' # using normalized density to equalize peak heights
#' ggplot(iris, aes(Petal.Length, Sepal.Length, fill = Species)) +
#'   geom_density_3d(aes(z = after_stat(ndensity), group = Species),
#'                   color = "black", alpha = .7, light = NULL) +
#'   coord_3d()
#'
#' # Same, but with extra padding to remove edge effects and
#' # with density filtering to remove rectangular artifacts
#' ggplot(iris, aes(Petal.Length, Sepal.Length, fill = Species)) +
#'   geom_density_3d(aes(z = after_stat(ndensity)),
#'                   pad = .3, min_ndensity = .001,
#'                   color = "black", alpha = .7, light = NULL) +
#'   coord_3d(ratio = c(3, 3, 1))
#'
#' }
#'
#' @seealso [ggplot2::stat_density_2d()] for 2D density contours, [stat_surface_3d()] for
#'   surfaces from existing grid data, [light()] for lighting specifications,
#'   [coord_3d()] for 3D coordinate systems.
#' @rdname stat_density_3d
#' @return A `Layer` object that can be added to a ggplot.
#' @export
geom_density_3d <- function(mapping = NULL, data = NULL,
                            stat = "density_3d",
                            position = "identity",
                            ...,
                            grid = "rectangle", n = 40, direction = "x", trim = TRUE,
                            h = NULL, adjust = 1,
                            pad = 0.1,
                            min_ndensity = 0,
                            light = NULL,
                            cull_backfaces = FALSE, sort_method = NULL,
                            force_convex = FALSE, scale_depth = TRUE,
                            na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomSurface3D,
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(n = n, grid = grid, direction = direction, trim = trim,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          h = h, adjust = adjust, pad = pad, min_ndensity = min_ndensity,
                          light = light, na.rm = na.rm, ...)
      )
}

#' @rdname stat_density_3d
#' @export
stat_density_3d <- function(mapping = NULL, data = NULL,
                            geom = "surface_3d",
                            position = "identity",
                            ...,
                            grid = "rectangle", n = 40, direction = "x", trim = TRUE,
                            h = NULL, adjust = 1,
                            pad = 0.1,
                            min_ndensity = 0,
                            light = NULL,
                            cull_backfaces = FALSE, sort_method = NULL,
                            force_convex = FALSE, scale_depth = TRUE,
                            na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = StatDensity3D, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(n = n, grid = grid, direction = direction, trim = trim,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          h = h, adjust = adjust, pad = pad, min_ndensity = min_ndensity,
                          light = light, na.rm = na.rm, ...)
      )
}

# Tick length for one axis, in points. ggplot2 separates the panel edge from
# axis text by the tick length plus the text margin, so the tick length is
# part of label clearance whether or not ticks are drawn.
#
# `axis.ticks.length.z` is not a registered element, so the axis-specific
# lookup is allowed to fail and fall back to the shared element. The final
# 2.75pt fallback is theme_grey()'s documented value (half_line / 2).
resolve_tick_length_points <- function(axis, theme) {
      element <- tryCatch(calc_element(paste0("axis.ticks.length.", axis), theme),
                          error = function(e) NULL)

      if (is.null(element)) {
            element <- tryCatch(calc_element("axis.ticks.length", theme),
                                error = function(e) NULL)
      }

      if (is.null(element) || !grid::is.unit(element)) return(2.75)

      value <- tryCatch(grid::convertUnit(element, "pt", valueOnly = TRUE),
                        error = function(e) NA_real_)

      if (length(value) != 1 || !is.finite(value)) return(2.75)

      value
}

extract_axis_theme_elements <- function(axis, theme) {
      axis_text_theme <- calc_element(paste0("axis.text.", axis), theme)
      axis_title_theme <- calc_element(paste0("axis.title.", axis), theme)
      parent_text_theme <- calc_element("axis.text", theme)
      parent_title_theme <- calc_element("axis.title", theme)
      default_theme <- theme_gray()
      default_axis_title <- calc_element(paste0("axis.title.", axis), default_theme)

      # Handle inheritance logic
      if (identical(axis_title_theme$margin, default_axis_title$margin)) {
            if (!is.null(parent_title_theme$margin)) {
                  axis_title_theme$margin <- parent_title_theme$margin
            }
      }

      axis_ticks_theme <- tryCatch(calc_element(paste0("axis.ticks.", axis), theme),
                                   error = function(e) NULL)
      if (is.null(axis_ticks_theme)) {
            axis_ticks_theme <- tryCatch(calc_element("axis.ticks", theme),
                                         error = function(e) NULL)
      }

      return(list(
            axis_text = axis_text_theme,
            axis_title = axis_title_theme,
            parent_text = parent_text_theme,
            axis_ticks = axis_ticks_theme,
            tick_length = resolve_tick_length_points(axis, theme)
      ))
}

# Resolve axis text and title margins, in points. Placement converts these to
# device space at draw time, so no plot-unit conversion happens here.
calculate_axis_offsets <- function(theme_elements, rotate_labels) {
      parent_margin <- theme_elements$parent_text$margin %||% margin()
      axis_margin <- theme_elements$axis_text$margin %||% margin()

      parent_margin_numeric <- as.numeric(parent_margin)
      axis_margin_numeric <- as.numeric(axis_margin)

      has_custom_parent_margin <- any(parent_margin_numeric != 0)

      if (has_custom_parent_margin) {
            non_zero_values <- axis_margin_numeric[axis_margin_numeric != 0]
            looks_like_default <- length(non_zero_values) <= 1 && all(non_zero_values <= 2.2)

            if (looks_like_default) {
                  margin_t <- parent_margin_numeric[1]
                  margin_r <- parent_margin_numeric[2]
                  margin_b <- parent_margin_numeric[3]
                  margin_l <- parent_margin_numeric[4]
            } else {
                  margin_t <- axis_margin_numeric[1]
                  margin_r <- axis_margin_numeric[2]
                  margin_b <- axis_margin_numeric[3]
                  margin_l <- axis_margin_numeric[4]
            }
      } else {
            margin_t <- axis_margin_numeric[1]
            margin_r <- axis_margin_numeric[2]
            margin_b <- axis_margin_numeric[3]
            margin_l <- axis_margin_numeric[4]
      }

      if (rotate_labels) {
            text_margin_points <- max(margin_l, margin_r)
      } else {
            text_margin_points <- max(margin_t, margin_r, margin_b, margin_l)
      }

      # Tick clearance is not added here. A tick is 3D geometry, so how far it
      # actually reaches from the edge is only known once projected; the drawn
      # extent is supplied by create_axis_ticks() at render time.
      #
      # ggplot2's 2.75pt margin assumes a vertical tick beside horizontal text.
      # Here labels hug the tick tip along the same line, so the same value
      # reads much tighter and needs a floor. Scaling with the text size keeps
      # spacing proportionate as base_size changes.
      text_size <- resolve_fontsize(theme_elements$axis_text$size, 8.5)
      text_offset_points <- max(text_margin_points, 0.6 * text_size)

      title_margin <- theme_elements$axis_title$margin %||% margin()
      title_margin_numeric <- as.numeric(title_margin)

      # ggplot2's 2.75pt title margin assumes a title in its own gtable band
      # below the axis text. Here the title sits diagonally beyond the labels
      # with no reserved strip, so it needs a floor, scaled with the title size
      # to stay proportionate as base_size changes.
      title_size <- resolve_fontsize(theme_elements$axis_title$size, 11)

      theme_margin_distance <- max(title_margin_numeric[1], title_margin_numeric[3])
      title_margin_points <- max(theme_margin_distance, 0.6 * title_size)

      return(list(
            text_offset = text_offset_points,
            title_margin = title_margin_points
      ))
}

# Device-space conversion --------------------------------------------------

# Placement geometry runs in points measured from the panel's lower-left
# corner. That space is isotropic, unlike npc, whose x and y units differ
# whenever the panel is not square.
make_device_context <- function(plot_bounds, panel_width_pt, panel_height_pt) {
      plot_width <- plot_bounds[2] - plot_bounds[1]
      plot_height <- plot_bounds[4] - plot_bounds[3]

      if (!is.finite(plot_width) || plot_width <= 0) plot_width <- 1
      if (!is.finite(plot_height) || plot_height <= 0) plot_height <- 1

      list(
            plot_bounds = plot_bounds,
            sx = panel_width_pt / plot_width,
            sy = panel_height_pt / plot_height,
            panel_width_pt = panel_width_pt,
            panel_height_pt = panel_height_pt
      )
}

plot_to_pt <- function(x, y, ctx) {
      list(x = (x - ctx$plot_bounds[1]) * ctx$sx,
           y = (y - ctx$plot_bounds[3]) * ctx$sy)
}

pt_to_npc <- function(x_pt, y_pt, ctx) {
      if (!is.finite(x_pt) || !is.finite(y_pt)) return(NULL)
      list(x = x_pt / ctx$panel_width_pt,
           y = y_pt / ctx$panel_height_pt)
}

# Convert a gridline data frame's projected coordinates into device points,
# leaving every other column untouched so downstream helpers work unchanged.
gridlines_to_pt <- function(gridlines, ctx) {
      converted <- plot_to_pt(gridlines$x, gridlines$y, ctx)
      gridlines$x <- converted$x
      gridlines$y <- converted$y
      gridlines
}

# Text measurement and placement -------------------------------------------

# Measure an unrotated label. Returns width and height in points. Must be
# called with an active device, i.e. from makeContent().
measure_text_box <- function(label, fontsize, family = "", face = "plain") {
      fallback <- list(
            width = 0.6 * fontsize * max(nchar(paste(deparse(label), collapse = "")), 1),
            height = fontsize
      )

      grob <- tryCatch({
            grid::textGrob(
                  label = as_grob_label(label),
                  gp = grid::gpar(fontsize = fontsize, fontfamily = family, fontface = face)
            )
      }, error = function(e) NULL)

      if (is.null(grob)) return(fallback)

      tryCatch({
            width <- grid::convertWidth(grid::grobWidth(grob), "pt", valueOnly = TRUE)
            height <- grid::convertHeight(grid::grobHeight(grob), "pt", valueOnly = TRUE)
            if (!is.finite(width) || width <= 0) width <- fallback$width
            if (!is.finite(height) || height <= 0) height <- fallback$height
            list(width = width, height = height)
      }, error = function(e) fallback)
}

# Corner offsets of a rotated text box relative to its justification point.
# grid rotates the box about that point, so corners are taken in the text's
# own frame and then rotated. Returns a 4x2 matrix of point offsets.
text_box_corner_offsets <- function(width, height, angle_degrees, hjust, vjust) {
      hjust <- if (is.null(hjust) || !is.finite(hjust)) 0.5 else hjust
      vjust <- if (is.null(vjust) || !is.finite(vjust)) 0.5 else vjust
      angle_degrees <- if (is.null(angle_degrees) || !is.finite(angle_degrees)) 0 else angle_degrees

      dx <- c(-hjust * width, (1 - hjust) * width)
      dy <- c(-vjust * height, (1 - vjust) * height)
      local <- as.matrix(expand.grid(dx = dx, dy = dy))

      angle <- angle_degrees * pi / 180
      cbind(
            local[, "dx"] * cos(angle) - local[, "dy"] * sin(angle),
            local[, "dx"] * sin(angle) + local[, "dy"] * cos(angle)
      )
}

# Unit normal to the axis edge, oriented to agree with the offset direction.
edge_normal <- function(axis_angle, direction) {
      u <- c(cos(axis_angle), sin(axis_angle))
      n <- c(-u[2], u[1])
      if (sum(n * direction) < 0) n <- -n
      n
}

# Distance to travel along `direction` so that every corner of the text box
# clears the axis edge by `margin`.
#
# The anchor lies on the edge, so a point P clears by (P - anchor) . normal.
# Solving min over corners of t * (direction . normal) + corner . normal >= margin
# gives the offset below. `min_cosine` caps the offset when the gridline runs
# nearly parallel to the edge and pushing outward gains almost no clearance.
#
# Pure function: no device or grid state, so it is testable with synthetic
# corner offsets.
place_axis_text <- function(direction, normal, margin, corner_offsets,
                            min_cosine = 0.15) {
      projections <- corner_offsets[, 1] * normal[1] + corner_offsets[, 2] * normal[2]
      inward_extent <- max(0, max(-projections))

      cosine <- sum(direction * normal)
      cosine <- max(cosine, min_cosine)

      (margin + inward_extent) / cosine
}

# How far a placed text box reaches from the edge, measured along the normal.
# Used to space titles beyond the labels they must clear.
text_box_reach <- function(offset, direction, normal, corner_offsets) {
      projections <- corner_offsets[, 1] * normal[1] + corner_offsets[, 2] * normal[2]
      offset * sum(direction * normal) + max(projections)
}

resolve_fontsize <- function(size, default) {
      if (is.null(size)) return(default)
      if (inherits(size, "unit")) size <- as.numeric(size)
      size <- suppressWarnings(as.numeric(size))
      if (length(size) != 1 || !is.finite(size) || size <= 0) return(default)
      size
}

# Geometry helpers ---------------------------------------------------------

# Helper function to calculate axis angle from gridlines
calculate_axis_angle <- function(axis_gridlines, axis_uses_start) {
      axis_gridlines_sorted <- axis_gridlines[order(axis_gridlines$break_value), ]
      first_gridline_group <- axis_gridlines[axis_gridlines$group == axis_gridlines_sorted$group[1], ]
      last_gridline_group <- axis_gridlines[axis_gridlines$group == axis_gridlines_sorted$group[nrow(axis_gridlines_sorted)], ]

      if (axis_uses_start) {
            axis_start_x <- first_gridline_group$x[1]
            axis_start_y <- first_gridline_group$y[1]
            axis_end_x <- last_gridline_group$x[1]
            axis_end_y <- last_gridline_group$y[1]
      } else {
            axis_start_x <- first_gridline_group$x[nrow(first_gridline_group)]
            axis_start_y <- first_gridline_group$y[nrow(first_gridline_group)]
            axis_end_x <- last_gridline_group$x[nrow(last_gridline_group)]
            axis_end_y <- last_gridline_group$y[nrow(last_gridline_group)]
      }

      axis_edge_dx <- axis_end_x - axis_start_x
      axis_edge_dy <- axis_end_y - axis_start_y
      axis_angle <- atan2(axis_edge_dy, axis_edge_dx)

      return(axis_angle)
}

# Helper function to calculate gridline position. Carries the endpoint's
# depth scaling factor so text can be sized by viewing distance.
calculate_gridline_position <- function(gridline_data, axis_uses_start) {
      index <- if (axis_uses_start) 1 else nrow(gridline_data)

      depth_scale <- 1
      if ("depth_scale" %in% names(gridline_data)) {
            candidate <- gridline_data$depth_scale[index]
            if (length(candidate) == 1 && is.finite(candidate) && candidate > 0) {
                  depth_scale <- candidate
            }
      }

      return(list(x = gridline_data$x[index],
                  y = gridline_data$y[index],
                  depth_scale = depth_scale))
}

# Helper function to calculate offset direction
calculate_offset_direction <- function(gridline_data, target_x, target_y) {
      gridline_center_x <- mean(gridline_data$x)
      gridline_center_y <- mean(gridline_data$y)

      center_to_target_dx <- target_x - gridline_center_x
      center_to_target_dy <- target_y - gridline_center_y
      center_to_target_length <- sqrt(center_to_target_dx^2 + center_to_target_dy^2)

      if (center_to_target_length == 0) {
            return(NULL)
      }

      return(list(
            dx = center_to_target_dx / center_to_target_length,
            dy = center_to_target_dy / center_to_target_length,
            center_x = gridline_center_x,
            center_y = gridline_center_y,
            length = center_to_target_length
      ))
}

# Helper function to calculate text rotation and justification
calculate_text_rotation_and_justification <- function(gridline_data, rotate_labels, theme_elements, is_title = FALSE, axis_angle = NULL) {
      if (is_title && !is.null(axis_angle)) {
            # For titles, use axis angle (parallel to axis edge)
            angle_radians <- axis_angle
      } else {
            # For labels, use gridline angle (parallel to gridline)
            gridline_dx <- gridline_data$x[nrow(gridline_data)] - gridline_data$x[1]
            gridline_dy <- gridline_data$y[nrow(gridline_data)] - gridline_data$y[1]
            angle_radians <- atan2(gridline_dy, gridline_dx)
      }

      angle_degrees <- angle_radians * 180 / pi

      # Ensure readable orientation
      if (abs(angle_degrees) > 90) {
            angle_degrees <- angle_degrees + 180
            if (angle_degrees > 180) angle_degrees <- angle_degrees - 360
      }

      if (rotate_labels) {
            result <- list(
                  angle = angle_degrees,
                  vjust = 0.5
            )
            # Add hjust calculation for labels (not titles)
            if (!is_title) {
                  result$hjust <- NULL  # Will be calculated in caller based on position
                  result$gridline_center_x <- mean(gridline_data$x)
                  result$gridline_center_y <- mean(gridline_data$y)
            } else {
                  element_type <- "axis_title"
                  result$hjust <- theme_elements[[element_type]]$hjust %||% 0.5
            }
            return(result)
      } else {
            element_type <- if (is_title) "axis_title" else "axis_text"
            return(list(
                  angle = theme_elements[[element_type]]$angle %||% 0,
                  hjust = theme_elements[[element_type]]$hjust %||% 0.5,
                  vjust = theme_elements[[element_type]]$vjust %||% 0.5
            ))
      }
}

# Anchor rotated text at its near end so it extends away from the cube.
resolve_end_anchor_hjust <- function(rotation_info, position_x, position_y) {
      offset_dx <- position_x - rotation_info$gridline_center_x
      offset_dy <- position_y - rotation_info$gridline_center_y
      angle_rad <- rotation_info$angle * pi / 180
      text_dir <- c(cos(angle_rad), sin(angle_rad))
      dot <- offset_dx * text_dir[1] + offset_dy * text_dir[2]
      if (dot > 0) 0 else 1
}

# Helper to normalise a single label into a value grid::textGrob accepts.
# Character and numeric labels pass through as strings; language objects
# (calls, names) and expression elements are preserved so plotmath renders.
# An expression vector of length >= 1 is reduced to its first element.
as_grob_label <- function(x) {
      if (is.expression(x)) {
            if (length(x) == 0) return("")
            return(x[[1]])
      }
      if (is.language(x)) {
            return(x)
      }
      as.character(x)
}

# Helper function to resolve label text. May return a character string or a
# language object (for plotmath labels); callers must not coerce the result.
resolve_label_text <- function(break_value, axis_labels, axis_breaks) {
      if (!is.null(axis_labels) && !is.null(axis_breaks)) {
            break_index <- match(break_value, axis_breaks)
            if (length(break_index) > 0 && !is.na(break_index) &&
                break_index <= length(axis_labels)) {
                  return(as_grob_label(axis_labels[[break_index]]))
            }
      }
      return(as.character(break_value))
}

# Helper function to scale coordinates to NPC
scale_to_npc_coordinates <- function(x, y = NULL, plot_bounds) {

      if(!is.null(y)){
            x_scaled <- (x - plot_bounds[1]) / (plot_bounds[2] - plot_bounds[1])
            y_scaled <- (y - plot_bounds[3]) / (plot_bounds[4] - plot_bounds[3])
            if (is.na(x_scaled) || is.na(y_scaled)) {
                  return(NULL)
            }
            return(list(x = x_scaled, y = y_scaled))
      }else{ # assume x is a data frame with x and y variables
            x$x <- (x$x - plot_bounds[1]) / (plot_bounds[2] - plot_bounds[1])
            x$y <- (x$y - plot_bounds[3]) / (plot_bounds[4] - plot_bounds[3])
            return(x)
      }
}

# Helper function to create text grob
create_text_grob <- function(text, x_npc, y_npc, rotation_info, theme_elements,
                             is_title = FALSE, fontsize = NULL) {
      element_type <- if (is_title) "axis_title" else "axis_text"

      # Get appropriate theme elements
      colour <- theme_elements[[element_type]]$colour %||% "black"
      fontfamily <- theme_elements[[element_type]]$family %||% ""
      fontface <- theme_elements[[element_type]]$face %||% "plain"

      if (is.null(fontsize)) {
            fontsize <- resolve_fontsize(theme_elements[[element_type]]$size,
                                         if (is_title) 11 else 8.5)
      }

      tryCatch({
            grid::textGrob(
                  label = as_grob_label(text),
                  x = as.numeric(x_npc),
                  y = as.numeric(y_npc),
                  hjust = as.numeric(rotation_info$hjust),
                  vjust = as.numeric(rotation_info$vjust),
                  rot = as.numeric(rotation_info$angle),
                  default.units = "npc",
                  gp = grid::gpar(
                        fontsize = fontsize,
                        col = colour,
                        fontfamily = fontfamily,
                        fontface = fontface
                  )
            )
      }, error = function(e) {
            NULL
      })
}

# Axis ticks ---------------------------------------------------------------

# Projected length, in points, of a one-unit vector along each cube axis,
# measured at the cube centre. Used to convert a tick length in points into
# cube space so that ticks are built as ordinary 3D geometry.
projected_axis_lengths_pt <- function(proj, ctx) {
      vapply(1:3, function(i) {
            coords <- matrix(0, nrow = 2, ncol = 3)
            coords[1, i] <- -0.5
            coords[2, i] <- 0.5

            projected <- transform_3d_standard(
                  data.frame(x = coords[, 1], y = coords[, 2], z = coords[, 3]), proj)

            start <- plot_to_pt(projected$x[1], projected$y[1], ctx)
            end <- plot_to_pt(projected$x[2], projected$y[2], ctx)

            sqrt((end$x - start$x)^2 + (end$y - start$y)^2)
      }, numeric(1))
}

# Tick length in cube units.
#
# A tick is 3D geometry, so its length lives in cube space and foreshortens
# with the view like everything else. Calibrating against the longest
# projected axis means a tick pointing along the least foreshortened axis
# renders at the requested point size and every other tick comes in shorter,
# so `axis.ticks.length` reads as an upper bound.
tick_length_cube_units <- function(tick_length_pt, proj, ctx) {
      axis_lengths <- projected_axis_lengths_pt(proj, ctx)
      reference <- max(axis_lengths)

      if (!is.finite(reference) || reference <= 0) return(NA_real_)

      tick_length_pt / reference
}

# Build the tick marks for one axis.
#
# Ticks continue their gridlines past the cube edge: the direction is the
# face's free axis, signed outward by the edge's fixed coordinate. Geometry is
# assembled in cube space and then projected, so ticks share the perspective
# and depth behaviour of the gridlines they extend.
create_axis_ticks <- function(axis, standard_gridlines, theme_elements,
                              panel_params, ctx, chosen_edge, chosen_face,
                              effective_ratios, axis_angle, on_hull = TRUE,
                              ticks_depth_strength = 1) {

      empty <- list(ticks = list(), tick_reach = 0)

      # Geometry is computed even when ticks are not drawn, so that blanking
      # axis.ticks leaves label spacing unchanged, as it does in 2D ggplot2.
      element <- theme_elements$axis_ticks
      draw <- !is.null(element) && !inherits(element, "element_blank")

      length_cube <- tick_length_cube_units(theme_elements$tick_length %||% 0,
                                            panel_params$proj, ctx)
      if (!is.finite(length_cube) || length_cube == 0) return(empty)

      normal_axis <- substr(chosen_face, 1, 1)
      free_axis <- setdiff(c("x", "y", "z"), c(axis, normal_axis))
      if (length(free_axis) != 1) return(empty)

      fixed <- chosen_edge$fixed_coords[[free_axis]]
      if (is.null(fixed) || !is.finite(fixed)) return(empty)

      # Outward from the face, flipping with the labels when the edge is
      # interior to the rendered silhouette so the two stay on the same side.
      direction <- sign(fixed)
      if (direction == 0) return(empty)
      if (!on_hull) direction <- -direction

      free_edge_coord <- fixed * effective_ratios[match(free_axis, c("x", "y", "z"))]
      free_tip_coord <- free_edge_coord + direction * length_cube

      groups <- unique(standard_gridlines$group)
      if (length(groups) == 0) return(empty)

      starts <- data.frame(x = numeric(0), y = numeric(0), z = numeric(0))
      ends <- starts

      for (group_id in groups) {
            rows <- standard_gridlines[standard_gridlines$group == group_id, ]
            if (nrow(rows) == 0) next

            base <- rows[1, c("x", "y", "z"), drop = FALSE]

            start <- base
            start[[free_axis]] <- free_edge_coord
            end <- base
            end[[free_axis]] <- free_tip_coord

            starts <- rbind(starts, start)
            ends <- rbind(ends, end)
      }

      if (nrow(starts) == 0) return(empty)

      starts_2d <- transform_3d_standard(starts, panel_params$proj)
      ends_2d <- transform_3d_standard(ends, panel_params$proj)

      # How far the drawn ticks reach from the edge, perpendicular to it, in
      # points. This is the quantity label clearance must be measured against:
      # the tick runs along its gridline, which is generally oblique to the
      # edge, so its perpendicular reach is shorter than its drawn length.
      start_pt_x <- (starts_2d$x - ctx$plot_bounds[1]) * ctx$sx
      start_pt_y <- (starts_2d$y - ctx$plot_bounds[3]) * ctx$sy
      end_pt_x <- (ends_2d$x - ctx$plot_bounds[1]) * ctx$sx
      end_pt_y <- (ends_2d$y - ctx$plot_bounds[3]) * ctx$sy

      tick_dx <- end_pt_x - start_pt_x
      tick_dy <- end_pt_y - start_pt_y

      tick_reach <- 0
      lengths <- sqrt(tick_dx^2 + tick_dy^2)
      longest <- which.max(replace(lengths, !is.finite(lengths), 0))

      if (length(longest) == 1 && is.finite(lengths[longest]) && lengths[longest] > 0) {
            unit_dir <- c(tick_dx[longest], tick_dy[longest]) / lengths[longest]
            normal <- edge_normal(axis_angle, unit_dir)
            reaches <- tick_dx * normal[1] + tick_dy * normal[2]
            reaches <- reaches[is.finite(reaches)]
            if (length(reaches) > 0) tick_reach <- max(0, max(reaches))
      }

      if (!draw) return(list(ticks = list(), tick_reach = tick_reach))

      # scale_to_npc_coordinates() guards with scalar `||`, so npc is computed
      # directly here rather than element by element.
      pb <- ctx$plot_bounds
      x0 <- (starts_2d$x - pb[1]) / (pb[2] - pb[1])
      y0 <- (starts_2d$y - pb[3]) / (pb[4] - pb[3])
      x1 <- (ends_2d$x - pb[1]) / (pb[2] - pb[1])
      y1 <- (ends_2d$y - pb[3]) / (pb[4] - pb[3])

      keep <- is.finite(x0) & is.finite(y0) & is.finite(x1) & is.finite(y1)
      if (!any(keep)) return(list(ticks = list(), tick_reach = tick_reach))

      # Depth taken from the tick itself rather than its parent gridline,
      # since a tick sits at one end rather than spanning the face.
      depth <- (starts_2d$depth_scale + ends_2d$depth_scale) / 2
      depth[!is.finite(depth) | depth <= 0] <- 1

      base_lwd <- (element$linewidth %||% 0.5) * .pt

      tick_grob <- tryCatch({
            grid::segmentsGrob(
                  x0 = x0[keep], y0 = y0[keep],
                  x1 = x1[keep], y1 = y1[keep],
                  default.units = "npc",
                  gp = grid::gpar(
                        col = element$colour %||% "grey20",
                        lwd = safe_lwd(base_lwd * apply_depth_strength(depth[keep],
                                                                       ticks_depth_strength),
                                       base_lwd),
                        lty = element$linetype %||% 1,
                        lineend = element$lineend %||% "butt"
                  ),
                  name = paste0("axis.ticks.", axis, ".3d")
            )
      }, error = function(e) NULL)

      if (is.null(tick_grob)) return(list(ticks = list(), tick_reach = tick_reach))

      list(ticks = list(tick_grob), tick_reach = tick_reach)
}


# Label and title construction ---------------------------------------------

# Build the tick labels for one axis. `edge_gridlines` must already be in
# device points. Returns the grobs plus the furthest distance any label
# reaches from the axis edge, which titles use for spacing.
create_axis_labels <- function(axis, edge_gridlines, theme_elements, offsets,
                               panel_params, rotate_labels, ctx, chosen_edge,
                               axis_uses_start, axis_angle, on_hull = TRUE,
                               tick_reach = 0, text_depth_strength = 1) {

      margin <- max(tick_reach, 0) + offsets$text_offset

      axis_labels <- panel_params$scale_info[[axis]]$labels %||% NULL
      axis_breaks <- panel_params$scale_info[[axis]]$breaks %||% NULL

      base_fontsize <- resolve_fontsize(theme_elements$axis_text$size, 8.5)
      fontfamily <- theme_elements$axis_text$family %||% ""
      fontface <- theme_elements$axis_text$face %||% "plain"

      all_labels <- list()
      label_reach <- 0

      for (group_id in unique(edge_gridlines$group)) {
            gridline_data <- edge_gridlines[edge_gridlines$group == group_id, ]
            if (nrow(gridline_data) < 2) next

            base_pos <- calculate_gridline_position(gridline_data, axis_uses_start)
            offset_dir <- calculate_offset_direction(gridline_data, base_pos$x, base_pos$y)
            if (is.null(offset_dir)) next

            direction <- c(offset_dir$dx, offset_dir$dy)
            if (!on_hull) direction <- -direction
            normal <- edge_normal(axis_angle, direction)

            rotation_info <- calculate_text_rotation_and_justification(
                  gridline_data, rotate_labels, theme_elements,
                  is_title = FALSE, axis_angle = NULL)

            if (rotate_labels && is.null(rotation_info$hjust)) {
                  rotation_info$hjust <- resolve_end_anchor_hjust(rotation_info,
                                                                  base_pos$x, base_pos$y)
            }

            label_text <- resolve_label_text(gridline_data$break_value[1],
                                             axis_labels, axis_breaks)

            fontsize <- base_fontsize * apply_depth_strength(base_pos$depth_scale,
                                                             text_depth_strength)
            if (!is.finite(fontsize) || fontsize <= 0) fontsize <- base_fontsize

            box <- measure_text_box(label_text, fontsize, fontfamily, fontface)
            corners <- text_box_corner_offsets(box$width, box$height,
                                               rotation_info$angle,
                                               rotation_info$hjust,
                                               rotation_info$vjust)

            offset <- place_axis_text(direction, normal, margin, corners)
            label_reach <- max(label_reach,
                               text_box_reach(offset, direction, normal, corners))

            position_npc <- pt_to_npc(base_pos$x + direction[1] * offset,
                                      base_pos$y + direction[2] * offset,
                                      ctx)
            if (is.null(position_npc)) next

            label_grob <- create_text_grob(label_text, position_npc$x, position_npc$y,
                                           rotation_info, theme_elements,
                                           is_title = FALSE, fontsize = fontsize)

            if (!is.null(label_grob)) {
                  all_labels[[length(all_labels) + 1]] <- label_grob
            }
      }

      return(list(labels = all_labels,
                  label_reach = label_reach))
}

# Build the title for one axis. `edge_gridlines` must already be in device
# points. `label_reach` is how far the axis labels extend from the edge.
create_axis_title <- function(axis, edge_gridlines, theme_elements, offsets,
                              panel_params, rotate_labels, ctx, chosen_edge,
                              axis_uses_start, on_hull = TRUE, axis_selection = NULL,
                              title_position = "auto", label_reach = 0) {

      axis_name <- panel_params$scale_info[[axis]]$name

      if (is.null(axis_name) || inherits(axis_name, "waiver")) {
            axis_name <- axis
      }

      if (is.null(axis_name) || (is.character(axis_name) && axis_name == "")) {
            return(list())
      }

      axis_gridlines <- edge_gridlines[edge_gridlines$break_axis == axis, ]

      if (nrow(axis_gridlines) == 0) {
            return(list())
      }

      title_fontsize <- resolve_fontsize(theme_elements$axis_title$size, 11)
      title_family <- theme_elements$axis_title$family %||% ""
      title_face <- theme_elements$axis_title$face %||% "plain"

      if (!on_hull && title_position != "center" && !is.null(axis_selection)) {
            # Place title at the near (peripheral) end of the axis edge
            p1_depth <- axis_selection$edge_p1_2d$depth
            p2_depth <- axis_selection$edge_p2_2d$depth
            if (p1_depth <= p2_depth) {
                  title_pos <- axis_selection$edge_p1_2d
            } else {
                  title_pos <- axis_selection$edge_p2_2d
            }

            title_pos_pt <- plot_to_pt(title_pos$x, title_pos$y, ctx)

            cube_center_2d <- transform_3d_standard(data.frame(x = 0, y = 0, z = 0),
                                                    panel_params$proj)
            cube_center_pt <- plot_to_pt(cube_center_2d$x, cube_center_2d$y, ctx)

            offset_dx <- title_pos_pt$x - cube_center_pt$x
            offset_dy <- title_pos_pt$y - cube_center_pt$y
            offset_len <- sqrt(offset_dx^2 + offset_dy^2)
            if (offset_len > 0) {
                  offset_dx <- offset_dx / offset_len
                  offset_dy <- offset_dy / offset_len
            }

            title_offset <- offsets$title_margin
            final_x <- title_pos_pt$x + offset_dx * title_offset
            final_y <- title_pos_pt$y + offset_dy * title_offset

            position_npc <- pt_to_npc(final_x, final_y, ctx)
            if (is.null(position_npc)) return(list())

            edge_p1_pt <- plot_to_pt(axis_selection$edge_p1_2d$x,
                                     axis_selection$edge_p1_2d$y, ctx)
            edge_p2_pt <- plot_to_pt(axis_selection$edge_p2_2d$x,
                                     axis_selection$edge_p2_2d$y, ctx)
            axis_angle <- atan2(edge_p2_pt$y - edge_p1_pt$y,
                                edge_p2_pt$x - edge_p1_pt$x)

            rotation_info <- calculate_text_rotation_and_justification(
                  edge_gridlines[edge_gridlines$group == edge_gridlines$group[1], ],
                  rotate_labels, theme_elements, is_title = TRUE, axis_angle = axis_angle
            )

            # hjust: anchor text so it extends away from the plot
            angle_rad <- rotation_info$angle * pi / 180
            text_dir <- c(cos(angle_rad), sin(angle_rad))
            dot <- offset_dx * text_dir[1] + offset_dy * text_dir[2]
            rotation_info$hjust <- if (dot > 0) 0 else 1

            title_grob <- create_text_grob(axis_name, position_npc$x, position_npc$y,
                                           rotation_info, theme_elements,
                                           is_title = TRUE, fontsize = title_fontsize)
            if (!is.null(title_grob)) return(list(title_grob))
            return(list())
      }

      # Find the gridline closest to the center of the axis range
      axis_breaks <- panel_params$scale_info[[axis]]$breaks
      if(!is.numeric(axis_breaks)) axis_breaks <- attr(axis_breaks, "pos")
      axis_center_value <- mean(range(axis_breaks))

      center_distances <- abs(axis_gridlines$break_pos - axis_center_value)
      center_group <- axis_gridlines$group[which.min(center_distances)]
      center_gridline <- axis_gridlines[axis_gridlines$group == center_group, ]

      if (nrow(center_gridline) < 2) {
            return(list())
      }

      axis_angle <- calculate_axis_angle(axis_gridlines, axis_uses_start)

      base_pos <- calculate_gridline_position(center_gridline, axis_uses_start)
      offset_dir <- calculate_offset_direction(center_gridline, base_pos$x, base_pos$y)
      if (is.null(offset_dir)) return(list())

      direction <- c(offset_dir$dx, offset_dir$dy)
      if (!on_hull) direction <- -direction
      normal <- edge_normal(axis_angle, direction)

      rotation_info <- calculate_text_rotation_and_justification(
            center_gridline, rotate_labels, theme_elements,
            is_title = TRUE, axis_angle = axis_angle)

      box <- measure_text_box(axis_name, title_fontsize, title_family, title_face)
      corners <- text_box_corner_offsets(box$width, box$height,
                                         rotation_info$angle,
                                         rotation_info$hjust,
                                         rotation_info$vjust)

      # label_reach already accounts for the ticks, since label clearance is
      # measured from the tick tip. When labels are not drawn, the caller
      # passes the tick reach through in its place.
      margin <- max(label_reach, 0) + offsets$title_margin
      offset <- place_axis_text(direction, normal, margin, corners)

      position_npc <- pt_to_npc(base_pos$x + direction[1] * offset,
                                base_pos$y + direction[2] * offset,
                                ctx)
      if (is.null(position_npc)) {
            return(list())
      }

      title_grob <- create_text_grob(axis_name, position_npc$x, position_npc$y,
                                     rotation_info, theme_elements,
                                     is_title = TRUE, fontsize = title_fontsize)

      if (!is.null(title_grob)) {
            return(list(title_grob))
      } else {
            return(list())
      }
}

# Draw-time assembly -------------------------------------------------------

# Build every axis label and title for the panel. Runs at draw time, so
# `panel_width_pt` and `panel_height_pt` are the panel's true dimensions and
# all text can be measured directly.
render_axis_text <- function(self, panel_params, theme, panel_width_pt, panel_height_pt) {
      tryCatch({
            all_labels <- list()
            all_titles <- list()
            all_ticks <- list()

            theme_elements <- panel_params$theme_elements %||% list()
            should_render_axis_text <- theme_elements$show_axis_text %||% !inherits(calc_element("axis.text", theme), "element_blank")
            should_render_axis_title <- theme_elements$show_axis_title %||% !inherits(calc_element("axis.title", theme), "element_blank")
            should_render_axis_ticks <- theme_elements$show_axis_ticks %||% !inherits(calc_element("axis.ticks", theme), "element_blank")

            if (!should_render_axis_text && !should_render_axis_title && !should_render_axis_ticks) {
                  return(list(labels = list(), titles = list(), ticks = list()))
            }

            ctx <- make_device_context(panel_params$plot_bounds,
                                       panel_width_pt, panel_height_pt)

            effective_ratios <- compute_effective_ratios(
                  list(x = panel_params$scale_info$x$limits,
                       y = panel_params$scale_info$y$limits,
                       z = panel_params$scale_info$z$limits),
                  panel_params$scales,
                  panel_params$ratio
            )

            for (axis in c("x", "y", "z")) {

                  axis_selection <- get_axis_selection(axis, self, panel_params, effective_ratios)

                  if (is.null(axis_selection)) {
                        next
                  }

                  chosen_edge <- axis_selection$edge
                  chosen_face <- axis_selection$face

                  theme_elements_axis <- extract_axis_theme_elements(axis, theme)

                  offsets <- calculate_axis_offsets(theme_elements_axis, self$rotate_labels)

                  edge_gridlines <- panel_params$grid_transformed[
                        panel_params$grid_transformed$face == chosen_face &
                              panel_params$grid_transformed$break_axis == axis, ]

                  if (nrow(edge_gridlines) == 0) next

                  edge_gridlines <- gridlines_to_pt(edge_gridlines, ctx)

                  # Placement direction depends on externality relative to the
                  # rendered panels, not the full cube silhouette.
                  on_hull <- axis_selection$on_panel_hull %||% TRUE
                  label_reach <- 0
                  tick_reach <- 0

                  axis_uses_start <- determine_endpoint_preference_by_boundary(chosen_edge, edge_gridlines)
                  axis_angle <- calculate_axis_angle(edge_gridlines, axis_uses_start)

                  # Ticks first: label clearance is measured from the tick tip,
                  # and the drawn extent is only known once projected.
                  if (!is.null(panel_params$grid_standard)) {
                        standard_gridlines <- panel_params$grid_standard[
                              panel_params$grid_standard$face == chosen_face &
                                    panel_params$grid_standard$break_axis == axis, ]

                        if (nrow(standard_gridlines) > 0) {
                              tick_result <- create_axis_ticks(
                                    axis, standard_gridlines, theme_elements_axis,
                                    panel_params, ctx, chosen_edge, chosen_face,
                                    effective_ratios, axis_angle, on_hull,
                                    depth_strength(self, "ticks"))

                              tick_reach <- tick_result$tick_reach
                              if (should_render_axis_ticks) {
                                    all_ticks <- c(all_ticks, tick_result$ticks)
                              }
                        }
                  }

                  if (should_render_axis_text) {
                        label_result <- create_axis_labels(axis, edge_gridlines, theme_elements_axis,
                                                           offsets, panel_params, self$rotate_labels,
                                                           ctx, chosen_edge, axis_uses_start,
                                                           axis_angle, on_hull, tick_reach,
                                                           depth_strength(self, "text"))
                        all_labels <- c(all_labels, label_result$labels)
                        label_reach <- label_result$label_reach
                  } else {
                        label_reach <- max(tick_reach, 0)
                  }

                  if (should_render_axis_title) {
                        title_result <- create_axis_title(axis, edge_gridlines, theme_elements_axis,
                                                          offsets, panel_params, self$rotate_labels,
                                                          ctx, chosen_edge, axis_uses_start,
                                                          on_hull, axis_selection,
                                                          self$title_position %||% "auto",
                                                          label_reach)
                        all_titles <- c(all_titles, title_result)
                  }
            }

            return(list(labels = all_labels, titles = all_titles, ticks = all_ticks))
      }, error = function(e) {
            warning("Axis label/title rendering failed: ", e$message)
            return(list(labels = list(), titles = list(), ticks = list()))
      })
}

# Deferred grob holding everything needed to build axis furniture. Resolution
# happens in makeContent(), inside the panel viewport, where the panel's real
# dimensions and text metrics are available.
axis_furniture_grob <- function(self, panel_params, theme, show_text, show_title,
                                show_ticks = TRUE, name = "axis.furniture.3d") {
      grid::gTree(
            coord = self,
            panel_params = panel_params,
            plot_theme = theme,
            show_text = show_text,
            show_title = show_title,
            show_ticks = show_ticks,
            name = name,
            cl = "ggcube_axis_furniture"
      )
}

#' Resolve 3D axis labels and titles at draw time
#'
#' @param x A `ggcube_axis_furniture` gTree.
#' @return The gTree with its children populated.
#' @importFrom grid makeContent
#' @export
#' @keywords internal
makeContent.ggcube_axis_furniture <- function(x) {
      children <- tryCatch({
            panel_width_pt <- grid::convertWidth(grid::unit(1, "npc"), "pt", valueOnly = TRUE)
            panel_height_pt <- grid::convertHeight(grid::unit(1, "npc"), "pt", valueOnly = TRUE)

            elements <- render_axis_text(x$coord, x$panel_params, x$plot_theme,
                                         panel_width_pt, panel_height_pt)

            kids <- list()

            if (isTRUE(x$show_ticks)) {
                  ticks <- elements$ticks[!vapply(elements$ticks, is.null, logical(1))]
                  if (length(ticks) > 0) {
                        kids[[length(kids) + 1]] <- do.call(
                              grid::grobTree, c(list(name = "axis.ticks.3d"), ticks))
                  }
            }

            if (x$show_text) {
                  labels <- elements$labels[!vapply(elements$labels, is.null, logical(1))]
                  if (length(labels) > 0) {
                        kids[[length(kids) + 1]] <- do.call(
                              grid::grobTree, c(list(name = "axis.labels.3d"), labels))
                  }
            }

            if (x$show_title) {
                  titles <- elements$titles[!vapply(elements$titles, is.null, logical(1))]
                  if (length(titles) > 0) {
                        kids[[length(kids) + 1]] <- do.call(
                              grid::grobTree, c(list(name = "axis.titles.3d"), titles))
                  }
            }

            kids
      }, error = function(e) {
            warning("Axis furniture rendering failed: ", e$message)
            list()
      })

      if (length(children) == 0) {
            return(grid::setChildren(x, grid::gList()))
      }

      grid::setChildren(x, do.call(grid::gList, children))
}

test_that("as_grob_label passes character and numeric through as strings", {
      expect_identical(as_grob_label("foo"), "foo")
      expect_identical(as_grob_label(5), "5")
      expect_identical(as_grob_label(NA_character_), NA_character_)
})

test_that("as_grob_label preserves language objects", {
      lab <- as_grob_label(quote(alpha))
      expect_true(is.language(lab))
      expect_false(is.expression(lab))
      expect_identical(lab, quote(alpha))

      call_lab <- as_grob_label(quote(beta[1]))
      expect_true(is.language(call_lab))
      expect_false(is.expression(call_lab))
})

test_that("as_grob_label reduces an expression vector to its first element", {
      lab <- as_grob_label(expression(alpha, beta, gamma))
      # Should be the bare first element, not the whole vector
      expect_true(is.language(lab))
      expect_false(is.expression(lab))
      expect_identical(lab, quote(alpha))
})

test_that("as_grob_label handles a single-element expression", {
      lab <- as_grob_label(expression(alpha))
      expect_true(is.language(lab))
      expect_false(is.expression(lab))
      expect_identical(lab, quote(alpha))
})

test_that("as_grob_label returns empty string for an empty expression", {
      expect_identical(as_grob_label(expression()), "")
})

test_that("resolve_label_text falls back to break value as character", {
      # No custom labels supplied
      expect_identical(resolve_label_text(2, NULL, NULL), "2")

      # Break not present among axis_breaks: fall back to the value itself
      expect_identical(
            resolve_label_text(99, c("a", "b"), c(1, 2)),
            "99"
      )
})

test_that("resolve_label_text returns character labels for character vectors", {
      out <- resolve_label_text(2, c("Low", "Med", "High"), c(1, 2, 3))
      expect_identical(out, "Med")
})

test_that("resolve_label_text returns the matched language element for plotmath", {
      labels <- expression(alpha, beta, gamma)
      breaks <- c(1, 2, 3)

      out2 <- resolve_label_text(2, labels, breaks)
      expect_true(is.language(out2))
      expect_false(is.expression(out2))
      expect_identical(out2, quote(beta))

      out3 <- resolve_label_text(3, labels, breaks)
      expect_identical(out3, quote(gamma))
})

test_that("resolve_label_text handles a label vector mixing math and plain text", {
      labels <- expression(alpha, "beta", gamma^2)
      breaks <- c(1, 2, 3)

      # Math element stays language
      expect_true(is.language(resolve_label_text(1, labels, breaks)))
      # The plain-text element is a character constant inside the expression;
      # indexing yields a length-1 character string
      expect_identical(resolve_label_text(2, labels, breaks), "beta")
      # Superscript element stays language
      expect_true(is.language(resolve_label_text(3, labels, breaks)))
})

test_that("resolve_label_text does not error when break has no match", {
      labels <- expression(alpha, beta)
      breaks <- c(1, 2)
      # match() returns NA; must fall back without indexing error
      expect_identical(resolve_label_text(5, labels, breaks), "5")
})

test_that("create_text_grob builds a grob for plain, language, and expression labels", {
      rot <- list(hjust = 0.5, vjust = 0.5, angle = 0)
      theme_elements <- list(axis_text = list())

      g_plain <- create_text_grob("foo", 0.5, 0.5, rot, theme_elements)
      expect_s3_class(g_plain, "text")

      g_lang <- create_text_grob(quote(alpha), 0.5, 0.5, rot, theme_elements)
      expect_s3_class(g_lang, "text")

      g_expr <- create_text_grob(expression(beta[1]), 0.5, 0.5, rot, theme_elements)
      expect_s3_class(g_expr, "text")
})

test_that("create_text_grob carries a language label into the grob unchanged", {
      rot <- list(hjust = 0.5, vjust = 0.5, angle = 0)
      g <- create_text_grob(quote(beta[1]), 0.5, 0.5, rot, list(axis_title = list()),
                            is_title = TRUE)
      # The grob's label should remain a language object, not a string
      expect_true(is.language(g$label))
})

# Text box geometry and placement -------------------------------------------

test_that("text_box_corner_offsets returns the four corners about the anchor", {
      corners <- text_box_corner_offsets(width = 10, height = 4,
                                         angle_degrees = 0,
                                         hjust = 0.5, vjust = 0.5)

      expect_equal(nrow(corners), 4)
      expect_equal(ncol(corners), 2)
      expect_equal(sort(unique(corners[, 1])), c(-5, 5))
      expect_equal(sort(unique(corners[, 2])), c(-2, 2))
})

test_that("text_box_corner_offsets honours justification", {
      # hjust = 0 anchors the left edge, so the box extends only rightward
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0, vjust = 0.5)
      expect_equal(min(corners[, 1]), 0)
      expect_equal(max(corners[, 1]), 10)

      # hjust = 1 anchors the right edge
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 1, vjust = 0.5)
      expect_equal(min(corners[, 1]), -10)
      expect_equal(max(corners[, 1]), 0)
})

test_that("text_box_corner_offsets rotates the box rather than its bounding box", {
      corners <- text_box_corner_offsets(10, 4, angle_degrees = 90,
                                         hjust = 0.5, vjust = 0.5)

      # A 90 degree rotation swaps the extents exactly; an axis-aligned
      # bounding box would not shrink the long dimension
      expect_equal(sort(unique(round(corners[, 1], 10))), c(-2, 2))
      expect_equal(sort(unique(round(corners[, 2], 10))), c(-5, 5))
})

test_that("text_box_corner_offsets falls back to sane values for bad input", {
      corners <- text_box_corner_offsets(10, 4, angle_degrees = NA,
                                         hjust = NULL, vjust = NA)
      expect_equal(nrow(corners), 4)
      expect_true(all(is.finite(corners)))
})

test_that("place_axis_text clears a centred box by margin plus half its extent", {
      # Offsetting straight along the edge normal, box 10 wide and 4 tall
      direction <- c(0, 1)
      normal <- c(0, 1)
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0.5, vjust = 0.5)

      # Nearest corner sits 2 below the anchor, so offset = margin + 2
      expect_equal(place_axis_text(direction, normal, margin = 5, corners), 7)
})

test_that("place_axis_text needs only the margin when the box extends outward", {
      # vjust = 0 anchors the bottom edge, so nothing lies below the anchor
      direction <- c(0, 1)
      normal <- c(0, 1)
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0.5, vjust = 0)

      expect_equal(place_axis_text(direction, normal, margin = 5, corners), 5)
})

test_that("place_axis_text divides by the direction-normal cosine", {
      # Direction 45 degrees off the normal: travel is margin / cos(45)
      direction <- c(sin(pi / 4), cos(pi / 4))
      normal <- c(0, 1)
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0.5, vjust = 0)

      expect_equal(place_axis_text(direction, normal, margin = 5, corners),
                   5 / cos(pi / 4))
})

test_that("place_axis_text caps the offset for near-parallel directions", {
      # Direction almost along the edge: without the clamp this diverges
      direction <- c(1, 1e-6)
      normal <- c(0, 1)
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0.5, vjust = 0)

      offset <- place_axis_text(direction, normal, margin = 5, corners,
                                min_cosine = 0.15)
      expect_equal(offset, 5 / 0.15)
      expect_true(is.finite(offset))
})

test_that("place_axis_text always returns at least the margin", {
      normal <- c(0, 1)
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0.5, vjust = 0)

      for (angle in seq(0, 80, by = 10)) {
            rad <- angle * pi / 180
            direction <- c(sin(rad), cos(rad))
            expect_gte(place_axis_text(direction, normal, margin = 5, corners), 5)
      }
})

test_that("text_box_reach measures the far edge of a placed box", {
      direction <- c(0, 1)
      normal <- c(0, 1)
      corners <- text_box_corner_offsets(10, 4, 0, hjust = 0.5, vjust = 0.5)

      offset <- place_axis_text(direction, normal, margin = 5, corners)
      # Box centre lands at 7, far edge is 2 beyond that
      expect_equal(text_box_reach(offset, direction, normal, corners), 9)
})

test_that("edge_normal is a unit vector agreeing with the offset direction", {
      n <- edge_normal(axis_angle = 0, direction = c(0, 1))
      expect_equal(sqrt(sum(n^2)), 1)
      expect_gt(sum(n * c(0, 1)), 0)

      # Flipping the offset direction flips the normal
      n_flipped <- edge_normal(axis_angle = 0, direction = c(0, -1))
      expect_equal(n_flipped, -n)
})

test_that("edge_normal is perpendicular to the edge for arbitrary angles", {
      for (angle in seq(0, 2 * pi, length.out = 9)) {
            u <- c(cos(angle), sin(angle))
            n <- edge_normal(angle, direction = c(-u[2], u[1]))
            expect_equal(sum(u * n), 0)
            expect_equal(sqrt(sum(n^2)), 1)
      }
})


# Device-space conversion ---------------------------------------------------

test_that("make_device_context round-trips through plot_to_pt and pt_to_npc", {
      ctx <- make_device_context(plot_bounds = c(-1, 1, -2, 2),
                                 panel_width_pt = 400,
                                 panel_height_pt = 800)

      pt <- plot_to_pt(0, 0, ctx)
      # Centre of the bounds sits at the centre of the panel
      expect_equal(pt$x, 200)
      expect_equal(pt$y, 400)

      npc <- pt_to_npc(pt$x, pt$y, ctx)
      expect_equal(npc$x, 0.5)
      expect_equal(npc$y, 0.5)
})

test_that("make_device_context guards degenerate plot bounds", {
      ctx <- make_device_context(c(0, 0, 0, 0), 400, 400)
      expect_true(is.finite(ctx$sx))
      expect_true(is.finite(ctx$sy))
})

test_that("pt_to_npc returns NULL for non-finite input", {
      ctx <- make_device_context(c(-1, 1, -1, 1), 400, 400)
      expect_null(pt_to_npc(NA_real_, 0, ctx))
      expect_null(pt_to_npc(0, Inf, ctx))
})


# Depth-scaled label sizing -------------------------------------------------

test_that("calculate_gridline_position picks the requested endpoint", {
      gridline <- data.frame(x = c(0, 10), y = c(0, 20),
                             depth_scale = c(0.8, 1.2))

      start <- calculate_gridline_position(gridline, axis_uses_start = TRUE)
      expect_equal(start$x, 0)
      expect_equal(start$y, 0)
      expect_equal(start$depth_scale, 0.8)

      end <- calculate_gridline_position(gridline, axis_uses_start = FALSE)
      expect_equal(end$x, 10)
      expect_equal(end$depth_scale, 1.2)
})

test_that("calculate_gridline_position defaults depth_scale when absent or invalid", {
      no_column <- data.frame(x = c(0, 1), y = c(0, 1))
      expect_equal(calculate_gridline_position(no_column, TRUE)$depth_scale, 1)

      bad <- data.frame(x = c(0, 1), y = c(0, 1), depth_scale = c(NA, -1))
      expect_equal(calculate_gridline_position(bad, TRUE)$depth_scale, 1)
      expect_equal(calculate_gridline_position(bad, FALSE)$depth_scale, 1)
})

test_that("resolve_fontsize falls back for missing or invalid sizes", {
      expect_equal(resolve_fontsize(NULL, 8.5), 8.5)
      expect_equal(resolve_fontsize(0, 8.5), 8.5)
      expect_equal(resolve_fontsize(-2, 8.5), 8.5)
      expect_equal(resolve_fontsize(c(1, 2), 8.5), 8.5)
      expect_equal(resolve_fontsize(12, 8.5), 12)
})


# Margin resolution ---------------------------------------------------------

test_that("calculate_axis_offsets floors the text margin against the text size", {
      # Tick clearance is added at render time from the drawn geometry, so a
      # small theme margin is raised to a fraction of the text size instead
      elements <- list(
            axis_text = list(margin = margin(2, 2, 2, 2), size = 10),
            axis_title = list(margin = margin(4, 4, 4, 4)),
            parent_text = list(margin = margin(0, 0, 0, 0))
      )

      offsets <- calculate_axis_offsets(elements, rotate_labels = TRUE)
      expect_equal(offsets$text_offset, 6)
})

test_that("calculate_axis_offsets honours a margin larger than the floor", {
      elements <- list(
            axis_text = list(margin = margin(20, 20, 20, 20), size = 10),
            axis_title = list(margin = margin(4, 4, 4, 4)),
            parent_text = list(margin = margin(0, 0, 0, 0))
      )

      offsets <- calculate_axis_offsets(elements, rotate_labels = TRUE)
      expect_equal(offsets$text_offset, 20)
})

test_that("calculate_axis_offsets falls back to a default text size", {
      elements <- list(
            axis_text = list(margin = margin(2, 2, 2, 2)),
            axis_title = list(margin = margin(4, 4, 4, 4)),
            parent_text = list(margin = margin(0, 0, 0, 0))
      )

      offsets <- calculate_axis_offsets(elements, rotate_labels = TRUE)
      expect_equal(offsets$text_offset, 0.6 * 8.5)
})

test_that("calculate_axis_offsets no longer consumes the nominal tick length", {
      base <- list(
            axis_text = list(margin = margin(2, 2, 2, 2), size = 10),
            axis_title = list(margin = margin(4, 4, 4, 4)),
            parent_text = list(margin = margin(0, 0, 0, 0))
      )

      with_tick <- base
      with_tick$tick_length <- 3

      expect_equal(calculate_axis_offsets(base, rotate_labels = TRUE)$text_offset,
                   calculate_axis_offsets(with_tick, rotate_labels = TRUE)$text_offset)
})

StatIdentity3D <- ggproto("StatIdentity3D", Stat,
                          required_aes = c("x", "y"),  # Require at least x and y like geom_point

                          compute_panel = function(data, scales, na.rm = FALSE, light = NULL,
                                                   cull_backfaces = FALSE) {

                                # Remove missing values if requested
                                if (na.rm) {
                                      complete_cases <- complete.cases(data)
                                      data <- data[complete_cases, ]
                                }

                                # Handle empty data
                                if (nrow(data) == 0) {
                                      return(data)
                                }

                                # Convert group values to hierarchical format for depth sorting
                                if (!"group" %in% names(data)) {
                                      data$group <- "-1__group"
                                } else {
                                      data$group <- paste0(data$group, "__group")
                                }

                                # Store original z values if z column exists (following stat_voxel_3d pattern)
                                if ("z" %in% names(data)) {
                                      data$z_raw <- data$z

                                      # Convert discrete z to numeric positions
                                      if (is.factor(data$z) || is.character(data$z)) {
                                            data$z <- as.numeric(as.factor(data$z))
                                      }
                                }

                                # Also handle x and y for completeness (some 3D geoms might need this)
                                if ("x" %in% names(data)) {
                                      data$x_raw <- data$x

                                      if (is.factor(data$x) || is.character(data$x)) {
                                            data$x <- as.numeric(as.factor(data$x))
                                      }
                                }

                                if ("y" %in% names(data)) {
                                      data$y_raw <- data$y

                                      if (is.factor(data$y) || is.character(data$y)) {
                                            data$y <- as.numeric(as.factor(data$y))
                                      }
                                }

                                if(!is.null(cull_backfaces)) data$cull_backfaces <- cull_backfaces

                                return(attach_light(data, light))
                          }
)

#' 3D-aware identity transformation
#'
#' This stat performs identity transformation (passes data through unchanged) while
#' properly handling discrete scales and lighting specifications for 3D coordinate
#' systems. It also converts group values to hierarchical format to prevent
#' reordering withing groups during depth sorting.
#'
#' This stat is primarily intended for use with 3D geoms that need discrete scale
#' or lighting support, following the same pattern as other ggcube stats.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data The data to be displayed in this layer.
#' @param geom The geometric object to use display the data.
#' @param position Position adjustment, defaults to "identity".
#' @inheritParams light_param
#' @param na.rm If `FALSE`, missing values are removed with a warning.
#' @param show.legend Logical indicating whether this layer should be included in legends.
#' @param inherit.aes If `FALSE`, overrides the default aesthetics.
#' @param ... Other arguments passed on to geom layer.
#'
#' @section Computed variables:
#' - `x_raw`, `y_raw`, `z_raw`: Original values before discrete-to-numeric conversion
#' - `group`: Converted to hierarchical format (e.g., "1__group", "2__group") for proper depth sorting.
#'
#' @seealso [geom_point_3d()], [geom_polygon_3d()] which use this stat for discrete scale support.
#' @return A `Layer` object that can be added to a ggplot.
#' @export
stat_identity_3d <- function(mapping = NULL, data = NULL,
                             geom = "point", position = "identity",
                             na.rm = FALSE, show.legend = NA, inherit.aes = TRUE,
                             light = NULL, ...) {

      layer(
            stat = StatIdentity3D, data = data, mapping = mapping, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, light = light, ...)
      )
}

#' Color guides showing lighting effects
#'
#' Creates color guides that show shading variation as gradients
#' within each color. Shows the full range of colors visible when
#' shading is enabled in 3D plots.
#'
#' When fill and color aesthetics map to the same variable (e.g.,
#' \code{aes(fill = z, color = z)}), ggplot2 creates a shared scale with a
#' single guide. In this case, use \code{guides(fill = guide_*_3d())} to
#' apply shading, **not** \code{guides(color = guide_*_3d())}, even if your layer
#' uses the color aesthetic. Only use the color guide when color and fill
#' map to different variables and you want separate guides for each.
#'
#' @param reverse_shade Logical. If TRUE, reverses the lighting gradient direction. By default,
#'   shadows are placed on the left, or on the bottom for horizontal colorbars.
#' @param shade_range Length-2 numeric vector in the range -1 to 1, giving the limits
#'   of the shading gradient. -1 is full shade, and 1 is full highlight. Default is `c(-.5, .5)`.
#' @param ... Additional arguments passed to \code{guide_colorbar()} or \code{guide_legend()}.
#' @return A guide object that displays shading effects
#' @examples
#' # continuous `colorbar` guide
#' ggplot(mountain, aes(x, y, z, fill = z)) +
#'    stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
#'    guides(fill = guide_colorbar_3d()) +
#'    scale_fill_gradientn(colors = c("tomato", "dodgerblue")) +
#'    coord_3d()
#'
#' # discrete `legend` guide
#' ggplot(mountain, aes(x, y, z, fill = x > .5, group = 1)) +
#'    stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
#'    guides(fill = guide_legend_3d()) +
#'    coord_3d()
#' @name guide_3d
#' @export
guide_colorbar_3d <- function(reverse_shade = FALSE, shade_range = c(-.5, .5), ...) {
      # Create normal guide
      guide <- guide_colorbar(...)

      # Store the original draw method
      original_draw <- guide$draw

      # Override the draw method with custom version
      guide$draw <- function(theme, position = NULL, direction = NULL, params = guide$params) {

            # Get original grob
            original_grob <- original_draw(theme, position, direction, params)

            # Get lighting info
            lighting_info <- extract_light_from_plot()
            if (!inherits(lighting_info, "light") || lighting_info$shade == "neither") {
                  return(original_grob)
            }

            # Create lighting gradients
            base_colors <- params$decor[["colour"]]
            grad <- create_light_gradients(base_colors, lighting_info, shade_range)
            if(reverse_shade){
                  if(is.null(params$direction) || params$direction == "vertical"){
                        grad <- grad[, ncol(grad):1]
                  }else{
                        grad <- grad[nrow(grad):1, ]
                  }
            }

            # Replace colors in grob
            shaded_grob <- replace_colorbar_colors(original_grob, grad)
            return(shaded_grob)
      }

      guide
}

#' @rdname guide_3d
#' @export
guide_legend_3d <- function(reverse_shade = FALSE, shade_range = c(-.5, .5), ...) {
      # Create normal guide
      guide <- guide_legend(...)

      # Store the original draw method
      original_draw <- guide$draw

      # Override the draw method with custom version
      guide$draw <- function(theme, position = NULL, direction = NULL, params = guide$params) {

            # Get original grob
            original_grob <- original_draw(theme, position, direction, params)

            # Get lighting info
            lighting_info <- extract_light_from_plot()
            if (!inherits(lighting_info, "light") || lighting_info$shade == "neither") {
                  return(original_grob)
            }

            # Create lighting gradients
            base_colors <- params$key[[params$aesthetic]]
            grad <- create_light_gradients(base_colors, lighting_info, shade_range)
            if(reverse_shade){
                  if(is.null(params$direction) || params$direction == "vertical"){
                        grad <- grad[, ncol(grad):1]
                  }else{
                        grad <- grad[nrow(grad):1, ]
                  }
            }

            # Replace colors in grob
            shaded_grob <- replace_legend_colors(original_grob, grad)
            return(shaded_grob)
      }

      guide
}



#' Extract lighting information from plot layers
#'
#' @return List with lighting specification or NULL if none found
#' @keywords internal
#' @noRd
extract_light_from_plot <- function() {
      # Walk up the call stack to find the plot object
      for (i in 1:20) {
            tryCatch({
                  env <- parent.frame(i)

                  if (exists("plot", envir = env)) {
                        plot_obj <- get("plot", envir = env)
                        if (inherits(plot_obj, "ggplot")) {

                              # Look for layers with lighting (existing behavior)
                              lighting_layers <- list()
                              for (j in seq_along(plot_obj$layers)) {
                                    layer <- plot_obj$layers[[j]]

                                    if ("light" %in% names(layer$stat_params)) {
                                          lighting_layers[[length(lighting_layers) + 1]] <- layer$stat_params$light
                                    }
                              }

                              # If layer lighting found, use it (maintains precedence)
                              if (length(lighting_layers) > 0) {
                                    if (length(lighting_layers) == 1) {
                                          return(lighting_layers[[1]])
                                    } else {
                                          warning("Multiple layers with different lighting found. Using first layer's lighting.")
                                          return(lighting_layers[[1]])
                                    }
                              }

                              # No layer lighting found - check the coord. Its
                              # `light` may be a `waiver()` when lighting was
                              # never specified, or was supplied by `+ light()`
                              # before `coord_3d()` and so is still stashed on
                              # the plot. Resolve both cases to a real spec.
                              if (!is.null(plot_obj$coordinates) && "light" %in% names(plot_obj$coordinates)) {
                                    coord_light <- plot_obj$coordinates$light
                                    if (inherits(coord_light, "waiver")) {
                                          pending <- attr(plot_obj, "ggcube_pending_light")
                                          # A waiver means lighting was never set
                                          # on the coord: either it came from
                                          # `+ light()` added before `coord_3d()`
                                          # and is still stashed, or nothing was
                                          # specified and the default applies.
                                          coord_light <- if (!is.null(pending)) pending else light()
                                    }
                                    if (!is.null(coord_light)) {
                                          return(coord_light)
                                    }
                              }

                              # No lighting found anywhere
                              return(NULL)
                        }
                  }
            }, error = function(e) {
                  # Skip frames that cause errors
            })
      }

      return(NULL)
}

#' Create shading gradients for colorbar
#'
#' @param base_colors Character vector of base colors
#' @param lighting_info Lighting specification object
#' @param n_steps Number of lighting gradient steps (default 10)
#' @return Matrix of colors with lighting applied
#' @keywords internal
#' @noRd
create_light_gradients <- function(base_colors,
                                   lighting_info,
                                   shade_range = c(-1, 1),
                                   n_steps = 20) {

      # Reverse order for proper gradient direction
      base_colors <- rev(base_colors)

      # Generate lighting values across range
      if(any(abs(shade_range) > 1)) stop("`shade_range` can't include values outside the range -1 to 1.")
      light <- seq(shade_range[1], shade_range[2], length.out = n_steps)

      # Transform to give appearance of convex surface
      light <- sin(light * pi/2)

      # Adjust range in accordance with lighting method
      if (lighting_info$method != "diffuse") {
            light <- scales::rescale(light, to = c(0, 1), from = c(-1, 1))
      }

      # Create matrix: rows = base colors, columns = lighting steps
      grad <- matrix(NA_character_, nrow = length(base_colors), ncol = n_steps)

      # Apply lighting to base colors across
      for (i in seq_along(base_colors)) {
            gradient_colors <- blend_light_with_colors(
                  rep(base_colors[i], n_steps),
                  light,
                  lighting_info
            )
            grad[i, ] <- gradient_colors
      }

      return(grad)
}

#' Replace colorbar colors with shading gradients
#'
#' @param grob Original colorbar grob
#' @param grad Matrix of gradient colors
#' @return Modified grob with gradients
#' @keywords internal
#' @noRd
replace_colorbar_colors <- function(grob, grad) {

      # Handle gtable (grid table) structure
      if (inherits(grob, "gtable")) {

            # Look through the grobs stored in the gtable
            for (i in seq_along(grob$grobs)) {
                  subgrob <- grob$grobs[[i]]

                  # Look for rastergrob which holds the colorbar image
                  if (inherits(subgrob, "rastergrob")) {

                        original_raster <- subgrob$raster

                        # Handle orientation
                        horizontal <- diff(dim(original_raster)) > 0
                        if(horizontal){
                              original_raster <- t(original_raster)
                              # grad <- t(grad)
                        }

                        # Create a new raster with lighting gradients
                        n_colors <- nrow(grad)
                        n_light_steps <- ncol(grad)

                        # Create wider raster to show lighting gradient
                        new_width <- n_light_steps
                        new_height <- nrow(original_raster)

                        # Create the 2D lighting raster
                        new_raster <- array(NA_character_, dim = c(new_height, new_width))

                        for (row in 1:new_height) {
                              # Map row to color (top to bottom = first to last color)
                              color_idx <- ceiling(row / new_height * n_colors)
                              color_idx <- max(1, min(color_idx, n_colors))

                              for (col in 1:new_width) {
                                    # Map column to lighting step (left to right = shadow to highlight)
                                    lighting_idx <- ceiling(col / new_width * n_light_steps)
                                    lighting_idx <- max(1, min(lighting_idx, n_light_steps))

                                    new_raster[row, col] <- grad[color_idx, lighting_idx]
                              }
                        }

                        # Correct orientation
                        if(horizontal){
                              new_raster <- t(new_raster) # transpose axes
                              new_raster <- new_raster[nrow(new_raster):1, ] # move shadow to bottom
                              new_raster <- new_raster[, ncol(new_raster):1] # flip color gradient
                        }

                        # Convert to raster object and replace
                        new_raster <- as.raster(new_raster)
                        grob$grobs[[i]]$raster <- new_raster

                        break  # Only modify the first rastergrob found
                  }
            }
      }

      return(grob)
}

#' Replace legend key colors with shading gradients
#'
#' Locates the colored key cells structurally rather than by matching a
#' specific gtable layout-name scheme, so it works across ggplot2 versions
#' whose internal naming and nesting differ. In ggplot2 >= 4.0 the legend is
#' wrapped in an outer "guide-box" gtable and each key cell is a gTree
#' containing both a background rect and the layer's drawn glyph rect; in
#' earlier versions the legend gtable is passed directly and key cells are
#' bare rects.
#'
#' @param grob Original legend grob
#' @param grad Matrix of gradient colors
#' @return Modified grob with gradients
#' @keywords internal
#' @noRd
replace_legend_colors <- function(grob, grad) {
      if (!inherits(grob, "gtable")) {
            return(grob)
      }

      # In ggplot2 >= 4.0 the legend gtable is nested inside an outer
      # "guide-box" wrapper. Descend into any child gtables that themselves
      # contain key cells, operate on them, and write them back.
      key_cells_here <- find_legend_key_cells(grob)
      if (length(key_cells_here) == 0) {
            modified <- FALSE
            for (i in seq_along(grob$grobs)) {
                  child <- grob$grobs[[i]]
                  if (inherits(child, "gtable") &&
                      length(find_legend_key_cells(child)) > 0) {
                        grob$grobs[[i]] <- replace_legend_colors(child, grad)
                        modified <- TRUE
                  }
            }
            return(grob)
      }

      # Order key cells by grid position (top-to-bottom, then left-to-right)
      # so cell i corresponds to row i of the gradient matrix, matching the
      # base-colour order regardless of layout naming.
      ord <- order(grob$layout$t[key_cells_here], grob$layout$l[key_cells_here])
      key_cells <- key_cells_here[ord]

      n <- min(length(key_cells), nrow(grad))
      for (i in seq_len(n)) {
            idx <- key_cells[i]
            grob$grobs[[idx]] <- replace_key_glyph(grob$grobs[[idx]], grad[i, ])
      }

      return(grob)
}

#' Identify the colored key cells within a legend gtable
#'
#' @param grob A gtable
#' @return Integer indices into grob$grobs / grob$layout for key cells
#' @keywords internal
#' @noRd
find_legend_key_cells <- function(grob) {
      if (!inherits(grob, "gtable") || is.null(grob$layout$name)) {
            return(integer(0))
      }
      nm <- grob$layout$name
      # ggplot2 >= 4.0: "key-1-1-bg"; ggplot2 3.5.x: "key-1-1-1" etc.
      grep("^key-\\d+-\\d+(-bg|-\\d+)?$", nm)
}

#' Replace the layer glyph in a single legend key cell with a gradient raster
#'
#' Handles both a bare rect cell (older ggplot2) and a gTree cell containing a
#' background rect plus the layer's drawn glyph (ggplot2 >= 4.0). For the gTree
#' case the layer glyph is identified via childrenOrder, where the background
#' carries an empty name and the layer glyph is named after its layer.
#'
#' @param cell A grob (rect or gTree) for one legend key
#' @param gradient_colors Character vector of colors for this key
#' @return Modified grob
#' @keywords internal
#' @noRd
replace_key_glyph <- function(cell, gradient_colors) {

      # Bare rect cell: replace directly.
      if (inherits(cell, "rect")) {
            return(legend_key_raster_grob(cell, gradient_colors))
      }

      # gTree cell: replace the layer glyph rect among the children.
      if (inherits(cell, "gTree") && length(cell$children) > 0) {
            child_names <- names(cell$children)
            order_names <- names(cell$childrenOrder)

            # The layer glyph is the child whose childrenOrder name is non-empty
            # (the key background's name is ""). Fall back to the last rect child.
            glyph_name <- NULL
            if (!is.null(order_names)) {
                  named <- cell$childrenOrder[nzchar(order_names)]
                  if (length(named) > 0) {
                        glyph_name <- unname(named[length(named)])
                  }
            }

            rect_children <- child_names[vapply(
                  cell$children, function(g) inherits(g, "rect"), logical(1)
            )]

            if (is.null(glyph_name) || !(glyph_name %in% child_names) ||
                !inherits(cell$children[[glyph_name]], "rect")) {
                  if (length(rect_children) == 0) return(cell)
                  glyph_name <- rect_children[length(rect_children)]
            }

            ref_rect <- cell$children[[glyph_name]]
            cell$children[[glyph_name]] <- legend_key_raster_grob(ref_rect, gradient_colors)
            return(cell)
      }

      return(cell)
}

#' Build a gradient raster grob matching a reference rect's geometry
#'
#' @param rect_grob Reference rect grob whose position/size/just are reused
#' @param gradient_colors Character vector of colors (shadow to highlight)
#' @return A rasterGrob
#' @keywords internal
#' @noRd
legend_key_raster_grob <- function(rect_grob, gradient_colors) {

      x <- rect_grob$x %||% grid::unit(0.5, "npc")
      y <- rect_grob$y %||% grid::unit(0.5, "npc")
      width <- rect_grob$width %||% grid::unit(1, "npc")
      height <- rect_grob$height %||% grid::unit(1, "npc")
      just <- rect_grob$just %||% c(0.5, 0.5)

      # Horizontal gradient raster (left to right = shadow to highlight)
      n_steps <- length(gradient_colors)
      raster_matrix <- matrix(gradient_colors, nrow = 1, ncol = n_steps)
      raster_obj <- as.raster(raster_matrix)

      grid::rasterGrob(
            image = raster_obj,
            x = x,
            y = y,
            width = width,
            height = height,
            just = just
      )
}

test_that("find_legend_key_cells matches both naming schemes and rejects non-key cells", {
      # Synthetic gtable whose layout names mimic the cells ggplot2 produces.
      # ggplot2 >= 4.0 uses "key-1-1-bg"; ggplot2 3.5.x uses "key-1-1-1".
      gt <- gtable::gtable(widths = grid::unit(1, "null"),
                           heights = grid::unit(1, "null"))
      gt$layout <- data.frame(
            t = c(1, 2, 3, 4, 5, 6),
            l = c(1, 1, 1, 1, 1, 1),
            b = c(1, 2, 3, 4, 5, 6),
            r = c(1, 1, 1, 1, 1, 1),
            z = c(1, 2, 3, 4, 5, 6),
            clip = "off",
            name = c("key-1-1-bg",   # 4.0.x key cell
                     "key-3-1-bg",   # 4.0.x key cell
                     "key-1-1-1",    # 3.5.x key cell
                     "label-1-2",    # not a key
                     "title",        # not a key
                     "background"),  # not a key
            stringsAsFactors = FALSE
      )

      idx <- find_legend_key_cells(gt)

      expect_setequal(gt$layout$name[idx],
                      c("key-1-1-bg", "key-3-1-bg", "key-1-1-1"))
})

test_that("replace_key_glyph replaces a bare rect cell with a raster grob", {
      cell <- grid::rectGrob(
            x = grid::unit(0.5, "npc"), y = grid::unit(0.5, "npc"),
            width = grid::unit(1, "npc"), height = grid::unit(1, "npc"),
            gp = grid::gpar(fill = "red")
      )

      out <- replace_key_glyph(cell, c("#000000", "#888888", "#ffffff"))

      expect_true(inherits(out, "rastergrob"))
})

test_that("replace_key_glyph replaces only the layer glyph inside a gTree cell", {
      # Mimic the ggplot2 >= 4.0 key cell: a background rect plus the layer's
      # drawn glyph rect. childrenOrder names the background "" and the glyph
      # after its layer.
      bg <- grid::rectGrob(gp = grid::gpar(fill = "grey95"),
                           name = "legend.key.rect")
      glyph <- grid::rectGrob(gp = grid::gpar(fill = "red"),
                              name = "layer.glyph")
      cell <- grid::gTree(children = grid::gList(bg, glyph),
                          childrenOrder = stats::setNames(
                                c("legend.key.rect", "layer.glyph"),
                                c("", "stat_surface_3d")))

      out <- replace_key_glyph(cell, c("#000000", "#888888", "#ffffff"))

      # The layer glyph becomes a raster; the background rect is preserved.
      expect_true(inherits(out$children[["layer.glyph"]], "rastergrob"))
      expect_true(inherits(out$children[["legend.key.rect"]], "rect"))
})

test_that("replace_legend_colors converts key glyphs to rasters in a rendered legend", {
      library(ggplot2)

      p <- ggplot(mountain, aes(x, y, z, fill = x > .5, group = 1)) +
            stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
            guides(fill = guide_legend_3d()) +
            coord_3d(ratio = c(2, 3, 1.5))

      gt <- ggplotGrob(p)
      box_idx <- which(grepl("guide-box", gt$layout$name))
      expect_gt(length(box_idx), 0)

      # Use the first non-empty guide-box (legend position varies by version).
      legend <- NULL
      for (i in box_idx) {
            g <- gt$grobs[[i]]
            if (inherits(g, "gtable")) {
                  legend <- g
                  break
            }
      }
      expect_false(is.null(legend))

      # Supply a gradient directly, bypassing extract_light_from_plot() (which
      # only resolves during a live render).
      grad <- matrix(
            c("#000000", "#444444", "#888888", "#bbbbbb", "#ffffff"),
            nrow = 2, ncol = 5, byrow = TRUE
      )

      shaded <- replace_legend_colors(legend, grad)

      # Recursively collect grob class names and assert at least one raster
      # appeared where the key glyphs were.
      collect_classes <- function(g) {
            classes <- class(g)
            if (inherits(g, "gtable") || inherits(g, "gTree")) {
                  kids <- g$grobs %||% g$children
                  for (k in kids) classes <- c(classes, collect_classes(k))
            }
            classes
      }

      expect_true("rastergrob" %in% collect_classes(shaded))
})

test_that("replace_colorbar_colors shades the raster in a rendered colorbar guide", {
      library(ggplot2)

      p <- ggplot(mountain, aes(x, y, z, fill = z)) +
            stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
            guides(fill = guide_colorbar_3d()) +
            scale_fill_gradientn(colors = c("tomato", "dodgerblue")) +
            coord_3d()

      gt <- ggplotGrob(p)
      box_idx <- which(grepl("guide-box", gt$layout$name))
      expect_gt(length(box_idx), 0)

      # Locate the colorbar gtable that replace_colorbar_colors actually
      # operates on: the one whose direct children include the rastergrob.
      # This mirrors the function's own (non-recursive) search depth, so the
      # test fails if the raster ever nests deeper than the function reaches.
      find_colorbar_gtable <- function(g) {
            if (inherits(g, "gtable")) {
                  for (sub in g$grobs) {
                        if (inherits(sub, "rastergrob")) return(g)
                  }
                  for (sub in g$grobs) {
                        hit <- find_colorbar_gtable(sub)
                        if (!is.null(hit)) return(hit)
                  }
            }
            NULL
      }

      colorbar <- NULL
      for (i in box_idx) {
            colorbar <- find_colorbar_gtable(gt$grobs[[i]])
            if (!is.null(colorbar)) break
      }
      expect_false(is.null(colorbar))

      # Capture the original raster width so we can confirm it changed.
      orig_raster <- NULL
      for (g in colorbar$grobs) {
            if (inherits(g, "rastergrob")) {
                  orig_raster <- g$raster
                  break
            }
      }
      expect_false(is.null(orig_raster))

      n_colors <- 6
      n_light_steps <- 8
      ramp <- grDevices::colorRampPalette(c("tomato", "dodgerblue"))(n_colors)
      grad <- matrix(rep(ramp, times = n_light_steps),
                     nrow = n_colors, ncol = n_light_steps)

      shaded <- replace_colorbar_colors(colorbar, grad)

      shaded_raster <- NULL
      for (g in shaded$grobs) {
            if (inherits(g, "rastergrob")) {
                  shaded_raster <- g$raster
                  break
            }
      }
      expect_false(is.null(shaded_raster))

      # A vertical colorbar gains lighting columns; a horizontal one gains rows.
      # Either way the raster's dimensions must change from the original.
      expect_false(identical(dim(shaded_raster), dim(orig_raster)))
      expect_true(n_light_steps %in% dim(shaded_raster))
})

test_that("replace_colorbar_colors widens the raster to the lighting steps", {
      # A vertical colorbar raster is a single column of colors (n rows x 1 col).
      n_colors <- 6
      ramp <- grDevices::colorRampPalette(c("tomato", "dodgerblue"))(n_colors)
      original_raster <- as.raster(matrix(ramp, ncol = 1))

      cbar <- grid::rasterGrob(original_raster)
      gt <- gtable::gtable(widths = grid::unit(1, "null"),
                           heights = grid::unit(1, "null"))
      gt <- gtable::gtable_add_grob(gt, cbar, t = 1, l = 1, name = "bar")

      # Gradient with a known number of lighting steps; columns drive new width.
      n_light_steps <- 8
      grad <- matrix(
            rep(ramp, times = n_light_steps),
            nrow = n_colors, ncol = n_light_steps
      )

      shaded <- replace_colorbar_colors(gt, grad)

      # Find the (modified) rastergrob and confirm it gained lighting columns.
      raster_grob <- NULL
      for (g in shaded$grobs) {
            if (inherits(g, "rastergrob")) {
                  raster_grob <- g
                  break
            }
      }
      expect_false(is.null(raster_grob))

      # Original was 1 column wide; shaded raster should span n_light_steps.
      expect_equal(ncol(raster_grob$raster), n_light_steps)
      expect_equal(nrow(raster_grob$raster), n_colors)
})

# Geom that tessellates point grid into polygon tiles for 3D surface rendering

GeomSurface3D <- ggproto("GeomSurface3D", GeomPolygon3D,

                         required_aes = c("x", "y", "z"),

                         default_aes = aes(
                               colour = NA,
                               fill = "grey60",
                               linewidth = 0.1,
                               linetype = 1,
                               alpha = NA
                         ),

                         extra_params = c("na.rm", "annotate", "method", "grid", "cull_backfaces",
                                          "sort_method", "scale_depth", "force_convex"),

                         setup_params = function(data, params) {
                               params$method <- params$method %||% "auto"
                               params$grid <- params$grid %||% "rectangle"
                               params$grid <- match.arg(params$grid, c("rectangle", "right1", "right2", "equilateral"))
                               params
                         },

                         setup_data = function(data, params) {

                               # Convert point grid to polygon tiles
                               data <- points_to_tiles(
                                     data,
                                     method = params$method,
                                     grid_type = params$grid,
                                     group_prefix = "surface__tile"
                               )

                               # Average aesthetics per polygon
                               data <- average_aesthetics(data)

                               setup_annotations(data, params)
                         }

                         # draw_panel inherited from GeomPolygon3D
)


# geom_surface_3d() function is defined in stat-surface-3d.R

#' Lighting specification for 3D surface rendering
#'
#' Creates a lighting specification object for use with 3D polygon layers.
#' Lighting modifies the brightness of fill and/or base color aesthetics based on surface
#' orientation (i.e., it implements form shadows but not cast shadows).
#' Various options are available to control light qualities and light source location.
#' The result can be added directly to a plot with `+ light()` to set lighting for
#' the whole plot. It can also be passed to individual `geom_*_3d()` layer functions,
#' which takes precedence over plot-level lighting.
#'
#' Note that light-like effects can also be achieved in some stats by mapping color
#' aesthestics to computed variables such as `after_stat(dzdx)`; see [geom_surface_3d()]
#' for examples.
#'
#' @param method Character string specifying lighting model:
#'   \itemize{
#'     \item \code{"diffuse"}: The default. Atmospheric lighting with soft shadows (only surfaces pointing
#'          directly away from light source are fully dark; base color occurs on surfaces perpendicular to light)
#'     \item \code{"direct"}: Direct lighting with hard shadows (all surfaces angled beyond 90 degrees from
#'          light source are fully dark; base color occurs on surfaces angled 45 degrees toward light)
#'     \item \code{"rgb"}: Map surface orientation to three dimensional color space. Rather than the light model
#'          darkening or brightening base colors as in the other models, this model generates a wholly new set of
#'           colors. If this option is selected, parameters like `mode` and `contrast` are ignored, but
#'           `fill`, `color`, and `direction` still apply.
#'     \item \code{"none"}: Disable lighting entirely. All other parameters are ignored.
#'          Equivalent to \code{light(fill = FALSE, color = FALSE)}, and to passing the
#'          bare string \code{"none"} to a \code{light} argument.
#'   }
#' @param mode Character string specifying color lighting mode:
#'   \itemize{
#'     \item \code{"hsv"}: The default. Modifies _value_ component of _HSV_ color
#'          (fades to bright colors at high end, black at low end).
#'     \item \code{"hsl"}: Modifies _lightness_ component of _HSL_ color
#'          (fades to white at high end, black at low end).
#'   }
#'   These two options give identical results for grayscale colors.
#' @param fill Logical indicating whether to apply lighting to fill colors. Default is TRUE.
#' @param color Logical indicating whether to apply lighting to border/line colors. Default is TRUE.
#' @param contrast Numeric value greater than zero controlling the intensity of lighting effects.
#'   1.0 (the default) gives full black-to-white range. Values less than 1 give subtler effects, while
#'   values greater than 1 give more dramatic effects.
#' @param direction Numeric vector of length 3 specifying direction in 3D space (x, y, z) that
#'   light comes from for directional lighting. Depending on `anchor`, direction is relative to
#'   either the plot axes or the viewing pane. The default is \code{c(-.5, 0, 1)}, giving
#'   diagonal lighting from the upper right edge with default rotation and anchor. At least
#'   one value must be non-zero. Values are automatically normalized, so magnitude doesn't
#'   matter, only sign and relative magnitude. Direction is specified in visual space (relative
#'   to the rendered cube's bounding box), not data space. This argument is ignored if
#'   \code{position} is provided.
#' @param anchor Character string specifying the reference frame for light direction:
#'   \itemize{
#'     \item \code{"scene"} (default): Light `direction` is fixed relative to the data/scene.
#'       Light rotates with the plot, so regardless of rotation, `direction = c(1, 0, 0)`
#'       lights surfaces facing the "xmax" cube face, for example.
#'     \item \code{"camera"}: Light `direction` is fixed relative to the camera/viewer.
#'       Light direction stays in place regardless of rotation, so `direction = c(1, 0, 0)`
#'       lights surfaces facing the right side of the plot, for example.
#'   }
#'   When all rotation parameters are zero, these options give the same result.
#' @param position Numeric vector of length 3 specifying light source position in
#'   data coordinate space for positional lighting. When specified, each face gets
#'   its own light direction calculated from the light position to the face center.
#'   Mutually exclusive with \code{direction}. Default is NULL (use directional lighting).
#' @param distance_falloff Logical indicating whether to apply distance-based
#'   intensity falloff for positional lighting using inverse square law
#'   (intensity proportional to 1/distance²). Only used when \code{position} is specified.
#'   Default is FALSE.
#' @param backface_scale,backface_offset Numeric values that determine how "frontface" light values get
#'   modified (scaled and then offset) to derive "backface" light values. A backface is the side of a
#'   polygon that faces the underside of a surface or the inside of a volume. Frontface light values are
#'   typically in the range `[-1, 1]` (unless `contrast` is boosted). The default scale of -1 gives
#'   backfaces highly contrasting lighting to frontfaces. To light backfaces the same as frontfaces,
#'   set scale to 1. To uniformly darken (brighten) all backfaces, use a negative (positive) offset.
#'
#' @return A \code{lighting} object that can be added to a plot or passed to polygon-based 3D layers
#'   or to `coord_3d()`.
#' @examples
#' # base plot used in examples
#' p <- ggplot(sphere_points, aes(x, y, z)) +
#'   geom_hull_3d(fill = "#9e2602", color = "#5e1600") +
#'   coord_3d(scales = "fixed")
#'
#' # Light qualities ----------------------------------------------------------
#'
#' # default `"diffuse"` lighting
#' p
#'
#' \donttest{
#'
#' # use `"direct"` lighting to apply full shade to unlit surfaces
#' p + light(method = "direct")
#'
#' # use `"hsl"` mode to fade highlights to white
#' p + light(mode = "hsl")
#'
#' # adjust lighting intensity with `contrast`
#' p + light(mode = "hsl", contrast = 1.5)
#'
#' # use `"rgb"` lighting to map face orientation to 3D color space
#' p + light(method = "rgb")
#'
#'
#' # Lighting targets ---------------------------------------------------------
#'
#' # use `fill` and `color` to select which aesthetics get modified by light
#' p + light(fill = TRUE, color = FALSE)
#'
#' # apply lighting to aesthetically mapped colors, with shaded guide to match
#' p + geom_hull_3d(aes(fill = x, color = x)) +
#'       scale_fill_viridis_c() +
#'       scale_color_viridis_c() +
#'       guides(fill = guide_colorbar_3d()) +
#'       light(mode = "hsl", contrast = .9)
#'
#' # disable lighting entirely
#' p + light("none")
#'
#' # if provided, layer-level lighting overrides plot-level lighting
#' p + light("direct") + # plot-level: affects original layer
#'   geom_hull_3d(aes(x = x + 2.5), fill = "#9e2602", color = "#5e1600",
#'       light = light("direct", mode = "hsl", direction = c(0, -1, 0)))
#'
#'
#' # Light sources ------------------------------------------------------------
#'
#' # directional light from the xmin-ymin-zmin direction
#' # (`direction` is relative to rotated axes with default `anchor = "scene"`)
#' p + light(direction = c(-1, -1, -1))
#'
#' # directional light from top-right corner of figure
#' # (`anchor = "camera"` makes `direction` fixed relative to the plot)
#' p + light(direction = c(1, 1, 0), anchor = "camera")
#'
#' # positional light source within plot
#' ggplot(mountain, aes(x, y, z)) +
#'   stat_surface_3d(fill = "red", color = "red") +
#'   light(position = c(.5, .7, 95), distance_falloff = TRUE,
#'         mode = "hsl", contrast = .9) +
#'   coord_3d(ratio = c(1.5, 2, 1))
#'
#'
#' # Backface lighting --------------------------------------------------------
#'
#' # backfaces get "opposite" lighting by default (`backface_scale = -1`)
#' p <- ggplot() +
#'   geom_function_3d(fun = function(x, y) x^2 + y^2,
#'     xlim = c(-3, 3), ylim = c(-3, 3),
#'     fill = "steelblue", color = "steelblue") +
#'     coord_3d(pitch = 0, roll = -70, yaw = 0)
#' p + light(mode = "hsl")
#'
#' # use `backface_scale = 1` to light backfaces as if they're frontfaces
#' p + light(backface_scale = 1, mode = "hsl")
#'
#' # use `backface_offset` to uniformly darken (or lighten) backfaces
#' p + light(backface_scale = 1, mode = "hsl", backface_offset = -.5)
#'
#' }
#'
#' @export
light <- function(method = "diffuse",
                  direction = c(-.5, 0, 1),
                  anchor = "scene",
                  position = NULL,
                  distance_falloff = FALSE,
                  fill = TRUE,
                  color = TRUE,
                  mode = "hsv",
                  contrast = 1.0,
                  backface_scale = -1,
                  backface_offset = 0) {

      # `method = "none"` disables lighting. Every other parameter describes how to
      # light a surface, so all of them are irrelevant here: return the disabled
      # spec directly without validating them. This makes `light("none")`
      # equivalent to the bare string `"none"` accepted by the `light` argument of
      # layer, coord, and plot-level functions.
      if (identical(method, "none")) {
            return(light(fill = FALSE, color = FALSE))
      }

      # Validate method
      valid_methods <- c("direct", "diffuse", "rgb")
      if (!method %in% valid_methods) {
            stop("method must be one of: ",
                 paste(c(valid_methods, "none"), collapse = ", "))
      }

      # Validate direction
      if (!is.numeric(direction) || length(direction) != 3) {
            stop("direction must be a numeric vector of length 3")
      }

      if (all(direction == 0)) {
            stop("direction must contain at least one nonzero value")
      }

      # Validate position if provided
      if (!is.null(position)) {
            if (!is.numeric(position) || length(position) != 3) {
                  stop("position must be a numeric vector of length 3")
            }
      }

      # Validate distance_falloff
      if (!is.logical(distance_falloff) || length(distance_falloff) != 1) {
            stop("distance_falloff must be a single logical value")
      }

      # Validate fill and color
      if (!is.logical(fill) || length(fill) != 1) {
            stop("fill must be a single logical value")
      }
      if (!is.logical(color) || length(color) != 1) {
            stop("color must be a single logical value")
      }

      # Validate contrast
      if (!is.numeric(contrast) || length(contrast) != 1 || contrast < 0) {
            stop("contrast must be a single non-negative numeric value")
      }

      # Validate mode
      valid_modes <- c("hsv", "hsl")
      if (!mode %in% valid_modes) {
            stop("mode must be one of: ", paste(valid_modes, collapse = ", "))
      }

      # Validate anchor
      valid_anchors <- c("scene", "camera")
      if (!anchor %in% valid_anchors) {
            stop("anchor must be one of: ", paste(valid_anchors, collapse = ", "))
      }

      # Convert new API to old internal representation for backward compatibility
      if (fill && color) {
            shade <- "both"
      } else if (fill) {
            shade <- "fill"
      } else if (color) {
            shade <- "colour"
      } else {
            shade <- "neither"
      }

      # Create lighting specification object using original internal structure
      structure(
            list(
                  method = method,
                  direction = direction,
                  position = position,
                  distance_falloff = distance_falloff,
                  shade = shade,
                  shade_strength = contrast,
                  shade_mode = mode,
                  anchor = anchor,
                  backface_scale = backface_scale,
                  backface_offset = backface_offset
            ),
            class = "light"
      )
}

#' @export
print.light <- function(x, ...) {
      cat("Lighting specification:\n")
      cat("  Method:", x$method, "\n")

      if (!is.null(x$position)) {
            cat("  Position: [", paste(x$position, collapse = ", "), "] (positional lighting)\n")
            if (x$distance_falloff) {
                  cat("  Distance falloff: enabled\n")
            } else {
                  cat("  Distance falloff: disabled\n")
            }
      } else {
            cat("  Direction: [", paste(x$direction, collapse = ", "), "] (directional lighting)\n")
      }

      cat("  anchor:", x$anchor, "\n")

      # Show what gets shaded using new API terms
      shading_targets <- c()
      if (x$shade %in% c("fill", "both")) shading_targets <- c(shading_targets, "fill")
      if (x$shade %in% c("colour", "both")) shading_targets <- c(shading_targets, "color")

      if (length(shading_targets) > 0) {
            cat("  Shading:", paste(shading_targets, collapse = " + "),
                "(contrast =", x$shade_strength, ", mode =", x$shade_mode, ")\n")
      } else {
            cat("  Shading: disabled\n")
      }

      invisible(x)
}


#' Apply lighting models to surface normals with positional light support
#'
#' Computes lighting values from surface normals using various lighting models.
#' Supports both directional lighting (parallel rays) and positional lighting
#' (point light sources with per-face light directions).
#'
#' @param normals Matrix with 3 columns (x, y, z normal components), where each
#'   row represents a face normal vector. Should be unit vectors (normalized).
#' @param lighting A lighting specification object created by \code{light()}
#' @param face_centers Matrix with 3 columns (x, y, z coordinates) representing
#'   the center position of each face in data coordinate space. Required for
#'   positional lighting, optional for directional lighting.
#' @return Vector of lighting values. For most methods, returns numeric values.
#'   For \code{method = "rgb"}, returns hex color strings with \code{I()}
#'   class for identity scaling.
#' @keywords internal
#' @noRd
compute_light <- function(normals, lighting, face_centers = NULL) {

      # Validate inputs
      if (!is.matrix(normals) || ncol(normals) != 3) {
            stop("normals must be a matrix with 3 columns (x, y, z)")
      }

      params <- lighting

      # Validate resolved parameters
      valid_light <- c("direct", "diffuse", "rgb", "normal_x", "normal_y", "normal_z")
      if (!params$method %in% valid_light) {
            stop("lighting method must be one of: ", paste(valid_light, collapse = ", "))
      }

      # Check if positional lighting is requested
      use_positional <- !is.null(params$position)

      if (use_positional) {
            # Positional lighting: need face centers
            if (is.null(face_centers)) {
                  stop("face_centers required for positional lighting")
            }
            if (!is.matrix(face_centers) || ncol(face_centers) != 3 || nrow(face_centers) != nrow(normals)) {
                  stop("face_centers must be a matrix with 3 columns and same number of rows as normals")
            }

            # Calculate per-face light directions
            light_directions <- calculate_positional_light_directions(face_centers, params$position)

            # Calculate per-face dot products
            dot_products <- rowSums(normals * light_directions)

            # Apply distance falloff if requested
            if (params$distance_falloff) {
                  distances <- sqrt(rowSums((face_centers - matrix(rep(params$position, nrow(face_centers)),
                                                                   nrow = nrow(face_centers), byrow = TRUE))^2))
                  # Prevent division by zero and extreme values
                  distances <- pmax(distances, 0.1)
                  falloff_factor <- 1 / (distances^2)
                  # Normalize falloff to prevent extreme values
                  falloff_factor <- falloff_factor / max(falloff_factor)

                  # For diffuse lighting, interpolate between -1 (dark) and full lighting value
                  if (params$method == "diffuse") {
                        dot_products <- -1 + falloff_factor * (dot_products + 1)
                  } else {
                        # For other methods, scale toward appropriate dark value
                        if (params$method == "direct") {
                              # Direct: interpolate between 0 (dark) and full value
                              dot_products <- dot_products * falloff_factor
                        } else {
                              # General case: scale the lighting values
                              dot_products <- dot_products * falloff_factor
                        }
                  }
            }

      } else {
            # Directional lighting: uniform direction for all faces
            if (!is.numeric(params$direction) || length(params$direction) != 3) {
                  stop("light direction must be a numeric vector of length 3")
            }

            # Normalize light direction
            light_dir_norm <- normalize_light_direction(params$direction)

            # Compute dot products between normals and light direction
            dot_products <- compute_light_dot_products(normals, light_dir_norm)
      }

      # Apply lighting model (same for both directional and positional)
      light <- switch(params$method,
                      direct = pmax(0, dot_products),             # Direct lighting with hard shadows
                      diffuse = dot_products,                     # Atmospheric lighting: full -1 to +1 range
                      rgb = {                              # Map normals to RGB hex colors
                            if (use_positional) {
                                  # For positional lighting, use average light direction for RGB mapping
                                  avg_light_dir <- colMeans(light_directions)
                                  compute_rgb_light(normals, avg_light_dir)
                            } else {
                                  light_dir_norm <- normalize_light_direction(params$direction)
                                  compute_rgb_light(normals, light_dir_norm)
                            }
                      },
                      normal_x = (normals[,1] + 1) / 2,           # X-normal as color (0-1 range)
                      normal_y = (normals[,2] + 1) / 2,           # Y-normal as color (0-1 range)
                      normal_z = (normals[,3] + 1) / 2,           # Z-normal as color (0-1 range)
                      stop("Unknown lighting method")
      )

      # Apply quantization if requested
      if (!is.null(params$quanta) && !params$method %in% c("rgb", "normal_x", "normal_y", "normal_z")) {
            light <- apply_quantization(light, params$method, params$quanta)
      }

      # Auto-wrap RGB colors with I() for identity scaling
      if (params$method == "rgb") {
            light <- I(light)
      }

      return(light)
}

#' Calculate per-face light directions for positional lighting
#'
#' @param face_centers Matrix with 3 columns (x, y, z face center coordinates)
#' @param light_position Numeric vector of length 3 (x, y, z light position)
#' @return Matrix with 3 columns (normalized light direction vectors)
#' @keywords internal
#' @noRd
calculate_positional_light_directions <- function(face_centers, light_position) {
      # Calculate light directions from each face center TO the light position
      light_vectors <- matrix(rep(light_position, nrow(face_centers)),
                              nrow = nrow(face_centers), byrow = TRUE) - face_centers

      # Normalize each light direction vector
      light_lengths <- sqrt(rowSums(light_vectors^2))
      # Prevent division by zero
      light_lengths <- pmax(light_lengths, 1e-10)

      # Return normalized light directions
      light_vectors / light_lengths
}

#' Normalize light direction vector
#'
#' @param light_dir Numeric vector of length 3
#' @return Normalized light direction vector
#' @keywords internal
#' @noRd
normalize_light_direction <- function(light_dir) {
      light_dir / sqrt(sum(light_dir^2))
}

#' Compute dot products between normals and light direction
#'
#' @param normals Matrix with 3 columns (x, y, z normal components)
#' @param light_dir_norm Normalized light direction vector
#' @return Vector of dot products
#' @keywords internal
#' @noRd
compute_light_dot_products <- function(normals, light_dir_norm) {
      rowSums(normals * matrix(rep(light_dir_norm, nrow(normals)),
                               nrow = nrow(normals), byrow = TRUE))
}

#' Compute normal-to-RGB mapping with light direction rotation
#'
#' Maps surface normals to RGB colors with rotation so that normals aligned
#' with light_dir map to white/bright colors.
#'
#' @param normals Matrix with 3 columns (x, y, z normal components)
#' @param light_dir_norm Normalized light direction vector
#' @return Character vector of hex color codes
#' @keywords internal
#' @noRd
compute_rgb_light <- function(normals, light_dir_norm) {

      # Target direction for white color in RGB space
      white_dir <- c(1, 1, 1) / sqrt(3)

      # Create rotation matrix using Rodrigues' formula
      # We want to rotate FROM white_dir TO light_dir_norm
      # So we can apply the inverse rotation to normals

      # Cross product for rotation axis (white_dir × light_dir_norm)
      v <- c(
            white_dir[2] * light_dir_norm[3] - white_dir[3] * light_dir_norm[2],
            white_dir[3] * light_dir_norm[1] - white_dir[1] * light_dir_norm[3],
            white_dir[1] * light_dir_norm[2] - white_dir[2] * light_dir_norm[1]
      )

      s <- sqrt(sum(v^2))  # sine of angle
      c <- sum(white_dir * light_dir_norm)  # cosine of angle

      if (s < 1e-10) {
            # Vectors are already aligned or opposite
            if (c > 0) {
                  rot_matrix <- diag(3)  # Identity matrix
            } else {
                  # 180 degree rotation - use any perpendicular axis
                  rot_matrix <- -diag(3)
            }
      } else {
            # Rodrigues' rotation formula
            # R = I + [v]× + [v]×^2 * (1-c)/s^2
            v_cross <- matrix(c(
                  0, -v[3], v[2],
                  v[3], 0, -v[1],
                  -v[2], v[1], 0
            ), nrow = 3, byrow = TRUE)

            rot_matrix <- diag(3) + v_cross + v_cross %*% v_cross * ((1 - c) / s^2)
      }

      # Apply rotation to normals
      rotated_normals <- normals %*% rot_matrix

      # Scale normals from [-1,1] to [0,1] for RGB
      r <- (rotated_normals[,1] + 1) / 2
      g <- (rotated_normals[,2] + 1) / 2
      b <- (rotated_normals[,3] + 1) / 2

      # Convert to hex colors
      rgb(r, g, b)
}


#' Compute lighting from standardized coordinates in coord_3d
#'
#' @param data Original data frame with lighting_spec column
#' @param standardized_coords Data frame with standardized x, y, z coordinates
#' @param scale_ranges List with x, y, z scale ranges
#' @param scales Character, either "free" or "fixed"
#' @param ratio Numeric vector of length 3
#' @param proj List with pitch, roll, yaw, persp, dist (projection parameters)
#' @return data frame with lighting values and normal components added
#' @keywords internal
#' @noRd
compute_light_in_coord <- function(data, standardized_coords, scale_ranges, scales, ratio, proj = NULL) {

      # Extract lighting specification from first row (all rows have same spec)
      light <- data$lighting_spec[[1]]

      # Transform light position from data space to visual space, if applicable
      light$position <- transform_light_position(light$position, scale_ranges, scales, ratio)

      # Camera-frame directional lighting keeps lighting geometry in scene space and
      # rotates the finished face normals into camera space (see rotate_normals_to_camera,
      # applied per branch below). Positional lighting stays entirely in scene space so
      # that face centers and the light position share one coordinate frame.
      camera_directional <- light$anchor == "camera" && !is.null(proj) && is.null(light$position)

      # Filter to polygon groups only — points and segments don't have
      # meaningful surface normals and should not receive lighting
      has_prim <- ".prim" %in% names(data)
      if (has_prim) {
            poly_mask <- is.na(data$.prim) | data$.prim == "polygon"
      } else {
            # Count vertices per group to identify polygons (3+ vertices)
            grp_sizes <- ave(seq_len(nrow(data)), data$group, FUN = length)
            poly_mask <- grp_sizes >= 3
      }

      if (!any(poly_mask)) {
            # No polygon data to light — return unchanged
            return(data)
      }

      poly_data <- data[poly_mask, , drop = FALSE]
      poly_coords <- standardized_coords[poly_mask, , drop = FALSE]

      # Replace original coordinates with standardized ones for lighting computation
      poly_data$x <- poly_coords$x
      poly_data$y <- poly_coords$y
      poly_data$z <- poly_coords$z

      # Get unique faces for normal/lighting computation
      faces <- poly_data %>%
            group_by(group) %>%
            slice(1) %>%  # One row per face
            ungroup()

      # Compute normals based on data type
      has_precomputed_normals <- all(c("normal_x", "normal_y", "normal_z") %in% names(faces))

      if (has_precomputed_normals) {
            # Pre-existing normals (e.g. from text stat): transform to standardized space
            normals <- transform_normals_to_standard(
                  as.matrix(faces[, c("normal_x", "normal_y", "normal_z")]),
                  scale_ranges, scales, ratio, light$anchor, proj
            )
      } else if (grepl("sh3d__hull", faces$group[1])) {
            # Hull/triangle data: compute normals from vertex coordinates
            normals <- compute_triangle_normals(poly_data, faces)
            if (camera_directional) normals <- rotate_normals_to_camera(normals, proj)
      } else if ("face_type" %in% names(faces)) {
            # Voxel/column data: axis-aligned normals (scene space)
            normals <- compute_axis_aligned_normals(faces)
            if (camera_directional) normals <- rotate_normals_to_camera(normals, proj)
      } else {
            # Surface-like data: normals from fitted plane gradients. The gradient fit
            # and (-dzdx, -dzdy, 1) normal are computed in scene space, where the surface
            # is single-valued and the +z orientation is a consistent "top" normal.
            gradients <- compute_surface_gradients_from_vertices(poly_data)
            normals <- compute_surface_normals(gradients)
            if (camera_directional) normals <- rotate_normals_to_camera(normals, proj)
      }

      # Compute face centers from standardized coordinates
      face_centers <- calculate_face_centers(poly_data)

      # Apply lighting models
      faces <- apply_surface_light(faces, normals, face_centers, light)

      # Remove pre-existing normal columns from data before join to prevent suffixes
      poly_data <- poly_data[, !names(poly_data) %in% c("normal_x", "normal_y", "normal_z")]

      # Restore original coordinates before join (standardized were temporary)
      poly_data$x <- data$x[poly_mask]
      poly_data$y <- data$y[poly_mask]
      poly_data$z <- data$z[poly_mask]

      # Merge face-level lighting vars back into vertex-level polygon data
      poly_data <- left_join(poly_data, faces, by = join_by(group))

      if (light$method == "rgb") {
            poly_data$light <- I(poly_data$light)
      }

      # Add lighting parameters for shade processing in geom
      poly_data$shade_enabled <- light$shade
      poly_data$shade_strength <- light$shade_strength
      poly_data$shade_mode <- light$shade_mode
      poly_data$lighting_method <- light$method

      # Recombine with non-polygon rows (which get no lighting columns)
      if (all(poly_mask)) {
            return(poly_data)
      }

      non_poly_data <- data[!poly_mask, , drop = FALSE]
      bind_rows(poly_data, non_poly_data)[order(c(which(poly_mask), which(!poly_mask))), ]
}


#' Transform pre-computed normals from data space to standardized lighting space
#'
#' Normals transform under non-uniform scaling by the inverse-transpose of
#' the scaling matrix. For the diagonal scaling used in standardization,
#' this means scaling each component by range_width / effective_ratio,
#' then renormalizing.
#'
#' @param normals Matrix with 3 columns (normal_x, normal_y, normal_z) in data space
#' @param scale_ranges List with x, y, z scale ranges
#' @param scales Character, either "free" or "fixed"
#' @param ratio Numeric vector of length 3
#' @param anchor Character, "scene" or "camera"
#' @param proj Projection parameters (needed for camera anchor rotation)
#' @return Matrix with 3 columns of transformed normals
#' @keywords internal
#' @noRd
transform_normals_to_standard <- function(normals, scale_ranges, scales, ratio, anchor, proj = NULL) {

      # Compute effective ratios (same logic as scale_to_standard multi-axis path)
      effective_ratios <- compute_effective_ratios(scale_ranges, scales, ratio)

      # Compute per-axis scale factors: standardization scales by effective_ratio / range_width
      # Normal transform uses inverse-transpose, which for diagonal = range_width / effective_ratio
      x_width <- diff(scale_ranges$x)
      y_width <- diff(scale_ranges$y)
      z_width <- diff(scale_ranges$z)
      if (x_width == 0 || is.na(x_width)) x_width <- 1
      if (y_width == 0 || is.na(y_width)) y_width <- 1
      if (z_width == 0 || is.na(z_width)) z_width <- 1

      inv_scale <- c(x_width / effective_ratios[1],
                     y_width / effective_ratios[2],
                     z_width / effective_ratios[3])

      # Apply inverse-transpose scaling
      normals[, 1] <- normals[, 1] * inv_scale[1]
      normals[, 2] <- normals[, 2] * inv_scale[2]
      normals[, 3] <- normals[, 3] * inv_scale[3]

      # Renormalize
      lengths <- sqrt(rowSums(normals^2))
      lengths <- pmax(lengths, 1e-10)
      normals <- normals / lengths

      # For camera-anchor lighting, apply the same transforms as the coordinates:
      # z-flip then rotation
      if (anchor == "camera" && !is.null(proj)) {
            normals[, 3] <- -normals[, 3]  # z-flip
            normals <- rotate_3d(normals, proj$pitch, proj$roll, proj$yaw)
      }

      return(normals)
}

compute_surface_gradients_from_vertices <- function(data) {
      # Compute both gradients in one pass using least squares normal equations
      gradients <- data %>%
            group_by(group) %>%
            summarise(
                  n_pts = n(),
                  # Sums needed for least squares: z = a*x + b*y + c
                  sum_x = sum(x), sum_y = sum(y), sum_z = sum(z),
                  sum_xx = sum(x*x), sum_yy = sum(y*y),
                  sum_xy = sum(x*y), sum_xz = sum(x*z), sum_yz = sum(y*z),
                  .groups = "drop"
            ) %>%
            mutate(
                  # Full bivariate plane fit: z = a*x + b*y + c
                  # Normal equations for centered data
                  Sxx = n_pts * sum_xx - sum_x^2,
                  Syy = n_pts * sum_yy - sum_y^2,
                  Sxy = n_pts * sum_xy - sum_x * sum_y,
                  Sxz = n_pts * sum_xz - sum_x * sum_z,
                  Syz = n_pts * sum_yz - sum_y * sum_z,
                  denom = Sxx * Syy - Sxy^2,
                  dzdx = ifelse(abs(denom) > 1e-10,
                                (Syy * Sxz - Sxy * Syz) / denom,
                                ifelse(abs(Sxx) > 1e-10, Sxz / Sxx, 0)),
                  dzdy = ifelse(abs(denom) > 1e-10,
                                (Sxx * Syz - Sxy * Sxz) / denom,
                                ifelse(abs(Syy) > 1e-10, Syz / Syy, 0))
            ) %>%
            select(group, dzdx, dzdy)

      return(gradients)
}


#' Rotate scene-space face normals into camera space
#'
#' For camera-anchored directional lighting, face normals are computed in scene
#' space (where surface orientation is well-defined) and then transformed into
#' camera space with the same operations applied to coordinates during projection:
#' a z-flip followed by the projection rotation. This mirrors the camera handling
#' in transform_normals_to_standard() so that all lit geoms share one convention
#' for entering camera space.
#'
#' @param normals Matrix with 3 columns (x, y, z normal components) in scene space
#' @param proj List with pitch, roll, yaw (projection parameters)
#' @return Matrix with 3 columns of camera-space normals
#' @keywords internal
#' @noRd
rotate_normals_to_camera <- function(normals, proj) {
      normals[, 3] <- -normals[, 3]  # z-flip, matching transform_3d_standard
      rotate_3d(normals, proj$pitch, proj$roll, proj$yaw)
}

#' Compute surface normals from gradients
#'
#' @param gradients Data frame with group, dzdx, dzdy columns
#' @return Matrix with 3 columns (x, y, z normal components)
#' @keywords internal
#' @noRd
compute_surface_normals <- function(gradients) {
      # Compute raw normals from gradients
      # For a surface z = f(x,y), the normal is (-dz/dx, -dz/dy, 1) (unnormalized)
      raw_normals <- cbind(
            -gradients$dzdx,
            -gradients$dzdy,
            rep(1, nrow(gradients))
      )

      # Normalize each normal vector
      lengths <- sqrt(rowSums(raw_normals^2))
      normals <- raw_normals / lengths

      return(normals)
}

#' Compute triangle normals from vertex coordinates
#'
#' @param data Full vertex data with x, y, z coordinates
#' @param faces Data frame with one row per face (group column required)
#' @return Matrix with 3 columns (x, y, z normal components)
#' @keywords internal
#' @noRd
compute_triangle_normals <- function(data, faces) {
      # For each face, get three vertices and compute cross product
      face_normals <- data %>%
            group_by(group) %>%
            slice(1:3) %>%
            summarise(
                  # Get first three vertices
                  x1 = x[1], y1 = y[1], z1 = z[1],
                  x2 = x[2], y2 = y[2], z2 = z[2],
                  x3 = x[3], y3 = y[3], z3 = z[3],
                  .groups = "drop"
            ) %>%
            mutate(
                  # Edge vectors
                  e1x = x2 - x1, e1y = y2 - y1, e1z = z2 - z1,
                  e2x = x3 - x1, e2y = y3 - y1, e2z = z3 - z1,
                  # Cross product (e1 × e2)
                  normal_x = e1y * e2z - e1z * e2y,
                  normal_y = e1z * e2x - e1x * e2z,
                  normal_z = e1x * e2y - e1y * e2x,
                  # Normalize
                  length = sqrt(normal_x^2 + normal_y^2 + normal_z^2),
                  normal_x = ifelse(length > 0, normal_x / length, 0),
                  normal_y = ifelse(length > 0, normal_y / length, 0),
                  normal_z = ifelse(length > 0, normal_z / length, 1)
            )

      # Return as matrix format expected by lighting functions
      normals <- as.matrix(face_normals[, c("normal_x", "normal_y", "normal_z")])

      return(normals)
}

#' Compute axis-aligned normals for voxel/column faces
#'
#' @param faces Data frame with face_type column indicating face orientation
#' @return Matrix with 3 columns (x, y, z normal components)
#' @keywords internal
#' @noRd
compute_axis_aligned_normals <- function(faces) {
      # Map face types to normal directions
      face_normals <- faces %>%
            mutate(
                  normal_x = case_when(
                        face_type == "xmin" ~ -1,
                        face_type == "xmax" ~ 1,
                        TRUE ~ 0
                  ),
                  normal_y = case_when(
                        face_type == "ymin" ~ -1,
                        face_type == "ymax" ~ 1,
                        TRUE ~ 0
                  ),
                  normal_z = case_when(
                        face_type == "zmin" ~ -1,
                        face_type == "zmax" ~ 1,
                        TRUE ~ 0
                  ),
                  # Handle NA face_type (shouldn't happen but be safe)
                  normal_x = ifelse(is.na(normal_x), 0, normal_x),
                  normal_y = ifelse(is.na(normal_y), 0, normal_y),
                  normal_z = ifelse(is.na(normal_z), 1, normal_z)
            )

      # Return as matrix format expected by lighting functions
      normals <- as.matrix(face_normals[, c("normal_x", "normal_y", "normal_z")])

      return(normals)
}

#' Calculate face centers from vertex data
#'
#' @param data Data frame with x, y, z coordinates and group column
#' @return Matrix with 3 columns (x, y, z face center coordinates)
#' @keywords internal
#' @noRd
calculate_face_centers <- function(data) {
      centers <- data %>%
            group_by(group) %>%
            summarise(
                  center_x = mean(x, na.rm = TRUE),
                  center_y = mean(y, na.rm = TRUE),
                  center_z = mean(z, na.rm = TRUE),
                  .groups = "drop"
            )

      as.matrix(centers[, c("center_x", "center_y", "center_z")])
}

#' Apply lighting models to surface normals
#'
#' @param face_data Data frame with unique faces (group column)
#' @param normals Matrix of surface normals
#' @param face_centers Matrix of face centers
#' @param light Lighting specification object
#' @return Data frame with lighting values and normal components
#' @keywords internal
#' @noRd
apply_surface_light <- function(face_data, normals, face_centers, light) {
      # Apply lighting models to the normals
      light_vals <- compute_light(normals, light, face_centers)

      face_data_with_light <- select(face_data, group) %>%
            mutate(light = light_vals,
                   normal_x = normals[, 1],
                   normal_y = normals[, 2],
                   normal_z = normals[, 3])

      return(face_data_with_light)
}


#' Convert RGB to HSL color space
#'
#' @param rgb_matrix 3xN matrix with RGB values in 0-1 range
#' @return 3xN matrix with HSL values (H in 0-1, S in 0-1, L in 0-1)
#' @keywords internal
#' @noRd
rgb2hsl <- function(rgb_matrix) {
      if (!is.matrix(rgb_matrix) || nrow(rgb_matrix) != 3) {
            stop("rgb_matrix must be a 3xN matrix")
      }

      r <- rgb_matrix[1, ]
      g <- rgb_matrix[2, ]
      b <- rgb_matrix[3, ]

      max_val <- pmax(r, g, b)
      min_val <- pmin(r, g, b)
      diff <- max_val - min_val

      # Lightness
      l <- (max_val + min_val) / 2

      # Initialize hue and saturation
      h <- rep(0, ncol(rgb_matrix))
      s <- rep(0, ncol(rgb_matrix))

      # Calculate saturation
      non_zero_diff <- diff > 0
      s[non_zero_diff] <- ifelse(
            l[non_zero_diff] < 0.5,
            diff[non_zero_diff] / (max_val[non_zero_diff] + min_val[non_zero_diff]),
            diff[non_zero_diff] / (2 - max_val[non_zero_diff] - min_val[non_zero_diff])
      )

      # Calculate hue
      if (any(non_zero_diff)) {
            # Red is max
            red_max <- non_zero_diff & (r == max_val)
            h[red_max] <- ((g[red_max] - b[red_max]) / diff[red_max]) %% 6

            # Green is max
            green_max <- non_zero_diff & (g == max_val)
            h[green_max] <- (b[green_max] - r[green_max]) / diff[green_max] + 2

            # Blue is max
            blue_max <- non_zero_diff & (b == max_val)
            h[blue_max] <- (r[blue_max] - g[blue_max]) / diff[blue_max] + 4

            # Normalize hue to [0, 1]
            h <- h / 6
      }

      rbind(h, s, l)
}

#' Convert HSL to RGB color space
#'
#' @param hsl_matrix 3xN matrix with HSL values (H in 0-1, S in 0-1, L in 0-1)
#' @return 3xN matrix with RGB values in 0-1 range
#' @keywords internal
#' @noRd
hsl2rgb <- function(hsl_matrix) {
      if (!is.matrix(hsl_matrix) || nrow(hsl_matrix) != 3) {
            stop("hsl_matrix must be a 3xN matrix")
      }

      h <- hsl_matrix[1, ]
      s <- hsl_matrix[2, ]
      l <- hsl_matrix[3, ]

      n <- ncol(hsl_matrix)
      r <- rep(0, n)
      g <- rep(0, n)
      b <- rep(0, n)

      for (i in seq_len(n)) {
            if (s[i] == 0) {
                  # Achromatic (gray)
                  r[i] <- g[i] <- b[i] <- l[i]
            } else {
                  q <- ifelse(l[i] < 0.5, l[i] * (1 + s[i]), l[i] + s[i] - l[i] * s[i])
                  p <- 2 * l[i] - q

                  hue_to_rgb <- function(p, q, t) {
                        if (t < 0) t <- t + 1
                        if (t > 1) t <- t - 1
                        if (t < 1/6) return(p + (q - p) * 6 * t)
                        if (t < 1/2) return(q)
                        if (t < 2/3) return(p + (q - p) * (2/3 - t) * 6)
                        return(p)
                  }

                  r[i] <- hue_to_rgb(p, q, h[i] + 1/3)
                  g[i] <- hue_to_rgb(p, q, h[i])
                  b[i] <- hue_to_rgb(p, q, h[i] - 1/3)
            }
      }

      rbind(r, g, b)
}


#' Apply shading to colors based on lighting values
#'
#' @param colors Vector of color values (hex or named colors)
#' @param light_values Numeric vector of lighting values
#' @param lighting Lighting specification with shade_mode and shade_strength
#' @return Vector of modified colors
#' @keywords internal
#' @noRd
blend_light_with_colors <- function(colors, light_values, lighting) {
      # Handle NA or NULL colors
      if (is.null(colors) || length(colors) == 0) {
            return(colors)
      }

      # RGB method produces colors directly - return them as-is
      if (lighting$method == "rgb") {
            return(I(light_values))
      }

      # Initialize result
      result_colors <- colors

      # Handle only valid colors (not NA)
      valid_mask <- !is.na(colors) & !is.na(light_values)

      if (!any(valid_mask)) {
            return(result_colors)
      }

      valid_base <- colors[valid_mask]
      valid_light <- light_values[valid_mask]

      # Normalize light values to [0, 1] if needed
      if (lighting$method == "diffuse") {
            valid_light <- (valid_light + 1) / 2
      }
      valid_light <- pmax(0, pmin(1, valid_light))

      # Convert to RGB matrix
      base_rgb <- col2rgb(valid_base)

      if (lighting$shade_mode == "hsl") {
            # HSL blending mode
            base_rgb_norm <- base_rgb / 255  # Normalize to [0,1] for HSL
            base_hsl <- rgb2hsl(base_rgb_norm)

            # Modify only lightness (L component)
            new_hsl <- base_hsl

            for (i in seq_along(valid_light)) {
                  h_val <- base_hsl[1, i]
                  s_val <- base_hsl[2, i]
                  l_val <- base_hsl[3, i]
                  light_val <- valid_light[i]

                  if (light_val > 0.5) {
                        # Brighten toward white (L = 1)
                        blend_factor <- (light_val - 0.5) * 2 * lighting$shade_strength
                        new_l <- l_val + blend_factor * (1 - l_val)
                  } else {
                        # Darken toward black (L = 0)
                        blend_factor <- (0.5 - light_val) * 2 * lighting$shade_strength
                        new_l <- l_val * (1 - blend_factor)
                  }

                  new_hsl[1, i] <- h_val  # Preserve hue
                  new_hsl[2, i] <- s_val  # Preserve saturation
                  new_hsl[3, i] <- pmax(0, pmin(1, new_l))  # Clamp lightness
            }

            # Convert back to RGB
            new_rgb <- hsl2rgb(new_hsl)
            new_colors <- rgb(new_rgb[1, ], new_rgb[2, ], new_rgb[3, ])

      } else {
            # HSV blending mode
            base_hsv <- rgb2hsv(base_rgb)

            # Modify only brightness (V component)
            new_hsv <- base_hsv

            for (i in seq_along(valid_light)) {
                  h_val <- base_hsv[1, i]
                  s_val <- base_hsv[2, i]
                  v_val <- base_hsv[3, i]
                  light_val <- valid_light[i]

                  # Handle NaN hue (gray colors)
                  if (is.nan(h_val)) h_val <- 0

                  if (light_val > 0.5) {
                        # Brighten
                        blend_factor <- (light_val - 0.5) * 2 * lighting$shade_strength
                        new_v <- v_val + blend_factor * (1 - v_val)
                  } else {
                        # Darken
                        blend_factor <- (0.5 - light_val) * 2 * lighting$shade_strength
                        new_v <- v_val * (1 - blend_factor)
                  }

                  new_hsv[1, i] <- h_val
                  new_hsv[2, i] <- s_val
                  new_hsv[3, i] <- pmax(0, pmin(1, new_v))
            }

            # Convert back to colors
            new_colors <- hsv(new_hsv[1, ], new_hsv[2, ], new_hsv[3, ])
      }

      result_colors[valid_mask] <- new_colors

      return(result_colors)
}


blend_light <- function(coords) {

      # Extract lighting parameters from special columns
      if ("shade_enabled" %in% names(coords)) {
            shade_val <- coords$shade_enabled[!is.na(coords$shade_enabled)][1]

            if (!is.na(shade_val) && shade_val != "neither") {
                  # Reconstruct lighting object from first non-NA row
                  light <- list(
                        shade = shade_val,
                        shade_strength = coords$shade_strength[!is.na(coords$shade_strength)][1],
                        shade_mode = coords$shade_mode[!is.na(coords$shade_mode)][1],
                        method = coords$lighting_method[!is.na(coords$lighting_method)][1]
                  )

                  # Apply shading if enabled
                  if (light$shade != "neither" && "light" %in% names(coords)) {
                        # Blend fill colors if requested
                        if (light$shade %in% c("fill", "both")) {
                              coords$fill <- blend_light_with_colors(coords$fill, coords$light, light)
                        }

                        # Blend colour/border colors if requested
                        if (light$shade %in% c("colour", "both")) {
                              coords$colour <- blend_light_with_colors(coords$colour, coords$light, light)
                        }
                  }
            }

            # Clean up lighting parameter columns
            coords <- coords[, !names(coords) %in% c("shade_enabled", "shade_strength", "shade_mode", "lighting_method")]
      }

      return(coords)
}

#' Convert light position from data units to visual space units
#'
#' @param position Numeric vector of length 3 (x, y, z in data coordinates)
#' @param scale_ranges List with x, y, z scale ranges
#' @param scales Character, either "free" or "fixed"
#' @param ratio Numeric vector of length 3
#' @return Position vector in visual space, or NULL if position is NULL
#' @keywords internal
#' @noRd
transform_light_position <- function(position, scale_ranges, scales, ratio) {
      if(is.null(position)) return(NULL)
      light_data <- data.frame(x = position[1], y = position[2], z = position[3])
      standardized <- scale_to_standard(light_data, scale_ranges, scales, ratio)
      return(c(standardized$x, standardized$y, standardized$z))
}

#' Resolve a lighting specification to its final value
#'
#' `waiver()` signals that no lighting was specified anywhere, in which case the
#' package default applies. `NULL` is a deliberate request for no lighting and is
#' preserved as-is.
#'
#' @param light A `light` object, `"none"`, `NULL`, or `waiver()`.
#' @return A `light` object, `"none"`, or `NULL`.
#' @keywords internal
#' @noRd
resolve_light <- function(light) {
      if (inherits(light, "waiver")) light() else light
}

#' Store lighting specification to data frame for downstream use
#'
#' @param data Data frame to attach lighting spec to
#' @param light Lighting specification object, "none", NULL, or waiver()
#' @return Data frame with lighting_spec column added (if light is not NULL)
#' @keywords internal
#' @noRd
attach_light <- function(data, light){
      light <- resolve_light(light)
      if(!is.null(light)){
            if(!inherits(light, "light") && identical(light, "none")) light <- light("none")
            if(! "lighting_spec" %in% names(data)) data$lighting_spec <- I(list(light))
      }

      data
}


# Plot-level lighting via `+ light()` -------------------------------------------

#' Take ownership of a plot's coord before mutating it
#'
#' `Coord` ggproto objects have reference semantics and are *not* deep-copied by
#' ggplot2's `plot_clone()`, so one coord object can be shared by several plots.
#' Mutating it in place would leak into every plot that shares it. This clones
#' once, marks the clone, and mutates in place on subsequent calls.
#'
#' The clone must be bound to a *new* name. `ggproto(NULL, parent)` captures
#' `parent` lazily, so `co <- ggproto(NULL, co)` makes the clone its own parent
#' and any inherited field lookup then recurses until the C stack overflows.
#'
#' @param plot A ggplot object whose `coordinates` is a `Coord3D`.
#' @return The plot, with a coord that is safe to mutate.
#' @keywords internal
#' @noRd
own_coord_3d <- function(plot) {
      co <- plot$coordinates
      if (!isTRUE(co$.ggcube_owned)) {
            new_co <- ggproto(NULL, co)
            new_co$.ggcube_owned <- TRUE
            plot$coordinates <- new_co
      }
      plot
}

#' Signal that lighting was specified in two places
#'
#' @return Throws an error.
#' @keywords internal
#' @noRd
abort_double_light <- function() {
      rlang::abort(c(
            "Lighting was specified twice.",
            "x" = "Found lighting in both `coord_3d(light = )` and `+ light()`.",
            "i" = "Supply it in one place only; `+ light()` is the recommended form."
      ), call = NULL)
}

#' Find a plot-level light stashed by `+ light()`
#'
#' `+ light()` added before `coord_3d()` has no coord to write to, so it stashes
#' the spec on the plot instead. `Coord3D$setup_panel_params()` receives only
#' scales and params, so the stash is recovered by reaching the plot itself.
#'
#' @return A `light` object, or `NULL` if none was stashed.
#' @keywords internal
#' @noRd
find_pending_light <- function() {
      p <- find_build_plot()
      if (is.null(p)) return(NULL)
      attr(p, "ggcube_pending_light")
}

#' Report that plot-level lighting is being replaced
#'
#' Mirrors ggplot2's behaviour when a coord or scale replaces an existing one:
#' a component that is discarded wholesale is worth reporting, since
#' `+ light(direction = ) + light(mode = )` looks like it accumulates settings
#' but in fact keeps only the second.
#'
#' @return Called for its side effect.
#' @keywords internal
#' @noRd
inform_light_replaced <- function() {
      rlang::inform(c(
            "Lighting is already present.",
            i = "Adding new lighting, which will replace the existing specification."
      ))
}

#' @export
ggplot_add.light <- function(object, plot, ...) {
      if (inherits(plot$coordinates, "Coord3D")) {
            if (isTRUE(plot$coordinates$light_explicit)) abort_double_light()
            # A light already written to the coord by an earlier `+ light()` is
            # about to be discarded. An unresolved `waiver()` is not a
            # replacement: it means no lighting has been specified yet.
            if (!inherits(plot$coordinates$light, "waiver")) inform_light_replaced()
            plot <- own_coord_3d(plot)
            plot$coordinates$light <- object
      } else {
            # No coord yet. Stash on the plot; Coord3D$setup_panel_params()
            # picks it up at build time, so `+ light()` is order-independent.
            if (!is.null(attr(plot, "ggcube_pending_light"))) inform_light_replaced()
            attr(plot, "ggcube_pending_light") <- object
      }
      plot
}

test_that("text_outlines produces net-CCW winding per glyph", {
      # Helper: signed area via shoelace formula (positive = CCW)
      signed_area <- function(x, y) {
            sum(x * dplyr::lead(y, default = y[1]) -
                      dplyr::lead(x, default = x[1]) * y)
      }

      # Word with holes (letters like "a", "o" have interior contours)
      # Individual hole contours may be CW (negative area), but the net
      # signed area per glyph (outer + holes) should be positive (CCW dominant)
      outlines_holes <- text_outlines("ao", family = "sans", size = 24)
      net_areas_holes <- outlines_holes %>%
            dplyr::group_by(poly_id) %>%
            dplyr::summarise(area = signed_area(x, y), .groups = "drop") %>%
            dplyr::mutate(glyph = sub("_.*", "", poly_id)) %>%
            dplyr::group_by(glyph) %>%
            dplyr::summarise(net_area = sum(area), .groups = "drop")
      expect_true(all(net_areas_holes$net_area > 0),
                  info = "Net signed area per glyph in 'ao' should be positive (CCW dominant)")

      # Word without holes — all contours should be CCW
      outlines_no_holes <- text_outlines("xx", family = "sans", size = 24)
      areas_no_holes <- outlines_no_holes %>%
            dplyr::group_by(poly_id) %>%
            dplyr::summarise(area = signed_area(x, y), .groups = "drop")
      expect_true(all(areas_no_holes$area > 0),
                  info = "All contours in 'xx' should be CCW (positive signed area)")

      # Multi-word: net area per glyph should be positive
      outlines_mixed <- text_outlines("Hello World", family = "sans", size = 24)
      net_areas_mixed <- outlines_mixed %>%
            dplyr::group_by(poly_id) %>%
            dplyr::summarise(area = signed_area(x, y), .groups = "drop") %>%
            dplyr::mutate(glyph = sub("_.*", "", poly_id)) %>%
            dplyr::group_by(glyph) %>%
            dplyr::summarise(net_area = sum(area), .groups = "drop")
      expect_true(all(net_areas_mixed$net_area > 0),
                  info = "Net signed area per glyph in 'Hello World' should be positive")
})


test_that("text_outlines outer contours are CCW", {
      signed_area <- function(x, y) {
            sum(x * dplyr::lead(y, default = y[1]) -
                      dplyr::lead(x, default = x[1]) * y)
      }

      # For letters with holes, the outer contour should have the largest
      # absolute area, and it should be positive (CCW)
      outlines <- text_outlines("ao", family = "sans", size = 24)
      contour_areas <- outlines %>%
            dplyr::group_by(poly_id) %>%
            dplyr::summarise(area = signed_area(x, y), .groups = "drop") %>%
            dplyr::mutate(glyph = sub("_.*", "", poly_id))

      # Within each glyph, the contour with largest absolute area (the outer)
      # should be positive
      outer_areas <- contour_areas %>%
            dplyr::group_by(glyph) %>%
            dplyr::slice_max(abs(area), n = 1, with_ties = FALSE) %>%
            dplyr::ungroup()
      expect_true(all(outer_areas$area > 0),
                  info = "Outer contours should be CCW (positive signed area)")
})


test_that("transform_normals_to_standard preserves axis-aligned normals under uniform scaling", {
      normals <- matrix(c(0, 0, 1), nrow = 1)
      scale_ranges <- list(x = c(0, 10), y = c(0, 10), z = c(0, 10))

      result <- ggcube:::transform_normals_to_standard(
            normals, scale_ranges, scales = "free", ratio = c(1, 1, 1),
            anchor = "scene", proj = NULL
      )

      expect_equal(result[1, ], c(0, 0, 1), tolerance = 1e-10)
})


test_that("transform_normals_to_standard applies inverse-transpose under non-uniform scaling", {
      normals <- matrix(c(1, 0, 0), nrow = 1)
      scale_ranges <- list(x = c(0, 10), y = c(0, 20), z = c(0, 10))

      result <- ggcube:::transform_normals_to_standard(
            normals, scale_ranges, scales = "free", ratio = c(1, 1, 1),
            anchor = "scene", proj = NULL
      )

      # Axis-aligned normal should remain axis-aligned regardless of scaling
      expect_equal(result[1, ], c(1, 0, 0), tolerance = 1e-10)

      # Diagonal normal should rotate under non-uniform scaling
      normals_diag <- matrix(c(1, 1, 0) / sqrt(2), nrow = 1)
      result_diag <- ggcube:::transform_normals_to_standard(
            normals_diag, scale_ranges, scales = "free", ratio = c(1, 1, 1),
            anchor = "scene", proj = NULL
      )

      # x gets scaled by 10, y by 20, so inverse-transpose scales x by 10, y by 20
      # After normalization: (10, 20, 0) / sqrt(500)
      expected <- c(10, 20, 0) / sqrt(10^2 + 20^2)
      expect_equal(result_diag[1, ], expected, tolerance = 1e-6)
})


test_that("transform_normals_to_standard handles fixed scales", {
      normals <- matrix(c(0, 0, 1), nrow = 1)
      scale_ranges <- list(x = c(0, 10), y = c(0, 5), z = c(0, 20))

      result <- ggcube:::transform_normals_to_standard(
            normals, scale_ranges, scales = "fixed", ratio = c(1, 1, 1),
            anchor = "scene", proj = NULL
      )

      expect_equal(result[1, 1], 0, tolerance = 1e-10)
      expect_equal(result[1, 2], 0, tolerance = 1e-10)
      expect_equal(abs(result[1, 3]), 1, tolerance = 1e-10)
})


test_that("transform_normals_to_standard applies camera rotation", {
      normals <- matrix(c(0, 0, 1), nrow = 1)
      scale_ranges <- list(x = c(0, 10), y = c(0, 10), z = c(0, 10))
      proj <- list(pitch = 0, roll = 0, yaw = 0)

      result_scene <- ggcube:::transform_normals_to_standard(
            normals, scale_ranges, scales = "free", ratio = c(1, 1, 1),
            anchor = "scene", proj = proj
      )

      result_camera <- ggcube:::transform_normals_to_standard(
            normals, scale_ranges, scales = "free", ratio = c(1, 1, 1),
            anchor = "camera", proj = proj
      )

      expect_equal(result_scene[1, ], c(0, 0, 1), tolerance = 1e-10)
      expect_equal(result_camera[1, ], c(0, 0, -1), tolerance = 1e-10)
})


test_that("process_backfaces correctly identifies frontfaces with .subgroup", {
      outer <- data.frame(
            x = c(0, 1, 1, 0),
            y = c(0, 0, 1, 1),
            group = "g1",
            .subgroup = "g1_outer"
      )
      hole <- data.frame(
            x = c(0.25, 0.75, 0.75, 0.25),
            y = c(0.25, 0.25, 0.75, 0.75),
            group = "g1",
            .subgroup = "g1_hole"
      )
      data <- rbind(outer, hole)

      result <- ggcube:::process_backfaces(data)

      expect_equal(nrow(result), 8)
      expect_false(".backface" %in% names(result))
})


test_that("process_backfaces correctly identifies backfaces with .subgroup", {
      outer <- data.frame(
            x = c(0, 0, 1, 1),
            y = c(0, 1, 1, 0),
            group = "g1",
            .subgroup = "g1_outer"
      )
      hole <- data.frame(
            x = c(0.25, 0.25, 0.75, 0.75),
            y = c(0.25, 0.75, 0.75, 0.25),
            group = "g1",
            .subgroup = "g1_hole"
      )
      data <- rbind(outer, hole)
      data$cull_backfaces <- TRUE

      result <- ggcube:::process_backfaces(data)

      expect_equal(nrow(result), 0)
})


test_that("process_backfaces works without .subgroup column", {
      data <- data.frame(
            x = c(0, 1, 0.5),
            y = c(0, 0, 1),
            group = "g1"
      )

      result <- ggcube:::process_backfaces(data)
      expect_equal(nrow(result), 3)

      data_cw <- data.frame(
            x = c(0, 0.5, 1),
            y = c(0, 1, 0),
            group = "g1",
            cull_backfaces = TRUE
      )

      result_cw <- ggcube:::process_backfaces(data_cw)
      expect_equal(nrow(result_cw), 0)
})


test_that("process_backfaces handles mixed front/back groups", {
      front <- data.frame(
            x = c(0, 1, 1, 0),
            y = c(0, 0, 1, 1),
            group = "front",
            .subgroup = "front_1"
      )
      back <- data.frame(
            x = c(2, 2, 3, 3),
            y = c(0, 1, 1, 0),
            group = "back",
            .subgroup = "back_1"
      )
      data <- rbind(front, back)
      data$cull_backfaces <- TRUE

      result <- ggcube:::process_backfaces(data)

      expect_equal(nrow(result), 4)
      expect_true(all(result$group == "front"))
})

# Tests for the lighting pipeline, focused on camera-anchor normal orientation.
#
# These verify the *analytic* invariant that a face's computed light equals the dot
# product of its correctly-oriented normal with the light direction, rather than
# snapshotting rendered pixels. The bugs these guard against all produced
# plausible-looking output, so the value is in checking orientation against ground
# truth, not against a frozen image.

# Extract per-face light and metadata from a built plot, after the full coord
# transform (projection, npc scaling, and backface handling) has run.
built_faces <- function(p) {
      b <- ggplot_build(p)
      d <- b$data[[1]]
      coord <- b$plot$coordinates
      pp <- b$layout$panel_params[[1]]
      tr <- coord$transform(d, pp)
      list(tr = tr, proj = pp$proj)
}

# Ground-truth light for an axis-aligned face under camera-anchor directional
# lighting: rotate the scene-space normal into camera space (z-flip + projection
# rotation), then dot with the normalised direction.
axis_normal <- function(face_type) {
      switch(face_type,
             xmin = c(-1, 0, 0), xmax = c(1, 0, 0),
             ymin = c(0, -1, 0), ymax = c(0, 1, 0),
             zmin = c(0, 0, -1), zmax = c(0, 0, 1),
             c(0, 0, 1))
}


test_that("rotate_normals_to_camera matches the coordinate transform convention", {
      proj <- list(pitch = 10, roll = -25, yaw = 60)
      n <- matrix(c(1, 0, 0,
                    0, 1, 0,
                    0, 0, 1,
                    -1, 0, 0), ncol = 3, byrow = TRUE)

      rotated <- rotate_normals_to_camera(n, proj)

      # Expected: z-flip then rotate_3d, identical to how coordinates enter camera space.
      expected <- n
      expected[, 3] <- -expected[, 3]
      expected <- rotate_3d(expected, proj$pitch, proj$roll, proj$yaw)

      expect_equal(rotated, expected)

      # Rotation preserves unit length.
      expect_equal(sqrt(rowSums(rotated^2)), rep(1, nrow(rotated)))
})


test_that("voxel camera-anchor lighting matches camera-space ground truth", {
      dir <- c(1, 1, 0)
      dirn <- dir / sqrt(sum(dir^2))

      for (cfg in list(c(60, -25, 10), c(30, 0, 0), c(-45, 20, -15))) {
            p <- ggplot(iris, aes(Species, Sepal.Length, fill = Species)) +
                  geom_bar_3d(bins = 20, width = c(.5, 1)) +
                  coord_3d(scales = "fixed", ratio = c(1, 3, .25),
                           yaw = cfg[1], roll = cfg[2], pitch = cfg[3],
                           light = light(mode = "hsl", anchor = "camera",
                                         direction = dir)) +
                  scale_z_continuous(expand = c(0, 0))

            bf <- built_faces(p)
            faces <- dplyr::summarise(dplyr::group_by(bf$tr, group),
                                      face_type = face_type[1],
                                      light = as.numeric(light)[1],
                                      .groups = "drop")

            N <- t(vapply(faces$face_type, axis_normal, numeric(3)))
            N[, 3] <- -N[, 3]
            Ncam <- rotate_3d(N, bf$proj$pitch, bf$proj$roll, bf$proj$yaw)
            truth <- as.numeric(Ncam %*% dirn)

            expect_equal(faces$light, truth,
                         tolerance = 1e-6,
                         info = paste("yaw/roll/pitch =",
                                      paste(cfg, collapse = "/")))
      }
})


test_that("voxel scene-anchor lighting is unchanged by the camera path", {
      dir <- c(1, 1, 0)
      dirn <- dir / sqrt(sum(dir^2))

      p <- ggplot(iris, aes(Species, Sepal.Length, fill = Species)) +
            geom_bar_3d(bins = 20, width = c(.5, 1)) +
            coord_3d(scales = "fixed", ratio = c(1, 3, .25),
                     yaw = 60, roll = -25, pitch = 10,
                     light = light(mode = "hsl", anchor = "scene",
                                   direction = dir)) +
            scale_z_continuous(expand = c(0, 0))

      bf <- built_faces(p)
      faces <- dplyr::summarise(dplyr::group_by(bf$tr, group),
                                face_type = face_type[1],
                                light = as.numeric(light)[1],
                                .groups = "drop")

      # Scene anchor: normals stay in scene space, direction applied directly.
      N <- t(vapply(faces$face_type, axis_normal, numeric(3)))
      truth <- as.numeric(N %*% dirn)

      expect_equal(faces$light, truth, tolerance = 1e-6)
})


# Note on hulls: geom_hull_3d also routes through rotate_normals_to_camera for
# camera-anchor lighting. Before the fix its triangle normals were computed via a
# coordinate z-flip (a reflection, which negates the cross-product normal) rather
# than a normal z-flip, leaving hull normals inverted. For a closed convex hull this
# inversion is fully masked by backface handling -- every visible face's polarity is
# flipped back -- so the final rendered light is identical with or without the fix.
# There is therefore no observable behaviour to pin with a regression test; the fix
# is an internal-consistency improvement so that hull normals share the single
# camera-space convention. The helper it relies on is covered by the
# rotate_normals_to_camera test above.


test_that("surface camera-anchor normals are oriented, not sign-forced", {
      # Reproduces the original white-patch report. Root cause: compute_surface_normals
      # builds (-dzdx, -dzdy, +1), forcing the camera-space normal z-component positive
      # for every face. That made faces curving past perpendicular to the sight line
      # receive frontface lighting, producing a bright patch where they should be dark.
      #
      # Once normals are computed in scene space and rotated into camera space, faces
      # whose top points away from the viewer have negative camera-space normal_z. The
      # pre-fix code could never produce this; its normal_z was clamped non-negative.
      p <- ggplot() +
            geom_function_3d(
                  fun = function(x, y) sin(x) * cos(y),
                  xlim = c(-pi, pi), ylim = c(-pi, pi),
                  fill = "steelblue", color = "steelblue") +
            coord_3d(light = light(mode = "hsl", anchor = "camera",
                                   direction = c(1, 1, 0)),
                     yaw = -30, roll = -40, pitch = 0)

      b <- ggplot_build(p)
      d <- b$data[[1]]
      coord <- b$plot$coordinates
      pp <- b$layout$panel_params[[1]]
      tr <- coord$transform(d, pp)

      faces <- dplyr::summarise(dplyr::group_by(tr, group),
                                normal_z = normal_z[1],
                                light = as.numeric(light)[1],
                                .groups = "drop")

      # Directly catches the +1 forcing: a rotated surface at this angle must have
      # faces whose camera-space normal points away from the viewer.
      expect_true(any(faces$normal_z < 0),
                  info = "no away-facing faces; normal_z appears sign-forced")

      # Backface handling engaged: some faces were flipped to negative (dark) light.
      # (backface_scale defaults to -1, so a face whose bottom faces the camera reads
      # dark.) This guards the polarity-flip half of the pipeline.
      expect_true(any(faces$light < 0))
})


# Plot-level lighting via `+ light()` -------------------------------------------
#
# `+ light()` sets lighting for a whole plot. It resolves through two different
# routes depending on where it appears: added *after* `coord_3d()` it writes
# straight to the coord, while added *before* it is stashed on the plot and
# retrieved during the build by `find_pending_light()`. Both routes are exercised
# here, since only one of them is covered by any given ordering.

# The light actually handed to the layers, read from the built panel params.
resolved_light <- function(p) {
      ggplot_build(p)$layout$panel_params[[1]]$light
}

# Minimal 3D plot used as a base throughout.
light_test_plot <- function() {
      ggplot(mtcars, aes(mpg, wt, disp)) + geom_point_3d()
}


test_that("`+ light()` applies whether added before or after coord_3d()", {
      p_after  <- light_test_plot() + coord_3d() + light(direction = c(1, 0, 0))
      p_before <- light_test_plot() + light(direction = c(1, 0, 0)) + coord_3d()

      expect_equal(resolved_light(p_after)$direction, c(1, 0, 0))

      # The pre-coord ordering is the one that depends on the stash-and-retrieve
      # path, so it is asserted separately rather than folded into the above.
      expect_equal(resolved_light(p_before)$direction, c(1, 0, 0))

      # Both orderings must agree, not merely each work.
      expect_equal(resolved_light(p_after), resolved_light(p_before))
})


test_that("plot-level lighting falls back to the default when unspecified", {
      p <- light_test_plot() + coord_3d()

      expect_s3_class(resolved_light(p), "light")
      expect_equal(resolved_light(p)$direction, light()$direction)
})


test_that("coord_3d(light = NULL) disables lighting and is not overwritten by the default", {
      p <- light_test_plot() + coord_3d(light = NULL)

      # NULL is a deliberate request for no lighting, distinct from the `waiver()`
      # that signals "unspecified". Resolving NULL to the default would silently
      # light plots that asked not to be.
      expect_null(resolved_light(p))
})


test_that("specifying lighting in both coord_3d() and `+ light()` is an error", {
      # Both orderings hit a different branch: the first errors in ggplot_add,
      # the second during the build in setup_panel_params.
      expect_error(
            print(light_test_plot() + coord_3d(light = light()) + light()),
            "specified twice"
      )
      expect_error(
            print(light_test_plot() + light() + coord_3d(light = light())),
            "specified twice"
      )
})


test_that("coord_3d(light = ) alone is unaffected by the double-specification check", {
      p <- light_test_plot() + coord_3d(light = light(direction = c(0, 1, 0)))
      expect_equal(resolved_light(p)$direction, c(0, 1, 0))
})


test_that("replacing plot-level lighting is reported", {
      # Mirrors ggplot2's behaviour for coords and scales: a component discarded
      # wholesale is reported. Both routes are checked, since a light added after
      # `coord_3d()` is replaced on the coord while one added before is replaced
      # in the stash.
      expect_message(
            light_test_plot() + coord_3d() + light() + light(),
            "already present"
      )
      expect_message(
            light_test_plot() + light() + light() + coord_3d(),
            "already present"
      )

      # A single light is not a replacement, by either route.
      expect_silent(light_test_plot() + coord_3d() + light())
      expect_silent(light_test_plot() + light() + coord_3d())
})


test_that("the last `+ light()` wins", {
      # Matches ggplot2's convention for repeated components (coords, labs, scales).
      p <- light_test_plot() + coord_3d() +
            light(direction = c(1, 0, 0)) +
            light(direction = c(0, 1, 0))

      expect_equal(resolved_light(p)$direction, c(0, 1, 0))
})


test_that("light('none') matches the bare string and the fill/color form", {
      expect_identical(light("none"), light(fill = FALSE, color = FALSE))
      expect_equal(light("none")$shade, "neither")

      d <- data.frame(x = 1)
      expect_identical(
            attach_light(d, "none")$lighting_spec[[1]],
            attach_light(d, light("none"))$lighting_spec[[1]]
      )

      # `"none"` short-circuits before validation, so parameters that would
      # otherwise error are ignored rather than checked.
      expect_silent(light("none", direction = c(0, 0, 0)))

      # The error message advertises "none" as a valid method.
      expect_error(light("bogus"), "none")
})


test_that("adding `+ light()` does not leak into other plots sharing a coord", {
      # Coord ggproto objects have reference semantics and are not deep-copied by
      # ggplot2's plot_clone(), so mutating one in place would affect every plot
      # built from the same coord object.
      shared <- coord_3d()

      p1 <- light_test_plot() + shared + light(direction = c(1, 0, 0))
      p2 <- light_test_plot() + shared

      expect_equal(resolved_light(p1)$direction, c(1, 0, 0))
      expect_equal(resolved_light(p2)$direction, light()$direction)
})


test_that("two plots sharing a coord can carry different plot-level lights", {
      shared <- coord_3d()

      p1 <- light_test_plot() + shared + light(direction = c(1, 0, 0))
      p2 <- light_test_plot() + shared + light(direction = c(0, 1, 0))

      # Order of evaluation matters here: building p1 first must not fix the
      # light that p2 subsequently resolves.
      l1 <- resolved_light(p1)
      l2 <- resolved_light(p2)

      expect_equal(l1$direction, c(1, 0, 0))
      expect_equal(l2$direction, c(0, 1, 0))
})


test_that("repeated `+ light()` does not build a self-referential coord", {
      # `ggproto(NULL, parent)` captures `parent` lazily. Rebinding the same name
      # (`co <- ggproto(NULL, co)`) makes the clone its own parent, and any
      # inherited field lookup then recurses until the C stack overflows. The
      # coord is cloned once and mutated in place thereafter to avoid this.
      p <- light_test_plot() + coord_3d()
      for (i in 1:25) p <- p + light(direction = c(i, 0, 1))

      co <- p$coordinates
      parent <- get("super", envir = as.environment(co), inherits = FALSE)
      if (is.function(parent)) parent <- parent()

      expect_false(identical(as.environment(parent), as.environment(co)))

      # An inherited-field read is what actually triggers the overflow, so read
      # one rather than only inspecting the chain.
      expect_false(isTRUE(co$light_explicit))
      expect_equal(resolved_light(p)$direction, c(25, 0, 1))
})


test_that("plot-level lighting survives faceting and the print path", {
      # setup_panel_params() runs once per unique scale combination, so faceting
      # exercises the retrieval path repeatedly within one build.
      p <- light_test_plot() + facet_wrap(~cyl) +
            light(direction = c(1, 0, 0)) + coord_3d()

      pp <- ggplot_build(p)$layout$panel_params
      expect_gt(length(pp), 1)
      for (panel in pp) expect_equal(panel$light$direction, c(1, 0, 0))

      # Printing takes a different route into the build than ggplot_build().
      tmp <- tempfile(fileext = ".png")
      grDevices::png(tmp)
      on.exit({grDevices::dev.off(); unlink(tmp)}, add = TRUE)
      expect_silent(print(p))
})


test_that("3D guides see the resolved light, not the unresolved waiver", {
      # extract_light_from_plot() reads the coord's `light` field directly. That
      # field holds a `waiver()` until the build resolves it, and `waiver()$shade`
      # is NULL, so an unguarded `$shade == "neither"` comparison yields NA and
      # errors when used as a condition.
      extract <- extract_light_from_plot
      probe <- function(plot) extract()

      # Unspecified lighting must resolve to the default, so guides shade the key.
      bare <- probe(light_test_plot() + coord_3d())
      expect_s3_class(bare, "light")
      expect_equal(bare$shade, light()$shade)

      # `+ light()` is visible whether it lands on the coord or is still stashed.
      after  <- probe(light_test_plot() + coord_3d() + light(direction = c(1, 0, 0)))
      before <- probe(light_test_plot() + light(direction = c(1, 0, 0)) + coord_3d())
      expect_equal(after$direction, c(1, 0, 0))
      expect_equal(before$direction, c(1, 0, 0))

      # Disabled lighting stays NULL so guides skip shading entirely.
      expect_null(probe(light_test_plot() + coord_3d(light = NULL)))
})


test_that("layer-level light overrides plot-level light", {
      # Layer-level lighting is genuine specificity rather than redundancy, so it
      # wins silently instead of erroring the way two plot-level sources do.
      p <- ggplot(mountain, aes(x, y, z)) +
            geom_surface_3d(light = light(direction = c(0, 0, 1))) +
            coord_3d() +
            light(direction = c(1, 0, 0))

      b <- ggplot_build(p)
      spec <- b$data[[1]]$lighting_spec[[1]]

      expect_equal(spec$direction, c(0, 0, 1))
})


# Coord light provenance ---------------------------------------------------

# `+ light()` added before `coord_3d()` is stashed on the plot rather than
# written to the coord, and is resolved at build time. The coord's `light`
# must survive that resolution as an unmodified `waiver()`, because
# `light_explicit` is derived from the sentinel at construction and anything
# that reads the coord afterwards (notably `animate_3d()` and `orbit_3d()`,
# which rebuild a coord from its fields) would otherwise see an inherited
# light as an explicitly supplied one.

light_test_plot <- function(...) {
      d <- expand.grid(x = 1:4, y = 1:4)
      d$z <- d$x + d$y
      ggplot(d, aes(x, y, z)) +
            geom_surface_3d() +
            light(mode = "hsl", direction = c(-1, 1, 0), anchor = "camera") +
            coord_3d(...)
}

light_test_data <- function() {
      d <- expand.grid(x = 1:4, y = 1:4)
      d$z <- d$x + d$y
      d
}

test_that("building a plot leaves an inherited `+ light()` off the coord", {
      p <- light_test_plot()
      expect_true(inherits(p$coordinates$light, "waiver"))
      expect_false(isTRUE(p$coordinates$light_explicit))

      invisible(ggplot_build(p))

      expect_true(inherits(p$coordinates$light, "waiver"))
      expect_false(isTRUE(p$coordinates$light_explicit))
})

test_that("a plot carrying `+ light()` can be built more than once", {
      p <- light_test_plot()
      invisible(ggplot_build(p))
      expect_no_error(invisible(ggplot_build(p)))
})

test_that("a plot carrying `+ light()` can be rendered more than once", {
      # Exercises Coord3D$transform(), which runs at draw time rather than
      # build time and is the consumer of the resolved light.
      p <- light_test_plot()
      invisible(ggplotGrob(p))
      expect_no_error(invisible(ggplotGrob(p)))
      expect_true(inherits(p$coordinates$light, "waiver"))
})

test_that("building a plot does not leak lighting into plots sharing its coord", {
      p <- light_test_plot()
      p2 <- p + theme_bw()

      invisible(ggplot_build(p))

      expect_true(inherits(p2$coordinates$light, "waiver"))
})

test_that("plot-level lighting reaches panel_params", {
      pp <- ggplot_build(light_test_plot())$layout$panel_params[[1]]

      expect_s3_class(pp$light, "light")
      expect_equal(pp$light$shade_mode, "hsl")
      expect_equal(pp$light$anchor, "camera")
      expect_equal(pp$light$direction, c(-1, 1, 0))
})

test_that("explicit `coord_3d(light = )` reaches panel_params", {
      p <- ggplot(light_test_data(), aes(x, y, z)) +
            geom_surface_3d() +
            coord_3d(light = light(mode = "hsl", contrast = 0.5))
      pp <- ggplot_build(p)$layout$panel_params[[1]]

      expect_s3_class(pp$light, "light")
      expect_equal(pp$light$shade_mode, "hsl")
      expect_equal(pp$light$shade_strength, 0.5)
})

test_that("`coord_3d(light = NULL)` disables lighting", {
      p <- ggplot(light_test_data(), aes(x, y, z)) +
            geom_surface_3d() +
            coord_3d(light = NULL)
      pp <- ggplot_build(p)$layout$panel_params[[1]]

      expect_null(pp$light)
      expect_no_error(invisible(ggplotGrob(p)))
})

test_that("unspecified lighting resolves to the package default", {
      p <- ggplot(light_test_data(), aes(x, y, z)) +
            geom_surface_3d() +
            coord_3d()
      pp <- ggplot_build(p)$layout$panel_params[[1]]

      expect_s3_class(pp$light, "light")
      expect_equal(pp$light$shade_mode, light()$shade_mode)
})

test_that("lighting specified in two places is an error", {
      # `+ light()` first: no coord to write to, so the clash surfaces at
      # build time when the stash meets an explicit coord light.
      p <- ggplot(light_test_data(), aes(x, y, z)) +
            geom_surface_3d() +
            light(mode = "hsl") +
            coord_3d(light = light(mode = "hsl"))
      expect_error(ggplot_build(p), "specified twice")

      # `coord_3d(light = )` first: the clash surfaces immediately, when
      # `+ light()` finds a coord that already holds an explicit spec.
      expect_error(
            ggplot(light_test_data(), aes(x, y, z)) +
                  geom_surface_3d() +
                  coord_3d(light = light(mode = "hsl")) +
                  light(mode = "hsl"),
            "specified twice"
      )
})

test_that("shaded guides find plot-level lighting from an untouched coord", {
      # guide_colorbar_3d() reads lighting off the plot rather than off the
      # resolved panel_params, so it must handle a coord whose light is still
      # a waiver and a spec still stashed on the plot.
      p <- light_test_plot() + guides(fill = guide_colorbar_3d())
      expect_no_error(invisible(ggplotGrob(p)))
})

.onLoad <- function(libname, pkgname) {
      # Define custom theme elements
      register_theme_elements(
            axis.text.z = element_text(),
            axis.title.z = element_text(),
            axis.ticks.z = element_line(),

            # No alpha set here intentionally: leaving the foreground default
            # without an explicit alpha lets it inherit from panel.background
            # at render time. The final 0.2 fallback (when neither layer has
            # an explicit alpha) lives in create_panel_polygons() in
            # R/panel-rendering.R.
            panel.foreground = element_rect(),

            panel.grid.foreground = element_line(),
            panel.grid.major.foreground = element_line(),
            panel.border.foreground = element_rect(),
            element_tree = list(
                  axis.text.z = el_def("element_text", "axis.text"),
                  axis.title.z = el_def("element_text", "axis.title"),
                  axis.ticks.z = el_def("element_line", "axis.ticks"),

                  # Registered in the element tree only. A unit-valued element
                  # has no sensible standalone default, so the value comes from
                  # axis.ticks.length by inheritance.
                  axis.ticks.length.z = el_def("unit", "axis.ticks.length"),
                  panel.foreground = el_def("element_rect", "panel.background"),
                  panel.grid.foreground = el_def("element_line", "panel.grid"),
                  panel.grid.major.foreground = el_def("element_line", "panel.grid.foreground"),
                  panel.border.foreground = el_def("element_rect", "panel.border")
            )
      )
}

#' ggcube: Extend ggplot2 into 3D
#'
#' ggcube provides 3D geoms, stats, and coordinate systems for ggplot2.
#' Add [coord_3d()] to any plot with a z aesthetic to render it in 3D
#' with rotation, perspective, lighting, and depth sorting. Layer types
#' include surfaces, hulls, contours, ridgelines, paths, points, segments,
#' text, bars, and columns.
#'
#' @docType package
#' @name ggcube
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import dplyr
#' @import ggplot2
#' @importFrom ggplot2 resolution
#' @importFrom grDevices as.raster
#' @importFrom grDevices chull
#' @importFrom grDevices col2rgb
#' @importFrom grDevices hsv
#' @importFrom grDevices rgb
#' @importFrom grDevices rgb2hsv
#' @importFrom grid convertHeight
#' @importFrom grid convertWidth
#' @importFrom grid gpar
#' @importFrom grid grobHeight
#' @importFrom grid grobTree
#' @importFrom grid grobWidth
#' @importFrom grid nullGrob
#' @importFrom grid polygonGrob
#' @importFrom grid rectGrob
#' @importFrom grid segmentsGrob
#' @importFrom grid textGrob
#' @importFrom isoband isobands
#' @importFrom labeling extended
#' @importFrom lifecycle deprecate_soft
#' @importFrom lifecycle is_present
#' @importFrom magrittr %>%
#' @importFrom polyclip polyclip
#' @importFrom polyclip polysimplify
#' @importFrom purrr map_dfr
#' @importFrom rlang %||%
#' @importFrom scales extended_breaks
#' @importFrom scales rescale
#' @importFrom stats ave density dnorm glm lm loess na.omit predict setNames weighted.mean
#' @importFrom stringr str_remove
#' @importFrom systemfonts glyph_outline
#' @importFrom systemfonts match_fonts
#' @importFrom systemfonts shape_string
#' @importFrom tidyr expand_grid
#' @importFrom tidyr gather
#' @importFrom tidyr unite
#' @importFrom utils modifyList
## usethis namespace: end
utils::globalVariables(c(
      ".", ".area", ".backface", ".prim_depth", ".render_order", ".subgroup",
      ".vertex_idx", ".vertex_order", ".x_round", ".y_round",
      "Sxx", "Sxy", "Sxz", "Syy", "Syz",
      "apply_quantization", "axis", "bin_area", "column", "cont_bin",
      "contour", "d_col", "d_row", "denom", "depth", "depth1", "depth_3d",
      "dzdx", "dzdy", "e",
      "e1x", "e1y", "e1z", "e2x", "e2y", "e2z",
      "element_type", "face", "facet", "far_corner", "foreground",
      "free_axis", "fx", "fz", "group", "id", "keep",
      "level1", "level2",
      "n_pts", "near_corner", "normal_x", "normal_y", "normal_z",
      "on_hull", "ref_circle_radius", "ref_circle_vertices",
      "simple", "sum_x", "sum_xx", "sum_xy", "sum_xz",
      "sum_y", "sum_yy", "sum_yz", "sum_z",
      "temp", "tile_col", "tile_id", "tile_row",
      "value", "vert", "vertex_order", "weight",
      "x", "x1", "x2", "x3", "x_bin",
      "y", "y1", "y2", "y3", "y_bin",
      "z", "z1", "z2", "z3", "zmean"
))
NULL

enumerate_axis_edges <- function(axis) {
      # Return all 4 possible edges for the given axis
      # Each edge is defined by the 2 coordinates that are fixed (not varying along the axis)

      if (axis == "x") {
            # X-axis edges: x varies from -0.5 to 0.5, y and z are fixed
            edges <- list(
                  list(fixed_coords = c(y = -0.5, z = -0.5), touching_faces = c("ymin", "zmin")),
                  list(fixed_coords = c(y = -0.5, z =  0.5), touching_faces = c("ymin", "zmax")),
                  list(fixed_coords = c(y =  0.5, z = -0.5), touching_faces = c("ymax", "zmin")),
                  list(fixed_coords = c(y =  0.5, z =  0.5), touching_faces = c("ymax", "zmax"))
            )
      } else if (axis == "y") {
            # Y-axis edges: y varies from -0.5 to 0.5, x and z are fixed
            edges <- list(
                  list(fixed_coords = c(x = -0.5, z = -0.5), touching_faces = c("xmin", "zmin")),
                  list(fixed_coords = c(x = -0.5, z =  0.5), touching_faces = c("xmin", "zmax")),
                  list(fixed_coords = c(x =  0.5, z = -0.5), touching_faces = c("xmax", "zmin")),
                  list(fixed_coords = c(x =  0.5, z =  0.5), touching_faces = c("xmax", "zmax"))
            )
      } else { # axis == "z"
            # Z-axis edges: z varies from -0.5 to 0.5, x and y are fixed
            edges <- list(
                  list(fixed_coords = c(x = -0.5, y = -0.5), touching_faces = c("xmin", "ymin")),
                  list(fixed_coords = c(x = -0.5, y =  0.5), touching_faces = c("xmin", "ymax")),
                  list(fixed_coords = c(x =  0.5, y = -0.5), touching_faces = c("xmax", "ymin")),
                  list(fixed_coords = c(x =  0.5, y =  0.5), touching_faces = c("xmax", "ymax"))
            )
      }

      # Add edge endpoints to each edge
      for (i in seq_along(edges)) {
            if (axis == "x") {
                  edges[[i]]$p1 <- c(x = -0.5, edges[[i]]$fixed_coords)
                  edges[[i]]$p2 <- c(x =  0.5, edges[[i]]$fixed_coords)
            } else if (axis == "y") {
                  edges[[i]]$p1 <- c(edges[[i]]$fixed_coords[1], y = -0.5, edges[[i]]$fixed_coords[2])
                  edges[[i]]$p2 <- c(edges[[i]]$fixed_coords[1], y =  0.5, edges[[i]]$fixed_coords[2])
            } else { # axis == "z"
                  edges[[i]]$p1 <- c(edges[[i]]$fixed_coords, z = -0.5)
                  edges[[i]]$p2 <- c(edges[[i]]$fixed_coords, z =  0.5)
            }
            edges[[i]]$axis <- axis
            edges[[i]]$edge_id <- i
      }

      return(edges)
}

# Calculate 2D projected area of a face
calculate_face_area_2d <- function(face_name, proj, effective_ratios) {
      corners_3d <- get_face_corners(face_name)

      # Scale by effective ratios to match coordinate system
      corners_3d$x <- corners_3d$x * effective_ratios[1]
      corners_3d$y <- corners_3d$y * effective_ratios[2]
      corners_3d$z <- corners_3d$z * effective_ratios[3]

      # Transform to 2D
      corners_2d <- transform_3d_standard(corners_3d, proj)

      # Reorder into perimeter sequence via convex hull
      hull_idx <- chull(corners_2d$x, corners_2d$y)
      x <- corners_2d$x[hull_idx]
      y <- corners_2d$y[hull_idx]

      # Shoelace formula
      n <- length(x)
      area <- 0.5 * abs(sum(x * c(y[-1], y[1]) - c(x[-1], x[1]) * y))

      return(area)
}

# Helper function to get which axes have gridlines on a given face
get_gridline_axes_for_face <- function(face) {
      # Each face has gridlines for the 2 axes that aren't the face normal
      face_axis <- substr(face, 1, 1)  # "x", "y", or "z"
      all_axes <- c("x", "y", "z")
      gridline_axes <- setdiff(all_axes, face_axis)
      return(gridline_axes)
}

# Face-specific perpendicularity calculation
calculate_face_perpendicularity_score <- function(edge, face, axis, proj, effective_ratios) {
      # Get 2D directions for all three axes using cube corners
      axis_directions_2d <- list()

      # X-axis direction: from (-0.5, 0, 0) to (0.5, 0, 0)
      x_3d <- data.frame(
            x = c(-0.5, 0.5) * effective_ratios[1],
            y = c(0, 0),
            z = c(0, 0)
      )
      x_2d <- transform_3d_standard(x_3d, proj)
      axis_directions_2d$x <- c(x_2d$x[2] - x_2d$x[1], x_2d$y[2] - x_2d$y[1])

      # Y-axis direction: from (0, -0.5, 0) to (0, 0.5, 0)
      y_3d <- data.frame(
            x = c(0, 0),
            y = c(-0.5, 0.5) * effective_ratios[2],
            z = c(0, 0)
      )
      y_2d <- transform_3d_standard(y_3d, proj)
      axis_directions_2d$y <- c(y_2d$x[2] - y_2d$x[1], y_2d$y[2] - y_2d$y[1])

      # Z-axis direction: from (0, 0, -0.5) to (0, 0, 0.5)
      z_3d <- data.frame(
            x = c(0, 0),
            y = c(0, 0),
            z = c(-0.5, 0.5) * effective_ratios[3]
      )
      z_2d <- transform_3d_standard(z_3d, proj)
      axis_directions_2d$z <- c(z_2d$x[2] - z_2d$x[1], z_2d$y[2] - z_2d$y[1])

      # Get this edge's 2D direction
      edge_dir_2d <- c(edge$p2_2d$x - edge$p1_2d$x, edge$p2_2d$y - edge$p1_2d$y)
      edge_length_2d <- sqrt(sum(edge_dir_2d^2))

      if (edge_length_2d == 0) return(0)  # Degenerate edge

      edge_dir_2d <- edge_dir_2d / edge_length_2d  # Normalize

      # Get which axes have gridlines on this face
      gridline_axes <- get_gridline_axes_for_face(face)

      # Remove the axis we're labeling (since that's parallel to the edge)
      perpendicular_axes <- setdiff(gridline_axes, axis)

      if (length(perpendicular_axes) == 0) return(0)  # No perpendicular gridlines

      # Calculate perpendicularity to the gridline axes that should be perpendicular
      perp_scores <- numeric(length(perpendicular_axes))
      for (i in seq_along(perpendicular_axes)) {
            perp_axis <- perpendicular_axes[i]
            perp_axis_dir <- axis_directions_2d[[perp_axis]]
            perp_axis_length <- sqrt(sum(perp_axis_dir^2))

            if (perp_axis_length > 0) {
                  perp_axis_dir <- perp_axis_dir / perp_axis_length  # Normalize

                  # Calculate angle between edge and this perpendicular axis
                  dot_product <- sum(edge_dir_2d * perp_axis_dir)
                  angle_rad <- acos(pmax(-1, pmin(1, abs(dot_product))))  # Clamp for numerical stability
                  angle_deg <- angle_rad * 180 / pi

                  # Score based on deviation from 90 degrees (perfect perpendicularity)
                  deviation_from_90 <- abs(angle_deg - 90)
                  perp_scores[i] <- 90 - deviation_from_90  # Higher score = more perpendicular
            }
      }

      # Return average perpendicularity score (0-90, higher is better)
      return(mean(perp_scores))
}

# Externality helpers ---------------------------------------------------------

# Projected convex hull of a set of 3D points, returned as ordered 2D
# vertices. NULL when fewer than three distinct points survive projection,
# which happens for a single face viewed exactly edge-on.
projected_hull <- function(corners_3d, proj) {
      if (is.null(corners_3d) || nrow(corners_3d) == 0) return(NULL)

      corners_2d <- transform_3d_standard(corners_3d, proj)
      distinct <- unique(data.frame(x = corners_2d$x, y = corners_2d$y))

      if (nrow(distinct) < 3) return(NULL)

      hull_idx <- chull(distinct$x, distinct$y)
      if (length(hull_idx) < 2) return(NULL)

      distinct[hull_idx, , drop = FALSE]
}

# Convex hull of the rendered panels. This is the reference for deciding
# which side of an edge faces away from the drawn figure, which is not the
# same as the cube silhouette whenever `panels` is a subset of the faces.
panel_silhouette_hull <- function(visible_faces, proj, effective_ratios) {
      if (is.null(visible_faces) || length(visible_faces) == 0) return(NULL)

      corners <- do.call(rbind, lapply(visible_faces, function(face) {
            corners_3d <- get_face_corners(face)
            corners_3d$x <- corners_3d$x * effective_ratios[1]
            corners_3d$y <- corners_3d$y * effective_ratios[2]
            corners_3d$z <- corners_3d$z * effective_ratios[3]
            corners_3d
      }))

      projected_hull(corners, proj)
}

# Shortest distance from a point to a line segment.
point_segment_distance <- function(px, py, x1, y1, x2, y2) {
      dx <- x2 - x1
      dy <- y2 - y1
      seg_sq <- dx * dx + dy * dy

      if (seg_sq == 0) return(sqrt((px - x1)^2 + (py - y1)^2))

      t <- ((px - x1) * dx + (py - y1) * dy) / seg_sq
      t <- max(0, min(1, t))

      sqrt((px - (x1 + t * dx))^2 + (py - (y1 + t * dy))^2)
}

# Is an edge on the boundary of the given hull?
#
# Tested via the edge's 2D midpoint rather than by matching endpoints against
# hull vertices. A boundary edge has its midpoint on a hull segment, while a
# diagonal between non-consecutive hull vertices, or an edge with an interior
# endpoint, has its midpoint strictly inside. The midpoint test is also
# robust to coincident projected corners, which occur under orthographic
# projection and otherwise make the result depend on chull() tie-breaking.
#
# A NULL hull means there is nothing to be inside of, so every edge counts.
edge_on_hull <- function(p1_2d, p2_2d, hull) {
      if (is.null(hull)) return(TRUE)

      mid_x <- (p1_2d$x + p2_2d$x) / 2
      mid_y <- (p1_2d$y + p2_2d$y) / 2

      hull_scale <- max(diff(range(hull$x)), diff(range(hull$y)))
      if (!is.finite(hull_scale) || hull_scale <= 0) return(TRUE)
      tol <- 1e-6 * hull_scale

      n <- nrow(hull)
      for (i in seq_len(n)) {
            j <- if (i == n) 1 else i + 1
            if (point_segment_distance(mid_x, mid_y,
                                       hull$x[i], hull$y[i],
                                       hull$x[j], hull$y[j]) <= tol) {
                  return(TRUE)
            }
      }

      FALSE
}

# Classify an edge against both silhouettes.
classify_edge_externality <- function(p1_2d, p2_2d, cube_hull, panels_hull) {
      list(
            on_cube_hull = edge_on_hull(p1_2d, p2_2d, cube_hull),
            on_panel_hull = edge_on_hull(p1_2d, p2_2d, panels_hull)
      )
}

# Score edge-face combinations
score_edge_face_combinations <- function(combinations, axis, proj, effective_ratios,
                                         visible_faces = NULL) {
      if (length(combinations) == 0) return(list())

      # Silhouette of the whole cube, and of the rendered panels only
      all_corners <- expand.grid(
            x = c(-0.5, 0.5) * effective_ratios[1],
            y = c(-0.5, 0.5) * effective_ratios[2],
            z = c(-0.5, 0.5) * effective_ratios[3]
      )
      all_corners_2d <- transform_3d_standard(all_corners, proj)
      cube_hull <- projected_hull(all_corners, proj)
      panels_hull <- panel_silhouette_hull(visible_faces, proj, effective_ratios)

      # Compute max x+y across all projected cube corners for corner score normalization
      max_corner_xy <- max(all_corners_2d$x + all_corners_2d$y)

      # Score each combination
      for (i in seq_along(combinations)) {
            edge <- combinations[[i]]$edge
            face <- combinations[[i]]$face

            # Apply effective ratios scaling to match hull corner scaling
            p1_scaled <- c(edge$p1[1] * effective_ratios[1],
                           edge$p1[2] * effective_ratios[2],
                           edge$p1[3] * effective_ratios[3])
            p2_scaled <- c(edge$p2[1] * effective_ratios[1],
                           edge$p2[2] * effective_ratios[2],
                           edge$p2[3] * effective_ratios[3])

            # Transform edge endpoints to 2D
            p1_df <- data.frame(x = p1_scaled[1], y = p1_scaled[2], z = p1_scaled[3])
            p2_df <- data.frame(x = p2_scaled[1], y = p2_scaled[2], z = p2_scaled[3])
            p1_2d <- transform_3d_standard(p1_df, proj)
            p2_2d <- transform_3d_standard(p2_df, proj)

            # Store 2D points in edge object
            combinations[[i]]$edge$p1_2d <- p1_2d
            combinations[[i]]$edge$p2_2d <- p2_2d

            # 1. Externality against both silhouettes. Placement depends on
            # panel externality; cube externality is preferred because such
            # edges cannot be occluded by the data.
            externality <- classify_edge_externality(p1_2d, p2_2d, cube_hull, panels_hull)

            # 2. Face-specific perpendicularity score
            perpendicularity_score <- calculate_face_perpendicularity_score(combinations[[i]]$edge, face, axis, proj, effective_ratios)

            # 3. Face area score
            face_area <- calculate_face_area_2d(face, proj, effective_ratios)

            # 4. Depth score (average z of endpoints - lower is better/closer)
            avg_depth <- (p1_2d$z + p2_2d$z) / 2

            # 5. Corner score (average x+y of endpoints - lower is better/more bottom-left)
            avg_corner <- ((p1_2d$x + p1_2d$y) + (p2_2d$x + p2_2d$y)) / 2

            # 6. Edge length
            edge_length <- sqrt((p2_2d$x - p1_2d$x)^2 + (p2_2d$y - p1_2d$y)^2)

            # Store scores in the combination
            combinations[[i]]$on_cube_hull <- externality$on_cube_hull
            combinations[[i]]$on_panel_hull <- externality$on_panel_hull
            combinations[[i]]$perpendicularity_score <- perpendicularity_score
            combinations[[i]]$face_area <- face_area
            combinations[[i]]$depth_score <- avg_depth
            combinations[[i]]$corner_score <- avg_corner
            combinations[[i]]$edge_length <- edge_length
            combinations[[i]]$max_corner_xy <- max_corner_xy
      }

      return(combinations)
}

# Select best edge-face combination with weighted scoring
select_best_edge_face_combination <- function(scored_combinations,
                                              weights = c(perp = 5,
                                                          length = 3,
                                                          area = 2,
                                                          position = 2)) {
      if (length(scored_combinations) == 0) return(NULL)

      # Validate and normalize weights
      if (length(weights) != 4) {
            stop("weights must be a vector of length 4: c(perpendicularity, length, area, position)")
      }
      if (any(weights < 0)) {
            stop("All weights must be non-negative")
      }

      # Normalize weights to sum to 1
      weights <- weights / sum(weights)
      perp_weight <- weights[1]
      length_weight <- weights[2]
      area_weight <- weights[3]
      position_weight <- weights[4]

      # STEP 1: Prefer edges that place correctly and cannot be occluded by
      # data, then edges that at least place correctly, then anything.
      # vapply rather than sapply: sapply returns list() on an empty input,
      # which is not a valid subscript.
      flagged <- function(combos, field) {
            if (length(combos) == 0) return(logical(0))
            vapply(combos, function(c) isTRUE(c[[field]]), logical(1))
      }

      panel_external <- scored_combinations[flagged(scored_combinations, "on_panel_hull")]
      both_external <- panel_external[flagged(panel_external, "on_cube_hull")]

      combinations <- if (length(both_external) > 0) {
            both_external
      } else if (length(panel_external) > 0) {
            panel_external
      } else {
            scored_combinations
      }

      # STEP 2: Calculate weighted scores with four factors
      # Find max values for normalization
      max_length <- max(sapply(combinations, function(c) c$edge_length))
      max_area <- max(sapply(combinations, function(c) c$face_area))
      if (max_area == 0) max_area <- 1  # Avoid NaN from 0/0

      # Corner normalization ceiling: max x+y across all projected cube corners
      max_corner_xy <- combinations[[1]]$max_corner_xy

      for(i in seq_along(combinations)) {
            c <- combinations[[i]]

            # Normalize to 0-1 range
            norm_perp <- c$perpendicularity_score / 90
            norm_length <- c$edge_length / max_length
            norm_area <- c$face_area / max_area

            # Invert corner: lower corner_score (more bottom-left) = higher norm_position
            inverted_corner <- max_corner_xy - c$corner_score
            max_inverted <- max_corner_xy - min(sapply(combinations, function(c) c$corner_score))
            norm_position <- inverted_corner / max_inverted

            # Multiplicative scoring - all factors must be decent
            c$combined_score <- (norm_perp^perp_weight) * (norm_length^length_weight) *
                  (norm_area^area_weight) * (norm_position^position_weight)

            combinations[[i]] <- c
      }

      # STEP 3: Sort by combined score (highest first), then depth, then corner as tiebreakers
      combo_data <- data.frame(
            index = seq_along(combinations),
            combined_score = sapply(combinations, function(c) c$combined_score),
            depth = sapply(combinations, function(c) c$depth_score),
            corner = sapply(combinations, function(c) c$corner_score)
      )
      combo_data <- combo_data[order(-combo_data$combined_score, combo_data$depth, combo_data$corner), ]

      best_combination <- combinations[[combo_data$index[1]]]

      return(best_combination)
}

select_axis_edge_and_face <- function(axis, visible_faces, proj, effective_ratios) {
      # Step 1: Enumerate all possible edges for this axis
      all_edges <- enumerate_axis_edges(axis)

      # Step 2: Create edge-face combinations
      edge_face_combinations <- list()
      for (edge in all_edges) {
            for (face in visible_faces) {
                  if (face %in% edge$touching_faces) {
                        edge_face_combinations[[length(edge_face_combinations) + 1]] <- list(
                              edge = edge,
                              face = face
                        )
                  }
            }
      }

      if (length(edge_face_combinations) == 0) {
            return(NULL)  # No valid combinations for this axis
      }

      # Step 3: Score each edge-face combination
      scored_combinations <- score_edge_face_combinations(edge_face_combinations, axis, proj,
                                                          effective_ratios, visible_faces)

      # Step 4: Select best combination
      best_combination <- select_best_edge_face_combination(scored_combinations)

      if (is.null(best_combination)) {
            return(NULL)
      }

      return(list(
            axis = axis,
            edge = best_combination$edge,
            face = best_combination$face,
            on_cube_hull = best_combination$on_cube_hull,
            on_panel_hull = best_combination$on_panel_hull,
            edge_p1_2d = best_combination$edge$p1_2d,
            edge_p2_2d = best_combination$edge$p2_2d
      ))
}

get_axis_selection <- function(axis, self, panel_params, effective_ratios) {
      # Get the appropriate text parameter for this axis
      axis_override <- switch(axis,
                              "x" = self$xlabels,
                              "y" = self$ylabels,
                              "z" = self$zlabels)

      if (length(axis_override) == 1 && axis_override == "auto") {
            # Use automatic selection (needs effective_ratios for scoring)
            return(select_axis_edge_and_face(axis, panel_params$visible_faces, panel_params$proj, effective_ratios))
      } else if (length(axis_override) == 2) {
            # Manual override (doesn't need effective_ratios)
            return(create_manual_axis_selection(axis, axis_override[1], axis_override[2],
                                                panel_params$proj, panel_params$visible_faces,
                                                effective_ratios))
      } else {
            warning("Invalid ", axis, "text specification, using auto")
            return(select_axis_edge_and_face(axis, panel_params$visible_faces, panel_params$proj, effective_ratios))
      }
}

create_manual_axis_selection <- function(axis, face1, face2, proj, visible_faces,
                                         effective_ratios = c(1, 1, 1)) {

      # Parse face names to get coordinate constraints
      parse_face <- function(face_name) {
            coord <- substr(face_name, 1, 1)  # "x", "y", or "z"
            value <- ifelse(substr(face_name, 2, 4) == "min", -0.5, 0.5)
            return(list(coord = coord, value = value))
      }

      f1 <- parse_face(face1)
      f2 <- parse_face(face2)

      # Verify faces are different and adjacent (share an edge)
      if (f1$coord == f2$coord) {
            stop("Faces ", face1, " and ", face2, " are parallel and don't share an edge")
      }

      # Find the shared edge: two coordinates are fixed, one varies
      # The varying coordinate should match the axis
      all_coords <- c("x", "y", "z")
      fixed_coords <- c(f1$coord, f2$coord)
      varying_coord <- setdiff(all_coords, fixed_coords)

      if (varying_coord != axis) {
            stop("Edge between ", face1, " and ", face2, " doesn't vary along ", axis, "-axis")
      }

      # Determine which face to use based on visibility
      face1_visible <- face1 %in% visible_faces
      face2_visible <- face2 %in% visible_faces

      if (face1_visible && face2_visible) {
            # Both visible - honor user's order (use face1)
            chosen_face <- face1
      } else if (face1_visible) {
            # Only face1 visible
            chosen_face <- face1
      } else if (face2_visible) {
            # Only face2 visible
            chosen_face <- face2
      } else {
            # Neither visible - warn and fall back to auto
            warning("Neither ", face1, " nor ", face2, " is visible, falling back to auto selection")
            return(select_axis_edge_and_face(axis, visible_faces, proj, effective_ratios))
      }

      # Create edge endpoints
      edge_coords <- list()
      edge_coords[[f1$coord]] <- f1$value
      edge_coords[[f2$coord]] <- f2$value

      # Edge varies along the axis coordinate
      p1 <- edge_coords
      p1[[axis]] <- -0.5
      p2 <- edge_coords
      p2[[axis]] <- 0.5

      # Create edge structure (similar to enumerate_axis_edges output)
      edge <- list(
            fixed_coords = setNames(c(edge_coords[[f1$coord]], edge_coords[[f2$coord]]),
                                    c(f1$coord, f2$coord)),
            touching_faces = c(face1, face2),
            p1 = c(p1$x, p1$y, p1$z),
            p2 = c(p2$x, p2$y, p2$z),
            axis = axis,
            edge_id = "manual"
      )

      # Transform to 2D for consistency with automatic selection. Endpoints are
      # scaled by the effective ratios first, matching the scoring path, so the
      # projected points correspond to the drawn geometry rather than to a
      # unit cube.
      p1_scaled <- c(edge$p1[1] * effective_ratios[1],
                     edge$p1[2] * effective_ratios[2],
                     edge$p1[3] * effective_ratios[3])
      p2_scaled <- c(edge$p2[1] * effective_ratios[1],
                     edge$p2[2] * effective_ratios[2],
                     edge$p2[3] * effective_ratios[3])

      edge_p1_3d <- data.frame(x = p1_scaled[1], y = p1_scaled[2], z = p1_scaled[3])
      edge_p2_3d <- data.frame(x = p2_scaled[1], y = p2_scaled[2], z = p2_scaled[3])
      edge_p1_2d <- transform_3d_standard(edge_p1_3d, proj)
      edge_p2_2d <- transform_3d_standard(edge_p2_3d, proj)

      all_corners <- expand.grid(
            x = c(-0.5, 0.5) * effective_ratios[1],
            y = c(-0.5, 0.5) * effective_ratios[2],
            z = c(-0.5, 0.5) * effective_ratios[3]
      )
      externality <- classify_edge_externality(
            edge_p1_2d, edge_p2_2d,
            projected_hull(all_corners, proj),
            panel_silhouette_hull(visible_faces, proj, effective_ratios)
      )

      # Return same structure as select_axis_edge_and_face
      return(list(
            axis = axis,
            edge = edge,
            face = chosen_face,
            on_cube_hull = externality$on_cube_hull,
            on_panel_hull = externality$on_panel_hull,
            edge_p1_2d = edge_p1_2d,
            edge_p2_2d = edge_p2_2d
      ))
}

test_that("point_segment_distance measures perpendicular distance", {
      expect_equal(point_segment_distance(0, 3, -5, 0, 5, 0), 3)
      expect_equal(point_segment_distance(2, 0, 0, 0, 10, 0), 0)
})

test_that("point_segment_distance clamps beyond the segment ends", {
      # Projection falls past the far end, so distance is to the endpoint
      expect_equal(point_segment_distance(13, 4, 0, 0, 10, 0), 5)
      expect_equal(point_segment_distance(-3, 4, 0, 0, 10, 0), 5)
})

test_that("point_segment_distance handles a degenerate segment", {
      expect_equal(point_segment_distance(3, 4, 0, 0, 0, 0), 5)
})


# A unit square in perimeter order, standing in for a projected silhouette
square_hull <- function() {
      data.frame(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
}

pt <- function(x, y) data.frame(x = x, y = y)

test_that("edge_on_hull accepts edges lying on the boundary", {
      hull <- square_hull()

      expect_true(edge_on_hull(pt(0, 0), pt(1, 0), hull))
      expect_true(edge_on_hull(pt(1, 0), pt(1, 1), hull))
      expect_true(edge_on_hull(pt(0, 0), pt(0, 1), hull))
})

test_that("edge_on_hull rejects a diagonal between hull vertices", {
      # Both endpoints are hull vertices, but the midpoint is interior.
      # This is the case that endpoint matching alone would get wrong.
      hull <- square_hull()

      expect_false(edge_on_hull(pt(0, 0), pt(1, 1), hull))
      expect_false(edge_on_hull(pt(1, 0), pt(0, 1), hull))
})

test_that("edge_on_hull rejects fully interior edges", {
      hull <- square_hull()

      expect_false(edge_on_hull(pt(0.25, 0.25), pt(0.75, 0.25), hull))
      expect_false(edge_on_hull(pt(0.5, 0.1), pt(0.5, 0.9), hull))
})

test_that("edge_on_hull accepts a sub-segment of a boundary edge", {
      # An edge shorter than the hull side it lies on is still external.
      # Vertex matching would reject this; the midpoint test does not.
      hull <- square_hull()

      expect_true(edge_on_hull(pt(0.2, 0), pt(0.8, 0), hull))
})

test_that("edge_on_hull treats a NULL hull as unconstrained", {
      expect_true(edge_on_hull(pt(0, 0), pt(1, 1), NULL))
})

test_that("edge_on_hull tolerance scales with hull size", {
      small <- square_hull()
      large <- data.frame(x = c(0, 1000, 1000, 0), y = c(0, 0, 1000, 1000))

      # An offset that is negligible relative to the large hull, but not the
      # small one, should be judged accordingly
      offset <- 0.01

      expect_false(edge_on_hull(pt(0, offset), pt(1, offset), small))
      expect_true(edge_on_hull(pt(0, 0), pt(1000, 0), large))
})

test_that("edge_on_hull handles a degenerate two-point hull", {
      # A face seen exactly edge-on collapses to a line segment
      line_hull <- data.frame(x = c(0, 10), y = c(0, 0))

      expect_true(edge_on_hull(pt(2, 0), pt(8, 0), line_hull))
      expect_false(edge_on_hull(pt(2, 5), pt(8, 5), line_hull))
})

test_that("classify_edge_externality reports both silhouettes independently", {
      cube_hull <- data.frame(x = c(-1, 1, 1, -1), y = c(-1, -1, 1, 1))
      panels_hull <- square_hull()

      # On the panel boundary, interior to the cube
      result <- classify_edge_externality(pt(0, 0), pt(1, 0), cube_hull, panels_hull)
      expect_true(result$on_panel_hull)
      expect_false(result$on_cube_hull)

      # On the cube boundary, outside and so not on the panel boundary
      result <- classify_edge_externality(pt(-1, -1), pt(1, -1), cube_hull, panels_hull)
      expect_true(result$on_cube_hull)
      expect_false(result$on_panel_hull)
})

#' Animate a 3D plot
#'
#' Renders an animated GIF or MP4 of a rotating ggplot with [coord_3d()],
#' smoothly interpolating rotation angles across frames.
#'
#' @param plot A ggplot object that uses [coord_3d()].
#' @param pitch,roll,yaw Rotation parameters to animate. Each can be:
#'   \itemize{
#'     \item \code{NULL} (default): Hold at the value specified in the plot's
#'       [coord_3d()] call.
#'     \item A single numeric value: Hold at this value for all frames
#'       (overriding the plot's coord).
#'     \item A numeric vector of length 2+: Keyframe values that are
#'       interpolated linearly across the animation. Keyframes are evenly
#'       spaced, so \code{c(0, 180, 180)} with 100 frames means: start at 0,
#'       reach 180 at frame 50, hold at 180 until frame 100.
#'   }
#' @param nframes,fps,duration Control animation length and speed. Specify
#'   any two of three; the third is computed as
#'   \code{duration = nframes / fps}. Defaults: \code{nframes = 100},
#'   \code{fps = 10}.
#' @param width,height Dimensions of the animation in pixels. Defaults to
#'   480 x 480.
#' @param res Resolution in ppi for the rendered frames. Default is 96.
#'   Controls the size of point-based text/lines/point elements relative to
#'   the plot. See `?grDevices::png` for details.
#' @param renderer A renderer function that combines image frames into
#'   an animation file. Built-in options:
#'   \itemize{
#'     \item \code{gifski_renderer_3d()} (default): Renders a GIF using the
#'       \pkg{gifski} package.
#'     \item \code{av_renderer_3d()}: Renders a video using the \pkg{av}
#'       package.
#'     \item \code{file_renderer_3d()}: Returns the frame image files
#'       without combining them.
#'   }
#'   Any function with signature \code{function(frames, fps)} can be used.
#' @param start_pause,end_pause Number of frames to hold at the beginning
#'   and end of the animation. Default is 0.
#' @param rewind Logical. If \code{TRUE}, the animation plays in reverse
#'   after completing, creating a seamless loop. Default is \code{FALSE}.
#' @param cores Number of CPU cores for parallel frame rendering. Default
#'   is 1 (sequential). Values greater than 1 use [parallel::parLapply()]
#'   to render frames in parallel, which can significantly speed up
#'   animation of complex plots. Note that progress reporting is not
#'   available during parallel rendering.
#' @param device The graphics device to use for rendering frames.
#'   Default is \code{"png"}. Other options include \code{"ragg_png"}
#'   (requires the \pkg{ragg} package).
#' @param progress Whether to print a progress bar. Ignored when `cores > 1`.
#'
#' @return The return value of the renderer function. For
#'   \code{gifski_renderer_3d()}, a \code{gif_3d} object (a file path with
#'   class attributes for display in RStudio and knitr). For
#'   \code{av_renderer_3d()}, a \code{video_3d} object (a file path with
#'   class attributes for display). For \code{file_renderer_3d()}, a
#'   character vector of file paths to the rendered frame images.
#'
#' @examplesIf requireNamespace("gifski", quietly = TRUE)
#' \donttest{
#' # Plot to animate. Note `anchor = "camera"` keeps light source from rotating.
#' p <- ggplot() +
#'   geom_function_3d(
#'     fun = function(x, y) sin(x) * cos(y),
#'     xlim = c(-pi, pi), ylim = c(-pi, pi),
#'     fill = "steelblue", color = "steelblue", linewidth = .5) +
#'   coord_3d(light = light(mode = "hsl", anchor = "camera",
#'                          direction = c(1, 1, 0))) +
#'   theme_void()
#'
#' # Simple turntable rotation
#' # (using small `nframes` to minimize runtime)
#' animate_3d(p, yaw = c(-30, 330), nframes = 12)
#'
#' # Multi-segment orbit with pitch change
#' animate_3d(p, yaw = c(0, 720), roll = c(-90, 0, -90),
#'            nframes = 12, fps = 15)
#'
#' # Save to file (writes to a temporary path; specify your own to keep it)
#' anim <- animate_3d(p, yaw = c(0, 360), nframes = 12)
#' anim_save_3d(anim, file.path(tempdir(), "rotating_surface.gif"))
#' }
#'
#' @seealso [anim_save_3d()] for saving animations to file.
#'   [gifski_renderer_3d()] and [file_renderer_3d()] for renderer options.
#'   [orbit_3d()] for interactive drag-to-rotate ggcubes.
#' @export
animate_3d <- function(plot,
                       pitch = NULL,
                       roll = NULL,
                       yaw = NULL,
                       nframes = NULL,
                       fps = NULL,
                       duration = NULL,
                       width = 480,
                       height = 480,
                       res = 96,
                       renderer = gifski_renderer_3d(),
                       start_pause = 0,
                       end_pause = 0,
                       rewind = FALSE,
                       cores = 1,
                       device = "png",
                       progress = interactive()) {

      # --- Validate input plot ---
      if (!inherits(plot, "gg")) {
            stop("`plot` must be a ggplot object.", call. = FALSE)
      }
      coord <- plot$coordinates
      if (!inherits(coord, "Coord3D")) {
            stop("`plot` must use coord_3d(). ",
                 "Add `+ coord_3d()` to your plot before animating.",
                 call. = FALSE)
      }

      # --- Validate cores ---
      cores <- as.integer(cores)
      if (cores < 1) stop("`cores` must be at least 1.", call. = FALSE)

      # --- Resolve nframes / fps / duration ---
      timing <- resolve_timing(nframes, fps, duration)
      nframes <- timing$nframes
      fps <- timing$fps

      # --- Build per-frame parameter sequences ---
      param_seqs <- build_param_sequences(
            coord = coord,
            pitch = pitch,
            roll = roll,
            yaw = yaw,
            nframes = nframes
      )

      # --- Apply start_pause, end_pause, rewind ---
      param_seqs <- apply_pause_and_rewind(param_seqs, start_pause, end_pause,
                                           rewind)

      # --- Render frames to PNG files ---
      frames <- render_3d_frames(
            plot = plot,
            coord = coord,
            param_seqs = param_seqs,
            width = width,
            height = height,
            res = res,
            device = device,
            cores = cores,
            progress = progress,
            prefix = "ggcube_anim_"
      )

      message("Assembling animation...")

      # --- Combine frames via renderer ---
      result <- renderer(frames$files, fps, width, height)

      # Clean up temp files (unless using file_renderer)
      if (!inherits(result, "frame_files_3d")) {
            unlink(frames$tmpdir, recursive = TRUE)
      }

      # Stash as last animation for anim_save_3d()
      .ggcube_last_anim$value <- result

      result
}


# Frame rendering ---------------------------------------------------------

#' Render a sequence of camera states to PNG frame files
#'
#' Shared rendering core for [animate_3d()] and [orbit_3d()]. Takes a
#' fully-resolved sequence of per-frame rotation parameters (one row per
#' frame) and renders each to a PNG on disk, computing a single fixed
#' viewport across all frames so the plot does not rescale between rotations.
#'
#' This is the \code{[resolved camera states -> image files]} seam: callers
#' are responsible for producing \code{param_seqs} (including any
#' pause/rewind/keyframe expansion), and receive back the list of rendered
#' files plus the temp directory holding them.
#'
#' @param plot A ggplot object using [coord_3d()].
#' @param coord The plot's Coord3D object (\code{plot$coordinates}).
#' @param param_seqs A data.frame with columns \code{pitch}, \code{roll},
#'   \code{yaw}; one row per frame.
#' @param width,height,res,device,cores,progress As in [animate_3d()].
#' @param prefix Prefix for the temporary frame directory.
#'
#' @return A list with elements \code{files} (character vector of PNG paths),
#'   \code{tmpdir} (the directory containing them), and \code{width},
#'   \code{height}.
#' @keywords internal
#' @noRd
render_3d_frames <- function(plot, coord, param_seqs, width, height, res,
                             device, cores, progress,
                             prefix = "ggcube_frames_") {
      total_frames <- nrow(param_seqs)

      # --- Set up output directory ---
      tmpdir <- tempfile(prefix)
      dir.create(tmpdir)
      frame_files <- file.path(tmpdir,
                               sprintf("frame_%04d.png", seq_len(total_frames)))

      # Extract static coord params to carry through every frame
      static_params <- extract_static_coord_params(coord)

      # --- Build plot once to extract actual data bounds and train z scale ---
      # This is needed because:
      # 1. Parallel workers need the trained z scale
      # 2. compute_global_bounds needs the actual standardized data extents,
      #    not just the theoretical [-0.5, 0.5] cube
      built <- ggplot_build(plot)

      # Extract the effective ratios from the built plot's panel_params
      # These account for both the ratio setting and any data-dependent scaling
      panel_params <- built$layout$panel_params[[1]]
      effective_ratios <- if (!is.null(panel_params$effective_ratios)) {
            panel_params$effective_ratios
      } else {
            static_params$ratio
      }

      # --- Compute global bounds across all frames ---
      # This prevents the plot from rescaling at each rotation angle.
      global_bounds <- compute_global_bounds(param_seqs, static_params,
                                             effective_ratios)

      # Apply zoom to global bounds
      zoom <- static_params$zoom %||% 1
      if (zoom != 1) {
            x_center <- mean(global_bounds[1:2])
            y_center <- mean(global_bounds[3:4])
            x_half <- diff(global_bounds[1:2]) / 2 / zoom
            y_half <- diff(global_bounds[3:4]) / 2 / zoom
            global_bounds <- c(
                  x_center - x_half, x_center + x_half,
                  y_center - y_half, y_center + y_half
            )
      }

      # --- Render frames ---
      if (cores > 1) {
            render_frames_parallel(plot, param_seqs, static_params, global_bounds,
                                   frame_files, width, height, res, device, cores)
      } else {
            render_frames_sequential(plot, param_seqs, static_params, global_bounds,
                                     frame_files, width, height, res, device,
                                     progress)
      }

      list(files = frame_files, tmpdir = tmpdir, width = width, height = height)
}


#' Render frames sequentially with progress bar
#' @keywords internal
#' @noRd
render_frames_sequential <- function(plot, param_seqs, static_params,
                                     global_bounds, frame_files,
                                     width, height, res, device, progress) {
      total_frames <- nrow(param_seqs)
      message(sprintf("Rendering %d frames...", total_frames))

      pb <- NULL
      if (progress && requireNamespace("progress", quietly = TRUE)) {
            pb <- progress::progress_bar$new(
                  format = "  frame :current/:total [:bar] :percent eta: :eta",
                  total = total_frames, clear = FALSE, width = 60
            )
      }

      for (i in seq_len(total_frames)) {
            if (!is.null(pb)) pb$tick()
            render_single_frame(plot, param_seqs, static_params, global_bounds,
                                frame_files[i], i, width, height, res, device)
      }
}


#' Render frames in parallel using a socket cluster
#' @keywords internal
#' @noRd
render_frames_parallel <- function(plot, param_seqs, static_params,
                                   global_bounds, frame_files,
                                   width, height, res, device, cores) {
      total_frames <- nrow(param_seqs)
      cores <- min(cores, total_frames)
      message(sprintf("Rendering %d frames across %d cores...",
                      total_frames, cores))

      cl <- parallel::makeCluster(cores)
      on.exit(parallel::stopCluster(cl), add = TRUE)

      # Load ggcube (which imports ggplot2) on each worker
      parallel::clusterEvalQ(cl, {
            library(ggcube)
      })

      # Export objects needed by the worker function. The z scale rides along
      # inside `plot`, so each worker's own build resolves it.
      parallel::clusterExport(cl, c("plot", "param_seqs", "static_params",
                                    "global_bounds", "frame_files",
                                    "width", "height", "res",
                                    "device"),
                              envir = environment())

      # Render in parallel. The worker function is defined inline so that
      # it closes over the exported variables directly, avoiding the need
      # to locate unexported package functions from the worker processes.
      parallel::parLapply(cl, seq_len(total_frames), function(i) {
            frame_coord_args <- modifyList(
                  static_params,
                  list(
                        pitch = param_seqs$pitch[i],
                        roll = param_seqs$roll[i],
                        yaw = param_seqs$yaw[i],
                        fixed_bounds = global_bounds
                  )
            )
            frame_coord <- do.call(ggcube::coord_3d, frame_coord_args)

            frame_plot <- plot
            frame_plot$coordinates <- frame_coord[[1]]

            device_fun <- switch(device,
                                 png = grDevices::png,
                                 ragg_png = ragg::agg_png,
                                 stop("Unsupported device: ", device, call. = FALSE)
            )

            device_fun(filename = frame_files[i],
                       width = width, height = height, res = res)
            on.exit(grDevices::dev.off(), add = TRUE)
            print(frame_plot)
      })

      invisible(NULL)
}


#' Render a single animation frame
#'
#' Builds the coord for frame \code{i}, swaps it onto the plot, and writes
#' the PNG. Used by both sequential and parallel rendering paths.
#' @keywords internal
#' @noRd
render_single_frame <- function(plot, param_seqs, static_params, global_bounds,
                                file, i, width, height, res, device) {
      frame_coord_args <- modifyList(
            static_params,
            list(
                  pitch = param_seqs$pitch[i],
                  roll = param_seqs$roll[i],
                  yaw = param_seqs$yaw[i],
                  fixed_bounds = global_bounds
            )
      )
      frame_coord <- do.call(coord_3d, frame_coord_args)

      # coord_3d() returns a list of (ggproto, theme); we only need the
      # ggproto since the theme is already applied from the original plot.
      frame_plot <- plot
      frame_plot$coordinates <- frame_coord[[1]]

      render_frame(frame_plot, file, width, height, res, device)
}


# Internal helpers ---------------------------------------------------------

#' Resolve nframes / fps / duration (any 2 of 3)
#' @keywords internal
#' @noRd
resolve_timing <- function(nframes, fps, duration) {
      provided <- c(!is.null(nframes), !is.null(fps), !is.null(duration))

      if (sum(provided) > 2) {
            stop("Specify at most two of `nframes`, `fps`, and `duration`.",
                 call. = FALSE)
      }

      # Apply defaults when fewer than 2 are specified
      if (sum(provided) == 0) {
            nframes <- 100
            fps <- 10
      } else if (sum(provided) == 1) {
            if (is.null(fps)) fps <- 10
            if (is.null(nframes) && is.null(duration)) nframes <- 100
      }

      # Compute the missing value
      if (is.null(duration)) {
            duration <- nframes / fps
      } else if (is.null(nframes)) {
            nframes <- round(duration * fps)
      } else if (is.null(fps)) {
            fps <- nframes / duration
      }

      if (nframes < 1) stop("`nframes` must be at least 1.", call. = FALSE)
      if (fps <= 0) stop("`fps` must be positive.", call. = FALSE)

      list(nframes = as.integer(nframes), fps = fps, duration = duration)
}


#' Build parameter sequences for each frame
#' @keywords internal
#' @noRd
build_param_sequences <- function(coord, pitch, roll, yaw, nframes) {

      make_seq <- function(user_val, coord_val, n) {
            if (is.null(user_val)) {
                  # Use coord's value for all frames
                  rep(coord_val, n)
            } else if (length(user_val) == 1) {
                  # Single value: hold constant
                  rep(user_val, n)
            } else {
                  # Multiple values: piecewise linear interpolation
                  # Keyframes are evenly spaced across the animation
                  n_keyframes <- length(user_val)
                  keyframe_positions <- seq(1, n, length.out = n_keyframes)

                  # Interpolate to all frames
                  stats::approx(
                        x = keyframe_positions,
                        y = user_val,
                        xout = seq_len(n),
                        method = "linear"
                  )$y
            }
      }

      data.frame(
            pitch = make_seq(pitch, coord$pitch, nframes),
            roll  = make_seq(roll,  coord$roll,  nframes),
            yaw   = make_seq(yaw,   coord$yaw,   nframes)
      )
}


#' Apply start_pause, end_pause, and rewind to parameter sequences
#' @keywords internal
#' @noRd
apply_pause_and_rewind <- function(param_seqs, start_pause, end_pause, rewind) {

      if (rewind) {
            # Append reversed frames (excluding the endpoints to avoid stutter)
            reversed <- param_seqs[rev(seq_len(nrow(param_seqs))), , drop = FALSE]
            if (nrow(reversed) > 2) {
                  reversed <- reversed[-c(1, nrow(reversed)), , drop = FALSE]
            }
            param_seqs <- rbind(param_seqs, reversed)
      }

      if (start_pause > 0) {
            first_row <- param_seqs[rep(1, start_pause), , drop = FALSE]
            param_seqs <- rbind(first_row, param_seqs)
      }

      if (end_pause > 0) {
            last_row <- param_seqs[rep(nrow(param_seqs), end_pause), , drop = FALSE]
            param_seqs <- rbind(param_seqs, last_row)
      }

      rownames(param_seqs) <- NULL
      param_seqs
}


#' Compute global plot bounds across all animation frames
#'
#' Projects the 8 cube corners at every unique rotation in the animation,
#' takes the union envelope, and adds padding. This gives a fixed viewport
#' that prevents the plot from rescaling as the rotation changes.
#' @keywords internal
#' @noRd
compute_global_bounds <- function(param_seqs, static_params, effective_ratios) {
      # Build cube corners using the effective ratios from the built plot.
      # These ratios account for both the user's ratio setting and any
      # data-dependent scaling (e.g., from scales = "fixed").
      # Data in the standardized domain lives in [-0.5, 0.5] * effective_ratios.
      corners <- as.matrix(expand.grid(
            x = c(-0.5, 0.5) * effective_ratios[1],
            y = c(-0.5, 0.5) * effective_ratios[2],
            z = c(-0.5, 0.5) * effective_ratios[3]
      ))

      # Find unique rotation combinations to avoid redundant projections
      unique_params <- unique(param_seqs)

      # Collect all projected x/y across all frames
      all_x <- numeric(0)
      all_y <- numeric(0)

      for (i in seq_len(nrow(unique_params))) {
            proj <- list(
                  pitch = unique_params$pitch[i],
                  roll = unique_params$roll[i],
                  yaw = unique_params$yaw[i],
                  persp = static_params$persp,
                  dist = static_params$dist
            )
            projected <- transform_3d_standard(
                  as.data.frame(corners), proj
            )
            all_x <- c(all_x, projected$x)
            all_y <- c(all_y, projected$y)
      }

      # Compute bounds with generous padding (15%) to accommodate labels
      x_range <- range(all_x)
      y_range <- range(all_y)
      x_pad <- diff(x_range) * 0.15
      y_pad <- diff(y_range) * 0.15

      c(x_range[1] - x_pad, x_range[2] + x_pad,
        y_range[1] - y_pad, y_range[2] + y_pad)
}


#' Extract static (non-animated) coord_3d parameters
#' @keywords internal
#' @noRd
extract_static_coord_params <- function(coord) {
      list(
            persp = coord$persp,
            dist = coord$dist,
            expand = coord$expand,
            clip = coord$clip,
            panels = coord$panels,
            xlabels = coord$xlabels,
            ylabels = coord$ylabels,
            zlabels = coord$zlabels,
            rotate_labels = coord$rotate_labels,
            title_position = coord$title_position,
            scales = coord$scales,
            scale_depth = coord$scale_depth,
            ratio = coord$ratio,
            zoom = coord$zoom,
            light = coord$light
      )
}


#' Render a single frame to a PNG file
#' @keywords internal
#' @noRd
render_frame <- function(plot, file, width, height, res, device) {
      device_fun <- switch(device,
                           png = grDevices::png,
                           ragg_png = {
                                 if (!requireNamespace("ragg", quietly = TRUE)) {
                                       stop("The ragg package is required for device = 'ragg_png'.",
                                            call. = FALSE)
                                 }
                                 ragg::agg_png
                           },
                           stop("Unsupported device: ", device, call. = FALSE)
      )

      device_fun(filename = file, width = width, height = height, res = res)
      on.exit(grDevices::dev.off(), add = TRUE)

      # Build gtable after opening our device so ggplot2 uses our dimensions
      gt <- ggplot2::ggplotGrob(plot)
      grid::grid.draw(gt)
}


# Renderers ---------------------------------------------------------

#' Animation renderers for animate_3d
#'
#' Renderer functions that combine individual frame images into a final
#' animation file. These are factory functions that return the actual
#' rendering function.
#'
#' @param file Output file path. If \code{NULL}, a temporary file is used.
#' @param loop Logical. Should the GIF loop? Default is \code{TRUE}.
#' @param dir Directory to copy frame files into. Defaults to a session
#'   temporary directory (\code{tempdir()}). For animations you want to keep,
#'   pass an explicit directory.
#' @param prefix Filename prefix for copied frame files.
#' @param overwrite Logical. Overwrite existing files? Default is \code{FALSE}.
#' @param vfilter A video filter string for ffmpeg (passed to \pkg{av}).
#' @param codec Video codec name. Default lets \pkg{av} choose.
#'
#' @return A function with signature \code{function(frames, fps, width, height)}
#'   that produces the animation output.
#'
#' @examplesIf requireNamespace("gifski", quietly = TRUE) && requireNamespace("av", quietly = TRUE)
#' \donttest{
#' p <- ggplot() +
#'   geom_function_3d(
#'     fun = function(x, y) sin(x) * cos(y),
#'     xlim = c(-pi, pi), ylim = c(-pi, pi),
#'     fill = "steelblue", color = "steelblue") +
#'   coord_3d()
#'
#' # GIF output (default)
#' animate_3d(p, yaw = c(0, 360), renderer = gifski_renderer_3d())
#'
#' # Video output
#' animate_3d(p, yaw = c(0, 360), renderer = av_renderer_3d())
#'
#' # Keep individual frame files in a chosen directory
#' animate_3d(p, yaw = c(0, 360),
#'     renderer = file_renderer_3d(dir = file.path(tempdir(), "my_frames")))
#' }
#'
#' @name renderers_3d
NULL

#' @rdname renderers_3d
#' @export
gifski_renderer_3d <- function(file = NULL, loop = TRUE) {
      function(frames, fps, width, height) {
            if (!requireNamespace("gifski", quietly = TRUE)) {
                  stop("The gifski package is required for GIF output. ",
                       "Install it with install.packages('gifski').",
                       call. = FALSE)
            }

            if (is.null(file)) {
                  file <- tempfile(fileext = ".gif")
            }

            gifski::gifski(frames, gif_file = file,
                           width = width, height = height,
                           delay = 1 / fps, loop = loop,
                           progress = FALSE)

            gif_3d(file, width = width, height = height)
      }
}

#' @rdname renderers_3d
#' @export
av_renderer_3d <- function(file = NULL, vfilter = "null", codec = NULL) {
      function(frames, fps, width, height) {
            if (!requireNamespace("av", quietly = TRUE)) {
                  stop("The av package is required for video output. ",
                       "Install it with install.packages('av').",
                       call. = FALSE)
            }

            if (is.null(file)) {
                  file <- tempfile(fileext = ".mp4")
            }

            # av auto-detects dimensions from input frames
            av::av_encode_video(frames, output = file, framerate = fps,
                                vfilter = vfilter, codec = codec)

            video_3d(file, width = width, height = height)
      }
}

#' @rdname renderers_3d
#' @export
file_renderer_3d <- function(dir = tempdir(), prefix = "ggcube_frame",
                             overwrite = FALSE) {
      function(frames, fps, width, height) {
            if (!dir.exists(dir)) {
                  dir.create(dir, showWarnings = FALSE, recursive = TRUE)
            }
            new_names <- file.path(dir,
                                   paste0(prefix, "_",
                                          sprintf("%04d", seq_along(frames)),
                                          ".png"))
            if (any(file.exists(new_names)) && !overwrite) {
                  stop("Output files already exist. Use `overwrite = TRUE` ",
                       "to replace them.", call. = FALSE)
            }
            file.copy(frames, new_names, overwrite = overwrite)

            result <- new_names
            class(result) <- c("frame_files_3d", "character")
            result
      }
}


# Output classes  ---------------------------------------------------------

#' Create a gif_3d output object
#' @param path Path to the GIF file.
#' @param width,height Dimensions of the GIF in pixels.
#' @return A \code{gif_3d} object.
#' @keywords internal
#' @noRd
gif_3d <- function(path, width = NULL, height = NULL) {
      structure(path,
                class = c("gif_3d", "character"),
                width = width,
                height = height)
}

#' @export
print.gif_3d <- function(x, ...) {
      viewer <- getOption("viewer")
      if (!is.null(viewer)) {
            viewer(x)
      } else {
            utils::browseURL(x)
      }
      invisible(x)
}

#' @exportS3Method knitr::knit_print gif_3d
knit_print.gif_3d <- function(x, options, ...) {
      # If the GIF is in a tempdir, copy it to knitr's figures directory
      # so it persists in the rendered output
      if (grepl(tempdir(), x, fixed = TRUE)) {
            fig_path <- knitr::fig_path(".gif")
            dir.create(dirname(fig_path), showWarnings = FALSE, recursive = TRUE)
            file.copy(x, fig_path, overwrite = TRUE)
            path <- fig_path
      } else {
            path <- x
      }

      # Use stored dimensions if chunk options don't override
      if (is.null(options$out.width) && !is.null(attr(x, "width"))) {
            options$out.width <- paste0(attr(x, "width"), "px")
      }
      if (is.null(options$out.height) && !is.null(attr(x, "height"))) {
            options$out.height <- paste0(attr(x, "height"), "px")
      }

      knitr::knit_print(knitr::include_graphics(path), options, ...)
}


#' Create a video_3d output object
#' @param path Path to the video file.
#' @param width,height Dimensions of the video in pixels.
#' @return A \code{video_3d} object.
#' @keywords internal
#' @noRd
video_3d <- function(path, width = NULL, height = NULL) {
      structure(path,
                class = c("video_3d", "character"),
                width = width,
                height = height)
}

#' @export
print.video_3d <- function(x, ...) {
      utils::browseURL(x)
      invisible(x)
}

#' @exportS3Method knitr::knit_print video_3d
knit_print.video_3d <- function(x, options, ...) {
      if (requireNamespace("base64enc", quietly = TRUE)) {
            ext <- tools::file_ext(x)
            mime <- switch(ext,
                           mp4 = "video/mp4",
                           webm = "video/webm",
                           "video/mp4")
            b64 <- base64enc::base64encode(x)
            html <- sprintf(
                  '<video controls autoplay loop><source src="data:%s;base64,%s" type="%s"></video>',
                  mime, b64, mime
            )
            knitr::asis_output(html)
      } else {
            # Fallback: copy tempfile to figures directory if needed
            if (grepl(tempdir(), x, fixed = TRUE)) {
                  fig_path <- knitr::fig_path(paste0(".", tools::file_ext(x)))
                  dir.create(dirname(fig_path), showWarnings = FALSE, recursive = TRUE)
                  file.copy(x, fig_path, overwrite = TRUE)
                  x <- fig_path
            }
            knitr::knit_print(knitr::include_graphics(x), options, ...)
      }
}



# Save helper  ---------------------------------------------------------

# Environment to store the last rendered animation
.ggcube_last_anim <- new.env(parent = emptyenv())
.ggcube_last_anim$value <- NULL

#' Save a 3D animation to a file
#'
#' Saves an animation object produced by [animate_3d()] to a file.
#' If no animation is provided, saves the most recently rendered animation.
#'
#' @param animation An animation object from [animate_3d()], or \code{NULL}
#'   to use the last rendered animation.
#' @param filename Output file path.
#'
#' @return Invisibly returns the output file path (a character string).
#'   Called primarily for its side effect of copying the animation to
#'   \code{filename}.
#'
#' @examplesIf requireNamespace("gifski", quietly = TRUE)
#' \donttest{
#' p <- ggplot() +
#'   geom_function_3d(
#'     fun = function(x, y) sin(x) * cos(y),
#'     xlim = c(-pi, pi), ylim = c(-pi, pi),
#'     fill = "steelblue", color = "steelblue") +
#'   coord_3d()
#'
#' animate_3d(p, yaw = c(0, 360))
#' anim_save_3d(filename = file.path(tempdir(), "my_animation.gif"))
#' }
#'
#' @export
anim_save_3d <- function(animation = NULL, filename) {
      if (is.null(animation)) {
            animation <- .ggcube_last_anim$value
            if (is.null(animation)) {
                  stop("No animation to save. Run `animate_3d()` first.",
                       call. = FALSE)
            }
      }

      if (inherits(animation, "frame_files_3d")) {
            stop("Cannot save frame files as a single animation. ",
                 "Use `gifski_renderer_3d()` or `av_renderer_3d()` instead.",
                 call. = FALSE)
      }

      file.copy(as.character(animation), filename, overwrite = TRUE)
      message("Animation saved to ", filename)
      invisible(filename)
}

# Geom that converts point grid to ridgeline polygons for Joy Division style rendering

GeomRidgeline3D <- ggproto("GeomRidgeline3D", GeomPolygon3D,

                           required_aes = c("x", "y", "z"),

                           default_aes = aes(
                                 colour = "black",
                                 fill = "white",
                                 linewidth = 0.1,
                                 linetype = 1,
                                 alpha = NA
                           ),

                           extra_params = c("na.rm", "annotate", "direction", "base", "cull_backfaces",
                                            "sort_method", "scale_depth", "force_convex"),

                           setup_params = function(data, params) {
                                 if(is.null(params$direction) && ".direction" %in% names(data)) params$direction <- data$.direction[1]
                                 params$direction <- params$direction %||% "x"
                                 params$direction <- match.arg(params$direction, c("x", "y"))

                                 # make pairwise the default, as painter is so often problematic for ridges
                                 if(is.null(params$sort_method)) params$sort_method <- "pairwise"
                                 params
                           },

                           setup_data = function(data, params) {
                                 # Validate grid structure
                                 n_unique_x <- length(unique(round(data$x, 10)))
                                 n_unique_y <- length(unique(round(data$y, 10)))

                                 if (params$direction == "x" && n_unique_y < 2) {
                                       stop("geom_ridgeline_3d() with direction='x' requires at least 2 unique y values")
                                 }
                                 if (params$direction == "y" && n_unique_x < 2) {
                                       stop("geom_ridgeline_3d() with direction='y' requires at least 2 unique x values")
                                 }

                                 # Warn if many slices
                                 slice_var <- if (params$direction == "x") "x" else "y"
                                 n_slices <- length(unique(round(data[[slice_var]], 10)))
                                 if (n_slices > 100) {
                                       message("Creating ", n_slices, " ridgeline polygons. ",
                                               "Consider reducing grid resolution.")
                                 }

                                 # Cleanup
                                 if (".direction" %in% names(data)) data <- select(data, -.direction)

                                 # Convert to ridgeline polygons
                                 data <- points_to_ridgelines(
                                       data,
                                       direction = params$direction,
                                       base = params$base,
                                       group_prefix = "ridgeline__"
                                 )

                                 setup_annotations(data, params)
                           }

                           # draw_panel inherited from GeomPolygon3D
)


#' 3D ridgeline plot from point grid
#'
#' Renders a point grid as ridgeline polygons (Joy Division style). Each slice
#' along one axis becomes a filled polygon, creating a cross-sectional view
#' of the surface.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data Point grid data with x, y, z coordinates.
#' @param stat Statistical transformation. Defaults to "identity".
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed to the layer.
#' @inheritParams ridgeline_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#'
#' @return A `Layer` object that can be added to a ggplot.
#'
#' @section Aesthetics:
#' \describe{
#'   \item{x, y, z}{Point coordinates (required)}
#'   \item{fill}{Polygon fill color (default: "grey70")}
#'   \item{colour}{Polygon border color (default: "black")}
#'   \item{alpha}{Transparency}
#'   \item{linewidth}{Border width}
#'   \item{linetype}{Border type}
#' }
#'
#' @examples
#' # From point grid
#' grid_data <- expand.grid(x = seq(-3, 3, 0.5), y = seq(-3, 3, 0.1))
#' grid_data$z <- with(grid_data, dnorm(y) * dnorm(x) * 10)
#'
#' ggplot(grid_data, aes(x, y, z = z)) +
#'   geom_ridgeline_3d() +
#'   coord_3d()
#'
#' # With fill
#' ggplot(grid_data, aes(x, y, z = z, fill = x)) +
#'   geom_ridgeline_3d(colour = "white", linewidth = 0.3) +
#'   scale_fill_viridis_c() +
#'   coord_3d()
#'
#' # Ridges in y direction
#' ggplot(grid_data, aes(x, y, z = z)) +
#'   geom_ridgeline_3d(direction = "y", fill = "steelblue") +
#'   coord_3d()
#'
#' # With stat_function_3d
#' ggplot() +
#'   stat_function_3d(fun = function(x, y) sin(x) * cos(y),
#'                    xlim = c(-pi, pi), ylim = c(-pi, pi),
#'                    n = 30,
#'                    geom = "ridgeline_3d") +
#'   coord_3d()
#'
#' @seealso [geom_surface_3d()], [stat_function_3d()], [coord_3d()]
#' @export
geom_ridgeline_3d <- function(mapping = NULL, data = NULL,
                              stat = "identity_3d", position = "identity",
                              ...,
                              direction = "x",
                              base = NULL,
                              cull_backfaces = FALSE,
                              sort_method = "pairwise",
                              scale_depth = TRUE,
                              force_convex = FALSE,
                              light = NULL,
                              na.rm = FALSE,
                              show.legend = NA,
                              inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = get_proto(stat, "stat"),
            geom = GeomRidgeline3D,
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  direction = direction,
                  base = base,
                  cull_backfaces = cull_backfaces,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  force_convex = force_convex,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}

#' Generate grid data using actual scale breaks with aspect ratio
#'
#' @param visible_faces Character vector of visible face names
#' @param scale_info List containing limits and breaks for x, y, z
#' @param scales Aspect ratio behavior ("free" or "fixed")
#' @param ratio Length-3 numeric vector of axis ratios
#' @return Data frame with grid lines in standard domain, including break values
#' @keywords internal
#' @noRd
make_scale_grid <- function(visible_faces, scale_info, scales = "free", ratio = c(1, 1, 1)) {

      # Calculate effective ratios using scale ranges (includes expansion)
      scale_ranges <- list(
            x = scale_info$x$limits,
            y = scale_info$y$limits,
            z = scale_info$z$limits
      )
      effective_ratios <- compute_effective_ratios(scale_ranges, scales, ratio)

      # Breaks and extents in standard domain
      vars <- c("x", "y", "z") %>% setNames(., .)
      breaks <- lapply(vars, function(v) breaks = scale_to_standard(scale_info[[v]]$breaks,
                                                                    scale_info[[v]]$limits) * effective_ratios[match(v, vars)])
      extents <- data.frame(x = c(-0.5, 0.5) * effective_ratios[1],
                            y = c(-0.5, 0.5) * effective_ratios[2],
                            z = c(-0.5, 0.5) * effective_ratios[3],
                            end = c("min", "max"))

      grid_data <- map_dfr(visible_faces, function(face){
            axis <- substr(face, 1, 1)
            end <- substr(face, 2, 4)
            fixed_val = extents[extents$end == end, axis]
            vrs <- setdiff(vars, axis)
            map_dfr(vrs, function(v){

                  brks <- scale_info[[v]]$breaks
                  break_value <- brks
                  if(is.numeric(break_value)){
                        break_pos <- break_value
                  }else{
                        break_pos <- attr(break_value, "pos")
                  }

                  b <- breaks[[v]]
                  brks <- data.frame(b = b,
                                     break_value = as.character(break_value),
                                     break_pos = break_pos)

                  vr <- setdiff(vrs, v)
                  expand_grid(f = fixed_val,
                              b = b,
                              e = extents[[vr]]) %>%
                        left_join(brks, by = join_by(b)) %>%
                        mutate(endpoint_index = as.integer(factor(e))) %>%
                        setNames(c(axis, v, vr, names(.)[4:ncol(.)])) %>%
                        mutate(face = face,
                               break_axis = v,
                               free_axis = vr)
            })
      }) %>%
            mutate(group = rep(1:(nrow(.)/2), each = 2)) %>%
            group_by(group) %>%
            mutate(start_boundaries = paste0(pmin(face, paste0(free_axis, "min")), "_",
                                             pmax(face, paste0(free_axis, "min"))),
                   end_boundaries = paste0(pmin(face, paste0(free_axis, "max")), "_",
                                           pmax(face, paste0(free_axis, "max")))) %>%
            ungroup()

      return(grid_data)
}

# Helper function to determine endpoint preference using boundary information
determine_endpoint_preference_by_boundary <- function(chosen_edge, gridlines) {
      # Check if boundary columns exist
      if (!("start_boundaries" %in% names(gridlines)) || !("end_boundaries" %in% names(gridlines))) {
            # Fallback to original distance-based method if boundary info is missing
            return(TRUE)  # Default to start
      }

      # Get the boundary signature for the chosen edge (touching faces)
      if (is.null(chosen_edge$touching_faces) || length(chosen_edge$touching_faces) == 0) {
            return(TRUE)  # Default to start
      }

      chosen_boundary <- paste(sort(chosen_edge$touching_faces), collapse = "_")

      start_preferences <- 0
      end_preferences <- 0

      for (group_id in unique(gridlines$group)) {
            group_data <- gridlines[gridlines$group == group_id, ]

            if (nrow(group_data) >= 2) {
                  start_boundary <- group_data$start_boundaries[1]
                  end_boundary <- group_data$end_boundaries[1]

                  # Check for NA or empty values
                  if (!is.na(start_boundary) && !is.na(end_boundary) &&
                      length(start_boundary) > 0 && length(end_boundary) > 0) {

                        # Check which endpoint's boundary matches the chosen edge's boundary
                        if (start_boundary == chosen_boundary) {
                              start_preferences <- start_preferences + 1
                        }
                        if (end_boundary == chosen_boundary) {
                              end_preferences <- end_preferences + 1
                        }
                  }
            }
      }

      # Return TRUE if majority prefer start, FALSE if majority prefer end
      return(start_preferences >= end_preferences)
}


make_face_panels <- function(visible_faces, scales = "free", ratio = c(1, 1, 1), scale_ranges = NULL){
      if (length(visible_faces) == 0) return(NULL)

      # Calculate effective ratios using scale ranges (includes expansion)
      effective_ratios <- compute_effective_ratios(scale_ranges, scales, ratio)

      panels <- list()
      corners <- expand.grid(x = c(-0.5, 0.5) * effective_ratios[1],
                             y = c(-0.5, 0.5) * effective_ratios[2],
                             z = c(-0.5, 0.5) * effective_ratios[3])

      for(face in visible_faces){
            # Extract the axis and min/max from face name
            face_axis <- substr(face, 1, 1)
            face_value <- ifelse(substr(face, 2, 4) == "min", -0.5, 0.5)

            # Find which axis index this corresponds to
            axis_index <- which(c("x", "y", "z") == face_axis)
            face_value <- face_value * effective_ratios[axis_index]

            # Select corners where this axis equals the face value
            crnrs <- corners[corners[face_axis] == face_value, ]
            crnrs$face <- face
            crnrs <- crnrs[c(1, 3, 4, 2)]
            panels[[length(panels) + 1]] <- crnrs
      }

      if(length(panels) == 0) return(NULL)
      do.call(rbind, panels)[,c(1,2,4,3)]
}


#' Create grid segments for rendering
#'
#' @param grid_data Transformed grid data
#' @param plot_bounds Bounds for the final plot `xmin, xmax, ymin, ymax`
#' @return Data frame with segment coordinates for grid grob
#' @keywords internal
#' @noRd
create_grid_segments <- function(grid_data, plot_bounds) {
      if (is.null(grid_data) || nrow(grid_data) == 0) {
            return(NULL)
      }

      # Scale to [0, 1] based on plot bounds
      x_scaled <- (grid_data$x - plot_bounds[1]) / (plot_bounds[2] - plot_bounds[1])
      y_scaled <- (grid_data$y - plot_bounds[3]) / (plot_bounds[4] - plot_bounds[3])

      # Create segments by connecting consecutive points in each group
      segments <- data.frame()
      for (g in unique(grid_data$group)) {
            group_data <- grid_data[grid_data$group == g, ]
            group_x <- x_scaled[grid_data$group == g]
            group_y <- y_scaled[grid_data$group == g]

            if (length(group_x) >= 2) {
                  for (i in 1:(length(group_x) - 1)) {
                        segments <- rbind(segments, data.frame(
                              x0 = group_x[i],
                              y0 = group_y[i],
                              x1 = group_x[i + 1],
                              y1 = group_y[i + 1],
                              depth = group_data$z_proj[i],
                              depth_scale = mean(group_data$depth_scale[i:(i+1)]),
                              face = group_data$face[i]
                        ))
                  }
            }
      }

      # Sort by depth (back to front)
      if (nrow(segments) > 0) {
            segments <- segments[order(-segments$depth), ]
      }

      return(segments)
}



create_panel_polygons = function(face_corners_transformed, panel_params, theme, layer = "background") {
      if (is.null(face_corners_transformed) || nrow(face_corners_transformed) == 0) {
            return(NULL)
      }

      # Get theme elements for panel styling
      panel_bg <- calc_element(switch(layer, background = "panel.background", foreground = "panel.foreground"), theme)
      panel_border <- calc_element("panel.border", theme)

      # Extract styling from theme with proper fallbacks
      # Priority: panel.border properties > panel.background properties > defaults

      # Fill color: prefer panel.background fill
      fill_color <- panel_bg$fill %||% "white"

      # Border: prefer panel.border if it exists, fallback to panel.background
      if (!is.null(panel_border) && !inherits(panel_border, "element_blank")) {
            # Use panel.border properties (themes like theme_bw, theme_light)
            border_color <- panel_border$colour %||% "black"
            border_width <- (panel_border$linewidth %||% 0.5) * .pt
            border_type <- panel_border$linetype %||% 1
      } else if (!is.null(panel_bg) && !inherits(panel_bg, "element_blank")) {
            # Use panel.background border properties (themes like theme_minimal)
            border_color <- panel_bg$colour %||% NA
            border_width <- (panel_bg$linewidth %||% 0) * .pt
            border_type <- panel_bg$linetype %||% 1
      } else {
            # Fallback defaults
            border_color <- "grey85"
            border_width <- 0.5 * .pt
            border_type <- 1
      }

      # Panel opacity. An explicit alpha attribute (set via
      # element_rect(alpha = ...)) is honoured on either layer. For foreground
      # panels, alpha additionally inherits from panel.background before
      # falling back to the package default 0.2; the 0.2 here is the canonical
      # location for that default (no longer set in .onLoad(), so leaving the
      # registered panel.foreground without an explicit alpha enables the
      # background-inheritance step above). Background panels are fully opaque
      # if no alpha is set.
      alpha_val <- attr(panel_bg, "ggcube_alpha")
      if (is.null(alpha_val) || is.na(alpha_val)) {
            if (layer == "foreground") {
                  bg_el <- calc_element("panel.background", theme)
                  alpha_val <- attr(bg_el, "ggcube_alpha")
                  if (is.null(alpha_val) || is.na(alpha_val)) {
                        alpha_val <- 0.2
                  }
            } else {
                  alpha_val <- 1
            }
      }

      # Scale to [0, 1] based on plot bounds
      x_scaled <- (face_corners_transformed$x - panel_params$plot_bounds[1]) /
            (panel_params$plot_bounds[2] - panel_params$plot_bounds[1])
      y_scaled <- (face_corners_transformed$y - panel_params$plot_bounds[3]) /
            (panel_params$plot_bounds[4] - panel_params$plot_bounds[3])

      # Get unique faces and their z-depths for sorting
      unique_faces <- unique(face_corners_transformed$face)
      face_depths <- sapply(unique_faces, function(f) {
            face_data <- face_corners_transformed[face_corners_transformed$face == f, ]
            mean(face_data$z, na.rm = TRUE)
      })

      # Sort faces by depth (back to front)
      sorted_faces <- unique_faces[order(-face_depths)]

      # Prepare data for single polygonGrob call with id parameter
      all_x <- numeric(0)
      all_y <- numeric(0)
      polygon_ids <- integer(0)

      polygon_id <- 1
      for (face in sorted_faces) {
            # Get data for this face
            face_indices <- which(face_corners_transformed$face == face)

            # Reorder vertices to form proper rectangles
            face_x <- x_scaled[face_indices][c(1, 3, 4, 2)]
            face_y <- y_scaled[face_indices][c(1, 3, 4, 2)]

            # Append to master lists
            all_x <- c(all_x, face_x)
            all_y <- c(all_y, face_y)
            polygon_ids <- c(polygon_ids, rep(polygon_id, length(face_x)))

            polygon_id <- polygon_id + 1
      }

      grid::polygonGrob(
            x = all_x,
            y = all_y,
            id = polygon_ids,
            default.units = "npc",
            gp = grid::gpar(
                  fill = scales::alpha(fill_color, alpha_val),
                  col = border_color,
                  lwd = border_width,
                  lty = border_type
            ),
            name = paste0("panel.faces.", layer, ".3d")
      )
}


render_cube <- function(self, panel_params, theme, layer = "background"){

      # Transparent base
      bg <- grid::rectGrob(
            x = 0.5, y = 0.5, width = 1, height = 1,
            gp = grid::gpar(fill = "transparent", col = NA),
            name = paste0("panel.", layer, ".transparent")
      )

      # Get theme elements
      panel_bg <- calc_element(switch(layer, background = "panel.background", foreground = "panel.foreground"), theme)
      panel_grid <- calc_element(switch(layer, background = "panel.grid", foreground = "panel.grid.foreground"), theme)
      panel_grid_major <- calc_element(switch(layer, background = "panel.grid.major", foreground = "panel.grid.major.foreground"), theme)
      panel_border <- calc_element("panel.border", theme)

      # Check theme element states from panel_params
      theme_elements <- panel_params$theme_elements %||% list()

      # Determine what to render based on theme
      show_panels <- if (layer == "background") {
            theme_elements$show_background_panels %||% !inherits(panel_bg, "element_blank")
      } else {
            theme_elements$show_foreground_panels %||% !inherits(panel_bg, "element_blank")
      }

      show_border <- !inherits(panel_border, "element_blank")
      grid_element <- panel_grid_major %||% panel_grid
      show_grid <- (theme_elements$show_grid %||% !inherits(grid_element, "element_blank"))

      visible_faces <- switch(layer,
                              background = panel_params$visible_faces_bg,
                              foreground = panel_params$visible_faces_fg)

      # 1. Render panels
      if (show_panels && !is.null(visible_faces)) {
            tryCatch({
                  face_corners <- make_face_panels(visible_faces, panel_params$scales, panel_params$ratio,
                                                   list(x = panel_params$scale_info$x$limits,
                                                        y = panel_params$scale_info$y$limits,
                                                        z = panel_params$scale_info$z$limits))
                  if (!is.null(face_corners) && nrow(face_corners) > 0) {
                        face_corners_transformed <- transform_3d_standard(face_corners, panel_params$proj)
                        face_corners_transformed$face <- face_corners$face

                        # Use background fill only
                        panel_grobs <- create_panel_polygons(face_corners_transformed, panel_params, theme, layer)
                        if (!is.null(panel_grobs)) {
                              bg <- grid::grobTree(bg, panel_grobs, name = paste0("panel.", layer, ".with.3d.panels"))
                        }
                  }
            }, error = function(e) {
                  warning("Panel rendering failed: ", e$message)
            })
      }

      # 2. Render gridlines
      if (show_grid && !is.null(panel_params$grid_transformed)) {
            tryCatch({
                  visible_grid <- panel_params$grid_transformed[panel_params$grid_transformed$face %in% visible_faces,]
                  segments <- create_grid_segments(visible_grid, panel_params$plot_bounds)
                  if (!is.null(segments) && nrow(segments) > 0) {
                        base_lwd <- (grid_element$linewidth %||% 0.5) * .pt
                        grid_grob <- grid::segmentsGrob(
                              x0 = segments$x0, y0 = segments$y0,
                              x1 = segments$x1, y1 = segments$y1,
                              default.units = "npc",
                              gp = grid::gpar(
                                    col = grid_element$colour %||% "grey90",
                                    lwd = safe_lwd(base_lwd * apply_depth_strength(
                                          segments$depth_scale,
                                          depth_strength(self, "grid")), base_lwd),
                                    lty = grid_element$linetype %||% 1
                              ),
                              name = paste0("grid.", layer, ".3d")
                        )
                        bg <- grid::grobTree(bg, grid_grob, name = paste0("panel.", layer, ".with.grid"))
                  }
            }, error = function(e) {
                  warning("Grid rendering failed: ", e$message)
            })
      }

      # 3. Render panel.border
      if (!is.null(visible_faces)) {
            # Get the appropriate border element based on layer
            if (layer == "foreground") {
                  panel_border_element <- calc_element("panel.border.foreground", theme)
            } else {
                  panel_border_element <- calc_element("panel.border", theme)
            }

            show_border <- !inherits(panel_border_element, "element_blank")

            if (show_border) {
                  tryCatch({
                        face_corners <- make_face_panels(visible_faces, panel_params$scales, panel_params$ratio,
                                                         list(x = panel_params$scale_info$x$limits,
                                                              y = panel_params$scale_info$y$limits,
                                                              z = panel_params$scale_info$z$limits))
                        if (!is.null(face_corners) && nrow(face_corners) > 0) {
                              face_corners_transformed <- transform_3d_standard(face_corners, panel_params$proj)
                              face_corners_transformed$face <- face_corners$face

                              # Render only polygon *outlines*
                              x_scaled <- (face_corners_transformed$x - panel_params$plot_bounds[1]) /
                                    (panel_params$plot_bounds[2] - panel_params$plot_bounds[1])
                              y_scaled <- (face_corners_transformed$y - panel_params$plot_bounds[3]) /
                                    (panel_params$plot_bounds[4] - panel_params$plot_bounds[3])

                              # Prepare outlines
                              unique_faces <- unique(face_corners_transformed$face)
                              face_depths <- sapply(unique_faces, function(f) {
                                    mean(face_corners_transformed$z[face_corners_transformed$face == f], na.rm = TRUE)
                              })
                              sorted_faces <- unique_faces[order(-face_depths)]

                              seg_x0 <- numeric()
                              seg_y0 <- numeric()
                              seg_x1 <- numeric()
                              seg_y1 <- numeric()
                              seg_depth_scale <- numeric()

                              for (face in sorted_faces) {
                                    idx <- which(face_corners_transformed$face == face)
                                    vx <- x_scaled[idx][c(1, 3, 4, 2)]
                                    vy <- y_scaled[idx][c(1, 3, 4, 2)]
                                    vds <- face_corners_transformed$depth_scale[idx][c(1, 3, 4, 2)]
                                    n <- length(vx)
                                    for (j in seq_len(n)) {
                                          k <- if (j == n) 1 else j + 1
                                          seg_x0 <- c(seg_x0, vx[j])
                                          seg_y0 <- c(seg_y0, vy[j])
                                          seg_x1 <- c(seg_x1, vx[k])
                                          seg_y1 <- c(seg_y1, vy[k])
                                          seg_depth_scale <- c(seg_depth_scale, mean(vds[c(j, k)]))
                                    }
                              }

                              base_lwd <- (panel_border_element$linewidth %||% 0.5) * .pt
                              border_grob <- grid::segmentsGrob(
                                    x0 = seg_x0, y0 = seg_y0,
                                    x1 = seg_x1, y1 = seg_y1,
                                    default.units = "npc",
                                    gp = grid::gpar(
                                          col = panel_border_element$colour %||% "black",
                                          lwd = safe_lwd(base_lwd * apply_depth_strength(
                                                seg_depth_scale,
                                                depth_strength(self, "border")), base_lwd),
                                          lty = panel_border_element$linetype %||% 1
                                    ),
                                    name = paste0("panel.border.", layer, ".3d")
                              )
                              bg <- grid::grobTree(bg, border_grob, name = paste0("panel.", layer, ".with.border"))
                        }
                  }, error = function(e) {
                        warning("Border rendering failed: ", e$message)
                  })
            }
      }


      # 4. Render axis text and titles
      if(layer == "background"){
            # Check if we should render axis text/titles based on theme
            should_render_axis_text <- theme_elements$show_axis_text %||% !inherits(calc_element("axis.text", theme), "element_blank")
            should_render_axis_title <- theme_elements$show_axis_title %||% !inherits(calc_element("axis.title", theme), "element_blank")
            should_render_axis_ticks <- theme_elements$show_axis_ticks %||% !inherits(calc_element("axis.ticks", theme), "element_blank")

            if (should_render_axis_text || should_render_axis_title || should_render_axis_ticks) {
                  furniture <- axis_furniture_grob(
                        self, panel_params, theme,
                        show_text = should_render_axis_text,
                        show_title = should_render_axis_title,
                        show_ticks = should_render_axis_ticks,
                        name = paste0("axis.furniture.", layer, ".3d")
                  )
                  bg <- grid::grobTree(bg, furniture,
                                       name = paste0("panel.", layer, ".with.axis.text"))
            }
      }

      return(bg)
}

#' Position for projecting 2D and 3D layers onto cube faces
#'
#' Enables layers to be projected onto 2D faces of the 3D coordinate cube.
#' It can be used to flatten 3D ggcube layers onto a single cube face as a way of
#' visualizing them in 2D, or to add certain natively 2D ggplot2 layers like
#' `geom_density_2d()` or `geom_smooth()` to a cube face.
#'
#' @param faces Character string or vector specifying which cube face(s) to project onto.
#'   Valid options are: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax", "3D".
#'   "3D" indicates the raw, non-flattened 3D position.
#'   Multiple faces and "3D" are only supported for 3D layers (when `axes = NULL`).
#' @param axes For 2D layers only: Character vector of length 2 specifying which
#'   3D dimensions the 2D layer's x and y aesthetics represent. For example,
#'   `c("x", "z")` means the 2D x-axis maps to the 3D x-axis and the 2D y-axis
#'   maps to the 3D z-axis. For 3D layers, use `NULL` (default) to preserve
#'   the existing x,y,z mapping.
#'
#' @details
#' This position adjustment supports both 2D and 3D layers:
#'
#' **For 3D layers** (`axes = NULL`):
#' - Data already has x,y,z coordinates in the correct order
#' - Simply adds projection metadata for coord_3d to place the layer on the specified face(s)
#' - The face coordinate will be overridden during coordinate transformation
#' - Multiple faces are supported - the layer will be duplicated on each specified face.
#'          All specified faces inherit aesthetics from the layer function; if you want different
#'          parameters for different faces and on the primary 3D layer, add each as a separate layer call.
#'
#' **For 2D layers** (`axes = c("dim1", "dim2")`):
#' - Renames the layer's x,y columns to match the specified 3D axes
#' - Adds the missing third dimension
#' - Adds projection metadata for coord_3d
#' - Only single faces are supported for 2D layers
#'
#' The actual projection happens during coordinate transformation in coord_3d.
#'
#' **Compatibility Note:** This position adjustment is not compatible with all 2D stats.
#' It works well with `stat_density_2d()` and other
#' stats that don't depend heavily on scale ranges during computation, but may cause
#' errors or rendering issues with `stat_density_2d_filled()` and similar stats that
#' generate polygons based on scale domains, as well as with layers like `stat_bin_2d()`
#' that return position variables other than x and y. For `geom_smooth()`, this position
#' works only if you set `se = FALSE` as shown in the example.
#'
#' @examples
#'
#' # 3D point layer in raw 3D form, and projected onto 2D face
#' ggplot(sphere_points, aes(x, y, z)) +
#'   geom_point_3d(position = position_on_face("zmin"), color = "red") +
#'   geom_point_3d(color = "black") + # add this layer last so it appears on top
#'   coord_3d()
#'
#' # 3D layer projected to multiple faces
#' set.seed(1)
#' d <- data.frame(x = round(rnorm(10)), y = round(rnorm(10)), z = round(rnorm(10)))
#' ggplot(d, aes(x, y, z)) +
#'   stat_voxel_3d(color = "black", fill = "steelblue",
#'     light = light(direction = c(1, 1, 0), mode = "hsv"),
#'     position = position_on_face(c("3D", "zmin", "xmax", "ymax"))) +
#'   coord_3d()
#'
#' # 3D layer projected differently on individual faces
#' ggplot(sphere_points, aes(x, y, z)) +
#'   stat_hull_3d(position = position_on_face("zmin"), fill = "black") +
#'   geom_point_3d(position = position_on_face("ymax")) +
#'   geom_path_3d(position = position_on_face("xmax")) +
#'   stat_hull_3d(color = "black") +
#'   coord_3d()
#'
#' # 2D density contour on a specific face
#' ggplot(iris, aes(Sepal.Length, Sepal.Width, Petal.Length, color = Species)) +
#'   stat_density_2d(position = position_on_face(faces = "zmin", axes = c("x", "y"))) +
#'   geom_point_3d() +
#'   coord_3d()
#'
#' # Distinct 2D layers projected to different faces
#' ggplot(mtcars) +
#'   geom_smooth(aes(mpg, qsec), color = "red", alpha = .5, se = FALSE,
#'     position = position_on_face(faces = "ymax", axes = c("x", "z"))) +
#'   geom_density_2d(aes(mpg, wt), alpha = .5,
#'     position = position_on_face(faces = "zmin", axes = c("x", "y"))) +
#'   geom_path(aes(wt, qsec), color = "forestgreen", alpha = .5,
#'     position = position_on_face(faces = "xmax", axes = c("y", "z"))) +
#'   geom_point_3d(aes(mpg, wt, qsec)) +
#'     coord_3d() +
#'   theme_light()
#'
#' @seealso [coord_3d()] for 3D coordinate systems, [ggplot2::stat_density_2d()] and other
#'   2D statistical transformations that can be projected onto faces.
#' @return A ggplot position object that can be supplied to the `position` argument
#'   to layer functions.
#' @export
position_on_face <- function(faces = "zmin", axes = NULL) {
      # Validate faces parameter
      valid_faces <- c("xmin", "xmax", "ymin", "ymax", "zmin", "zmax", "3D")
      if (!all(faces %in% valid_faces)) {
            invalid_faces <- setdiff(faces, valid_faces)
            stop("Invalid face names: ", paste(invalid_faces, collapse = ", "),
                 ". Valid faces are: ", paste(valid_faces, collapse = ", "))
      }

      # Validate axes parameter
      if (!is.null(axes)) {
            if (length(axes) != 2) {
                  stop("axes must be NULL (for 3D layers) or a character vector of length 2 (for 2D layers)")
            }

            valid_axes <- c("x", "y", "z")
            if (!all(axes %in% valid_axes)) {
                  stop("axes must contain only 'x', 'y', or 'z'")
            }

            if (length(unique(axes)) != 2) {
                  stop("axes must specify two different dimensions")
            }

            # Multiple faces not allowed for 2D case
            if (length(faces) > 1) {
                  stop("Multiple faces are only supported for 3D layers (axes = NULL). ",
                       "For 2D layers, specify a single face.")
            }

            # Check that face and axes are compatible (only for 2D case)
            face_dim <- substr(faces[1], 1, 1)  # Use first (and only) face
            missing_dim <- setdiff(c("x", "y", "z"), axes)

            if (length(missing_dim) != 1) {
                  stop("axes must specify exactly two different dimensions from 'x', 'y', 'z'")
            }

            if (face_dim != missing_dim) {
                  stop("Face '", faces[1], "' is incompatible with axes = c('", axes[1], "', '", axes[2], "'). ",
                       "The face dimension ('", face_dim, "') must match the missing dimension ('", missing_dim, "').")
            }
      }

      ggproto(NULL, PositionOnFace,
              faces = faces,
              axes = axes
      )
}

#' Validate stat compatibility with position_on_face
#'
#' This function walks parent frames to find the layer's stat and checks if it's
#' compatible with position_on_face. Some stats have scale training conflicts that
#' cause incorrect rendering when used with position_on_face.
#'
#' @return NULL if compatible, otherwise issues a warning
#' @keywords internal
#' @noRd
validate_position_stat <- function() {
      # Known incompatible stats
      incompatible_stats <- c(
            "StatDensity2dFilled",
            "StatBin2d",
            "StatBinhex"
      )

      # Walk parent frames to find the layer
      layer_obj <- NULL
      tryCatch({
            for (i in 1:25) {
                  env <- parent.frame(i)

                  # Look for layer being constructed
                  if (exists("layer", envir = env)) {
                        potential_layer <- get("layer", envir = env)
                        if (inherits(potential_layer, "function") &&
                            "stat" %in% names(formals(potential_layer))) {
                              # This is the layer() function call
                              # Try to get the stat argument
                              call_obj <- sys.call(i)
                              if (!is.null(call_obj) && "stat" %in% names(call_obj)) {
                                    stat_arg <- eval(call_obj$stat, envir = env)
                                    if (inherits(stat_arg, "Stat")) {
                                          layer_obj <- list(stat = stat_arg)
                                          break
                                    }
                              }
                        }
                  }

                  # Also look for already-constructed layer objects
                  for (var_name in ls(envir = env)) {
                        obj <- get(var_name, envir = env)
                        if (inherits(obj, "Layer")) {
                              layer_obj <- obj
                              break
                        }
                  }
                  if (!is.null(layer_obj)) break
            }
      }, error = function(e) {
            # If frame walking fails, skip validation
            return(NULL)
      })

      # Extract stat class name
      if (!is.null(layer_obj) && !is.null(layer_obj$stat)) {
            stat_class <- class(layer_obj$stat)[1]

            if (stat_class %in% incompatible_stats) {
                  # Generate helpful warning message
                  stat_name <- switch(stat_class,
                                      "StatDensity2dFilled" = "stat_density_2d_filled()",
                                      "StatBin2d" = "stat_bin_2d()",
                                      paste0(stat_class, "()"))

                  alternative <- switch(stat_class,
                                        "StatDensity2dFilled" = "stat_density_2d()",
                                        "StatBin2d" = "stat_bin()",
                                        "a compatible stat")

                  warning("position_on_face() is incompatible with ", stat_name, ". ",
                          "Stats that have position aesthetics beyond x and y, or that employ certain scale training routines, are not currently supported. ",
                          "Consider using ", alternative, " instead.",
                          call. = FALSE)
            }
      }

      return(NULL)
}

PositionOnFace <- ggproto("PositionOnFace", Position,
                          faces = NULL,
                          axes = NULL,

                          compute_layer = function(self, data, params, layout) {
                                # Only validate stat compatibility for 2D case (axes is not NULL)
                                if (!is.null(self$axes)) {
                                      validate_position_stat()
                                }

                                # Collect results for all faces
                                all_face_data <- list()

                                for (face in self$faces) {
                                      if (is.null(self$axes)) {
                                            # 3D case: data should already have x, y, z columns
                                            # Just add projection metadata
                                            if (!all(c("x", "y", "z") %in% names(data))) {
                                                  warning("position_on_face with axes = NULL requires x, y, and z aesthetics for 3D layers")
                                                  return(data)
                                            }

                                            face_data <- data
                                            # Add projection metadata
                                            if(face == "3D") face <- NA # transform leaves NA values alone
                                            face_data$project_to_face <- face

                                      } else {
                                            # 2D case
                                            # If the layer doesn't have both x and y aesthetics, return unchanged
                                            if (!all(c("x", "y") %in% names(data))) {
                                                  warning("position_on_face requires both x and y aesthetics")
                                                  return(data)
                                            }

                                            if(face == "3D") stop("the selected stat or `axes` specification is incompatible with `faces = '3D'` .")

                                            face_data <- data
                                            # Map the 2D x,y columns to the correct 3D dimensions
                                            # Rename columns based on axes specification
                                            names(face_data)[match(c("x", "y"), names(face_data))] <- self$axes

                                            # Add the missing dimension with placeholder value
                                            missing_dim <- setdiff(c("x", "y", "z"), self$axes)
                                            if (length(missing_dim) == 1) {
                                                  # Check if the missing dimension already exists (e.g., from stat_contour)
                                                  if (missing_dim %in% names(face_data)) {
                                                        warning("Data already contains a '", missing_dim, "' column (possibly computed by the stat). ",
                                                                "This will be overwritten for face projection. ",
                                                                "If you need the original values, consider using after_stat() before applying this position.")
                                                  }
                                                  # Use -Inf as placeholder for face-projected values
                                                  # (because NA errors, and finite numbers pollute scale training)
                                                  face_data[[missing_dim]] <- -Inf
                                            }

                                            # Add projection metadata
                                            face_data$project_to_face <- face
                                      }

                                      # Add hierarchical grouping to prevent depth sorting within groups
                                      # This ensures contour lines and other connected geometries maintain vertex order
                                      if ("group" %in% names(face_data)) {
                                            face_data$group <- paste0("face_", face, "__", face_data$group)
                                      }

                                      all_face_data[[length(all_face_data) + 1]] <- face_data
                                }

                                # Combine all face data
                                if (length(all_face_data) == 1) {
                                      return(all_face_data[[1]])
                                } else {
                                      return(do.call(rbind, all_face_data))
                                }
                          }
)

#' Expand tetrahedra into their constituent faces
#'
#' Each tetrahedron contributes four triangular faces, one per omitted vertex.
#' The omitted vertex is retained as an interior reference point: it lies on the
#' interior side of that face by construction.
#'
#' @param tetra Integer matrix with 4 columns of vertex indices
#' @return List with `faces` (n*4 x 3 matrix of sorted vertex indices) and
#'   `opposite` (length n*4 vector of the omitted vertex index)
#' @keywords internal
#' @noRd
expand_tetra_faces <- function(tetra) {
      m <- nrow(tetra)
      faces <- matrix(nrow = m * 4L, ncol = 3L)
      opposite <- integer(m * 4L)

      for (k in 1:4) {
            rows <- ((k - 1L) * m + 1L):(k * m)
            faces[rows, ] <- tetra[, setdiff(1:4, k), drop = FALSE]
            opposite[rows] <- tetra[, k]
      }

      # Sort each face's vertex indices so faces can be matched regardless of winding
      faces <- sort_triples(faces)

      list(faces = faces, opposite = opposite)
}

#' Row-wise sort of a three-column integer matrix
#'
#' Equivalent to `t(apply(m, 1, sort))` but vectorized: for exactly three
#' values the sorted triple can be expressed with pairwise min and max,
#' avoiding a per-row R-level call.
#'
#' @param m Integer or numeric matrix with 3 columns
#' @return Matrix of the same dimensions with each row sorted ascending
#' @keywords internal
#' @noRd
sort_triples <- function(m) {
      a <- m[, 1]; b <- m[, 2]; cc <- m[, 3]
      lo <- pmin.int(a, b)
      hi <- pmax.int(a, b)
      cbind(pmin.int(lo, cc),
            pmin.int(hi, pmax.int(lo, cc)),
            pmax.int(hi, cc))
}

#' Encode sorted vertex triples as unique scalar keys
#'
#' Matching on numeric keys is substantially faster than on pasted strings.
#' Indices are packed in base `n`, which is exact in double precision while
#' `n^3` stays below 2^53; beyond that a string key is used instead.
#'
#' @param triples Integer matrix with 3 columns of sorted vertex indices
#' @param n Number of vertices (the base for packing)
#' @return Numeric or character vector of keys
#' @keywords internal
#' @noRd
triple_keys <- function(triples, n) {
      if (n <= 200000) {
            nn <- as.numeric(n)
            (triples[, 1] - 1) * nn * nn + (triples[, 2] - 1) * nn + triples[, 3]
      } else {
            paste(triples[, 1], triples[, 2], triples[, 3], sep = "_")
      }
}

#' Find an interior reference point for each boundary triangle
#'
#' For a triangle on the boundary of an alpha complex, exactly one adjacent
#' tetrahedron lies inside the complex. That tetrahedron's fourth vertex is a
#' point known to be on the interior side of the face, which determines the
#' outward normal direction. This is a local test, so unlike a global centroid
#' it remains correct for concave shapes and shapes with holes.
#'
#' @param tri Integer matrix with 3 columns of vertex indices (boundary triangles)
#' @param tetra Integer matrix with 4 columns of vertex indices (interior tetrahedra)
#' @return Integer vector of vertex indices, NA where no interior tetrahedron was found
#' @keywords internal
#' @noRd
find_interior_reference <- function(tri, tetra) {
      if (is.null(tetra) || nrow(tetra) == 0) {
            return(rep(NA_integer_, nrow(tri)))
      }

      expanded <- expand_tetra_faces(tetra)

      n <- max(max(tetra), max(tri))
      face_keys <- triple_keys(expanded$faces, n)
      tri_keys <- triple_keys(sort_triples(tri), n)

      # Match faces against triangles rather than the reverse: the expanded face
      # table is far larger than the boundary triangle set, and match() builds its
      # hash table from the second argument, so this hashes the smaller vector.
      hit <- match(face_keys, tri_keys)
      matched <- !is.na(hit)

      reference <- rep(NA_integer_, nrow(tri))
      if (any(matched)) {
            slot <- hit[matched]
            vert <- expanded$opposite[matched]
            # Keep the first matching tetrahedron per triangle
            keep <- !duplicated(slot)
            reference[slot[keep]] <- vert[keep]
      }

      reference
}

#' Compute outward-facing normals and consistent winding for a triangulation
#'
#' Computes face normals by cross product, then flips each normal (and the
#' corresponding triangle winding) so that it points away from the interior.
#' The interior side is determined per-face from `reference_points` when
#' available, falling back to the global centroid otherwise.
#'
#' @param coords Numeric matrix of vertex coordinates (n x 3)
#' @param tri Integer matrix with 3 columns of vertex indices
#' @param reference_points Integer vector of per-face interior vertex indices,
#'   or NULL to use the centroid for every face
#' @return List with `tri` (winding-corrected) and `normals` (unit outward normals)
#' @keywords internal
#' @noRd
orient_triangles <- function(coords, tri, reference_points = NULL) {
      v1 <- coords[tri[, 1], , drop = FALSE]
      v2 <- coords[tri[, 2], , drop = FALSE]
      v3 <- coords[tri[, 3], , drop = FALSE]

      edge1 <- v2 - v1
      edge2 <- v3 - v1

      normals <- cbind(
            edge1[, 2] * edge2[, 3] - edge1[, 3] * edge2[, 2],
            edge1[, 3] * edge2[, 1] - edge1[, 1] * edge2[, 3],
            edge1[, 1] * edge2[, 2] - edge1[, 2] * edge2[, 1]
      )

      face_centers <- (v1 + v2 + v3) / 3

      # Vector from the interior reference point toward the face. A normal
      # pointing outward has a positive dot product with this vector.
      if (is.null(reference_points)) {
            outward <- sweep(face_centers, 2, colMeans(coords), "-")
      } else {
            interior <- coords[reference_points, , drop = FALSE]
            outward <- face_centers - interior

            # Faces with no interior tetrahedron fall back to the centroid
            missing <- is.na(reference_points)
            if (any(missing)) {
                  centroid_dir <- sweep(face_centers[missing, , drop = FALSE], 2,
                                        colMeans(coords), "-")
                  outward[missing, ] <- centroid_dir
            }
      }

      # Flip inward-pointing normals and their triangle winding
      inward <- rowSums(normals * outward) < 0
      if (any(inward)) {
            normals[inward, ] <- -normals[inward, ]
            tri[inward, c(2, 3)] <- tri[inward, c(3, 2)]
      }

      lengths <- sqrt(rowSums(normals^2))
      degenerate <- lengths == 0 | !is.finite(lengths)
      normals <- normals / lengths
      if (any(degenerate)) {
            normals[degenerate, ] <- matrix(c(0, 0, 1), nrow = sum(degenerate),
                                            ncol = 3, byrow = TRUE)
      }

      list(tri = tri, normals = normals)
}

StatHull3D <- ggproto("StatHull3D", Stat,
                      required_aes = c("x", "y", "z"),
                      compute_group = function(data, scales,
                                               method = "convex", radius = NULL,
                                               singular = FALSE,
                                               cull_backfaces = NULL,
                                               light = NULL) {
                            coords <- as.matrix(data[, c("x", "y", "z")])

                            reference_points <- NULL
                            singular_faces <- NULL

                            # Get triangle indices
                            tri <- switch(method,
                                          alpha = {
                                                if (!requireNamespace("alphashape3d", quietly = TRUE)) {
                                                      stop('The alphashape3d package is required for method = "alpha". ',
                                                           'Install it with install.packages("alphashape3d").',
                                                           call. = FALSE)
                                                }
                                                if(is.null(radius)){
                                                      radius <- signif(.5 * mean(apply(coords, 2, function(x) diff(range(na.omit(x))))), 3)
                                                      message("Alpha shape `radius` parameter is NULL; defaulting to ", radius, " based on data ranges.")
                                                }
                                                ashape <- alphashape3d::ashape3d(coords, alpha = radius)

                                                # Triangle membership codes live in the column following the
                                                # standard 8, one per alpha value. Codes: 0 = exterior,
                                                # 1 = interior, 2 = regular (boundary), 3 = singular.
                                                tri_all <- ashape$triang
                                                tri_code <- tri_all[, 9]

                                                keep_codes <- if (isTRUE(singular)) c(2, 3) else 2
                                                selected <- tri_all[tri_code %in% keep_codes, 1:3, drop = FALSE]

                                                if (nrow(selected) == 0) {
                                                      stop("No valid triangle set found - try increasing `radius`")
                                                }

                                                # Singular faces bound no volume, so no outward direction exists
                                                singular_faces <- tri_code[tri_code %in% keep_codes] == 3

                                                # Tetrahedron membership codes follow the 4 vertex columns
                                                # and rhoT, one per alpha value: 1 = in the alpha complex.
                                                tetra_all <- ashape$tetra
                                                interior_tetra <- tetra_all[tetra_all[, 6] == 1, 1:4, drop = FALSE]

                                                reference_points <- find_interior_reference(selected, interior_tetra)

                                                unresolved <- sum(is.na(reference_points) & !singular_faces)
                                                if (unresolved > 0) {
                                                      warning(unresolved, " of ", nrow(selected),
                                                              " alpha shape faces had no adjacent interior tetrahedron; ",
                                                              "falling back to centroid orientation for these faces.",
                                                              call. = FALSE)
                                                }

                                                selected
                                          },
                                          convex = {
                                                if (!requireNamespace("geometry", quietly = TRUE)) {
                                                      stop('The geometry package is required for method = "convex". ',
                                                           'Install it with install.packages("geometry").',
                                                           call. = FALSE)
                                                }
                                                geometry::convhulln(coords)
                                          },
                                          stop("Unknown method: use 'alpha' or 'convex'")
                            )

                            # Check that we got valid triangulation
                            if (is.null(tri) || nrow(tri) == 0) {
                                  stop("No triangles generated - try adjusting alpha parameter or check data")
                            }

                            # Standardize triangle winding order and compute outward normals.
                            # A convex hull is star-shaped about its centroid, so the centroid
                            # test is exact there; alpha shapes need the per-face interior
                            # reference derived from the alpha complex.
                            oriented <- orient_triangles(coords, tri, reference_points)
                            tri <- oriented$tri
                            normals <- oriented$normals

                            # Assemble triangle vertices with corrected winding order
                            data <- data[as.vector(t(tri)), ]

                            # Create unique face IDs that won't collide across groups,
                            # using a random suffix to ensure uniqueness across compute_group calls
                            group_suffix <- sample(10000:99999, 1)
                            data$group <- rep(paste0("sh3d__hull", group_suffix, "_tri", 1:nrow(tri)), each = 3)

                            # Add computed normal variables (replicated for each vertex in triangle)
                            data$normal_x <- rep(normals[, 1], each = 3)
                            data$normal_y <- rep(normals[, 2], each = 3)
                            data$normal_z <- rep(normals[, 3], each = 3)

                            # Singular faces have no consistent outward orientation, so culling
                            # them by winding would discard them arbitrarily
                            if (!is.null(singular_faces) && any(singular_faces)) {
                                  data$cull_backfaces <- rep(!singular_faces & isTRUE(cull_backfaces), each = 3)
                            } else {
                                  data$cull_backfaces <- cull_backfaces
                            }

                            # add computed variables and lighting info
                            data <- data %>%
                                  average_aesthetics() %>%
                                  attach_light(light)
                            return(data)
                      }
)


#' 3D convex and alpha hulls
#'
#' Turns 3D point clouds into surface hulls consisting of triangular polygons,
#' using either convex hull or alpha shape algorithms.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. The required
#'   aesthetics are `x`, `y`, and `z`. Additional aesthetics can use computed
#'   variables with [ggplot2::after_stat()].
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use on the data. Defaults to `StatHull3D`.
#' @param geom The geometric object used to display the data. Defaults to `GeomPolygon3D.`
#'
#' @param method Triangulation method. Either:
#'   - `"convex"`: Convex hull triangulation (default). Requires the \pkg{geometry} package.
#'   - `"alpha"`: Alpha shape triangulation (can capture non-convex topologies)
#' @param radius The "alpha" parameter when alpha method is used.
#'   A face is included in the resulting alpha shape if it can be "exposed" by a sphere of this radius.
#'   If NULL (the default), a simple heuristic based on the data scale is used to calculate a radius value.
#'   Note that alpha shapes are quite sensitive to the coordinate scales of your data. See Details section.
#' @param singular Whether to include "singular" faces when `method = "alpha"`.
#'   These are faces that do not bound any enclosed volume, appearing as isolated
#'   sheets dangling from the surface, and are usually a sign that `radius` is
#'   slightly too small. Because they enclose nothing, they have no well-defined
#'   outward direction: their lighting is arbitrary and they are never culled.
#'   Defaults to `FALSE`.
#'
#' @inheritParams light_param
#' @inheritParams polygon_params
#' @inheritParams position_param
#'
#' @section Grouping:
#' Hulls respect ggplot2 grouping aesthetics. To create separate hulls for different
#' subsets of your data, use `aes(group = category_variable)` or similar grouping aesthetics.
#' Each group will get its own independent hull.
#'
#' @section Alpha scale sensitivity:
#' Alpha shape method is highly sensitive to coordinate scales. `radius` is expressed
#' in data units, so a value that works for data scaled 0-1 will likely fail for data
#' scaled 0-1000.
#' Guidelines for choosing radius:
#' - Start at a small fraction of your data's overall extent and adjust based on results
#' - For data with mixed scales (e.g., x: 0-1, y: 0-1000), consider rescaling your data first
#' - Larger radius values → smoother, more connected surfaces
#' - Smaller radius values → more detailed surfaces, but may fragment
#' - If you get no triangles, try increasing radius by 10x
#' - If surface fills unwanted holes, try decreasing radius by 10x
#'
#' @section Aesthetics:
#' `stat_hull_3d()` requires the following aesthetics:
#' - **x**: X coordinate
#' - **y**: Y coordinate
#' - **z**: Z coordinate
#'
#' @section Computed variables:
#' - `normal_x`, `normal_y`, `normal_z`: Surface normal components
#'
#' @examplesIf requireNamespace("geometry", quietly = TRUE)
#' # A solid torus: a shape whose hole a convex hull cannot represent
#' set.seed(1)
#' n <- 2000
#' theta <- runif(n, 0, 2 * pi)  # position around the ring
#' phi <- runif(n, 0, 2 * pi)    # position around the tube
#' rho <- sqrt(runif(n))         # distance from the tube center
#' torus <- data.frame(
#'   x = (3 + rho * cos(phi)) * cos(theta),
#'   y = (3 + rho * cos(phi)) * sin(theta),
#'   z = rho * sin(phi)
#' )
#'
#' # Convex hull: the hole is bridged over
#' ggplot(torus, aes(x, y, z)) +
#'   geom_hull_3d(method = "convex", fill = "gray40") +
#'   coord_3d(scales = "fixed")
#'
#' # Alpha shape: the hole is preserved
#' \donttest{
#' ggplot(torus, aes(x, y, z)) +
#'   geom_hull_3d(method = "alpha", radius = 0.6, fill = "gray40") +
#'   coord_3d(scales = "fixed")
#' }
#'
#' # Use `cull_backfaces = FALSE` to render far side of hull
#' \donttest{
#' ggplot(torus, aes(x, y, z)) +
#'   geom_hull_3d( # default culling for comparison
#'     method = "alpha", radius = 0.6, light = NULL,
#'     fill = "steelblue", color = "darkred", linewidth = .5, alpha = .5) +
#'   geom_hull_3d( # culling disabled
#'     aes(x = x + 9), cull_backfaces = FALSE,
#'     method = "alpha", radius = 0.6, light = NULL,
#'     fill = "steelblue", color = "darkred", linewidth = .5, alpha = .5) +
#'   coord_3d(scales = "fixed")
#' }
#'
#' # Use grouping to build separate hulls for data subsets
#' ggplot(iris, aes(Petal.Length, Sepal.Length, Sepal.Width,
#'                  color = Species, fill = Species)) +
#'       geom_hull_3d() +
#'       coord_3d(scales = "fixed")
#'
#' @seealso [coord_3d()] for 3D coordinate systems, [geom_polygon_3d] for the
#'   default geometry with depth sorting, [light()] for lighting specifications.
#' @rdname geom_hull_3d
#' @return A `Layer` object that can be added to a ggplot.
#' @export
geom_hull_3d <- function(mapping = NULL, data = NULL,
                         stat = StatHull3D,
                         position = "identity",
                         ...,
                         method = "convex", radius = NULL, singular = FALSE, light = NULL,
                         cull_backfaces = TRUE, sort_method = NULL, scale_depth = TRUE,
                         inherit.aes = TRUE, show.legend = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomPolygon3D,
            position = position, inherit.aes = inherit.aes, show.legend = show.legend,
            params = list(method = method, radius = radius, singular = singular, light = light,
                          cull_backfaces = cull_backfaces, sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}

#' @rdname geom_hull_3d
#' @export
stat_hull_3d <- function(mapping = NULL, data = NULL,
                         geom = GeomPolygon3D,
                         position = "identity",
                         ...,
                         method = "convex", radius = NULL, singular = FALSE, light = NULL,
                         cull_backfaces = TRUE, sort_method = NULL, scale_depth = TRUE,
                         inherit.aes = TRUE, show.legend = TRUE) {

      layer(data = data, mapping = mapping, stat = StatHull3D, geom = get_proto(geom, "geom"),
            position = position, inherit.aes = inherit.aes, show.legend = show.legend,
            params = list(method = method, radius = radius, singular = singular, light = light,
                          cull_backfaces = cull_backfaces, sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}

facets <- function(proj, effective_ratios = c(1, 1, 1)){

      # Apply aspect ratios to cube corners
      corners <- expand.grid(x = c(-.5, .5) * effective_ratios[1],
                             y = c(-.5, .5) * effective_ratios[2],
                             z = c(-.5, .5) * effective_ratios[3])

      corners_p <- transform_3d_standard(corners, proj)

      facet_grid <- expand.grid(x = c("xmin", "xmax"),
                                y = c("ymin", "ymax"),
                                z = c("zmin", "zmax"),
                                stringsAsFactors = FALSE)  # Prevent factors

      f <- bind_cols(setNames(facet_grid, c("fx", "fy", "fz")), corners_p) %>%
            gather(facet, value, fx:fz) %>%
            mutate(facet = str_remove(facet, "f"),
                   value = str_remove(value, facet),
                   x = scales::rescale(x),
                   y = scales::rescale(y),
                   z = scales::rescale(z)) %>%
            unite("facet", facet, value, sep = "") %>%
            arrange(facet)
      f
}

vertices <- function(proj, effective_ratios = c(1, 1, 1)){
      f <- facets(proj, effective_ratios)

      v <- f %>%
            mutate(on_hull = 1:nrow(.) %in% chull(x, y)) %>%
            group_by(x,y,z) %>%
            mutate(on_hull = any(on_hull)) %>%
            group_by(facet) %>%
            mutate(far_corner = any(z == 1),
                   near_corner = any(z == 0)) %>%
            ungroup()
      v <- v %>%
            select(-facet) %>%
            filter(on_hull) %>%
            distinct() %>%
            mutate(vert = 1:nrow(.)) %>%
            full_join(v, by = join_by(x, y, z, on_hull,
                                      near_corner, far_corner))
      v
}

face_info <- function(proj, effective_ratios = c(1, 1, 1)){

      v <- vertices(proj, effective_ratios)
      h <- v %>%
            group_by(facet) %>%
            summarize(xmin = min(x), xmean = mean(x), xmax = max(x),
                      ymin = min(y), ymean = mean(y), ymax = max(y),
                      zmin = min(z), zmean = mean(z), zmax = max(z),

                      xfar = mean(x[z == min(z)]), # x' coordinate of facet corner farthest from viewer
                      xnear = mean(x[z == max(z)]),
                      yfar = mean(y[z == min(z)]),
                      ynear = mean(y[z == max(z)]),

                      nvchull = length(unique(na.omit(vert))),
                      near = any(near_corner),
                      far = any(far_corner),
                      .groups = "drop")

      # Foreground/background logic:
      # -FACE VIEW (hull has 4 corners): nearest face is fg, all others bg
      # -EDGE VIEW (hull has 6 corners, two faces have 4 hull corners): faces with 4 hull corners are fg, all others bg
      # -CORNER VIEW (hull has 6 corners, all faces have 3 hull corners):
      #       faces touching near corner are fg, faces touching far corner are bg
      view <- case_when((sum(v$on_hull) / 3) == 4 ~ "face",
                        length(unique(h$nvchull)) == 1 ~ "corner",
                        TRUE ~ "edge")

      h <- h %>%
            mutate(foreground = case_when(
                  view == "face" & zmean == min(zmean) ~ TRUE,
                  view == "edge" & nvchull == 4 ~ TRUE,
                  view == "edge" & nvchull != 4 ~ FALSE,
                  view == "corner" & zmin == min(zmin) ~ TRUE,
                  TRUE ~ FALSE))

      h %>% arrange(zmean, foreground) # rendering order
}

select_visible_faces <- function(faces_param, proj, effective_ratios = c(1, 1, 1)){

      rules <- c("all", "none", "near", "far", "background", "foreground",
                 "left", "right", "top", "bottom", "front", "back")
      facets <- c("xmin", "xmax", "ymin", "ymax", "zmin", "zmax")
      faces <- match.arg(faces_param, c(rules, facets), several.ok = TRUE)
      rules <- rules[rules %in% faces]
      facets <- facets[facets %in% faces]

      f <- face_info(proj, effective_ratios)
      rules <- sapply(rules, function(x){
            switch(x,
                   all = f$facet,
                   none = NULL,
                   near = f$facet[f$near],
                   far = f$facet[f$far],
                   background = f$facet[!f$foreground],
                   foreground = f$facet[f$foreground],
                   left = f$facet[f$xmean == min(f$xmean)][1],
                   right = f$facet[f$xmean == max(f$xmean)][1],
                   top = f$facet[f$ymean == max(f$ymean)][1],
                   bottom = f$facet[f$ymean == min(f$ymean)][1],
                   front = f$facet[f$zmean == max(f$zmean)][1],
                   back = f$facet[f$zmean == min(f$zmean)][1]
            )
      })
      faces <- unique(c(rules, facets))

      list(bg =  f$facet[f$facet %in% faces & !f$foreground],
           fg =  f$facet[f$facet %in% faces & f$foreground])
}

get_face_corners <- function(face_name) {
      # Return the 4 corners of a cube face
      if (face_name == "xmin") {
            return(data.frame(
                  x = c(-0.5, -0.5, -0.5, -0.5),
                  y = c(-0.5,  0.5, -0.5,  0.5),
                  z = c(-0.5, -0.5,  0.5,  0.5)
            ))
      } else if (face_name == "xmax") {
            return(data.frame(
                  x = c(0.5, 0.5, 0.5, 0.5),
                  y = c(-0.5, 0.5, -0.5, 0.5),
                  z = c(-0.5, -0.5, 0.5, 0.5)
            ))
      } else if (face_name == "ymin") {
            return(data.frame(
                  x = c(-0.5,  0.5, -0.5,  0.5),
                  y = c(-0.5, -0.5, -0.5, -0.5),
                  z = c(-0.5, -0.5,  0.5,  0.5)
            ))
      } else if (face_name == "ymax") {
            return(data.frame(
                  x = c(-0.5, 0.5, -0.5, 0.5),
                  y = c(0.5, 0.5, 0.5, 0.5),
                  z = c(-0.5, -0.5, 0.5, 0.5)
            ))
      } else if (face_name == "zmin") {
            return(data.frame(
                  x = c(-0.5,  0.5, -0.5,  0.5),
                  y = c(-0.5, -0.5,  0.5,  0.5),
                  z = c(-0.5, -0.5, -0.5, -0.5)
            ))
      } else if (face_name == "zmax") {
            return(data.frame(
                  x = c(-0.5, 0.5, -0.5, 0.5),
                  y = c(-0.5, -0.5, 0.5, 0.5),
                  z = c(0.5, 0.5, 0.5, 0.5)
            ))
      }
}

StatBar3D <- ggproto("StatBar3D", Stat,
                     required_aes = c("x", "y"),
                     optional_aes = c("weight"),
                     default_aes = aes(z = after_stat(count),
                                       group = after_stat(group)),

                     compute_panel = function(data, scales, na.rm = FALSE,
                                              bins = 10, binwidth = NULL, drop = TRUE,
                                              width = 1.0, faces = "all",
                                              light = NULL, cull_backfaces = NULL) {

                           # Remove missing values if requested
                           if (na.rm) {
                                 data <- data[complete.cases(data[c("x", "y")]), ]
                           }

                           # Check we have enough data
                           if (nrow(data) < 1) {
                                 stop("stat_bar_3d requires at least 1 point")
                           }

                           # Handle weight aesthetic
                           if (!"weight" %in% names(data)) {
                                 data$weight <- 1
                           }

                           # Detect discrete vs continuous
                           x_discrete <- inherits(scales$x, "ScaleDiscrete")
                           y_discrete <- inherits(scales$y, "ScaleDiscrete")

                           # Normalize bins/binwidth to length-2 vectors
                           if (!is.null(binwidth)) {
                                 if (length(binwidth) == 1) binwidth <- c(binwidth, binwidth)
                           }
                           if (length(bins) == 1) bins <- c(bins, bins)

                           # Normalize width to length-2 vector
                           if (length(width) == 1) width <- c(width, width)

                           # Dispatch to appropriate aggregation method
                           if (x_discrete && y_discrete) {
                                 agg_data <- compute_count_2d(data, drop)
                           } else if (!x_discrete && !y_discrete) {
                                 agg_data <- compute_bin_2d(data, bins, binwidth, drop)
                           } else {
                                 agg_data <- compute_bin_mixed(data, bins, binwidth, drop, x_discrete)
                           }

                           # Check if aggregation produced any data
                           if (nrow(agg_data) == 0) {
                                 return(data.frame(
                                       x = numeric(0), y = numeric(0), z = numeric(0),
                                       group = character(0), col_id = integer(0), face_type = character(0),
                                       count = numeric(0), proportion = numeric(0), ncount = numeric(0),
                                       density = numeric(0), ndensity = numeric(0)
                                 ))
                           }

                           # Compute normalized stat variables
                           agg_data <- compute_bar_stats(agg_data)

                           # Calculate grid spacing
                           x_spacing <- resolution(agg_data$x, zero = FALSE)
                           y_spacing <- resolution(agg_data$y, zero = FALSE)

                           # Validate and process faces parameter
                           selected_faces <- select_faces(faces)
                           if (length(selected_faces) == 0) {
                                 return(data.frame(
                                       x = numeric(0), y = numeric(0), z = numeric(0),
                                       group = character(0), col_id = integer(0), face_type = character(0),
                                       count = numeric(0), proportion = numeric(0), ncount = numeric(0),
                                       density = numeric(0), ndensity = numeric(0)
                                 ))
                           }

                           # Create columns using shared infrastructure
                           # Note: z values will be overwritten by after_stat(count), so we use z0 metadata
                           # to mark which vertices should be at baseline (z=0)
                           col_faces <- create_bar_cols(agg_data, x_spacing, y_spacing, width, selected_faces)

                           if (nrow(col_faces) == 0) {
                                 return(data.frame(
                                       x = numeric(0), y = numeric(0), z = numeric(0),
                                       group = character(0), col_id = integer(0), face_type = character(0),
                                       count = numeric(0), proportion = numeric(0), ncount = numeric(0),
                                       density = numeric(0), ndensity = numeric(0)
                                 ))
                           }

                           col_faces %>%
                                 average_aesthetics() %>%
                                 mutate(cull_backfaces = cull_backfaces) %>%
                                 attach_light(light) %>%
                                 return()
                     }
)


#' Create column faces from aggregated bar data
#'
#' Similar to create_cols but accepts separate x/y widths and adds z0 metadata
#' to mark baseline vertices (since after_stat will overwrite z values)
#'
#' @param data Data frame with x, y, count columns
#' @param x_spacing Grid spacing in x direction
#' @param y_spacing Grid spacing in y direction
#' @param width Vector of length 2 giving width factors for x and y
#' @param selected_faces Character vector of face names to render
#' @return Data frame with column face vertices including z0 column
#' @keywords internal
#' @noRd
create_bar_cols <- function(data, x_spacing, y_spacing, width, selected_faces) {

      # Calculate actual column dimensions (separate x and y widths)
      col_width_x <- x_spacing * width[1]
      col_width_y <- y_spacing * width[2]
      half_x <- col_width_x / 2
      half_y <- col_width_y / 2

      all_faces <- list()

      if (nrow(data) == 0) {
            warning("create_bar_cols: No data points provided")
            return(data.frame())
      }

      if (length(selected_faces) == 0) {
            warning("create_bar_cols: No faces selected")
            return(data.frame())
      }

      for (i in 1:nrow(data)) {
            point <- data[i, ]
            cx <- point$x
            cy <- point$y

            # Define bounds
            xmin_bound <- cx - half_x
            xmax_bound <- cx + half_x
            ymin_bound <- cy - half_y
            ymax_bound <- cy + half_y
            # Note: we store whether each corner is at z=0 (bottom) or z=height (top)
            # The actual z values will be set by after_stat(), then corrected in coord

            # Define the 8 corners of the column
            # Each corner: c(x, y, is_bottom)
            # is_bottom = TRUE means this vertex should be at z=0
            corners <- list(
                  list(x = xmin_bound, y = ymin_bound, z0 = TRUE),   # 1: left-front-bottom
                  list(x = xmax_bound, y = ymin_bound, z0 = TRUE),   # 2: right-front-bottom
                  list(x = xmax_bound, y = ymax_bound, z0 = TRUE),   # 3: right-back-bottom
                  list(x = xmin_bound, y = ymax_bound, z0 = TRUE),   # 4: left-back-bottom
                  list(x = xmin_bound, y = ymin_bound, z0 = FALSE),  # 5: left-front-top
                  list(x = xmax_bound, y = ymin_bound, z0 = FALSE),  # 6: right-front-top
                  list(x = xmax_bound, y = ymax_bound, z0 = FALSE),  # 7: right-back-top
                  list(x = xmin_bound, y = ymax_bound, z0 = FALSE)   # 8: left-back-top
            )

            # Define faces using corner indices (ccw when viewed from outside)
            face_definitions <- list(
                  zmin = c(1, 4, 3, 2),
                  zmax = c(5, 6, 7, 8),
                  xmin = c(1, 5, 8, 4),
                  xmax = c(2, 3, 7, 6),
                  ymin = c(1, 2, 6, 5),
                  ymax = c(4, 8, 7, 3)
            )

            for (face_name in selected_faces) {
                  if (face_name %in% names(face_definitions)) {
                        corner_indices <- face_definitions[[face_name]]
                        hierarchical_group <- paste0("col", i, "__", face_name)

                        face_vertices <- data.frame(
                              x = sapply(corner_indices, function(idx) corners[[idx]]$x),
                              y = sapply(corner_indices, function(idx) corners[[idx]]$y),
                              z0 = sapply(corner_indices, function(idx) corners[[idx]]$z0),
                              group = hierarchical_group,
                              col_id = i,
                              face_type = face_name
                        )

                        # Preserve all columns from aggregated data (count, density, etc.)
                        for (col_name in names(point)) {
                              if (!col_name %in% names(face_vertices)) {
                                    face_vertices[[col_name]] <- rep(point[[col_name]], 4)
                              }
                        }

                        all_faces[[length(all_faces) + 1]] <- face_vertices
                  }
            }
      }

      if (length(all_faces) == 0) {
            return(data.frame())
      }

      result <- do.call(rbind, all_faces)
      rownames(result) <- NULL
      return(result)
}


#' Count discrete x-y combinations
#'
#' @param data Data frame with x, y, weight columns
#' @param drop Whether to drop empty combinations
#' @return Data frame with x, y, count, bin_area columns
#' @keywords internal
#' @noRd
compute_count_2d <- function(data, drop = TRUE) {

      # Aggregate counts
      agg <- data %>%
            group_by(x, y) %>%
            summarise(count = sum(weight), .groups = "drop")

      # Expand to all combinations if drop = FALSE
      if (!drop) {
            all_combos <- expand.grid(
                  x = unique(data$x),
                  y = unique(data$y),
                  stringsAsFactors = FALSE
            )
            agg <- merge(all_combos, agg, by = c("x", "y"), all.x = TRUE)
            agg$count[is.na(agg$count)] <- 0
      }

      # For discrete data, bin_area is 1 (for density calculation consistency)
      agg$bin_area <- 1

      # preserve aesthetics (e.g. fill, color) mapped to x or y
      agg <- data %>% select(-weight, -group) %>% distinct() %>%
            left_join(agg, ., by = join_by(x, y))

      return(agg)
}


#' Bin continuous x and y data into 2D grid
#'
#' @param data Data frame with x, y, weight columns
#' @param bins Vector of length 2 giving number of bins in x and y
#' @param binwidth Vector of length 2 giving bin widths (overrides bins if provided)
#' @param drop Whether to drop empty bins
#' @return Data frame with x, y, count, bin_area columns
#' @keywords internal
#' @noRd
compute_bin_2d <- function(data, bins, binwidth, drop = TRUE) {

      # Compute bin edges
      x_range <- range(data$x, na.rm = TRUE)
      y_range <- range(data$y, na.rm = TRUE)

      if (!is.null(binwidth)) {
            x_breaks <- seq(x_range[1], x_range[2] + binwidth[1], by = binwidth[1])
            y_breaks <- seq(y_range[1], y_range[2] + binwidth[2], by = binwidth[2])
      } else {
            x_breaks <- seq(x_range[1], x_range[2], length.out = bins[1] + 1)
            y_breaks <- seq(y_range[1], y_range[2], length.out = bins[2] + 1)
      }

      # Compute bin widths
      x_width <- diff(x_breaks[1:2])
      y_width <- diff(y_breaks[1:2])
      bin_area <- x_width * y_width

      # Assign data to bins
      data$x_bin <- cut(data$x, breaks = x_breaks, include.lowest = TRUE, labels = FALSE)
      data$y_bin <- cut(data$y, breaks = y_breaks, include.lowest = TRUE, labels = FALSE)

      # Compute bin centers
      x_centers <- (x_breaks[-length(x_breaks)] + x_breaks[-1]) / 2
      y_centers <- (y_breaks[-length(y_breaks)] + y_breaks[-1]) / 2

      # Aggregate counts
      agg <- data %>%
            filter(!is.na(x_bin), !is.na(y_bin)) %>%
            group_by(x_bin, y_bin) %>%
            summarise(count = sum(weight), .groups = "drop")

      # Expand to all bins if drop = FALSE
      if (!drop) {
            all_bins <- expand.grid(
                  x_bin = seq_along(x_centers),
                  y_bin = seq_along(y_centers)
            )
            agg <- merge(all_bins, agg, by = c("x_bin", "y_bin"), all.x = TRUE)
            agg$count[is.na(agg$count)] <- 0
      }

      # Convert bin indices to center coordinates
      agg$x <- x_centers[agg$x_bin]
      agg$y <- y_centers[agg$y_bin]
      agg$bin_area <- bin_area

      # Clean up
      agg <- agg %>% select(-x_bin, -y_bin)

      return(agg)
}


#' Bin mixed discrete/continuous data
#'
#' @param data Data frame with x, y, weight columns
#' @param bins Vector of length 2 giving number of bins
#' @param binwidth Vector of length 2 giving bin widths
#' @param drop Whether to drop empty bins
#' @param x_discrete Logical indicating if x is discrete (y is continuous) or vice versa
#' @return Data frame with x, y, count, bin_area columns
#' @keywords internal
#' @noRd
compute_bin_mixed <- function(data, bins, binwidth, drop = TRUE, x_discrete) {

      if (x_discrete) {
            # x is discrete, bin y
            cont_var <- "y"
            disc_var <- "x"
            cont_bins <- bins[2]
            cont_binwidth <- if (!is.null(binwidth)) binwidth[2] else NULL
      } else {
            # y is discrete, bin x
            cont_var <- "x"
            disc_var <- "y"
            cont_bins <- bins[1]
            cont_binwidth <- if (!is.null(binwidth)) binwidth[1] else NULL
      }

      # Compute bin edges for continuous variable (globally across all discrete groups)
      cont_range <- range(data[[cont_var]], na.rm = TRUE)

      if (!is.null(cont_binwidth)) {
            cont_breaks <- seq(cont_range[1], cont_range[2] + cont_binwidth, by = cont_binwidth)
      } else {
            cont_breaks <- seq(cont_range[1], cont_range[2], length.out = cont_bins + 1)
      }

      cont_width <- diff(cont_breaks[1:2])
      cont_centers <- (cont_breaks[-length(cont_breaks)] + cont_breaks[-1]) / 2

      # Assign data to bins
      data$cont_bin <- cut(data[[cont_var]], breaks = cont_breaks, include.lowest = TRUE, labels = FALSE)

      # Aggregate counts
      agg <- data %>%
            filter(!is.na(cont_bin)) %>%
            group_by(across(all_of(disc_var)), cont_bin) %>%
            summarise(count = sum(weight), .groups = "drop")

      # Expand to all combinations if drop = FALSE
      if (!drop) {
            all_combos <- expand.grid(
                  disc_val = unique(data[[disc_var]]),
                  cont_bin = seq_along(cont_centers),
                  stringsAsFactors = FALSE
            )
            names(all_combos)[1] <- disc_var
            agg <- merge(all_combos, agg, by = c(disc_var, "cont_bin"), all.x = TRUE)
            agg$count[is.na(agg$count)] <- 0
      }

      # Convert bin indices to center coordinates
      agg[[cont_var]] <- cont_centers[agg$cont_bin]

      # bin_area is just the width of the continuous bin (discrete dimension has width 1)
      agg$bin_area <- cont_width

      # Clean up
      agg <- agg %>% select(-cont_bin)

      # preserve any aesthetics (e.g. fill, color) mapped to discrete x/y var
      if(cont_var == "x"){
            agg <- data %>% select(-x, -weight, -cont_bin) %>% distinct() %>%
                  left_join(agg, ., by = join_by(y))
      }else{
            agg <- data %>% select(-y, -weight, -cont_bin) %>% distinct() %>%
                  left_join(agg, ., by = join_by(x))
      }

      return(agg)
}


#' Compute normalized bar statistics
#'
#' @param agg_data Data frame with count and bin_area columns
#' @return Data frame with added proportion, ncount, density, ndensity columns
#' @keywords internal
#' @noRd
compute_bar_stats <- function(agg_data) {

      n <- sum(agg_data$count)

      agg_data <- agg_data %>%
            mutate(
                  proportion = count / n,
                  ncount = count / max(count, na.rm = TRUE),
                  density = count / (n * bin_area),
                  ndensity = density / max(density, na.rm = TRUE)
            )

      # Handle edge cases where max is 0 or n is 0
      agg_data$ncount[is.nan(agg_data$ncount)] <- 0
      agg_data$ndensity[is.nan(agg_data$ndensity)] <- 0
      agg_data$proportion[is.nan(agg_data$proportion)] <- 0
      agg_data$density[is.nan(agg_data$density)] <- 0

      return(agg_data)
}


#' 3D bar chart with automatic counting or binning
#'
#' Creates 3D bar charts by automatically counting discrete data or binning continuous data.
#' This is the 3D analogue of [ggplot2::geom_bar()] and [ggplot2::geom_histogram()].
#'
#' The stat automatically detects whether x and y are discrete or continuous:
#' \itemize{
#'   \item **Both discrete**: Counts occurrences of each (x, y) combination
#'   \item **Both continuous**: Performs 2D binning (like a 3D histogram)
#'   \item **Mixed**: Bins the continuous axis while preserving discrete groups
#' }
#'
#' For pre-computed bar heights, use [geom_col_3d()] instead.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use. Defaults to `StatBar3D`.
#' @param geom The geometric object used to display the data. Defaults to `GeomPolygon3D`.
#'
#' @param bins Number of bins in each dimension. Either a single value (used for both x and y)
#'   or a vector of length 2 giving `c(bins_x, bins_y)`. Default is 10. Ignored for discrete
#'   variables.
#' @param binwidth Bin width in each dimension. Either a single value or a vector of length 2
#'   giving `c(binwidth_x, binwidth_y)`. If provided, overrides `bins`. Ignored for discrete
#'   variables.
#' @param drop If `TRUE` (the default), empty bins/combinations are not rendered. If `FALSE`,
#'   empty bins render as zero-height columns.
#' @param width Column width as a fraction of bin spacing. Either a single value (used for both
#'   x and y) or a vector of length 2 giving `c(width_x, width_y)`. Default is 1.0
#'   (columns touch). Use values less than 1 for gaps between columns.
#'
#' @inheritParams col_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' `stat_bar_3d()` requires the following aesthetics:
#' - **x**: X coordinate
#' - **y**: Y coordinate
#'
#' And optionally understands:
#' - **weight**: Observation weights for counting
#'
#' @section Computed variables:
#' These variables can be used with `after_stat()` to map to aesthetics:
#' - `count`: Number of observations in each bin (default for z)
#' - `proportion`: Count divided by total count (sums to 1)
#' - `ncount`: Count scaled to maximum of 1
#' - `density`: Count divided by (total count × bin area); integrates to 1 for continuous data
#' - `ndensity`: Density scaled to maximum of 1
#'
#' @examples
#' # Discrete x and y: count combinations
#' d_discrete <- data.frame(
#'   x = sample(letters[1:4], 200, replace = TRUE),
#'   y = sample(LETTERS[1:3], 200, replace = TRUE)
#' )
#' ggplot(d_discrete, aes(x, y)) +
#'   geom_bar_3d() +
#'   coord_3d()
#'
#' # Continuous x and y: 2D histogram
#' d_cont <- data.frame(x = rnorm(1000), y = rnorm(1000))
#' ggplot(d_cont, aes(x, y)) +
#'   geom_bar_3d() +
#'   coord_3d()
#'
#' # Mixed: one discrete, one continuous
#' d_mixed <- data.frame(
#'   group = rep(c("A", "B", "C"), each = 100),
#'   value = c(rnorm(100, 2), rnorm(100, 1, 2), rnorm(100, 0))
#' )
#' ggplot(d_mixed, aes(x = group, y = value, fill = group)) +
#'   geom_bar_3d(bins = 20, width = c(.5, 1)) +
#'   coord_3d(scales = "fixed", ratio = c(1, 1, .1))
#'
#' # Use density instead of count for z
#' ggplot(d_mixed,
#'       aes(x = group, y = value, z = after_stat(density))) +
#'   geom_bar_3d(bins = 20, width = c(.5, 1)) +
#'   coord_3d()
#'
#' # Show empty bins with drop = FALSE
#' ggplot(d_cont, aes(x, y)) +
#'   geom_bar_3d(drop = FALSE) +
#'   coord_3d()
#'
#' @seealso [geom_col_3d()] for pre-computed heights, [coord_3d()] for 3D coordinate systems,
#'   [light()] for lighting specifications.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_bar_3d
#' @export
geom_bar_3d <- function(mapping = NULL, data = NULL,
                        stat = StatBar3D,
                        position = "identity",
                        ...,
                        bins = 10, binwidth = NULL, drop = TRUE,
                        width = 1.0, faces = "all",
                        light = NULL,
                        cull_backfaces = TRUE, sort_method = NULL,
                        scale_depth = TRUE, force_convex = FALSE,
                        na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomPolygon3D,
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, bins = bins, binwidth = binwidth, drop = drop,
                          width = width, faces = faces, light = light,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}


#' @rdname geom_bar_3d
#' @export
stat_bar_3d <- function(mapping = NULL, data = NULL,
                        geom = GeomPolygon3D,
                        position = "identity",
                        ...,
                        bins = 10, binwidth = NULL, drop = TRUE,
                        width = 1.0, faces = "all",
                        light = NULL,
                        cull_backfaces = TRUE, sort_method = NULL,
                        scale_depth = TRUE, force_convex = FALSE,
                        na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = StatBar3D, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, bins = bins, binwidth = binwidth, drop = drop,
                          width = width, faces = faces, light = light,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}

GeomPolygon3D <- ggproto("GeomPolygon3D", Geom,

                         required_aes = c("x", "y", "z", "group"),

                         optional_aes = c("subgroup"),

                         extra_params = c("na.rm", "annotate"),

                         default_aes = aes(
                               fill = "grey50", colour = "grey50", linewidth = 0.1, linetype = 1, alpha = 1
                         ),

                         setup_data = function(data, params) {
                               setup_annotations(data, params)
                         },

                         draw_panel = function(data, panel_params, coord,
                                               sort_method = "auto",
                                               scale_depth = TRUE,
                                               force_convex = FALSE,
                                               rule = "evenodd",
                                               annotate = NULL,
                                               point.colour = NULL, point.fill = NULL,
                                               point.size = NULL, point.shape = NULL,
                                               point.alpha = NULL, point.stroke = NULL,
                                               residual.colour = NULL, residual.linewidth = NULL,
                                               residual.linetype = NULL, residual.alpha = NULL) {

                               # Parameter validation
                               validate_coord3d(coord)

                               # Resolve annotation sentinels that need panel range info
                               # (no-op for v1 types; will expand plane extents when added)
                               data <- prepare_annotations(data, panel_params)

                               # Transform data
                               sort_method <- match.arg(sort_method, c("auto", "pairwise", "painter"))
                               data$.sort_method <- sort_method
                               if (!".prim" %in% names(data)) {
                                     data$.prim <- "polygon"
                               } else {
                                     data$.prim[is.na(data$.prim)] <- "polygon"
                               }
                               coords <- coord$transform(data, panel_params)

                               # Enforce convexity if requested (polygons only)
                               coords <- drop_nonconvex_vertices(coords, force_convex)

                               # Apply smooth overlay point styling (post-transform, pre-depth-scale).
                               # Values come from geom_smooth_3d point.* params.
                               # Only applied when params are non-NULL (i.e. when smooth overlay
                               # is active). Annotation points are styled separately via
                               # apply_annotation_styles().
                               if (".prim" %in% names(coords) && any(coords$.prim == "point", na.rm = TRUE)) {
                                     pt_mask <- coords$.prim == "point" & !is.na(coords$.prim)

                                     # Initialize point-specific columns if they don't exist
                                     for (col in c("size", "shape", "stroke")) {
                                           if (!col %in% names(coords)) coords[[col]] <- NA_real_
                                     }

                                     if (!is.null(point.colour)) coords$colour[pt_mask] <- point.colour
                                     if (!is.null(point.fill))   coords$fill[pt_mask]   <- point.fill
                                     if (!is.null(point.size))   coords$size[pt_mask]   <- point.size
                                     if (!is.null(point.shape))  coords$shape[pt_mask]  <- point.shape
                                     if (!is.null(point.alpha))  coords$alpha[pt_mask]  <- point.alpha
                                     if (!is.null(point.stroke)) coords$stroke[pt_mask] <- point.stroke
                               }

                               # Apply smooth overlay segment styling (residual lines)
                               if (".prim" %in% names(coords) && any(coords$.prim == "segment", na.rm = TRUE)) {
                                     seg_mask <- coords$.prim == "segment" & !is.na(coords$.prim)

                                     if (!is.null(residual.colour))    coords$colour[seg_mask]    <- residual.colour
                                     if (!is.null(residual.alpha))     coords$alpha[seg_mask]     <- residual.alpha
                                     if (!is.null(residual.linewidth)) coords$linewidth[seg_mask] <- residual.linewidth
                                     if (!is.null(residual.linetype))  coords$linetype[seg_mask]  <- residual.linetype
                               }

                               # Apply annotate_3d styles (overrides blanket smooth overlay styling above for annotate_3d rows only)
                               coords <- apply_annotation_styles(coords)

                               # Scale linewidths/sizes by depth
                               coords <- apply_depth_scaling(coords, scale_depth)

                               # Apply light blending to colors
                               coords <- blend_light(coords)

                               if (!"group" %in% names(coords)) {
                                     warning("No group column found in polygon data")
                                     return(grid::nullGrob())
                               }

                               if (nrow(coords) == 0) {
                                     return(grid::nullGrob())
                               }

                               render_mixed_grobs(coords, rule = rule)
                         },

                         draw_key = draw_key_polygon
)


#' 3D polygon geometry with depth sorting
#'
#' `geom_polygon_3d()` renders 3D polygons with proper depth sorting for realistic
#' 3D surface visualization. It's designed to work with surface data
#' from [stat_hull_3d()] and [stat_surface_3d()], as well as regular polygon data like maps.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data The data to be displayed in this layer. Note that if you specify `light` or
#'   `cull_backfaces`, behavior will depend on the "winding order" of polygon vertices, with
#'   the counter-clockwise face considered the "front".
#' @param stat The statistical transformation to use on the data. Defaults to `"identity_3d"`.
#' @param rule Either `"evenodd"` or `"winding"`. If polygons with holes are
#'   being drawn (using the `subgroup` aesthetic) this argument defines how the
#'   hole coordinates are interpreted. See [ggplot2::geom_polygon()] for reference.
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' `geom_polygon_3d()` requires:
#' - **x**: X coordinate
#' - **y**: Y coordinate
#' - **z**: Z coordinate (for depth sorting)
#' - **group**: Polygon grouping variable
#'
#' And understands these additional aesthetics:
#' - `fill`: Polygon fill color
#' - `colour`: Border color
#' - `linewidth`: Border line width
#' - `linetype`: Border line type
#' - `alpha`: Transparency
#' - `subgroup`: Secondary grouping for polygons with holes
#' - `order`: Vertex order within polygons (for proper polygon construction)
#'
#' @examples
#' # Typically used via stats such as stat_surface_3d() or stat_hull_3d()
#' ggplot(sphere_points, aes(x, y, z)) +
#'   stat_hull_3d(method = "convex", fill = "dodgerblue",
#'             light = light(fill = TRUE, mode = "hsl")) +
#'   coord_3d()
#'
#' # Can be used directly with properly structured data
#' triangles <- data.frame(x = rep(c(3, 2, 1), 3),
#'                             y = rep(c(1, 3, 1), 3),
#'                             z = rep(1:3, each = 3),
#'                             shape = rep(letters[1:3], each = 3))
#' ggplot(triangles, aes(x, y, z, fill = shape)) +
#'   geom_polygon_3d(color = "black") +
#'   coord_3d()
#'
#' # Use `sort_method` to choose between depth sorting algorithms
#' d <- data.frame(group = rep(letters[1:3], each = 4),
#'                 x = c(1, 1, 2, 2,   1, 1, 3, 3,   2, 2, 3, 3),
#'                 y = rep(c(1, 2, 2, 1), 3),
#'                 z = rep(c(1, 1.5, 2), each = 4))
#' p <- ggplot(d, aes(x, y, z, group = group, fill = group)) +
#'       coord_3d(pitch = 50, roll = 20, yaw = 0,
#'                scales = "fixed", light = "none") +
#'       theme_light()
#'
#' # fast, but rendering order is incorrect in this particular example
#' p + geom_polygon_3d(color = "black", linewidth = 1, alpha = .75,
#'       sort_method = "painter")
#'
#' # correct rendering order (but slower for large data sets)
#' p + geom_polygon_3d(color = "black", linewidth = 1, alpha = .75,
#'       sort_method = "pairwise")
#'
#' @return A `Layer` object that can be added to a ggplot.
#' @export
geom_polygon_3d <- function(mapping = NULL, data = NULL, stat = StatIdentity3D,
                            position = "identity",
                            ...,
                            rule = "evenodd",
                            sort_method = "auto", scale_depth = TRUE, force_convex = FALSE,
                            cull_backfaces = FALSE,
                            light = NULL,
                            na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(
            geom = GeomPolygon3D, mapping = mapping, data = data, stat = get_proto(stat, "stat"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(rule = rule,
                          sort_method = sort_method,
                          scale_depth = scale_depth,
                          force_convex = force_convex,
                          cull_backfaces = cull_backfaces,
                          light = light,
                          na.rm = na.rm, ...)
      )
}

# force polygons to be convex, to prevent artifacts in surfaces that wrap beyond visible horizon
drop_nonconvex_vertices <- function(data, force_convex){
      if(!force_convex) return(data)
      data %>%
            group_by(group) %>%
            mutate(simple = n() <= 3,
                   keep = simple[1] | 1:n() %in% chull(x, y)) %>%
            ungroup() %>%
            filter(keep) %>%
            select(-keep, -simple) %>%
            return()
}

#' 3D line segments
#'
#' `geom_segment_3d()` and `stat_segment_3d()` draw line segments in 3D space
#' with automatic depth-based linewidth scaling and proper depth sorting.
#' Each segment is defined by start coordinates (x, y, z) and end
#' coordinates (xend, yend, zend).
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. Requires x, y, z
#'   for start coordinates and xend, yend, zend for end coordinates.
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use on the data. Defaults to `"segment_3d"`.
#' @param geom The geometric object used to display the data. Defaults to `"segment_3d"`.
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed on to [ggplot2::layer()].
#' @param na.rm If `FALSE`, missing values are removed with a warning.
#' @param show.legend Logical indicating whether this layer should be included in legends.
#' @param inherit.aes If `FALSE`, overrides the default aesthetics.
#' @inheritParams sort_params
#' @inheritParams depth_params
#' @param arrow Specification for arrow heads, created by [arrow()].
#' @param lineend Line end style, one of "round", "butt", "square".
#'
#' @section Aesthetics:
#' `geom_segment_3d()` understands the following aesthetics:
#' - **x, y, z**: Start coordinates (required)
#' - **xend, yend, zend**: End coordinates (required)
#' - `colour`: Line color
#' - `linewidth`: Line width (gets depth-scaled when `scale_depth = TRUE`)
#' - `linetype`: Line type
#' - `alpha`: Transparency
#'
#' @examples
#' # Basic 3D segments
#' ggplot(sphere_points,
#'       aes(x, y, z, xend = 0, yend = 0, zend = 0)) +
#'   geom_segment_3d() +
#'   coord_3d()
#'
#' # 3D vector field
#' data <- expand.grid(x = -1:2, y = -1:2, z = -1:2)
#' data2 <- data + seq(-.5, .5, length.out = length(as.matrix(data)))
#' data <- cbind(data, setNames(data2, c("x2", "y2", "z2")))
#' ggplot(data, aes(x, y, z,
#'       xend = x2, yend = y2, zend = z2, color = x)) +
#'   geom_segment_3d(arrow = arrow(length = unit(0.1, "inches"),
#'                   type = "closed", angle = 15),
#'                   linewidth = .5) +
#'   coord_3d()
#'
#' @seealso [geom_path_3d()] for connected paths.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_segment_3d
#' @export
geom_segment_3d <- function(mapping = NULL, data = NULL,
                            stat = StatSegment3D, position = "identity",
                            ...,
                            sort_method = "painter",
                            scale_depth = TRUE, arrow = NULL, lineend = "butt",
                            na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomSegment3D,
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(
                  na.rm = na.rm,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  arrow = arrow,
                  lineend = lineend,
                  ...
            )
      )
}

#' @rdname geom_segment_3d
#' @export
stat_segment_3d <- function(mapping = NULL, data = NULL,
                            geom = GeomSegment3D, position = "identity",
                            ...,
                            sort_method = "painter",
                            scale_depth = TRUE, arrow = NULL, lineend = "butt",
                            na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = StatSegment3D, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(
                  na.rm = na.rm,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  arrow = arrow,
                  lineend = lineend,
                  ...
            )
      )
}


StatSegment3D <- ggproto("StatSegment3D", Stat,
                         required_aes = c("x", "y", "z", "xend", "yend", "zend"),

                         compute_group = function(data, scales, na.rm = FALSE) {

                               # Remove missing values if requested
                               if (na.rm) {
                                     complete_cases <- complete.cases(data[c("x", "y", "z", "xend", "yend", "zend")])
                                     data <- data[complete_cases, ]
                               }

                               if (nrow(data) == 0) {
                                     return(data)
                               }

                               # Handle discrete scale conversion for start points
                               if ("z" %in% names(data)) {
                                     data$z_raw <- data$z
                                     if (is.factor(data$z) || is.character(data$z)) {
                                           data$z <- as.numeric(as.factor(data$z))
                                     }
                               }

                               if ("x" %in% names(data)) {
                                     data$x_raw <- data$x
                                     if (is.factor(data$x) || is.character(data$x)) {
                                           data$x <- as.numeric(as.factor(data$x))
                                     }
                               }

                               if ("y" %in% names(data)) {
                                     data$y_raw <- data$y
                                     if (is.factor(data$y) || is.character(data$y)) {
                                           data$y <- as.numeric(as.factor(data$y))
                                     }
                               }

                               # Handle discrete scale conversion for end points
                               if ("zend" %in% names(data)) {
                                     data$zend_raw <- data$zend
                                     if (is.factor(data$zend) || is.character(data$zend)) {
                                           data$zend <- as.numeric(as.factor(data$zend))
                                     }
                               }

                               if ("xend" %in% names(data)) {
                                     data$xend_raw <- data$xend
                                     if (is.factor(data$xend) || is.character(data$xend)) {
                                           data$xend <- as.numeric(as.factor(data$xend))
                                     }
                               }

                               if ("yend" %in% names(data)) {
                                     data$yend_raw <- data$yend
                                     if (is.factor(data$yend) || is.character(data$yend)) {
                                           data$yend <- as.numeric(as.factor(data$yend))
                                     }
                               }

                               # Create hierarchical group IDs if not already present
                               if (!"group" %in% names(data) || !any(grepl("__", data$group))) {
                                     original_group <- data$group[1] %||% "-1"
                                     data$group <- paste0(original_group, "__seg", 1:nrow(data))
                               }

                               data <- average_aesthetics(data)

                               return(data)
                         }
)

GeomSegment3D <- ggproto("GeomSegment3D", Geom,
                         required_aes = c("x", "y", "z", "xend", "yend", "zend"),
                         default_aes = aes(
                               colour = "black", linewidth = 0.5, linetype = 1, alpha = 1
                         ),

                         extra_params = c("na.rm", "annotate"),

                         setup_data = function(data, params) {
                               # Convert from wide format (x/y/z + xend/yend/zend) to
                               # long format (two rows per segment) for the rendering
                               # pipeline. Each segment becomes a pair of start/end rows
                               # sharing a group ID.
                               n_segments <- nrow(data)

                               start_data <- data.frame(
                                     x = data$x,
                                     y = data$y,
                                     z = data$z,
                                     group = data$group,
                                     segment_id = seq_len(n_segments),
                                     point_type = "start",
                                     stringsAsFactors = FALSE
                               )

                               end_data <- data.frame(
                                     x = data$xend,
                                     y = data$yend,
                                     z = data$zend,
                                     group = data$group,
                                     segment_id = seq_len(n_segments),
                                     point_type = "end",
                                     stringsAsFactors = FALSE
                               )

                               # Preserve aesthetics and other columns for both points
                               other_cols <- setdiff(names(data),
                                                     c("x", "y", "z", "xend", "yend", "zend", "group"))
                               for (col_name in other_cols) {
                                     start_data[[col_name]] <- data[[col_name]]
                                     end_data[[col_name]] <- data[[col_name]]
                               }

                               long_data <- rbind(start_data, end_data)
                               long_data$.prim <- "segment"

                               setup_annotations(long_data, params)
                         },

                         draw_panel = function(data, panel_params, coord,
                                               annotate = NULL,
                                               sort_method = "painter",
                                               scale_depth = TRUE,
                                               arrow = NULL, lineend = "butt", na.rm = FALSE) {

                               validate_coord3d(coord)

                               if (nrow(data) == 0) {
                                     return(grid::nullGrob())
                               }

                               data <- prepare_annotations(data, panel_params)

                               sort_method <- match.arg(sort_method, c("auto", "pairwise", "painter"))
                               data$.sort_method <- sort_method

                               # Transform all points together (handles depth sorting)
                               coords <- coord$transform(data, panel_params)

                               # Apply annotate_3d styles
                               coords <- apply_annotation_styles(coords)

                               # Apply depth scaling to linewidth
                               coords <- apply_depth_scaling(coords, scale_depth)

                               if (nrow(coords) == 0) {
                                     return(grid::nullGrob())
                               }

                               render_mixed_grobs(coords, arrow = arrow, lineend = lineend)
                         },

                         draw_key = draw_key_path
)

#' Create point rows from raw data for mixed-geometry rendering
#'
#' Takes raw data points and prepares them for inclusion alongside polygon
#' surfaces in a mixed-geometry layer. Assigns unique group IDs and sets
#' `.prim = "point"`. Aesthetic columns are stripped so that annotation
#' points don't contaminate scale training; styling is applied at render
#' time in the geom's `draw_panel` via override parameters.
#'
#' @param data Data frame with x, y, z and aesthetic columns.
#' @param group_prefix Prefix for point group IDs.
#' @return Data frame with one row per point, tagged with `.prim = "point"`.
#' @keywords internal
#' @noRd
make_point_rows <- function(data, group_prefix = "data_point") {

      n <- nrow(data)

      # Keep only positional and structural columns — strip mapped aesthetics
      # so annotation points don't contaminate scale training
      keep_cols <- intersect(names(data),
                             c("x", "y", "z", "fitted", "se", "level"))
      pt_data <- data[, keep_cols, drop = FALSE]

      pt_data$group <- paste0(group_prefix, "__pt", seq_len(n))
      pt_data$.prim <- "point"

      return(pt_data)
}


#' Create segment rows connecting data points to a surface
#'
#' For each data point at (x, y, z), creates a two-row segment group
#' connecting the point to the surface at (x, y, fitted). Used for
#' residual lines in geom_smooth_3d. Aesthetic columns are stripped so
#' annotations don't contaminate scale training; styling is applied at
#' render time via override parameters.
#'
#' @param data Data frame with x, y, z columns (raw data points).
#' @param fitted Numeric vector of fitted z values at each data point.
#' @param group_prefix Prefix for segment group IDs.
#' @return Data frame with two rows per segment, tagged with `.prim = "segment"`.
#' @keywords internal
#' @noRd
make_residual_segment_rows <- function(data, fitted,
                                       group_prefix = "residual_line") {

      n <- nrow(data)
      seg_groups <- paste0(group_prefix, "__seg", seq_len(n))

      # Start points (raw data)
      start_rows <- data.frame(
            x = data$x, y = data$y, z = data$z,
            group = seg_groups,
            point_type = "start",
            segment_id = seq_len(n)
      )

      # End points (on the surface)
      end_rows <- data.frame(
            x = data$x, y = data$y, z = fitted,
            group = seg_groups,
            point_type = "end",
            segment_id = seq_len(n)
      )

      # Combine — no aesthetic columns carried forward
      seg_data <- rbind(start_rows, end_rows)
      seg_data$.prim <- "segment"

      return(seg_data)
}


#' Bind rows with automatic type coercion for aesthetic columns
#'
#' Coerces shared colour/fill columns to character before calling `bind_rows`,
#' preventing type conflicts when one data frame has logical `NA` (from
#' unmapped aesthetics) and the other has character values.
#'
#' @param ... Data frames to bind.
#' @return Combined data frame.
#' @keywords internal
#' @noRd
safe_bind_rows <- function(...) {
      dfs <- list(...)
      colour_cols <- c("colour", "fill")
      for (col in colour_cols) {
            has_col <- vapply(dfs, function(df) col %in% names(df), logical(1))
            if (any(has_col)) {
                  for (i in which(has_col)) {
                        dfs[[i]][[col]] <- as.character(dfs[[i]][[col]])
                  }
            }
      }
      do.call(bind_rows, dfs)
}




# Aggregate numeric values per polygon group.
# Averages all numeric columns except positional/structural ones, so that
# computed variables (e.g. `fitted`, `density`, `se`) and visual aesthetics
# (e.g. `fill`, `colour`) get per-polygon values before ggplot2 resolves
# `after_stat()` mappings or picks a single vertex's value.
average_aesthetics <- function(data,
                               exclude = c("x", "y", "z", "order",
                                           ".vertex_order", "PANEL",
                                           "row", "column")){

      numeric_cols <- names(data)[sapply(data, is.numeric)]
      avg_cols <- setdiff(numeric_cols, exclude)

      if (length(avg_cols) > 0) {
            data <- data %>%
                  group_by(group) %>%
                  mutate(across(all_of(avg_cols),
                                ~ if(is.numeric(.x)) mean(.x, na.rm = TRUE) else .x)) %>%
                  ungroup()
      }

      return(data)
}

get_proto <- function(x, type = c("geom", "stat")){
      if(!inherits(x, "character")) return(x)
      if(!grepl("_3d", x)) return(x)
      cap <- function(s) paste0(toupper(substr(s, 1, 1)), substr(s, 2, nchar(s)))
      get(paste0(cap(type), cap(gsub("_3d", "3D", x))))
}


#' Find the plot object currently being built
#'
#' Several pieces of plot-level state -- the labels, a `+ light()` stash, the
#' z scale -- live on the plot, but `Coord3D` methods and their helpers are
#' handed only scales and params. This walks the calling frames to reach the
#' plot that `ggplot_build()` is working on.
#'
#' The walk stops at the global environment, so a user object that happens to
#' be named `plot` can't be picked up when the call chain runs out before
#' `max_frames` does.
#'
#' @param max_frames Maximum number of calling frames to inspect.
#' @return The `ggplot` object being built, or `NULL` if it can't be reached.
#' @keywords internal
#' @noRd
find_build_plot <- function(max_frames = 40) {
      found <- NULL
      tryCatch({
            for (i in seq_len(max_frames)) {
                  env <- parent.frame(i)
                  if (identical(env, globalenv())) break
                  if (exists("plot", envir = env, inherits = FALSE)) {
                        candidate <- get("plot", envir = env, inherits = FALSE)
                        if (inherits(candidate, "ggplot")) {
                              found <- candidate
                              break
                        }
                  }
            }
      }, error = function(e) {
            # If frame walking fails, callers fall back to their defaults
      })
      found
}

#' Rotate points in 3D space
#'
#' @param xyz Matrix of x, y, z coordinates
#' @param pitch Rotation around x-axis in degrees
#' @param roll Rotation around y-axis in degrees
#' @param yaw Rotation around z-axis in degrees
#' @return Rotated coordinates matrix
#' @keywords internal
#' @noRd
rotate_3d <- function(xyz, pitch, roll, yaw) {
      # Convert angles to radians
      pitch_rad <- pitch * pi / 180
      roll_rad <- roll * pi / 180
      yaw_rad <- yaw * pi / 180

      # Create rotation matrix
      rot <- matrix(c(
            cos(yaw_rad) * cos(pitch_rad),
            cos(yaw_rad) * sin(pitch_rad) * sin(roll_rad) - sin(yaw_rad) * cos(roll_rad),
            cos(yaw_rad) * sin(pitch_rad) * cos(roll_rad) + sin(yaw_rad) * sin(roll_rad),

            sin(yaw_rad) * cos(pitch_rad),
            sin(yaw_rad) * sin(pitch_rad) * sin(roll_rad) + cos(yaw_rad) * cos(roll_rad),
            sin(yaw_rad) * sin(pitch_rad) * cos(roll_rad) - cos(yaw_rad) * sin(roll_rad),

            -sin(pitch_rad),
            cos(pitch_rad) * sin(roll_rad),
            cos(pitch_rad) * cos(roll_rad)
      ), nrow = 3, byrow = TRUE)

      # Apply rotation
      rotated <- xyz %*% rot

      return(rotated)
}

#' Apply perspective effect to 3D points
#'
#' @param rotated Matrix of rotated x, y, z coordinates
#' @param dist Distance from viewer to center of scene
#' @return Matrix with perspective effect applied
#' @keywords internal
#' @noRd
apply_perspective <- function(rotated, dist) {
      z_offset <- rotated[, 3] + dist
      perspective_factor <- dist / z_offset
      rotated[, 1] <- rotated[, 1] * perspective_factor
      rotated[, 2] <- rotated[, 2] * perspective_factor
      return(rotated)
}

#' Transform 3D points using rotation and optional perspective with viewpoint distance
#'
#' @param data Data frame with x, y, z columns (in standard `[-0.5, 0.5]` domain)
#' @param proj A list of projection parameters
#' @return Data frame with transformed coordinates, depth for sorting, and depth_scale for size scaling
#' @keywords internal
#' @noRd
transform_3d_standard <- function(data, proj = list(pitch = 0, roll = 0, yaw = 0, persp = TRUE, dist = 2)) {

      pitch <- proj$pitch
      roll <- proj$roll
      yaw <- proj$yaw
      persp <- proj$persp
      dist <- proj$dist

      # Flip z coordinates, to support a more intuitive default projection
      # in which all axes increase toward top-right plot corner
      data[,"z"] <- data[,"z"] * -1

      # Apply rotation to data points
      rotated <- rotate_3d(as.matrix(data[, c("x", "y", "z")]), pitch, roll, yaw)

      # Calculate reference depth for point (0,0,0) - the center of the cube
      reference_point <- matrix(c(0, 0, 0), nrow = 1)
      rotated_ref <- rotate_3d(reference_point, pitch, roll, yaw)

      if (persp) {
            # Euclidean distance from viewpoint for sorting
            depth <- sqrt(rotated[, 1]^2 + rotated[, 2]^2 + (rotated[, 3] + dist)^2)

            # Z-axis depth for size scaling (distance along view axis only)
            z_depth <- rotated[, 3] + dist
            reference_z_depth <- rotated_ref[1, 3] + dist

            # Apply perspective transformation
            transformed <- apply_perspective(rotated, dist)

            # Return transformed coordinates with viewpoint depth and depth scaling
            return(data.frame(
                  x = transformed[, 1],
                  y = transformed[, 2],
                  z = transformed[, 3],  # Keep z for face visibility calculations
                  depth = depth,         # True viewpoint distance for sorting
                  depth_scale = reference_z_depth / z_depth  # Scaling factor for size/linewidth
            ))
      } else {
            # Orthographic: no depth-based size scaling (parallel projection)
            return(data.frame(
                  x = rotated[, 1],
                  y = rotated[, 2],
                  z = rotated[, 3],
                  depth = rotated[, 3],
                  depth_scale = 1
            ))
      }
}

#' Scale data to standard domain with aspect ratio control
#'
#' @param values Vector of values to scale (for single axis) OR data frame with x,y,z columns (for multi-axis)
#' @param data_range Original range of the data `c(min, max)` (for single axis) OR list with x,y,z scale ranges (for multi-axis)
#' @param scales Aspect ratio behavior ("free" or "fixed") (only used for multi-axis)
#' @param ratio Length-3 numeric vector of axis ratios (only used for multi-axis)
#' @return Scaled values in `[-0.5, 0.5]` domain (single axis) OR data frame with scaled coordinates (multi-axis)
#' @keywords internal
#' @noRd
scale_to_standard <- function(values, data_range, scales = "free", ratio = c(1, 1, 1)) {

      # Handle single-axis case
      if (is.numeric(data_range) && length(data_range) == 2) {
            # Handle case where data has zero range (single point or constant values)
            range_width <- diff(data_range)
            if (range_width == 0 || is.na(range_width)) {
                  # If all values are the same, center them at 0 in standard domain
                  return(rep(0, length(values)))
            }

            # for discrete scales, get the position rather than the raw value
            if(!is.numeric(values)) values <- attr(values, "pos")

            # Scale to [0, 1] then shift to [-0.5, 0.5]
            return((values - data_range[1]) / range_width - 0.5)
      }

      # Handle multi-axis case
      data <- values
      scale_ranges <- data_range  # This is actually scale limits, not raw data ranges

      # Compute effective ratios based on aspect setting
      effective_ratios <- compute_effective_ratios(scale_ranges, scales, ratio)

      # Apply standard scaling for each axis
      return(data.frame(x = scale_to_standard(data$x, scale_ranges$x) * effective_ratios[1],
                        y = scale_to_standard(data$y, scale_ranges$y) * effective_ratios[2],
                        z = scale_to_standard(data$z, scale_ranges$z) * effective_ratios[3]))
}

# Helper function to compute effective ratios
compute_effective_ratios <- function(scale_ranges, scales, ratio) {
      if (scales == "free") {
            # Apply ratios directly to standardized coordinates
            effective_ratios <- ratio
      } else {
            # scales == "fixed": adjust ratios based on scale ranges (including expansion)
            x_span <- diff(scale_ranges$x)
            y_span <- diff(scale_ranges$y)
            z_span <- diff(scale_ranges$z)

            # Handle zero-width ranges
            if (x_span == 0) x_span <- 1
            if (y_span == 0) y_span <- 1
            if (z_span == 0) z_span <- 1

            # Apply user ratios to scale-space proportions
            scale_proportional <- c(x_span, y_span, z_span)
            effective_ratios <- ratio * scale_proportional
      }

      # Normalize so largest dimension gets [-0.5, 0.5]
      max_ratio <- max(effective_ratios)
      return(effective_ratios / max_ratio)
}

#' Convert text to polygon outlines
#'
#' Converts a text string into polygon vertices suitable for plotting with
#' ggplot2's geom_polygon. Uses systemfonts for font matching and glyph extraction.
#'
#' @param text Character string to convert
#' @param family Font family name (e.g., "Arial", "Helvetica", "sans")
#' @param weight Font weight ("normal", "bold", "thin", "light", etc.)
#' @param italic Logical; use italic variant
#' @param size Font size in points
#' @param tolerance Bezier curve tolerance; lower values give more detailed outlines
#' @param spacing Additional letter spacing in em units
#' @param hjust,vjust The justification of the textbox surrounding the text
#' @param width Maximum line width in inches for word wrapping; NA disables wrapping
#' @param align Text alignment: "left", "center", or "right"
#' @param lineheight Line height multiplier
#' @param internal_size Internal rendering size for kerning precision; generally
#'   shouldn't need adjustment
#'
#' @return A data frame with columns: x, y, contour, glyph, letter, poly_id. X and y
#'    are vertex coordinates, in point units.
#'
#' @examples
#' data <- text_outlines("Howdy, partner", family = "Arial")
#'
#' ggplot2::ggplot(data, ggplot2::aes(x, y, group = letter, subgroup = poly_id)) +
#'   ggplot2::geom_polygon() +
#'   ggplot2::coord_fixed()
#'
#' @export
text_outlines <- function(text,
                          family = "sans",
                          weight = "normal",
                          italic = FALSE,
                          size = 12,
                          tolerance = 0.01,
                          spacing = 0,
                          hjust = 0.5,
                          vjust = 0.5,
                          width = NA,
                          align = "left",
                          lineheight = 1,
                          internal_size = 144) {
      scale_factor <- size / internal_size
      font_match <- systemfonts::match_fonts(family, weight = weight, italic = italic)
      f_path <- font_match$path
      f_index <- font_match$index
      internal_width <- if (!is.na(width)) width / scale_factor else NA
      shaped <- systemfonts::shape_string(
            text,
            path = f_path,
            index = f_index,
            size = internal_size,
            tracking = spacing * 1000,
            hjust = hjust, vjust = vjust,
            width = internal_width,
            align = align,
            lineheight = lineheight
      )
      glyph_positions <- shaped$shape
      all_glyphs <- lapply(seq_len(nrow(glyph_positions)), function(i) {
            outlines <- systemfonts::glyph_outline(
                  glyph = glyph_positions$index[i],
                  path = f_path,
                  index = f_index,
                  size = internal_size,
                  tolerance = tolerance / scale_factor
            )
            if (is.null(outlines) || nrow(outlines) == 0) {
                  return(NULL)
            }
            outlines$x <- (outlines$x + glyph_positions$x_offset[i]) * scale_factor
            outlines$y <- (outlines$y + glyph_positions$y_offset[i]) * scale_factor
            outlines$letter <- glyph_positions$glyph[i]
            outlines$poly_id <- paste0(i, "_", outlines$contour)
            outlines
      })

      result <- dplyr::bind_rows(all_glyphs)

      # Ensure proper winding order for backface identification
      result <- result %>%
            group_by(contour) %>%
            arrange(desc(row_number())) %>%
            ungroup()

      result
}


#' Compute camera-facing direction for 3D text
#'
#' Creates a specification for camera-facing "billboard" text that will compute
#' the appropriate facing direction for each label position.
#'
#' @param coord Optional objected created by `coord_3d`. If provided, coord
#'   parameters are drawn from this and the `pitch`, `roll`, `yaw`, `dist` arguments
#'   are ignored.
#' @param pitch,roll,yaw Rotation angles in degrees, matching the values
#'   you'll use in `coord_3d()`.
#' @param dist Distance from viewer to center of the data cube, matching
#'   the value you'll use in `coord_3d()`. Default is 2. Only used when
#'   `mode = "point"`.
#' @param mode Character string specifying how labels should face the camera:
#'   \itemize{
#'     \item `"plane"` (default): All labels face parallel to the view plane,
#'       like billboards. This is typically easier to read.
#'     \item `"point"`: Each label faces toward the camera position. Labels
#'       at the edges will angle inward, giving perspective-correct orientation.
#'   }
#'
#' @return A `camera_facing_spec` object to pass to the `facing` parameter
#'   of `geom_text_3d()`.
#'
#' @details
#' With `mode = "plane"`, all text labels face the same direction (parallel
#' to the viewing plane), similar to traditional billboard rendering. This
#' produces clean, readable labels.
#'
#' With `mode = "point"`, each label faces directly toward the camera's
#' position in 3D space. Labels near the edges of the plot will angle inward,
#' which is geometrically correct but can look less clean.
#'
#' Since `geom_text_3d()` computes vertex positions before the view is known,
#' you must specify the view parameters here and use the same values in
#' `coord_3d()`.
#'
#' @examples
#' \donttest{
#' # Parallel billboard text (default) - all labels face same direction
#' df <- expand.grid(x = c("H", "B"), y = c("a", "o", "u"), z = c("g", "t"))
#' df$label <- paste0(df$x, df$y, df$z)
#'
#' ggplot(df, aes(x, y, z = z, label = label)) +
#'   geom_text_3d(method = "polygon",
#'                facing = camera_facing(pitch = 20, roll = -60, yaw = -30)) +
#'   coord_3d(pitch = 20, roll = -60, yaw = -30)
#'
#' # Point-facing text - labels angle toward camera position
#' ggplot(df, aes(x, y, z = z, label = label)) +
#'   geom_text_3d(method = "polygon",
#'                facing = camera_facing(pitch = 20, roll = -60, yaw = -30,
#'                                       mode = "point")) +
#'   coord_3d(pitch = 20, roll = -60, yaw = -30)
#' }
#'
#' @seealso [geom_text_3d()] for rendering 3D text labels
#'
#' @export
camera_facing <- function(coord = NULL,
                          pitch = 0, roll = -60, yaw = -30, dist = 2,
                          mode = c("plane", "point")) {
      mode <- match.arg(mode)

      if(!is.null(coord)){
            pitch <- coord[[1]]$pitch
            roll <- coord[[1]]$roll
            yaw <- coord[[1]]$yaw
            dist <- coord[[1]]$dist
      }

      # For plane mode, use infinite distance so all labels
      # face parallel to the view plane
      effective_dist <- if (mode == "plane") 1e6 else dist

      structure(
            list(pitch = pitch, roll = roll, yaw = yaw, dist = effective_dist),
            class = "camera_facing_spec"
      )
}


#' Compute per-point camera facing vectors
#'
#' @param anchors Data frame with x, y, z columns for anchor positions
#' @param spec A camera_facing_spec object
#' @param coord Optional coord_3d object for proper axis scaling
#' @return Matrix with one row per anchor, columns for facing x, y, z
#' @keywords internal
#' @noRd
compute_camera_facing_vectors <- function(anchors, spec, coord = NULL) {
      # Extract view parameters
      pitch <- spec$pitch
      roll <- spec$roll
      yaw <- spec$yaw
      dist <- spec$dist

      # Convert angles to radians
      pitch_rad <- pitch * pi / 180
      roll_rad <- roll * pi / 180
      yaw_rad <- yaw * pi / 180

      # Build rotation matrix (same as in rotate_3d / transform_3d_standard)
      rot <- matrix(c(
            cos(yaw_rad) * cos(pitch_rad),
            cos(yaw_rad) * sin(pitch_rad) * sin(roll_rad) - sin(yaw_rad) * cos(roll_rad),
            cos(yaw_rad) * sin(pitch_rad) * cos(roll_rad) + sin(yaw_rad) * sin(roll_rad),

            sin(yaw_rad) * cos(pitch_rad),
            sin(yaw_rad) * sin(pitch_rad) * sin(roll_rad) + cos(yaw_rad) * cos(roll_rad),
            sin(yaw_rad) * sin(pitch_rad) * cos(roll_rad) - cos(yaw_rad) * sin(roll_rad),

            -sin(pitch_rad),
            cos(pitch_rad) * sin(roll_rad),
            cos(pitch_rad) * cos(roll_rad)
      ), nrow = 3, byrow = TRUE)

      # Get data ranges
      x_range <- range(anchors$x, na.rm = TRUE)
      y_range <- range(anchors$y, na.rm = TRUE)
      z_range <- range(anchors$z, na.rm = TRUE)

      # Compute effective ratios based on coord settings
      if (!is.null(coord)) {
            ratio <- coord$ratio %||% c(1, 1, 1)
            scales_mode <- coord$scales %||% "free"

            if (scales_mode == "free") {
                  effective_ratios <- ratio
            } else {
                  # Fixed scales: ratios depend on data ranges
                  x_span <- diff(x_range)
                  y_span <- diff(y_range)
                  z_span <- diff(z_range)

                  if (x_span == 0 || is.na(x_span)) x_span <- 1
                  if (y_span == 0 || is.na(y_span)) y_span <- 1
                  if (z_span == 0 || is.na(z_span)) z_span <- 1

                  scale_proportional <- c(x_span, y_span, z_span)
                  effective_ratios <- ratio * scale_proportional
            }

            # Normalize
            effective_ratios <- effective_ratios / max(effective_ratios)
      } else {
            effective_ratios <- c(1, 1, 1)
      }

      # Scale anchor points to standard domain with effective ratios applied
      scale_to_standard <- function(vals, rng, eff_ratio) {
            width <- diff(rng)
            if (width == 0 || is.na(width)) {
                  rep(0, length(vals))
            } else {
                  ((vals - mean(rng)) / width) * eff_ratio
            }
      }

      std_x <- scale_to_standard(anchors$x, x_range, effective_ratios[1])
      std_y <- scale_to_standard(anchors$y, y_range, effective_ratios[2])
      std_z <- scale_to_standard(anchors$z, z_range, effective_ratios[3])

      # Apply z-flip (as done in transform_3d_standard)
      std_z <- -std_z

      # Apply rotation to get points in view space
      std_coords <- rbind(std_x, std_y, std_z)

      # rotate_3d does: rotated <- xyz %*% rot
      # So for column vectors we need: t(t(std_coords) %*% rot) = t(rot) %*% std_coords
      rotated <- t(rot) %*% std_coords

      # Camera is at (0, 0, -dist) in rotated space
      # Direction from each point toward camera
      cam_x <- 0 - rotated[1, ]
      cam_y <- 0 - rotated[2, ]
      cam_z <- (-dist) - rotated[3, ]

      # Normalize these direction vectors
      lengths <- sqrt(cam_x^2 + cam_y^2 + cam_z^2)
      cam_x <- cam_x / lengths
      cam_y <- cam_y / lengths
      cam_z <- cam_z / lengths

      # These directions are in rotated space - transform back to standard space
      rotated_dirs <- rbind(cam_x, cam_y, cam_z)
      std_dirs <- rot %*% rotated_dirs

      # Undo z-flip
      std_dirs[3, ] <- -std_dirs[3, ]

      # Transform from standard space back to data space
      # Need to undo the effective ratio scaling
      data_dirs <- std_dirs
      data_dirs[1, ] <- data_dirs[1, ] / effective_ratios[1]
      data_dirs[2, ] <- data_dirs[2, ] / effective_ratios[2]
      data_dirs[3, ] <- data_dirs[3, ] / effective_ratios[3]

      # Re-normalize after undoing the scaling
      lengths <- sqrt(data_dirs[1, ]^2 + data_dirs[2, ]^2 + data_dirs[3, ]^2)
      data_dirs[1, ] <- data_dirs[1, ] / lengths
      data_dirs[2, ] <- data_dirs[2, ] / lengths
      data_dirs[3, ] <- data_dirs[3, ] / lengths

      # Return as data frame for easier handling
      data.frame(
            facing_x = data_dirs[1, ],
            facing_y = data_dirs[2, ],
            facing_z = data_dirs[3, ]
      )
}


# Helper functions for 3D text rendering --------------------------------------

#' Convert facing specification to normal vector
#'
#' @param facing Character string ("xmin", "xmax", "ymin", "ymax", "zmin", "zmax")
#'   or numeric vector of length 3
#' @return Normalized vector of length 3
#' @keywords internal
#' @noRd
facing_to_normal <- function(facing) {
      if (is.numeric(facing)) {
            if (length(facing) != 3) {
                  stop("`facing` must be a length-3 numeric vector or a character string")
            }
            # Normalize
            return(facing / sqrt(sum(facing^2)))
      }

      if (!is.character(facing) || length(facing) != 1) {
            stop("`facing` must be a length-3 numeric vector or a character string")
      }

      switch(facing,
             xmin = c(-1, 0, 0),
             xmax = c(1, 0, 0),
             ymin = c(0, -1, 0),
             ymax = c(0, 1, 0),
             zmin = c(0, 0, -1),
             zmax = c(0, 0, 1),
             stop("`facing` must be one of: 'xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax', or a numeric vector")
      )
}


#' Compute text rotation matrix with intuitive "up" direction
#'
#' Creates a rotation matrix that orients text to face a given direction,
#' with angle=0 producing the most natural baseline orientation (horizontal
#' when possible).
#'
#' @param facing_normal Unit vector for direction text should face
#' @param angle Additional rotation around facing axis (degrees)
#' @return 3x3 rotation matrix
#' @keywords internal
#' @noRd
compute_text_rotation_matrix <- function(facing_normal, angle) {
      # We need to build an orthonormal basis where:
      # - The "forward" direction (original +z) maps to facing_normal
      # - The "up" direction (original +y) is as close to world +z as possible
      # - The "right" direction (original +x) completes the basis
      #
      # Text starts with:
      # - baseline along x-axis (right)
      # - ascenders along y-axis (up)
      # - face pointing along z-axis (forward)

      forward <- facing_normal

      # Choose world up hint: prefer +z, but if facing is parallel to z, use +y
      world_up <- c(0, 0, 1)
      if (abs(abs(sum(forward * world_up)) - 1) < 1e-10) {
            # Facing is parallel to z-axis, use y as up hint
            world_up <- c(0, 1, 0)
      }

      # Compute "right" as cross(world_up, forward), then normalize
      right <- c(
            world_up[2] * forward[3] - world_up[3] * forward[2],
            world_up[3] * forward[1] - world_up[1] * forward[3],
            world_up[1] * forward[2] - world_up[2] * forward[1]
      )
      right_len <- sqrt(sum(right^2))
      if (right_len < 1e-10) {
            # Fallback if cross product is degenerate
            right <- c(1, 0, 0)
      } else {
            right <- right / right_len
      }

      # Compute true "up" as cross(forward, right) to ensure orthogonality
      up <- c(
            forward[2] * right[3] - forward[3] * right[2],
            forward[3] * right[1] - forward[1] * right[3],
            forward[1] * right[2] - forward[2] * right[1]
      )
      # Should already be normalized, but just in case
      up <- up / sqrt(sum(up^2))

      # Build rotation matrix that maps:
      # (1,0,0) -> right
      # (0,1,0) -> up
      # (0,0,1) -> forward
      # This is just the matrix with right, up, forward as columns
      base_rotation <- cbind(right, up, forward)

      # Apply additional angle rotation around the facing axis
      if (angle != 0) {
            angle_rot <- rotation_matrix_around_axis(forward, angle)
            return(angle_rot %*% base_rotation)
      }

      base_rotation
}


#' Rotation matrix for rotation around an axis
#'
#' @param axis Unit vector defining rotation axis
#' @param angle Angle in degrees
#' @return 3x3 rotation matrix
#' @keywords internal
#' @noRd
rotation_matrix_around_axis <- function(axis, angle) {
      if (angle == 0) return(diag(3))

      theta <- angle * pi / 180
      c <- cos(theta)
      s <- sin(theta)
      t <- 1 - c

      x <- axis[1]
      y <- axis[2]
      z <- axis[3]

      matrix(c(
            t*x*x + c,   t*x*y - s*z, t*x*z + s*y,
            t*x*y + s*z, t*y*y + c,   t*y*z - s*x,
            t*x*z - s*y, t*y*z + s*x, t*z*z + c
      ), nrow = 3, byrow = TRUE)
}


#' Compute aspect ratio correction for text
#'
#' Estimates the aspect ratio correction needed so text doesn't appear stretched
#' when the coord_3d scales axes differently. Returns a scalar to multiply
#' the text's y-coordinates by.
#'
#' @param data_ranges List with x, y, z elements, each a length-2 vector of `c(min, max)`
#' @param coord A coord_3d object, or NULL
#' @return Scalar aspect ratio correction (1 = no correction needed)
#' @keywords internal
#' @noRd
compute_text_aspect_ratios <- function(data_ranges, coord) {
      if (is.null(coord)) {
            return(1)
      }

      # Extract coord parameters
      ratio <- coord$ratio %||% c(1, 1, 1)
      scales_mode <- coord$scales %||% "free"

      # Compute effective ratios similar to compute_effective_ratios in transform-3d.R
      if (scales_mode == "free") {
            # With free scales, each axis is normalized independently
            # The ratio parameter applies directly
            effective_ratios <- ratio
      } else {
            # With fixed scales, ratios depend on data ranges
            x_span <- diff(data_ranges$x)
            y_span <- diff(data_ranges$y)
            z_span <- diff(data_ranges$z)

            # Handle zero-width ranges
            if (x_span == 0 || is.na(x_span)) x_span <- 1
            if (y_span == 0 || is.na(y_span)) y_span <- 1
            if (z_span == 0 || is.na(z_span)) z_span <- 1

            # Combine user ratios with scale-space proportions
            scale_proportional <- c(x_span, y_span, z_span)
            effective_ratios <- ratio * scale_proportional
      }

      # Normalize so largest dimension is 1
      max_ratio <- max(effective_ratios)
      effective_ratios <- effective_ratios / max_ratio

      # For text aspect correction, we need to know how much the x and y axes
      # differ in their effective scaling. Return ratio of y to x.
      # If y is compressed relative to x, we need to stretch text vertically.
      # Note: This is approximate since text can face any direction in 3D.
      # For now, use x/y ratio as a reasonable default.
      if (effective_ratios[1] == 0) return(1)

      return(effective_ratios[2] / effective_ratios[1])
}


#' Rotate and position text vertices in 3D space
#'
#' Takes 2D text vertices (in x-y plane) and transforms them to face the
#' specified direction at the specified position. With angle=0, the text
#' baseline will be as horizontal as possible (parallel to x-y plane when
#' the facing direction allows).
#'
#' @param vertices Data frame with vertex_x, vertex_y columns (text outline vertices in points)
#' @param facing_normal Unit vector for the direction the text should face
#' @param angle Rotation angle around the facing axis (degrees)
#' @param anchor_x,anchor_y,anchor_z Position to place the text
#' @param size_mm Text size in mm (same units as geom_text)
#' @param data_span The span of the data in data units (used to compute conversion)
#' @param aspect_adjust Scalar multiplier for text y-coordinates (height).
#'   Values > 1 make text taller, values < 1 make text wider.
#' @return Data frame with x, y, z columns in data coordinates
#' @keywords internal
#' @noRd
rotate_text_to_facing <- function(vertices, facing_normal, angle,
                                  anchor_x, anchor_y, anchor_z,
                                  size_mm, data_span, aspect_adjust = 1) {
      # Convert text vertices from points to data units
      #
      # Goal: text specified at size_mm should appear roughly that size on a

      # typical 6" plot, similar to billboard text.
      #
      # For billboard: 3.88mm text on 152.4mm plot ≈ 2.5% of plot height
      # For polygon: we want text to be ~2.5% of the data span
      #
      # text_outlines() returns vertices in points. For size_pt point text,
      # the character height is roughly size_pt points.
      #
      # We want: (text height in data units) / data_span = size_mm / 152.4
      # So: text height in data units = size_mm * data_span / 152.4
      #
      # If vertices are at size_pt points, and we want height = size_mm * data_span / 152.4,
      # then scale factor = (size_mm * data_span / 152.4) / size_pt
      #
      # But size_pt = size_mm * 72.27 / 25.4 (mm to points conversion)
      # So scale factor = (size_mm * data_span / 152.4) / (size_mm * 72.27 / 25.4)
      #                 = (data_span / 152.4) * (25.4 / 72.27)
      #                 = data_span / 152.4 / 2.845
      #                 = data_span / 433.5
      #
      # Hmm, that's independent of size_mm, which is wrong.
      #
      # Let's think differently: vertices come in at size_pt points.
      # We want the final text height to be (size_mm / 152.4) * data_span data units.
      # Current vertex height is roughly size_pt points.
      # So we need to scale by: target_height / current_height
      #   = (size_mm / 152.4 * data_span) / size_pt
      #
      # Since size_pt = size_mm * 2.845:
      #   = (size_mm / 152.4 * data_span) / (size_mm * 2.845)
      #   = data_span / (152.4 * 2.845)
      #   = data_span / 433.5
      #
      # This is still size-independent! That's because larger size_mm gives
      # proportionally larger size_pt, and we're scaling both.
      #
      # The issue is that text_outlines already scaled the vertices.
      # Let's just use a simple empirical scale factor.

      # Simple approach: make text height roughly 10% of data_span at default size
      # Default size is 3.88mm. Scale proportionally from there.
      reference_size <- 3.88
      base_fraction <- 0.10  # text is 10% of data span at reference size

      # Target text height in data units
      target_height <- (size_mm / reference_size) * base_fraction * data_span

      # Vertices from text_outlines are in points. Approximate height is size_pt.
      size_pt <- size_mm * 72.27 / 25.4

      # Scale factor to convert from points to data units
      scale_factor <- target_height / size_pt

      vx <- vertices$vertex_x * scale_factor
      vy <- vertices$vertex_y * scale_factor * aspect_adjust
      vz <- rep(0, length(vx))

      # Get combined rotation matrix (handles both facing and angle)
      rotation <- compute_text_rotation_matrix(facing_normal, angle)

      # Apply rotation
      coords <- rbind(vx, vy, vz)
      rotated <- rotation %*% coords

      # Translate to anchor position
      data.frame(
            x = rotated[1, ] + anchor_x,
            y = rotated[2, ] + anchor_y,
            z = rotated[3, ] + anchor_z
      )
}

# Geom (Polygon method) -------------------------------------------------------

GeomText3DPolygon <- ggproto("GeomText3DPolygon", GeomPolygon3D,
                             default_aes = aes(
                                   fill = "black",
                                   colour = NA,
                                   linewidth = 0,
                                   linetype = 1,
                                   alpha = 1
                             )
)

# Stat (Polygon method) -------------------------------------------------------

StatText3DPolygon <- ggproto("StatText3DPolygon", Stat,
                             required_aes = c("x", "y", "z", "label"),

                             compute_panel = function(data, scales, na.rm = FALSE,
                                                      family = "sans",
                                                      weight = "normal",
                                                      italic = FALSE,
                                                      size = 3.88,
                                                      hjust = 0.5,
                                                      vjust = 0.5,
                                                      lineheight = 1.2,
                                                      spacing = 0,
                                                      tolerance = 0.01,
                                                      facing = "zmax",
                                                      angle = 0,
                                                      nudge_x = 0,
                                                      nudge_y = 0,
                                                      nudge_z = 0,
                                                      coord = NULL,
                                                      aspect_adjust = 1,
                                                      light = NULL) {

                                   # Remove missing values if requested
                                   if (na.rm) {
                                         data <- data[complete.cases(data[c("x", "y", "z", "label")]), ]
                                   }

                                   if (nrow(data) == 0) {
                                         return(data.frame())
                                   }

                                   # Handle discrete scale conversion (preserve raw values for scale training)
                                   if ("x" %in% names(data)) {
                                         data$x_raw <- data$x
                                         if (is.factor(data$x) || is.character(data$x)) {
                                               data$x <- as.numeric(as.factor(data$x))
                                         }
                                   }
                                   if ("y" %in% names(data)) {
                                         data$y_raw <- data$y
                                         if (is.factor(data$y) || is.character(data$y)) {
                                               data$y <- as.numeric(as.factor(data$y))
                                         }
                                   }
                                   if ("z" %in% names(data)) {
                                         data$z_raw <- data$z
                                         if (is.factor(data$z) || is.character(data$z)) {
                                               data$z <- as.numeric(as.factor(data$z))
                                         }
                                   }

                                   # Compute anchor positions (with nudge applied)
                                   anchors <- data.frame(
                                         x = data$x + nudge_x,
                                         y = data$y + nudge_y,
                                         z = data$z + nudge_z
                                   )

                                   # Handle facing = "camera" when coord is provided
                                   if (is.character(facing) && facing == "camera") {
                                         if (is.null(coord)) {
                                               stop('facing = "camera" requires the `coord` parameter to be set')
                                         }
                                         # Create camera_facing_spec from coord params
                                         facing <- camera_facing(
                                               pitch = coord$pitch %||% 0,
                                               roll = coord$roll %||% -60,
                                               yaw = coord$yaw %||% -30,
                                               dist = coord$dist %||% 2
                                         )
                                   }

                                   # Handle camera_facing_spec: compute per-row facing vectors
                                   if (inherits(facing, "camera_facing_spec")) {
                                         facing_vectors <- compute_camera_facing_vectors(anchors, facing, coord)
                                         per_row_facing <- TRUE
                                   } else {
                                         # Fixed facing for all rows
                                         facing_normal <- facing_to_normal(facing)
                                         per_row_facing <- FALSE
                                   }

                                   # Compute aspect adjustment for text correction
                                   # Start with user-specified adjustment
                                   final_aspect_adjust <- aspect_adjust

                                   # If coord is provided, compute additional correction from coord params
                                   if (!is.null(coord)) {
                                         coord_aspect <- compute_text_aspect_ratios(
                                               data_ranges = list(
                                                     x = range(data$x, na.rm = TRUE),
                                                     y = range(data$y, na.rm = TRUE),
                                                     z = range(data$z, na.rm = TRUE)
                                               ),
                                               coord = coord
                                         )
                                         # Multiply user adjustment by coord-based correction
                                         final_aspect_adjust <- aspect_adjust * coord_aspect
                                   }

                                   # Compute data_span for size conversion
                                   # Use the maximum span across all axes as reference
                                   x_span <- diff(range(data$x, na.rm = TRUE))
                                   y_span <- diff(range(data$y, na.rm = TRUE))
                                   z_span <- diff(range(data$z, na.rm = TRUE))
                                   # Handle zero spans (single point)
                                   if (x_span == 0 || is.na(x_span)) x_span <- 1
                                   if (y_span == 0 || is.na(y_span)) y_span <- 1
                                   if (z_span == 0 || is.na(z_span)) z_span <- 1
                                   data_span <- max(x_span, y_span, z_span)

                                   # Process each label
                                   all_text <- lapply(seq_len(nrow(data)), function(i) {
                                         row <- data[i, ]
                                         label <- as.character(row$label)

                                         # Skip empty labels
                                         if (is.na(label) || label == "") {
                                               return(NULL)
                                         }

                                         # Get text outlines
                                         # Convert size from mm to points for text_outlines
                                         # 1 mm ≈ 2.835 points (72.27 pt/inch / 25.4 mm/inch)
                                         size_pt <- size * 72.27 / 25.4

                                         outlines <- text_outlines(
                                               text = label,
                                               family = family,
                                               weight = weight,
                                               italic = italic,
                                               size = size_pt,
                                               tolerance = tolerance,
                                               spacing = spacing,
                                               hjust = hjust,
                                               vjust = vjust,
                                               lineheight = lineheight
                                         )

                                         if (is.null(outlines) || nrow(outlines) == 0) {
                                               return(NULL)
                                         }

                                         # Add text-specific identifiers
                                         outlines$text_id <- i

                                         # Store vertex positions for rotation (already at correct size in points)
                                         outlines$vertex_x <- outlines$x
                                         outlines$vertex_y <- outlines$y

                                         # subgroup identifies contours within a text label (for holes)
                                         outlines$.subgroup <- outlines$poly_id

                                         # Get anchor position (with nudge already applied)
                                         anchor_x <- anchors$x[i]
                                         anchor_y <- anchors$y[i]
                                         anchor_z <- anchors$z[i]

                                         # Get facing normal for this row
                                         if (per_row_facing) {
                                               facing_normal <- c(
                                                     facing_vectors$facing_x[i],
                                                     facing_vectors$facing_y[i],
                                                     facing_vectors$facing_z[i]
                                               )
                                         }

                                         # Rotate and position vertices in data space
                                         transformed <- rotate_text_to_facing(
                                               vertices = outlines,
                                               facing_normal = facing_normal,
                                               angle = angle,
                                               anchor_x = anchor_x,
                                               anchor_y = anchor_y,
                                               anchor_z = anchor_z,
                                               size_mm = size,
                                               data_span = data_span,
                                               aspect_adjust = final_aspect_adjust
                                         )

                                         # Store final positions as x, y, z for scale training
                                         outlines$x <- transformed$x
                                         outlines$y <- transformed$y
                                         outlines$z <- transformed$z

                                         # Store normal vectors for lighting (same for all vertices in this label)
                                         outlines$normal_x <- facing_normal[1]
                                         outlines$normal_y <- facing_normal[2]
                                         outlines$normal_z <- facing_normal[3]

                                         # Carry over aesthetics from original row
                                         for (col in setdiff(names(row), c("x", "y", "z", "label"))) {
                                               outlines[[col]] <- row[[col]]
                                         }

                                         outlines
                                   })

                                   result <- dplyr::bind_rows(all_text)

                                   if (nrow(result) == 0) {
                                         return(data.frame())
                                   }

                                   # Create hierarchical group for depth sorting
                                   # Group identifies each text label; subgroup identifies contours within
                                   if (!"group" %in% names(result) || all(result$group == -1)) {
                                         result$group <- paste0("1__text", result$text_id)
                                   } else {
                                         result$group <- paste0(result$group, "__text", result$text_id)
                                   }

                                   # Remove intermediate columns not needed downstream
                                   result$contour <- NULL
                                   result$glyph <- NULL
                                   result$letter <- NULL
                                   result$poly_id <- NULL
                                   result$vertex_x <- NULL
                                   result$vertex_y <- NULL

                                   # Attach lighting specification for downstream processing
                                   result <- attach_light(result, light)

                                   result
                             }
)



# Stat (Billboard method) -----------------------------------------------------

StatText3DBillboard <- ggproto("StatText3DBillboard", Stat,
                               required_aes = c("x", "y", "z", "label"),

                               default_aes = aes(
                                     colour = "black",
                                     size = 3.88,
                                     alpha = NA,
                                     family = "",
                                     fontface = 1,
                                     hjust = 0.5,
                                     vjust = 0.5,
                                     angle = 0,
                                     lineheight = 1.2
                               ),

                               compute_panel = function(data, scales, na.rm = FALSE,
                                                        nudge_x = 0,
                                                        nudge_y = 0,
                                                        nudge_z = 0,
                                                        size = 3.88,
                                                        hjust = 0.5,
                                                        vjust = 0.5,
                                                        angle = 0,
                                                        lineheight = 1.2,
                                                        family = "",
                                                        fontface = 1) {

                                     # Remove missing values if requested
                                     if (na.rm) {
                                           data <- data[complete.cases(data[c("x", "y", "z", "label")]), ]
                                     }

                                     if (nrow(data) == 0) {
                                           return(data.frame())
                                     }

                                     # Handle discrete scale conversion (preserve raw values for scale training)
                                     if ("x" %in% names(data)) {
                                           data$x_raw <- data$x
                                           if (is.factor(data$x) || is.character(data$x)) {
                                                 data$x <- as.numeric(as.factor(data$x))
                                           }
                                     }
                                     if ("y" %in% names(data)) {
                                           data$y_raw <- data$y
                                           if (is.factor(data$y) || is.character(data$y)) {
                                                 data$y <- as.numeric(as.factor(data$y))
                                           }
                                     }
                                     if ("z" %in% names(data)) {
                                           data$z_raw <- data$z
                                           if (is.factor(data$z) || is.character(data$z)) {
                                                 data$z <- as.numeric(as.factor(data$z))
                                           }
                                     }

                                     # Apply nudge
                                     data$x <- data$x + nudge_x
                                     data$y <- data$y + nudge_y
                                     data$z <- data$z + nudge_z

                                     # Set defaults for text parameters if not in data
                                     if (!"size" %in% names(data)) data$size <- size
                                     if (!"hjust" %in% names(data)) data$hjust <- hjust
                                     if (!"vjust" %in% names(data)) data$vjust <- vjust
                                     if (!"angle" %in% names(data)) data$angle <- angle
                                     if (!"lineheight" %in% names(data)) data$lineheight <- lineheight
                                     if (!"family" %in% names(data)) data$family <- family
                                     if (!"fontface" %in% names(data)) data$fontface <- fontface

                                     data
                               }
)


# Geom (Billboard method) -----------------------------------------------------

GeomText3DBillboard <- ggproto("GeomText3DBillboard", Geom,
                               required_aes = c("x", "y", "z", "label"),

                               default_aes = aes(
                                     colour = "black",
                                     size = 3.88,
                                     alpha = NA,
                                     family = "",
                                     fontface = 1,
                                     hjust = 0.5,
                                     vjust = 0.5,
                                     angle = 0,
                                     lineheight = 1.2
                               ),

                               draw_panel = function(data, panel_params, coord,
                                                     scale_depth = TRUE,
                                                     na.rm = FALSE) {

                                     # Validate coord
                                     validate_coord3d(coord)

                                     if (nrow(data) == 0) {
                                           return(grid::nullGrob())
                                     }

                                     # Transform through coord (3D -> 2D projection)
                                     coords <- coord$transform(data, panel_params)

                                     # Scale sizes by depth if enabled
                                     text_strength <- resolve_layer_scale_depth(scale_depth)
                                     if (text_strength > 0 && "depth_scale" %in% names(coords)) {
                                           coords$size <- coords$size *
                                                 apply_depth_strength(coords$depth_scale, text_strength)
                                     }

                                     # Convert size from mm to grid fontsize
                                     # ggplot2 size aesthetic is in mm, we need to convert appropriately
                                     fontsize_pts <- coords$size * .pt

                                     # Sort by depth (back to front) for proper rendering order
                                     if ("depth" %in% names(coords)) {
                                           coords <- coords[order(coords$depth, decreasing = TRUE), ]
                                     }

                                     # Create text grobs for each label
                                     grobs <- lapply(seq_len(nrow(coords)), function(i) {
                                           grid::textGrob(
                                                 label = coords$label[i],
                                                 x = coords$x[i],
                                                 y = coords$y[i],
                                                 default.units = "npc",
                                                 hjust = coords$hjust[i],
                                                 vjust = coords$vjust[i],
                                                 rot = coords$angle[i],
                                                 gp = grid::gpar(
                                                       col = alpha(coords$colour[i], coords$alpha[i]),
                                                       fontsize = fontsize_pts[i],
                                                       fontfamily = coords$family[i],
                                                       fontface = coords$fontface[i],
                                                       lineheight = coords$lineheight[i]
                                                 )
                                           )
                                     })

                                     do.call(grid::grobTree, grobs)
                               },

                               draw_key = ggplot2::draw_key_text
)


# User-facing functions -------------------------------------------------------

#' 3D text labels
#'
#' `geom_text_3d()` renders text labels in 3D space. Two rendering methods are
#' available: `"billboard"` (default) renders text as flat labels that always
#' face the camera, while `"polygon"` renders text as filled polygons that can
#' be oriented in any direction.
#'
#' @section Methods:
#' \itemize{
#'   \item `"billboard"` (default): Fast, simple text rendering using native
#'     fonts. Text is always parallel to the view plane (like a billboard).
#'     Only uses the `colour` aesthetic for text color.
#'   \item `"polygon"`: Renders text as filled polygon outlines. Text can be
#'     oriented to face any direction using the `facing` parameter. Uses both
#'     `fill` (text fill) and `colour` (text outline) aesthetics.
#' }
#'
#' @section Orientation (polygon method only):
#' The `facing` parameter controls which direction the text faces:
#' \itemize{
#'   \item `"zmax"` (default): Text faces upward (readable from above)
#'   \item `"xmin"`, `"xmax"`, `"ymin"`, `"ymax"`, `"zmin"`: Text faces
#'     toward the specified cube face
#'   \item A numeric vector `c(x, y, z)`: Text faces the specified direction
#'   \item `camera_facing(...)`: Text faces the camera
#' }
#'
#' The `angle` parameter rotates the text around its facing axis (in degrees).
#' With `angle = 0`, the text baseline is oriented as horizontally as possible
#' (parallel to the x-y plane when the facing direction allows).
#'
#' @section Camera-facing text:
#' For billboard method, text automatically faces the camera - no configuration
#' needed.
#'
#' For polygon method, there are two ways to make text face the camera:
#'
#' 1. Use `facing = "camera"` with the `coord` parameter (recommended):
#' ```
#' my_coord <- coord_3d(pitch = 20, roll = -60, yaw = -30)
#' ggplot(df, aes(x, y, z = z, label = label)) +
#'   geom_text_3d(method = "polygon", facing = "camera", coord = my_coord) +
#'   my_coord
#' ```
#'
#' 2. Pass `camera_facing()` directly to `facing`:
#' ```
#' ggplot(df, aes(x, y, z = z, label = label)) +
#'   geom_text_3d(method = "polygon",
#'                facing = camera_facing(pitch = 20, roll = -60, yaw = -30)) +
#'   coord_3d(pitch = 20, roll = -60, yaw = -30)
#' ```
#'
#' @section Aspect Ratio Correction:
#' For polygon method, text can appear stretched if the x, y, and z axes have
#' different data ranges. Providing a `coord_3d()` object via the `coord`
#' parameter enables automatic aspect ratio correction based on the coord's
#' `ratio` and `scales` settings.
#'
#' If other layers in your plot have different data ranges that affect the
#' final scale limits, you can fine-tune the correction with `aspect_adjust`:
#' - `aspect_adjust = 1` (default): Use the computed correction
#' - `aspect_adjust > 1`: Make text taller
#' - `aspect_adjust < 1`: Make text wider
#'
#' @section Size and Scaling:
#' Text size is specified in mm via the `size` parameter, matching the units
#' used by [ggplot2::geom_text()]. The default size of 3.88 corresponds to
#' approximately 11pt text.
#'
#' Both methods use the same size units, so switching between `"billboard"`
#' and `"polygon"` methods with the same `size` value will produce similarly
#' sized text (assuming a typical 6 inch plot).
#'
#' When depth scaling is enabled (the default), text farther from the viewer
#' appears smaller, creating a natural perspective effect. For billboard
#' method, this scales the font size. For polygon method, this scales the
#' outline linewidth (the polygon fill is always depth-scaled by the 3D
#' projection).
#'
#' @param mapping Set of aesthetic mappings. See [ggplot2::aes()].
#' @param data A data frame containing the variables to plot.
#' @param position Position adjustment. Defaults to `"identity"`.
#' @param ... Other arguments passed to the layer.
#' @param method Rendering method: `"billboard"` (default) for flat camera-facing
#'   text, or `"polygon"` for 3D-oriented filled text.
#' @param facing (Polygon method only) Direction the text should face. One of
#'   `"xmin"`, `"xmax"`, `"ymin"`, `"ymax"`, `"zmin"`, `"zmax"` (default),
#'   `"camera"`, a numeric vector of length 3, or a `camera_facing()` specification.
#'   Use `"camera"` with the `coord` parameter to automatically face the camera.
#'   Ignored with a warning for billboard method.
#' @param coord (Polygon method only) A `coord_3d()` object. When provided:
#'   (1) if `facing = "camera"`, extracts rotation parameters automatically;
#'   (2) uses the coord's `ratio` and `scales` parameters to correct text aspect
#'   ratio so text doesn't appear stretched. Ignored for billboard method.
#' @param aspect_adjust (Polygon method only) Scalar multiplier for text height
#'   relative to width. Default is 1 (normal proportions). Values > 1 make text
#'   taller, values < 1 make text wider. When `coord` is provided, this is
#'   multiplied by the computed aspect correction. Ignored for billboard method.
#' @param light (Polygon method only) Either a lighting specification created by
#'   [light()], `"none"` (the default) or `light("none")` for no lighting, or
#'   `NULL` to inherit plot-level lighting. When provided, text polygons are
#'   shaded based on their facing direction. Ignored for billboard method.
#' @param angle Rotation angle in degrees. For polygon method, rotates around
#'   the facing axis. For billboard method, rotates in the view plane.
#' @param nudge_x,nudge_y,nudge_z Offset to apply to text position in data units.
#' @param family Font family.
#' @param fontface Font face (for billboard method): "plain", "bold", "italic",
#'   or "bold.italic".
#' @param weight (Polygon method only) Font weight: "normal", "bold", "thin",
#'   "light", etc.
#' @param italic (Polygon method only) Logical; use italic font variant.
#' @param size Font size in mm (same units as [ggplot2::geom_text()]). Default
#'   is 3.88, which corresponds to approximately 11pt text.
#' @param hjust,vjust Horizontal and vertical justification of text (0-1).
#' @param lineheight Line height multiplier for multi-line text.
#' @param spacing (Polygon method only) Additional letter spacing in em units.
#' @param tolerance (Polygon method only) Bezier curve tolerance for text
#'   outlines. Lower values give smoother curves but more vertices.
#' @inheritParams depth_params
#' @param na.rm If TRUE, silently remove missing values.
#' @param show.legend Logical. Should this layer be included in the legends?
#' @param inherit.aes If TRUE, inherit aesthetics from the plot's default.
#'
#' @section Aesthetics:
#' `geom_text_3d()` understands the following aesthetics (required in bold):
#'
#' **Both methods:**
#' \itemize{
#'   \item \strong{x}, \strong{y}, \strong{z}, \strong{label}
#'   \item `alpha`
#' }
#'
#' **Billboard method only:**
#' \itemize{
#'   \item `colour` - text color
#' }
#'
#' **Polygon method only:**
#' \itemize{
#'   \item `fill` - text fill color
#'   \item `colour` - text outline color
#'   \item `linewidth` - outline thickness
#'   \item `linetype` - outline line type
#'   \item `group`
#' }
#'
#' Other text properties (`size`, `hjust`, `vjust`, `angle`, `family`,
#' `fontface`, `lineheight`, and for the polygon method `weight`, `italic`,
#' `spacing`, `tolerance`) are currently accepted only as fixed values passed
#' as arguments to `geom_text_3d()`, not as mappable aesthetics in `aes()`.
#'
#' @return A ggplot2 layer
#'
#' @examples
#' df <- expand.grid(x = c("B", "H", "N"), y = c("a", "o", "u"), z = c("g", "t"))
#' df$label <- paste0(df$x, df$y, df$z)
#'
#' # Billboard text (default) - automatically faces camera
#' ggplot(df, aes(x, y, z, label = label)) +
#'   geom_text_3d() +
#'   coord_3d(scales = "fixed")
#'
#' # Polygon text - can face any direction
#' ggplot(df, aes(x, y, z, label = label, fill = z)) +
#'   geom_text_3d(method = "polygon", facing = "zmax") +
#'   coord_3d(scales = "fixed")
#'
#' # Polygon text facing camera (using coord parameter)
#' my_coord <- coord_3d(pitch = 0, roll = -60, yaw = -30, scales = "fixed")
#' ggplot(df, aes(x, y, z, label = label)) +
#'   geom_text_3d(method = "polygon", facing = "camera", coord = my_coord,
#'                color = "red", linewidth = .1) +
#'   my_coord
#'
#' # Text with styling
#' ggplot(df, aes(x, y, z, label = label, fill = factor(x))) +
#'   geom_text_3d(method = "polygon", facing = "xmin",
#'                family = "serif", weight = "bold", italic = TRUE,
#'                size = 6, aspect_adjust = 2) +
#'   coord_3d(scales = "fixed")
#'
#' @seealso [camera_facing()] for camera-facing specifications, [text_outlines()]
#'   for text-to-polygon conversion
#' @export
geom_text_3d <- function(mapping = NULL, data = NULL,
                         position = "identity",
                         ...,
                         method = c("billboard", "polygon"),
                         # Polygon parameters
                         facing = "zmax",
                         coord = NULL,
                         aspect_adjust = 1,
                         light = "none",
                         # Shared parameters
                         angle = 0,
                         nudge_x = 0,
                         nudge_y = 0,
                         nudge_z = 0,
                         size = 3.88,
                         hjust = 0.5,
                         vjust = 0.5,
                         lineheight = 1.2,
                         scale_depth = TRUE,
                         # Polygon-only parameters
                         family = "sans",
                         weight = "normal",
                         italic = FALSE,
                         fontface = "plain",
                         spacing = 0,
                         tolerance = 0.01,
                         # Common layer parameters
                         na.rm = FALSE,
                         show.legend = NA,
                         inherit.aes = TRUE) {

      method <- match.arg(method)

      if (method == "billboard") {
            # Warn if polygon-specific parameters are non-default
            if (!missing(facing) && !identical(facing, "zmax")) {
                  warning("`facing` is ignored for billboard method. ",
                          "Billboard text automatically faces the camera.", call. = FALSE)
            }
            if (!is.null(coord)) {
                  warning("`coord` is ignored for billboard method.", call. = FALSE)
            }
            if (aspect_adjust != 1) {
                  warning("`aspect_adjust` is ignored for billboard method.", call. = FALSE)
            }
            if (!missing(light) && !identical(light, "none")) {
                  warning("`light` is ignored for billboard method.", call. = FALSE)
            }

            # Convert fontface string to numeric if needed
            fontface_num <- switch(as.character(fontface),
                                   "plain" = 1, "bold" = 2,
                                   "italic" = 3, "bold.italic" = 4,
                                   as.numeric(fontface))

            layer(
                  data = data,
                  mapping = mapping,
                  stat = StatText3DBillboard,
                  geom = GeomText3DBillboard,
                  position = position,
                  show.legend = show.legend,
                  inherit.aes = inherit.aes,
                  params = list(
                        nudge_x = nudge_x,
                        nudge_y = nudge_y,
                        nudge_z = nudge_z,
                        size = size,
                        hjust = hjust,
                        vjust = vjust,
                        angle = angle,
                        lineheight = lineheight,
                        family = family,
                        fontface = fontface_num,
                        scale_depth = scale_depth,
                        na.rm = na.rm,
                        ...
                  )
            )
      } else {
            # Polygon method
            layer(
                  data = data,
                  mapping = mapping,
                  stat = StatText3DPolygon,
                  geom = GeomText3DPolygon,
                  position = position,
                  show.legend = show.legend,
                  inherit.aes = inherit.aes,
                  params = list(
                        facing = facing,
                        coord = coord,
                        aspect_adjust = aspect_adjust,
                        light = light,
                        angle = angle,
                        nudge_x = nudge_x,
                        nudge_y = nudge_y,
                        nudge_z = nudge_z,
                        family = family,
                        weight = weight,
                        italic = italic,
                        size = size,
                        hjust = hjust,
                        vjust = vjust,
                        lineheight = lineheight,
                        spacing = spacing,
                        tolerance = tolerance,
                        scale_depth = scale_depth,
                        sort_method = "pairwise",
                        force_convex = FALSE,
                        na.rm = na.rm,
                        ...
                  )
            )
      }
}


#' @rdname geom_text_3d
#' @param geom The geometric object to use. Defaults based on method.
#' @export
stat_text_3d <- function(mapping = NULL, data = NULL,
                         geom = NULL, position = "identity",
                         ...,
                         method = c("billboard", "polygon"),
                         # Polygon parameters
                         facing = "zmax",
                         coord = NULL,
                         aspect_adjust = 1,
                         light = "none",
                         # Shared parameters
                         angle = 0,
                         nudge_x = 0,
                         nudge_y = 0,
                         nudge_z = 0,
                         size = 3.88,
                         hjust = 0.5,
                         vjust = 0.5,
                         lineheight = 1.2,
                         scale_depth = TRUE,
                         # Polygon-only parameters
                         family = "sans",
                         weight = "normal",
                         italic = FALSE,
                         fontface = "plain",
                         spacing = 0,
                         tolerance = 0.01,
                         # Common layer parameters
                         na.rm = FALSE,
                         show.legend = NA,
                         inherit.aes = TRUE) {

      method <- match.arg(method)

      if (method == "billboard") {
            if (!missing(facing) && !identical(facing, "zmax")) {
                  warning("`facing` is ignored for billboard method. ",
                          "Billboard text automatically faces the camera.", call. = FALSE)
            }
            if (!is.null(coord)) {
                  warning("`coord` is ignored for billboard method.", call. = FALSE)
            }
            if (aspect_adjust != 1) {
                  warning("`aspect_adjust` is ignored for billboard method.", call. = FALSE)
            }
            if (!missing(light) && !identical(light, "none")) {
                  warning("`light` is ignored for billboard method.", call. = FALSE)
            }

            fontface_num <- switch(as.character(fontface),
                                   "plain" = 1, "bold" = 2,
                                   "italic" = 3, "bold.italic" = 4,
                                   as.numeric(fontface))

            geom <- geom %||% GeomText3DBillboard

            layer(
                  data = data,
                  mapping = mapping,
                  stat = StatText3DBillboard,
                  geom = get_proto(geom, "geom"),
                  position = position,
                  show.legend = show.legend,
                  inherit.aes = inherit.aes,
                  params = list(
                        nudge_x = nudge_x,
                        nudge_y = nudge_y,
                        nudge_z = nudge_z,
                        size = size,
                        hjust = hjust,
                        vjust = vjust,
                        angle = angle,
                        lineheight = lineheight,
                        family = family,
                        fontface = fontface_num,
                        scale_depth = scale_depth,
                        na.rm = na.rm,
                        ...
                  )
            )
      } else {
            geom <- geom %||% GeomText3DPolygon

            layer(
                  data = data,
                  mapping = mapping,
                  stat = StatText3DPolygon,
                  geom = get_proto(geom, "geom"),
                  position = position,
                  show.legend = show.legend,
                  inherit.aes = inherit.aes,
                  params = list(
                        facing = facing,
                        coord = coord,
                        aspect_adjust = aspect_adjust,
                        light = light,
                        angle = angle,
                        nudge_x = nudge_x,
                        nudge_y = nudge_y,
                        nudge_z = nudge_z,
                        family = family,
                        weight = weight,
                        italic = italic,
                        size = size,
                        hjust = hjust,
                        vjust = vjust,
                        lineheight = lineheight,
                        spacing = spacing,
                        tolerance = tolerance,
                        scale_depth = scale_depth,
                        sort_method = "pairwise",
                        force_convex = FALSE,
                        na.rm = na.rm,
                        ...
                  )
            )
      }
}

# Depth sorting for 3D rendering
#
# Implements back-to-front ordering of primitives for painter's algorithm
# rendering. Supports mixed primitive types (polygon, point, segment, text)
# identified by the `.prim` column.
#
# Two sorting modes:
#   - "painter": Fast approximate sort by mean depth per group, with
#     hierarchical support for `__`-separated group names.
#   - "pairwise": Accurate O(k^2) sort that compares every pair of groups
#     using type-specific geometric tests to determine occlusion order.
#     Segments that interpenetrate polygons in 3D are automatically split
#     at the crossing point before sorting.
#
# Pairwise comparison coverage:
#   polygon-polygon: SAT overlap test, depth interpolation at intersection centroid
#   polygon-point:   point-in-polygon test, depth interpolation at point xy
#   polygon-segment: polyclip intersection, depth interpolation at clipped midpoint
#   segment-segment: line intersection, depth interpolation at crossing point
#   point-segment:   bbox proximity test, depth at closest point on segment
#   point-point:     no constraint (painter fallback, zero-area overlap)
#
# When primitives are coplanar (depth delta ~ 0), a tiebreaker renders
# smaller primitives on top: point > segment > polygon.

interpolate_z <- function(train, new, beta = NULL) {
      if (!is.null(beta)) {
            return(beta[1] + beta[2] * new[1] + beta[3] * new[2])
      }
      train <- as.matrix(train)
      if(nrow(unique(train)) >= 3) {
            x <- cbind(1, as.matrix(train[,1:2]))
            beta <- tryCatch(
                  solve(crossprod(x), crossprod(x, train[,3])),
                  error = function(e) NULL
            )
            if (!is.null(beta)) {
                  return(beta[1] + beta[2] * new[1] + beta[3] * new[2])
            }
      }
      # Fallback: inverse-distance-weighted interpolation
      # for degenerate polygons (e.g. collinear vertices from face projection,
      # zero-height pillar faces)
      weights <- 1 / sqrt((train[,1] - new[1])^2 + (train[,2] - new[2])^2)
      z <- weighted.mean(train[,3], weights)
      return(z)
}

# Quick check for definite cycles
definitely_has_cycles <- function(adj_logical) {
      in_degrees <- colSums(adj_logical, na.rm = TRUE)
      return(all(in_degrees > 0))  # Every node has incoming edges
}

# Test if graph is acyclic using topological sort
is_acyclic <- function(adj_logical) {
      n <- nrow(adj_logical)
      in_degree <- colSums(adj_logical, na.rm = TRUE)
      queue <- which(in_degree == 0)
      processed <- 0

      while (length(queue) > 0) {
            current <- queue[1]
            queue <- queue[-1]
            processed <- processed + 1

            # Remove edges from current node
            neighbors <- which(adj_logical[current, ])
            for (neighbor in neighbors) {
                  in_degree[neighbor] <- in_degree[neighbor] - 1
                  if (in_degree[neighbor] == 0) {
                        queue <- c(queue, neighbor)
                  }
            }
      }

      return(processed == n)
}

# Break cycles by removing weakest constraints
break_cycles <- function(delta_matrix) {
      max_iterations <- nrow(delta_matrix)^2  # Safety limit
      iteration <- 0

      while (iteration < max_iterations) {
            iteration <- iteration + 1

            # Convert to logical adjacency matrix
            adj_logical <- (delta_matrix > 0) & !is.na(delta_matrix)

            # Quick cycle check first
            if (!definitely_has_cycles(adj_logical)) {
                  # Still might have cycles, need full check
                  if (is_acyclic(adj_logical)) break
            }

            # Find weakest constraint to remove
            has_constraint <- !is.na(delta_matrix)
            if (!any(has_constraint)) {
                  warning("No constraints left but graph still has cycles")
                  break
            }

            abs_deltas <- abs(delta_matrix)
            abs_deltas[!has_constraint] <- Inf  # Ignore NA entries

            min_delta <- min(abs_deltas, na.rm = TRUE)
            if (!is.finite(min_delta)) break

            weakest <- which(abs_deltas == min_delta, arr.ind = TRUE)[1, ]

            # Remove this constraint (set back to NA)
            delta_matrix[weakest[1], weakest[2]] <- NA
            delta_matrix[weakest[2], weakest[1]] <- NA
      }

      if (iteration >= max_iterations) {
            warning("Cycle breaking exceeded maximum iterations")
      }

      return(delta_matrix)
}


topological_sort <- function(adj_matrix, depths = NULL) {
      n <- nrow(adj_matrix)
      in_degree <- colSums(adj_matrix)
      queue <- which(in_degree == 0)
      result <- integer(0)

      # Sort initial queue by depth (farthest first) if depths provided
      if (!is.null(depths) && length(queue) > 1) {
            queue <- queue[order(-depths[queue])]
      }

      while (length(queue) > 0) {
            # Process node with no dependencies
            current <- queue[1]
            queue <- queue[-1]
            result <- c(result, current)

            # Remove edges from current node
            neighbors <- which(adj_matrix[current, ])
            new_free <- integer(0)
            for (neighbor in neighbors) {
                  in_degree[neighbor] <- in_degree[neighbor] - 1
                  if (in_degree[neighbor] == 0) {
                        new_free <- c(new_free, neighbor)
                  }
            }

            # Insert newly freed nodes into queue, sorted by depth
            if (length(new_free) > 0) {
                  if (!is.null(depths) && length(new_free) > 1) {
                        new_free <- new_free[order(-depths[new_free])]
                  }
                  # Merge into queue maintaining depth order
                  if (length(queue) == 0 || is.null(depths)) {
                        queue <- c(queue, new_free)
                  } else {
                        combined <- c(queue, new_free)
                        queue <- combined[order(-depths[combined])]
                  }
            }
      }

      # If we couldn't sort all nodes, there's a cycle (shouldn't happen after break_cycles)
      if (length(result) < n) {
            warning("Topological sort failed - remaining cycles detected")
            remaining <- setdiff(1:n, result)
            result <- c(result, remaining)
      }

      return(result)
}

# calculate edge normals
calculate_normals <- function(polygon_vertices) {
      # Input: polygon_vertices is a matrix where each row is a vertex (x, y)
      # Example: matrix(c(0,0, 1,0, 1,1, 0,1), ncol=2, byrow=TRUE) for a square

      num_vertices <- nrow(polygon_vertices)
      edge_vectors <- polygon_vertices[c(2:num_vertices, 1), ] - polygon_vertices
      normals <- cbind(-edge_vectors[, 2], edge_vectors[, 1])
      norm_values <- sqrt(rowSums(normals^2))
      normalized_normals <- normals / norm_values

      return(normalized_normals)
}

# Project vertices onto an axis
project_onto_axis <- function(vertices, axis) {
      # Input: vertices is a matrix of polygon vertices (x, y)
      #        axis is a 1x2 vector representing the separating axis

      # Project each vertex onto the axis using the dot product
      # Matrix multiplication is efficient for this
      projections <- vertices %*% axis

      # Find the minimum and maximum projection values
      min_projection <- min(projections)
      max_projection <- max(projections)

      return(list(min = min_projection, max = max_projection))
}

# separating axis theorem test for overlap
sat_overlap <- function(poly1, poly2, tol = 1e-9) {

      for(v in 1:2){
            if(v == 1){
                  v1 <- poly1
                  v2 <- poly2
            }else{
                  v1 <- poly2
                  v2 <- poly1
            }

            axes <- na.omit(calculate_normals(v1))
            for (i in 1:nrow(axes)) {
                  proj1 <- project_onto_axis(v1, axes[i,])
                  proj2 <- project_onto_axis(v2, axes[i,])
                  if (proj1$max <= proj2$min + tol || proj2$max <= proj1$min + tol) {
                        return(FALSE) # found separating axis, polygons don't touch
                  }
            }
      }

      return(TRUE) # no separating axis found, polygons overlap
}

get_centroid <- function(p) {
      x <- p$x
      y <- p$y

      # Close the loop if it's not already closed
      n <- length(x)
      if (x[1] != x[n] || y[1] != y[n]) {
            x <- c(x, x[1])
            y <- c(y, y[1])
            n <- n + 1
      }

      # Pre-calculate the cross products
      i <- 1:(n - 1)
      cross_prod <- x[i] * y[i+1] - x[i+1] * y[i]

      # Calculate Area
      area <- sum(cross_prod) / 2

      # Calculate Centroid Coordinates
      cx <- sum((x[i] + x[i+1]) * cross_prod) / (6 * area)
      cy <- sum((y[i] + y[i+1]) * cross_prod) / (6 * area)

      return(c(x = cx, y = cy))
}

# Get centroid of polygon intersection
intersection_centroid <- function(poly1, poly2) {
      clipped <- polyclip::polyclip(list(as.data.frame(poly1)),
                                    list(as.data.frame(poly2)))
      if(length(clipped) == 0) return(NULL) # non-intersecting polygons

      centroids <- do.call("rbind", lapply(clipped, get_centroid))

      # if multiple intersection polys, average their centroids
      return(colMeans(centroids))
}


# Point-in-polygon test using ray casting
point_in_polygon <- function(px, py, poly_x, poly_y) {
      n <- length(poly_x)
      inside <- FALSE
      j <- n
      for (i in seq_len(n)) {
            if (((poly_y[i] > py) != (poly_y[j] > py)) &&
                (px < (poly_x[j] - poly_x[i]) * (py - poly_y[i]) /
                 (poly_y[j] - poly_y[i]) + poly_x[i])) {
                  inside <- !inside
            }
            j <- i
      }
      inside
}


# Pre-compute plane coefficients for polygon depth interpolation ---------------

#' Compute least-squares plane betas for all polygon groups
#'
#' Returns a list parallel to grp_data where polygon entries contain a
#' length-3 numeric vector (intercept, x_coef, y_coef) and non-polygon
#' entries are NULL.
#'
#' @param grp_data List of per-group matrices (x, y, z, bbox...).
#' @param prim_lookup Named character vector of primitive types per group.
#' @return List of beta vectors (or NULL for non-polygons / degenerate cases).
#' @keywords internal
#' @noRd
compute_poly_betas <- function(grp_data, prim_lookup) {
      betas <- vector("list", length(grp_data))
      for (i in seq_along(grp_data)) {
            if (prim_lookup[i] != "polygon") next
            m <- grp_data[[i]]
            if (nrow(m) < 3) next
            xmat <- cbind(1, m[, 1], m[, 2])
            betas[[i]] <- tryCatch(
                  solve(crossprod(xmat), crossprod(xmat, m[, 3])),
                  error = function(e) NULL
            )
      }
      betas
}


# Pairwise comparison modules -------------------------------------------------
# Each returns a depth delta (positive = i in front of j) or NA (no constraint)

# Compare two polygon groups
pw_compare_poly_poly <- function(di, dj, bbi, bbj, beta_i = NULL, beta_j = NULL) {
      # Bounding box overlap check
      if(bbi[1] >= bbj[2]) return(NA)
      if(bbi[2] <= bbj[1]) return(NA)
      if(bbi[3] >= bbj[4]) return(NA)
      if(bbi[4] <= bbj[3]) return(NA)
      if(!sat_overlap(di[,1:2], dj[,1:2])) return(NA)

      # xy overlap but no z overlap: standard depth precedence
      if(bbi[5] >= bbj[6] | bbj[5] >= bbi[6]){
            return(mean(bbi[5:6]) - mean(bbj[5:6]))
      }

      # Overlapping in all dimensions: sample depth at intersection centroid
      int <- intersection_centroid(di[, 1:2], dj[, 1:2])
      if(is.null(int)) return(NA)
      zi <- interpolate_z(di[, 1:3], int, beta = beta_i)
      zj <- interpolate_z(dj[, 1:3], int, beta = beta_j)
      zi - zj
}

# Compare a polygon group with a point group
pw_compare_poly_point <- function(poly_data, point_data, poly_beta = NULL) {
      px <- point_data[1, 1]
      py <- point_data[1, 2]
      pz <- point_data[1, 3]

      # Check if point falls within polygon's bounding box
      poly_bb <- poly_data[1, 4:9]
      if (px < poly_bb[1] || px > poly_bb[2] ||
          py < poly_bb[3] || py > poly_bb[4]) {
            return(NA)
      }

      # Point-in-polygon test
      if (!point_in_polygon(px, py, poly_data[, 1], poly_data[, 2])) {
            return(NA)
      }

      # Point is inside polygon's xy projection: interpolate polygon depth at point location
      poly_z <- interpolate_z(poly_data[, 1:3], c(px, py), beta = poly_beta)
      poly_z - pz
}

# Compare a polygon group with a segment group
# Uses polyclip to find the segment's intersection with the polygon,
# then samples depth at the midpoint of the clipped portion.
pw_compare_poly_segment <- function(poly_data, seg_data, poly_beta = NULL) {
      seg_x <- seg_data[, 1]
      seg_y <- seg_data[, 2]

      poly_bb <- poly_data[1, 4:9]

      # Quick bounding box check
      seg_xmin <- min(seg_x); seg_xmax <- max(seg_x)
      seg_ymin <- min(seg_y); seg_ymax <- max(seg_y)

      if (seg_xmin >= poly_bb[2] || seg_xmax <= poly_bb[1] ||
          seg_ymin >= poly_bb[4] || seg_ymax <= poly_bb[3]) {
            return(NA)
      }

      # Clip segment against polygon to find overlap region
      seg_line <- list(list(x = seg_x, y = seg_y))
      poly_clip <- list(list(x = poly_data[, 1], y = poly_data[, 2]))
      clipped <- tryCatch(
            polyclip::polyclip(seg_line, poly_clip, op = "intersection"),
            error = function(e) list()
      )

      if (length(clipped) == 0) return(NA)

      # Compute midpoint of the clipped portion(s)
      all_x <- unlist(lapply(clipped, `[[`, "x"))
      all_y <- unlist(lapply(clipped, `[[`, "y"))
      if (length(all_x) == 0) return(NA)

      sample_xy <- c(mean(all_x), mean(all_y))

      # Interpolate depths at sample point
      poly_z <- interpolate_z(poly_data[, 1:3], sample_xy, beta = poly_beta)
      seg_z <- interpolate_z_on_segment(seg_data[, 1:3], sample_xy)

      poly_z - seg_z
}

# Compare two segment groups
# Finds xy intersection point, interpolates depth of each segment there.
pw_compare_seg_seg <- function(seg_i, seg_j) {
      int <- segment_intersection(
            seg_i[1, 1], seg_i[1, 2], seg_i[2, 1], seg_i[2, 2],
            seg_j[1, 1], seg_j[1, 2], seg_j[2, 1], seg_j[2, 2]
      )
      if (is.null(int)) return(NA)

      zi <- interpolate_z_on_segment(seg_i[, 1:3], int)
      zj <- interpolate_z_on_segment(seg_j[, 1:3], int)
      zi - zj
}

# Compare a point group with a segment group
# Triggers when point is within segment's bounding box; compares depths
# at the point on the segment closest to the point center.
pw_compare_point_seg <- function(point_data, seg_data) {
      px <- point_data[1, 1]
      py <- point_data[1, 2]
      pz <- point_data[1, 3]

      # Bounding box check
      seg_xmin <- min(seg_data[, 1]); seg_xmax <- max(seg_data[, 1])
      seg_ymin <- min(seg_data[, 2]); seg_ymax <- max(seg_data[, 2])

      if (px < seg_xmin || px > seg_xmax ||
          py < seg_ymin || py > seg_ymax) {
            return(NA)
      }

      # Project point onto segment axis, clamp to endpoints
      closest <- closest_point_on_segment(
            px, py,
            seg_data[1, 1], seg_data[1, 2],
            seg_data[2, 1], seg_data[2, 2]
      )

      # Interpolate segment depth at the closest point
      seg_z <- interpolate_z_on_segment(seg_data[, 1:3], closest)
      seg_z - pz
}


# Geometry helpers for pairwise comparisons ------------------------------------

# Interpolate depth along a segment at a given xy point
# Uses linear interpolation based on projection parameter t
interpolate_z_on_segment <- function(seg_xyz, xy) {
      dx <- seg_xyz[2, 1] - seg_xyz[1, 1]
      dy <- seg_xyz[2, 2] - seg_xyz[1, 2]
      len_sq <- dx^2 + dy^2

      if (len_sq < 1e-20) {
            # Degenerate segment (zero length): return mean depth
            return(mean(seg_xyz[, 3]))
      }

      t <- ((xy[1] - seg_xyz[1, 1]) * dx + (xy[2] - seg_xyz[1, 2]) * dy) / len_sq
      t <- max(0, min(1, t))
      seg_xyz[1, 3] + t * (seg_xyz[2, 3] - seg_xyz[1, 3])
}

# Find intersection point of two line segments in 2D
# Returns c(x, y) or NULL if segments don't intersect
segment_intersection <- function(x1, y1, x2, y2, x3, y3, x4, y4) {
      dx1 <- x2 - x1; dy1 <- y2 - y1
      dx2 <- x4 - x3; dy2 <- y4 - y3

      denom <- dx1 * dy2 - dy1 * dx2
      if (abs(denom) < 1e-12) return(NULL)  # Parallel or collinear

      t <- ((x3 - x1) * dy2 - (y3 - y1) * dx2) / denom
      u <- ((x3 - x1) * dy1 - (y3 - y1) * dx1) / denom

      # Check that intersection is within both segments
      if (t < 0 || t > 1 || u < 0 || u > 1) return(NULL)

      c(x1 + t * dx1, y1 + t * dy1)
}

# Find closest point on a line segment to a given point
# Returns c(x, y) of the closest point, clamped to segment endpoints
closest_point_on_segment <- function(px, py, x1, y1, x2, y2) {
      dx <- x2 - x1; dy <- y2 - y1
      len_sq <- dx^2 + dy^2

      if (len_sq < 1e-20) return(c(x1, y1))  # Degenerate segment

      t <- ((px - x1) * dx + (py - y1) * dy) / len_sq
      t <- max(0, min(1, t))
      c(x1 + t * dx, y1 + t * dy)
}

# Find t-values where a segment crosses polygon edges in 2D
# Returns sorted numeric vector of t-values along the segment, or NULL
segment_polygon_edge_ts <- function(sx1, sy1, sx2, sy2, poly_x, poly_y) {
      n <- length(poly_x)
      ts <- numeric(0)
      dx_seg <- sx2 - sx1
      dy_seg <- sy2 - sy1
      seg_len_sq <- dx_seg^2 + dy_seg^2
      if (seg_len_sq < 1e-20) return(NULL)

      j <- n
      for (i in seq_len(n)) {
            # Polygon edge j -> i
            dx_edge <- poly_x[i] - poly_x[j]
            dy_edge <- poly_y[i] - poly_y[j]

            denom <- dx_seg * dy_edge - dy_seg * dx_edge
            if (abs(denom) > 1e-12) {
                  ox <- poly_x[j] - sx1
                  oy <- poly_y[j] - sy1
                  t_seg <- (ox * dy_edge - oy * dx_edge) / denom
                  u_edge <- (ox * dy_seg - oy * dx_seg) / denom

                  if (t_seg > 1e-9 && t_seg < (1 - 1e-9) &&
                      u_edge >= 0 && u_edge <= 1) {
                        ts <- c(ts, t_seg)
                  }
            }
            j <- i
      }
      if (length(ts) == 0) return(NULL)
      sort(ts)
}


# Primitive type priority for coplanar tiebreaking (higher = renders on top)
.prim_priority <- c(polygon = 1, segment = 2, point = 3, text = 3)

# Coplanar tiebreaker tolerance
.coplanar_tol <- 1e-8

# Dispatch pairwise comparison based on primitive types.
# Returns depth delta (positive = i in front of j) or NA (no constraint).
# When primitives are coplanar (delta ~ 0), smaller primitives render on top.
pw_compare_pair <- function(di, dj, type_i, type_j, beta_i = NULL, beta_j = NULL) {

      dlt <- pw_compare_pair_raw(di, dj, type_i, type_j, beta_i, beta_j)

      # Coplanar tiebreaker: smaller primitives on top
      if (!is.na(dlt) && abs(dlt) < .coplanar_tol) {
            pi <- .prim_priority[type_i] %||% 1
            pj <- .prim_priority[type_j] %||% 1
            if (pi != pj) return(pj - pi)  # higher priority j => positive => i behind j
      }

      dlt
}

# Raw pairwise comparison without tiebreaker
pw_compare_pair_raw <- function(di, dj, type_i, type_j, beta_i = NULL, beta_j = NULL) {

      # Polygon vs polygon
      if (type_i == "polygon" && type_j == "polygon") {
            bbi <- di[1, 4:9]
            bbj <- dj[1, 4:9]
            return(pw_compare_poly_poly(di, dj, bbi, bbj, beta_i = beta_i, beta_j = beta_j))
      }

      # Polygon vs point/text
      if (type_i == "polygon" && type_j %in% c("point", "text")) {
            return(pw_compare_poly_point(di, dj, poly_beta = beta_i))
      }
      if (type_j == "polygon" && type_i %in% c("point", "text")) {
            dlt <- pw_compare_poly_point(dj, di, poly_beta = beta_j)
            if (!is.na(dlt)) return(-dlt) else return(NA)
      }

      # Polygon vs segment
      if (type_i == "polygon" && type_j == "segment") {
            return(pw_compare_poly_segment(di, dj, poly_beta = beta_i))
      }
      if (type_j == "polygon" && type_i == "segment") {
            dlt <- pw_compare_poly_segment(dj, di, poly_beta = beta_j)
            if (!is.na(dlt)) return(-dlt) else return(NA)
      }

      # Segment vs segment
      if (type_i == "segment" && type_j == "segment") {
            return(pw_compare_seg_seg(di, dj))
      }

      # Point/text vs segment
      if (type_i %in% c("point", "text") && type_j == "segment") {
            return(pw_compare_point_seg(di, dj))
      }
      if (type_j %in% c("point", "text") && type_i == "segment") {
            dlt <- pw_compare_point_seg(dj, di)
            if (!is.na(dlt)) return(-dlt) else return(NA)
      }

      # Point vs point: no geometric constraint
      NA
}


# Pairwise render order with mixed-geometry support ----------------------------
# Uses base R for performance in the hot path.

pw_render_order <- function(data,
                            depth_var = "depth"
){
      has_prim <- ".prim" %in% names(data)

      grp <- data$group
      x <- data$x
      y <- data$y
      z <- data[[depth_var]]

      ugrp <- unique(grp)
      grp_idx <- split(seq_along(grp), grp)
      grp_idx <- grp_idx[ugrp]

      if (has_prim) {
            prim_vec <- data$.prim
            prim_lookup <- vapply(grp_idx, function(idx) prim_vec[idx[1]], character(1))
      } else {
            prim_lookup <- rep("polygon", length(ugrp))
            names(prim_lookup) <- ugrp
      }

      grp_data <- lapply(grp_idx, function(idx) {
            gx <- x[idx]; gy <- y[idx]; gz <- z[idx]
            bb <- c(min(gx), max(gx), min(gy), max(gy), min(gz), max(gz))
            m <- cbind(x = gx, y = gy, z = gz)
            cbind(m, matrix(bb, nrow = nrow(m), ncol = 6, byrow = TRUE))
      })

      poly_betas <- compute_poly_betas(grp_data, prim_lookup)

      # Cache betas by group name (for cheap reconstitution after splitting)
      names(poly_betas) <- names(grp_data)

      # Split segments that interpenetrate polygons
      data <- split_interpenetrating_segments(data, grp_data, grp_idx,
                                              prim_lookup, poly_betas)

      # Incrementally add new groups from splitting (only if data changed)
      new_grp <- data$group
      new_groups <- setdiff(unique(new_grp), ugrp)

      if (length(new_groups) > 0) {
            # Remove entries for original groups that were split
            split_originals <- unique(sub("__split[0-9]+$", "", new_groups))
            drop_idx <- which(names(grp_data) %in% split_originals)
            if (length(drop_idx) > 0) {
                  grp_data <- grp_data[-drop_idx]
                  prim_lookup <- prim_lookup[-drop_idx]
                  poly_betas <- poly_betas[-drop_idx]
            }

            # Build grp_data entries for new groups only
            new_x <- data$x; new_y <- data$y; new_z <- data[[depth_var]]
            new_mask <- new_grp %in% new_groups
            new_seq <- which(new_mask)
            new_grp_idx <- split(new_seq, new_grp[new_mask])

            for (ng in new_groups) {
                  idx <- new_grp_idx[[ng]]
                  gx <- new_x[idx]; gy <- new_y[idx]; gz <- new_z[idx]
                  bb <- c(min(gx), max(gx), min(gy), max(gy), min(gz), max(gz))
                  m <- cbind(x = gx, y = gy, z = gz)
                  grp_data[[ng]] <- cbind(m, matrix(bb, nrow = nrow(m), ncol = 6, byrow = TRUE))
                  prim_lookup[[ng]] <- "segment"
                  poly_betas[[ng]] <- NULL
            }
      }

      k <- length(grp_data)
      delta <- matrix(NA_real_, k, k)
      groups <- names(grp_data)

      for (i in 1:k) {
            for (j in 1:k) {
                  if (i >= j) next

                  di <- grp_data[[i]]
                  dj <- grp_data[[j]]

                  ti <- prim_lookup[i]
                  tj <- prim_lookup[j]

                  dlt <- pw_compare_pair(di, dj, ti, tj,
                                         beta_i = poly_betas[[i]],
                                         beta_j = poly_betas[[j]])

                  if (!is.na(dlt)) {
                        delta[i, j] <- dlt
                        delta[j, i] <- -dlt
                  }
            }
      }

      delta_pruned <- break_cycles(delta)
      adj_logical <- (delta_pruned > 0) & !is.na(delta_pruned)

      grp_depths <- vapply(grp_data, function(m) mean(m[, 3]), numeric(1))

      render_order <- groups[topological_sort(adj_logical, grp_depths)]
      list(data = data, render_order = match(data$group, render_order))
}


# Hierarchical depth sorting with mixed-geometry support
#
# Sorts data back-to-front for painter's algorithm rendering, supporting
# mixed primitive types (polygon, point, segment, text) via the `.prim`
# column. Preserves vertex order within multi-vertex groups (polygons).
#
# Sort hierarchy for groups with `__` separators:
#   level1 (e.g. "col1") sorted by max depth across all child groups
#   level2 (e.g. "col1__zmax") sorted by representative depth
#
# Pairwise sorting handles all primitive types via type-specific comparison
# modules. When primitives are coplanar, smaller types render on top.
sort_by_depth <- function(data) {

      if(".sort_method" %in% names(data)){
            method <- data$.sort_method[1]
            if(method == "auto"){
                  method <- if(nrow(data) <= 500) "pairwise" else "painter"
            }
      }else{
            method <- "painter"
      }

      # Vertex order within each group (preserves polygon winding)
      data <- data %>%
            group_by(group) %>%
            mutate(.vertex_order = row_number()) %>%
            ungroup()

      # Compute per-group representative depth
      data <- compute_prim_depth(data)

      if (method == "pairwise") {
            result <- pw_render_order(data, "depth")
            data <- result$data
            # Recompute .vertex_order and .prim_depth for any split rows
            data <- data %>%
                  group_by(group) %>%
                  mutate(.vertex_order = row_number(),
                         .prim_depth = mean(depth, na.rm = TRUE)) %>%
                  ungroup()
            data$.render_order <- result$render_order
            data <- data %>%
                  arrange(.render_order, .vertex_order) %>%
                  select(-.render_order)

      } else if(any(grepl("__", data$group))) {

            has_mixed_prims <- ".prim" %in% names(data) &&
                  length(unique(data$.prim[!is.na(data$.prim)])) > 1

            if (has_mixed_prims) {
                  # Mixed primitive types: flat sort by per-group depth to allow
                  # interleaving (hierarchical sort would block by level1)
                  data <- data %>%
                        arrange(desc(.prim_depth), group, .vertex_order)
            } else {
                  # Single primitive type: hierarchical sort preserves object grouping
                  data <- data %>%
                        tidyr::separate(group, c("level1", "level2"), sep = "__",
                                        remove = FALSE, extra = "merge") %>%
                        group_by(level1) %>% mutate(depth1 = max(depth)) %>%
                        ungroup() %>%
                        arrange(desc(depth1), desc(.prim_depth), group, .vertex_order) %>%
                        select(-level1, -level2, -depth1)
            }

      } else {
            # Simple painter sorting
            data <- data %>%
                  arrange(desc(.prim_depth), group, .vertex_order)
      }

      data <- data %>% select(-.vertex_order, -.prim_depth)
      return(data)
}


#' Compute per-group representative depth for sorting
#'
#' Uses mean depth of all vertices in each group, which works correctly
#' for all primitive types (polygon, point, segment, text).
#'
#' @param data Data frame with `group`, `depth`, and optionally `.prim`.
#' @return Data frame with `.prim_depth` column added.
#' @keywords internal
#' @noRd
compute_prim_depth <- function(data) {
      depth_col <- if("depth_3d" %in% names(data)) "depth_3d" else "depth"

      data <- data %>%
            group_by(group) %>%
            mutate(.prim_depth = mean(.data[[depth_col]], na.rm = TRUE)) %>%
            ungroup()

      return(data)
}


# Segment-polygon interpenetration splitting -----------------------------------

#' Split segments at polygon interpenetration points
#'
#' For each segment, finds all polygons it passes through in 3D (i.e., the
#' segment is in front of the polygon at one end and behind at the other).
#' Computes the parameter t where segment depth equals polygon depth, and
#' splits the segment at all such crossing points.
#'
#' Handles three overlap cases:
#' \itemize{
#'   \item Both segment endpoints inside polygon footprint
#'   \item One endpoint inside, one outside
#'   \item Both endpoints outside but segment crosses through polygon
#' }
#'
#' @param data Data frame with columns x, y, depth (post-transform), group,
#'   .prim, and all aesthetic columns.
#' @param grp_data List of per-group matrices from pw_render_order.
#' @param grp_idx List of row index vectors per group, keyed by group name.
#' @param prim_lookup Named character vector of primitive types per group.
#' @param poly_betas List of pre-computed plane beta vectors.
#' @return Data frame with interpenetrating segments replaced by sub-segments.
#'   Non-segment rows are returned unchanged.
#' @keywords internal
#' @noRd
split_interpenetrating_segments <- function(data, grp_data, grp_idx,
                                            prim_lookup, poly_betas) {

      seg_which <- which(prim_lookup == "segment")
      poly_which <- which(prim_lookup == "polygon")

      if (length(seg_which) == 0 || length(poly_which) == 0) return(data)

      # Tolerance for near-surface detection: if either endpoint's depth
      # difference from the polygon is below this, the segment is touching
      # rather than crossing (e.g. residuals originating at their surface)
      .split_tol <- 0.01

      # Pre-extract polygon xy and bboxes for fast access
      n_poly <- length(poly_which)
      poly_px <- vector("list", n_poly)
      poly_py <- vector("list", n_poly)
      poly_bb <- matrix(NA_real_, nrow = n_poly, ncol = 4)

      for (pi in seq_along(poly_which)) {
            m <- grp_data[[poly_which[pi]]]
            poly_px[[pi]] <- m[, 1]
            poly_py[[pi]] <- m[, 2]
            poly_bb[pi, ] <- m[1, 4:7]  # xmin, xmax, ymin, ymax
      }

      # For each segment, collect all split t-values across all polygons
      splits_needed <- FALSE
      seg_split_ts <- vector("list", length(seg_which))

      for (si in seq_along(seg_which)) {
            m <- grp_data[[seg_which[si]]]
            if (nrow(m) != 2) next

            sx <- m[, 1]; sy <- m[, 2]; sz <- m[, 3]

            seg_xmin <- min(sx); seg_xmax <- max(sx)
            seg_ymin <- min(sy); seg_ymax <- max(sy)

            split_ts <- numeric(0)

            for (pi in seq_along(poly_which)) {
                  # Quick bbox rejection
                  if (seg_xmin > poly_bb[pi, 2] || seg_xmax < poly_bb[pi, 1] ||
                      seg_ymin > poly_bb[pi, 4] || seg_ymax < poly_bb[pi, 3]) {
                        next
                  }

                  ppx <- poly_px[[pi]]
                  ppy <- poly_py[[pi]]

                  p1_in <- point_in_polygon(sx[1], sy[1], ppx, ppy)
                  p2_in <- point_in_polygon(sx[2], sy[2], ppx, ppy)

                  if (!p1_in && !p2_in) {
                        # Neither endpoint inside: check if segment crosses polygon edges
                        edge_ts <- segment_polygon_edge_ts(sx[1], sy[1], sx[2], sy[2], ppx, ppy)
                        if (is.null(edge_ts) || length(edge_ts) < 2) next

                        # Segment enters and exits polygon; check depth at the entry/exit
                        t_enter <- edge_ts[1]
                        t_exit <- edge_ts[length(edge_ts)]

                        xy_enter <- c(sx[1] + t_enter * (sx[2] - sx[1]),
                                      sy[1] + t_enter * (sy[2] - sy[1]))
                        xy_exit <- c(sx[1] + t_exit * (sx[2] - sx[1]),
                                     sy[1] + t_exit * (sy[2] - sy[1]))

                        seg_z_enter <- sz[1] + t_enter * (sz[2] - sz[1])
                        seg_z_exit <- sz[1] + t_exit * (sz[2] - sz[1])

                        beta <- poly_betas[[poly_which[pi]]]
                        poly_z_enter <- interpolate_z(grp_data[[poly_which[pi]]][, 1:3],
                                                      xy_enter, beta = beta)
                        poly_z_exit <- interpolate_z(grp_data[[poly_which[pi]]][, 1:3],
                                                     xy_exit, beta = beta)

                        d_enter <- seg_z_enter - poly_z_enter
                        d_exit <- seg_z_exit - poly_z_exit

                        if (d_enter * d_exit >= 0 ||
                            min(abs(d_enter), abs(d_exit)) < .split_tol) next

                        # Crossing within [t_enter, t_exit] range
                        # Linear solve restricted to this sub-interval
                        denom <- (d_exit - d_enter)
                        if (abs(denom) < 1e-12) next
                        t_local <- -d_enter / denom  # fractional position within [t_enter, t_exit]
                        t_cross <- t_enter + t_local * (t_exit - t_enter)

                  } else {
                        # At least one endpoint inside polygon
                        beta <- poly_betas[[poly_which[pi]]]
                        pz1 <- interpolate_z(grp_data[[poly_which[pi]]][, 1:3],
                                             c(sx[1], sy[1]), beta = beta)
                        pz2 <- interpolate_z(grp_data[[poly_which[pi]]][, 1:3],
                                             c(sx[2], sy[2]), beta = beta)

                        d1 <- sz[1] - pz1
                        d2 <- sz[2] - pz2

                        if (d1 * d2 >= 0 ||
                            min(abs(d1), abs(d2)) < .split_tol) next

                        denom <- (sz[2] - sz[1]) - (pz2 - pz1)
                        if (abs(denom) < 1e-12) next

                        t_cross <- (pz1 - sz[1]) / denom
                  }

                  if (t_cross > 0.001 && t_cross < 0.999) {
                        split_ts <- c(split_ts, t_cross)
                  }
            }

            if (length(split_ts) > 0) {
                  seg_split_ts[[si]] <- sort(unique(split_ts))
                  splits_needed <- TRUE
            }
      }

      if (!splits_needed) return(data)

      # Perform the splits
      rows_to_drop <- integer(0)
      new_rows_list <- vector("list", length(seg_which))
      cols <- names(data)

      # Columns to interpolate at split points
      interp_cols <- c("x", "y", "depth")
      if ("depth_scale" %in% cols) interp_cols <- c(interp_cols, "depth_scale")
      if ("depth_3d" %in% cols) interp_cols <- c(interp_cols, "depth_3d")

      for (si in seq_along(seg_which)) {
            ts <- seg_split_ts[[si]]
            if (is.null(ts) || length(ts) == 0) next

            idx <- grp_idx[[seg_which[si]]]
            seg_rows <- data[idx, , drop = FALSE]

            if ("point_type" %in% cols) {
                  start_row <- seg_rows[seg_rows$point_type == "start", , drop = FALSE]
                  end_row <- seg_rows[seg_rows$point_type == "end", , drop = FALSE]
            } else {
                  start_row <- seg_rows[1, , drop = FALSE]
                  end_row <- seg_rows[2, , drop = FALSE]
            }

            orig_group <- start_row$group[1]
            orig_seg_id <- if ("segment_id" %in% cols) start_row$segment_id[1] else NA

            breaks <- c(0, ts, 1)
            n_sub <- length(breaks) - 1

            sub_rows <- vector("list", n_sub)

            for (k in seq_len(n_sub)) {
                  t0 <- breaks[k]
                  t1 <- breaks[k + 1]

                  new_start <- start_row
                  new_end <- end_row

                  for (col in interp_cols) {
                        v0 <- start_row[[col]]
                        v1 <- end_row[[col]]
                        if (is.numeric(v0) && is.numeric(v1)) {
                              new_start[[col]] <- v0 + t0 * (v1 - v0)
                              new_end[[col]] <- v0 + t1 * (v1 - v0)
                        }
                  }

                  suffix <- paste0("__split", k)
                  new_start$group <- paste0(orig_group, suffix)
                  new_end$group <- paste0(orig_group, suffix)

                  if ("segment_id" %in% cols && !is.na(orig_seg_id)) {
                        new_id <- orig_seg_id + (k - 1) * 0.01
                        new_start$segment_id <- new_id
                        new_end$segment_id <- new_id
                  }

                  sub_rows[[k]] <- rbind(new_start, new_end)
            }

            new_rows_list[[si]] <- do.call(rbind, sub_rows)
            rows_to_drop <- c(rows_to_drop, idx)
      }

      new_rows <- do.call(rbind, new_rows_list[!vapply(new_rows_list, is.null, logical(1))])

      if (is.null(new_rows) || nrow(new_rows) == 0) return(data)

      data_kept <- data[-rows_to_drop, , drop = FALSE]
      rbind(data_kept, new_rows)
}

# Stat that evaluates a function over a grid and outputs points with gradients

StatFunction3D <- ggproto("StatFunction3D", Stat,
                          required_aes = character(0),
                          default_aes = aes(x = after_stat(x),
                                            y = after_stat(y),
                                            z = after_stat(z),
                                            fill = after_stat(z)),

                          setup_data = function(data, params) {
                                if (nrow(data) == 0) {
                                      data <- data.frame(x = numeric(0), y = numeric(0), z = numeric(0))
                                }
                                data
                          },

                          compute_panel = function(data, scales, fun = NULL,
                                                   grid = "rectangle", n = 40, direction = "x", xlim = NULL, ylim = NULL, trim = TRUE,
                                                   cull_backfaces = FALSE, light = NULL, na.rm = FALSE) {

                                # Validate function
                                if (is.null(fun)) {
                                      stop("stat_function_3d requires a `fun` parameter")
                                }
                                if (!is.function(fun)) {
                                      stop("`fun` must be a function")
                                }

                                # Get x and y ranges
                                if (is.null(xlim) || is.null(ylim)) {
                                      if (is.null(scales$x) || is.null(scales$y)) {
                                            stop("xlim and ylim must be specified when no scale info is available.\n",
                                                 "Try: stat_function_3d(..., xlim = c(min, max), ylim = c(min, max))")
                                      }
                                      xlim <- xlim %||% scales$x$dimension()
                                      ylim <- ylim %||% scales$y$dimension()
                                }

                                # Validate ranges
                                if (length(xlim) != 2 || !is.numeric(xlim) || xlim[1] >= xlim[2]) {
                                      stop("xlim must be a numeric vector of length 2 with xlim[1] < xlim[2]")
                                }
                                if (length(ylim) != 2 || !is.numeric(ylim) || ylim[1] >= ylim[2]) {
                                      stop("ylim must be a numeric vector of length 2 with ylim[1] < ylim[2]")
                                }

                                # Generate point grid
                                grid_data <- make_point_grid(grid, n, direction, xlim, ylim, trim)

                                # Evaluate function
                                tryCatch({
                                      z_values <- fun(grid_data$x, grid_data$y)
                                }, error = function(e) {
                                      stop("Error evaluating function: ", e$message)
                                })

                                if (length(z_values) != nrow(grid_data)) {
                                      stop("Function must return one z value per (x,y) pair")
                                }
                                if (!is.numeric(z_values)) {
                                      stop("Function must return numeric values")
                                }

                                grid_data$z <- z_values

                                # Remove NAs if requested
                                if (na.rm) {
                                      grid_data <- grid_data[complete.cases(grid_data[c("x", "y", "z")]), ]
                                      if (nrow(grid_data) < 3) {
                                            stop("After removing NAs, insufficient points remain")
                                      }
                                }

                                # Compute point-level gradients
                                grid_data <- compute_point_gradients(grid_data)

                                # Add group if not present
                                if (!"group" %in% names(grid_data)) {
                                      grid_data$group <- 1
                                }

                                # Attach rendering parameters
                                grid_data$cull_backfaces <- cull_backfaces
                                grid_data <- attach_light(grid_data, light)

                                return(grid_data)
                          }
)


#' 3D surface from a function
#'
#' Evaluates a function f(x,y) = z over a regular grid and renders the result
#' as a 3D surface or ridgeline plot.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. By default, fill
#'   is mapped to `after_stat(z)`.
#' @param data Ignored; this stat generates its own data.
#' @param xlim,ylim Length-2 numeric vectors giving the x and y ranges over which to
#'   plot the function.
#' @param geom Geom to use for rendering. Defaults to `"surface_3d"` for mesh
#'   surfaces. Use `"ridgeline_3d"` for ridgeline rendering.
#' @param stat Statistical transformation to use on the data. Defaults to `"surface_3d"`.
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed to the layer.
#' @param fun Function to evaluate. Must accept (x, y) and return numeric z values.
#' @inheritParams grid_params
#' @inheritParams ridgeline_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#'
#' @return A `Layer` object that can be added to a ggplot.
#'
#' @section Computed variables:
#' \describe{
#'   \item{x, y, z}{Grid coordinates and function values}
#'   \item{dzdx, dzdy}{Partial derivatives at each point}
#'   \item{slope}{Gradient magnitude: sqrt(dzdx^2 + dzdy^2)}
#'   \item{aspect}{Direction of steepest slope: atan2(dzdy, dzdx)}
#' }
#'
#' @examples
#' # Basic function surface
#' ggplot() +
#'   geom_function_3d(fun = function(x, y) sin(x) * cos(y),
#'                    xlim = c(-pi, pi), ylim = c(-pi, pi)) +
#'   coord_3d()
#'
#' \donttest{
#'
#' # Fill by slope
#' ggplot() +
#'   geom_function_3d(fun = function(x, y) x^2 + y^2,
#'                    xlim = c(-2, 2), ylim = c(-2, 2),
#'                    aes(fill = after_stat(slope)),
#'                    grid = "equilateral") +
#'   scale_fill_viridis_c() +
#'   coord_3d()
#'
#' # As ridgelines
#' ggplot() +
#'   stat_function_3d(fun = function(x, y) dnorm(x) * dnorm(y) * 10,
#'                    xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), n = c(15, 30),
#'                    geom = "ridgeline_3d", base = 0, light = "none",
#'                    fill = "black", color = "white") +
#'   coord_3d()
#'
#' }
#'
#' @seealso [geom_surface_3d()], [geom_ridgeline_3d()], [coord_3d()]
#' @export
stat_function_3d <- function(mapping = NULL, data = ensure_nonempty_data,
                             geom = "surface_3d", position = "identity",
                             ...,
                             fun = NULL,
                             xlim = NULL, ylim = NULL,
                             grid = "rectangle", n = 40, direction = "x", trim = TRUE,
                             cull_backfaces = FALSE,
                             light = NULL,
                             na.rm = FALSE,
                             show.legend = NA,
                             inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = StatFunction3D,
            geom = get_proto(geom, "geom"),
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  fun = fun,
                  xlim = xlim, ylim = ylim,
                  grid = grid, n = n, direction = direction, trim = trim,
                  cull_backfaces = cull_backfaces,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}


#' @rdname stat_function_3d
#' @export
geom_function_3d <- function(mapping = NULL, data = ensure_nonempty_data,
                             stat = "function_3d", position = "identity",
                             ...,
                             fun = NULL,
                             xlim = NULL, ylim = NULL,
                             grid = "rectangle", n = 40, direction = "x", trim = TRUE,
                             cull_backfaces = FALSE,
                             light = NULL,
                             na.rm = FALSE,
                             show.legend = NA,
                             inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = get_proto(stat, "stat"),
            geom = GeomSurface3D,
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  fun = fun,
                  xlim = xlim, ylim = ylim,
                  grid = grid, n = n, direction = direction, trim = trim,
                  cull_backfaces = cull_backfaces,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}


#' Ensure non-empty data for stats that generate their own data
#' @keywords internal
#' @noRd
ensure_nonempty_data <- function(data) {
      if (is.null(data) || nrow(data) == 0 || ncol(data) == 0 || inherits(data, "waiver")) {
            data.frame(group = -1L)
      } else {
            data
      }
}

# Shared documentation for ggcube functions


# sorting ----------------------------------------------------

#' Depth sorting method
#'
#' The `sort_method` argument determines the order in which elements are rendered.
#' This controls depth sorting for all geometry types within a layer, including
#' polygons, points, segments, and text. Default varies by geometry type.
#'
#' @section Options:
#' \describe{
#'   \item{\code{"painter"}}{Elements are sorted by the mean depth (distance from
#'     viewer after rotation) of their vertices. This is fast, but can give incorrect
#'     results when primitives overlap in screen space at different depths.}
#'   \item{\code{"pairwise"}}{A more intensive sorting algorithm that compares every
#'     pair of elements to determine occlusion order. Uses type-specific geometric
#'     tests: polygon overlap detection for polygon-polygon pairs, point-in-polygon
#'     tests for polygon-point pairs, line clipping for polygon-segment pairs, and
#'     line intersection for segment-segment pairs. When elements are coplanar,
#'     smaller primitives (points, segments) render on top of larger ones (polygons).
#'     Slower but more accurate.}
#'   \item{\code{"auto"}}{Uses pairwise if the data has fewer than 500 rows, and
#'     painter otherwise.}
#' }
#'
#' @name sorting_methods
NULL

#' Sorting parameters
#'
#' @param sort_method Depth sorting algorithm. See [sorting_methods] for details.
#' @name sort_params
#' @keywords internal
NULL


#' Depth scaling parameter
#'
#' Generic wording for layers whose depth-scaled property is something other
#' than polygon linewidth. Polygon layers document `scale_depth` through
#' [polygon_params] instead, which describes the mean-depth behaviour specific
#' to them. Keep the two in step when the semantics change.
#'
#' @param scale_depth Controls depth-based scaling, drawing closer elements
#'   larger or thicker and farther ones smaller or thinner. `TRUE` (the default)
#'   applies full scaling and `FALSE` disables it; a number sets the strength
#'   directly, where `1` matches `TRUE`, `0` matches `FALSE`, values below 1
#'   subdue the effect and values above 1 exaggerate it. This affects the layer
#'   only; use `coord_3d(scale_depth = )` for panel and axis components.
#' @name depth_params
#' @keywords internal
NULL


# polygons ----------------------------------------------------

#' Polygon rendering parameters
#'
#' Parameters controlling depth sorting, backface culling, depth scaling, and
#' convexity enforcement for polygon-based 3D layers.
#'
#' @param sort_method Depth sorting algorithm. See [sorting_methods] for details.
#' @param cull_backfaces Logical indicating whether to remove back-facing polygons
#'   from rendering. This is primarily for performance optimization but may be useful
#'   for aesthetic reasons in some situations. Backfaces are determined using
#'   screen-space winding order after 3D transformation. Defaults vary by geometry
#'   type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
#'   voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
#'   or explicitly disabled).
#' @param scale_depth Controls whether polygon linewidths are scaled to make closer lines
#'   wider and farther lines narrower, based on the mean depth of a polygon. `TRUE`
#'   (the default) applies full scaling and `FALSE` disables it; a number sets the
#'   strength directly, where `1` matches `TRUE`, `0` matches `FALSE`, values below
#'   1 subdue the effect and values above 1 exaggerate it. This affects the layer
#'   only; use `coord_3d(scale_depth = )` for gridlines, cube borders, axis ticks,
#'   and axis text.
#' @param force_convex Logical indicating whether to remove polygon vertices that are not part of the
#'   convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
#'   that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
#'   voxels, FALSE is safe because polygons fill always be convex.
#' @param na.rm If `FALSE`, missing values are removed.
#' @param show.legend Logical indicating whether this layer should be included in legends.
#' @param inherit.aes If `FALSE`, overrides the default aesthetics.
#' @param ... Other arguments passed on to the layer function (typically GeomPolygon3D), such as
#'   aesthetics like `colour`, `fill`, `linewidth`, `annotate = annotate_3d(...)`, etc.
#' @name polygon_params
#' @keywords internal
NULL


#' Shared column and voxel params
#'
#' @param width Numeric value controlling box width as a fraction of grid spacing.
#'   Default is 1.0 (volumes touch each other). Use 0.8 for small gaps, 1.2 for overlap.
#'   Grid spacing is determined automatically using `ggplot2::resolution()`
#' @param faces Character vector specifying which faces to render. Options:
#'   \itemize{
#'     \item \code{"all"} (default): Render all 6 faces
#'     \item \code{"none"}: Render no faces
#'     \item Vector of face names: \code{c("zmax", "xmin", "ymax")}, etc.
#'   }
#'   Valid face names: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
#'   Note that this setting acts jointly with backface culling, which removes faces whose
#'   interior faces the viewer -- e.g., when `cull_backfaces = TRUE` and `faces = "all"`
#'   (the default), only front faces are rendered.
#' @name col_params
#' @keywords internal
NULL

#' Shared ridgeline params
#'
#' @param direction Direction of ridges:
#'   \describe{
#'     \item{"x"}{One ridge per unique x value; ridge varies in y (default)}
#'     \item{"y"}{One ridge per unique y value; ridge varies in x}
#'   }
#' @param base Z-value for ridge polygon bottoms. If NULL, uses min(z).
#' @name ridgeline_params
#' @keywords internal
NULL


# grids ----------------------------------------------------

#' Grid generation
#'
#' Parameters defining the geometry, resolution, and orientation of a regular grid
#' of surface tiles, as used by various ggcube layer functions including
#' [geom_smooth_3d()], [geom_function_3d()], [geom_density_3d()], and
#' [geom_surface_3d()].
#'
#' @section Arguments:
#' \describe{
#'   \item{`grid`}{Character specifying tile geometry. Options:
#'     \describe{
#'       \item{`"rectangle"`}{Rectangular grid (the default).}
#'       \item{`"right1"`}{Rectangular grid with each quad split into two right triangles
#'         along the bottom-left to top-right diagonal.}
#'       \item{`"right2"`}{Like `"right1"` but split along the bottom-right to top-left
#'         diagonal.}
#'       \item{`"equilateral"`}{Equilateral triangular lattice (tessellated via Delaunay
#'         triangulation). Can prevent lighting artifacts where a surface curves past
#'         parallel with the sight line.}
#'     }
#'   }
#'   \item{`n`}{Either a single integer specifying grid resolution in both dimensions,
#'     or a vector of length 2 specifying `c(nx, ny)` for separate x and y resolutions.
#'     Default is `40`. Higher values create smoother surfaces but slower rendering.}
#'   \item{`direction`}{Either `"x"` (the default) or `"y"`, specifying the orientation
#'     of tile rows. Ignored for rectangular grids.}
#'   \item{`trim`}{Logical. Only relevant for `grid = "equilateral"`. If `TRUE`
#'     (default), trims edge vertices so that grid boundaries are straight lines. If
#'     `FALSE`, preserves the full lattice, resulting in a grid with irregular edges.}
#' }
#'
#' @name grid_generation
NULL

#' Grid generation parameters
#'
#' @param grid,n,direction,trim Parameters determining the geometry, resolution, and
#'   orientation of the surface grid. See [grid_generation] for details.
#' @name grid_params
#' @keywords internal
NULL


# computed variables ----------------------------------------------------


#' Computed variables for surfaces
#'
#' @section Computed variables:
#' The following computed variables are available via `after_stat()`:
#' - `x`, `y`, `z`: Grid coordinates and function values
#' - `normal_x`, `normal_y`, `normal_z`: Surface normal components
#' - `slope`: Gradient magnitude from surface calculations
#' - `aspect`: Direction of steepest slope from surface calculations
#' - `dzdx`, `dzdy`: Partial derivatives from surface calculation
#'
#' @name surface_computed_vars
#' @keywords internal
NULL



# others ----------------------------------------------------


#' Position param
#'
#' @param position Position adjustment, defaults to "identity". To collapse the result
#'   onto one 2D surface, use `position_on_face()`.
#' @name position_param
#' @keywords internal
NULL


#' Light param
#'
#' @param light A lighting specification object created by [light()], `light("none")`
#'   or the string `"none"` to disable lighting, or `NULL` to inherit plot-level
#'   lighting. Set plot-level lighting by adding [light()] to the plot, and
#'   layer-specific lighting via the `light` argument of `geom_*_3d()` functions,
#'   which takes precedence.
#' @name light_param
#' @keywords internal
NULL

# Stat that takes user-provided xyz data, validates it, and computes gradients

StatSurface3D <- ggproto("StatSurface3D", Stat,
                         required_aes = c("x", "y", "z"),
                         default_aes = aes(fill = after_stat(z)),

                         compute_group = function(data, scales, na.rm = FALSE,
                                                  cull_backfaces = FALSE, light = NULL) {

                               # Remove NAs if requested
                               if (na.rm) {
                                     data <- data[complete.cases(data[c("x", "y", "z")]), ]
                               }

                               if (nrow(data) < 3) {
                                     warning("stat_surface_3d requires at least 3 points; returning empty data")
                                     return(data.frame())
                               }

                               data <- compute_point_gradients(data)

                               # Attach rendering parameters
                               data$cull_backfaces <- cull_backfaces
                               data <- attach_light(data, light)

                               return(data)
                         }
)


#' 3D surface from point data
#'
#' Takes user-provided (x, y, z) point data and prepares it for surface
#' rendering. If data form a regular grid, can render either a GeomSurface3D
#' of rectangular or right-triangular tiles, or a GeomRidgeline3D or GeomContour3D
#' set of surface slices; otherwise, renders irregular triangular tiles via Delaunay
#' trianuglation.
#'
#'  For regular grids, computes point-level gradients. Works with
#' both [geom_surface_3d()] for mesh rendering and [geom_ridgeline_3d()] for
#' ridgeline rendering.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. Must include
#'   `x`, `y`, and `z`. By default, fill is mapped to `after_stat(z)`.
#' @param data Data frame containing point coordinates.
#' @param geom Geom to use for rendering. Defaults to `"surface_3d"` for mesh
#'   surfaces. Use `"ridgeline_3d"` for ridgeline rendering.
#' @param stat Statistical transformation to use on the data. Defaults to `"surface_3d"`.
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed to the layer.
#' @inheritParams ridgeline_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#'
#' @return A `Layer` object that can be added to a ggplot.
#'
#' @section Computed variables:
#' For regular grid data:
#' \describe{
#'   \item{dzdx, dzdy}{Partial derivatives at each point}
#'   \item{slope}{Gradient magnitude: sqrt(dzdx^2 + dzdy^2)}
#'   \item{aspect}{Direction of steepest slope: atan2(dzdy, dzdx)}
#' }
#'
#' For irregular data, gradient variables are NA.
#'
#' @section Grid detection:
#' The stat automatically detects whether data forms a regular grid by checking
#' if `nrow(data) == length(unique(x)) * length(unique(y))`. Regular grids
#' get point-level gradient computation; irregular point clouds are passed
#' through for Delaunay tessellation by the geom.
#'
#' @examples

#' # Regular grid data ------------------------------------------
#'
#' # simulated data and base plot for basic surface
#' d <- dplyr::mutate(tidyr::expand_grid(x = -10:10, y = -10:10),
#'                    z = sqrt(x^2 + y^2) / 1.5,
#'                    z = cos(z) - z)
#' p <- ggplot(d, aes(x, y, z)) +
#'       coord_3d(light = light(mode = "hsl", direction = c(1, 0, 0)))
#'
#' # surface with 3d lighting
#' p + geom_surface_3d(fill = "steelblue", color = "steelblue", linewidth = .2)
#'
#' \donttest{
#'
#' # mesh wireframe (`fill = NA`) with aes line color
#' p + geom_surface_3d(aes(color = z), fill = NA,
#'                     linewidth = .5, light = "none") +
#'       scale_color_gradientn(colors = c("black", "blue", "red"))
#'
#' # triangulated surface (can prevent lighting flaws)
#' p + geom_surface_3d(fill = "#9e2602", color = "black", grid = "right2")
#'
#' # use after_stat to access computed surface-orientation variables
#' p + geom_surface_3d(aes(fill = after_stat(slope)), grid = "right2") +
#'       scale_fill_viridis_c() +
#'       guides(fill = guide_colorbar_3d())
#'
#' # use `group` to plot data for multiple surfaces
#' d <- expand.grid(x = -5:5, y = -5:5)
#' d$z <- d$x^2 - d$y^2
#' d$g <- "a"
#' d2 <- d
#' d2$z <- d$z + 15
#' d2$g <- "b"
#' ggplot(rbind(d, d2), aes(x, y, z, group = g, fill = g)) +
#'       coord_3d(light = "none") +
#'       geom_surface_3d(color = "black", alpha = .5, light = NULL)
#'
#' # terrain surface with topographic hillshade and elevational fill
#' ggplot(mountain, aes(x, y, z, fill = z, color = z)) +
#'       geom_surface_3d(light = light(direction = c(1, 0, .5),
#'                                     mode = "hsv", contrast = 1.5),
#'                       linewidth = .2) +
#'       coord_3d(ratio = c(1, 1.5, .75)) +
#'       theme_light() +
#'       scale_fill_gradientn(colors = c("darkgreen", "rosybrown4", "gray60")) +
#'       scale_color_gradientn(colors = c("darkgreen", "rosybrown4", "gray60")) +
#'       guides(fill = guide_colorbar_3d())
#'
#' # stack of flat surfaces (e.g. a time series of raster maps)
#' d <- expand.grid(lon = 1:20, lat = 1:20, time = 1:6)
#' d$value <- sin(-d$lon * d$lat / 10) + d$time / 2
#' ggplot(d, aes(lon, lat, time, group = time,
#'               fill = value, color = value)) +
#'    geom_surface_3d(light = "none") +
#'    coord_3d() +
#'    scale_color_viridis_c() + scale_fill_viridis_c()
#'
#' # stat_surface_3d with alternative geoms ----------------------------
#'
#' # horizontal slices with geom_ridgeline_3d
#' ggplot(mountain, aes(x, y, z)) +
#'   stat_surface_3d(geom = "ridgeline_3d",
#'                   fill = "black", color = "white",
#'                   light = "none", linewidth = .1) +
#'       coord_3d(ratio = c(1, 1.5, .75), yaw = 45)
#'
#' # elevation contours with geom_contour_3d
#' ggplot(mountain, aes(x, y, z, fill = z)) +
#'   stat_surface_3d(geom = "contour_3d", light = "none",
#'                   bins = 25, sort_method = "pairwise",
#'                   color = "black") +
#'       coord_3d(ratio = c(1, 1.5, .75), yaw = 45) +
#'       scale_fill_viridis_c(option = "B")
#'
#'
#' # Irregular point data ---------------------------------------
#'
#' set.seed(42)
#' pts <- data.frame(x = runif(200, -2, 2), y = runif(200, -2, 2))
#' pts$z <- with(pts, sin(x) * cos(y))
#'
#' ggplot(pts, aes(x, y, z = z, fill = z)) +
#'   stat_surface_3d(sort_method = "pairwise") +
#'   scale_fill_viridis_c() +
#'   coord_3d(light = "none")
#'
#' }
#'
#' @seealso [geom_surface_3d()], [geom_ridgeline_3d()], [stat_function_3d()],
#'   [coord_3d()]
#' @export
stat_surface_3d <- function(mapping = NULL, data = NULL,
                            geom = "surface_3d", position = "identity",
                            ...,
                            cull_backfaces = FALSE,
                            light = NULL,
                            na.rm = FALSE,
                            show.legend = NA,
                            inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = StatSurface3D,
            geom = get_proto(geom, "geom"),
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  cull_backfaces = cull_backfaces,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}


#' @rdname stat_surface_3d
#' @param method Tessellation method passed to [geom_surface_3d()]:
#'   "auto" (default), "grid", or "delaunay".
#' @param grid Tile geometry for regular grids: "rectangle" (default),
#'   "right1", or "right2".
#' @export
geom_surface_3d <- function(mapping = NULL, data = NULL,
                            stat = "surface_3d", position = "identity",
                            ...,
                            method = "auto",
                            grid = NULL,
                            cull_backfaces = FALSE,
                            sort_method = "auto",
                            scale_depth = TRUE,
                            force_convex = TRUE,
                            light = NULL,
                            na.rm = FALSE,
                            show.legend = NA,
                            inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = get_proto(stat, "stat"),
            geom = GeomSurface3D,
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  method = method,
                  grid = grid,
                  cull_backfaces = cull_backfaces,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  force_convex = force_convex,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}

# Geom that converts point grid to contour band polygons for "layer cake" 3D rendering

# Geom ggproto object ---------------------------------------------------------

GeomContour3D <- ggproto("GeomContour3D", GeomPolygon3D,

                         required_aes = c("x", "y", "z"),

                         default_aes = aes(
                               colour = "white",
                               fill = "black",
                               linewidth = 0.1,
                               linetype = 1,
                               alpha = NA
                         ),

                         extra_params = c("na.rm", "annotate", "bins", "binwidth", "breaks",
                                          "cull_backfaces", "sort_method", "scale_depth", "force_convex"),

                         setup_params = function(data, params) {
                               params$bins <- params$bins %||% 10
                               params
                         },

                         setup_data = function(data, params) {

                               # Preserve aesthetic columns for later merging
                               aes_cols <- intersect(names(data),
                                                     c("fill", "colour", "color", "alpha",
                                                       "linewidth", "linetype", "PANEL",
                                                       "cull_backfaces", "lighting_spec"))

                               all_data <- split(data, data$group)

                               all_contour_data <- lapply(all_data, function(data){

                                     # Validate grid structure
                                     n_unique_x <- length(unique(round(data$x, 10)))
                                     n_unique_y <- length(unique(round(data$y, 10)))

                                     if (n_unique_x < 2 || n_unique_y < 2) {
                                           stop("geom_contour_3d() requires at least a 2x2 grid of points")
                                     }

                                     # Convert point grid to contour band polygons
                                     contour_data <- points_to_contours(
                                           data,
                                           breaks = params$breaks,
                                           bins = params$bins,
                                           binwidth = params$binwidth
                                     )

                                     # Transfer point-level aesthetic columns to contour data
                                     breaks <- c(unique(contour_data$z), Inf)

                                     distill <- function(x){
                                           ux <- unique(x)
                                           if(length(ux) == 1) return(ux)
                                           if(is.numeric(x)) return(mean(x))
                                           return(x[1]) # fallback
                                     }
                                     ds <- data %>%
                                           mutate(z = cut(z, breaks, include.lowest = TRUE),
                                                  z = breaks[as.integer(z)]) %>%
                                           group_by(z) %>%
                                           summarize(across(all_of(aes_cols), distill))
                                     contour_data <- left_join(contour_data, ds, by = join_by(z))

                                     return(na.omit(contour_data))
                               }) %>%
                                     bind_rows() %>%
                                     mutate(group = gsub("__", "_", group),
                                            group = paste0("contour__layer", group))

                               setup_annotations(all_contour_data, params)
                         }
)

#' Convert point grid to contour band polygons
#'
#' @param data Data frame with x, y, z columns forming a regular grid.
#' @param breaks Numeric vector of break points for contour levels. If NULL,
#'   breaks are computed automatically using `bins` or `binwidth`.
#' @param bins Number of contour bins. Ignored if `breaks` or `binwidth` is provided.
#' @param binwidth Width of contour bins. Ignored if `breaks` is provided.
#'
#' @return Data frame with polygon vertices for contour band rendering, including
#'   `group` and `subgroup` columns for proper hole handling.
#'
#' @keywords internal
#' @noRd
points_to_contours <- function(data,
                               breaks = NULL,
                               bins = NULL,
                               binwidth = NULL) {

      # Get unique x and y values to form the grid
      x_vals <- sort(unique(data$x))
      y_vals <- sort(unique(data$y))

      nx <- length(x_vals)
      ny <- length(y_vals)

      # Check if data forms a complete regular grid
      expected_n <- nx * ny
      if (nrow(data) != expected_n) {
            warning("Data does not form a complete regular grid. ",
                    "Expected ", expected_n, " points, got ", nrow(data), ". ",
                    "Missing values will be treated as NA.")
      }

      # Create z matrix in the format isoband expects
      # isoband expects z[i, j] where i indexes y (rows) and j indexes x (columns)
      z_matrix <- matrix(NA_real_, nrow = ny, ncol = nx)

      # Map data points to matrix positions
      x_idx <- match(data$x, x_vals)
      y_idx <- match(data$y, y_vals)

      for (i in seq_len(nrow(data))) {
            z_matrix[y_idx[i], x_idx[i]] <- data$z[i]
      }

      # Compute breaks if not provided
      z_range <- range(data$z, na.rm = TRUE)

      if (is.null(breaks)) {
            if (!is.null(binwidth)) {
                  # Extend range slightly to include endpoints
                  breaks <- seq(
                        floor(z_range[1] / binwidth) * binwidth,
                        ceiling(z_range[2] / binwidth) * binwidth,
                        by = binwidth
                  )
            } else {
                  # Use bins (default 10)
                  bins <- bins %||% 10
                  breaks <- pretty(z_range, n = bins)
            }
      }

      # Ensure breaks span the data range
      if (min(breaks) > z_range[1]) {
            breaks <- c(z_range[1], breaks)
      }
      if (max(breaks) < z_range[2]) {
            breaks <- c(breaks, z_range[2])
      }

      breaks <- sort(unique(breaks))

      if (length(breaks) < 2) {
            stop("Need at least 2 breaks to create contour bands")
      }

      # For solid discs (layer cake), each level goes from data minimum to that level
      # This means we compute isobands from z_min to each break point
      z_min <- z_range[1]

      if (length(breaks) == 0) {
            warning("No contour levels above data minimum")
            return(data.frame(
                  x = numeric(0), y = numeric(0), z = numeric(0),
                  order = integer(0), group = character(0), subgroup = character(0),
                  face_type = character(0)
            ))
      }

      bands <- isoband::isobands(
            x = x_vals,
            y = y_vals,
            z = z_matrix,
            levels_high = rep(Inf, length(breaks)),
            levels_low = breaks
      )

      bands <- lapply(1:length(breaks), function(i){
            mutate(as.data.frame(bands[[i]]),
                   z = breaks[i],
                   group = paste0(data$group[1], "::", breaks[i]))}) %>% bind_rows() %>%
            rename(.subgroup = id)

      return(bands)
}




# Layer function --------------------------------------------------------------

#' Contours of a 3D surface
#'
#' Renders a surface as stacked horizontal contour bands. Each contour band is a
#' polygon placed at its corresponding z-level.
#'
#' This geom takes point grid data (like that produced by [stat_surface_3d()],
#' [stat_function_3d()], [stat_smooth_3d()], or [stat_density_3d()]) and converts
#' it to filled contour polygons using the `isoband` package.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data Point grid data with x, y, z coordinates.
#' @param stat Statistical transformation. Defaults to `"surface_3d"`.
#' @param geom Geometry function used to display the data. Defaults to `"surface_3d"`.
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed to the layer.
#' @param bins Number of contour levels. Default is 20. Ignored if `breaks` or
#'   `binwidth` is provided.
#' @param binwidth Width of each contour band. Overrides `bins` if provided.
#' @param breaks Numeric vector specifying exact contour break points. Overrides
#'   both `bins` and `binwidth` if provided.
#' @inheritParams polygon_params
#' @inheritParams light_param
#'
#' @section Aesthetics:
#' `geom_contour_3d()` requires:
#' \describe{
#'   \item{x, y, z}{Point coordinates forming a regular grid}
#' }
#'
#' And understands these additional aesthetics:
#' \describe{
#'   \item{fill}{Band fill color. For automatic coloring by elevation, use
#'     `aes(fill = after_stat(z))`. Default is "grey60".}
#'   \item{colour}{Band border color (default: "grey30")}
#'   \item{alpha}{Transparency}
#'   \item{linewidth}{Border width (default: 0.1)}
#'   \item{linetype}{Border line type}
#' }
#'
#' @section Computed variables:
#' Each contour band is placed at its corresponding z-level (the upper boundary
#' of the band). To color by elevation, use `aes(fill = after_stat(z))`.
#'
#' @examples
#' # Basic usage with volcano data
#' ggplot(mountain, aes(x, y, z)) +
#'       geom_contour_3d(color = "white", fill = "black") +
#'       coord_3d(light = "none", ratio = c(1.5, 2, 1))
#'
#' # Map fill to elevation and customize number of levels
#' ggplot(mountain, aes(x, y, z, fill = after_stat(z))) +
#'       geom_contour_3d(bins = 12, color = "white") +
#'       scale_fill_viridis_c() +
#'       coord_3d(light = "none", ratio = c(1.5, 2, 1))
#'
#' # Specify exact breaks
#' ggplot(mountain, aes(x, y, z, fill = after_stat(z))) +
#'   geom_contour_3d(breaks = seq(0, 200, by = 5)) +
#'   scale_fill_viridis_c() +
#'   coord_3d(light = "none")
#'
#' # With stat_density_3d
#' ggplot(faithful, aes(eruptions, waiting)) +
#'   stat_density_3d(geom = "contour_3d",
#'     sort_method = "pairwise") +
#'   coord_3d()
#'
#' # With stat_function_3d
#' ggplot() +
#'   stat_function_3d(
#'     fun = function(x, y) sin(x) * cos(y),
#'     xlim = c(-pi, pi), ylim = c(-pi, pi),
#'     geom = "contour_3d",
#'     bins = 50, color = "black"
#'   ) +
#'   scale_fill_viridis_c(option = "B") +
#'   coord_3d(light = "none")
#'
#' @seealso [geom_surface_3d()] for continuous surface rendering,
#'   [geom_ridgeline_3d()] for cross-sectional ridgeline rendering,
#'   [stat_function_3d()] for mathematical function surfaces,
#'   [stat_smooth_3d()] for fitted model surfaces,
#'   [coord_3d()] for 3D coordinate systems,
#'   [ggplot2::geom_contour_filled()] for the 2D equivalent.
#'
#' @return A `Layer` object that can be added to a ggplot.
#' @export
geom_contour_3d <- function(mapping = NULL, data = NULL,
                            stat = "surface_3d", position = "identity",
                            ...,
                            bins = 20,
                            binwidth = NULL,
                            breaks = NULL,
                            cull_backfaces = FALSE,
                            sort_method = "pairwise",
                            scale_depth = TRUE,
                            force_convex = FALSE,
                            light = NULL,
                            na.rm = FALSE,
                            show.legend = NA,
                            inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = get_proto(stat, "stat"),
            geom = GeomContour3D,
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  bins = bins,
                  binwidth = binwidth,
                  breaks = breaks,
                  cull_backfaces = cull_backfaces,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  force_convex = force_convex,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}


#' @rdname geom_contour_3d
#' @export
stat_contour_3d <- function(mapping = NULL, data = NULL,
                            geom = "contour_3d", position = "identity",
                            ...,
                            bins = 20,
                            binwidth = NULL,
                            breaks = NULL,
                            cull_backfaces = FALSE,
                            sort_method = "pairwise",
                            scale_depth = TRUE,
                            force_convex = FALSE,
                            light = NULL,
                            na.rm = FALSE,
                            show.legend = NA,
                            inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = StatSurface3D,
            geom = get_proto(geom, "geom"),
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  bins = bins,
                  binwidth = binwidth,
                  breaks = breaks,
                  cull_backfaces = cull_backfaces,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  force_convex = force_convex,
                  light = light,
                  na.rm = na.rm,
                  ...
            )
      )
}

#' Continuous z-axis scale for 3D plots
#'
#' \code{scale_z_continuous} creates a continuous scale for the z aesthetic in 3D plots.
#' It works similarly to \code{scale_x_continuous} and \code{scale_y_continuous}, providing
#' control over axis breaks, labels, limits, and transformations for the z dimension.
#'
#' @param name The name of the scale, used as the axis title. Use \code{waiver()} for
#'   the default, or \code{NULL} to omit the title.
#' @param breaks One of:
#'   \itemize{
#'     \item \code{NULL} for no breaks
#'     \item \code{waiver()} for the default breaks
#'     \item A numeric vector of positions
#'     \item A function that takes the limits as input and returns breaks as output
#'   }
#' @param minor_breaks One of:
#'   \itemize{
#'     \item \code{NULL} for no minor breaks
#'     \item \code{waiver()} for the default minor breaks
#'     \item A numeric vector of positions
#'     \item A function that takes the limits as input and returns minor breaks as output
#'   }
#' @param n.breaks An integer guiding the number of major breaks. The algorithm
#'   may choose a slightly different number to ensure nice break labels.
#' @param labels One of:
#'   \itemize{
#'     \item \code{NULL} for no labels
#'     \item \code{waiver()} for the default labels
#'     \item A character vector giving labels (must be same length as breaks)
#'     \item A function that takes the breaks as input and returns labels as output
#'   }
#' @param limits A numeric vector of length two providing limits of the scale.
#'   Use \code{NA} to refer to the existing minimum or maximum.
#' @param expand For position scales, a vector of range expansion constants used to add some
#'   padding around the data to ensure that they are placed some distance away from the axes.
#' @param oob One of:
#'   \itemize{
#'     \item Function that handles limits outside the scale limits (out of bounds).
#'     \item \code{scales::censor} for replacing out of bounds values with \code{NA}
#'     \item \code{scales::squish} for squishing out of bounds values into range
#'   }
#' @param na.value Missing values will be replaced with this value.
#' @param transform The name of a transformation object or the object itself. Default is "identity",
#'   but works with standard transform options such as "log10", "sqrt", and "reverse",
#'   detailed in the documentation for [`ggplot2::scale_x_continuous()`].
#' @param guide A function used to create a guide or its name. Since z-axis guides are not
#'   yet supported, this defaults to \code{"none"}.
#' @param ... Other arguments passed on to \code{continuous_scale()}.
#'
#' @return A ggplot2 scale object for the z aesthetic.
#'
#' @examples
#' # Custom breaks, labels, and limits
#' ggplot(mtcars, aes(mpg, wt, z = qsec)) +
#'   geom_point() +
#'   scale_z_continuous(
#'     breaks = c(15, 17, 19, 21),
#'     labels = c("Fast", "Medium", "Slow", "Very Slow"),
#'     limits = c(10, NA)) +
#'   coord_3d()
#'
#' # Works with standard scale transformations like "reverse", "log10", etc.
#' ggplot(mtcars, aes(mpg, wt, z = qsec)) +
#'   geom_point() +
#'   scale_z_continuous(transform = "reverse") +
#'   coord_3d()
#'
#' @seealso \code{\link{zlim}} for a shorthand way to set z-axis limits,
#'   \code{\link{coord_3d}} for the 3D coordinate system
#' @family 3D scale functions
#' @export
scale_z_continuous <- function(name = waiver(), breaks = waiver(), minor_breaks = waiver(),
                               n.breaks = NULL, labels = waiver(), limits = NULL,
                               expand = waiver(), oob = scales::censor, na.value = NA_real_,
                               transform = "identity", guide = "none", ...) {

      scale_obj <- continuous_scale(
            aesthetics = "z",
            palette = identity,
            name = name,
            breaks = breaks,
            minor_breaks = minor_breaks,
            n.breaks = n.breaks,
            labels = labels,
            limits = limits,
            expand = expand,
            oob = oob,
            na.value = na.value,
            transform = transform,
            guide = guide,
            super = ScaleContinuousPosition,
            ...
      )
      return(scale_obj)
}

#' Discrete z-axis scale for 3D plots
#'
#' \code{scale_z_discrete} creates a discrete scale for the z aesthetic in 3D plots.
#' It works with categorical/factor data, positioning each level at integer coordinates
#' (1, 2, 3, ...) in 3D space.
#'
#' @param name The name of the scale, used as the axis title. Use \code{waiver()} for
#'   the default, or \code{NULL} to omit the title.
#' @param breaks One of:
#'   \itemize{
#'     \item \code{NULL} for no breaks
#'     \item \code{waiver()} for all breaks
#'     \item A character vector of breaks
#'     \item A function that takes the limits as input and returns breaks as output
#'   }
#' @param labels One of:
#'   \itemize{
#'     \item \code{NULL} for no labels
#'     \item \code{waiver()} for the default labels
#'     \item A character vector giving labels (must be same length as breaks)
#'     \item A function that takes the breaks as input and returns labels as output
#'   }
#' @param limits A character vector that defines possible values of the scale and their order.
#' @param expand A numeric vector of length two giving multiplicative and additive expansion constants.
#'   These constants ensure that data is placed some distance away from the axes. The defaults are
#'   to expand by 0.6 units on each side for discrete scales.
#' @param guide A function used to create a guide or its name. Since z-axis guides are not
#'   yet supported, this defaults to \code{"none"}.
#' @param na.translate Unlike continuous scales, discrete scales can easily show missing values,
#'   and do so by default. If you want to remove missing values from a discrete scale, specify
#'   \code{na.translate = FALSE}.
#' @param na.value If \code{na.translate = TRUE}, what aesthetic value should the missing values be
#'   displayed as? Does not apply to position scales where \code{NA} is always placed at the far right.
#' @param drop Should unused factor levels be omitted from the scale? The default, \code{TRUE}, uses
#'   the levels that appear in the data; \code{FALSE} uses all the levels in the factor.
#' @param ... Other arguments passed on to \code{discrete_scale()}.
#'
#' @return A ggplot2 scale object for the z aesthetic.
#'
#' @examples
#' library(ggplot2)
#'
#' # Basic usage to control order, breaks, labels, expansion, etc.
#' ggplot(mpg, aes(displ, cty, drv, color = drv)) +
#'       geom_point() +
#'       scale_z_discrete(limits = c("f", "r", "4"), # change default order
#'                        breaks = c("f", "r", "4"),
#'                        labels = c("front", "rear", "4-wheel"),
#'                        expand = expansion(.5)) +
#'       coord_3d()
#'
#' @seealso \code{\link{scale_z_continuous}} for continuous z-axis scaling,
#'   \code{\link{coord_3d}} for the 3D coordinate system
#' @family 3D scale functions
#' @export
scale_z_discrete <- function(name = waiver(), breaks = waiver(), labels = waiver(),
                             limits = NULL, expand = waiver(), guide = "none",
                             na.translate = TRUE, na.value = NA_real_, drop = TRUE,
                             ...) {


      # Create the scale object like ggplot2's scale_x_discrete
      sc <- discrete_scale(
            aesthetics = "z",
            palette = seq_len,  # Use seq_len like ggplot2
            name = name,
            breaks = breaks,
            labels = labels,
            limits = limits,
            expand = expand,
            guide = guide,
            na.translate = na.translate,
            na.value = na.value,
            drop = drop,
            super = ScaleDiscretePosition,
            ...
      )

      # Add the missing range_c component, using a template scale to copy the range_c structure
      template_scale <- scale_x_discrete()
      sc$range_c <- template_scale$range_c$clone()
      continuous.limits <- NULL
      sc$continuous_limits <- continuous.limits

      return(sc)
}

#' Set z-axis limits
#'
#' This is a shorthand for \code{scale_z_continuous(limits = c(min, max))}.
#' It's a convenient way to set the z-axis limits without specifying other
#' scale parameters.
#'
#' @param min,max The minimum and maximum values for the z-axis.
#' @param ... Additional arguments passed to \code{scale_z_continuous()}.
#'
#' @return A ggplot2 scale object for the z aesthetic with specified limits.
#'
#' @examples
#' library(ggplot2)
#'
#' # Set z-axis limits
#' ggplot(mtcars, aes(mpg, wt, z = qsec)) +
#'   geom_point() +
#'   zlim(15, 20) +
#'   coord_3d()
#'
#' # Equivalent to:
#' ggplot(mtcars, aes(mpg, wt, z = qsec)) +
#'   geom_point() +
#'   scale_z_continuous(limits = c(15, 20)) +
#'   coord_3d()
#'
#' @seealso \code{\link{scale_z_continuous}} for more control over z-axis scaling,
#'   \code{\link{xlim}}, \code{\link{ylim}} for x and y axis limits
#' @family 3D scale functions
#' @export
zlim <- function(min, max, ...) {
      scale_z_continuous(limits = c(min, max), ...)
}

# Geom -------------------------------------------------------------------------

GeomPoint3D <- ggproto("GeomPoint3D", GeomPoint,

                       extra_params = c("na.rm", "annotate"),

                       setup_data = function(data, params) {
                             setup_annotations(data, params)
                       },

                       draw_panel = function(data, panel_params, coord, na.rm = FALSE,
                                             annotate = NULL,
                                             sort_method = "painter",
                                             scale_depth = TRUE,
                                             ref_line_colour = NULL, ref_line_linewidth = 0.25, ref_line_linetype = NULL, ref_line_alpha = NULL,
                                             ref_point_colour = NULL, ref_point_fill = NULL, ref_point_alpha = NULL, ref_point_size = NULL, ref_point_stroke = NULL, ref_point_shape = NULL) {

                             validate_coord3d(coord)

                             # Seed annotation rows with columns the point pipeline expects
                             # (coord$transform calls points_to_circles which needs these)
                             if (".ann" %in% names(data) && any(!is.na(data$.ann) & data$.ann)) {
                                   ann_mask <- !is.na(data$.ann) & data$.ann
                                   if (!"element_type" %in% names(data)) data$element_type <- NA_character_
                                   if (!"ref_circle_radius" %in% names(data)) data$ref_circle_radius <- NA_real_
                                   if (!"ref_circle_vertices" %in% names(data)) data$ref_circle_vertices <- NA_integer_
                                   if (!"project_to_face" %in% names(data)) data$project_to_face <- NA_character_
                                   data$element_type[ann_mask] <- "raw_point"
                                   data$ref_circle_radius[ann_mask] <- 0
                                   data$ref_circle_vertices[ann_mask] <- 0L
                                   data$project_to_face[ann_mask] <- NA_character_
                             }

                             # Point-specific preprocessing
                             data$ref_circle_radius <- data$ref_circle_radius / 100
                             data <- prepare_annotations(data, panel_params)
                             sort_method <- match.arg(sort_method, c("auto", "pairwise", "painter"))
                             data$.sort_method <- sort_method
                             coords <- coord$transform(data, panel_params)

                             # Scale points, strokes, linewidths by depth
                             if(!"linewidth" %in% names(coords)) coords$linewidth <- 0.5
                             coords <- apply_depth_scaling(coords, scale_depth)

                             if (nrow(coords) == 0) {
                                   return(grid::nullGrob())
                             }

                             # Save annotation .prim values (resolve_point_aesthetics
                             # overwrites .prim based on element_type)
                             ann_prim <- NULL
                             if (".ann" %in% names(coords)) {
                                   ann_idx <- which(!is.na(coords$.ann) & coords$.ann)
                                   if (length(ann_idx) > 0) {
                                         ann_prim <- coords$.prim[ann_idx]
                                   }
                             }

                             # Resolve ref_* styling and assign .prim types
                             coords <- resolve_point_aesthetics(
                                   coords,
                                   ref_line_colour = ref_line_colour,
                                   ref_line_linewidth = ref_line_linewidth,
                                   ref_line_linetype = ref_line_linetype,
                                   ref_line_alpha = ref_line_alpha,
                                   ref_point_colour = ref_point_colour,
                                   ref_point_fill = ref_point_fill,
                                   ref_point_alpha = ref_point_alpha,
                                   ref_point_size = ref_point_size,
                                   ref_point_stroke = ref_point_stroke,
                                   ref_point_shape = ref_point_shape
                             )

                             # Restore annotation .prim values
                             if (!is.null(ann_prim)) {
                                   coords$.prim[ann_idx] <- ann_prim
                             }

                             # Apply annotate_3d styles (overrides resolve_point_aesthetics
                             # for annotation rows only)
                             coords <- apply_annotation_styles(coords)

                             render_mixed_grobs(coords)
                       }
)

# Ref element aesthetic resolution ---------------------------------------------

#' Resolve reference element aesthetics and assign .prim types
#'
#' Walks through the data and applies ref_* styling inheritance logic,
#' writing final visual property values into standard aesthetic columns.
#' Also maps element_type to .prim for the shared renderer.
#'
#' @param coords Transformed coordinate data with element_type column.
#' @param ref_line_colour,ref_line_linewidth,ref_line_linetype,ref_line_alpha
#'   Overrides for reference line styling.
#' @param ref_point_colour,ref_point_fill,ref_point_alpha,ref_point_size,ref_point_stroke,ref_point_shape
#'   Overrides for reference point/circle styling.
#' @return Data frame with resolved aesthetics and .prim column.
#' @keywords internal
#' @noRd
resolve_point_aesthetics <- function(coords,
                                     ref_line_colour = NULL,
                                     ref_line_linewidth = 0.25,
                                     ref_line_linetype = NULL,
                                     ref_line_alpha = NULL,
                                     ref_point_colour = NULL,
                                     ref_point_fill = NULL,
                                     ref_point_alpha = NULL,
                                     ref_point_size = NULL,
                                     ref_point_stroke = NULL,
                                     ref_point_shape = NULL) {

      if (!"element_type" %in% names(coords)) {
            # No element_type column — treat everything as raw points
            coords$.prim <- "point"
            return(coords)
      }

      # Extract base object IDs for grouping (everything before "__")
      coords$object_id <- sub("__.*", "", coords$group)

      # Detect shape complexity and alpha mapping per object
      obj_info <- coords[!duplicated(coords$object_id), ]
      is_complex <- setNames(
            is.numeric(obj_info$shape) & obj_info$shape %in% 21:25,
            obj_info$object_id
      )
      alpha_mapped <- setNames(
            !sapply(split(obj_info$alpha, obj_info$object_id),
                    function(a) all(a == 1, na.rm = TRUE)),
            obj_info$object_id
      )

      # Initialize .prim column
      coords$.prim <- NA_character_

      # Ensure linetype column exists (point data may not have it)
      if (!"linetype" %in% names(coords)) coords$linetype <- 1

      # Process each element type
      types <- unique(coords$element_type)

      for (et in types) {
            mask <- coords$element_type == et
            et_data <- coords[mask, ]
            oids <- et_data$object_id

            if (et == "raw_point") {
                  # Raw points keep their aesthetics as-is; no changes needed
                  coords$.prim[mask] <- "point"

            } else if (et == "ref_circle") {
                  # Ref circles are polygons — apply ref_point inheritance
                  complex <- is_complex[oids]
                  amapped <- alpha_mapped[oids]

                  coords$fill[mask] <- resolve_ref_fill(
                        ref_point_fill, et_data$fill, et_data$colour, complex)
                  coords$colour[mask] <- resolve_ref_colour(
                        ref_point_colour, et_data$colour, complex)
                  coords$alpha[mask] <- resolve_ref_alpha(
                        ref_point_alpha, et_data$alpha, amapped, default = 0.5)
                  coords$linetype[mask] <- 1

                  # Stroke/linewidth for circle border
                  avg_ds <- ave(et_data$depth_scale, et_data$object_id, FUN = mean)
                  coords$linewidth[mask] <- resolve_ref_circle_linewidth(
                        ref_point_stroke, et_data$stroke, complex, avg_ds)

                  coords$.prim[mask] <- "polygon"

            } else if (et == "ref_line") {
                  # Ref lines are segments — apply ref_line inheritance
                  amapped <- alpha_mapped[oids]

                  coords$colour[mask] <- ref_line_colour %||% et_data$colour
                  coords$linetype[mask] <- ref_line_linetype %||% 1
                  coords$alpha[mask] <- resolve_ref_alpha(
                        ref_line_alpha, et_data$alpha, amapped, default = 0.5)

                  # Linewidth with depth scaling
                  avg_ds <- ave(et_data$depth_scale, et_data$object_id, FUN = mean)
                  coords$linewidth[mask] <- ref_line_linewidth * avg_ds

                  coords$.prim[mask] <- "segment"

            } else if (et == "ref_point") {
                  # Ref points are points — apply ref_point inheritance
                  complex <- is_complex[oids]
                  amapped <- alpha_mapped[oids]

                  coords$colour[mask] <- ref_point_colour %||% et_data$colour
                  coords$fill[mask] <- resolve_ref_fill(
                        ref_point_fill, et_data$fill, et_data$colour, complex)
                  coords$alpha[mask] <- resolve_ref_alpha(
                        ref_point_alpha, et_data$alpha, amapped, default = 0.5)
                  coords$shape[mask] <- ref_point_shape %||% et_data$shape

                  # Size: ref_point_size scales by depth, otherwise 1/3 of original
                  if (!is.null(ref_point_size)) {
                        coords$size[mask] <- ref_point_size * et_data$depth_scale
                  } else {
                        coords$size[mask] <- et_data$size * (1/3)
                  }

                  coords$stroke[mask] <- ref_point_stroke %||% et_data$stroke

                  coords$.prim[mask] <- "point"
            }
      }

      coords$object_id <- NULL
      return(coords)
}


# Ref aesthetic helpers --------------------------------------------------------

resolve_ref_fill <- function(override, fill, colour, is_complex) {
      if (!is.null(override)) return(rep(override, length(fill)))
      ifelse(is_complex, fill, colour)
}

resolve_ref_colour <- function(override, colour, is_complex) {
      if (!is.null(override)) return(rep(override, length(colour)))
      ifelse(is_complex, colour, NA_character_)
}

resolve_ref_alpha <- function(override, alpha, alpha_mapped, default = 0.5) {
      if (!is.null(override)) return(rep(override, length(alpha)))
      ifelse(alpha_mapped, alpha, default)
}

resolve_ref_circle_linewidth <- function(override_stroke, stroke, is_complex, avg_depth_scale) {
      if (!is.null(override_stroke)) {
            return(override_stroke * avg_depth_scale)
      }
      ifelse(is_complex, (stroke %||% 0.5) * avg_depth_scale, 0)
}


# Stat -------------------------------------------------------------------------

StatPoint3D <- ggproto("StatPoint3D", Stat,
                       required_aes = c("x", "y", "z"),

                       compute_panel = function(data, scales, na.rm = FALSE,
                                                raw_points = TRUE,
                                                ref_lines = FALSE,
                                                ref_points = "circles",
                                                ref_faces = "zmin",
                                                ref_circle_radius = 0.02,
                                                ref_circle_vertices = 16) {

                             # Validate parameters
                             if (!is.logical(raw_points)) raw_points <- TRUE
                             if (!is.logical(ref_lines)) ref_lines <- FALSE

                             if (is.logical(ref_points)) {
                                   ref_points <- if(ref_points) "circles" else FALSE
                             }

                             # Handle empty data
                             if (nrow(data) == 0) return(data)

                             # Handle discrete scale conversion
                             if ("z" %in% names(data)) {
                                   data$z_raw <- data$z
                                   if (is.factor(data$z) || is.character(data$z)) {
                                         data$z <- as.numeric(as.factor(data$z))
                                   }
                             }

                             if ("x" %in% names(data)) {
                                   data$x_raw <- data$x
                                   if (is.factor(data$x) || is.character(data$x)) {
                                         data$x <- as.numeric(as.factor(data$x))
                                   }
                             }

                             if ("y" %in% names(data)) {
                                   data$y_raw <- data$y
                                   if (is.factor(data$y) || is.character(data$y)) {
                                         data$y <- as.numeric(as.factor(data$y))
                                   }
                             }

                             # Determine which faces to use for projections
                             valid_faces <- c("xmin", "xmax", "ymin", "ymax", "zmin", "zmax")
                             ref_faces <- intersect(ref_faces, valid_faces)

                             result <- generate_point_elements(data, raw_points,
                                                               ref_faces, ref_points, ref_lines,
                                                               ref_circle_radius, ref_circle_vertices)

                             return(result)
                       }
)


# Ref element generation -------------------------------------------------------

#' Generate reference elements with simplified approach
#'
#' @param data Original point data
#' @param raw_points Whether to include raw points
#' @param ref_faces Character vector of face names
#' @param ref_points Type of reference points: FALSE, "circles", or "points"
#' @param ref_lines Whether to include reference lines
#' @param ref_circle_radius Radius for circular reference points
#' @param ref_circle_vertices Number of vertices for circular reference points
#' @return Data frame with reference elements
#' @keywords internal
#' @noRd
generate_point_elements <- function(data, raw_points,
                                    ref_faces, ref_points, ref_lines,
                                    ref_circle_radius, ref_circle_vertices) {

      # Build result data frame
      result <- data.frame()
      id <- 1:nrow(data)

      # Add primary points if requested
      if(raw_points) {
            result <- data %>%
                  mutate(element_type = "raw_point",
                         group = paste0("point_raw_", id, "__grouped"),
                         project_to_face = NA_character_,
                         ref_circle_radius = NA_real_,
                         ref_circle_vertices = NA_integer_) %>%
                  bind_rows(result)
      }

      for(face in ref_faces){
            if(ref_points == "circles"){
                  result <- data %>%
                        mutate(element_type = "ref_circle",
                               group = paste0("circle_", face, "_", id, "__grouped"),
                               project_to_face = face,
                               ref_circle_radius = ref_circle_radius,
                               ref_circle_vertices = as.integer(ref_circle_vertices)) %>%
                        bind_rows(result)
            } else if(ref_points == "points") {
                  result <- data %>%
                        mutate(element_type = "ref_point",
                               group = paste0("point_", face, "_", id, "__grouped"),
                               project_to_face = face,
                               ref_circle_radius = NA_real_,
                               ref_circle_vertices = NA_integer_) %>%
                        bind_rows(result)
            }

            if(ref_lines){
                  result <- data %>%
                        mutate(element_type = "ref_line",
                               group = paste0("segment_", face, "_", id, "__grouped"),
                               project_to_face = NA_character_,
                               ref_circle_radius = NA_real_,
                               ref_circle_vertices = NA_integer_) %>%
                        bind_rows(mutate(., project_to_face = face)) %>%
                        bind_rows(result)
            }
      }

      return(result)
}



# User-facing functions --------------------------------------------------------

#' 3D scatter plot with 2D reference elements
#'
#' `geom_point_3d()` creates scatter plots in 3D space with automatic depth-based
#' size scaling. Points closer to the viewer appear larger, while points farther
#' away appear smaller, creating realistic perspective effects. Optionally adds
#' reference lines and points projecting to cube faces, to illustrate point locations
#' in 2D.
#'
#' StatPoint3D performs identity transformation (passes data through unchanged) while
#' properly handling discrete scales for 3D coordinate systems. It can optionally
#' generate reference lines and points projecting to cube faces.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. In addition to
#'   the standard point aesthetics, `geom_point_3d()` requires x, y, and z coordinates.
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use on the data. Defaults to `"point_3d"`.
#' @param geom The geometric object used to display the data. Defaults to `"point_3d"`.
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed on to [ggplot2::layer()].
#' @param na.rm If `FALSE`, missing values are removed with a warning.
#' @param show.legend Logical indicating whether this layer should be included in legends.
#' @param inherit.aes If `FALSE`, overrides the default aesthetics.
#' @inheritParams sort_params
#' @inheritParams depth_params
#' @param raw_points Logical indicating whether to show the original 3D points.
#'   Default is `TRUE`.
#' @param ref_lines Logical indicating whether to show reference lines projecting
#'   from points to cube faces. Default is `FALSE`.
#' @param ref_points Type of reference points to create. Options:
#'   \itemize{
#'     \item \code{FALSE}: No reference points (default)
#'     \item \code{TRUE or "circles"}: Circular reference points that project properly
#'     \item \code{"points"}: Single-point references (renders faster and supports non-circular shapes,
#'     but point shape is not 3D-transformed)
#'   }
#' @param ref_faces Character vector specifying which cube faces to project to.
#'   Valid face names are: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
#'   Default is "zmin" (bottom face). Multiple faces can be specified.
#' @param ref_circle_radius Radius for circular reference points as percentage of standardized coordinate space.
#' @param ref_circle_vertices Number of vertices for circular reference points (higher = smoother).
#' @param ref_line_color,ref_line_colour,ref_line_linewidth,ref_line_linetype,ref_line_alpha
#'   Styling overrides for reference lines. When NULL, inherits from point aesthetics or uses defaults.
#'   Both American (`ref_line_color`) and British (`ref_line_colour`) spellings are accepted.
#' @param ref_point_color,ref_point_colour,ref_point_fill,ref_point_size,ref_point_alpha,ref_point_stroke,ref_point_shape
#'   Styling overrides for reference points and circles. When NULL, inherits from raw point aesthetics
#'   with shape-aware logic (complex shapes 21-25 vs simple shapes 0-20/characters).
#'   Both American (`ref_point_color`) and British (`ref_point_colour`) spellings are accepted.
#'
#' @section Aesthetics:
#' `geom_point_3d()` supports all the same aesthetics as [ggplot2::geom_point()], plus `z`:
#' - **x**: X coordinate (required)
#' - **y**: Y coordinate (required)
#' - **z**: Z coordinate (required)
#' - `alpha`: Transparency
#' - `colour`: Point border color
#' - `fill`: Point fill color (for certain shapes)
#' - `shape`: Point shape
#' - `size`: Point size (gets depth-scaled when `scale_depth = TRUE`)
#' - `stroke`: Border width for shapes with borders
#'
#' @section StatPoint3D returns the following computed variables:
#' - `x_raw`, `y_raw`, `z_raw`: Original values before discrete-to-numeric conversion
#' - `element_type`: Type of element ("raw_point", "ref_point", "ref_line", "ref_circle")
#' - `segment_id`: ID linking the two endpoints of reference lines
#' - `ref_face`: Which face reference elements project to
#' - `ref_circle_radius`: Radius for circular reference points
#' - `ref_circle_vertices`: Number of vertices for circular reference points
#'
#' @section Depth Scaling:
#' The depth scaling uses an inverse relationship with distance, following the
#' mathematical relationship: `apparent_size = base_size * reference_distance / actual_distance`
#'
#' This creates realistic perspective where:
#' - Objects twice as far appear half as big
#' - Objects twice as close appear twice as big
#' - The center of the plot volume renders at exactly the user-specified size
#'
#' @section Point Rendering:
#' ggcube uses shape-aware rendering for improved stroke behavior:
#' - **Simple shapes (0-20, characters):** Use `size` for fontsize, `stroke` for border width, no fill
#' - **Complex shapes (21-25):** Use ggplot2's approach: `size + 0.5*stroke` for fontsize to prevent gaps
#' - Both size and stroke are depth-scaled when `scale_depth = TRUE`
#' - All shapes preserve stroke depth scaling and parameter control
#'
#' @section Reference Features:
#' Reference lines and points help visualize the 3D relationships by projecting
#' data points onto cube faces:
#' - **Reference lines**: Connect each 3D point to its projection on specified faces
#' - **Reference points**: Show the projected location on the faces
#' - **Reference circles**: Circular projections that appear as realistic 3D shadows
#' - **Depth sorting**: All elements (original points, reference lines, reference points)
#'   are automatically depth-sorted for proper 3D rendering
#'
#' @section Aesthetic Inheritance:
#' Reference elements intelligently inherit styling from raw points:
#' - **Shape-aware**: Complex shapes (21-25) with fill/colour are handled differently from simple shapes (0-20/characters)
#' - **Alpha detection**: If all points have alpha=1, uses 0.5 default for ref elements; otherwise inherits mapped alpha
#' - **Priority**: Explicit ref_* parameters > mapped aesthetics > smart defaults
#'
#' @examples
#' library(ggplot2)
#'
#' # Basic 3D scatter plot with depth scaling
#' ggplot(expand.grid(x = 1:5, y = 1:5, z = 1:5),
#'        aes(x, y, z, fill = z)) +
#'   geom_point_3d(size = 10, shape = 21, color = "white", stroke = .1) +
#'   coord_3d(pitch = 40, roll = 5, yaw = 0, dist = 1.5) +
#'   scale_fill_viridis_c()
#'
#' # Add circular reference points on 2D face panel
#' ggplot(mtcars, aes(mpg, wt, qsec)) +
#'   geom_point_3d(size = 3,
#'     ref_points = TRUE, ref_faces = c("ymax", "xmax")) +
#'   coord_3d()
#'
#' @seealso [geom_segment_3d()] for 3D segments, [geom_path_3d()] for 3D paths,
#'   [coord_3d()] for 3D coordinate systems.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_point_3d
#' @export
geom_point_3d <- function(mapping = NULL, data = NULL,
                          stat = StatPoint3D, position = "identity",
                          ...,
                          na.rm = FALSE,
                          sort_method = "painter",
                          scale_depth = TRUE,
                          raw_points = TRUE,
                          ref_lines = FALSE,
                          ref_points = FALSE,
                          ref_faces = "zmin",
                          ref_circle_radius = 1.5,
                          ref_circle_vertices = 16,
                          ref_line_color = NULL, ref_line_colour = NULL,
                          ref_line_linewidth = 0.25,
                          ref_line_linetype = NULL,
                          ref_line_alpha = NULL,
                          ref_point_color = NULL, ref_point_colour = NULL,
                          ref_point_fill = NULL,
                          ref_point_size = NULL,
                          ref_point_alpha = NULL,
                          ref_point_stroke = NULL,
                          ref_point_shape = NULL,
                          inherit.aes = TRUE,
                          show.legend = NA) {

      # Handle both American and British spellings
      if (!is.null(ref_line_colour) && is.null(ref_line_color)) {
            ref_line_color <- ref_line_colour
      }
      if (!is.null(ref_point_colour) && is.null(ref_point_color)) {
            ref_point_color <- ref_point_colour
      }

      layer(
            data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomPoint3D,
            position = position,  show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(
                  na.rm = na.rm,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  raw_points = raw_points,
                  ref_lines = ref_lines,
                  ref_points = ref_points,
                  ref_faces = ref_faces,
                  ref_circle_radius = ref_circle_radius,
                  ref_circle_vertices = ref_circle_vertices,
                  ref_line_color = ref_line_color,
                  ref_line_linewidth = ref_line_linewidth,
                  ref_line_linetype = ref_line_linetype,
                  ref_line_alpha = ref_line_alpha,
                  ref_point_color = ref_point_color,
                  ref_point_fill = ref_point_fill,
                  ref_point_size = ref_point_size,
                  ref_point_alpha = ref_point_alpha,
                  ref_point_stroke = ref_point_stroke,
                  ref_point_shape = ref_point_shape,
                  ...
            )
      )
}

#' @rdname geom_point_3d
#' @export
stat_point_3d <- function(mapping = NULL, data = NULL,
                          geom = GeomPoint3D, position = "identity",
                          ...,
                          na.rm = FALSE,
                          sort_method = "painter",
                          scale_depth = TRUE,
                          raw_points = TRUE,
                          ref_lines = FALSE,
                          ref_points = FALSE,
                          ref_faces = "zmin",
                          ref_circle_radius = 1.5,
                          ref_circle_vertices = 16,
                          ref_line_color = NULL, ref_line_colour = NULL,
                          ref_line_linewidth = 0.25,
                          ref_line_linetype = NULL,
                          ref_line_alpha = NULL,
                          ref_point_color = NULL, ref_point_colour = NULL,
                          ref_point_fill = NULL,
                          ref_point_size = NULL,
                          ref_point_alpha = NULL,
                          ref_point_stroke = NULL,
                          ref_point_shape = NULL,
                          inherit.aes = TRUE,
                          show.legend = NA) {

      # Handle both American and British spellings
      if (!is.null(ref_line_colour) && is.null(ref_line_color)) {
            ref_line_color <- ref_line_colour
      }
      if (!is.null(ref_point_colour) && is.null(ref_point_color)) {
            ref_point_color <- ref_point_colour
      }

      layer(
            data = data, mapping = mapping, stat = StatPoint3D, geom = get_proto(geom, "geom"),
            position = position,  show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(
                  na.rm = na.rm,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  raw_points = raw_points,
                  ref_lines = ref_lines,
                  ref_points = ref_points,
                  ref_faces = ref_faces,
                  ref_circle_radius = ref_circle_radius,
                  ref_circle_vertices = ref_circle_vertices,
                  ref_line_color = ref_line_color,
                  ref_line_linewidth = ref_line_linewidth,
                  ref_line_linetype = ref_line_linetype,
                  ref_line_alpha = ref_line_alpha,
                  ref_point_color = ref_point_color,
                  ref_point_fill = ref_point_fill,
                  ref_point_size = ref_point_size,
                  ref_point_alpha = ref_point_alpha,
                  ref_point_stroke = ref_point_stroke,
                  ref_point_shape = ref_point_shape,
                  ...
            )
      )
}

#' Sphere surface points
#'
#' A toy data set representing points on a sphere surface,
#' useful for testing hull functions
#'
#' @format ## `sphere_points`
#' A data frame with 200 rows and 3 columns (x, y, and z)
"sphere_points"


#' Volcano terrain data
#'
#' A terrain data set representing elevational variation over
#' an x-y grid.
#'
#' @format ## `mountain`
#' A data frame with 1290 rows and 3 columns (x, y, and z)
#' @source derived from the `datasets::volcano` data set
"mountain"

#' Create an interactive drag-to-rotate ggcube
#'
#' Creates an interactive HTML widget that you can rotate by dragging.
#' Pre-renders a set of ggcube plots across a 1D or 2D grid of rotation angles
#' and packages them as an \pkg{htmlwidgets} widget. The widget displays in the
#' RStudio Viewer, embeds in R Markdown and Quarto documents, works in Shiny
#' (see [orbit3dOutput()]), and can be written to a self-contained HTML
#' file with `file =`. When viewing it, navigate through the views by dragging
#' the image or using the keyboard arrows.
#'
#' @section Specifying rotation:
#' Each of \code{pitch}, \code{roll}, and \code{yaw} is given either as a
#' single value (held fixed) or as a length-2 range \code{c(from, to)} that
#' is a variable, draggable axis. The number of ranges determines the
#' interaction:
#' \itemize{
#'   \item One range gives a 1D orbit: horizontal drag (or left/right
#'     keyboard arrows) vary that one parameter across its range.
#'   \item Two ranges give a 2D orbit: horizontal drag (or left/right
#'     keyboard arrows) vary one parameter while vertical drag/arrows vary
#'     the other.
#' }
#' At least one range is required. Supplying three ranges is an error.
#'
#' For a 2D orbit, the horizontal (x) and vertical (y) axes are assigned
#' from whichever parameters are ranged, by priority. The x axis takes the
#' highest-priority ranged parameter in the order \code{yaw}, \code{pitch},
#' \code{roll}; the y axis takes the highest-priority remaining ranged
#' parameter in the order \code{roll}, \code{pitch}. In the common case
#' \code{yaw = c(0, 360), roll = c(...)} this puts yaw on horizontal drag and
#' roll on vertical drag.
#'
#' The range direction follows the order the values are written: for example
#' \code{yaw = c(360, 0)} rotates the opposite way from \code{yaw = c(0, 360)}.
#'
#' @section Looping:
#' When a variable axis returns to its starting view (for example a full
#' \code{yaw = c(0, 360)} turn), that axis loops seamlessly: dragging past
#' either end wraps around. This is detected per axis by comparing the first
#' and last angle modulo 360, so a full-turn axis wraps while a partial range
#' (such as \code{roll = c(-90, 90)}) clamps. \code{loop} can force this on or
#' off for all axes at once.
#'
#' @section Keyboard:
#' Once the widget has focus, the arrow keys step one frame: left/right move
#' the horizontal axis, and (in a 2D orbit) up/down move the vertical axis.
#'
#' @inheritParams animate_3d
#' @param pitch,roll,yaw Rotation angles in degrees. Each is either a single
#'   value (held fixed) or a length-2 range \code{c(from, to)} that becomes a
#'   draggable axis. For `yaw`, using a larger value for `from` than `to` gives
#'   more natural results. Values specified here replace the ones specified in
#'   `coord_3d()`. See Details.
#' @param n Grid resolution. For a 1D orbit, a single number giving the
#'   frame count. For a 2D orbit, a length-2 vector giving the counts for
#'   the horizontal and vertical screen axes respectively (in that order, x then y).
#'   Its length must match the number of ranged parameters. Default is 36.
#' @param loop Controls seamless wraparound. \code{NULL} (default)
#'   auto-detects looping independently for each axis by checking whether that
#'   axis returns to its starting view. \code{TRUE} or \code{FALSE} force
#'   wrapping on or off for all axes.
#' @param start Initial viewing angle(s) the orbit opens at, given in
#'   degrees as a named vector matching the variable axes, e.g.
#'   \code{start = c(yaw = 30)} or \code{start = c(yaw = 30, roll = 45)}. For a
#'   1D orbit a bare scalar (e.g. \code{start = 30}) is also accepted. The
#'   orbit opens at the frame nearest each given angle. Any variable axis not
#'   named in \code{start} opens at the midpoint of its range, as does every
#'   axis when \code{start = NULL} (the default). Naming an axis that is
#'   fixed rather than variable throas an error.
#' @param max_frames Safety cap on the total number of rendered frames
#'   (\code{prod(n)}). A 2D grid can become very large; if the requested grid
#'   exceeds this cap the function errors rather than rendering. Raise it to
#'   allow larger grids. Default is 500.
#' @param file Optional path to write a self-contained HTML file (via
#'   [htmlwidgets::saveWidget()]). If `NULL` (default), no file is written and
#'   the widget is simply returned for display or embedding.
#'
#' @return A `orbit_3d` htmlwidget. Printed at the console it opens in the
#'   RStudio Viewer or browser; in a knitr document it embeds automatically;
#'   in Shiny use [orbit3dOutput()] and [renderOrbit3d()].
#'
#' @examplesIf requireNamespace("base64enc", quietly = TRUE) && requireNamespace("htmlwidgets", quietly = TRUE)
#' \donttest{
#' p <- ggplot(mountain, aes(x, y, z, fill = z, color = z)) +
#'   geom_surface_3d(linewidth = .25) +
#'   coord_3d(light = light(anchor = "camera",direction = c(1, 1, 0)),
#'            ratio = c(1.5, 2, 1)) +
#'   scale_fill_viridis_c() + scale_color_viridis_c() +
#'   theme_void()
#'
#' # 1D drag-to-rotate turntable (small n to minimize runtime)
#' orbit_3d(p, yaw = c(360, 0), n = 12)
#'
#' # 2D orbit: drag horizontally to spin (yaw), vertically to tilt (roll)
#' # (note you could increase `cores` to speed up compute time)
#' orbit_3d(p, yaw = c(360, 0), roll = c(-90, 0), n = c(20, 10),
#'             width = 1000, height = 1000, cores = 1)
#'
#' # Save to a chosen path
#' orbit_3d(p, yaw = c(360, 0), n = 12,
#'             file = file.path(tempdir(), "surface.html"))
#' }
#'
#' @seealso [animate_3d()] for GIF/MP4 output.
#'   [orbit_3d-shiny] for including orbits in  Shiny apps.
#' @export
orbit_3d <- function(plot,
                     pitch = NULL,
                     roll = NULL,
                     yaw = NULL,
                     n = 36,
                     loop = NULL,
                     start = NULL,
                     max_frames = 500,
                     width = 480,
                     height = 480,
                     res = 96,
                     cores = 1,
                     device = "png",
                     progress = interactive(),
                     file = NULL) {

      # --- Validate input plot ---
      if (!inherits(plot, "gg")) {
            stop("`plot` must be a ggplot object.", call. = FALSE)
      }
      coord <- plot$coordinates
      if (!inherits(coord, "Coord3D")) {
            stop("`plot` must use coord_3d(). ",
                 "Add `+ coord_3d()` to your plot before making a orbit.",
                 call. = FALSE)
      }

      # --- Validate cores / loop ---
      cores <- as.integer(cores)
      if (cores < 1) stop("`cores` must be at least 1.", call. = FALSE)
      if (!is.null(loop) && !(is.logical(loop) && length(loop) == 1)) {
            stop("`loop` must be NULL, TRUE, or FALSE.", call. = FALSE)
      }

      # --- Classify rotation parameters and assign drag axes ---
      # Each parameter is absent (NULL), fixed (length 1), or a variable range
      # (length 2). The set of ranged parameters determines dimensionality and
      # the x/y axis assignment.
      spec <- resolve_orbit_axes(pitch = pitch, roll = roll, yaw = yaw)
      ndim <- length(spec$axes)

      # --- Validate n against dimensionality, and enforce the frame cap ---
      n <- as.integer(n)
      if (any(is.na(n)) || any(n < 1)) {
            stop("`n` must be one or more positive integers.", call. = FALSE)
      }
      if (length(n) != ndim) {
            stop(sprintf(
                  "`n` has length %d but %d rotation %s ranged (needs length %d). %s",
                  length(n), ndim, if (ndim == 1) "parameter is" else "parameters are",
                  ndim,
                  "Give one count per ranged parameter (x then y)."),
                 call. = FALSE)
      }
      total <- prod(n)
      if (total > max_frames) {
            stop(sprintf(
                  paste0("This orbit would render %d frames (%s), exceeding ",
                         "max_frames = %d. Reduce `n`, or raise `max_frames` to ",
                         "allow a larger grid."),
                  total, paste(n, collapse = " x "), max_frames),
                 call. = FALSE)
      }

      # --- Build the (flattened) parameter grid ---
      # Each variable axis gets its own 1D range, with per-axis loop detection
      # and duplicate-endpoint handling; the axes are then crossed row-major
      # (x varying fastest) into a flat param_seqs data.frame. Fixed and absent
      # parameters are held constant using the supplied value or the coord's.
      grid <- build_param_grid(
            coord = coord,
            spec = spec,
            n = n,
            loop = loop,
            start = start
      )
      param_seqs <- grid$param_seqs
      dims <- grid$dims
      loop_flags <- grid$loop_flags
      start_cell <- grid$start

      # --- Render frames to PNG files ---
      frames <- render_3d_frames(
            plot = plot,
            coord = coord,
            param_seqs = param_seqs,
            width = width,
            height = height,
            res = res,
            device = device,
            cores = cores,
            progress = progress,
            prefix = "ggcube_flip_"
      )

      message("Assembling orbit...")

      # --- Encode frames and build the htmlwidget ---
      # The frame/parameter logic above is unchanged; only the output layer
      # differs. Frames are base64-encoded into data URIs and handed to the
      # widget along with the grid topology (`dims`), per-axis wraparound
      # (`loop_flags`), and the initial cell (`start`). htmlwidgets then
      # handles all rendering contexts (RStudio Viewer, R Markdown / Quarto
      # via its own knit_print, and Shiny) and, for `file=`, self-contained
      # HTML via saveWidget().
      if (!requireNamespace("base64enc", quietly = TRUE)) {
            stop("The base64enc package is required to build a orbit. ",
                 "Install it with install.packages('base64enc').",
                 call. = FALSE)
      }
      data_uris <- vapply(frames$files, function(f) {
            paste0("data:image/png;base64,", base64enc::base64encode(f))
      }, character(1), USE.NAMES = FALSE)

      # Frames are now inlined as data URIs; the temp PNGs are no longer needed.
      unlink(frames$tmpdir, recursive = TRUE)

      x <- list(
            frames = as.list(data_uris),
            dims = as.list(as.integer(dims)),
            loop_flags = as.list(as.logical(loop_flags)),
            start = as.list(as.integer(start_cell))
      )

      widget <- htmlwidgets::createWidget(
            name = "orbit_3d",
            x = x,
            width = frames$width,
            height = frames$height,
            package = "ggcube",
            sizingPolicy = htmlwidgets::sizingPolicy(
                  viewer.padding = 0,
                  browser.fill = FALSE,
                  knitr.figure = FALSE
            )
      )

      # If a file path is requested, write a self-contained HTML file. This
      # preserves the single-portable-file output of earlier versions.
      #
      # saveWidget(selfcontained = TRUE) has a long-standing quirk: it inlines
      # dependencies into the HTML but then fails to remove the temporary
      # "<name>_files" directory when the output path is not the current working
      # directory (htmlwidgets issue #296), leaving a stray sidecar folder next
      # to an otherwise self-contained file. Running saveWidget from within the
      # output directory, with a bare filename, makes its cleanup path resolve
      # correctly and yields a true single file.
      if (!is.null(file)) {
            file <- normalizePath(file, mustWork = FALSE)
            out_dir <- dirname(file)
            out_base <- basename(file)
            if (!dir.exists(out_dir)) {
                  dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)
            }
            old_wd <- setwd(out_dir)
            on.exit(setwd(old_wd), add = TRUE)
            htmlwidgets::saveWidget(widget, out_base, selfcontained = TRUE)
      }

      widget
}


# Axis resolution & grid construction -------------------------------------

#' Classify rotation parameters and assign drag axes
#'
#' Each of pitch/roll/yaw is absent (NULL), fixed (length 1), or a variable range
#' (length 2). Ranged parameters become draggable axes; the x axis takes the
#' highest-priority ranged parameter in order yaw > pitch > roll, and the y
#' axis takes the highest-priority remaining ranged parameter in order
#' roll > pitch. Fixed values are recorded to be held constant.
#'
#' @param pitch,roll,yaw User-supplied rotation arguments.
#' @return A list with \code{axes} (character vector of ranged parameter
#'   names, in x-then-y order), \code{ranges} (named list of the length-2
#'   ranges for those axes), and \code{fixed} (named list of scalar values for
#'   parameters the user pinned to a single value).
#' @keywords internal
#' @noRd
resolve_orbit_axes <- function(pitch = NULL, roll = NULL, yaw = NULL) {
      args <- list(pitch = pitch, roll = roll, yaw = yaw)

      ranged <- character(0)
      fixed <- list()
      for (nm in names(args)) {
            v <- args[[nm]]
            if (is.null(v)) next
            if (length(v) == 1) {
                  fixed[[nm]] <- v
            } else if (length(v) == 2) {
                  ranged <- c(ranged, nm)
            } else {
                  stop(sprintf(
                        paste0("`%s` has length %d. Each rotation parameter must ",
                               "be a single value (held fixed) or a length-2 ",
                               "range c(from, to)."),
                        nm, length(v)), call. = FALSE)
            }
      }

      if (length(ranged) == 0) {
            stop(paste0("A orbit needs at least one rotation parameter given ",
                        "as a range c(from, to). For example, yaw = c(0, 360)."),
                 call. = FALSE)
      }
      if (length(ranged) > 2) {
            stop(paste0("Three ranged rotation parameters were given. A orbit ",
                        "supports at most two draggable axes; for motion through ",
                        "three angles use animate_3d()."),
                 call. = FALSE)
      }

      # Assign axes by priority.
      x_priority <- c("yaw", "pitch", "roll")
      y_priority <- c("roll", "pitch")

      x_axis <- x_priority[x_priority %in% ranged][1]
      remaining <- setdiff(ranged, x_axis)
      y_axis <- if (length(remaining) > 0) {
            y_priority[y_priority %in% remaining][1]
      } else {
            NA_character_
      }

      axes <- c(x_axis, if (!is.na(y_axis)) y_axis)
      ranges <- args[axes]

      list(axes = axes, ranges = ranges, fixed = fixed)
}


#' Build the flattened parameter grid for a orbit
#'
#' Produces one 1D range per variable axis (with per-axis loop detection and
#' duplicate-endpoint handling), then crosses the axes row-major with the x
#' axis varying fastest. Fixed and absent parameters are held constant.
#'
#' @param coord The plot's Coord3D object (source of held-constant defaults).
#' @param spec Output of resolve_orbit_axes().
#' @param n Integer resolution vector, one entry per axis, in x-then-y order.
#' @param loop \code{NULL} to auto-detect per axis, or a single logical to
#'   force all axes.
#' @return A list with \code{param_seqs} (flat data.frame of pitch/roll/yaw,
#'   one row per frame, row-major with x fastest), \code{dims} (integer vector
#'   of axis lengths after endpoint handling), and \code{loop_flags} (logical
#'   vector, one per axis).
#' @keywords internal
#' @noRd
build_param_grid <- function(coord, spec, n, loop, start = NULL) {
      axes <- spec$axes

      # Build each axis's 1D range of angle values, applying loop detection and
      # dropping the duplicate endpoint when the axis wraps.
      axis_values <- vector("list", length(axes))
      loop_flags <- logical(length(axes))
      dims <- integer(length(axes))

      for (i in seq_along(axes)) {
            rng <- spec$ranges[[axes[i]]]
            count <- n[i]

            # Determine whether this axis loops.
            is_loop <- if (is.null(loop)) {
                  angles_return_home(rng[1], rng[2])
            } else {
                  isTRUE(loop)
            }

            if (is_loop && count > 1L) {
                  # Sample count + 1 inclusive points, drop the duplicate end,
                  # leaving `count` unique points spanning the half-open range.
                  vals <- seq(rng[1], rng[2], length.out = count + 1L)
                  vals <- vals[-length(vals)]
            } else {
                  vals <- seq(rng[1], rng[2], length.out = count)
                  is_loop <- is_loop && count > 1L
            }

            axis_values[[i]] <- vals
            loop_flags[i] <- is_loop
            dims[i] <- length(vals)
      }

      # Cross the axes row-major with x (axis 1) varying fastest.
      if (length(axes) == 1) {
            grid_idx <- data.frame(x = seq_along(axis_values[[1]]))
      } else {
            # expand.grid varies its first argument fastest, which is what we
            # want for x.
            grid_idx <- expand.grid(
                  x = seq_along(axis_values[[1]]),
                  y = seq_along(axis_values[[2]])
            )
      }

      total <- nrow(grid_idx)

      # Start every parameter at its held-constant value, then overwrite the
      # axis parameters with their variable values.
      const <- function(nm) {
            if (!is.null(spec$fixed[[nm]])) spec$fixed[[nm]] else coord[[nm]]
      }
      out <- data.frame(
            pitch = rep(const("pitch"), total),
            roll  = rep(const("roll"),  total),
            yaw   = rep(const("yaw"),   total)
      )

      out[[axes[1]]] <- axis_values[[1]][grid_idx$x]
      if (length(axes) == 2) {
            out[[axes[2]]] <- axis_values[[2]][grid_idx$y]
      }

      rownames(out) <- NULL

      # Determine the 0-based starting cell per axis from the requested start
      # angles (or the midpoint when unspecified).
      start_cell <- orbit_start_cell(axis_values, axes, start)

      list(param_seqs = out, dims = dims, loop_flags = loop_flags,
           start = start_cell)
}


#' Choose the orbit's initial frame cell from requested start angles
#'
#' Returns a 0-based index per variable axis for the frame the player shows first.
#' \code{start} is an optional set of viewing angles in degrees: a named vector
#' whose names are variable axis names (e.g. \code{c(yaw = 30, roll = 45)}), or, for
#' a 1D orbit, a bare scalar applied to the single variable axis. For each variable
#' axis named in \code{start}, the orbit opens at the frame nearest that angle
#' (nearest by circular distance, modulo 360). Any variable axis not named opens at
#' the index midpoint of its range, as does every axis when \code{start} is
#' \code{NULL}.
#'
#' @param axis_values List of per-axis sampled angle vectors.
#' @param axes Character vector of variable axis names (x then y).
#' @param start \code{NULL}, a named numeric vector of angles, or (1D only) a
#'   bare scalar.
#' @return Integer vector of 0-based cell indices, one per axis.
#' @keywords internal
#' @noRd
orbit_start_cell <- function(axis_values, axes, start = NULL) {
      midpoint <- function() {
            vapply(axis_values, function(v) as.integer(floor(length(v) / 2)),
                   integer(1))
      }

      if (is.null(start)) return(midpoint())

      # Normalize `start` to a named list of angles keyed by axis.
      if (is.null(names(start))) {
            # Unnamed: only valid as a single value for a 1D orbit.
            if (length(axes) == 1 && length(start) == 1) {
                  targets <- stats::setNames(list(start[[1]]), axes)
            } else {
                  stop("`start` must be a named vector of angles (e.g. ",
                       "c(yaw = 30)); a bare value is only allowed for a 1D ",
                       "orbit.", call. = FALSE)
            }
      } else {
            bad <- setdiff(names(start), axes)
            if (length(bad)) {
                  stop(sprintf(
                        "`start` names %s, which %s fixed rather than variable. Variable %s: %s.",
                        paste(bad, collapse = ", "),
                        if (length(bad) == 1) "is" else "are",
                        if (length(axes) == 1) "axis is" else "axes are",
                        paste(axes, collapse = ", ")),
                       call. = FALSE)
            }
            targets <- as.list(start)
      }

      circ_dist <- function(a, b) abs(((a - b) + 180) %% 360 - 180)

      out <- integer(length(axes))
      for (i in seq_along(axes)) {
            ax <- axes[i]
            vals <- axis_values[[i]]
            if (is.null(targets[[ax]]) || length(vals) == 1) {
                  out[i] <- as.integer(floor(length(vals) / 2))  # midpoint
            } else {
                  out[i] <- which.min(circ_dist(vals, targets[[ax]])) - 1L
            }
      }
      out
}


#' Do two angles (degrees) describe a full-circle return?
#'
#' Compares two angles modulo 360, so 0 and 360 (or 0 and 720) count as
#' returning home.
#'
#' @param a,b Angles in degrees.
#' @param tol Angular tolerance in degrees.
#' @return A single logical.
#' @keywords internal
#' @noRd
angles_return_home <- function(a, b, tol = 1e-6) {
      d <- abs(((a - b) + 180) %% 360 - 180)
      d <= tol
}


# Shiny bindings ----------------------------------------------------------

#' Shiny bindings for orbit_3d
#'
#' Output and render functions for using [orbit_3d()] within Shiny
#' applications and interactive R Markdown documents.
#'
#' @param outputId Output variable to read from.
#' @param width,height Must be a valid CSS unit (like `"100%"`, `"400px"`,
#'   `"auto"`) or a number, which will be coerced to a string and have `"px"`
#'   appended.
#' @param expr An expression that generates a orbit (a call to
#'   [orbit_3d()]).
#' @param env The environment in which to evaluate `expr`.
#' @param quoted Is `expr` a quoted expression (with `quote()`)? This is useful
#'   if you want to save an expression in a variable.
#'
#' @examplesIf interactive() && requireNamespace("shiny", quietly = TRUE) && requireNamespace("base64enc", quietly = TRUE)
#' library(shiny)
#'
#' p <- ggplot(mountain, aes(x, y, z, fill = z, color = z)) +
#'   geom_surface_3d(linewidth = .25) +
#'   coord_3d(light = light(anchor = "camera", direction = c(1, 1, 0)),
#'            ratio = c(1.5, 2, 1)) +
#'   scale_fill_viridis_c() + scale_color_viridis_c() +
#'   theme_void()
#'
#' ui <- fluidPage(
#'   titlePanel("Interactive ggcube"),
#'   orbit3dOutput("fb", width = "500px", height = "500px")
#' )
#'
#' server <- function(input, output, session) {
#'   output$fb <- renderOrbit3d({
#'     orbit_3d(p, yaw = c(360, 0), n = 12)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @name orbit_3d-shiny
#' @export
orbit3dOutput <- function(outputId, width = "100%", height = "400px") {
      htmlwidgets::shinyWidgetOutput(outputId, "orbit_3d", width, height,
                                     package = "ggcube")
}

#' @rdname orbit_3d-shiny
#' @export
renderOrbit3d <- function(expr, env = parent.frame(), quoted = FALSE) {
      if (!quoted) expr <- substitute(expr)
      htmlwidgets::shinyRenderWidget(expr, orbit3dOutput, env,
                                     quoted = TRUE)
}

# Tests for orbit_3d front-end logic: axis assignment, validation, grid
# construction, and per-axis loop detection. These exercise the pure helpers
# and do not render (grob/render behavior is covered elsewhere / skipped on
# CRAN).

# A minimal stand-in for a Coord3D carrying default rotation angles, used as
# the source of held-constant values in build_param_grid().
fake_coord <- function(pitch = 10, roll = 20, yaw = 30) {
      list(pitch = pitch, roll = roll, yaw = yaw)
}

# --- angles_return_home ------------------------------------------------------

test_that("angles_return_home detects full-circle returns modulo 360", {
      expect_true(angles_return_home(0, 360))
      expect_true(angles_return_home(0, 720))
      expect_true(angles_return_home(0, 0))
      expect_true(angles_return_home(10, 370))
      expect_false(angles_return_home(0, 90))
      expect_false(angles_return_home(-45, 45))
})

# --- resolve_orbit_axes: classification and axis assignment ---------------

test_that("a single ranged parameter yields a 1D orbit on that axis", {
      expect_equal(resolve_orbit_axes(yaw = c(0, 360))$axes, "yaw")
      expect_equal(resolve_orbit_axes(roll = c(0, 360))$axes, "roll")
      expect_equal(resolve_orbit_axes(pitch = c(0, 90))$axes, "pitch")
})

test_that("two ranges assign x and y by priority", {
      # yaw is top x priority; roll is top y priority.
      expect_equal(resolve_orbit_axes(yaw = c(0, 360), roll = c(-45, 45))$axes,
                   c("yaw", "roll"))
      # yaw on x, pitch falls to y.
      expect_equal(resolve_orbit_axes(yaw = c(0, 360), pitch = c(0, 90))$axes,
                   c("yaw", "pitch"))
      # no yaw: pitch takes x (next x priority), roll takes y.
      expect_equal(resolve_orbit_axes(pitch = c(0, 90), roll = c(-45, 45))$axes,
                   c("pitch", "roll"))
})

test_that("fixed parameters are recorded and not treated as axes", {
      spec <- resolve_orbit_axes(yaw = c(0, 360), pitch = 15, roll = c(-45, 45))
      expect_equal(spec$axes, c("yaw", "roll"))
      expect_equal(spec$fixed$pitch, 15)
})

test_that("resolve_orbit_axes errors on invalid range counts", {
      # No ranges at all.
      expect_error(resolve_orbit_axes(pitch = 5),
                   "at least one rotation parameter")
      # A length-3 parameter is not a valid range.
      expect_error(resolve_orbit_axes(yaw = c(0, 180, 360)),
                   "length 3")
      # Three ranged parameters.
      expect_error(
            resolve_orbit_axes(yaw = c(0, 360), pitch = c(0, 90),
                                  roll = c(-1, 1)),
            "at most two")
})

# --- build_param_grid: dimensions, ordering, constants -----------------------

test_that("1D grid has the requested number of rows and holds others constant", {
      spec <- resolve_orbit_axes(yaw = c(0, 90))
      grid <- build_param_grid(fake_coord(pitch = 10, roll = 20, yaw = 30),
                               spec, n = 5, loop = NULL)
      expect_equal(nrow(grid$param_seqs), 5)
      expect_equal(grid$dims, 5L)
      # Non-swept params held at coord defaults.
      expect_true(all(grid$param_seqs$pitch == 10))
      expect_true(all(grid$param_seqs$roll == 20))
      # Swept param spans the range inclusively (non-looping).
      expect_equal(grid$param_seqs$yaw, seq(0, 90, length.out = 5))
})

test_that("2D grid is the row-major outer product with x varying fastest", {
      spec <- resolve_orbit_axes(yaw = c(0, 100), roll = c(-10, 10))
      grid <- build_param_grid(fake_coord(), spec, n = c(3, 2), loop = NULL)
      # 3 x 2 = 6 rows.
      expect_equal(nrow(grid$param_seqs), 6)
      expect_equal(grid$dims, c(3L, 2L))
      # x (yaw) varies fastest: first 3 rows sweep yaw at the first roll value.
      yaw_vals <- seq(0, 100, length.out = 3)
      roll_vals <- seq(-10, 10, length.out = 2)
      expect_equal(grid$param_seqs$yaw, rep(yaw_vals, times = 2))
      expect_equal(grid$param_seqs$roll, rep(roll_vals, each = 3))
})

test_that("fixed parameter overrides the coord default in the grid", {
      spec <- resolve_orbit_axes(yaw = c(0, 90), pitch = 42)
      grid <- build_param_grid(fake_coord(pitch = 10), spec, n = 4, loop = NULL)
      expect_true(all(grid$param_seqs$pitch == 42))
})

# --- Per-axis loop detection and endpoint handling ---------------------------

test_that("a full-turn axis loops and drops its duplicate endpoint", {
      spec <- resolve_orbit_axes(yaw = c(0, 360))
      grid <- build_param_grid(fake_coord(), spec, n = 4, loop = NULL)
      expect_true(grid$loop_flags[1])
      # 4 unique frames over [0, 360): 0, 90, 180, 270 (no 360 duplicate).
      expect_equal(grid$param_seqs$yaw, c(0, 90, 180, 270))
      expect_equal(grid$dims, 4L)
})

test_that("a partial-sweep axis clamps and samples inclusively", {
      spec <- resolve_orbit_axes(roll = c(-90, 90))
      grid <- build_param_grid(fake_coord(), spec, n = 5, loop = NULL)
      expect_false(grid$loop_flags[1])
      expect_equal(grid$param_seqs$roll, seq(-90, 90, length.out = 5))
})

test_that("per-axis auto-detect mixes loop and clamp in 2D", {
      # yaw full turn (loops) x roll partial (clamps).
      spec <- resolve_orbit_axes(yaw = c(0, 360), roll = c(-45, 45))
      grid <- build_param_grid(fake_coord(), spec, n = c(6, 4), loop = NULL)
      expect_true(grid$loop_flags[1])
      expect_false(grid$loop_flags[2])
})

test_that("loop = FALSE forces clamp even on a full turn", {
      spec <- resolve_orbit_axes(yaw = c(0, 360))
      grid <- build_param_grid(fake_coord(), spec, n = 4, loop = FALSE)
      expect_false(grid$loop_flags[1])
      # Inclusive sampling keeps the endpoint (360 present).
      expect_equal(grid$param_seqs$yaw, seq(0, 360, length.out = 4))
})

test_that("loop = TRUE forces wrap and endpoint drop", {
      spec <- resolve_orbit_axes(roll = c(-90, 90))
      grid <- build_param_grid(fake_coord(), spec, n = 4, loop = TRUE)
      expect_true(grid$loop_flags[1])
      expect_equal(nrow(grid$param_seqs), 4)
      # Endpoint dropped: 4 samples over [-90, 90).
      expect_equal(grid$param_seqs$roll, seq(-90, 90, length.out = 5)[-5])
})

# --- Starting-frame selection ------------------------------------------------

test_that("start defaults to the sweep midpoint per axis", {
      spec <- resolve_orbit_axes(yaw = c(0, 350))
      grid <- build_param_grid(fake_coord(), spec, n = 36, loop = FALSE,
                               start = NULL)
      expect_equal(grid$start, 18L)  # floor(36 / 2)

      spec2 <- resolve_orbit_axes(yaw = c(0, 100), roll = c(-90, 90))
      grid2 <- build_param_grid(fake_coord(), spec2, n = c(6, 4), loop = FALSE,
                                start = NULL)
      expect_equal(grid2$start, c(3L, 2L))  # floor(6/2), floor(4/2)
})

test_that("start snaps to the frame nearest a named angle", {
      spec <- resolve_orbit_axes(yaw = c(0, 350))
      # 36 frames every 10 degrees; yaw = 30 -> index 3 (0-based).
      grid <- build_param_grid(fake_coord(), spec, n = 36, loop = TRUE,
                               start = c(yaw = 30))
      expect_equal(grid$start, 3L)
      # Nearest-frame snapping: 32 still rounds to the 30 frame.
      grid2 <- build_param_grid(fake_coord(), spec, n = 36, loop = TRUE,
                                start = c(yaw = 32))
      expect_equal(grid2$start, 3L)
})

test_that("a bare scalar start is accepted for a 1D orbit", {
      spec <- resolve_orbit_axes(yaw = c(0, 350))
      grid <- build_param_grid(fake_coord(), spec, n = 36, loop = TRUE,
                               start = 30)
      expect_equal(grid$start, 3L)
})

test_that("an unnamed swept axis in start falls back to its midpoint", {
      spec <- resolve_orbit_axes(yaw = c(0, 100), roll = c(-90, 90))
      grid <- build_param_grid(fake_coord(), spec, n = c(6, 4), loop = FALSE,
                               start = c(yaw = 0))
      expect_equal(grid$start[1], 0L)  # yaw = 0 -> first frame
      expect_equal(grid$start[2], 2L)  # roll unspecified -> midpoint
})

test_that("start naming a non-swept axis is an error", {
      spec <- resolve_orbit_axes(yaw = c(0, 350))
      expect_error(
            build_param_grid(fake_coord(), spec, n = 36, loop = TRUE,
                             start = c(roll = 10)),
            "fixed rather than variable")
})

test_that("a bare scalar start errors for a 2D orbit", {
      spec <- resolve_orbit_axes(yaw = c(0, 100), roll = c(-90, 90))
      expect_error(
            build_param_grid(fake_coord(), spec, n = c(6, 4), loop = FALSE,
                             start = 30),
            "named vector")
})

StatVoxel3D <- ggproto("StatVoxel3D", Stat,
                     required_aes = c("x", "y", "z"),
                     default_aes = aes(group = after_stat(group)),

                     compute_panel = function(data, scales, na.rm = FALSE,
                                              width = 1.0, faces = "all",
                                              light = NULL,
                                              cull_backfaces = NULL) {

                           # Remove missing values if requested
                           if (na.rm) {
                                 data <- data[complete.cases(data[c("x", "y", "z")]), ]
                           }

                           # Check we have enough data
                           if (nrow(data) < 1) {
                                 stop("stat_voxel_3d requires at least 1 point")
                           }

                           # Generate numeric positions before calculating spacing
                           data <- convert_to_numeric(data)

                           # Calculate voxel spacing using resolution (works for sparse 3D grids)
                           x_spacing <- resolution(data$x, zero = FALSE)
                           y_spacing <- resolution(data$y, zero = FALSE)
                           z_spacing <- resolution(data$z, zero = FALSE)

                           # Validate and process faces parameter
                           selected_faces <- select_faces(faces)

                           # Create voxels
                           voxel_faces <- create_voxels(data, x_spacing, y_spacing, z_spacing, width, selected_faces)

                           voxel_faces %>%
                                 average_aesthetics() %>%
                                 mutate(cull_backfaces = cull_backfaces) %>%
                                 attach_light(light) %>%
                                 return()
                     }
)

#' Create voxel faces from 3D sparse data
#'
#' @param data Data frame with x, y, z columns
#' @param x_spacing Grid spacing in x direction
#' @param y_spacing Grid spacing in y direction
#' @param z_spacing Grid spacing in z direction
#' @param width Width factor (1.0 = full grid spacing)
#' @param selected_faces Character vector of face names to render
#' @return Data frame with voxel face vertices
#' @keywords internal
#' @noRd
create_voxels <- function(data, x_spacing, y_spacing, z_spacing, width, selected_faces) {

      voxel_width_x <- x_spacing * width
      voxel_width_y <- y_spacing * width
      voxel_width_z <- z_spacing * width
      half_x <- voxel_width_x / 2
      half_y <- voxel_width_y / 2
      half_z <- voxel_width_z / 2

      all_faces <- list()

      # Debug: check inputs
      if (nrow(data) == 0) {
            warning("create_voxels: No data points provided")
            return(data.frame())
      }

      if (length(selected_faces) == 0) {
            warning("create_voxels: No faces selected")
            return(data.frame())
      }

      for (i in 1:nrow(data)) {
            point <- data[i, ]
            cx <- point$x  # Center x
            cy <- point$y  # Center y
            cz <- point$z  # Center z

            l <- cx - half_x
            r <- cx + half_x
            f <- cy - half_y
            k <- cy + half_y
            t <- cz + half_z
            b <- cz - half_z

            # Define the 8 corners of the voxel
            corners <- list(
                  c(l, k, b),  # 1: left-back-bottom
                  c(r, k, b),  # 2: right-back
                  c(r, f, b),  # 3: right-front
                  c(l, f, b),  # 4: left-front
                  c(l, k, t),  # 5: left-back-top
                  c(r, k, t),  # 6: right-back
                  c(r, f, t),  # 7: right-front
                  c(l, f, t)  # 8: left-front
            )

            # Define faces using corner indices (order: ccw from outside)
            face_definitions <- list(
                  zmin = c(1, 2, 3, 4),  # Bottom
                  zmax = c(5, 8, 7, 6),  # Top
                  xmin = c(1, 4, 8, 5),  # Left
                  xmax = c(3, 2, 6, 7),  # Right
                  ymin = c(1, 5, 6, 2),  # Back
                  ymax = c(4, 3, 7, 8)   # Front
            )

            # Create requested faces
            for (face_name in selected_faces) {
                  if (face_name %in% names(face_definitions)) {
                        corner_indices <- face_definitions[[face_name]]

                        # Create 2-level hierarchical group ID: voxel_id__face_type
                        hierarchical_group <- paste0("voxel", i, "__", face_name)

                        # Create face vertices with hierarchical grouping
                        face_vertices <- data.frame(
                              x = sapply(corner_indices, function(idx) corners[[idx]][1]),
                              y = sapply(corner_indices, function(idx) corners[[idx]][2]),
                              z = sapply(corner_indices, function(idx) corners[[idx]][3]),
                              z_raw = point$z_raw,
                              group = hierarchical_group,
                              voxel_id = i,
                              face_type = face_name
                        )

                        # Preserve all non-coordinate columns
                        non_coord_cols <- setdiff(names(point), names(face_vertices))
                        for (col_name in non_coord_cols) {
                              face_vertices[[col_name]] <- rep(point[[col_name]], 4)
                        }

                        all_faces[[length(all_faces) + 1]] <- face_vertices
                  }
            }
      }

      if (length(all_faces) == 0) {
            return(data.frame())
      }

      # Combine all faces
      result <- do.call(rbind, all_faces)
      rownames(result) <- NULL

      return(result)
}

# Validate and process faces parameter
# (Helper used by multiple stats)
select_faces <- function(faces) {
      valid_faces <- c("xmin", "xmax", "ymin", "ymax", "zmin", "zmax")
      if (length(faces) == 1 && faces == "all") {
            selected_faces <- valid_faces
      } else if (length(faces) == 1 && faces == "none") {
            selected_faces <- character(0)
      } else {
            invalid_faces <- setdiff(faces, valid_faces)
            if (length(invalid_faces) > 0) {
                  stop("Invalid face names: ", paste(invalid_faces, collapse = ", "),
                       ". Valid faces are: ", paste(valid_faces, collapse = ", "))
            }
            selected_faces <- faces
      }
      return(selected_faces)
}

# Convert categorical position columns to numeric positions
# (Helper used by multiple stats)
convert_to_numeric <- function(data) {
      data$z_raw <- data$z # (stash a copy of original z values first)
      if (is.factor(data$x) || is.character(data$x)) {
            data$x <- as.numeric(as.factor(data$x))
      }
      if (is.factor(data$y) || is.character(data$y)) {
            data$y <- as.numeric(as.factor(data$y))
      }
      if (is.factor(data$z) || is.character(data$z)) {
            data$z <- as.numeric(as.factor(data$z))
      }
      data
}


#' 3D voxel visualization from sparse 3D data
#'
#' Creates 3D voxel visualizations from sparse 3D point data.
#' Each data point becomes a fixed-size cube centered on its coordinates.
#' Useful for volumetric data and 3D pixel art.
#'
#' Note that voxel geometries sometimes require pairwise depth sorting for correct rendering.
#' This is the default for smaller data sets, but not for larger data sets due to compute speed;
#' in those cases you may wish to manually specify `sort_method = "pairwise"`.
#'
#' @param mapping Set of aesthetic mappings created by [aes()].
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use on the data. Defaults to `StatVoxel3D`.
#' @param geom The geometric object used to display the data. Defaults to `GeomPolygon3D`.
#'
#' @inheritParams col_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' Voxel 3D requires the following aesthetics:
#' - **x**: X coordinate (voxel center position)
#' - **y**: Y coordinate (voxel center position)
#' - **z**: Z coordinate (voxel center position)
#'
#' @section Computed variables:
#' - `normal_x`, `normal_y`, `normal_z`: Face normal components
#' - `voxel_id`: Sequential voxel number
#' - `face_type`: Face name ("zmax", "xmin", etc.)
#'
#' @examples
#' # Sparse 3D voxel data
#' voxel_data <- data.frame(
#'   x = round(rnorm(100, 0, 2)),
#'   y = round(rnorm(100, 0, 2)),
#'   z = round(rnorm(100, 0, 2))
#' )
#'
#' p <- ggplot(voxel_data, aes(x, y, z)) + coord_3d()
#'
#' # Basic 3D voxel plot
#' p + geom_voxel_3d(fill = "steelblue")
#'
#' # With aesthetic fill
#' p + stat_voxel_3d(aes(fill = z)) +
#'   scale_fill_viridis_c() +
#'   guides(fill = guide_colorbar_3d())
#'
#' # Show only some voxel faces
#' p + geom_voxel_3d(fill = "steelblue",
#'                   faces = c("zmax", "ymin", "xmin"))
#'
#' @seealso [stat_col_3d()] for variable-height columns, [stat_surface_3d()] for smooth surfaces,
#'   [coord_3d()] for 3D coordinate systems, [light()] for lighting specifications.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_voxel_3d
#' @export
geom_voxel_3d <- function(mapping = NULL, data = NULL,
                          stat = StatVoxel3D,
                          position = "identity",
                          ...,
                          width = 1.0, faces = "all",
                          light = NULL,
                          cull_backfaces = TRUE, sort_method = NULL,
                          scale_depth = TRUE, force_convex = FALSE,
                          na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomPolygon3D,
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, width = width, faces = faces, light = light,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}

#' @rdname geom_voxel_3d
#' @export
stat_voxel_3d <- function(mapping = NULL, data = NULL,
                          geom = GeomPolygon3D,
                          position = "identity",
                          ...,
                          width = 1.0, faces = "all",
                          light = NULL,
                          cull_backfaces = TRUE, sort_method = NULL,
                          scale_depth = TRUE, force_convex = FALSE,
                          na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = StatVoxel3D, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(na.rm = na.rm, width = width, faces = faces, light = light,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          ...)
      )
}

# ggprotos ---------------

StatSmooth3D <- ggproto("StatSmooth3D", Stat,
                        required_aes = c("x", "y", "z"),
                        default_aes = aes(x = after_stat(x),
                                          y = after_stat(y),
                                          z = after_stat(z),
                                          fill = after_stat(fitted)),

                        extra_params = c("na.rm", "points", "residuals"),

                        compute_group = function(data, scales, method = "loess", formula = NULL,
                                                 method.args = list(), cull_backfaces = FALSE,
                                                 xlim = NULL, ylim = NULL,
                                                 n = NULL, grid = NULL, direction = NULL, trim = NULL,
                                                 light = NULL, na.rm = FALSE, domain = "chull",
                                                 se = FALSE, level = 0.95, se.fill = NULL, se.colour = NULL,
                                                 se.alpha = 0.5, se.linewidth = NULL,
                                                 points = FALSE, residuals = FALSE) {

                              # Remove missing values if requested
                              if (na.rm) {
                                    complete_cases <- complete.cases(data[c("x", "y", "z")])
                                    data <- data[complete_cases, ]
                              }

                              # Check we have enough data
                              if (nrow(data) < 4) {
                                    stop("stat_smooth_3d requires at least 4 complete cases")
                              }

                              # Validate method
                              valid_methods <- c("loess", "lm", "glm", "gam")
                              if (!method %in% valid_methods) {
                                    stop("method must be one of: ", paste(valid_methods, collapse = ", "))
                              }

                              # Get model specification
                              model_func <- switch(method,
                                                   lm = lm_model,
                                                   glm = glm_model,
                                                   loess = loess_model,
                                                   gam = gam_model,
                                                   stop("invalid model"))

                              # Handle formula defaults
                              if (is.null(formula)) {
                                    model_spec <- model_func()
                                    if("default_formula" %in% names(model_spec)) {
                                          formula <- model_spec$default_formula
                                    } else {
                                          formula <- z ~ x + y
                                    }
                              }

                              # Determine confidence levels to compute
                              z_score <- qnorm(1 - (1 - level)/2)
                              if(se) {
                                    surface_types <- c("upper", "fitted", "lower")
                              } else {
                                    surface_types <- "fitted"
                              }

                              # Generate point grid and expand to tile polygons
                              if(is.null(xlim)) xlim <- range(data$x, na.rm = TRUE)
                              if(is.null(ylim)) ylim <- range(data$y, na.rm = TRUE)
                              if(is.null(trim)) trim <- TRUE
                              points_grid <- make_point_grid(grid, n, direction, xlim, ylim, trim)
                              points_grid$z <- 0 # placeholder for points_to_tiles
                              points_grid$group <- data$group[1]
                              polys <- points_to_tiles(points_grid,
                                                       method = "grid",
                                                       grid_type = grid %||% "rectangle",
                                                       group_prefix = "surface__tile")
                              polys$group <- paste0(polys$group, "::grp", data$group[1])

                              # Clip to hull if applicable (before prediction, since
                              # clipping changes vertex positions and creates new ones)
                              if(domain == "chull") polys <- clip_polys_to_chull(polys, data)

                              # Fit model and predict at polygon vertex positions
                              predictions <- fit_and_predict(data, polys, method, formula,
                                                             method.args, se = se, level = level)
                              polys$fitted <- predictions$fitted
                              polys$se <- predictions$se

                              # Create surfaces for each level
                              surfaces <- surface_types %>%
                                    lapply(function(surface_type, data = polys) {
                                          data$level <- factor(surface_type,
                                                               levels = c("upper", "fitted", "lower"),
                                                               labels = c(paste0("upper ", level*100, "% CI"),
                                                                          "fitted",
                                                                          paste0("lower ", level*100, "% CI")))
                                          data$group <- paste0(data$group, "-", surface_type)
                                          data$z <- switch(surface_type,
                                                           fitted = data$fitted,
                                                           lower = data$fitted - z_score * data$se,
                                                           upper = data$fitted + z_score * data$se)

                                          # Apply confidence band styling
                                          if(surface_type != "fitted"){
                                                if (!is.null(se.fill)) {
                                                      data$fill <- se.fill
                                                }
                                                if (!is.null(se.colour)) {
                                                      data$colour <- se.colour
                                                }
                                                if (!is.null(se.linewidth)) {
                                                      data$linewidth <- se.linewidth
                                                }
                                                data$alpha <- se.alpha
                                          }

                                          return(data)
                                    }) %>%
                                    bind_rows()

                              # Add computed variables and light info
                              surfaces <- surfaces %>%
                                    compute_surface_vars() %>%
                                    average_aesthetics() %>%
                                    mutate(cull_backfaces = cull_backfaces) %>%
                                    attach_light(light)

                              # Optionally append data points
                              if (points) {
                                    pt_rows <- make_point_rows(
                                          data,
                                          group_prefix = paste0("smooth_data__grp", data$group[1])
                                    )

                                    # Ensure computed variables exist so after_stat() mappings
                                    # don't produce NA (e.g. fill = after_stat(fitted))
                                    if (!"fitted" %in% names(pt_rows)) pt_rows$fitted <- pt_rows$z
                                    if (!"se" %in% names(pt_rows)) pt_rows$se <- 0
                                    if (!"level" %in% names(pt_rows)) pt_rows$level <- "fitted"

                                    # Drop PANEL to prevent NA conflicts in bind_rows
                                    # (ggplot2 re-adds it after compute_group returns)
                                    pt_rows$PANEL <- NULL

                                    surfaces <- safe_bind_rows(surfaces, pt_rows)
                              }

                              # Optionally append residual lines connecting points to surface
                              if (residuals) {
                                    data_fitted <- fit_and_predict(data, data, method, formula,
                                                                   method.args, se = FALSE)

                                    seg_rows <- make_residual_segment_rows(
                                          data,
                                          fitted = data_fitted$fitted,
                                          group_prefix = paste0("smooth_resid__grp", data$group[1])
                                    )

                                    if (!"fitted" %in% names(seg_rows)) seg_rows$fitted <- seg_rows$z
                                    if (!"se" %in% names(seg_rows)) seg_rows$se <- 0
                                    if (!"level" %in% names(seg_rows)) seg_rows$level <- "fitted"

                                    surfaces <- safe_bind_rows(surfaces, seg_rows)
                              }

                              return(surfaces)

                        }
)


# helpers ---------------

clip_polys_to_chull <- function(d, point_data){

      hull_ind <- grDevices::chull(point_data[, c("x", "y")])
      hull <- as.matrix(point_data[hull_ind, c("x", "y")])

      d <- d %>%
            group_by(group) %>%
            reframe(poly = bind_cols(polyclip::polyclip(A = list(x = x, y = y),
                                                        B = list(x = hull[,1], y = hull[,2])))) %>%
            mutate(.vertex_order = 1:n()) %>%
            ungroup()

      xy <- d[[2]]
      colnames(xy) <- c("x", "y")
      d <- d %>% select(group) %>% bind_cols(xy, .)
      d <- d %>%
            group_by(group) %>%
            mutate(.vertex_order = 1:n()) %>%
            ungroup()
}

#' Fit and predict with 3D smoothing models
#'
#' @param data Data frame with x, y, z columns
#' @param new_data Data frame with x, y columns for prediction
#' @param method Smoothing method ("loess", "lm", "glm", "gam")
#' @param formula Model formula (NULL for default)
#' @param method.args Additional arguments for fitting function
#' @param se Logical, whether to compute standard errors
#' @param level Confidence level (not used here, passed to create_confidence_surfaces)
#' @return List with $fitted and optionally $se vectors
#' @keywords internal
#' @noRd
fit_and_predict <- function(data, new_data, method, formula, method.args,
                            se = FALSE, level = 0.95){

      model <- switch(method,
                      lm = lm_model(),
                      glm = glm_model(),
                      loess = loess_model(),
                      gam = gam_model(),
                      stop("invalid model"))

      # Formula precedence: user > method default > global default
      if (is.null(formula)) {
            if("default_formula" %in% names(model)) {
                  formula <- model$default_formula
            } else {
                  formula <- z ~ x + y
            }
      }

      # Default arguments for fitting
      default_args <- list(data = data, formula = formula)

      # Merge with user arguments (user args take precedence)
      args <- modifyList(default_args, method.args)

      # Fit model
      tryCatch({
            fit <- do.call(model$fit_function, args)
      }, error = function(e) {
            stop("Error fitting model: ", e$message)
      })

      # Predict with or without standard errors
      tryCatch({
            predictions <- model$predict_function(fit, new_data)
      }, error = function(e) {
            stop("Error predicting fitted values: ", e$message)
      })

      # Validate predictions
      if (!is.data.frame(predictions) || is.null(predictions$fitted)) {
            stop("Model prediction returned invalid structure")
      }

      if (length(predictions$fitted) != nrow(new_data)) {
            stop("Model prediction returned wrong number of values")
      }

      if (!is.numeric(predictions$fitted)) {
            stop("Model prediction must return numeric values")
      }

      predictions
}

lm_model <- function(){
      list(
            fit_function = lm,
            predict_function = function(fit, data){
                  pred <- predict(fit, data, se.fit = TRUE)
                  data.frame(fitted = pred$fit,
                             se = pred$se.fit)
            }
      )
}

glm_model <- function(){
      list(
            fit_function = glm,
            predict_function = function(fit, data){
                  pred <- predict(fit, data, type = "response", se.fit = TRUE)
                  data.frame(fitted = pred$fit,
                             se = pred$se.fit)
            }
      )
}

loess_model <- function(){
      list(
            fit_function = loess,
            predict_function = function(fit, data){
                  pred <- predict(fit, data, se = TRUE)
                  data.frame(fitted = as.vector(pred$fit),
                             se = as.vector(pred$se.fit))
            }
      )
}

gam_model <- function(){
      list(
            fit_function = function(...) {
                  if (!requireNamespace("mgcv", quietly = TRUE)) {
                        stop("GAM method requires the 'mgcv' package. Install with: install.packages('mgcv')")
                  }
                  mgcv::gam(...)
            },
            predict_function = function(fit, data){
                  pred <- predict(fit, data, se.fit = TRUE)
                  data.frame(fitted = pred$fit,
                             se = pred$se.fit)
            },
            default_formula = z ~ s(x) + s(y)
      )
}

# exported layer functions ------------------------------

#' 3D surface from smoothed conditional means
#'
#' A 3D version of `ggplot2::geom_smooth()`.
#' Creates surfaces by fitting smoothing models to scattered (x,y,z) data points.
#' The fitted statistical model is evaluated on a regular grid and rendered as a
#' 3D surface with optional standard error surfaces.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. This stat
#'   requires `x`, `y`, and `z` aesthetics from the input data. By default, fill is
#'   mapped to `after_stat(fitted)`.
#' @param data The data to be displayed in this layer. Must contain x, y, z columns.
#' @param stat The statistical transformation to use on the data. Defaults to `StatSmooth3D`.
#' @param geom The geometric object used to display the data. Defaults to `GeomPolygon3D`.
#'
#' @param method Smoothing method to use. Currently supported:
#'   \itemize{
#'     \item \code{"loess"} (default): Local polynomial regression
#'     \item \code{"lm"}: Linear model
#'     \item \code{"glm"}: Generalized linear model
#'     \item \code{"gam"}: Generalized additive model (requires \code{mgcv} package)
#'   }
#' @param formula Model formula. If `NULL` (default), uses method-appropriate defaults:
#'   `z ~ x + y` for lm and glm, `z ~ s(x) + s(y)` for gam, auto for loess.
#' @param method.args List of additional arguments passed to the fitting function.
#'   For loess, this might include `span` or `degree`. For lm, this might include `weights`.
#'   For glm, this might include `family` (defaults to `gaussian()`). For gam, this might
#'   include smoothing parameters or basis specifications.
#' @param domain Character indicating the x-y domain over which to visualize the surface.
#'   The default, `"bbox"`, shows predictions over the full rectangular bounding box of
#'   the predictors.The alternative, `"chull"`, shows predictions only within the convex
#'   hull of the input data, which prevents extrapolation into unoccupied corners of predictor space.
#' @param xlim,ylim Numeric vectors of length 2 giving the range for prediction grid.
#'   If `NULL` (default), uses the exact data range with no extrapolation.
#' @param se Logical indicating whether to display confidence interval bands around
#'   the smooth; if `TRUE`, these are rendered as additional surfaces; they inherit
#'   aesthetics from the primary smooth layer unless otherwise specified.
#'   Defaults to `FALSE`.
#' @param level Level of confidence interval to use (0.95 by default).
#' @param se_fill Fill colour for confidence interval bands. If `NULL`, inherits from
#'   the main surface `fill` aesthetic.
#' @param se_colour,se_color Color for confidence interval band borders. If `NULL`,
#'   inherits from the main surface `color` aesthetic.
#' @param se_alpha Alpha transparency for confidence interval bands. Defaults to 0.5.
#' @param se_linewidth Line width for confidence interval band borders. If `NULL`,
#'   inherits from the main surface `linewidth` aesthetic.
#' @param points Logical indicating whether to overlay the original data points
#'   on the fitted surface. Points are depth-sorted together with the surface
#'   polygons for proper 3D rendering. Point styling is controlled via the
#'   `point_*` parameters; mapped aesthetics from the layer are not inherited
#'   by annotation points. Default is `FALSE`.
#' @param point_colour,point_color Color for data points. Defaults to `"black"`.
#' @param point_fill Fill color for data points (only relevant for shapes 21-25).
#'   Defaults to `NA` (transparent).
#' @param point_size Size of data points.
#' @param point_shape Shape of data points.
#' @param point_alpha Alpha transparency for data points.
#' @param point_stroke Stroke width for data points.
#' @param residuals Logical indicating whether to draw residual lines connecting
#'   data points to the fitted surface. Default is `FALSE`. Residual styling is
#'   controlled via the `residual_*` parameters. Note: residual lines may render
#'   incorrectly when combined with `se = TRUE`, as the lines intersect the
#'   confidence interval surfaces and cannot be split at intersection points.
#' @param residual_colour,residual_color Color for residual lines.
#' @param residual_linewidth Line width for residual lines.
#' @param residual_linetype Line type for residual lines.
#' @param residual_alpha Alpha transparency for residual lines.
#'
#' @inheritParams grid_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' `stat_smooth_3d()` requires the following aesthetics from input data:
#' - **x**: X coordinate
#' - **y**: Y coordinate
#' - **z**: Z coordinate (response variable to be smoothed)
#'
#' @inheritSection surface_computed_vars Computed variables
#'
#' @section Computed variables specific to StatSmooth3D:
#' - `level`: Type of surface ("fitted", "upper CI", or "lower CI" for confidence bands)
#' - `fitted`: Smoothed predictions (same as `z` when `level == "fitted"`)
#' - `se`: Standard errors of the fitted values (available when `se = TRUE`)
#'
#' @examples
#' # Generate scattered 3D data
#' set.seed(123)
#' d <- data.frame(
#'   x = runif(100, -1, 3),
#'   y = runif(100, -3, 3)
#' )
#' d$z <- abs(1 + d$x^2 - d$y^2 + rnorm(100, 0, 1))
#'
#' # Base plot
#' p <- ggplot(d, aes(x, y, z)) +
#'   coord_3d(light = NULL) +
#'   scale_fill_viridis_c()
#'
#' # Basic smooth surface with default loess model
#' p + geom_smooth_3d()
#'
#' \donttest{
#'
#' # Show data points
#' p + geom_smooth_3d(points = TRUE)
#'
#' # Show data points with residual lines
#' p + geom_smooth_3d(points = TRUE, residuals = TRUE,
#'       point_color = "red", alpha = .8)
#'
#' # Linear model surface with 90% confidence intervals
#' p + geom_smooth_3d(aes(fill = after_stat(level)),
#'       method = "lm", color = "black", se = TRUE,
#'       level = 0.99, se_alpha = .7, n = 10) +
#'       scale_fill_manual(values = c("red", "darkorchid4", "steelblue"))
#'
#' # Linear model surface with custom model formula
#' p + geom_smooth_3d(method = "lm", n = 10,
#'       formula = z ~ poly(x, 2) + poly(y, 2) + x:y)
#'
#' # Loess with custom span parameter, and lighting effects
#' p + geom_smooth_3d(
#'       method = "loess", method.args = list(span = 0.3),
#'       fill = "steelblue", color = "white", n = 20,
#'       light = light(direction = c(0, -1, 0), color = FALSE))
#'
#' # GLM with gamma family and log link
#' p + geom_smooth_3d(
#'       method = "glm", n = 10,
#'       method.args = list(family = Gamma(link = "log")),
#'       formula = z ~ poly(x, 2) + poly(y, 2))
#'
#' # Visualize uncertainty with computed "standard error" variable
#' p + geom_smooth_3d(aes(fill = after_stat(se * 2))) +
#'   scale_fill_viridis_c()
#'
#' # Extend surface beyond training data range (explicit extrapolation)
#' p + geom_smooth_3d(method = "lm", xlim = c(-5, 5), ylim = c(-5, 5))
#'
#' # Clip surface to predictor convex hull
#' # to prevent extrapolation into corner areas
#' p + geom_smooth_3d(method = "lm", domain = "chull")
#'
#' # Specify alternative grid geometry
#' p + geom_smooth_3d(grid = "right1", n = 30, direction = "y")
#'
#' # Separate fits for data subgroups
#' ggplot(mtcars, aes(wt, mpg, qsec, fill = factor(cyl))) +
#'   geom_smooth_3d(method = "lm", alpha = .7,
#'     xlim = c(0, 5), ylim = c(0, 40)) + # specify shared domain
#'   coord_3d() + theme_light()
#'
#' }
#'
#' @seealso [stat_surface_3d()] for surfaces from existing grid data,
#'   [stat_function_3d()] for mathematical function surfaces.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_smooth_3d
#' @export
geom_smooth_3d <- function(mapping = NULL, data = NULL, stat = StatSmooth3D,
                           position = "identity",
                           ...,
                           method = "loess",
                           formula = NULL,
                           method.args = list(),
                           xlim = NULL,
                           ylim = NULL,
                           n = NULL, grid = NULL, direction = NULL, trim = NULL,
                           domain = c("bbox", "chull"),
                           se = FALSE,
                           level = 0.95,
                           se_fill = NULL,
                           se_colour = NULL,
                           se_color = NULL,
                           se_alpha = 0.5,
                           se_linewidth = NULL,
                           points = FALSE,
                           point_colour = "black", point_color = NULL,
                           point_fill = NA,
                           point_size = 1.5,
                           point_shape = 19,
                           point_alpha = 1,
                           point_stroke = 0.5,
                           residuals = FALSE,
                           residual_colour = "black", residual_color = NULL,
                           residual_linewidth = 0.5,
                           residual_linetype = 1,
                           residual_alpha = 1,
                           light = NULL,
                           cull_backfaces = FALSE, sort_method = NULL,
                           force_convex = TRUE, scale_depth = TRUE,
                           na.rm = FALSE,
                           show.legend = NA,
                           inherit.aes = TRUE) {

      # Resolve colour spellings (American -> British)
      se_colour <- se_colour %||% se_color
      if (!is.null(point_color)) point_colour <- point_color
      if (!is.null(residual_color)) residual_colour <- residual_color

      domain <- match.arg(domain)

      # Convert underscore param names to periods for layer internals
      # (avoids ggplot2 aesthetic name standardization on stat params)
      layer(
            geom = GeomPolygon3D, mapping = mapping, data = data, stat = get_proto(stat, "stat"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(method = method, formula = formula, method.args = method.args,
                          xlim = xlim, ylim = ylim, domain = domain, n = n, se = se, level = level,
                          grid = grid, direction = direction, trim = trim,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          se.fill = se_fill, se.colour = se_colour, se.alpha = se_alpha,
                          se.linewidth = se_linewidth,
                          points = points,
                          point.colour = point_colour,
                          point.fill = point_fill, point.size = point_size,
                          point.shape = point_shape, point.alpha = point_alpha,
                          point.stroke = point_stroke,
                          residuals = residuals,
                          residual.colour = residual_colour,
                          residual.linewidth = residual_linewidth,
                          residual.linetype = residual_linetype,
                          residual.alpha = residual_alpha,
                          light = light, na.rm = na.rm, ...)
      )
}

#' @rdname geom_smooth_3d
#' @export
stat_smooth_3d <- function(mapping = NULL, data = NULL,
                           geom = GeomPolygon3D,
                           position = "identity",
                           ...,
                           method = "loess",
                           formula = NULL,
                           method.args = list(),
                           xlim = NULL,
                           ylim = NULL,
                           n = NULL, grid = NULL, direction = NULL, trim = NULL,
                           domain = c("bbox", "chull"),
                           se = FALSE,
                           level = 0.95,
                           se_fill = NULL,
                           se_colour = NULL,
                           se_color = NULL,
                           se_alpha = 0.5,
                           se_linewidth = NULL,
                           points = FALSE,
                           point_colour = "black", point_color = NULL,
                           point_fill = NA,
                           point_size = 1.5,
                           point_shape = 19,
                           point_alpha = 1,
                           point_stroke = 0.5,
                           residuals = FALSE,
                           residual_colour = "black", residual_color = NULL,
                           residual_linewidth = 0.5,
                           residual_linetype = 1,
                           residual_alpha = 1,
                           light = NULL,
                           cull_backfaces = FALSE, sort_method = NULL,
                           force_convex = TRUE, scale_depth = TRUE,
                           na.rm = FALSE,
                           show.legend = NA,
                           inherit.aes = TRUE) {

      # Resolve colour spellings (American -> British)
      se_colour <- se_colour %||% se_color
      if (!is.null(point_color)) point_colour <- point_color
      if (!is.null(residual_color)) residual_colour <- residual_color

      domain <- match.arg(domain)

      # Convert underscore param names to periods for layer internals
      layer(
            stat = StatSmooth3D, data = data, mapping = mapping, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(method = method, formula = formula, method.args = method.args,
                          xlim = xlim, ylim = ylim, domain = domain, n = n, se = se, level = level,
                          grid = grid, direction = direction, trim = trim,
                          force_convex = force_convex, cull_backfaces = cull_backfaces,
                          sort_method = sort_method, scale_depth = scale_depth,
                          se.fill = se_fill, se.colour = se_colour, se.alpha = se_alpha,
                          se.linewidth = se_linewidth,
                          points = points,
                          point.colour = point_colour,
                          point.fill = point_fill, point.size = point_size,
                          point.shape = point_shape, point.alpha = point_alpha,
                          point.stroke = point_stroke,
                          residuals = residuals,
                          residual.colour = residual_colour,
                          residual.linewidth = residual_linewidth,
                          residual.linetype = residual_linetype,
                          residual.alpha = residual_alpha,
                          light = light, na.rm = na.rm, ...)
      )
}

# Stat for visualizing multiple 1D density distributions as 3D ridgelines

StatDistributions3D <- ggproto("StatDistributions3D", Stat,
                               required_aes = c("x", "y"),
                               default_aes = aes(z = after_stat(density)),

                               setup_params = function(data, params) {
                                     # Auto-detect direction if not specified
                                     if (is.null(params$direction)) {
                                           x_discrete <- is.factor(data$x) || is.character(data$x)
                                           y_discrete <- is.factor(data$y) || is.character(data$y) || inherits(data$y, "mapped_discrete")

                                           if (x_discrete && !y_discrete) {
                                                 params$direction <- "x"
                                           } else if (y_discrete && !x_discrete) {
                                                 params$direction <- "y"
                                           } else {
                                                 # Both continuous, both discrete, or can't determine - default to "x"
                                                 params$direction <- "x"
                                           }
                                     }

                                     # Compute joint bandwidth across all data if requested
                                     if (isTRUE(params$joint_bandwidth) && is.character(params$bw)) {
                                           # Get the density variable based on direction
                                           if (params$direction == "x") {
                                                 density_var <- data$y
                                           } else {
                                                 density_var <- data$x
                                           }

                                           # Remove non-finite values
                                           density_var <- density_var[is.finite(density_var)]

                                           if (length(density_var) >= 2) {
                                                 # Compute bandwidth using the specified method
                                                 bw_fun <- switch(params$bw,
                                                                  nrd0 = stats::bw.nrd0,
                                                                  nrd = stats::bw.nrd,
                                                                  ucv = stats::bw.ucv,
                                                                  bcv = stats::bw.bcv,
                                                                  SJ = stats::bw.SJ,
                                                                  `SJ-ste` = function(x) stats::bw.SJ(x, method = "ste"),
                                                                  `SJ-dpi` = function(x) stats::bw.SJ(x, method = "dpi"),
                                                                  stats::bw.nrd0  # fallback
                                                 )
                                                 params$bw <- bw_fun(density_var) * (params$adjust %||% 1)
                                                 message("Picking joint bandwidth of ", signif(params$bw, 3))
                                           }
                                     }

                                     params
                               },

                               compute_group = function(data, scales,
                                                        direction = "auto",
                                                        n = 512,
                                                        bw = "nrd0",
                                                        adjust = 1,
                                                        kernel = "gaussian",
                                                        trim = FALSE,
                                                        bounds = c(-Inf, Inf),
                                                        rel_min_height = 0,
                                                        joint_bandwidth = FALSE,
                                                        light = NULL,
                                                        cull_backfaces = FALSE,
                                                        na.rm = FALSE) {

                                     # Determine which variable is the density variable vs position variable
                                     if (direction == "x") {
                                           # direction = "x" means ridges march along x, density computed along y
                                           # Each unique x value gets a density curve in the y-z plane
                                           density_vals <- data$y
                                           position_val <- data$x[1]  # Should be constant within group
                                     } else {
                                           # direction = "y" means ridges march along y, density computed along x
                                           # Each unique y value gets a density curve in the x-z plane
                                           density_vals <- data$x
                                           position_val <- data$y[1]  # Should be constant within group
                                     }

                                     # Remove non-finite values
                                     density_vals <- density_vals[is.finite(density_vals)]

                                     # Need at least 2 points for density estimation
                                     if (length(density_vals) < 2) {
                                           warning("Group has fewer than 2 points, skipping density estimation")
                                           return(data.frame())
                                     }

                                     # Handle bounds (following ggplot2 pattern)
                                     if (any(is.finite(bounds))) {
                                           # Remove data outside bounds
                                           in_bounds <- density_vals >= bounds[1] & density_vals <= bounds[2]
                                           if (!all(in_bounds)) {
                                                 warning("Removed ", sum(!in_bounds), " observations outside `bounds`")
                                                 density_vals <- density_vals[in_bounds]
                                           }
                                     }

                                     if (length(density_vals) < 2) {
                                           warning("Group has fewer than 2 points after bounds filtering, skipping")
                                           return(data.frame())
                                     }

                                     # Determine range for density evaluation
                                     data_range <- range(density_vals)

                                     # Compute density using stats::density
                                     # bw can be numeric or a character rule like "nrd0", "SJ", etc.
                                     # When trim = TRUE, restrict to data range; otherwise let density() choose
                                     if (trim) {
                                           dens <- stats::density(
                                                 density_vals,
                                                 bw = bw,
                                                 adjust = adjust,
                                                 kernel = kernel,
                                                 n = n,
                                                 from = data_range[1],
                                                 to = data_range[2]
                                           )
                                     } else {
                                           dens <- stats::density(
                                                 density_vals,
                                                 bw = bw,
                                                 adjust = adjust,
                                                 kernel = kernel,
                                                 n = n
                                           )
                                     }

                                     # Apply boundary correction if bounds are finite (following ggplot2)
                                     if (any(is.finite(bounds))) {
                                           # Zero out density outside bounds
                                           if (is.finite(bounds[1])) {
                                                 dens$y[dens$x < bounds[1]] <- 0
                                           }
                                           if (is.finite(bounds[2])) {
                                                 dens$y[dens$x > bounds[2]] <- 0
                                           }
                                     }

                                     # Build output data frame
                                     n_obs <- length(density_vals)
                                     max_density <- max(dens$y, na.rm = TRUE)

                                     result <- data.frame(
                                           density = dens$y,
                                           ndensity = if (max_density > 0) dens$y / max_density else 0,
                                           count = dens$y * n_obs,
                                           n = n_obs,
                                           bw = dens$bw
                                     )

                                     # Set coordinates based on direction
                                     if (direction == "x") {
                                           result$x <- position_val
                                           result$y <- dens$x
                                           result$z <- dens$y
                                     } else {
                                           result$x <- dens$x
                                           result$y <- position_val
                                           result$z <- dens$y
                                     }

                                     # Apply rel_min_height filter (like ggridges)
                                     if (rel_min_height > 0 && nrow(result) > 0) {
                                           overall_max <- max(result$density, na.rm = TRUE)
                                           result <- result[result$density >= rel_min_height * overall_max, ]
                                     }

                                     # Preserve non-coordinate columns from input (like fill, colour, etc.)
                                     non_coord_cols <- setdiff(names(data), c("x", "y", "z", "PANEL", "group"))
                                     if (length(non_coord_cols) > 0) {
                                           for (col_name in non_coord_cols) {
                                                 result[[col_name]] <- data[[col_name]][1]
                                           }
                                     }

                                     # Add ggcube-specific columns
                                     if (!is.null(cull_backfaces)) result$cull_backfaces <- cull_backfaces

                                     # Add direction param (since the geom inherits the original un-resolved version)
                                     result$.direction <- direction

                                     result <- attach_light(result, light)

                                     return(result)
                               }
)


#' 3D ridgeline distributions
#'
#' Computes 1D kernel density estimates for each group and arranges them as
#' ridgeline polygons in 3D space. Similar to `ggridges::geom_density_ridges()`,
#' but rendered as 3D surfaces using [geom_ridgeline_3d()].
#'
#' This stat is modeled after [ggplot2::stat_density()], with similar
#' parametrization for bandwidth selection, kernel choice, and boundary handling.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. This stat
#'   requires `x` and `y` aesthetics. One of these serves as the grouping/position
#'   variable (determined by `direction`), and the other provides values for
#'   density estimation.
#' @param data The data to be displayed in this layer.
#' @param geom The geometric object to use to display the data. Defaults to
#'   [geom_ridgeline_3d()].
#' @param stat Statistical transformation to use on the data. Defaults to
#'   [stat_distributions_3d()].
#' @param ... Other arguments passed to the layer.
#' @param bw The smoothing bandwidth to be used. If numeric, the standard
#'   deviation of the smoothing kernel. If character, a rule to choose the
#'   bandwidth, as listed in [stats::bw.nrd()]. Options include `"nrd0"` (default),
#'   `"nrd"`, `"ucv"`, `"bcv"`, `"SJ"`, `"SJ-ste"`, and `"SJ-dpi"`. Note that
#'
#'   automatic calculation is performed per-group unless `joint_bandwidth = TRUE`.
#' @param adjust A multiplicative bandwidth adjustment. This makes it possible
#'   to adjust the bandwidth while still using a bandwidth estimator. For
#'   example, `adjust = 1/2` means use half of the default bandwidth. Default is 1.
#' @param kernel Kernel function to use. One of `"gaussian"` (default),
#'   `"rectangular"`, `"triangular"`, `"epanechnikov"`, `"biweight"`, `"cosine"`,
#'   or `"optcosine"`. See [stats::density()] for details.
#' @param n Number of equally spaced points at which the density is estimated.
#'   Should be a power of two for efficiency. Default is 512.
#' @param trim If `FALSE` (the default), each density is computed on the full
#'   range of the data (extended by a factor based on bandwidth). If `TRUE`,
#'   each density is computed over the range of that group only.
#' @param bounds A numeric vector of length 2 giving the lower and upper bounds
#'   for bounded density estimation. Density values outside bounds are set to
#'   zero. Data points outside bounds are removed with a warning.
#'   Default is `c(-Inf, Inf)` (unbounded).
#' @param rel_min_height Lines with heights below this cutoff will be removed.
#'   The cutoff is measured relative to the maximum height within each group.
#'   For example, `rel_min_height = 0.01` removes points with density less than
#'   1% of the peak. Default is 0 (no removal). This is similar to the
#'   parameter of the same name in `ggridges::geom_density_ridges()`.
#' @param joint_bandwidth If `TRUE`, bandwidth is computed jointly across all
#'   groups using the specified `bw` method, ensuring consistent smoothing
#'   across all density curves. This matches the behavior of
#'   [ggridges::stat_density_ridges()]. If `FALSE` (the default), bandwidth is
#'   computed separately for each group, matching [ggplot2::stat_density()].
#'   Only applies when `bw` is a character string (bandwidth rule), not when
#'   `bw` is provided as a numeric value.
#' @inheritParams ridgeline_params
#' @inheritParams polygon_params
#' @inheritParams light_param
#' @inheritParams position_param
#'
#' @section Aesthetics:
#' `stat_distributions_3d()` understands the following aesthetics (required
#' aesthetics are in bold):
#' \describe{
#'   \item{**x**}{X coordinate - either density variable or position variable
#'     depending on `direction`}
#'   \item{**y**}{Y coordinate - either position variable or density variable
#'     depending on `direction`}
#'   \item{group}{Grouping variable (typically derived from the position aesthetic)}
#'   \item{fill, colour, alpha, linewidth, linetype}{Passed to [geom_ridgeline_3d()]}
#' }
#'
#' @section Direction:
#' The `direction` parameter determines how the data is interpreted:
#' \describe{
#'   \item{`direction = NULL` (default)}{Automatically detects direction based
#'     on whether `x` or `y` is discrete (factor/character). If `x` is discrete
#'     and `y` is continuous, uses `"x"`; if `y` is discrete and `x` is
#'     continuous, uses `"y"`. Falls back to `"x"` if ambiguous.}
#'   \item{`direction = "x"`}{Ridges march along the x-axis. Each unique
#'     x value defines a group, and density is computed from the y values within
#'     that group. The resulting density curves lie in the y-z plane.}
#'   \item{`direction = "y"`}{Ridges march along the y-axis. Each unique y value
#'     defines a group, and density is computed from the x values. The resulting
#'     density curves lie in the x-z plane.}
#' }
#'
#' @section Computed variables:
#' The following variables are computed and available via `after_stat()`:
#' \describe{
#'   \item{density}{The kernel density estimate at each point}
#'   \item{ndensity}{Density normalized to a maximum of 1 within each group}
#'   \item{count}{Density multiplied by number of observations (expected count)}
#'   \item{n}{Number of observations in the group}
#'   \item{bw}{Bandwidth actually used for this group}
#' }
#'
#' @examples
#' library(ggplot2)
#'
#' # Basic usage with iris data
#' p <- ggplot(iris, aes(y = Sepal.Length, x = Species, fill = Species)) +
#'   coord_3d() +
#'   scale_z_continuous(expand = expansion(mult = c(0, NA))) + # remove gap beneath ridges
#'   theme(legend.position = "none")
#' p + stat_distributions_3d()
#'
#' # Normalize max ridge heights
#' p + stat_distributions_3d(aes(z = after_stat(ndensity)))
#'
#' # Adjust smoothing bandwidth
#' p + stat_distributions_3d(adjust = 0.5)
#'
#' # Use joint bandwidth for consistent smoothing across groups
#' p + stat_distributions_3d(joint_bandwidth = TRUE)
#'
#' # Different bandwidth selection rules
#' p + stat_distributions_3d(bw = "SJ")
#'
#' # Remove tails with rel_min_height
#' p + stat_distributions_3d(rel_min_height = 0.05)
#'
#' # Trim to data range
#' p + stat_distributions_3d(trim = TRUE)
#'
#' # Rotated to reduce perspective distortion
#' p + stat_distributions_3d(alpha = .7) +
#'    coord_3d(pitch = 0, roll = -90, yaw = 90, dist = 5,
#'       panels = c("zmin", "xmin"))
#'
#' @seealso [geom_ridgeline_3d()] for rendering pre-computed ridgeline data,
#'   [stat_density_3d()] for 2D kernel density surfaces,
#'   [ggplot2::stat_density()] for the parametrization this stat follows,
#'   `ggridges::geom_density_ridges()` for the 2D ridgeline equivalent.
#'
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname stat_distributions_3d
#' @export
stat_distributions_3d <- function(mapping = NULL, data = NULL,
                                  geom = "ridgeline_3d",
                                  position = "identity",
                                  ...,
                                  direction = NULL,
                                  bw = "nrd0",
                                  adjust = 1,
                                  kernel = "gaussian",
                                  n = 512,
                                  trim = FALSE,
                                  bounds = c(-Inf, Inf),
                                  rel_min_height = 0,
                                  joint_bandwidth = FALSE,
                                  base = 0,
                                  light = NULL,
                                  cull_backfaces = FALSE,
                                  sort_method = NULL,
                                  scale_depth = TRUE,
                                  force_convex = FALSE,
                                  na.rm = FALSE,
                                  show.legend = NA,
                                  inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = StatDistributions3D,
            geom = get_proto(geom, "geom"),
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  direction = direction,
                  bw = bw,
                  adjust = adjust,
                  kernel = kernel,
                  n = n,
                  trim = trim,
                  bounds = bounds,
                  rel_min_height = rel_min_height,
                  joint_bandwidth = joint_bandwidth,
                  base = base,
                  light = light,
                  cull_backfaces = cull_backfaces,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  force_convex = force_convex,
                  na.rm = na.rm,
                  ...
            )
      )
}


#' @rdname stat_distributions_3d
#' @export
geom_distributions_3d <- function(mapping = NULL, data = NULL,
                                  stat = "distributions_3d",
                                  position = "identity",
                                  ...,
                                  direction = NULL,
                                  bw = "nrd0",
                                  adjust = 1,
                                  kernel = "gaussian",
                                  n = 512,
                                  trim = FALSE,
                                  bounds = c(-Inf, Inf),
                                  rel_min_height = 0,
                                  joint_bandwidth = FALSE,
                                  base = 0,
                                  light = NULL,
                                  cull_backfaces = FALSE,
                                  sort_method = NULL,
                                  scale_depth = TRUE,
                                  force_convex = FALSE,
                                  na.rm = FALSE,
                                  show.legend = NA,
                                  inherit.aes = TRUE) {

      layer(
            data = data,
            mapping = mapping,
            stat = get_proto(stat, "stat"),
            geom = GeomRidgeline3D,
            position = position,
            show.legend = show.legend,
            inherit.aes = inherit.aes,
            params = list(
                  direction = direction,
                  bw = bw,
                  adjust = adjust,
                  kernel = kernel,
                  n = n,
                  trim = trim,
                  bounds = bounds,
                  rel_min_height = rel_min_height,
                  joint_bandwidth = joint_bandwidth,
                  base = base,
                  light = light,
                  cull_backfaces = cull_backfaces,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  force_convex = force_convex,
                  na.rm = na.rm,
                  ...
            )
      )
}

#' 3D paths connecting observations
#'
#' Connects observations in 3D space in the order they appear
#' in the data. It converts path data into individual segments for proper depth
#' sorting while maintaining the appearance of connected paths. Each path is
#' divided into segments that can be depth-sorted independently.
#'
#' @param mapping Set of aesthetic mappings created by [aes()]. Requires x, y, z
#'   coordinates. Grouping aesthetics determine separate paths.
#' @param data The data to be displayed in this layer.
#' @param stat The statistical transformation to use on the data. Defaults to `"path_3d"`.
#' @param geom The geometric object used to display the data. Defaults to `"segment_3d"`.
#' @param position Position adjustment, defaults to "identity".
#' @param ... Other arguments passed on to [ggplot2::layer()].
#' @param na.rm If `FALSE`, missing values are removed with a warning.
#' @param show.legend Logical indicating whether this layer should be included in legends.
#' @param inherit.aes If `FALSE`, overrides the default aesthetics.
#' @inheritParams sort_params
#' @inheritParams depth_params
#' @param arrow Specification for arrow heads, created by [arrow()].
#' @param lineend Line end style, one of "round", "butt", "square".
#'
#' @section Aesthetics:
#' `geom_path_3d()` understands the following aesthetics:
#' - **x, y, z**: Coordinates (required)
#' - `group`: Grouping variable to create separate paths
#' - `colour`: Line color
#' - `linewidth`: Line width (gets depth-scaled when `scale_depth = TRUE`)
#' - `linetype`: Line type
#' - `alpha`: Transparency
#'
#' @section Computed variables for StatSegment3D:
#' - `x`, `y`, `z`: Start coordinates of each segment
#' - `xend`, `yend`, `zend`: End coordinates of each segment
#' - `group`: Hierarchical group identifier preserving original grouping
#'
#' @section Grouping:
#' Multiple paths are created based on grouping aesthetics (group, colour, etc.).
#' Each group forms a separate path, and segments from different paths can be
#' interleaved during depth sorting for proper 3D rendering.
#'
#' @examples
#' library(ggplot2)
#'
#' x <- seq(0, 20*pi, pi/16)
#' spiral <- data.frame(
#'   x = x,
#'   y = sin(x),
#'   z = cos(x))
#'
#' # Basic path
#' ggplot(spiral, aes(x, y, z)) +
#'   geom_path_3d() +
#'   coord_3d()
#'
#' # With aesthetic coloring
#' ggplot(spiral, aes(x, y, z, color = y)) +
#'   geom_path_3d(linewidth = 1, lineend = "round") +
#'   coord_3d() +
#'   scale_color_gradientn(colors = c("red", "purple", "blue"))
#'
#' # With grouping
#' ggplot(spiral, aes(x, y, z, color = x > 30)) +
#'   geom_path_3d(linewidth = 1, lineend = "round") +
#'   coord_3d()
#'
#' @seealso [geom_segment_3d()] for individual segments.
#' @return A `Layer` object that can be added to a ggplot.
#' @rdname geom_path_3d
#' @export
geom_path_3d <- function(mapping = NULL, data = NULL,
                         stat = StatPath3D, position = "identity",
                         ...,
                         sort_method = "painter",
                         scale_depth = TRUE, arrow = NULL, lineend = "butt",
                         na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = get_proto(stat, "stat"), geom = GeomSegment3D,
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(
                  na.rm = na.rm,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  arrow = arrow,
                  lineend = lineend,
                  ...
            )
      )
}

#' @rdname geom_path_3d
#' @export
stat_path_3d <- function(mapping = NULL, data = NULL,
                         geom = GeomSegment3D, position = "identity",
                         ...,
                         sort_method = "painter",
                         scale_depth = TRUE, arrow = NULL, lineend = "butt",
                         na.rm = FALSE, show.legend = NA, inherit.aes = TRUE) {

      layer(data = data, mapping = mapping, stat = StatPath3D, geom = get_proto(geom, "geom"),
            position = position, show.legend = show.legend, inherit.aes = inherit.aes,
            params = list(
                  na.rm = na.rm,
                  sort_method = sort_method,
                  scale_depth = scale_depth,
                  arrow = arrow,
                  lineend = lineend,
                  ...
            )
      )
}

StatPath3D <- ggproto("StatPath3D", Stat,
                      required_aes = c("x", "y", "z"),

                      compute_group = function(data, scales, na.rm = FALSE) {

                            # Remove missing values if requested
                            if (na.rm) {
                                  complete_cases <- complete.cases(data[c("x", "y", "z")])
                                  data <- data[complete_cases, ]
                            }

                            # Need at least 2 points to make a path
                            if (nrow(data) < 2) {
                                  return(data.frame())
                            }

                            # Handle discrete scale conversion (following StatPoint3D pattern)
                            if ("z" %in% names(data)) {
                                  data$z_raw <- data$z
                                  if (is.factor(data$z) || is.character(data$z)) {
                                        data$z <- as.numeric(as.factor(data$z))
                                  }
                            }

                            if ("x" %in% names(data)) {
                                  data$x_raw <- data$x
                                  if (is.factor(data$x) || is.character(data$x)) {
                                        data$x <- as.numeric(as.factor(data$x))
                                  }
                            }

                            if ("y" %in% names(data)) {
                                  data$y_raw <- data$y
                                  if (is.factor(data$y) || is.character(data$y)) {
                                        data$y <- as.numeric(as.factor(data$y))
                                  }
                            }

                            # Convert path to segments
                            n_points <- nrow(data)
                            n_segments <- n_points - 1

                            if (n_segments == 0) {
                                  return(data.frame())
                            }

                            # Create segments from consecutive points
                            segments <- data.frame(
                                  x = data$x[1:n_segments],
                                  y = data$y[1:n_segments],
                                  z = data$z[1:n_segments],
                                  xend = data$x[2:n_points],
                                  yend = data$y[2:n_points],
                                  zend = data$z[2:n_points]
                            )

                            # Create hierarchical group IDs
                            original_group <- data$group[1]
                            segments$group <- paste0(original_group, "__seg", 1:n_segments)

                            # Preserve other aesthetics from the first point of each segment
                            other_cols <- setdiff(names(data), c("x", "y", "z", "group"))
                            for (col_name in other_cols) {
                                  segments[[col_name]] <- data[[col_name]][1:n_segments]
                            }

                            segments <- average_aesthetics(segments)

                            return(segments)
                      }
)

#' Lorenz butterfly attractor generator
#'
#' @keywords internal
#' @noRd
lorenz_attractor <- function(n_points = 5000, dt = 0.01,
                             sigma = 10, rho = 28, beta = 8/3) {

      # Initialize vectors
      x <- numeric(n_points)
      y <- numeric(n_points)
      z <- numeric(n_points)

      # Initial conditions (slightly off the unstable equilibrium)
      x[1] <- 1
      y[1] <- 1
      z[1] <- 1

      # Integrate the Lorenz equations using Euler method
      for (i in 2:n_points) {
            # Lorenz equations:
            # dx/dt = σ(y - x)
            # dy/dt = x(ρ - z) - y
            # dz/dt = xy - βz

            dx_dt <- sigma * (y[i-1] - x[i-1])
            dy_dt <- x[i-1] * (rho - z[i-1]) - y[i-1]
            dz_dt <- x[i-1] * y[i-1] - beta * z[i-1]

            x[i] <- x[i-1] + dx_dt * dt
            y[i] <- y[i-1] + dy_dt * dt
            z[i] <- z[i-1] + dz_dt * dt
      }

      data.frame(x = x, y = y, z = z, time = (1:n_points) * dt)
}

# annotate-3d.R
# In-layer 3D annotation system
#
# Provides annotate_3d() for embedding fixed-position annotations (points,
# text labels, segments) within a 3D layer so they participate in the same
# depth sorting pass as the primary geometry.


# Constructor ------------------------------------------------------------------

#' Create a 3D annotation specification
#'
#' Defines fixed-position annotations to be embedded within a 3D layer.
#' Annotations are depth-sorted together with the primary geometry, unlike
#' ggplot2's `annotate()` which creates a separate layer.
#'
#' Parameters can be vectors to create multiple annotations of the same type
#' in one call, with scalar values recycled as needed (matching
#' `ggplot2::annotate()` behaviour).
#'
#' @param type Character string specifying the annotation type. One of
#'   `"point"`, `"text"`, or `"segment"`.
#' @param ... Type-specific parameters. See Details.
#'
#' @details
#' ## Point annotations
#' Requires `x`, `y`, `z`. Optional styling: `colour`/`color`, `fill`, `size`,
#' `shape`, `alpha`, `stroke`.
#'
#' ## Text annotations
#' Requires `x`, `y`, `z`, `label`. Optional styling: `colour`/`color`, `size`,
#' `alpha`, `family`, `fontface`, `hjust`, `vjust`, `angle`, `lineheight`.
#'
#' ## Segment annotations
#' Requires `x`, `y`, `z`, `xend`, `yend`, `zend`. Optional styling:
#' `colour`/`color`, `linewidth`, `linetype`, `alpha`.
#'
#' @return An S3 object of class `"annotate_3d"`.
#'
#' @examples
#' p <- ggplot(mountain, aes(x, y, z)) +
#'   coord_3d(ratio = c(2, 3, 1),
#'            light = light(mode = "hsl", direction = c(-1, 0, 0)))
#'
#' # Basic point annotation
#' p + geom_surface_3d(
#'   annotate = annotate_3d("point", x = .5, y = .25, z = 100,
#'                          color = "red", size = 3))
#'
#' # Vectorized: multiple points in one call
#' p + geom_surface_3d(
#'   annotate = annotate_3d("point", x = .5, y = .25, z = seq(50, 100, 5),
#'                          color = "red", size = 3))
#'
#' # Multiple annotation types
#' p + geom_surface_3d(
#'   annotate = list(
#'     annotate_3d("point", x = .5, y = .25, z = 100,
#'                 color = "red", size = 3),
#'     annotate_3d("text", x = .5, y = .25, z = 100,
#'                 color = "red", label = "Look here!",
#'                 vjust = -1, fontface = "bold"),
#'     annotate_3d("segment", x = .5, y = .25, z = 100,
#'                 xend = .5, yend = .25, zend = 50,
#'                 color = "red", size = 3)))
#'
#' @export
annotate_3d <- function(type, ...) {

      valid_types <- c("point", "text", "segment")
      if (!type %in% valid_types) {
            stop("annotate_3d() type must be one of: ",
                 paste(valid_types, collapse = ", "),
                 call. = FALSE)
      }

      params <- list(...)

      # Normalize color -> colour (American -> British, matching ggplot2)
      if ("color" %in% names(params)) {
            params$colour <- params$colour %||% params$color
            params$color <- NULL
      }

      # Validate required parameters
      required <- switch(type,
                         point   = c("x", "y", "z"),
                         text    = c("x", "y", "z", "label"),
                         segment = c("x", "y", "z", "xend", "yend", "zend")
      )
      missing <- setdiff(required, names(params))
      if (length(missing) > 0) {
            stop("annotate_3d('", type, "') requires: ",
                 paste(missing, collapse = ", "),
                 call. = FALSE)
      }

      # Expand and recycle all params via data.frame(). This handles
      # vectorized inputs (e.g. x = 1:5, colour = "red") with standard
      # R recycling rules and throws clear errors for incompatible lengths.
      params_df <- tryCatch(
            data.frame(params, stringsAsFactors = FALSE),
            error = function(e) {
                  stop("annotate_3d(): incompatible parameter lengths.\n",
                       conditionMessage(e), call. = FALSE)
            }
      )

      structure(
            list(type = type, data = params_df),
            class = "annotate_3d"
      )
}


# Build annotations ------------------------------------------------------------

#' Build annotation rows from annotation specs
#'
#' Takes annotation specifications (single or list) and generates data-space
#' rows tagged with `.prim` and dot-prefixed style columns, ready to be
#' bound onto primary data in `setup_data` so that annotation positions
#' participate in scale training.
#'
#' For annotation types that need panel range information (e.g. planes),
#' positional columns may contain sentinel values (NA) that are resolved
#' later in `draw_panel` via `resolve_annotations()`.
#'
#' @param annotate A single `annotate_3d` object, or a list of them.
#' @param max_group Numeric. The maximum existing group number, used to
#'   generate non-colliding group IDs for annotation rows.
#' @return A data frame of annotation rows, or `NULL` if `annotate` is NULL.
#' @keywords internal
#' @noRd
build_annotations <- function(annotate, max_group = 0) {

      if (is.null(annotate)) return(NULL)

      # Normalize to a flat list of specs
      specs <- normalize_annotation_specs(annotate)

      if (length(specs) == 0) return(NULL)

      # Generate rows for each spec
      all_rows <- list()
      group_offset <- max_group

      for (spec in specs) {
            result <- switch(spec$type,
                             point   = make_annotation_point_rows(spec, group_offset),
                             text    = make_annotation_text_rows(spec, group_offset),
                             segment = make_annotation_segment_rows(spec, group_offset),
                             stop("Unknown annotation type: ", spec$type, call. = FALSE)
            )
            all_rows <- c(all_rows, list(result$data))
            group_offset <- result$next_offset
      }

      dplyr::bind_rows(all_rows)
}


#' Normalize annotation input to a flat list of specs
#'
#' Accepts a single annotate_3d object, a list of them, or nested
#' combinations, and returns a flat list of annotate_3d objects.
#'
#' @param annotate Input annotation spec(s).
#' @return A flat list of `annotate_3d` objects.
#' @keywords internal
#' @noRd
normalize_annotation_specs <- function(annotate) {

      if (inherits(annotate, "annotate_3d")) {
            return(list(annotate))
      }

      if (is.list(annotate)) {
            # Flatten and validate
            specs <- list()
            for (item in annotate) {
                  if (inherits(item, "annotate_3d")) {
                        specs <- c(specs, list(item))
                  } else if (is.list(item)) {
                        # Recurse for nested lists
                        specs <- c(specs, normalize_annotation_specs(item))
                  } else {
                        warning("Ignoring non-annotate_3d element in annotation list",
                                call. = FALSE)
                  }
            }
            return(specs)
      }

      warning("annotate must be an annotate_3d object or list of them",
              call. = FALSE)
      return(list())
}


# Geom integration helpers -----------------------------------------------------

#' Build and bind annotation rows in setup_data
#'
#' One-liner helper for geom `setup_data` methods. Checks for annotation
#' specs in params, builds annotation rows, handles PANEL assignment and
#' type coercion, and binds onto the primary data. Returns data unchanged
#' if no annotations are present.
#'
#' @param data Primary data frame from the layer.
#' @param params Layer params list (checked for `$annotate`).
#' @return Data frame with annotation rows appended, or original data if
#'   no annotations.
#' @keywords internal
#' @noRd
setup_annotations <- function(data, params) {
      annotate <- params$annotate
      if (is.null(annotate)) return(data)

      max_group <- max(as.numeric(factor(data$group)), na.rm = TRUE)
      ann_rows <- build_annotations(annotate, max_group = max_group)
      if (is.null(ann_rows) || nrow(ann_rows) == 0) return(data)

      ann_rows$PANEL <- data$PANEL[1]

      # Coerce overlapping columns to compatible types so bind_rows
      # doesn't choke on mismatches (e.g. factor vs character group).
      # bind_rows fills missing columns with NA automatically.
      shared_cols <- intersect(names(data), names(ann_rows))
      for (col in shared_cols) {
            if (is.factor(data[[col]]) && !is.factor(ann_rows[[col]])) {
                  ann_rows[[col]] <- as.character(ann_rows[[col]])
                  data[[col]] <- as.character(data[[col]])
            }
      }

      dplyr::bind_rows(data, ann_rows)
}


#' Resolve annotation sentinels in draw_panel
#'
#' One-liner helper for geom `draw_panel` methods, called before coord
#' transformation. For annotation types that need panel range information
#' (e.g. planes needing extent on free axes), replaces sentinel NA values
#' with actual ranges from `panel_params`.
#'
#' For v1 annotation types (point, text, segment), this is a no-op since
#' all positions are fully specified at build time.
#'
#' @param data Data frame potentially containing annotation rows with
#'   sentinel values.
#' @param panel_params Panel parameters from the coord, containing scale
#'   ranges.
#' @return Data frame with sentinel values resolved.
#' @keywords internal
#' @noRd
prepare_annotations <- function(data, panel_params) {
      # v1: no-op. Point, text, and segment annotations are fully specified.
      # When plane annotations are added, this function will resolve NA extents
      # using panel_params$scale_info.
      data
}


# Apply annotation styles ------------------------------------------------------

#' Apply annotation styles from dot-prefixed columns
#'
#' After coord transformation, copies `.ann_*` style columns to the standard
#' aesthetic columns for annotation rows only (identified by `.ann` marker).
#' This allows annotation styling to survive scale training without
#' contaminating colour/fill/etc. scales.
#'
#' @param data Transformed data frame potentially containing annotation rows.
#' @return Data frame with annotation styles applied.
#' @keywords internal
#' @noRd
apply_annotation_styles <- function(data) {

      if (!".ann" %in% names(data)) return(data)

      ann_mask <- !is.na(data$.ann) & data$.ann
      if (!any(ann_mask)) return(data)

      # Find all .ann_ prefixed columns
      ann_cols <- grep("^\\.ann_", names(data), value = TRUE)

      for (ann_col in ann_cols) {
            # Map .ann_colour -> colour, .ann_fill -> fill, etc.
            target_col <- sub("^\\.ann_", "", ann_col)

            # Ensure target column exists
            if (!target_col %in% names(data)) {
                  data[[target_col]] <- NA
            }

            # Apply annotation values only to annotation rows
            ann_values <- data[[ann_col]][ann_mask]
            non_na <- !is.na(ann_values)
            if (any(non_na)) {
                  idx <- which(ann_mask)[non_na]
                  data[[target_col]][idx] <- ann_values[non_na]
            }
      }

      data
}


# Row generators ---------------------------------------------------------------

#' Generate point annotation rows
#' @keywords internal
#' @noRd
make_annotation_point_rows <- function(spec, group_offset) {

      d <- spec$data
      n <- nrow(d)
      ids <- seq_len(n) + group_offset

      rows <- data.frame(
            x = d$x,
            y = d$y,
            z = d$z,
            group = paste0("ann__pt", ids),
            .prim = "point",
            .ann = TRUE,
            stringsAsFactors = FALSE
      )

      # Point style defaults
      style_defaults <- list(
            colour = "black",
            fill   = NA_character_,
            size   = 2,
            shape  = 19,
            alpha  = 1,
            stroke = 0.5
      )

      # Apply user values (already recycled in spec$data) or defaults
      for (col in names(style_defaults)) {
            val <- if (col %in% names(d)) d[[col]] else style_defaults[[col]]
            rows[[paste0(".ann_", col)]] <- val
      }

      list(data = rows, next_offset = group_offset + n)
}


#' Generate text annotation rows
#' @keywords internal
#' @noRd
make_annotation_text_rows <- function(spec, group_offset) {

      d <- spec$data
      n <- nrow(d)
      ids <- seq_len(n) + group_offset

      rows <- data.frame(
            x = d$x,
            y = d$y,
            z = d$z,
            group = paste0("ann__text", ids),
            label = d$label,
            .prim = "text",
            .ann = TRUE,
            stringsAsFactors = FALSE
      )

      # Text style defaults
      style_defaults <- list(
            colour     = "black",
            size       = 3.88,
            alpha      = 1,
            family     = "",
            fontface   = 1,
            hjust      = 0.5,
            vjust      = 0.5,
            angle      = 0,
            lineheight = 1.2
      )

      for (col in names(style_defaults)) {
            val <- if (col %in% names(d)) d[[col]] else style_defaults[[col]]
            rows[[paste0(".ann_", col)]] <- val
      }

      list(data = rows, next_offset = group_offset + n)
}


#' Generate segment annotation rows
#' @keywords internal
#' @noRd
make_annotation_segment_rows <- function(spec, group_offset) {

      d <- spec$data
      n <- nrow(d)
      ids <- seq_len(n) + group_offset
      seg_groups <- paste0("ann__seg", ids)

      start_rows <- data.frame(
            x = d$x,
            y = d$y,
            z = d$z,
            group = seg_groups,
            point_type = "start",
            segment_id = ids,
            .prim = "segment",
            .ann = TRUE,
            stringsAsFactors = FALSE
      )

      end_rows <- data.frame(
            x = d$xend,
            y = d$yend,
            z = d$zend,
            group = seg_groups,
            point_type = "end",
            segment_id = ids,
            .prim = "segment",
            .ann = TRUE,
            stringsAsFactors = FALSE
      )

      rows <- rbind(start_rows, end_rows)

      # Segment style defaults
      style_defaults <- list(
            colour    = "black",
            linewidth = 0.5,
            linetype  = 1,
            alpha     = 1
      )

      for (col in names(style_defaults)) {
            val <- if (col %in% names(d)) rep(d[[col]], 2) else style_defaults[[col]]
            rows[[paste0(".ann_", col)]] <- val
      }

      list(data = rows, next_offset = group_offset + n)
}

# test-annotate-3d.R

# Constructor ------------------------------------------------------------------

test_that("annotate_3d creates valid specs", {
      pt <- annotate_3d("point", x = 1, y = 2, z = 3)
      expect_s3_class(pt, "annotate_3d")
      expect_equal(pt$type, "point")
      expect_equal(nrow(pt$data), 1)

      txt <- annotate_3d("text", x = 0, y = 0, z = 0, label = "hi")
      expect_equal(txt$type, "text")
      expect_equal(txt$data$label, "hi")

      seg <- annotate_3d("segment", x = 0, y = 0, z = 0, xend = 1, yend = 1, zend = 1)
      expect_equal(seg$type, "segment")
      expect_equal(seg$data$xend, 1)
})

test_that("annotate_3d validates required params", {
      expect_error(annotate_3d("point", x = 1, y = 2), "requires")
      expect_error(annotate_3d("text", x = 0, y = 0, z = 0), "requires.*label")
      expect_error(annotate_3d("segment", x = 0, y = 0, z = 0), "requires")
})

test_that("annotate_3d rejects invalid types", {
      expect_error(annotate_3d("plane", z = 0), "must be one of")
})

test_that("annotate_3d normalizes color to colour", {
      pt <- annotate_3d("point", x = 0, y = 0, z = 0, color = "red")
      expect_equal(pt$data$colour, "red")
      expect_false("color" %in% names(pt$data))
})

test_that("annotate_3d supports vectorized params", {
      pt <- annotate_3d("point", x = 1:3, y = 0, z = 0)
      expect_equal(nrow(pt$data), 3)
      expect_equal(pt$data$x, 1:3)
      expect_equal(pt$data$y, c(0, 0, 0))

      pt2 <- annotate_3d("point", x = 1:3, y = 0, z = 0, colour = c("red", "green", "blue"))
      expect_equal(nrow(pt2$data), 3)
      expect_equal(pt2$data$colour, c("red", "green", "blue"))
})

test_that("annotate_3d errors on incompatible lengths", {
      expect_error(annotate_3d("point", x = 1:3, y = 1:2, z = 0), "incompatible")
})


# Build annotations ------------------------------------------------------------

test_that("build_annotations handles NULL", {
      expect_null(build_annotations(NULL))
})

test_that("build_annotations handles single spec", {
      pt <- annotate_3d("point", x = 1, y = 2, z = 3)
      rows <- build_annotations(pt)
      expect_equal(nrow(rows), 1)
      expect_equal(rows$.prim, "point")
      expect_true(rows$.ann)
})

test_that("build_annotations handles list of mixed types", {
      specs <- list(
            annotate_3d("point", x = 0, y = 0, z = 0),
            annotate_3d("segment", x = 0, y = 0, z = 0, xend = 1, yend = 1, zend = 1),
            annotate_3d("text", x = 0, y = 0, z = 0, label = "hi")
      )
      rows <- build_annotations(specs)
      expect_equal(nrow(rows), 4) # 1 point + 2 segment rows + 1 text
      expect_equal(sum(rows$.prim == "point"), 1)
      expect_equal(sum(rows$.prim == "segment"), 2)
      expect_equal(sum(rows$.prim == "text"), 1)
      expect_true(all(rows$.ann))
})

test_that("build_annotations assigns unique groups", {
      specs <- list(
            annotate_3d("point", x = 1:3, y = 0, z = 0),
            annotate_3d("point", x = 4, y = 0, z = 0)
      )
      rows <- build_annotations(specs)
      expect_equal(nrow(rows), 4)
      expect_equal(length(unique(rows$group)), 4)
})


# Apply annotation styles ------------------------------------------------------

test_that("apply_annotation_styles copies .ann_ columns to standard columns", {
      data <- data.frame(
            x = c(0, 1), y = c(0, 1),
            colour = c("blue", "blue"),
            .ann = c(FALSE, TRUE),
            .ann_colour = c(NA, "red"),
            stringsAsFactors = FALSE
      )
      result <- apply_annotation_styles(data)
      expect_equal(result$colour, c("blue", "red"))
})

test_that("apply_annotation_styles is a no-op without .ann column", {
      data <- data.frame(x = 1, y = 1, colour = "blue", stringsAsFactors = FALSE)
      result <- apply_annotation_styles(data)
      expect_equal(result$colour, "blue")
})


# Integration: rendering without error -----------------------------------------

test_that("annotations render on polygon-based geoms", {
      ann <- annotate_3d("point", x = 0, y = 0, z = 0, colour = "red")

      expect_no_error({
            ggplot(sphere_points, aes(x, y, z)) +
                  geom_hull_3d(fill = "dodgerblue", annotate = ann) +
                  coord_3d()
      })

      expect_no_error({
            ggplot(sphere_points, aes(x, y, z)) +
                  geom_polygon_3d(annotate = ann) +
                  coord_3d()
      })
})

test_that("annotations render on geom_point_3d", {
      d <- expand.grid(x = 1:3, y = 1:3, z = 1:3)
      ann <- annotate_3d("text", x = 2, y = 2, z = 2, label = "center")

      expect_no_error({
            ggplot(d, aes(x, y, z)) +
                  geom_point_3d(annotate = ann) +
                  coord_3d()
      })
})

test_that("annotations render on geom_segment_3d", {
      d <- data.frame(x = 0, y = 0, z = 0, xend = 1, yend = 1, zend = 1)
      ann <- annotate_3d("point", x = 0.5, y = 0.5, z = 0.5, colour = "red", size = 5)

      expect_no_error({
            ggplot(d, aes(x, y, z, xend = xend, yend = yend, zend = zend)) +
                  geom_segment_3d(annotate = ann) +
                  coord_3d()
      })
})

test_that("mixed annotation types render together", {
      ann <- list(
            annotate_3d("point", x = 0, y = 0, z = 1, colour = "red", size = 3),
            annotate_3d("text", x = 0, y = 0, z = 1.2, label = "top"),
            annotate_3d("segment", x = 0, y = 0, z = -1,
                        xend = 0, yend = 0, zend = 1, colour = "grey50")
      )

      expect_no_error({
            ggplot(sphere_points, aes(x, y, z)) +
                  geom_hull_3d(fill = "dodgerblue", annotate = ann) +
                  coord_3d()
      })
})

test_that("vectorized annotations render", {
      ann <- annotate_3d("point", x = c(-1, 0, 1), y = 0, z = 0,
                         colour = c("red", "white", "blue"), size = 4)

      expect_no_error({
            ggplot(sphere_points, aes(x, y, z)) +
                  geom_hull_3d(fill = "dodgerblue", annotate = ann) +
                  coord_3d()
      })
})

test_that("annotations expand scale ranges", {
      # Annotation at z = 5 should expand z scale beyond sphere_points range (~-1 to 1)
      ann <- annotate_3d("point", x = 0, y = 0, z = 5, colour = "red")
      p <- ggplot(sphere_points, aes(x, y, z)) +
            geom_hull_3d(fill = "dodgerblue", annotate = ann) +
            coord_3d()
      built <- ggplot_build(p)
      z_range <- built$layout$panel_params[[1]]$scale_info$z$limits
      expect_true(z_range[2] >= 5)
})

#' Aesthetic mapping with positional z support
#'
#' This function extends \code{ggplot2::aes()} to support positional mapping
#' to the z aesthetic. It maintains full backward compatibility with the original
#' \code{aes()} function while enabling the convenient \code{aes(x, y, z)} syntax
#' for 3D plots.
#'
#' @param x Variable to map to x aesthetic (required)
#' @param y Variable to map to y aesthetic (required)
#' @param z Variable to map to z aesthetic (optional)
#' @param ... Other aesthetic mappings (color, size, etc.)
#'
#' @return An aesthetic mapping object, same as \code{ggplot2::aes()}
#'
#' @details
#' This function is a lightweight wrapper around \code{ggplot2::aes()} that:
#' \itemize{
#'   \item Maintains full backward compatibility with existing 2D plots
#'   \item Enables positional z mapping: \code{aes(x_var, y_var, z_var)}
#'   \item Works with any geom that uses the z aesthetic (contour, raster, 3D plots)
#'   \item Passes through all other aesthetics unchanged
#' }
#'
#' @examples
#' library(ggplot2)
#'
#' # 2D plots work like regular ggplot2
#' ggplot(mtcars, aes(mpg, wt)) + geom_point()
#'
#' # 3D plots can use positional syntax, or explicitly map to z
#' # same as: ggplot(mtcars, aes(mpg, wt, z = qsec)) + geom_point() + coord_3d()
#' ggplot(mtcars, aes(mpg, wt, qsec)) + geom_point() + coord_3d()
#'
#' # Also works with non-ggcube z-aesthetic geoms
#' ggplot(mountain, aes(x, y, z)) + geom_contour()
#'
#' @seealso \code{\link[ggplot2]{aes}} for the original aesthetic mapping function,
#'   \code{\link{coord_3d}} for 3D coordinate systems
#' @export
aes <- function(x, y, z, ...) {
      if (missing(z)) {
            # 2D case - pass through to ggplot2::aes with positional x, y
            ggplot2::aes({{x}}, {{y}}, ...)
      } else {
            # 3D case - map x, y, z positionally and pass other aesthetics
            ggplot2::aes(x = {{x}}, y = {{y}}, z = {{z}}, ...)
      }
}

#' Using themes to style 3D panels and axis labels
#'
#' In ggcube, standard ggplot2 themes generally influence 3D plots as expected, including adding
#' complete themes like `ggplot2::theme_dark()` and modifying theme elements like
#' `theme(panel.background = element_rect(fill = "darkblue")`. However, ggcube also
#' provides additional theme elements that control 3D-specific styling of panels
#' and labels.
#'
#' @section Text elements:
#' - `axis.text.z`: Styling for z-axis tick labels (inherits from `axis.text`)
#' - `axis.title.z`: Styling for z-axis title (inherits from `axis.title`)
#' - `axis.text`, `axis.title`: Standard styling with `element_text()`.
#'
#' Use `element_text(margin = margin(...))` to adjust text padding, with left/right
#' margins affecting axis text and top/bottom margins affecting axis titles;
#' since placement and justification of these elements varies dynamically,
#' no distinction is made between left and right margins, or between top and
#' bottom margins -- you can set either, and the maximum of the two will be used.
#'
#' @section Axis ticks:
#' - `axis.ticks.z`: Styling for z-axis ticks (inherits from `axis.ticks`)
#' - `axis.ticks`, `axis.ticks.x`, `axis.ticks.y`: Standard styling with `element_line()`.
#' - `axis.ticks.length`, `axis.ticks.length.x/y/z`: Tick length.
#'
#' Ticks are 3D geometry rather than screen-space annotation: each one continues
#' its gridline past the cube edge, and is projected like everything else. They
#' therefore foreshorten with the view, and a tick pointing away from the viewer
#' is drawn shorter than one lying across it.
#'
#' Because of this, `axis.ticks.length` is an upper bound rather than an exact
#' length. It is calibrated so that a tick along the least foreshortened axis
#' renders at the requested size, with ticks along the other axes coming in
#' shorter. Tick length does not vary with axis `ratio`; a stretched axis gets
#' the same ticks as a compressed one.
#'
#' As in 2D ggplot2, `axis.ticks.length` also contributes to the spacing between
#' the cube and the axis text, whether or not ticks are drawn. Setting
#' `axis.ticks = element_blank()` removes the ticks but leaves that spacing
#' intact; set `axis.ticks.length = unit(0, "pt")` to close the gap as well.
#' Negative lengths point ticks inward, and do not displace the axis text.
#'
#' @section Panel elements:
#' - `panel.foreground`: Styling for cube faces rendered in front of data (inherits from `panel.background`).
#'          Foreground panels default to 20% opacity to keep them from obscuring the data; this can be
#'          overridden by setting `alpha` on either `panel.foreground` or `panel.background`.
#' - `panel.border.foreground`: Styling for cube faces rendered in front of data (inherits from `panel.border`)
#' - `panel.grid.foreground`: Styling for grid lines on foreground faces (inherits from `panel.grid`)
#' - `panel.grid.major.foreground`: Major grid lines on foreground faces (inherits from `panel.grid.foreground`)
#'
#' Background panels use standard `panel.background`, `panel.border`, `panel.grid`, etc., while foreground panels
#' use the `*.foreground` variants listed above. Since the foreground elements inherit from the standard background
#' and grid elements, you can use `panel.background`, etc. to style both background and foreground faces simultaneously.
#' This also extends to `alpha` set via the enhanced `element_rect()` below: setting `alpha` on `panel.background`
#' will tint both background and foreground panels (overriding the foreground's 20% default), unless an explicit
#' `alpha` is also set on `panel.foreground`.
#'
#' @section Depth scaling:
#' Under perspective projection, gridlines, borders, ticks, and axis text are (by default)
#' sized by their distance from the viewer. An element at the center of the cube
#' renders at exactly the size the theme specifies, with nearer elements drawn
#' larger and farther ones smaller. Axis titles are never scaled this way, and
#' orthographic plots are unaffected. You can use `coord_3d(scale_depth = )` to reduce or
#' disable the depth-scaling effect, per element or globally.
#'
#' @section Enhanced elements:
#' - `element_rect()` extends `ggplot2::element_rect()` by adding an `alpha` parameter for transparency effects
#' on `coord_3d()`'s foreground and background panels.
#'
#' @examples
#' # example code
#' p <- ggplot(sphere_points, aes(x, y, z)) +
#'   geom_hull_3d() +
#'   coord_3d(panels = "all") +
#'   theme(panel.background = element_rect(color = "black"),
#'           panel.border = element_rect(color = "black"),
#'           panel.foreground = element_rect(alpha = .3),
#'           panel.grid.foreground = element_line(color = "gray", linewidth = .25),
#'           axis.text = element_text(color = "darkblue"),
#'           axis.text.z = element_text(color = "darkred"),
#'           axis.title = element_text(margin = margin(t = 30)), # add padding
#'           axis.title.x = element_text(color = "magenta"))
#'
#' @name cube_theming
NULL


#' Enhanced rectangle element with alpha support
#'
#' This function extends \code{ggplot2::element_rect()} to support transparency
#' via an \code{alpha} parameter, which is attached to the returned element as
#' an R attribute (\code{"ggcube_alpha"}). The attribute survives ggplot2's
#' theme inheritance machinery, so the alpha value set here is available to
#' the renderer at draw time. The \code{alpha} parameter is consumed only by
#' \code{coord_3d()}'s panel rendering and has no effect on other theme
#' elements.
#'
#' @param fill Fill colour for the rectangle.
#' @param alpha Transparency level applied to the fill of foreground or
#'   background panels in \code{coord_3d()}, ranging from 0 (completely
#'   transparent) to 1 (completely opaque). For \code{panel.foreground}, the
#'   alpha resolves in order: an explicit value here, then alpha inherited
#'   from \code{panel.background}, then a default of 0.2. For
#'   \code{panel.background}, an explicit value is used directly; otherwise
#'   the panel is fully opaque. Has no effect on other theme elements or in
#'   plots without a 3D coordinate system.
#' @param ... Additional arguments passed to \code{\link[ggplot2]{element_rect}},
#'   such as \code{colour}, \code{linewidth}, \code{linetype}, and
#'   \code{inherit.blank}.
#'
#' @return A \code{ggplot2::element_rect} theme element with an attached
#'   \code{"ggcube_alpha"} attribute.
#'
#' @examples
#' # Basic 3D plot with semi-transparent foreground panels
#' ggplot(mountain, aes(x, y, z)) +
#'   stat_surface_3d(fill = "darkblue", color = "lightblue", linewidth = .1) +
#'   coord_3d(panels = c("background", "ymin")) +
#'   theme(panel.foreground = element_rect(alpha = 0.6))
#'
#' # Completely transparent foreground panels
#' ggplot(mtcars, aes(mpg, wt, qsec)) +
#'   geom_point() +
#'   coord_3d(panels = "all") +
#'   theme(panel.border = element_rect(color = "black"),
#'         panel.foreground = element_rect(fill = "blue", alpha = 0))
#'
#' @seealso \code{\link[ggplot2]{element_rect}} for the original function,
#'   \code{\link{coord_3d}} for 3D coordinate systems that utilize foreground panels
#' @export
element_rect <- function(fill = NULL, alpha = NA, ...) {
      el <- ggplot2::element_rect(fill = fill, ...)
      attr(el, "ggcube_alpha") <- alpha
      el
}

test_that("element_rect() returns a valid ggplot2 element", {
      el <- element_rect(fill = "blue")
      expect_s3_class(el, "element_rect")
      expect_s3_class(el, "element")
      expect_no_error(ggplot2::element_grob(el))
})

test_that("element_rect() attaches alpha as an attribute", {
      expect_equal(attr(element_rect(alpha = 0.5), "ggcube_alpha"), 0.5)
      expect_equal(attr(element_rect(fill = "blue", alpha = 0.3), "ggcube_alpha"), 0.3)
      # No alpha specified: attribute defaults to NA so the renderer can
      # detect "use the package default" via the same code path that handles
      # inherited elements stripped of their attribute by merge_element().
      expect_true(is.na(attr(element_rect(fill = "blue"), "ggcube_alpha")))
})

test_that("element_rect() honours the color/colour alias", {
      expect_equal(element_rect(color = "blue")[["colour"]], "blue")
})

#' 3D coordinate system
#'
#' \code{coord_3d} is a 3D coordinate system that creates a 2D view of 3D data.
#' This is the essential core component of any plot made with `ggcube`.
#' It supports rotation, perspective projection, and options for controlling plot
#' aspect ratios, panel selection, axis label placement, and lighting.
#'
#' @param roll,pitch,yaw Rotation around x, y, and z axes, respectively, in degrees.
#'    Positive values rotate the near face of the plot "downward", "rightward", and clockwise, respectively.
#' @param persp Logical indicating whether to apply perspective projection. When \code{TRUE} (the default),
#'   objects farther from the viewer appear smaller. When \code{FALSE}, produces an orthographic projection in which
#'   lines that are parallel in 3D space render as parallel in the plot.
#' @param dist Distance from viewer to center of the data cube. Only used when \code{persp = TRUE}.
#'   Larger values create less perspective distortion. Default is 2.
#'   Values less than 1 are allowed but can be problematic for rendering.
#' @param zoom Numeric value controlling the framing of the plot. Values greater
#'   than 1 zoom in (tighter framing, may crop edges), values less than 1
#'   zoom out (more whitespace around the plot). Default is 1.
#' @param expand Logical indicating whether to expand axis ranges beyond the data range,
#'   similar to standard ggplot2 behavior. If \code{TRUE} (the default), expansion behavior
#'   can be controlled using standard axis scaling functions, e.g.
#'   \code{... + scale_x_continuous(expand = expansion(.5))}.
#' @param clip Character string indicating clipping behavior. Use \code{"off"} (the default, recommended
#'   for some 3D plots) to allow drawing outside the plot panel.
#' @param panels Character vector specifying which panels to render, including one or more of the following:
#'   \itemize{
#'     \item \code{"background"} (the default), \code{"foreground"}: faces laying behind or in front of
#'       the cube's interior volume, respectively. These panels vary depending on plot rotation.
#'     \item \code{"xmin"}, \code{"ymax"}, etc.: names of specific cube faces.
#'     \item \code{"all"}, \code{"none"}: display the full cube or remove all faces.
#'   }
#'   See [cube_theming] for details on panel styling, including transparency of foreground panels.
#' @param xlabels,ylabels,zlabels Character strings or length-2 character vectors specifying
#'   axis label (text and title) placement. Labels are placed inline with grid lines for
#'   the selected panel face. For each axis, there are four potential panels where labels
#'   could be placed, and two potential edges for each panel. Labels can only be placed on
#'   visible faces (see `panels` argument). Each parameter accepts:
#'   \itemize{
#'     \item \code{"auto"} (default): Automatic edge selection based on an algorithm
#'       that prioritizes edges that are visible on the periphery of the plot and considers
#'       several attributes of face geometry for better readability.
#'     \item \code{c("face1", "face2")}: Manual edge specification using two adjacent
#'       face names (e.g., \code{c("xmin", "ymin")} selects the edge shared by the
#'       xmin and ymin faces). The \strong{first face} in the vector determines which
#'       face the axis labels will be aligned with, while the second face
#'       identifies which edge of this face gets labelled. Available face names are:
#'       "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
#'   }
#'   See [cube_theming] for details on axis label styling.
#' @param title_position Character string controlling axis title placement. Currently only
#'    affects titles for internal axes (not on the plot periphery); titles for axes on the
#'    plot periphery are always centered along the axis edge.
#'   \itemize{
#'     \item \code{"auto"} (default): Internal axis titles are placed at the near
#'       end of the axis, outside the plot area.
#'     \item \code{"center"}: Internal axis titles are centered along the axis edge.
#'   }
#' @param rotate_labels Logical indicating whether axis labels (text and titles) should automatically
#'   rotate to align with the projected axis directions. When \code{FALSE}, uses theme
#'   text and title angle settings.
#' @param scale_depth Controls how axis-related elements scale with viewing
#'   distance under perspective projection, in which objects nearer the viewer are drawn
#'   larger. Accepts:
#'   \itemize{
#'     \item \code{TRUE} (default) or \code{FALSE}: full depth scaling, or none.
#'     \item A single number: applied to every element. \code{1} is the default
#'       strength, \code{0} disables scaling, values below 1 subdue the effect and
#'       values above 1 exaggerate it. Negative values are an error.
#'     \item A named numeric vector, to set elements individually. Names are
#'       \code{"grid"} (panel gridlines), \code{"border"} (cube edges), and
#'       \code{"text"} (axis tick labels), after the theme elements they govern.
#'       Unnamed elements keep their default.
#'   }
#'   Strength \code{s} maps a depth factor \code{d} to \code{d^s}, so the effect
#'   scales geometrically and never produces a non-positive size. This affects axis-related
#'   elements only; layers have their own \code{scale_depth} parameters. Axis
#'   titles are never depth-scaled, since a single title has no series of sizes to
#'   read as perspective. Has no effect when \code{persp = FALSE}, where all depth
#'   factors are 1.
#' @param scales Character string specifying aspect ratio behavior:
#'   \itemize{
#'     \item \code{"free"} (default): Each axis scales independently to fill cube space,
#'       then \code{ratio} applies to standardized coordinates. This gives maximum
#'       visual range for each dimension.
#'     \item \code{"fixed"}: Maintains proportional relationships in raw data values,
#'       as scaled by \code{ratio}. Similar to \code{coord_fixed()} but for 3D
#'       (visual ratios match the labeled axis ranges).
#'   }
#' @param ratio Numeric vector of length 3 specifying relative axis lengths as
#'   \code{c(x, y, z)}. Defaults to \code{c(1, 1, 1)} for equal proportions.
#'   \itemize{
#'     \item With \code{scales = "free"}: Ratios apply to scaled cube coordinates
#'     \item With \code{scales = "fixed"}: Ratios apply to original data coordinates
#'   }
#' @param light Lighting specification for the plot. Usually left unset: add
#'   [light()] to the plot instead, which is the recommended way to set plot-level
#'   lighting and can appear anywhere in the plot expression. Supplying `light`
#'   here as well as adding [light()] to the plot is an error. Use `NULL` to
#'   disable lighting, or `light("none")` for an equivalent lighting object.
#' @param ... Additional arguments reserved for internal use.
#'
#' @examples
#' # base plot used in examples
#' p <- ggplot() +
#'   geom_function_3d(
#'     aes(fill = after_stat(z), color = after_stat(z)),
#'     fun = function(x, y) sin(x) * cos(y),
#'     xlim = c(-pi, pi), ylim = c(-pi, pi), n = 50) +
#'   light("direct", contrast = .7) +
#'   scale_fill_viridis_c() +
#'   scale_color_viridis_c() +
#'   theme(legend.position = "none")
#'
#' # 3D plot with default coord settings
#' p + coord_3d()
#'
#' \donttest{
#'
#' # Use `pitch`, `roll`, `yaw` to control plot rotation ----------------------
#'
#' # zero rotation gives view from x-y face
#' p + coord_3d(pitch = 0, roll = 0, yaw = 0)
#'
#' # pitch rotates plot around y axis
#' p + coord_3d(pitch = 30, roll = 0, yaw = 0)
#'
#' # roll rotates plot around x axis
#' p + coord_3d(pitch = 0, roll = 30, yaw = 0)
#'
#' # yaw rotates plot around z axis
#' p + coord_3d(pitch = 0, roll = 0, yaw = 30)
#'
#' # combine them to achieve arbitrary rotations
#' p + coord_3d(pitch = 20, roll = 40, yaw = 60)
#'
#'
#' # Use `persp` and `dist` to control perspective effects --------------------
#'
#' # strong perspective effect as if seen from very close
#' p + coord_3d(dist = 1)
#'
#' # weaker perspective effects as if seen from far away
#' p + coord_3d(dist = 3)
#'
#' # orthographic projection (`dist = Inf` would be equivalent but it errors)
#' p + coord_3d(persp = FALSE)
#'
#'
#' # Use `scales` and `ratio` to control aspect ratio -------------------------
#'
#' # The default "free" scales shown above give cube with maximum visual range.
#' # Use "fixed" scales to make figure match data scales, like coord_fixed.
#' p + coord_3d(scales = "fixed")
#'
#' # Custom aspect ratios: make y twice as long visually
#' p + coord_3d(scales = "free", ratio = c(1, 2, 1))
#'
#' # Combine behaviors: fix scales and make y twice as long
#' p + coord_3d(scales = "fixed", ratio = c(1, 2, 1))
#'
#'
#' # Use `panels` to select which cube faces to render ------------------------
#'
#' p + coord_3d(panels = c("xmin", "xmax", "zmax"))
#'
#' # foreground panels default to 20% opaque (can be styled using `theme()`)
#' p + coord_3d(panels = "all")
#'
#'
#' # Use label params to control axis text placement and rotation -------------
#'
#' p + coord_3d(xlabels = c("ymax", "zmax"),
#'              zlabels = c("xmax", "ymin"))
#'
#' p + coord_3d(rotate_labels = FALSE)
#'
#' }
#'
#' @return A `Coord` object that can be added to a ggplot.
#' @seealso [light()] for lighting specification, [cube_theming] for panel and
#'   text styling, [polygon_params] for 3D-related parameters for polygon layers.
#' @export
coord_3d <- function(pitch = 0, roll = -60, yaw = -30,
                     persp = TRUE, dist = 2,
                     expand = TRUE, clip = "off",
                     panels = "background",
                     xlabels = "auto", ylabels = "auto", zlabels = "auto",
                     title_position = c("auto", "center"),
                     rotate_labels = TRUE,
                     scales = "free",
                     scale_depth = TRUE,
                     ratio = c(1, 1, 1),
                     zoom = 1,
                     light = waiver(),
                     ...) {

      # Capture internal-only params from ...
      dots <- list(...)
      fixed_bounds <- dots$fixed_bounds

      # Validate parameters
      if (!scales %in% c("free", "fixed")) {
            stop("scales must be 'free' or 'fixed'")
      }

      if (!is.numeric(ratio) || length(ratio) != 3 || any(ratio <= 0)) {
            stop("ratio must be a positive numeric vector of length 3")
      }

      if (!is.numeric(zoom) || length(zoom) != 1 || zoom <= 0) {
            stop("zoom must be a positive numeric value")
      }

      if(persp && dist < 1) {
            warning("Although `dist` values less than 1 are allowed, they often produce nonsensical plots.")
      }

      title_position <- match.arg(title_position)

      scale_depth <- resolve_scale_depth(scale_depth)

      list(
            ggproto(NULL, Coord3D,
                    pitch = pitch, roll = roll, yaw = yaw,
                    persp = persp, dist = dist,
                    expand = expand, clip = clip,
                    panels = panels,
                    rotate_labels = rotate_labels,
                    scales = scales,
                    scale_depth = scale_depth,
                    ratio = ratio,
                    zoom = zoom,
                    xlabels = xlabels, ylabels = ylabels, zlabels = zlabels,
                    light = light,
                    light_explicit = !inherits(light, "waiver"),
                    title_position = title_position,
                    fixed_bounds = fixed_bounds
            ),
            theme(plot.margin = margin(20, 20, 20, 20, "pt"))
      )
}

# Depth scaling -----------------------------------------------------------

# Axis furniture whose size responds to viewing distance, named after the
# theme elements they govern. Axis titles are deliberately absent.
ggcube_depth_elements <- c("grid", "border", "ticks", "text")

# Apply a strength exponent to a depth factor.
#
# depth_scale is a ratio (reference distance over actual distance), so
# interpolating it geometrically rather than linearly keeps the result
# positive for every strength and makes near and far effects symmetric.
# Strength 1 is the identity; 0 disables scaling.
apply_depth_strength <- function(depth_scale, strength = 1) {
      if (is.null(strength) || length(strength) != 1 || !is.finite(strength)) {
            strength <- 1
      }
      if (strength == 1) return(depth_scale)

      result <- pmax(depth_scale, 0) ^ strength
      result[!is.finite(result)] <- 1
      result
}

# Look up one element's strength on a coord, tolerating coords built before
# this parameter existed.
depth_strength <- function(coord, element) {
      strengths <- coord$scale_depth
      if (is.null(strengths)) return(1)

      value <- unname(strengths[element])
      if (length(value) != 1 || is.na(value)) return(1)
      value
}

# Guard a computed linewidth against non-finite or negative values.
safe_lwd <- function(lwd, fallback) {
      invalid <- !is.finite(lwd) | lwd < 0
      if (any(invalid)) lwd[invalid] <- fallback
      lwd
}

# Normalise the user-facing `scale_depth` argument into a named numeric
# vector covering every element.
resolve_scale_depth <- function(scale_depth) {
      defaults <- stats::setNames(rep(1, length(ggcube_depth_elements)),
                                  ggcube_depth_elements)

      if (is.null(scale_depth)) return(defaults)

      if (is.logical(scale_depth)) {
            if (length(scale_depth) != 1 || is.na(scale_depth)) {
                  rlang::abort("`scale_depth` must be TRUE, FALSE, a single number, or a named numeric vector.")
            }
            return(stats::setNames(rep(as.numeric(scale_depth), length(defaults)),
                                   names(defaults)))
      }

      if (!is.numeric(scale_depth) || length(scale_depth) == 0) {
            rlang::abort("`scale_depth` must be TRUE, FALSE, a single number, or a named numeric vector.")
      }

      if (any(!is.finite(scale_depth))) {
            rlang::abort("`scale_depth` values must be finite.")
      }

      if (any(scale_depth < 0)) {
            rlang::abort(c(
                  "`scale_depth` values must be non-negative.",
                  i = "Negative values would invert the depth cue, drawing distant elements larger than near ones."
            ))
      }

      if (is.null(names(scale_depth))) {
            if (length(scale_depth) != 1) {
                  rlang::abort(c(
                        "`scale_depth` must be a single number, or a named numeric vector.",
                        i = paste0("Valid names are ",
                                   paste0("`", ggcube_depth_elements, "`", collapse = ", "), ".")
                  ))
            }
            return(stats::setNames(rep(scale_depth, length(defaults)), names(defaults)))
      }

      if (any(names(scale_depth) == "")) {
            rlang::abort("`scale_depth` must be either fully named or a single unnamed number.")
      }

      if (anyDuplicated(names(scale_depth)) > 0) {
            rlang::abort("`scale_depth` has duplicate names.")
      }

      unknown <- setdiff(names(scale_depth), ggcube_depth_elements)
      if (length(unknown) > 0) {
            rlang::abort(c(
                  paste0("Unknown `scale_depth` element", if (length(unknown) > 1) "s" else "",
                         ": ", paste0("`", unknown, "`", collapse = ", "), "."),
                  i = paste0("Valid names are ",
                             paste0("`", ggcube_depth_elements, "`", collapse = ", "), ".")
            ))
      }

      defaults[names(scale_depth)] <- scale_depth
      defaults
}


#' Detect if a scale transformation flips direction
#'
#' @param scale_obj A ggplot2 scale object
#' @return Logical indicating if the scale transform flips direction
#' @keywords internal
#' @noRd
detect_scale_direction_flip <- function(scale_obj) {
      if (is.null(scale_obj)) return(FALSE)

      transform <- scale_obj$trans
      if (is.null(transform)) return(FALSE)

      test_output <- transform$transform(c(1, 2))

      # If we can't compare the outputs, assume no flip
      if (length(test_output) != 2 || any(is.na(test_output))) {
            return(FALSE)
      }

      return(test_output[1] > test_output[2])
}

#' Translate face names to account for scale direction flips
#'
#' @param face_names Character vector of face names
#' @param scale_flips Scales flipped for each axis
#' @return Character vector of translated face names
#' @keywords internal
#' @noRd
translate_face_names_from_flips <- function(face_names, scale_flips) {
      if (length(face_names) == 0) return(face_names)

      # Only translate actual face names, leave other values unchanged
      valid_faces <- c("xmin", "xmax", "ymin", "ymax", "zmin", "zmax")
      if (!any(face_names %in% valid_faces)) {
            return(face_names)  # No face names to translate
      }

      # Get flip indicators
      x_flipped <- scale_flips$x
      y_flipped <- scale_flips$y
      z_flipped <- scale_flips$z

      # Create translation mapping
      translated <- face_names

      for (i in seq_along(translated)) {
            face <- translated[i]

            if (x_flipped && face == "xmin") {
                  translated[i] <- "xmax"
            } else if (x_flipped && face == "xmax") {
                  translated[i] <- "xmin"
            } else if (y_flipped && face == "ymin") {
                  translated[i] <- "ymax"
            } else if (y_flipped && face == "ymax") {
                  translated[i] <- "ymin"
            } else if (z_flipped && face == "zmin") {
                  translated[i] <- "zmax"
            } else if (z_flipped && face == "zmax") {
                  translated[i] <- "zmin"
            }
      }

      return(translated)
}

translate_face_names <- function(faces, scale_x, scale_y, scale_z) {
      if (length(faces) == 0) return(faces)

      # Detect direction flips for each axis
      flips <- list(
            x = detect_scale_direction_flip(scale_x),
            y = detect_scale_direction_flip(scale_y),
            z = detect_scale_direction_flip(scale_z)
      )

      # Use the same logic as the flips version
      return(translate_face_names_from_flips(faces, flips))
}

Coord3D <- ggproto("Coord3D", CoordCartesian,
                   # Parameters
                   pitch = 0,
                   roll = 120,
                   yaw = 30,
                   persp = FALSE,
                   dist = 2,
                   expand = TRUE,
                   clip = "off",
                   panels = "background",
                   rotate_labels = TRUE,
                   scales = "free",
                   ratio = c(1, 1, 1),
                   zoom = 1,
                   light = waiver(),
                   light_explicit = FALSE,

                   plot_bounds = c(0, 1, 0, 1),  # [xmin, xmax, ymin, ymax]

                   fixed_bounds = NULL, # set by animate_3d() to lock bounds

                   setup_panel_params = function(self, scale_x, scale_y, params = list()) {

                         # Check if theme is void-like and override panels if so
                         theme_obj <- NULL
                         tryCatch({
                               for (i in 1:25) {
                                     env <- parent.frame(i)
                                     if (exists("theme", envir = env)) {
                                           potential_theme <- get("theme", envir = env)
                                           if (is.list(potential_theme)) {
                                                 theme_obj <- potential_theme
                                                 break
                                           }
                                     }
                               }
                         }, error = function(e) {
                               # Ignore errors - theme_obj will remain NULL
                         })

                         # Override panels to "none" if theme_void-like
                         original_panels <- self$panels
                         if (!is.null(theme_obj) && is_theme_void_like(theme_obj)) {
                               self$panels <- "none"
                         }

                         # Get standard panel params from parent
                         panel_params <- ggproto_parent(CoordCartesian, self)$setup_panel_params(scale_x, scale_y, params)

                         # Recover and train the plot's own z scale
                         scale_z <- find_plot_z_scale()
                         if (is.null(scale_z)) {
                               # The plot has no z scale, which happens when z is
                               # stat-computed rather than mapped (e.g. with
                               # stat_function_3d()) and so ggplot2 never added a
                               # default. Build one for this panel.
                               scale_z <- scale_z_continuous()
                         }
                         train_z_scale(scale_z)

                         # Translate face names to account for scale direction flips
                         self$panels <- translate_face_names(self$panels, scale_x, scale_y, scale_z)
                         self$xlabels <- translate_face_names(self$xlabels, scale_x, scale_y, scale_z)
                         self$ylabels <- translate_face_names(self$ylabels, scale_x, scale_y, scale_z)
                         self$zlabels <- translate_face_names(self$zlabels, scale_x, scale_y, scale_z)

                         # Store scale flip information for later use in transform()
                         panel_params$scale_flips <- list(
                               x = detect_scale_direction_flip(scale_x),
                               y = detect_scale_direction_flip(scale_y),
                               z = detect_scale_direction_flip(scale_z)
                         )

                         # Scale info (including axis names)
                         panel_params$scales <- self$scales
                         panel_params$scale_info <- list(
                               x = get_scale_info(scale_x, expand = self$expand, axis_name = "x"),
                               y = get_scale_info(scale_y, expand = self$expand, axis_name = "y"),
                               z = get_scale_info(scale_z, expand = self$expand, axis_name = "z")
                         )

                         # Blank out native scale names (prevents them from showing in standard ggplot2 rendering)
                         scale_x$name <- ""
                         scale_y$name <- ""

                         # Projection info
                         panel_params$proj <- list(pitch = self$pitch, roll = self$roll, yaw = self$yaw,
                                                   persp = self$persp, dist = self$dist)

                         # Aspect ratio info
                         panel_params$ratio <- self$ratio
                         effective_ratios <- compute_effective_ratios(
                               list(x = panel_params$scale_info$x$limits,
                                    y = panel_params$scale_info$y$limits,
                                    z = panel_params$scale_info$z$limits),
                               panel_params$scales,
                               panel_params$ratio
                         )
                         panel_params$effective_ratios <- effective_ratios

                         # Visible faces (using translated panel names)
                         visible_faces_fgbg <- select_visible_faces(self$panels, panel_params$proj, effective_ratios)
                         visible_faces <- do.call("c", visible_faces_fgbg)
                         panel_params$visible_faces <- visible_faces
                         panel_params$visible_faces_fg <- visible_faces_fgbg$fg
                         panel_params$visible_faces_bg <- visible_faces_fgbg$bg

                         # Calculate plot bounds using SCALE BREAKS and TITLE POSITIONS
                         if (length(visible_faces) > 0) {
                               all_faces <- c("xmin", "xmax", "ymin", "ymax", "zmin", "zmax")
                               full_grid <- make_scale_grid(all_faces, panel_params$scale_info,
                                                            panel_params$scales, panel_params$ratio)

                               if (!is.null(full_grid)) {
                                     # Transform the FULL grid to get base bounds
                                     full_grid_transformed <- transform_3d_standard(full_grid, panel_params$proj)

                                     # Calculate bounds including potential title positions
                                     all_bounds_x <- full_grid_transformed$x
                                     all_bounds_y <- full_grid_transformed$y

                                     bounds_info <- calculate_plot_bounds(all_bounds_x, all_bounds_y)
                                     panel_params$plot_bounds <- bounds_info$bounds
                                     self$bounds_aspect <- bounds_info$aspect
                               } else {
                                     panel_params$plot_bounds <- c(-1, 1, -1, 1)
                               }

                               # Override bounds if fixed (e.g. during animation)
                               if (!is.null(self$fixed_bounds)) {
                                     panel_params$plot_bounds <- self$fixed_bounds
                                     self$bounds_aspect <- diff(self$fixed_bounds[3:4]) /
                                           diff(self$fixed_bounds[1:2])
                               }

                               # Apply zoom: scale bounds around center
                               # zoom < 1 zooms in (tighter framing), zoom > 1 zooms out (more padding)
                               if (!is.null(self$zoom) && self$zoom != 1) {
                                     bounds <- panel_params$plot_bounds
                                     x_center <- mean(bounds[1:2])
                                     y_center <- mean(bounds[3:4])
                                     x_half <- diff(bounds[1:2]) / 2 / self$zoom
                                     y_half <- diff(bounds[3:4]) / 2 / self$zoom
                                     panel_params$plot_bounds <- c(
                                           x_center - x_half, x_center + x_half,
                                           y_center - y_half, y_center + y_half
                                     )
                                     # Aspect ratio doesn't change with uniform zoom
                               }

                               # Generate grid for selected faces using real scale breaks
                               if (length(visible_faces) > 0) {
                                     selected_grid <- make_scale_grid(visible_faces, panel_params$scale_info,
                                                                      panel_params$scales, panel_params$ratio)

                                     if (!is.null(selected_grid)) {
                                           selected_grid_transformed <- transform_3d_standard(selected_grid, panel_params$proj)

                                           # transform_3d_standard() overwrites x/y/z in place, so keep
                                           # the pre-projection copy: axis ticks are built in cube space
                                           # and need the unprojected break coordinates.
                                           panel_params$grid_standard <- selected_grid

                                           panel_params$grid_transformed <- selected_grid_transformed
                                           panel_params$grid_transformed$face <- selected_grid$face
                                           panel_params$grid_transformed$group <- selected_grid$group
                                           panel_params$grid_transformed$z_proj <- selected_grid_transformed$z
                                           panel_params$grid_transformed$depth_scale <- selected_grid_transformed$depth_scale
                                           panel_params$grid_transformed$break_value <- selected_grid$break_value
                                           panel_params$grid_transformed$break_pos <- selected_grid$break_pos
                                           panel_params$grid_transformed$break_axis <- selected_grid$break_axis
                                           panel_params$grid_transformed$start_boundaries <- selected_grid$start_boundaries
                                           panel_params$grid_transformed$end_boundaries <- selected_grid$end_boundaries
                                     } else {
                                           panel_params$grid_transformed <- NULL
                                           panel_params$grid_standard <- NULL
                                     }
                               } else {
                                     # No visible faces - no grid to render
                                     panel_params$grid_transformed <- NULL
                                     panel_params$grid_standard <- NULL
                               }
                         } else {
                               # No visible faces - use minimal bounds (just cube corners)
                               aspect_cube <- data.frame(
                                     x = c(-0.5, 0.5, -0.5, 0.5, -0.5, 0.5, -0.5, 0.5) * effective_ratios[1],
                                     y = c(-0.5, -0.5, 0.5, 0.5, -0.5, -0.5, 0.5, 0.5) * effective_ratios[2],
                                     z = c(-0.5, -0.5, -0.5, -0.5, 0.5, 0.5, 0.5, 0.5) * effective_ratios[3]
                               )
                               cube_transformed <- transform_3d_standard(aspect_cube, panel_params$proj)

                               bounds_info <- calculate_plot_bounds(cube_transformed$x, cube_transformed$y)
                               panel_params$plot_bounds <- bounds_info$bounds
                               self$bounds_aspect <- bounds_info$aspect
                               panel_params$grid_transformed <- NULL
                               panel_params$grid_standard <- NULL
                         }

                         # Resolve lighting. `coord_3d(light = )` is explicit and
                         # takes precedence; otherwise a plot-level `+ light()`
                         # applies; otherwise the package default. `NULL` is
                         # preserved as an explicit request for no lighting.
                         # The resolved spec goes to `panel_params`, never back
                         # to `self`: the coord is shared by reference with every
                         # plot derived from this one, and overwriting the
                         # `waiver()` sentinel would make an inherited
                         # `+ light()` indistinguishable from an explicit
                         # `coord_3d(light = )` on any subsequent build.
                         pending <- find_pending_light()
                         if (!is.null(pending) && isTRUE(self$light_explicit)) {
                               abort_double_light()
                         }
                         effective_light <- if (is.null(pending)) self$light else pending
                         panel_params$light <- resolve_light(effective_light)

                         return(panel_params)
                   },

                   # Force 1:1 aspect ratio
                   aspect = function(self, ranges) {
                         if (!is.null(self$fixed_bounds)) {
                               return(diff(self$fixed_bounds[3:4]) / diff(self$fixed_bounds[1:2]))
                         }
                         if (!is.null(self$bounds_aspect)) {
                               return(self$bounds_aspect)
                         }
                         return(1)
                   },

                   # Handle back transformation
                   backtransform_range = function(self, panel_params) {
                         return(list(x = c(0, 1), y = c(0, 1)))
                   },

                   render_bg = function(self, panel_params, theme) {
                         # Store theme element states in panel_params for render_cube to use
                         panel_params$theme_elements <- list(
                               show_background_panels = !inherits(calc_element("panel.background", theme), "element_blank"),
                               show_grid = !inherits(calc_element("panel.grid", theme), "element_blank") &&
                                     !inherits(calc_element("panel.grid.major", theme), "element_blank"),
                               show_axis_text = !inherits(calc_element("axis.text", theme), "element_blank"),
                               show_axis_title = !inherits(calc_element("axis.title", theme), "element_blank")
                         )

                         render_cube(self, panel_params, theme, layer = "background")
                   },

                   transform = function(self, data, panel_params) {

                         # Add light specs if applicable
                         data <- attach_light(data, panel_params$light)

                         # Restore z=0 for baseline vertices, if applicable (stat_bar_3d)
                         if ("z0" %in% names(data)) {
                               data$z[data$z0] <- 0
                         }

                         # Translate project_to_face names if they exist
                         if ("project_to_face" %in% names(data) && !is.null(panel_params$scale_flips)) {
                               data$project_to_face <- translate_face_names_from_flips(data$project_to_face, panel_params$scale_flips)
                         }

                         # Scale data to standard domain with aspect ratio
                         scale_ranges <- list(x = panel_params$scale_info$x$limits,
                                              y = panel_params$scale_info$y$limits,
                                              z = panel_params$scale_info$z$limits)
                         result <- scale_to_standard(data[c("x", "y", "z")], scale_ranges,
                                                     panel_params$scales, panel_params$ratio)

                         # lighting computation
                         if ("lighting_spec" %in% names(data)) {
                               data <- compute_light_in_coord(data, result, scale_ranges,
                                                              panel_params$scales, panel_params$ratio,
                                                              panel_params$proj)
                         }

                         # Project data onto cube face, if applicable
                         result <- project_to_face(data, result, panel_params$proj, panel_params$effective_ratios)

                         # Expand ref_circle points to circular polygons, if applicable
                         data <- points_to_circles(data, result, panel_params$effective_ratios)

                         # Apply 3D transformation (returns x, y, z, depth, depth_scale)
                         result <- transform_3d_standard(data, panel_params$proj)

                         # Combine transformed coordinates and additional variables
                         result <- bind_cols(select(result, x, y, z, depth, depth_scale),
                                             select(data, -x, -y, -z))

                         # Apply final coordinate transformation to fit plot bounds
                         result <- scale_to_npc_coordinates(result, plot_bounds = panel_params$plot_bounds)

                         # Backface culling and light modification
                         result <- process_backfaces(result)

                         # Hierarchical depth sorting
                         result <- sort_by_depth(result)

                         # convert group to integer to prevent downstream rendering errors
                         result$group <- as.integer(factor(result$group))

                         return(result)
                   },

                   render_fg = function(self, panel_params, theme) {
                         # Store theme element states for foreground rendering
                         panel_params$theme_elements$show_foreground_panels = !inherits(calc_element("panel.foreground", theme), "element_blank")

                         render_cube(self, panel_params, theme, layer = "foreground")
                   },

                   # Disable standard grids
                   render_grid = function(self, panel_params, theme) {
                         grid::nullGrob()
                   },

                   # Handle z-ordering
                   map_data = function(self, data, panel_params) {
                         data <- self$transform(data, panel_params)

                         # Sort by z-depth
                         if ("z_proj" %in% names(data)) {
                               if ("group" %in% names(data)) {
                                     data <- data[order(data$group, -data$z_proj), ]
                               } else {
                                     data <- data[order(-data$z_proj), ]
                               }
                         }

                         return(data)
                   },

                   # Suppress irrelevant warnings about axis guides
                   modify_scales = function(self, scales_x, scales_y) {
                         # Set guides to none to prevent position guide warnings
                         for (scale in scales_x) {
                               if (!is.null(scale)) {
                                     scale$guide <- "none"
                               }
                         }
                         for (scale in scales_y) {
                               if (!is.null(scale)) {
                                     scale$guide <- "none"
                               }
                         }
                         return(list(x = scales_x, y = scales_y))
                   }
)

# Helper function to compute plot bounds and aspect ratio
calculate_plot_bounds <- function(all_bounds_x, all_bounds_y){
      x_bounds <- range(all_bounds_x, na.rm = TRUE)
      y_bounds <- range(all_bounds_y, na.rm = TRUE)

      # Add padding
      x_padding <- diff(x_bounds) * 0.05
      y_padding <- diff(y_bounds) * 0.05
      x_bounds <- c(x_bounds[1] - x_padding, x_bounds[2] + x_padding)
      y_bounds <- c(y_bounds[1] - y_padding, y_bounds[2] + y_padding)

      # Calculate aspect ratio
      bounds_aspect <- diff(y_bounds) / diff(x_bounds)

      # Return both bounds and aspect ratio
      list(
            bounds = c(x_bounds[1], x_bounds[2], y_bounds[1], y_bounds[2]),
            aspect = bounds_aspect
      )
}


#' Extract variable names from aesthetic mappings
#'
#' @param plot_obj A ggplot object
#' @return A list with x, y, z character vectors of variable names
#' @keywords internal
#' @noRd
extract_aesthetic_vars <- function(plot_obj) {
      x_vars <- character(0)
      y_vars <- character(0)
      z_vars <- character(0)

      # Helper function to safely extract variables from a mapping
      extract_vars <- function(mapping) {
            vars <- list(x = character(0), y = character(0), z = character(0))

            tryCatch({
                  if (!is.null(mapping$x)) {
                        vars$x <- all.vars(mapping$x)
                  }
                  if (!is.null(mapping$y)) {
                        vars$y <- all.vars(mapping$y)
                  }
                  if (!is.null(mapping$z)) {
                        vars$z <- all.vars(mapping$z)
                  }
            }, error = function(e) {
                  # If all.vars() fails, ignore this mapping
            })

            return(vars)
      }

      # Plot-level mappings
      if (!is.null(plot_obj$mapping)) {
            plot_vars <- extract_vars(plot_obj$mapping)
            x_vars <- c(x_vars, plot_vars$x)
            y_vars <- c(y_vars, plot_vars$y)
            z_vars <- c(z_vars, plot_vars$z)
      }

      # Layer-level mappings
      if (!is.null(plot_obj$layers)) {
            for (layer in plot_obj$layers) {
                  if (!is.null(layer$mapping)) {
                        layer_vars <- extract_vars(layer$mapping)
                        x_vars <- c(x_vars, layer_vars$x)
                        y_vars <- c(y_vars, layer_vars$y)
                        z_vars <- c(z_vars, layer_vars$z)
                  }
            }
      }

      return(list(
            x = unique(x_vars),
            y = unique(y_vars),
            z = unique(z_vars)
      ))
}


#' Get axis name for a single scale with proper fallback hierarchy
#'
#' @param scale_obj Scale object (can be NULL for z-axis)
#' @param axis_name Axis name ("x", "y", or "z")
#' @return Single axis name string
#' @keywords internal
#' @noRd
get_scale_names <- function(scale_obj, axis_name) {

      # CAPTURE SCALE NAME BEFORE BLANKING IT OUT
      scale_name <- if (!is.null(scale_obj)) scale_obj$name %||% waiver() else waiver()

      # TRY TO FIND PLOT OBJECT AND EXTRACT BOTH LABELS AND AESTHETIC VARS
      plot_obj <- find_build_plot()
      plot_labels <- NULL
      aesthetic_vars <- NULL

      if (!is.null(plot_obj)) {
            plot_labels <- plot_obj$labels
            tryCatch({
                  aesthetic_vars <- extract_aesthetic_vars(plot_obj)
            }, error = function(e) {
                  # Ignore errors - will use defaults
            })
      }

      # RESOLVE FINAL NAME WITH FALLBACK HIERARCHY:
      # 1. Explicit scale name (from scale constructors like scale_x_continuous(name = "..."))
      # 2. Plot labels (automatic from aes() expressions OR user-set from labs())
      # 3. Simple aesthetic variable names (fallback if plot labels missing)
      # 4. Default name (axis_name)

      final_name <- axis_name  # default

      if (!inherits(scale_name, "waiver") && !is.null(scale_name) && scale_name != "") {
            # Priority 1: Explicit scale name
            final_name <- scale_name
      } else if (!is.null(plot_labels) && !is.null(plot_labels[[axis_name]]) && !inherits(plot_labels[[axis_name]], "waiver")) {
            # Priority 2: Plot labels (from aes expressions or labs())
            final_name <- plot_labels[[axis_name]]
      } else if (!is.null(aesthetic_vars) && length(aesthetic_vars[[axis_name]]) > 0) {
            # Priority 3: Aesthetic variable names
            final_name <- aesthetic_vars[[axis_name]][1]
      }

      return(final_name)
}

#' Find the z scale belonging to the plot being built
#'
#' The z scale rides on the plot like any other scale: `+ scale_z_continuous()`
#' lands in `plot$scales`, and ggplot2 constructs a default one itself when `z`
#' is mapped but no scale was supplied. `setup_panel_params()` is handed only
#' the x and y scales, so this reaches the plot to recover the z scale that
#' belongs to it. Because `ggplot_build()` clones the scales list before
#' building, the returned scale is a per-build object that is safe to train.
#'
#' @return A z scale object, or `NULL` if the plot has none.
#' @keywords internal
#' @noRd
find_plot_z_scale <- function() {
      p <- find_build_plot()
      if (is.null(p)) return(NULL)
      scales <- p$scales
      if (is.null(scales)) return(NULL)
      scales$get_scales("z")
}

#' Train a z scale on the layer data of the plot being built
#'
#' `Layout$setup_panel_params()` runs before ggplot2 trains non-position
#' scales, so the z scale still has an empty range when `coord_3d()` needs its
#' limits. The layer data has no channel into the coord either, so it too is
#' reached by walking parent frames.
#'
#' Trains on `zend`, `zmin` and a zero baseline where present, since those
#' aren't among the scale's aesthetics and so would never be trained by
#' ggplot2 itself.
#'
#' @param scale_z A z scale object.
#' @return Called for its side effect on `scale_z`.
#' @keywords internal
#' @noRd
train_z_scale <- function(scale_z){

      if (is.null(scale_z)) return(invisible(NULL))

      # Walk parent frames to find layer data
      data <- NULL
      tryCatch({
            for (i in 1:25) {
                  env <- parent.frame(i)
                  if (exists("data", envir = env)) {
                        potential_data <- get("data", envir = env)
                        if (is.list(potential_data) && length(potential_data) > 0 &&
                            is.data.frame(potential_data[[1]])) {
                              data <- potential_data
                              break
                        }
                  }
            }
      }, error = function(e) {})

      if (!is.null(data)) {
            for (layer_data in data) {

                  # Train on z column
                  if ("z" %in% names(layer_data) && nrow(layer_data) > 0) {

                        if(inherits(scale_z, "ScaleContinuousPosition")) {
                              scale_z$train(layer_data$z)
                        }

                        if(inherits(scale_z, "ScaleDiscretePosition")) {
                              if("z_raw" %in% names(layer_data)) {
                                    scale_z$range_c$train(layer_data$z)
                                    scale_z$train(layer_data$z_raw)
                              } else {
                                    scale_z$range_c$train(as.integer(factor(layer_data$z)))
                                    scale_z$train(layer_data$z)
                              }
                        }
                  }

                  # Train on zend column if present (for segment geoms)
                  if ("zend" %in% names(layer_data) && nrow(layer_data) > 0) {

                        if(inherits(scale_z, "ScaleContinuousPosition")) {
                              scale_z$train(layer_data$zend)
                        }

                        if(inherits(scale_z, "ScaleDiscretePosition")) {
                              if("zend_raw" %in% names(layer_data)) {
                                    scale_z$range_c$train(layer_data$zend)
                                    scale_z$train(layer_data$zend_raw)
                              } else {
                                    scale_z$range_c$train(as.integer(factor(layer_data$zend)))
                                    scale_z$train(layer_data$zend)
                              }
                        }
                  }

                  # Train on zmin column if present (for col geoms with variable baseline)
                  if ("zmin" %in% names(layer_data) && nrow(layer_data) > 0) {
                        if(inherits(scale_z, "ScaleContinuousPosition")) {
                              scale_z$train(layer_data$zmin)
                        }
                  }

                  # Train on 0 if z0 column exists (for bar geoms with fixed baseline)
                  if ("z0" %in% names(layer_data) && nrow(layer_data) > 0) {
                        if(inherits(scale_z, "ScaleContinuousPosition")) {
                              scale_z$train(0)
                        }
                  }
            }
      }

      invisible(NULL)
}

#' Check if theme is void-like (has multiple key elements set to element_blank)
#'
#' @param theme_obj Theme object or theme list
#' @return Logical indicating if theme appears to be void-like
#' @keywords internal
#' @noRd
is_theme_void_like <- function(theme_obj) {
      if (is.null(theme_obj)) return(FALSE)

      # Create a temporary theme to test elements
      temp_theme <- theme_obj

      # Count how many key theme elements are element_blank
      blank_count <- 0
      key_elements <- c("panel.background", "panel.grid", "panel.grid.major",
                        "axis.text", "axis.title")

      for (element_name in key_elements) {
            tryCatch({
                  element_val <- calc_element(element_name, temp_theme)
                  if (inherits(element_val, "element_blank")) {
                        blank_count <- blank_count + 1
                  }
            }, error = function(e) {
                  # If we can't evaluate the element, skip it
            })
      }

      # If 3 or more key elements are blank, consider it void-like
      return(blank_count >= 3)
}

get_scale_info <- function(scale_obj, expand = TRUE, axis_name = NULL) {

      default_expansion <- utils::getFromNamespace("default_expansion", "ggplot2")
      expand_limits_scale <- utils::getFromNamespace("expand_limits_scale", "ggplot2")

      expansion <- default_expansion(scale_obj, expand = expand)
      limits <- scale_obj$get_limits()
      # Pass NULL for coord_limits since coord_3d doesn't support them yet
      expanded_range <- expand_limits_scale(scale_obj, expansion, limits, coord_limits = NULL)

      # Get all breaks and labels
      all_breaks <- scale_obj$get_breaks()
      all_labels <- scale_obj$get_labels()

      # Filter breaks to be within expanded limits
      if (is.numeric(all_breaks)) {
            # For continuous scales, filter numerically
            valid_mask <- !is.na(all_breaks) &
                  all_breaks >= expanded_range[1] &
                  all_breaks <= expanded_range[2]
            valid_breaks <- all_breaks[valid_mask]

            # Filter corresponding labels
            if (length(all_labels) == length(all_breaks)) {
                  valid_labels <- all_labels[valid_mask]
            } else {
                  valid_labels <- all_labels  # Let ggplot2 handle label mismatch
            }
      } else {
            # For discrete scales, keep all breaks/labels as-is
            # (discrete scales shouldn't have out-of-bounds issues)
            valid_breaks <- all_breaks
            valid_labels <- all_labels
      }

      # Get the scale name if axis_name is provided
      scale_name <- if (!is.null(axis_name)) get_scale_names(scale_obj, axis_name) else NULL

      result <- list(limits = expanded_range,
                     breaks = valid_breaks,
                     labels = valid_labels)

      # Add name if provided
      if (!is.null(scale_name)) {
            result$name <- scale_name
      }

      return(result)
}

# Project data onto cube face, if applicable
project_to_face <- function(data, data_std, proj, effective_ratios = c(1, 1, 1)){
      if(! "project_to_face" %in% names(data) || all(is.na(data$project_to_face))) return(data_std)
      face_extent <- setNames(effective_ratios * 0.5, c("x", "y", "z"))
      data_std %>%
            mutate(face = data$project_to_face,
                   depth_3d = transform_3d_standard(data_std, proj)$depth,
                   axis = substr(face, 1, 1),
                   value = ifelse(substr(face, 2, 4) == "min", -1, 1) *
                         face_extent[axis],
                   x = ifelse(is.na(face) | axis != "x", x, value),
                   y = ifelse(is.na(face) | axis != "y", y, value),
                   z = ifelse(is.na(face) | axis != "z", z, value)) %>%
            select(x, y, z, depth_3d)
}


#' Convert ref_circle points to circular polygons
#'
#' @param data Original data frame
#' @param data_std Standardized data frame
#' @return Data frame with circular polygons replacing ref_circle points
#' @keywords internal
#' @noRd
points_to_circles <- function(data, data_std, effective_ratios = c(1, 1, 1)) {

      # Keep std coords and all other vars
      d <- bind_cols(select(data, -x, -y, -z),
                     select(data_std, x, y, z))
      if("depth_3d" %in% names(data_std)) d$depth_3d <- data_std$depth_3d

      # Check if we have any ref_circle elements
      if (!"element_type" %in% names(d) || !any(d$element_type == "ref_circle")) {
            return(d)
      }

      result <- filter(d, element_type == "ref_circle") %>%
            rowwise() %>%
            reframe(generate_circle_vertices(x, y, z, project_to_face,
                                             ref_circle_radius, ref_circle_vertices,
                                             effective_ratios),
                    vertex_order = 1:ref_circle_vertices,
                    group = group) %>%
            full_join(select(filter(d, element_type == "ref_circle"), -x, -y, -z),
                      by = join_by(group)) %>%
            arrange(group, vertex_order) %>%
            bind_rows(filter(d, element_type != "ref_circle"))

      return(result)
}

#' Generate circle vertices in 3D space for a given face
#'
#' @param x_std,y_std,z_std Standardized point coordinates
#' @param face Face name (e.g., "zmin", "xmax")
#' @param radius Circle radius in standardized units
#' @param n_vertices Number of vertices for the circle
#' @return Data frame with x, y, z coordinates for circle vertices
#' @keywords internal
#' @noRd
generate_circle_vertices <- function(x_std, y_std, z_std, face, radius, n_vertices,
                                     effective_ratios = c(1, 1, 1)) {
      # Generate angles for circle vertices
      angles <- seq(0, 2 * pi, length.out = n_vertices + 1)[-(n_vertices + 1)]

      # Get the face axis and value
      face_axis <- substr(face, 1, 1)
      axis_index <- match(face_axis, c("x", "y", "z"))
      face_value <- ifelse(substr(face, 2, 4) == "min", -0.5, 0.5) * effective_ratios[axis_index]

      # Generate circle vertices based on the face
      if (face_axis == "z") { # Circle in x-y plane
            x_coords <- x_std + radius * cos(angles)
            y_coords <- y_std + radius * sin(angles)
            z_coords <- rep(face_value, n_vertices)
      } else if (face_axis == "x") { # Circle in y-z plane
            x_coords <- rep(face_value, n_vertices)
            y_coords <- y_std + radius * cos(angles)
            z_coords <- z_std + radius * sin(angles)
      } else if (face_axis == "y") { # Circle in x-z plane
            x_coords <- x_std + radius * cos(angles)
            y_coords <- rep(face_value, n_vertices)
            z_coords <- z_std + radius * sin(angles)
      } else {
            stop("Invalid face: ", face)
      }

      return(data.frame(x = x_coords, y = y_coords, z = z_coords))
}

validate_coord3d <- function(coord){
      stopifnot("Did you forget to add `coord_3d()` to your plot?" = inherits(coord, "Coord3D"))
}

# Identify backfaces using signed area test
is_backface <- function(x, y){
      sum(x * lead(y, default = first(y)) -
                lead(x, default = first(x)) * y) < 0
}

process_backfaces <- function(data) {

      if (".subgroup" %in% names(data)) {
            # For data with subgroups (e.g. text with holes), compute signed area
            # per subgroup to avoid cross-contour stitching artifacts from lead()
            # wrapping across contour boundaries, then sum per group.
            group_areas <- data %>%
                  mutate(y = y*1000, x = x*1000) %>%
                  group_by(group, .subgroup) %>%
                  summarise(.area = sum(x * lead(y, default = first(y)) -
                                              lead(x, default = first(x)) * y),
                            .groups = "drop") %>%
                  group_by(group) %>%
                  summarise(.backface = sum(.area) < 0,
                            .area = sum(.area)) %>%
                  ungroup()
            data <- left_join(data, group_areas, by = "group")
      } else {
            # Standard case: signed area over whole group
            data <- data %>%
                  group_by(group) %>%
                  mutate(.backface = is_backface(x, y)) %>%
                  ungroup()
      }

      # Apply culling if requested
      if("cull_backfaces" %in% names(data) && data$cull_backfaces[1]) {
            data <- filter(data, !.backface)
      }

      # Apply lighting effects if present
      if("lighting_spec" %in% names(data) && "light" %in% names(data)) {
            scl <- data$lighting_spec[[1]]$backface_scale %||% 1
            off <- data$lighting_spec[[1]]$backface_offset %||% 0
            if((scl != 1 || off != 0) &
               data$lighting_spec[[1]]$method != "rgb") {
                  data <- mutate(data, light = ifelse(.backface, light * scl + off, light))
            }
      }

      return(select(data, -.backface))
}

test_that("resolve_scale_depth defaults every element to full strength", {
      resolved <- resolve_scale_depth(TRUE)

      expect_named(resolved, ggcube_depth_elements)
      expect_true(all(resolved == 1))
})

test_that("resolve_scale_depth treats FALSE as off everywhere", {
      resolved <- resolve_scale_depth(FALSE)

      expect_named(resolved, ggcube_depth_elements)
      expect_true(all(resolved == 0))
})

test_that("resolve_scale_depth broadcasts a single number", {
      resolved <- resolve_scale_depth(0.4)

      expect_true(all(resolved == 0.4))
      expect_length(resolved, length(ggcube_depth_elements))
})

test_that("resolve_scale_depth applies named values and defaults the rest", {
      resolved <- resolve_scale_depth(c(text = 0.4))

      expect_equal(unname(resolved["text"]), 0.4)
      expect_equal(unname(resolved["grid"]), 1)
      expect_equal(unname(resolved["border"]), 1)
})

test_that("resolve_scale_depth accepts values above one", {
      resolved <- resolve_scale_depth(c(grid = 2.5))
      expect_equal(unname(resolved["grid"]), 2.5)
})

test_that("resolve_scale_depth is idempotent", {
      once <- resolve_scale_depth(c(text = 0.4))
      twice <- resolve_scale_depth(once)

      expect_equal(once, twice)
})

test_that("resolve_scale_depth rejects negative values", {
      expect_error(resolve_scale_depth(-1), "non-negative")
      expect_error(resolve_scale_depth(c(text = -0.5)), "non-negative")
})

test_that("resolve_scale_depth rejects non-finite values", {
      expect_error(resolve_scale_depth(NA_real_), "finite")
      expect_error(resolve_scale_depth(c(grid = Inf)), "finite")
})

test_that("resolve_scale_depth rejects unknown element names", {
      expect_error(resolve_scale_depth(c(title = 1)), "title")
      expect_error(resolve_scale_depth(c(grid = 1, panel = 1)), "panel")
})

test_that("resolve_scale_depth covers every documented element", {
      resolved <- resolve_scale_depth(c(grid = 0.2, border = 0.4,
                                        ticks = 0.6, text = 0.8))

      expect_equal(unname(resolved["ticks"]), 0.6)
      expect_setequal(names(resolved), c("grid", "border", "ticks", "text"))
})

test_that("resolve_scale_depth rejects malformed input", {
      expect_error(resolve_scale_depth("grid"))
      expect_error(resolve_scale_depth(c(1, 2)))
      expect_error(resolve_scale_depth(c(grid = 1, 2)))
      expect_error(resolve_scale_depth(c(grid = 1, grid = 2)), "duplicate")
      expect_error(resolve_scale_depth(c(TRUE, FALSE)))
      expect_error(resolve_scale_depth(NA))
})

test_that("apply_depth_strength is the identity at strength one", {
      depths <- c(0.5, 0.8, 1, 1.25, 2)
      expect_equal(apply_depth_strength(depths, 1), depths)
})

test_that("apply_depth_strength flattens to one at strength zero", {
      depths <- c(0.5, 0.8, 1, 1.25, 2)
      expect_true(all(apply_depth_strength(depths, 0) == 1))
})

test_that("apply_depth_strength interpolates geometrically", {
      expect_equal(apply_depth_strength(4, 0.5), 2)
      expect_equal(apply_depth_strength(0.25, 0.5), 0.5)
})

test_that("apply_depth_strength composes multiplicatively", {
      depths <- c(0.7, 1, 1.75)
      chained <- apply_depth_strength(apply_depth_strength(depths, 0.5), 0.4)

      expect_equal(chained, apply_depth_strength(depths, 0.5 * 0.4))
})

test_that("apply_depth_strength treats near and far symmetrically", {
      # A depth factor and its reciprocal stay reciprocal at any strength
      expect_equal(apply_depth_strength(2, 0.3) * apply_depth_strength(0.5, 0.3), 1)
})

test_that("apply_depth_strength stays positive and finite for large strengths", {
      depths <- c(0.5, 0.7, 1, 1.75)
      result <- apply_depth_strength(depths, 10)

      expect_true(all(result > 0))
      expect_true(all(is.finite(result)))
})

test_that("apply_depth_strength falls back to identity for invalid strength", {
      depths <- c(0.8, 1.2)
      expect_equal(apply_depth_strength(depths, NA), depths)
      expect_equal(apply_depth_strength(depths, NULL), depths)
      expect_equal(apply_depth_strength(depths, c(1, 2)), depths)
})

test_that("depth_strength reads an element off a coord", {
      coord <- list(scale_depth = c(grid = 1, border = 0.5, text = 0))

      expect_equal(depth_strength(coord, "grid"), 1)
      expect_equal(depth_strength(coord, "border"), 0.5)
      expect_equal(depth_strength(coord, "text"), 0)
})

test_that("depth_strength defaults to full strength when unavailable", {
      # Coord predating the parameter
      expect_equal(depth_strength(list(), "text"), 1)
      # Element not present in the stored vector
      expect_equal(depth_strength(list(scale_depth = c(grid = 0)), "text"), 1)
})

test_that("safe_lwd replaces invalid linewidths with the fallback", {
      expect_equal(safe_lwd(c(1, 2, 3), fallback = 5), c(1, 2, 3))
      expect_equal(safe_lwd(c(1, NA, -1, Inf), fallback = 5), c(1, 5, 5, 5))
})

test_that("coord_3d stores a resolved scale_depth vector", {
      coord <- coord_3d()[[1]]

      expect_named(coord$scale_depth, ggcube_depth_elements)
      expect_true(all(coord$scale_depth == 1))

      subdued <- coord_3d(scale_depth = c(text = 0.4))[[1]]
      expect_equal(unname(subdued$scale_depth["text"]), 0.4)
})

test_that("coord_3d rejects invalid scale_depth at construction", {
      expect_error(coord_3d(scale_depth = -1), "non-negative")
      expect_error(coord_3d(scale_depth = c(title = 1)), "title")
})

# Mixed-geometry rendering pipeline
#
# Shared rendering function that produces grobs from depth-sorted data
# containing multiple primitive types (polygon, point, segment, text).
# Data arrives already transformed to NPC coordinates, depth-sorted, and
# with resolved aesthetic values. This function handles only the final
# grob assembly step.


# Depth scaling ----------------------------------------------------------------

#' Validate a layer's `scale_depth` argument
#'
#' Layers take a single strength rather than the coord's per-element vector.
#' `TRUE` and `FALSE` are accepted as 1 and 0 so that existing code keeps
#' working unchanged.
#'
#' @param scale_depth User-supplied value.
#' @return A single non-negative number.
#' @keywords internal
#' @noRd
resolve_layer_scale_depth <- function(scale_depth) {
      if (is.null(scale_depth)) return(1)

      if (is.logical(scale_depth)) {
            if (length(scale_depth) != 1 || is.na(scale_depth)) {
                  rlang::abort("`scale_depth` must be TRUE, FALSE, or a single number.")
            }
            return(as.numeric(scale_depth))
      }

      if (!is.numeric(scale_depth) || length(scale_depth) != 1) {
            rlang::abort("`scale_depth` must be TRUE, FALSE, or a single number.")
      }

      if (!is.finite(scale_depth)) {
            rlang::abort("`scale_depth` must be finite.")
      }

      if (scale_depth < 0) {
            rlang::abort(c(
                  "`scale_depth` must be non-negative.",
                  i = "Negative values would invert the depth cue, drawing distant elements larger than near ones."
            ))
      }

      scale_depth
}

#' Apply depth scaling to point sizes, strokes, and linewidths
#'
#' Scales visual properties by the depth_scale factor so that closer objects
#' appear larger/thicker and farther objects appear smaller/thinner. The
#' strength exponent is shared with `coord_3d(scale_depth = )`, so `1` is full
#' strength, `0` disables scaling, and intermediate values subdue it.
#'
#' @param coords Data frame with aesthetic columns and optionally `depth_scale`.
#' @param strength Depth scaling strength. `TRUE`/`FALSE` are accepted as 1/0.
#' @return Data frame with scaled aesthetics.
#' @keywords internal
#' @noRd
apply_depth_scaling <- function(coords, strength){
      strength <- resolve_layer_scale_depth(strength)

      if (strength > 0 && "depth_scale" %in% names(coords)) {
            factor <- apply_depth_strength(coords$depth_scale, strength)

            if("size" %in% names(coords)) {
                  coords$size <- coords$size * factor
            }
            if("stroke" %in% names(coords)) {
                  coords$stroke <- coords$stroke * factor
            }
            if("linewidth" %in% names(coords)) {
                  coords$linewidth <- coords$linewidth * factor
            }
      }
      return(coords)
}


# Grob assembly ----------------------------------------------------------------


#' Render mixed-geometry data into a grob tree
#'
#' Takes depth-sorted, transformed data with a `.prim` column indicating
#' primitive type and produces a single grob (or grobTree) for the layer.
#' Consecutive runs of the same primitive type are batched into single
#' vectorized grob calls for efficiency.
#'
#' @param data Data frame with transformed NPC coordinates, depth-sorted.
#'   Must contain a `.prim` column with values in `"polygon"`, `"point"`,
#'   `"segment"`, `"text"`.
#' @param rule Polygon fill rule: `"evenodd"` or `"winding"`. Only used
#'   for polygons with subgroups (holes).
#' @param arrow Arrow specification for segments, created by [grid::arrow()].
#' @param lineend Line end style for segments.
#' @return A grid grob.
#' @keywords internal
#' @noRd
render_mixed_grobs <- function(data, rule = "evenodd",
                               arrow = NULL, lineend = "butt") {

      if (nrow(data) == 0) return(grid::nullGrob())

      # If no .prim column, treat everything as polygon (backward compat)
      if (!".prim" %in% names(data)) {
            data$.prim <- "polygon"
      }

      # Identify runs of consecutive same-type primitives
      runs <- compute_prim_runs(data)

      if (nrow(runs) == 1) {
            # Single type: fast path
            return(render_single_type(data, runs$.prim[1], rule, arrow, lineend))
      }

      # Multiple types: render each run, assemble into grobTree
      grobs <- vector("list", nrow(runs))
      for (i in seq_len(nrow(runs))) {
            run_data <- data[runs$start[i]:runs$end[i], ]
            grobs[[i]] <- render_single_type(run_data, runs$.prim[i],
                                             rule, arrow, lineend)
      }

      # Drop nullGrobs
      grobs <- grobs[!vapply(grobs, inherits, logical(1), "null")]
      if (length(grobs) == 0) return(grid::nullGrob())
      if (length(grobs) == 1) return(grobs[[1]])
      do.call(grid::grobTree, grobs)
}


#' Compute runs of consecutive same-type primitives
#'
#' Groups are the atomic unit: all rows in a group share a .prim type.
#' We detect run boundaries by looking at where .prim changes across
#' consecutive rows (groups are contiguous after depth sorting).
#'
#' @param data Data frame with `.prim` and `group` columns.
#' @return Data frame with columns `.prim`, `start`, `end` (row indices).
#' @keywords internal
#' @noRd
compute_prim_runs <- function(data) {
      prims <- data$.prim
      # Detect where .prim changes
      changes <- c(TRUE, prims[-1] != prims[-length(prims)])
      run_id <- cumsum(changes)

      # Build run table
      run_starts <- which(changes)
      run_ends <- c(run_starts[-1] - 1L, nrow(data))
      run_types <- prims[run_starts]

      data.frame(.prim = run_types, start = run_starts, end = run_ends,
                 stringsAsFactors = FALSE)
}


# Type-specific renderers ----------------------------------------------------

#' Dispatch to the correct renderer for a single primitive type
#' @keywords internal
#' @noRd
render_single_type <- function(data, prim_type, rule, arrow, lineend) {
      switch(prim_type,
             polygon = render_polygons(data, rule),
             point   = render_points(data),
             segment = render_segments(data, arrow, lineend),
             text    = render_texts(data),
             {
                   warning("Unknown .prim type: ", prim_type, "; skipping")
                   grid::nullGrob()
             }
      )
}


#' Render polygon primitives
#'
#' Handles both simple polygons and polygons with holes (subgroups).
#' All polygons in the batch are rendered in a single vectorized grob call.
#'
#' @param data Polygon data with x, y, group, and aesthetic columns.
#' @param rule Fill rule for polygons with holes.
#' @return A polygonGrob or pathGrob.
#' @keywords internal
#' @noRd
render_polygons <- function(data, rule = "evenodd") {
      if (nrow(data) == 0) return(grid::nullGrob())

      # Detect subgroup column
      has_subgroup <- any(c("subgroup", ".subgroup") %in% names(data))
      if (has_subgroup) {
            sg_col <- if (".subgroup" %in% names(data)) ".subgroup" else "subgroup"
      }

      # Map group to integer IDs preserving data order (not alphabetical)
      unique_groups <- unique(data$group)
      group_ids <- match(data$group, unique_groups)

      # Extract per-group aesthetics (one value per group, in data order)
      group_first_idx <- !duplicated(data$group)
      group_col <- data$colour[group_first_idx]
      group_fill <- data$fill[group_first_idx]
      group_lwd <- data$linewidth[group_first_idx] * .pt
      group_lty <- if ("linetype" %in% names(data)) data$linetype[group_first_idx] else 1
      group_lty <- ifelse(is.na(group_lty), 1, group_lty)
      group_alpha <- data$alpha[group_first_idx]
      group_alpha <- ifelse(is.na(group_alpha), 1, group_alpha)

      if (has_subgroup) {
            # Ensure rows are contiguous by group and ordered by subgroup within,
            # while preserving the inter-group depth-sorted order.
            grp_rank <- match(data$group, unique_groups)
            data <- data[order(grp_rank, data[[sg_col]]), ]

            # Recompute after reordering
            unique_groups <- unique(data$group)
            group_ids <- match(data$group, unique_groups)
            unique_subgroups <- unique(data[[sg_col]])
            subgroup_ids <- match(data[[sg_col]], unique_subgroups)

            group_first_idx <- !duplicated(data$group)
            group_col <- data$colour[group_first_idx]
            group_fill <- data$fill[group_first_idx]
            group_lwd <- data$linewidth[group_first_idx] * .pt
            group_lty <- if ("linetype" %in% names(data)) data$linetype[group_first_idx] else 1
            group_lty <- ifelse(is.na(group_lty), 1, group_lty)
            group_alpha <- data$alpha[group_first_idx]
            group_alpha <- ifelse(is.na(group_alpha), 1, group_alpha)

            grid::pathGrob(
                  x = data$x,
                  y = data$y,
                  id = subgroup_ids,
                  pathId = group_ids,
                  rule = rule,
                  default.units = "npc",
                  gp = grid::gpar(
                        col = group_col,
                        fill = group_fill,
                        lwd = group_lwd,
                        lty = group_lty,
                        alpha = group_alpha
                  )
            )
      } else {
            grid::polygonGrob(
                  x = data$x,
                  y = data$y,
                  id = group_ids,
                  default.units = "npc",
                  gp = grid::gpar(
                        col = group_col,
                        fill = group_fill,
                        lwd = group_lwd,
                        lty = group_lty,
                        alpha = group_alpha
                  )
            )
      }
}


#' Render point primitives
#'
#' Each group is a single point. Points are rendered in a single vectorized
#' pointsGrob call when all share the same shape, or as individual grobs
#' when shapes differ (required by grid's pointsGrob which takes scalar pch).
#'
#' @param data Point data with x, y, and aesthetic columns.
#' @return A pointsGrob or grobTree.
#' @keywords internal
#' @noRd
render_points <- function(data) {
      if (nrow(data) == 0) return(grid::nullGrob())

      # Points: one row per group (take first row of each group)
      pts <- data[!duplicated(data$group), ]

      # Detect shape complexity for fill handling
      shapes <- pts$shape
      is_complex <- is.numeric(shapes) & shapes %in% 21:25

      # Compute fontsize: complex shapes use size + 0.5*stroke, simple use size
      stroke <- if ("stroke" %in% names(pts)) pts$stroke else rep(0.5, nrow(pts))
      fontsize <- ifelse(is_complex,
                         (pts$size + 0.5 * stroke) * .pt,
                         pts$size * .pt)

      # Fill: complex shapes use fill column, simple shapes use NA
      fill <- ifelse(is_complex,
                     ifelse(is.na(pts$fill), NA_character_, pts$fill),
                     NA_character_)

      alpha <- ifelse(is.na(pts$alpha), 1, pts$alpha)

      # If all shapes are the same, single vectorized call
      if (length(unique(shapes)) == 1) {
            grid::pointsGrob(
                  x = pts$x,
                  y = pts$y,
                  pch = shapes[1],
                  default.units = "npc",
                  gp = grid::gpar(
                        col = pts$colour,
                        fill = fill,
                        fontsize = fontsize,
                        lwd = stroke * .pt,
                        alpha = alpha
                  )
            )
      } else {
            # Mixed shapes: individual grobs (grid limitation on pch)
            grobs <- lapply(seq_len(nrow(pts)), function(i) {
                  grid::pointsGrob(
                        x = pts$x[i],
                        y = pts$y[i],
                        pch = shapes[i],
                        default.units = "npc",
                        gp = grid::gpar(
                              col = pts$colour[i],
                              fill = fill[i],
                              fontsize = fontsize[i],
                              lwd = stroke[i] * .pt,
                              alpha = alpha[i]
                        )
                  )
            })
            do.call(grid::grobTree, grobs)
      }
}


#' Render segment primitives
#'
#' Each group is a segment with exactly two rows (start and end points).
#' Segments are rendered in a single vectorized segmentsGrob call.
#'
#' @param data Segment data with x, y, group, and aesthetic columns.
#'   Each group must have exactly 2 rows with `point_type` "start"/"end",
#'   or simply two rows in order.
#' @param arrow Arrow specification.
#' @param lineend Line end style.
#' @return A segmentsGrob.
#' @keywords internal
#' @noRd
render_segments <- function(data, arrow = NULL, lineend = "butt") {
      if (nrow(data) == 0) return(grid::nullGrob())

      # Segments are stored as two rows per group: start and end
      # Identify start/end using point_type column if present,
      # otherwise use first/second row of each group
      if ("point_type" %in% names(data)) {
            start_rows <- data[data$point_type == "start", ]
            end_rows <- data[data$point_type == "end", ]
            # Match end rows to start rows by group, preserving start row order
            end_rows <- end_rows[match(start_rows$group, end_rows$group), ]
      } else {
            # Generic: first row = start, second row = end within each group
            groups <- unique(data$group)
            start_idx <- integer(length(groups))
            end_idx <- integer(length(groups))
            for (i in seq_along(groups)) {
                  rows <- which(data$group == groups[i])
                  if (length(rows) >= 2) {
                        start_idx[i] <- rows[1]
                        end_idx[i] <- rows[2]
                  }
            }
            valid <- start_idx > 0 & end_idx > 0
            start_rows <- data[start_idx[valid], ]
            end_rows <- data[end_idx[valid], ]
      }

      if (nrow(start_rows) == 0) return(grid::nullGrob())

      lty <- if ("linetype" %in% names(start_rows)) start_rows$linetype else 1

      grid::segmentsGrob(
            x0 = start_rows$x, y0 = start_rows$y,
            x1 = end_rows$x, y1 = end_rows$y,
            default.units = "npc",
            arrow = arrow,
            gp = grid::gpar(
                  col = start_rows$colour,
                  lwd = start_rows$linewidth * .pt,
                  lty = lty,
                  lineend = lineend,
                  alpha = ifelse(is.na(start_rows$alpha), 1, start_rows$alpha)
            )
      )
}


#' Render text primitives (billboard)
#'
#' Each group is a single text label. Rendered as individual textGrobs
#' assembled into a grobTree.
#'
#' @param data Text data with x, y, label, and aesthetic columns.
#' @return A grobTree of textGrobs.
#' @keywords internal
#' @noRd
render_texts <- function(data) {
      if (nrow(data) == 0) return(grid::nullGrob())

      # One row per label (take first row of each group)
      labels <- data[!duplicated(data$group), ]

      fontsize <- labels$size * .pt
      alpha <- ifelse(is.na(labels$alpha), 1, labels$alpha)

      grobs <- lapply(seq_len(nrow(labels)), function(i) {
            grid::textGrob(
                  label = labels$label[i],
                  x = labels$x[i],
                  y = labels$y[i],
                  default.units = "npc",
                  hjust = labels$hjust[i] %||% 0.5,
                  vjust = labels$vjust[i] %||% 0.5,
                  rot = labels$angle[i] %||% 0,
                  gp = grid::gpar(
                        col = alpha(labels$colour[i], alpha[i]),
                        fontsize = fontsize[i],
                        fontfamily = labels$family[i] %||% "",
                        fontface = labels$fontface[i] %||% 1,
                        lineheight = labels$lineheight[i] %||% 1.2
                  )
            )
      })

      if (length(grobs) == 1) return(grobs[[1]])
      do.call(grid::grobTree, grobs)
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi = 150,
  warning = FALSE,
  message = FALSE
)

## ----setup, message = FALSE---------------------------------------------------
library(ggcube)

## ----basics-------------------------------------------------------------------
ggplot(mpg, aes(x = displ, y = hwy, z = drv, color = class)) +
      geom_point() +
      coord_3d()

## ----rotation, fig.height = 4-------------------------------------------------
ggplot(mpg, aes(displ, hwy, drv, color = class)) +
      geom_point() +
      coord_3d(pitch = 0, roll = 60, yaw = 0, dist = 1.4,
               ratio = c(2, 1, 1), panels = "all") +
      theme(panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .1))

## ----surface------------------------------------------------------------------
ggplot(mountain, aes(x, y, z)) +
      geom_surface_3d(aes(fill = z, color = z)) +
      scale_fill_viridis_c() + scale_color_viridis_c() +
      coord_3d(ratio = c(1.5, 2, 1), expand = FALSE, panels = "zmin") +
      light(direction = c(1, 0, 0)) +
      guides(fill = guide_colorbar_3d()) +
      theme_light()

## ----points-------------------------------------------------------------------
ggplot(mpg, aes(x = displ, y = hwy, z = drv, fill = class)) +
      geom_point_3d(size = 3, shape = 21, color = "black", stroke = .1,
                    ref_lines = TRUE, ref_faces = c("ymax", "xmax")) +
      coord_3d()

## ----paths--------------------------------------------------------------------
x <- seq(0, 20*pi, pi/16)
spiral <- data.frame(x = x, y = sin(x), z = cos(x), time = 1:length(x))
ggplot(spiral, aes(x, y, z, color = time)) +
      geom_path_3d() +
      scale_color_gradientn(colors = c("blue", "purple", "red", "orange")) +
      coord_3d() +
      theme_light()

## ----bar----------------------------------------------------------------------
ggplot(iris, aes(Species, Sepal.Length, fill = Species)) +
      geom_bar_3d(bins = 20, width = c(.5, 1)) +
      coord_3d(scales = "fixed", ratio = c(1, 3, .25), yaw = 60) +
      scale_z_continuous(expand = c(0, 0)) +
      theme(legend.position = "none")

## ----hull---------------------------------------------------------------------
ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d(method = "convex", fill = "#9e2602", color = "#5e1600") +
      coord_3d()

## ----distributions------------------------------------------------------------
ggplot(iris, aes(y = Sepal.Length, x = Species, fill = Species)) +
      stat_distributions_3d() +
      scale_z_continuous(expand = expansion(mult = c(0, NA))) +
      coord_3d() +
      theme(legend.position = "none")

## ----text---------------------------------------------------------------------
df <- expand.grid(x = c("H", "B"), y = c("a", "o", "u"), z = c("g", "t"))
df$label <- paste0(df$x, df$y, df$z)
ggplot(df, aes(x, y, z, label = label, fill = x)) +
      geom_text_3d(method = "polygon", facing = "zmax",
                   size = 5, weight = "bold") +
      coord_3d(scales = "fixed", light = NULL)

## ----lighting-----------------------------------------------------------------
p <- ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d(fill = "#9e2602", color = "#5e1600") +
      coord_3d()

p + light(method = "direct", mode = "hsl", direction = c(0, 0, 1))

## ----zlim, fig.show = "hide"--------------------------------------------------
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
      geom_point() +
      zlim(15, 20) +
      coord_3d()

## ----guide, fig.show = "hide"-------------------------------------------------
ggplot(mountain, aes(x, y, z, fill = z)) +
      stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
      guides(fill = guide_colorbar_3d()) +
      scale_fill_gradientn(colors = c("tomato", "dodgerblue")) +
      coord_3d()

## ----theme--------------------------------------------------------------------
ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d() +
      coord_3d(panels = "all") +
      theme(panel.background = element_rect(color = "black"),
            panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .3),
            panel.grid.foreground = element_line(color = "gray", linewidth = .25),
            axis.text = element_text(color = "darkblue"),
            axis.text.z = element_text(color = "darkred"),
            axis.title = element_text(margin = margin(t = 30)),
            axis.title.x = element_text(color = "magenta"))

## ----annotate, eval = FALSE---------------------------------------------------
# summit <- filter(mountain, z == max(z))
# ggplot(mountain, aes(x, y, z)) +
#       geom_contour_3d(
#             annotate = list(
#                   annotate_3d("point", x = summit$x, y = summit$y, z = summit$z, color = "red"),
#                   annotate_3d("text", x = summit$x, y = summit$y, z = summit$z, color = "red",
#                               label = "Summit", fontface = "bold", vjust = -1)
#             ), fill = "black"
#       ) +
#       coord_3d(ratio = c(2, 3, 1.5), light = "none")

## ----face_projection----------------------------------------------------------
ggplot(iris, aes(Sepal.Length, Sepal.Width, Petal.Length,
                 color = Species, fill = Species)) +
      coord_3d() + xlim(4, 8) +
      stat_density_2d(position = position_on_face(faces = "zmin", axes = c("x", "y")),
                      geom = "polygon", alpha = .1, linewidth = .25) +
      geom_hull_3d(position = position_on_face("ymax"), alpha = .5) +
      geom_point_3d(shape = 21, color = "black", stroke = .25)

## ----animation, eval = FALSE--------------------------------------------------
# p <- ggplot(mountain, aes(x, y, z)) +
#       geom_contour_3d(fill = "black", color = "white", linewidth = .5, light = "none") +
#       coord_3d(ratio = c(1.5, 2, 1), zoom = 1.25) +
#       theme_void()
# 
# orbit_3d(p, yaw = c(360, 0), roll = c(-90, 0), n = c(24, 8),
#          start = c(yaw = 300, roll = -60))

# Shared grid utilities --------------------------------

# Flip orientation if applicable
transpose_grid <- function(grid, direction){
      if(direction == "y"){
            grid <- grid %>%
                  mutate(temp = x,
                         x = y,
                         y = temp) %>%
                  select(-temp)
      }
      return(grid)
}

# Scale to target range
rescale_grid <- function(grid, xlim, ylim){
      grid %>%
            mutate(x = scales::rescale(x, xlim),
                   y = scales::rescale(y, ylim))
}

# Add computed surface orientation variables
compute_surface_vars <- function(tiles){
      left_join(tiles,
                compute_surface_gradients_from_vertices(tiles),
                by = join_by(group)) %>%
            mutate(slope = sqrt(dzdy^2 + dzdx^2),
                   aspect = atan2(dzdy, dzdx))
}


# Point grids --------------------------------------


#' Generate a grid of unique vertex points
#'
#' Creates a regular grid of unique vertex positions for use by surface-generating
#' stats. The grid is tessellated into polygon tiles in the geom layer (or in the
#' stat for [stat_smooth_3d()]).
#'
#' @inheritParams grid_params
#' @param xlim,ylim Length-two numeric vectors defining bounding box.
#'
#' @return A data frame with columns `x`, `y`.
#'
#' @keywords internal
#' @noRd
make_point_grid <- function(grid = c("rectangle", "right1", "right2", "equilateral"),
                            n = 40,
                            direction = c("x", "y"),
                            xlim, ylim,
                            trim = TRUE) {

      grid <- match.arg(grid)
      direction <- match.arg(direction)

      if (is.null(n)) n <- 40
      n <- as.integer(n)
      if (any(n < 2)) stop("`n` must be at least 2")
      if (!length(n) %in% 1:2) stop("`n` must be a vector of length 1 or 2")

      d <- switch(grid,
                  "rectangle" = make_rect_grid(n),
                  "right1" = make_rect_grid(n),
                  "right2" = make_rect_grid(n),
                  "equilateral" = make_equilateral_grid(n, trim))

      d <- d %>%
            transpose_grid(direction) %>%
            rescale_grid(xlim, ylim)

      return(d)
}

make_rect_grid <- function(n){
      if (length(n) == 1) n <- c(n, n)
      tidyr::expand_grid(x = 1:n[1], y = 1:n[2])
}

make_equilateral_grid <- function(n, trim = FALSE){

      side_length <- 1
      tri_height <- side_length * sqrt(3) / 2

      if (length(n) == 1) {
            # Equilateral triangles
            n_x <- n
            n_y <- n * side_length / tri_height
      } else {
            # Non-equilateral case - user specifies both dimensions
            n_x <- n[1]
            n_y <- n[2]
      }

      x <- seq(0, by = side_length, length.out = n_x)
      y <- seq(0, by = tri_height, length.out = n_y)

      grid <- expand_grid(row = 1:n_y, column = 1:n_x) %>%
            mutate(x = x[column],
                   y = y[row],
                   x = ifelse(row %% 2 == 0, x, x + side_length/2)) %>%
            select(-row, -column)

      if(trim){
            xmax <- filter(grid, x == max(x)) %>%
                  mutate(x = min(grid$x))
            grid <- grid %>%
                  mutate(x = ifelse(x == max(x), max(x[x != max(x)]), x)) %>%
                  bind_rows(xmax)
      }

      return(grid)
}

is_regular_grid <- function(data){
      if(! "group" %in% names(data)) data$group <- 1

      # If row/column indices are present, trust them as grid structure.
      # This allows downstream tiling to handle grid subsets (e.g. data filtered
      # by min_ndensity) without falling back to Delaunay triangulation.
      if(all(c("row", "column") %in% names(data))) return(TRUE)

      data <- split(data, data$group)
      reg <- sapply(data, function(d){
            x_vals <- sort(unique(d$x))
            y_vals <- sort(unique(d$y))
            nrow(d) == length(x_vals) * length(y_vals)
      })
      all(reg)
}

#' Compute gradients at grid points using finite differences
#'
#' For rectangular grids, computes dz/dx and dz/dy at each grid point using
#' central differences where possible, forward/backward differences at edges.
#'
#' @param data Data frame with x, y, z and row, column indices
#'
#' @return Data frame with added columns: dzdx, dzdy, slope, aspect
#'
#' @keywords internal
#' @noRd
compute_point_gradients <- function(data) {

      if(is_regular_grid(data)){
            data <- compute_grid_point_gradients(data)
      }else{
            data <- compute_irregular_point_gradients(data)
      }

      # Compute derived variables
      data$slope <- sqrt(data$dzdx^2 + data$dzdy^2)
      data$aspect <- atan2(data$dzdy, data$dzdx)

      return(data)
}

compute_grid_point_gradients <- function(data){
      # Add row/column indices for regular grid
      x_vals <- sort(unique(data$x))
      y_vals <- sort(unique(data$y))
      data <- data %>%
            mutate(column = match(x, x_vals),
                   row = match(y, y_vals))

      # Get grid dimensions and spacing
      n_col <- max(data$column)
      n_row <- max(data$row)

      x_vals <- sort(unique(data$x))
      y_vals <- sort(unique(data$y))
      dx <- if (length(x_vals) > 1) mean(diff(x_vals)) else 1
      dy <- if (length(y_vals) > 1) mean(diff(y_vals)) else 1

      # Create lookup matrix for z values
      z_matrix <- matrix(NA_real_, nrow = n_row, ncol = n_col)
      for (i in 1:nrow(data)) {
            z_matrix[data$row[i], data$column[i]] <- data$z[i]
      }

      # Compute gradients using finite differences
      dzdx_matrix <- matrix(NA_real_, nrow = n_row, ncol = n_col)
      dzdy_matrix <- matrix(NA_real_, nrow = n_row, ncol = n_col)

      for (r in 1:n_row) {
            for (c in 1:n_col) {
                  # dz/dx (central difference, forward/backward at edges)
                  if (c == 1) {
                        dzdx_matrix[r, c] <- (z_matrix[r, c + 1] - z_matrix[r, c]) / dx
                  } else if (c == n_col) {
                        dzdx_matrix[r, c] <- (z_matrix[r, c] - z_matrix[r, c - 1]) / dx
                  } else {
                        dzdx_matrix[r, c] <- (z_matrix[r, c + 1] - z_matrix[r, c - 1]) / (2 * dx)
                  }

                  # dz/dy (central difference, forward/backward at edges)
                  if (r == 1) {
                        dzdy_matrix[r, c] <- (z_matrix[r + 1, c] - z_matrix[r, c]) / dy
                  } else if (r == n_row) {
                        dzdy_matrix[r, c] <- (z_matrix[r, c] - z_matrix[r - 1, c]) / dy
                  } else {
                        dzdy_matrix[r, c] <- (z_matrix[r + 1, c] - z_matrix[r - 1, c]) / (2 * dy)
                  }
            }
      }

      # Add gradients to data
      data$dzdx <- mapply(function(r, c) dzdx_matrix[r, c], data$row, data$column)
      data$dzdy <- mapply(function(r, c) dzdy_matrix[r, c], data$row, data$column)

      data
}


compute_irregular_point_gradients <- function(data) {

      # Extract coordinates
      x_vals <- data$x
      y_vals <- data$y
      z_vals <- data$z

      # Perform Delaunay triangulation
      points_2d <- cbind(x_vals, y_vals)
      triangles <- geometry::delaunayn(points_2d)

      # Initialize gradient storage
      n_triangles <- nrow(triangles)
      triangle_dzdx <- numeric(n_triangles)
      triangle_dzdy <- numeric(n_triangles)
      triangle_centroids <- matrix(0, nrow = n_triangles, ncol = 2)

      # Compute gradient for each triangle
      for (i in seq_len(n_triangles)) {
            idx <- triangles[i, ]

            # Get triangle vertices
            x1 <- x_vals[idx[1]]; y1 <- y_vals[idx[1]]; z1 <- z_vals[idx[1]]
            x2 <- x_vals[idx[2]]; y2 <- y_vals[idx[2]]; z2 <- z_vals[idx[2]]
            x3 <- x_vals[idx[3]]; y3 <- y_vals[idx[3]]; z3 <- z_vals[idx[3]]

            # Edge vectors
            v1 <- c(x2 - x1, y2 - y1, z2 - z1)
            v2 <- c(x3 - x1, y3 - y1, z3 - z1)

            # Normal vector via cross product
            normal <- c(
                  v1[2] * v2[3] - v1[3] * v2[2],
                  v1[3] * v2[1] - v1[1] * v2[3],
                  v1[1] * v2[2] - v1[2] * v2[1]
            )

            # Plane equation: normal[1]*x + normal[2]*y + normal[3]*z = d
            # Gradient: dz/dx = -normal[1]/normal[3], dz/dy = -normal[2]/normal[3]
            if (abs(normal[3]) > 1e-10) {
                  triangle_dzdx[i] <- -normal[1] / normal[3]
                  triangle_dzdy[i] <- -normal[2] / normal[3]
            } else {
                  # Vertical or near-vertical triangle
                  triangle_dzdx[i] <- NA
                  triangle_dzdy[i] <- NA
            }

            # Centroid
            triangle_centroids[i, 1] <- (x1 + x2 + x3) / 3
            triangle_centroids[i, 2] <- (y1 + y2 + y3) / 3
      }

      # For each point, find adjacent triangles and compute weighted average
      n_points <- nrow(data)
      point_dzdx <- numeric(n_points)
      point_dzdy <- numeric(n_points)

      for (i in seq_len(n_points)) {
            # Find triangles containing this point
            adj_triangles <- which(apply(triangles, 1, function(tri) i %in% tri))

            if (length(adj_triangles) == 0) {
                  point_dzdx[i] <- NA
                  point_dzdy[i] <- NA
                  next
            }

            # Get gradients for adjacent triangles (remove NAs)
            valid <- !is.na(triangle_dzdx[adj_triangles])
            adj_triangles <- adj_triangles[valid]

            if (length(adj_triangles) == 0) {
                  point_dzdx[i] <- NA
                  point_dzdy[i] <- NA
                  next
            }

            # Calculate distances to centroids
            px <- x_vals[i]
            py <- y_vals[i]
            distances <- sqrt(
                  (triangle_centroids[adj_triangles, 1] - px)^2 +
                        (triangle_centroids[adj_triangles, 2] - py)^2
            )

            # Use inverse distance weighting (add small epsilon to avoid division by zero)
            epsilon <- 1e-10
            weights <- 1 / (distances + epsilon)
            weights <- weights / sum(weights)

            # Weighted average
            point_dzdx[i] <- sum(weights * triangle_dzdx[adj_triangles])
            point_dzdy[i] <- sum(weights * triangle_dzdy[adj_triangles])
      }

      # Add gradients to original data
      data$dzdx <- point_dzdx
      data$dzdy <- point_dzdy

      return(data)
}


#' Convert point grid to polygon tiles
#'
#' Takes a grid of points and tessellates into polygon tiles.
#'
#' @param data Data frame with x, y, z columns (and optionally row, column).
#' @param method Tessellation method: "grid" for regular grid, "delaunay" for
#'   triangulation, "auto" to detect.
#' @param grid_type Tile geometry: "rectangle", "right1", "right2", or
#'   "equilateral". The equilateral option routes to Delaunay tessellation.
#' @param group_prefix String prefix for polygon group IDs.
#'
#' @return Data frame with polygon vertices including `group` and `order` columns.
#'
#' @keywords internal
#' @noRd
points_to_tiles <- function(data,
                            method = "auto",
                            grid_type = "rectangle",
                            group_prefix = "surface__tile") {

      method <- match.arg(method, c("auto", "grid", "delaunay"))

      # Auto-detect method
      if (method == "auto") {
            if (is_regular_grid(data)) {
                  method <- "grid"
            } else {
                  method <- "delaunay"
            }

      }

      if (method == "grid") {
            if (grid_type == "equilateral") {
                  tiles <- delaunay_points_to_tiles(data, group_prefix)
            } else {
                  tiles <- rect_points_to_tiles(data, grid_type, group_prefix)
            }
      } else {
            tiles <- delaunay_points_to_tiles(data, group_prefix)
      }

      return(tiles)
}


#' Convert regular grid points to tiles
#' @keywords internal
#' @noRd
rect_points_to_tiles <- function(data, grid_type, group_prefix) {

      grid_type <- match.arg(grid_type, c("rectangle", "right1", "right2"))

      # Ensure row/column indices exist
      if (!all(c("row", "column") %in% names(data))) {
            x_vals <- sort(unique(data$x))
            y_vals <- sort(unique(data$y))
            data <- data %>%
                  mutate(column = match(x, x_vals),
                         row = match(y, y_vals))
      }

      n_col <- max(data$column)
      n_row <- max(data$row)

      if (n_col < 2 || n_row < 2) {
            stop("Need at least 2 points in each dimension to create tiles")
      }

      # Create tile_id for each cell
      tiles <- tidyr::expand_grid(tile_col = 1:(n_col - 1), tile_row = 1:(n_row - 1)) %>%
            mutate(tile_id = row_number())

      # Get four corners for each tile in CCW order when viewed from above:
      # bottom-left -> bottom-right -> top-right -> top-left
      corner_offsets <- data.frame(
            d_col = c(0, 1, 1, 0),
            d_row = c(0, 0, 1, 1),
            order = 1:4
      )

      tile_vertices <- tiles %>%
            tidyr::crossing(corner_offsets) %>%
            mutate(column = tile_col + d_col,
                   row = tile_row + d_row) %>%
            left_join(data, by = c("column", "row")) %>%
            select(-d_col, -d_row, -tile_col, -tile_row)

      # Handle triangle splitting before culling so that each triangle can be
      # evaluated independently; with right1/right2 a quad with one missing
      # corner still yields one valid triangle.
      if (grid_type %in% c("right1", "right2")) {
            tile_vertices <- split_quads_to_triangles(tile_vertices, grid_type)
      }

      # Drop tiles with missing corners. When the input is a sparse subset of a
      # regular grid (e.g. data filtered by min_ndensity in stat_density_3d),
      # the left_join leaves NA coordinates at corners that don't exist in the
      # data; removing those tiles preserves the rectangular tiling on the
      # surviving cells without falling back to Delaunay triangulation.
      incomplete_tiles <- unique(tile_vertices$tile_id[is.na(tile_vertices$x) |
                                                             is.na(tile_vertices$y)])
      if (length(incomplete_tiles) > 0) {
            tile_vertices <- filter(tile_vertices, !tile_id %in% incomplete_tiles)
            if (nrow(tile_vertices) == 0) {
                  stop("No complete tiles remain after dropping cells with missing corners.")
            }
      }

      # Create hierarchical group column
      if ("group" %in% names(data) && !all(data$group == data$group[1])) {
            tile_vertices <- tile_vertices %>%
                  mutate(group = paste0(group_prefix, tile_id, "::", group))
      } else {
            tile_vertices <- tile_vertices %>%
                  mutate(group = paste0(group_prefix, tile_id))
      }

      tile_vertices <- tile_vertices %>%
            select(-column, -row, -tile_id) %>%
            arrange(group, order)

      return(tile_vertices)
}


#' Split quadrilateral tiles into triangles
#' @keywords internal
#' @noRd
split_quads_to_triangles <- function(tile_vertices, grid_type) {
      # Input quad vertices in CCW order when viewed from above:
      # 1 = bottom-left, 2 = bottom-right, 3 = top-right, 4 = top-left
      #
      # right1: diagonal from bottom-left(1) to top-right(3)
      #   Triangle A: 1 -> 2 -> 3 (CCW from above)
      #   Triangle B: 1 -> 3 -> 4 (CCW from above)
      #
      # right2: diagonal from bottom-right(2) to top-left(4)
      #   Triangle A: 1 -> 2 -> 4 (CCW from above)
      #   Triangle B: 2 -> 3 -> 4 (CCW from above)

      i <- switch(grid_type,
                  right1 = c(1, 2, 3, 1, 3, 4),
                  right2 = c(1, 2, 4, 2, 3, 4))

      tile_vertices %>%
            group_by(tile_id) %>%
            arrange(order) %>%
            reframe(
                  across(everything(), ~ .x[i]),
                  tile_id = tile_id[1] + rep(c(0.1, 0.2), each = 3),
                  order = rep(1:3, 2)
            ) %>%
            ungroup()
}


#' Convert irregular points to triangular tiles via Delaunay triangulation
#' @keywords internal
#' @noRd
delaunay_points_to_tiles <- function(data, group_prefix) {

      if (!requireNamespace("geometry", quietly = TRUE)) {
            stop("Delaunay triangulation requires the 'geometry' package.\n",
                 "Install with: install.packages('geometry')")
      }

      all_data <- data

      tile_vertices <- lapply(unique(all_data$group), function(i){
            data <- filter(all_data, group == i)

            if (nrow(data) < 3) {
                  stop("Need at least 3 points for triangulation")
            }

            # Normalize coordinates to avoid numerical issues
            x_range <- range(data$x, na.rm = TRUE)
            y_range <- range(data$y, na.rm = TRUE)
            x_scale <- max(diff(x_range), 1e-10)
            y_scale <- max(diff(y_range), 1e-10)

            x_norm <- (data$x - x_range[1]) / x_scale
            y_norm <- (data$y - y_range[1]) / y_scale

            # Perform Delaunay triangulation with robust options
            tri <- tryCatch({
                  geometry::delaunayn(cbind(x_norm, y_norm), options = "Qt Qc Qz Qbb")
            }, error = function(e) {
                  # Fallback without options
                  geometry::delaunayn(cbind(x_norm, y_norm))
            })

            n_tri <- nrow(tri)
            if (is.null(n_tri) || n_tri == 0) {
                  stop("Delaunay triangulation produced no triangles")
            }

            # Ensure CCW winding order for all triangles
            for (i in 1:n_tri) {
                  idx <- tri[i, ]
                  x1 <- data$x[idx[1]]; y1 <- data$y[idx[1]]
                  x2 <- data$x[idx[2]]; y2 <- data$y[idx[2]]
                  x3 <- data$x[idx[3]]; y3 <- data$y[idx[3]]

                  # Signed area: positive = CCW, negative = CW
                  signed_area <- (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)

                  if (signed_area < 0) {
                        tri[i, ] <- tri[i, c(1, 3, 2)]
                  }
            }

            # Build tile data
            tile_vertices <- data.frame(
                  tile_id = rep(1:n_tri, each = 3),
                  order = rep(1:3, n_tri),
                  .vertex_idx = as.vector(t(tri))
            )

            # Join with original data
            data$.vertex_idx <- 1:nrow(data)
            tile_vertices <- tile_vertices %>%
                  left_join(data, by = ".vertex_idx") %>%
                  select(-.vertex_idx)

            # Create group column
            tile_vertices <- tile_vertices %>%
                  mutate(group = paste0(group_prefix, tile_id, "::", group))
      })

      tile_vertices <- bind_rows(tile_vertices) %>%
            select(-tile_id) %>%
            arrange(group, order)

      return(tile_vertices)
}


#' Convert point grid to ridgeline polygons
#'
#' Takes a grid of points and creates ridgeline polygons where each slice
#' along one axis becomes a filled polygon.
#'
#' @param data Data frame with x, y, z columns.
#' @param direction Direction of ridges: "x" means one ridge per unique x value
#'   (ridge varies in y), "y" means one ridge per unique y value (ridge varies in x).
#' @param base Z-value for the bottom of ridgeline polygons. If NULL, uses min(z).
#' @param group_prefix String prefix for polygon group IDs.
#'
#' @return Data frame with polygon vertices for ridgeline rendering.
#'
#' @keywords internal
#' @noRd
points_to_ridgelines <- function(data,
                                 direction = "x",
                                 base = NULL,
                                 group_prefix = "ridgeline__") {

      direction <- match.arg(direction, c("x", "y"))

      # Get unique points (in case input has duplicates)
      points <- data %>%
            mutate(.x_round = round(x, 10),
                   .y_round = round(y, 10)) %>%
            distinct(.x_round, .y_round, .keep_all = TRUE) %>%
            select(-.x_round, -.y_round)

      # Compute base if not provided
      if (is.null(base)) {
            base <- min(points$z, na.rm = TRUE)
      }

      # Determine slice and ridge variables
      if (direction == "x") {
            slice_var <- "x"
            ridge_var <- "y"
      } else {
            slice_var <- "y"
            ridge_var <- "x"
      }

      # Get unique slice values
      slice_vals <- sort(unique(round(points[[slice_var]], 10)))

      if (length(slice_vals) < 1) {
            stop("No data points found for ridgeline generation")
      }

      # Build ridgeline polygons
      all_ridges <- vector("list", length(slice_vals))

      for (i in seq_along(slice_vals)) {
            slice_val <- slice_vals[i]

            # Get points for this slice
            slice_data <- points[abs(points[[slice_var]] - slice_val) < 1e-9, ]

            if (nrow(slice_data) < 2) {
                  warning("Slice at ", slice_var, " = ", slice_val,
                          " has fewer than 2 points; skipping")
                  next
            }

            # Sort by ridge variable
            slice_data <- slice_data[order(slice_data[[ridge_var]]), ]

            ridge_min <- min(slice_data[[ridge_var]])
            ridge_max <- max(slice_data[[ridge_var]])
            zmax <- max(slice_data$z, na.rm = TRUE)
            n_ridge_pts <- nrow(slice_data)

            # Build closed polygon vertices in CCW order when viewed from front with default projection
            # For direction="x", front faces negative x (toward typical viewer)
            # For direction="y", front faces negative y (toward typical viewer)

            if (direction == "x") {
                  # Polygon in y-z plane at constant x, front faces -x
                  # Looking from -x toward +x: y increases right, z increases up
                  # CCW from -x view: (y_max, base) -> ridge points decreasing y -> (y_min, base)
                  poly_x <- rep(slice_val, n_ridge_pts + 2)
                  poly_y <- c(ridge_min, slice_data$y, ridge_max)
                  poly_z <- c(base, slice_data$z, base)
            } else {
                  # Polygon in x-z plane at constant y, front faces -y
                  # Looking from -y toward +y: x increases right, z increases up
                  # CCW from -y view: (x_max, base) -> ridge points decreasing x -> (x_min, base)
                  poly_x <- c(ridge_min, slice_data$x, ridge_max)
                  poly_y <- rep(slice_val, n_ridge_pts + 2)
                  poly_z <- c(base, slice_data$z, base)
            }

            ridge_df <- data.frame(
                  x = poly_x,
                  y = poly_y,
                  z = poly_z,
                  zmax = zmax,
                  stringsAsFactors = FALSE
            ) %>% distinct()
            nr <- nrow(ridge_df)

            ridge_df$order <- 1:nr

            # Add face_type for axis-aligned normal computation
            # Ridgelines are vertical planes, so normals point along the slice axis
            # We want front faces to point toward typical viewer (negative x/y)
            ridge_df$face_type <- if (direction == "x") "xmin" else "ymin"

            # Preserve ggplot2 internal columns
            internal_cols <- c("PANEL", "cull_backfaces", "lighting_spec")
            for (int_col in internal_cols) {
                  if (int_col %in% names(slice_data)) {
                        ridge_df[[int_col]] <- slice_data[[int_col]][1]
                  }
            }

            # Create group name
            if ("group" %in% names(slice_data)) {
                  base_group <- sub(".*::", "", slice_data$group[1])
                  ridge_df$group <- paste0(group_prefix, i, "::", base_group)
            } else {
                  ridge_df$group <- paste0(group_prefix, i)
            }

            # Preserve aesthetic columns
            aes_cols <- intersect(names(slice_data),
                                  c("fill", "colour", "color", "alpha",
                                    "linewidth", "linetype", "size"))
            for (aes_col in aes_cols) {
                  vals <- slice_data[[aes_col]]
                  ridge_df[[aes_col]] <- if (is.numeric(vals)) mean(vals, na.rm = TRUE) else vals[1]
            }

            # split self-intersecting polygons
            if(direction == "x"){
                  polys <- polyclip::polysimplify(
                        A = list(x = ridge_df$y, y = ridge_df$z),
                        fillRule = "evenodd")
                  polys <- polys %>% lapply(function(p)
                        data.frame(x = ridge_df$x[1],
                                   y = p$x,
                                   z = p$y,
                                   order = 1:length(p$x)))
            } else {
                  polys <- polyclip::polysimplify(
                        A = list(x = ridge_df$x, y = ridge_df$z),
                        fillRule = "evenodd")
                  polys <- polys %>% lapply(function(p)
                        data.frame(x = p$x,
                                   y = ridge_df$y[1],
                                   z = p$y,
                                   order = 1:length(p$x)))
            }
            for(j in 1:length(polys)) polys[[j]]$group <- paste0(ridge_df$group[1], "::", j)
            ridge_df <- do.call("rbind", polys) %>%
                  bind_cols(distinct(select(ridge_df, -x, -y, -z, -order, -group)))

            all_ridges[[i]] <- ridge_df
      }

      # Combine results
      all_ridges <- all_ridges[!sapply(all_ridges, is.null)]
      if (length(all_ridges) == 0) {
            stop("No valid ridgeline slices could be created")
      }

      result <- do.call(rbind, all_ridges)
      rownames(result) <- NULL
      return(result)
}

---
title: "Introduction to ggcube"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to ggcube}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  dpi = 150,
  warning = FALSE,
  message = FALSE
)
```

ggcube is a ggplot2 extension for 3D plotting. It provides 3D geoms, stats, and a
3D coordinate system that let you build 3D visualizations using the familiar ggplot2
grammar: map your data to aesthetics, add layers, and customize the result with scales,
guides, and themes.

3D plots can be excellent for exploration, communication, and art. But it's important to
note that they can also give a distorted or obscured view of the data --- 2D alternative 
are generally recommended when precise quantitative data representation is important.


```{r setup, message = FALSE}
library(ggcube)
```


## The basics

The essential ingredient of a ggcube plot is `coord_3d()`. Adding it to a standard
ggplot that includes a `z` aesthetic creates a 3D plot:

```{r basics}
ggplot(mpg, aes(x = displ, y = hwy, z = drv, color = class)) +
      geom_point() +
      coord_3d()
```

Some standard ggplot2 layers like `geom_point()` work in 3D automatically. ggcube also
provides 3D-specific layer functions for surfaces, paths, volumes, and text, described
below.

Because ggcube works within ggplot2's 2D graphics engine, there are some important things
to understand about how rendering works. Each 3D layer projects its geometry onto a 2D
plane, with depth sorting to determine which elements appear in front of others. This
sorting happens *within* each layer, but not *across* layers — just as in standard
ggplot2, later layers are drawn on top of earlier ones regardless of their 3D depth.
This means that layer order in your code matters, and complex multi-layer scenes may
require some thought about stacking.


## Controlling the view

`coord_3d()` controls how the 3D scene is projected onto 2D. Rotation is specified via
three angles --- `pitch` (tilt around the x-axis), `roll` (tilt around the y-axis), and
`yaw` (spin around the z-axis):

```{r rotation, fig.height = 4}
ggplot(mpg, aes(displ, hwy, drv, color = class)) +
      geom_point() +
      coord_3d(pitch = 0, roll = 60, yaw = 0, dist = 1.4,
               ratio = c(2, 1, 1), panels = "all") +
      theme(panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .1))
```

Perspective projection is on by default, making distant objects appear smaller. The
`dist` parameter controls the strength of this effect (larger values = less distortion),
and `persp = FALSE` switches to orthographic projection where parallel lines stay
parallel.

The `scales` parameter controls how axis ranges map to visual size. `"free"` (the
default) stretches each axis independently to fill the cube, while `"fixed"` preserves
the relative scale of the data (like `coord_fixed()` in 2D). The `ratio` parameter lets
you set custom proportions for the three axes. And `zoom` adjusts overall framing without
changing the rotation or projection.

See the [3D view](https://matthewkling.github.io/ggcube/articles/coord_3d.html) article for a comprehensive guide to 
all view parameters.


## 3D layers

Most standard ggplot2 geoms are designed around 2D coordinate assumptions, and won't
render correctly with `coord_3d()`. (An exception is `geom_point()`, which works out
of the box as shown above, albeit with ordering and sizing limitations.) ggcube provides 
a range of 3D-native layer functions that cover common 3D plot types, including points, 
surfaces, bars, paths, and text.

Here's a quick tour of the main categories.

### Surfaces

Several geoms and stats work together to render surfaces. `geom_surface_3d()` renders
data as a tessellated surface, `geom_contour_3d()` creates layer-cake
contour stacks, and `geom_ridgeline_3d()` shows cross-sectional slices:

```{r surface}
ggplot(mountain, aes(x, y, z)) +
      geom_surface_3d(aes(fill = z, color = z)) +
      scale_fill_viridis_c() + scale_color_viridis_c() +
      coord_3d(ratio = c(1.5, 2, 1), expand = FALSE, panels = "zmin") +
      light(direction = c(1, 0, 0)) +
      guides(fill = guide_colorbar_3d()) +
      theme_light()
```

Stats like `stat_function_3d()`, `stat_smooth_3d()`, and `stat_density_3d()` generate
surface data from functions, model fits, or kernel density estimates. These stats can be
paired with any of the surface geoms. See the [3D surfaces](https://matthewkling.github.io/ggcube/articles/surfaces.html)
article for more detail on surface plotting options.

### Points

`geom_point_3d()` extends scatter plots with depth-scaled point sizes (closer points shown larger), 
proper depth sorting (closer points plotted on top), and optional reference lines and points that 
project onto cube faces:

```{r points}
ggplot(mpg, aes(x = displ, y = hwy, z = drv, fill = class)) +
      geom_point_3d(size = 3, shape = 21, color = "black", stroke = .1,
                    ref_lines = TRUE, ref_faces = c("ymax", "xmax")) +
      coord_3d()
```

### Paths

`geom_path_3d()` connects observations with depth-sorted, depth-scaled line segments:

```{r paths}
x <- seq(0, 20*pi, pi/16)
spiral <- data.frame(x = x, y = sin(x), z = cos(x), time = 1:length(x))
ggplot(spiral, aes(x, y, z, color = time)) +
      geom_path_3d() +
      scale_color_gradientn(colors = c("blue", "purple", "red", "orange")) +
      coord_3d() +
      theme_light()
```

`geom_segment_3d()` is also available for drawing individual segments defined by start
and end coordinates.

### Prisms

`geom_col_3d()` produces 3D column charts, `geom_bar_3d()` creates 3D histograms with
automatic binning, and `geom_voxel_3d()` renders arrays of cubes:

```{r bar}
ggplot(iris, aes(Species, Sepal.Length, fill = Species)) +
      geom_bar_3d(bins = 20, width = c(.5, 1)) +
      coord_3d(scales = "fixed", ratio = c(1, 3, .25), yaw = 60) +
      scale_z_continuous(expand = c(0, 0)) +
      theme(legend.position = "none")
```

### Hulls

`geom_hull_3d()` computes and renders triangulated hulls from 3D point clouds, including convex and alpha hulls:

```{r hull}
ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d(method = "convex", fill = "#9e2602", color = "#5e1600") +
      coord_3d()
```

### Distributions

`stat_distributions_3d()` computes 1D kernel density estimates per group and arranges
them as ridgeline surfaces, the 3D analog of `ggridges::geom_density_ridges()`:

```{r distributions}
ggplot(iris, aes(y = Sepal.Length, x = Species, fill = Species)) +
      stat_distributions_3d() +
      scale_z_continuous(expand = expansion(mult = c(0, NA))) +
      coord_3d() +
      theme(legend.position = "none")
```

### Text

`geom_text_3d()` renders text in 3D, either as "billboard" labels that always face the
camera, or as polygon outlines that can be oriented in any direction:

```{r text}
df <- expand.grid(x = c("H", "B"), y = c("a", "o", "u"), z = c("g", "t"))
df$label <- paste0(df$x, df$y, df$z)
ggplot(df, aes(x, y, z, label = label, fill = x)) +
      geom_text_3d(method = "polygon", facing = "zmax",
                   size = 5, weight = "bold") +
      coord_3d(scales = "fixed", light = NULL)
```


## Lighting

Lighting modifies the fill and/or color of polygon faces based on their orientation
relative to a light source, giving surfaces a sense of depth and shape. It's controlled
via the `light()` function, which can be added to a plot or passed to `coord_3d()`
(applying to all layers) or passed to individual layer functions (overriding the 
plot-level setting):

```{r lighting}
p <- ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d(fill = "#9e2602", color = "#5e1600") +
      coord_3d()

p + light(method = "direct", mode = "hsl", direction = c(0, 0, 1))
```

Use `light("none")` to disable lighting entirely, or `light = NULL` in a layer to
inherit the coord-level setting. For a comprehensive guide to lighting methods, color
modes, light direction, and backface handling, see the [lighting](https://matthewkling.github.io/ggcube/articles/lighting.html) article.


## Scales, guides, and themes

### Z-axis scales

ggcube provides `scale_z_continuous()` and `scale_z_discrete()` for controlling the
z-axis, with the same interface as their x/y counterparts. `zlim()` is a shorthand for
setting z-axis limits:

```{r zlim, fig.show = "hide"}
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
      geom_point() +
      zlim(15, 20) +
      coord_3d()
```

### Shaded guides

When lighting is active, standard color guides don't reflect the shading visible in the
plot. `guide_colorbar_3d()` and `guide_legend_3d()` create guides that show the range of
shaded colors:

```{r guide, fig.show = "hide"}
ggplot(mountain, aes(x, y, z, fill = z)) +
      stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
      guides(fill = guide_colorbar_3d()) +
      scale_fill_gradientn(colors = c("tomato", "dodgerblue")) +
      coord_3d()
```

### Panels and themes

The `panels` argument to `coord_3d()` controls which cube faces are drawn. Faces behind
the data are "background" panels; faces in front are "foreground" panels (which default to
semi-transparent so they don't obscure the data):

```{r theme}
ggplot(sphere_points, aes(x, y, z)) +
      geom_hull_3d() +
      coord_3d(panels = "all") +
      theme(panel.background = element_rect(color = "black"),
            panel.border = element_rect(color = "black"),
            panel.foreground = element_rect(alpha = .3),
            panel.grid.foreground = element_line(color = "gray", linewidth = .25),
            axis.text = element_text(color = "darkblue"),
            axis.text.z = element_text(color = "darkred"),
            axis.title = element_text(margin = margin(t = 30)),
            axis.title.x = element_text(color = "magenta"))
```

Standard ggplot2 themes and `theme()` customization work as expected. ggcube adds
foreground-specific elements (`panel.foreground`, `panel.grid.foreground`,
`panel.border.foreground`) and z-axis text elements (`axis.text.z`, `axis.title.z`).
ggcube's `element_rect()` extends ggplot2's version with an `alpha` parameter for
transparency. For details, see the theming section of the 
[3D coordinates](https://matthewkling.github.io/ggcube/articles/coord_3d.html)
article.


## Annotations

`annotate_3d()` adds reference geometry --- points, text labels, or segments --- to any
3D layer. Unlike adding a separate layer, annotations are embedded within
the host layer so they participate in the same depth sorting:

```{r annotate, eval = FALSE}
summit <- filter(mountain, z == max(z))
ggplot(mountain, aes(x, y, z)) +
      geom_contour_3d(
            annotate = list(
                  annotate_3d("point", x = summit$x, y = summit$y, z = summit$z, color = "red"),
                  annotate_3d("text", x = summit$x, y = summit$y, z = summit$z, color = "red",
                              label = "Summit", fontface = "bold", vjust = -1)
            ), fill = "black"
      ) +
      coord_3d(ratio = c(2, 3, 1.5), light = "none")
```


## Mixing 2D and 3D

`position_on_face()` lets you project layers onto cube faces, enabling a mix of 3D and
2D content. You can flatten a 3D layer onto a face, or place a natively 2D layer (like
`stat_density_2d()`) onto a specific face using the `axes` parameter:

```{r face_projection}
ggplot(iris, aes(Sepal.Length, Sepal.Width, Petal.Length,
                 color = Species, fill = Species)) +
      coord_3d() + xlim(4, 8) +
      stat_density_2d(position = position_on_face(faces = "zmin", axes = c("x", "y")),
                      geom = "polygon", alpha = .1, linewidth = .25) +
      geom_hull_3d(position = position_on_face("ymax"), alpha = .5) +
      geom_point_3d(shape = 21, color = "black", stroke = .25)
```


## Interactivity & animation

A major limitation of 3D figures is that you can't get a full view of the data from any 
single angle. One way to mitigate this is by rotating the figure to view the data from 
different directions. ggcube offers two options for rotating 3D plots. `animate_3d()`
makes a gif or movie of a spinning plot across a specified sequence of rotation angles,
while `orbit_3d()` produces an interactive HTML widget that you can drag to rotate. 
Both functions take a pre-made plot and a range of rotation angles:

```{r animation, eval = FALSE}
p <- ggplot(mountain, aes(x, y, z)) +
      geom_contour_3d(fill = "black", color = "white", linewidth = .5, light = "none") +
      coord_3d(ratio = c(1.5, 2, 1), zoom = 1.25) +
      theme_void()

orbit_3d(p, yaw = c(360, 0), roll = c(-90, 0), n = c(24, 8),
         start = c(yaw = 300, roll = -60))
```

See the [animation and interaction](https://matthewkling.github.io/ggcube/articles/animation.html) 
article for more examples.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/orbit-3d.R
\name{orbit_3d}
\alias{orbit_3d}
\title{Create an interactive drag-to-rotate ggcube}
\usage{
orbit_3d(
  plot,
  pitch = NULL,
  roll = NULL,
  yaw = NULL,
  n = 36,
  loop = NULL,
  start = NULL,
  max_frames = 500,
  width = 480,
  height = 480,
  res = 96,
  cores = 1,
  device = "png",
  progress = interactive(),
  file = NULL
)
}
\arguments{
\item{plot}{A ggplot object that uses \code{\link[=coord_3d]{coord_3d()}}.}

\item{pitch, roll, yaw}{Rotation angles in degrees. Each is either a single
value (held fixed) or a length-2 range \code{c(from, to)} that becomes a
draggable axis. For \code{yaw}, using a larger value for \code{from} than \code{to} gives
more natural results. Values specified here replace the ones specified in
\code{coord_3d()}. See Details.}

\item{n}{Grid resolution. For a 1D orbit, a single number giving the
frame count. For a 2D orbit, a length-2 vector giving the counts for
the horizontal and vertical screen axes respectively (in that order, x then y).
Its length must match the number of ranged parameters. Default is 36.}

\item{loop}{Controls seamless wraparound. \code{NULL} (default)
auto-detects looping independently for each axis by checking whether that
axis returns to its starting view. \code{TRUE} or \code{FALSE} force
wrapping on or off for all axes.}

\item{start}{Initial viewing angle(s) the orbit opens at, given in
degrees as a named vector matching the variable axes, e.g.
\code{start = c(yaw = 30)} or \code{start = c(yaw = 30, roll = 45)}. For a
1D orbit a bare scalar (e.g. \code{start = 30}) is also accepted. The
orbit opens at the frame nearest each given angle. Any variable axis not
named in \code{start} opens at the midpoint of its range, as does every
axis when \code{start = NULL} (the default). Naming an axis that is
fixed rather than variable throas an error.}

\item{max_frames}{Safety cap on the total number of rendered frames
(\code{prod(n)}). A 2D grid can become very large; if the requested grid
exceeds this cap the function errors rather than rendering. Raise it to
allow larger grids. Default is 500.}

\item{width, height}{Dimensions of the animation in pixels. Defaults to
480 x 480.}

\item{res}{Resolution in ppi for the rendered frames. Default is 96.
Controls the size of point-based text/lines/point elements relative to
the plot. See \code{?grDevices::png} for details.}

\item{cores}{Number of CPU cores for parallel frame rendering. Default
is 1 (sequential). Values greater than 1 use \code{\link[parallel:parLapply]{parallel::parLapply()}}
to render frames in parallel, which can significantly speed up
animation of complex plots. Note that progress reporting is not
available during parallel rendering.}

\item{device}{The graphics device to use for rendering frames.
Default is \code{"png"}. Other options include \code{"ragg_png"}
(requires the \pkg{ragg} package).}

\item{progress}{Whether to print a progress bar. Ignored when \code{cores > 1}.}

\item{file}{Optional path to write a self-contained HTML file (via
\code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}}). If \code{NULL} (default), no file is written and
the widget is simply returned for display or embedding.}
}
\value{
A \code{orbit_3d} htmlwidget. Printed at the console it opens in the
RStudio Viewer or browser; in a knitr document it embeds automatically;
in Shiny use \code{\link[=orbit3dOutput]{orbit3dOutput()}} and \code{\link[=renderOrbit3d]{renderOrbit3d()}}.
}
\description{
Creates an interactive HTML widget that you can rotate by dragging.
Pre-renders a set of ggcube plots across a 1D or 2D grid of rotation angles
and packages them as an \pkg{htmlwidgets} widget. The widget displays in the
RStudio Viewer, embeds in R Markdown and Quarto documents, works in Shiny
(see \code{\link[=orbit3dOutput]{orbit3dOutput()}}), and can be written to a self-contained HTML
file with \verb{file =}. When viewing it, navigate through the views by dragging
the image or using the keyboard arrows.
}
\section{Specifying rotation}{

Each of \code{pitch}, \code{roll}, and \code{yaw} is given either as a
single value (held fixed) or as a length-2 range \code{c(from, to)} that
is a variable, draggable axis. The number of ranges determines the
interaction:
\itemize{
\item One range gives a 1D orbit: horizontal drag (or left/right
keyboard arrows) vary that one parameter across its range.
\item Two ranges give a 2D orbit: horizontal drag (or left/right
keyboard arrows) vary one parameter while vertical drag/arrows vary
the other.
}
At least one range is required. Supplying three ranges is an error.

For a 2D orbit, the horizontal (x) and vertical (y) axes are assigned
from whichever parameters are ranged, by priority. The x axis takes the
highest-priority ranged parameter in the order \code{yaw}, \code{pitch},
\code{roll}; the y axis takes the highest-priority remaining ranged
parameter in the order \code{roll}, \code{pitch}. In the common case
\code{yaw = c(0, 360), roll = c(...)} this puts yaw on horizontal drag and
roll on vertical drag.

The range direction follows the order the values are written: for example
\code{yaw = c(360, 0)} rotates the opposite way from \code{yaw = c(0, 360)}.
}

\section{Looping}{

When a variable axis returns to its starting view (for example a full
\code{yaw = c(0, 360)} turn), that axis loops seamlessly: dragging past
either end wraps around. This is detected per axis by comparing the first
and last angle modulo 360, so a full-turn axis wraps while a partial range
(such as \code{roll = c(-90, 90)}) clamps. \code{loop} can force this on or
off for all axes at once.
}

\section{Keyboard}{

Once the widget has focus, the arrow keys step one frame: left/right move
the horizontal axis, and (in a 2D orbit) up/down move the vertical axis.
}

\examples{
\dontshow{if (requireNamespace("base64enc", quietly = TRUE) && requireNamespace("htmlwidgets", quietly = TRUE)) withAutoprint(\{ # examplesIf}
\donttest{
p <- ggplot(mountain, aes(x, y, z, fill = z, color = z)) +
  geom_surface_3d(linewidth = .25) +
  coord_3d(light = light(anchor = "camera",direction = c(1, 1, 0)),
           ratio = c(1.5, 2, 1)) +
  scale_fill_viridis_c() + scale_color_viridis_c() +
  theme_void()

# 1D drag-to-rotate turntable (small n to minimize runtime)
orbit_3d(p, yaw = c(360, 0), n = 12)

# 2D orbit: drag horizontally to spin (yaw), vertically to tilt (roll)
# (note you could increase `cores` to speed up compute time)
orbit_3d(p, yaw = c(360, 0), roll = c(-90, 0), n = c(20, 10),
            width = 1000, height = 1000, cores = 1)

# Save to a chosen path
orbit_3d(p, yaw = c(360, 0), n = 12,
            file = file.path(tempdir(), "surface.html"))
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=animate_3d]{animate_3d()}} for GIF/MP4 output.
\link{orbit_3d-shiny} for including orbits in  Shiny apps.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{position_param}
\alias{position_param}
\title{Position param}
\arguments{
\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}
}
\description{
Position param
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-distributions-3d.R
\name{stat_distributions_3d}
\alias{stat_distributions_3d}
\alias{geom_distributions_3d}
\title{3D ridgeline distributions}
\usage{
stat_distributions_3d(
  mapping = NULL,
  data = NULL,
  geom = "ridgeline_3d",
  position = "identity",
  ...,
  direction = NULL,
  bw = "nrd0",
  adjust = 1,
  kernel = "gaussian",
  n = 512,
  trim = FALSE,
  bounds = c(-Inf, Inf),
  rel_min_height = 0,
  joint_bandwidth = FALSE,
  base = 0,
  light = NULL,
  cull_backfaces = FALSE,
  sort_method = NULL,
  scale_depth = TRUE,
  force_convex = FALSE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

geom_distributions_3d(
  mapping = NULL,
  data = NULL,
  stat = "distributions_3d",
  position = "identity",
  ...,
  direction = NULL,
  bw = "nrd0",
  adjust = 1,
  kernel = "gaussian",
  n = 512,
  trim = FALSE,
  bounds = c(-Inf, Inf),
  rel_min_height = 0,
  joint_bandwidth = FALSE,
  base = 0,
  light = NULL,
  cull_backfaces = FALSE,
  sort_method = NULL,
  scale_depth = TRUE,
  force_convex = FALSE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. This stat
requires \code{x} and \code{y} aesthetics. One of these serves as the grouping/position
variable (determined by \code{direction}), and the other provides values for
density estimation.}

\item{data}{The data to be displayed in this layer.}

\item{geom}{The geometric object to use to display the data. Defaults to
\code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed to the layer.}

\item{direction}{Direction of ridges:
\describe{
\item{"x"}{One ridge per unique x value; ridge varies in y (default)}
\item{"y"}{One ridge per unique y value; ridge varies in x}
}}

\item{bw}{The smoothing bandwidth to be used. If numeric, the standard
deviation of the smoothing kernel. If character, a rule to choose the
bandwidth, as listed in \code{\link[stats:bw.nrd]{stats::bw.nrd()}}. Options include \code{"nrd0"} (default),
\code{"nrd"}, \code{"ucv"}, \code{"bcv"}, \code{"SJ"}, \code{"SJ-ste"}, and \code{"SJ-dpi"}. Note that

automatic calculation is performed per-group unless \code{joint_bandwidth = TRUE}.}

\item{adjust}{A multiplicative bandwidth adjustment. This makes it possible
to adjust the bandwidth while still using a bandwidth estimator. For
example, \code{adjust = 1/2} means use half of the default bandwidth. Default is 1.}

\item{kernel}{Kernel function to use. One of \code{"gaussian"} (default),
\code{"rectangular"}, \code{"triangular"}, \code{"epanechnikov"}, \code{"biweight"}, \code{"cosine"},
or \code{"optcosine"}. See \code{\link[stats:density]{stats::density()}} for details.}

\item{n}{Number of equally spaced points at which the density is estimated.
Should be a power of two for efficiency. Default is 512.}

\item{trim}{If \code{FALSE} (the default), each density is computed on the full
range of the data (extended by a factor based on bandwidth). If \code{TRUE},
each density is computed over the range of that group only.}

\item{bounds}{A numeric vector of length 2 giving the lower and upper bounds
for bounded density estimation. Density values outside bounds are set to
zero. Data points outside bounds are removed with a warning.
Default is \code{c(-Inf, Inf)} (unbounded).}

\item{rel_min_height}{Lines with heights below this cutoff will be removed.
The cutoff is measured relative to the maximum height within each group.
For example, \code{rel_min_height = 0.01} removes points with density less than
1\% of the peak. Default is 0 (no removal). This is similar to the
parameter of the same name in \code{ggridges::geom_density_ridges()}.}

\item{joint_bandwidth}{If \code{TRUE}, bandwidth is computed jointly across all
groups using the specified \code{bw} method, ensuring consistent smoothing
across all density curves. This matches the behavior of
\code{\link[ggridges:stat_density_ridges]{ggridges::stat_density_ridges()}}. If \code{FALSE} (the default), bandwidth is
computed separately for each group, matching \code{\link[ggplot2:stat_density]{ggplot2::stat_density()}}.
Only applies when \code{bw} is a character string (bandwidth rule), not when
\code{bw} is provided as a numeric value.}

\item{base}{Z-value for ridge polygon bottoms. If NULL, uses min(z).}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{stat}{Statistical transformation to use on the data. Defaults to
\code{\link[=stat_distributions_3d]{stat_distributions_3d()}}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Computes 1D kernel density estimates for each group and arranges them as
ridgeline polygons in 3D space. Similar to \code{ggridges::geom_density_ridges()},
but rendered as 3D surfaces using \code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}}.
}
\details{
This stat is modeled after \code{\link[ggplot2:stat_density]{ggplot2::stat_density()}}, with similar
parametrization for bandwidth selection, kernel choice, and boundary handling.
}
\section{Aesthetics}{

\code{stat_distributions_3d()} understands the following aesthetics (required
aesthetics are in bold):
\describe{
\item{\strong{x}}{X coordinate - either density variable or position variable
depending on \code{direction}}
\item{\strong{y}}{Y coordinate - either position variable or density variable
depending on \code{direction}}
\item{group}{Grouping variable (typically derived from the position aesthetic)}
\item{fill, colour, alpha, linewidth, linetype}{Passed to \code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}}}
}
}

\section{Direction}{

The \code{direction} parameter determines how the data is interpreted:
\describe{
\item{\code{direction = NULL} (default)}{Automatically detects direction based
on whether \code{x} or \code{y} is discrete (factor/character). If \code{x} is discrete
and \code{y} is continuous, uses \code{"x"}; if \code{y} is discrete and \code{x} is
continuous, uses \code{"y"}. Falls back to \code{"x"} if ambiguous.}
\item{\code{direction = "x"}}{Ridges march along the x-axis. Each unique
x value defines a group, and density is computed from the y values within
that group. The resulting density curves lie in the y-z plane.}
\item{\code{direction = "y"}}{Ridges march along the y-axis. Each unique y value
defines a group, and density is computed from the x values. The resulting
density curves lie in the x-z plane.}
}
}

\section{Computed variables}{

The following variables are computed and available via \code{after_stat()}:
\describe{
\item{density}{The kernel density estimate at each point}
\item{ndensity}{Density normalized to a maximum of 1 within each group}
\item{count}{Density multiplied by number of observations (expected count)}
\item{n}{Number of observations in the group}
\item{bw}{Bandwidth actually used for this group}
}
}

\examples{
library(ggplot2)

# Basic usage with iris data
p <- ggplot(iris, aes(y = Sepal.Length, x = Species, fill = Species)) +
  coord_3d() +
  scale_z_continuous(expand = expansion(mult = c(0, NA))) + # remove gap beneath ridges
  theme(legend.position = "none")
p + stat_distributions_3d()

# Normalize max ridge heights
p + stat_distributions_3d(aes(z = after_stat(ndensity)))

# Adjust smoothing bandwidth
p + stat_distributions_3d(adjust = 0.5)

# Use joint bandwidth for consistent smoothing across groups
p + stat_distributions_3d(joint_bandwidth = TRUE)

# Different bandwidth selection rules
p + stat_distributions_3d(bw = "SJ")

# Remove tails with rel_min_height
p + stat_distributions_3d(rel_min_height = 0.05)

# Trim to data range
p + stat_distributions_3d(trim = TRUE)

# Rotated to reduce perspective distortion
p + stat_distributions_3d(alpha = .7) +
   coord_3d(pitch = 0, roll = -90, yaw = 90, dist = 5,
      panels = c("zmin", "xmin"))

}
\seealso{
\code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}} for rendering pre-computed ridgeline data,
\code{\link[=stat_density_3d]{stat_density_3d()}} for 2D kernel density surfaces,
\code{\link[ggplot2:stat_density]{ggplot2::stat_density()}} for the parametrization this stat follows,
\code{ggridges::geom_density_ridges()} for the 2D ridgeline equivalent.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotate-3d.R
\name{annotate_3d}
\alias{annotate_3d}
\title{Create a 3D annotation specification}
\usage{
annotate_3d(type, ...)
}
\arguments{
\item{type}{Character string specifying the annotation type. One of
\code{"point"}, \code{"text"}, or \code{"segment"}.}

\item{...}{Type-specific parameters. See Details.}
}
\value{
An S3 object of class \code{"annotate_3d"}.
}
\description{
Defines fixed-position annotations to be embedded within a 3D layer.
Annotations are depth-sorted together with the primary geometry, unlike
ggplot2's \code{annotate()} which creates a separate layer.
}
\details{
Parameters can be vectors to create multiple annotations of the same type
in one call, with scalar values recycled as needed (matching
\code{ggplot2::annotate()} behaviour).
\subsection{Point annotations}{

Requires \code{x}, \code{y}, \code{z}. Optional styling: \code{colour}/\code{color}, \code{fill}, \code{size},
\code{shape}, \code{alpha}, \code{stroke}.
}

\subsection{Text annotations}{

Requires \code{x}, \code{y}, \code{z}, \code{label}. Optional styling: \code{colour}/\code{color}, \code{size},
\code{alpha}, \code{family}, \code{fontface}, \code{hjust}, \code{vjust}, \code{angle}, \code{lineheight}.
}

\subsection{Segment annotations}{

Requires \code{x}, \code{y}, \code{z}, \code{xend}, \code{yend}, \code{zend}. Optional styling:
\code{colour}/\code{color}, \code{linewidth}, \code{linetype}, \code{alpha}.
}
}
\examples{
p <- ggplot(mountain, aes(x, y, z)) +
  coord_3d(ratio = c(2, 3, 1),
           light = light(mode = "hsl", direction = c(-1, 0, 0)))

# Basic point annotation
p + geom_surface_3d(
  annotate = annotate_3d("point", x = .5, y = .25, z = 100,
                         color = "red", size = 3))

# Vectorized: multiple points in one call
p + geom_surface_3d(
  annotate = annotate_3d("point", x = .5, y = .25, z = seq(50, 100, 5),
                         color = "red", size = 3))

# Multiple annotation types
p + geom_surface_3d(
  annotate = list(
    annotate_3d("point", x = .5, y = .25, z = 100,
                color = "red", size = 3),
    annotate_3d("text", x = .5, y = .25, z = 100,
                color = "red", label = "Look here!",
                vjust = -1, fontface = "bold"),
    annotate_3d("segment", x = .5, y = .25, z = 100,
                xend = .5, yend = .25, zend = 50,
                color = "red", size = 3)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{col_params}
\alias{col_params}
\title{Shared column and voxel params}
\arguments{
\item{width}{Numeric value controlling box width as a fraction of grid spacing.
Default is 1.0 (volumes touch each other). Use 0.8 for small gaps, 1.2 for overlap.
Grid spacing is determined automatically using \code{ggplot2::resolution()}}

\item{faces}{Character vector specifying which faces to render. Options:
\itemize{
\item \code{"all"} (default): Render all 6 faces
\item \code{"none"}: Render no faces
\item Vector of face names: \code{c("zmax", "xmin", "ymax")}, etc.
}
Valid face names: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
Note that this setting acts jointly with backface culling, which removes faces whose
interior faces the viewer -- e.g., when \code{cull_backfaces = TRUE} and \code{faces = "all"}
(the default), only front faces are rendered.}
}
\description{
Shared column and voxel params
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{depth_params}
\alias{depth_params}
\title{Depth scaling parameter}
\arguments{
\item{scale_depth}{Controls depth-based scaling, drawing closer elements
larger or thicker and farther ones smaller or thinner. \code{TRUE} (the default)
applies full scaling and \code{FALSE} disables it; a number sets the strength
directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below 1
subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for panel and axis components.}
}
\description{
Generic wording for layers whose depth-scaled property is something other
than polygon linewidth. Polygon layers document \code{scale_depth} through
\link{polygon_params} instead, which describes the mean-depth behaviour specific
to them. Keep the two in step when the semantics change.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-density-3d.R
\name{geom_density_3d}
\alias{geom_density_3d}
\alias{stat_density_3d}
\title{3D surface from 2D density estimate}
\usage{
geom_density_3d(
  mapping = NULL,
  data = NULL,
  stat = "density_3d",
  position = "identity",
  ...,
  grid = "rectangle",
  n = 40,
  direction = "x",
  trim = TRUE,
  h = NULL,
  adjust = 1,
  pad = 0.1,
  min_ndensity = 0,
  light = NULL,
  cull_backfaces = FALSE,
  sort_method = NULL,
  force_convex = FALSE,
  scale_depth = TRUE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_density_3d(
  mapping = NULL,
  data = NULL,
  geom = "surface_3d",
  position = "identity",
  ...,
  grid = "rectangle",
  n = 40,
  direction = "x",
  trim = TRUE,
  h = NULL,
  adjust = 1,
  pad = 0.1,
  min_ndensity = 0,
  light = NULL,
  cull_backfaces = FALSE,
  sort_method = NULL,
  force_convex = FALSE,
  scale_depth = TRUE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. This stat
requires \code{x} and \code{y} aesthetics. By default, \code{fill} is mapped to
\code{after_stat(density)} and \code{z} is mapped to \code{after_stat(density)}.}

\item{data}{The data to be displayed in this layer. Must contain x and y columns
with point coordinates.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{StatDensity3D}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}

\item{grid, n, direction, trim}{Parameters determining the geometry, resolution, and
orientation of the surface grid. See \link{grid_generation} for details.}

\item{h}{Bandwidth vector. If \code{NULL} (default), uses automatic bandwidth selection
via \code{MASS::bandwidth.nrd()}. Can be a single number (used for both dimensions)
or a vector of length 2 for different bandwidths in x and y directions.}

\item{adjust}{Multiplicative bandwidth adjustment factor. Values greater than 1
produce smoother surfaces; values less than 1 produce more detailed surfaces.
Default is 1.}

\item{pad}{Proportional range expansion factor. The computed density grid extends
this proportion of the raw data range beyond each data limit. Default is 0.1.}

\item{min_ndensity}{Lower cutoff for normalized density (computed variable \code{ndensity}
described below), below which to filter out results. This is particularly useful for
removing low-density corners of rectangular density grids when density surfaces are
shown for multiple groups, as in the example below. Default is 0 (no filtering).}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{The geometric object used to display the data. Defaults to \code{GeomPolygon3D}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
A 3D version of \code{ggplot2::stat_density_2d()}.
Creates surfaces from 2D point data using kernel density estimation.
The density values become the z-coordinates of the surface, allowing
visualization of data concentration as peaks and valleys in 3D space.
}
\section{Aesthetics}{

\code{stat_density_3d()} requires the following aesthetics from input data:
\itemize{
\item \strong{x}: X coordinate of data points
\item \strong{y}: Y coordinate of data points
}

And optionally understands:
\itemize{
\item \strong{group}: Grouping variable for computing separate density surfaces
\item Additional aesthetics are passed through for surface styling
}
}

\section{Computed variables specific to StatDensity3D}{

\itemize{
\item \code{density}: The kernel density estimate at each grid point
\item \code{ndensity}: Density estimate scaled to maximum of 1 within each group
\item \code{count}: Density estimate × number of observations in group (expected count)
\item \code{n}: Number of observations in each group
}
}

\section{Grouping}{

When aesthetics like \code{colour} or \code{fill} are mapped to categorical variables,
\code{stat_density_3d()} computes separate density surfaces for each group, just
like \code{stat_density_2d()}. Each group gets its own density calculation with
proper \code{count} and \code{n} values.
}

\section{Computed variables}{

The following computed variables are available via \code{after_stat()}:
\itemize{
\item \code{x}, \code{y}, \code{z}: Grid coordinates and function values
\item \code{normal_x}, \code{normal_y}, \code{normal_z}: Surface normal components
\item \code{slope}: Gradient magnitude from surface calculations
\item \code{aspect}: Direction of steepest slope from surface calculations
\item \code{dzdx}, \code{dzdy}: Partial derivatives from surface calculation
}
}

\examples{
# Basic density surface from scattered points
p <- ggplot(faithful, aes(eruptions, waiting)) +
  coord_3d() +
  scale_fill_viridis_c()
p + geom_density_3d() + guides(fill = guide_colorbar_3d())

\donttest{

# Specify alternative grid geometry and light model
p + geom_density_3d(grid = "equilateral", n = 30, direction = "y",
                    light = light("direct"),
                    color = "white", linewidth = .1) +
  guides(fill = guide_colorbar_3d())

# Color by alternative density metric
p + geom_density_3d(aes(fill = after_stat(count)))

# Adjust bandwidth for smoother or more detailed surfaces
p + geom_density_3d(adjust = 0.5, n = 100, color = "white")  # More detail
p + geom_density_3d(adjust = 2, color = "white")   # Smoother

# As contour plot instead of default surface plot
p + stat_density_3d(geom = "contour_3d", light = "none",
                    color = "black", bins = 25,
                    sort_method = "pairwise")

# Multiple density surfaces by group,
# using normalized density to equalize peak heights
ggplot(iris, aes(Petal.Length, Sepal.Length, fill = Species)) +
  geom_density_3d(aes(z = after_stat(ndensity), group = Species),
                  color = "black", alpha = .7, light = NULL) +
  coord_3d()

# Same, but with extra padding to remove edge effects and
# with density filtering to remove rectangular artifacts
ggplot(iris, aes(Petal.Length, Sepal.Length, fill = Species)) +
  geom_density_3d(aes(z = after_stat(ndensity)),
                  pad = .3, min_ndensity = .001,
                  color = "black", alpha = .7, light = NULL) +
  coord_3d(ratio = c(3, 3, 1))

}

}
\seealso{
\code{\link[ggplot2:stat_density_2d]{ggplot2::stat_density_2d()}} for 2D density contours, \code{\link[=stat_surface_3d]{stat_surface_3d()}} for
surfaces from existing grid data, \code{\link[=light]{light()}} for lighting specifications,
\code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/animate-3d.R
\name{animate_3d}
\alias{animate_3d}
\title{Animate a 3D plot}
\usage{
animate_3d(
  plot,
  pitch = NULL,
  roll = NULL,
  yaw = NULL,
  nframes = NULL,
  fps = NULL,
  duration = NULL,
  width = 480,
  height = 480,
  res = 96,
  renderer = gifski_renderer_3d(),
  start_pause = 0,
  end_pause = 0,
  rewind = FALSE,
  cores = 1,
  device = "png",
  progress = interactive()
)
}
\arguments{
\item{plot}{A ggplot object that uses \code{\link[=coord_3d]{coord_3d()}}.}

\item{pitch, roll, yaw}{Rotation parameters to animate. Each can be:
\itemize{
\item \code{NULL} (default): Hold at the value specified in the plot's
\code{\link[=coord_3d]{coord_3d()}} call.
\item A single numeric value: Hold at this value for all frames
(overriding the plot's coord).
\item A numeric vector of length 2+: Keyframe values that are
interpolated linearly across the animation. Keyframes are evenly
spaced, so \code{c(0, 180, 180)} with 100 frames means: start at 0,
reach 180 at frame 50, hold at 180 until frame 100.
}}

\item{nframes, fps, duration}{Control animation length and speed. Specify
any two of three; the third is computed as
\code{duration = nframes / fps}. Defaults: \code{nframes = 100},
\code{fps = 10}.}

\item{width, height}{Dimensions of the animation in pixels. Defaults to
480 x 480.}

\item{res}{Resolution in ppi for the rendered frames. Default is 96.
Controls the size of point-based text/lines/point elements relative to
the plot. See \code{?grDevices::png} for details.}

\item{renderer}{A renderer function that combines image frames into
an animation file. Built-in options:
\itemize{
\item \code{gifski_renderer_3d()} (default): Renders a GIF using the
\pkg{gifski} package.
\item \code{av_renderer_3d()}: Renders a video using the \pkg{av}
package.
\item \code{file_renderer_3d()}: Returns the frame image files
without combining them.
}
Any function with signature \code{function(frames, fps)} can be used.}

\item{start_pause, end_pause}{Number of frames to hold at the beginning
and end of the animation. Default is 0.}

\item{rewind}{Logical. If \code{TRUE}, the animation plays in reverse
after completing, creating a seamless loop. Default is \code{FALSE}.}

\item{cores}{Number of CPU cores for parallel frame rendering. Default
is 1 (sequential). Values greater than 1 use \code{\link[parallel:parLapply]{parallel::parLapply()}}
to render frames in parallel, which can significantly speed up
animation of complex plots. Note that progress reporting is not
available during parallel rendering.}

\item{device}{The graphics device to use for rendering frames.
Default is \code{"png"}. Other options include \code{"ragg_png"}
(requires the \pkg{ragg} package).}

\item{progress}{Whether to print a progress bar. Ignored when \code{cores > 1}.}
}
\value{
The return value of the renderer function. For
\code{gifski_renderer_3d()}, a \code{gif_3d} object (a file path with
class attributes for display in RStudio and knitr). For
\code{av_renderer_3d()}, a \code{video_3d} object (a file path with
class attributes for display). For \code{file_renderer_3d()}, a
character vector of file paths to the rendered frame images.
}
\description{
Renders an animated GIF or MP4 of a rotating ggplot with \code{\link[=coord_3d]{coord_3d()}},
smoothly interpolating rotation angles across frames.
}
\examples{
\dontshow{if (requireNamespace("gifski", quietly = TRUE)) withAutoprint(\{ # examplesIf}
\donttest{
# Plot to animate. Note `anchor = "camera"` keeps light source from rotating.
p <- ggplot() +
  geom_function_3d(
    fun = function(x, y) sin(x) * cos(y),
    xlim = c(-pi, pi), ylim = c(-pi, pi),
    fill = "steelblue", color = "steelblue", linewidth = .5) +
  coord_3d(light = light(mode = "hsl", anchor = "camera",
                         direction = c(1, 1, 0))) +
  theme_void()

# Simple turntable rotation
# (using small `nframes` to minimize runtime)
animate_3d(p, yaw = c(-30, 330), nframes = 12)

# Multi-segment orbit with pitch change
animate_3d(p, yaw = c(0, 720), roll = c(-90, 0, -90),
           nframes = 12, fps = 15)

# Save to file (writes to a temporary path; specify your own to keep it)
anim <- animate_3d(p, yaw = c(0, 360), nframes = 12)
anim_save_3d(anim, file.path(tempdir(), "rotating_surface.gif"))
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=anim_save_3d]{anim_save_3d()}} for saving animations to file.
\code{\link[=gifski_renderer_3d]{gifski_renderer_3d()}} and \code{\link[=file_renderer_3d]{file_renderer_3d()}} for renderer options.
\code{\link[=orbit_3d]{orbit_3d()}} for interactive drag-to-rotate ggcubes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-text-3d.R
\name{geom_text_3d}
\alias{geom_text_3d}
\alias{stat_text_3d}
\title{3D text labels}
\usage{
geom_text_3d(
  mapping = NULL,
  data = NULL,
  position = "identity",
  ...,
  method = c("billboard", "polygon"),
  facing = "zmax",
  coord = NULL,
  aspect_adjust = 1,
  light = "none",
  angle = 0,
  nudge_x = 0,
  nudge_y = 0,
  nudge_z = 0,
  size = 3.88,
  hjust = 0.5,
  vjust = 0.5,
  lineheight = 1.2,
  scale_depth = TRUE,
  family = "sans",
  weight = "normal",
  italic = FALSE,
  fontface = "plain",
  spacing = 0,
  tolerance = 0.01,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_text_3d(
  mapping = NULL,
  data = NULL,
  geom = NULL,
  position = "identity",
  ...,
  method = c("billboard", "polygon"),
  facing = "zmax",
  coord = NULL,
  aspect_adjust = 1,
  light = "none",
  angle = 0,
  nudge_x = 0,
  nudge_y = 0,
  nudge_z = 0,
  size = 3.88,
  hjust = 0.5,
  vjust = 0.5,
  lineheight = 1.2,
  scale_depth = TRUE,
  family = "sans",
  weight = "normal",
  italic = FALSE,
  fontface = "plain",
  spacing = 0,
  tolerance = 0.01,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings. See \code{\link[ggplot2:aes]{ggplot2::aes()}}.}

\item{data}{A data frame containing the variables to plot.}

\item{position}{Position adjustment. Defaults to \code{"identity"}.}

\item{...}{Other arguments passed to the layer.}

\item{method}{Rendering method: \code{"billboard"} (default) for flat camera-facing
text, or \code{"polygon"} for 3D-oriented filled text.}

\item{facing}{(Polygon method only) Direction the text should face. One of
\code{"xmin"}, \code{"xmax"}, \code{"ymin"}, \code{"ymax"}, \code{"zmin"}, \code{"zmax"} (default),
\code{"camera"}, a numeric vector of length 3, or a \code{camera_facing()} specification.
Use \code{"camera"} with the \code{coord} parameter to automatically face the camera.
Ignored with a warning for billboard method.}

\item{coord}{(Polygon method only) A \code{coord_3d()} object. When provided:
(1) if \code{facing = "camera"}, extracts rotation parameters automatically;
(2) uses the coord's \code{ratio} and \code{scales} parameters to correct text aspect
ratio so text doesn't appear stretched. Ignored for billboard method.}

\item{aspect_adjust}{(Polygon method only) Scalar multiplier for text height
relative to width. Default is 1 (normal proportions). Values > 1 make text
taller, values < 1 make text wider. When \code{coord} is provided, this is
multiplied by the computed aspect correction. Ignored for billboard method.}

\item{light}{(Polygon method only) Either a lighting specification created by
\code{\link[=light]{light()}}, \code{"none"} (the default) or \code{light("none")} for no lighting, or
\code{NULL} to inherit plot-level lighting. When provided, text polygons are
shaded based on their facing direction. Ignored for billboard method.}

\item{angle}{Rotation angle in degrees. For polygon method, rotates around
the facing axis. For billboard method, rotates in the view plane.}

\item{nudge_x, nudge_y, nudge_z}{Offset to apply to text position in data units.}

\item{size}{Font size in mm (same units as \code{\link[ggplot2:geom_text]{ggplot2::geom_text()}}). Default
is 3.88, which corresponds to approximately 11pt text.}

\item{hjust, vjust}{Horizontal and vertical justification of text (0-1).}

\item{lineheight}{Line height multiplier for multi-line text.}

\item{scale_depth}{Controls depth-based scaling, drawing closer elements
larger or thicker and farther ones smaller or thinner. \code{TRUE} (the default)
applies full scaling and \code{FALSE} disables it; a number sets the strength
directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below 1
subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for panel and axis components.}

\item{family}{Font family.}

\item{weight}{(Polygon method only) Font weight: "normal", "bold", "thin",
"light", etc.}

\item{italic}{(Polygon method only) Logical; use italic font variant.}

\item{fontface}{Font face (for billboard method): "plain", "bold", "italic",
or "bold.italic".}

\item{spacing}{(Polygon method only) Additional letter spacing in em units.}

\item{tolerance}{(Polygon method only) Bezier curve tolerance for text
outlines. Lower values give smoother curves but more vertices.}

\item{na.rm}{If TRUE, silently remove missing values.}

\item{show.legend}{Logical. Should this layer be included in the legends?}

\item{inherit.aes}{If TRUE, inherit aesthetics from the plot's default.}

\item{geom}{The geometric object to use. Defaults based on method.}
}
\value{
A ggplot2 layer
}
\description{
\code{geom_text_3d()} renders text labels in 3D space. Two rendering methods are
available: \code{"billboard"} (default) renders text as flat labels that always
face the camera, while \code{"polygon"} renders text as filled polygons that can
be oriented in any direction.
}
\section{Methods}{

\itemize{
\item \code{"billboard"} (default): Fast, simple text rendering using native
fonts. Text is always parallel to the view plane (like a billboard).
Only uses the \code{colour} aesthetic for text color.
\item \code{"polygon"}: Renders text as filled polygon outlines. Text can be
oriented to face any direction using the \code{facing} parameter. Uses both
\code{fill} (text fill) and \code{colour} (text outline) aesthetics.
}
}

\section{Orientation (polygon method only)}{

The \code{facing} parameter controls which direction the text faces:
\itemize{
\item \code{"zmax"} (default): Text faces upward (readable from above)
\item \code{"xmin"}, \code{"xmax"}, \code{"ymin"}, \code{"ymax"}, \code{"zmin"}: Text faces
toward the specified cube face
\item A numeric vector \code{c(x, y, z)}: Text faces the specified direction
\item \code{camera_facing(...)}: Text faces the camera
}

The \code{angle} parameter rotates the text around its facing axis (in degrees).
With \code{angle = 0}, the text baseline is oriented as horizontally as possible
(parallel to the x-y plane when the facing direction allows).
}

\section{Camera-facing text}{

For billboard method, text automatically faces the camera - no configuration
needed.

For polygon method, there are two ways to make text face the camera:
\enumerate{
\item Use \code{facing = "camera"} with the \code{coord} parameter (recommended):
}

\if{html}{\out{<div class="sourceCode">}}\preformatted{my_coord <- coord_3d(pitch = 20, roll = -60, yaw = -30)
ggplot(df, aes(x, y, z = z, label = label)) +
  geom_text_3d(method = "polygon", facing = "camera", coord = my_coord) +
  my_coord
}\if{html}{\out{</div>}}
\enumerate{
\item Pass \code{camera_facing()} directly to \code{facing}:
}

\if{html}{\out{<div class="sourceCode">}}\preformatted{ggplot(df, aes(x, y, z = z, label = label)) +
  geom_text_3d(method = "polygon",
               facing = camera_facing(pitch = 20, roll = -60, yaw = -30)) +
  coord_3d(pitch = 20, roll = -60, yaw = -30)
}\if{html}{\out{</div>}}
}

\section{Aspect Ratio Correction}{

For polygon method, text can appear stretched if the x, y, and z axes have
different data ranges. Providing a \code{coord_3d()} object via the \code{coord}
parameter enables automatic aspect ratio correction based on the coord's
\code{ratio} and \code{scales} settings.

If other layers in your plot have different data ranges that affect the
final scale limits, you can fine-tune the correction with \code{aspect_adjust}:
\itemize{
\item \code{aspect_adjust = 1} (default): Use the computed correction
\item \code{aspect_adjust > 1}: Make text taller
\item \code{aspect_adjust < 1}: Make text wider
}
}

\section{Size and Scaling}{

Text size is specified in mm via the \code{size} parameter, matching the units
used by \code{\link[ggplot2:geom_text]{ggplot2::geom_text()}}. The default size of 3.88 corresponds to
approximately 11pt text.

Both methods use the same size units, so switching between \code{"billboard"}
and \code{"polygon"} methods with the same \code{size} value will produce similarly
sized text (assuming a typical 6 inch plot).

When depth scaling is enabled (the default), text farther from the viewer
appears smaller, creating a natural perspective effect. For billboard
method, this scales the font size. For polygon method, this scales the
outline linewidth (the polygon fill is always depth-scaled by the 3D
projection).
}

\section{Aesthetics}{

\code{geom_text_3d()} understands the following aesthetics (required in bold):

\strong{Both methods:}
\itemize{
\item \strong{x}, \strong{y}, \strong{z}, \strong{label}
\item \code{alpha}
}

\strong{Billboard method only:}
\itemize{
\item \code{colour} - text color
}

\strong{Polygon method only:}
\itemize{
\item \code{fill} - text fill color
\item \code{colour} - text outline color
\item \code{linewidth} - outline thickness
\item \code{linetype} - outline line type
\item \code{group}
}

Other text properties (\code{size}, \code{hjust}, \code{vjust}, \code{angle}, \code{family},
\code{fontface}, \code{lineheight}, and for the polygon method \code{weight}, \code{italic},
\code{spacing}, \code{tolerance}) are currently accepted only as fixed values passed
as arguments to \code{geom_text_3d()}, not as mappable aesthetics in \code{aes()}.
}

\examples{
df <- expand.grid(x = c("B", "H", "N"), y = c("a", "o", "u"), z = c("g", "t"))
df$label <- paste0(df$x, df$y, df$z)

# Billboard text (default) - automatically faces camera
ggplot(df, aes(x, y, z, label = label)) +
  geom_text_3d() +
  coord_3d(scales = "fixed")

# Polygon text - can face any direction
ggplot(df, aes(x, y, z, label = label, fill = z)) +
  geom_text_3d(method = "polygon", facing = "zmax") +
  coord_3d(scales = "fixed")

# Polygon text facing camera (using coord parameter)
my_coord <- coord_3d(pitch = 0, roll = -60, yaw = -30, scales = "fixed")
ggplot(df, aes(x, y, z, label = label)) +
  geom_text_3d(method = "polygon", facing = "camera", coord = my_coord,
               color = "red", linewidth = .1) +
  my_coord

# Text with styling
ggplot(df, aes(x, y, z, label = label, fill = factor(x))) +
  geom_text_3d(method = "polygon", facing = "xmin",
               family = "serif", weight = "bold", italic = TRUE,
               size = 6, aspect_adjust = 2) +
  coord_3d(scales = "fixed")

}
\seealso{
\code{\link[=camera_facing]{camera_facing()}} for camera-facing specifications, \code{\link[=text_outlines]{text_outlines()}}
for text-to-polygon conversion
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{ridgeline_params}
\alias{ridgeline_params}
\title{Shared ridgeline params}
\arguments{
\item{direction}{Direction of ridges:
\describe{
\item{"x"}{One ridge per unique x value; ridge varies in y (default)}
\item{"y"}{One ridge per unique y value; ridge varies in x}
}}

\item{base}{Z-value for ridge polygon bottoms. If NULL, uses min(z).}
}
\description{
Shared ridgeline params
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{light_param}
\alias{light_param}
\title{Light param}
\arguments{
\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}
}
\description{
Light param
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme.R
\name{cube_theming}
\alias{cube_theming}
\title{Using themes to style 3D panels and axis labels}
\description{
In ggcube, standard ggplot2 themes generally influence 3D plots as expected, including adding
complete themes like \code{ggplot2::theme_dark()} and modifying theme elements like
\verb{theme(panel.background = element_rect(fill = "darkblue")}. However, ggcube also
provides additional theme elements that control 3D-specific styling of panels
and labels.
}
\section{Text elements}{

\itemize{
\item \code{axis.text.z}: Styling for z-axis tick labels (inherits from \code{axis.text})
\item \code{axis.title.z}: Styling for z-axis title (inherits from \code{axis.title})
\item \code{axis.text}, \code{axis.title}: Standard styling with \code{element_text()}.
}

Use \code{element_text(margin = margin(...))} to adjust text padding, with left/right
margins affecting axis text and top/bottom margins affecting axis titles;
since placement and justification of these elements varies dynamically,
no distinction is made between left and right margins, or between top and
bottom margins -- you can set either, and the maximum of the two will be used.
}

\section{Axis ticks}{

\itemize{
\item \code{axis.ticks.z}: Styling for z-axis ticks (inherits from \code{axis.ticks})
\item \code{axis.ticks}, \code{axis.ticks.x}, \code{axis.ticks.y}: Standard styling with \code{element_line()}.
\item \code{axis.ticks.length}, \code{axis.ticks.length.x/y/z}: Tick length.
}

Ticks are 3D geometry rather than screen-space annotation: each one continues
its gridline past the cube edge, and is projected like everything else. They
therefore foreshorten with the view, and a tick pointing away from the viewer
is drawn shorter than one lying across it.

Because of this, \code{axis.ticks.length} is an upper bound rather than an exact
length. It is calibrated so that a tick along the least foreshortened axis
renders at the requested size, with ticks along the other axes coming in
shorter. Tick length does not vary with axis \code{ratio}; a stretched axis gets
the same ticks as a compressed one.

As in 2D ggplot2, \code{axis.ticks.length} also contributes to the spacing between
the cube and the axis text, whether or not ticks are drawn. Setting
\code{axis.ticks = element_blank()} removes the ticks but leaves that spacing
intact; set \code{axis.ticks.length = unit(0, "pt")} to close the gap as well.
Negative lengths point ticks inward, and do not displace the axis text.
}

\section{Panel elements}{

\itemize{
\item \code{panel.foreground}: Styling for cube faces rendered in front of data (inherits from \code{panel.background}).
Foreground panels default to 20\% opacity to keep them from obscuring the data; this can be
overridden by setting \code{alpha} on either \code{panel.foreground} or \code{panel.background}.
\item \code{panel.border.foreground}: Styling for cube faces rendered in front of data (inherits from \code{panel.border})
\item \code{panel.grid.foreground}: Styling for grid lines on foreground faces (inherits from \code{panel.grid})
\item \code{panel.grid.major.foreground}: Major grid lines on foreground faces (inherits from \code{panel.grid.foreground})
}

Background panels use standard \code{panel.background}, \code{panel.border}, \code{panel.grid}, etc., while foreground panels
use the \verb{*.foreground} variants listed above. Since the foreground elements inherit from the standard background
and grid elements, you can use \code{panel.background}, etc. to style both background and foreground faces simultaneously.
This also extends to \code{alpha} set via the enhanced \code{element_rect()} below: setting \code{alpha} on \code{panel.background}
will tint both background and foreground panels (overriding the foreground's 20\% default), unless an explicit
\code{alpha} is also set on \code{panel.foreground}.
}

\section{Depth scaling}{

Under perspective projection, gridlines, borders, ticks, and axis text are (by default)
sized by their distance from the viewer. An element at the center of the cube
renders at exactly the size the theme specifies, with nearer elements drawn
larger and farther ones smaller. Axis titles are never scaled this way, and
orthographic plots are unaffected. You can use \code{coord_3d(scale_depth = )} to reduce or
disable the depth-scaling effect, per element or globally.
}

\section{Enhanced elements}{

\itemize{
\item \code{element_rect()} extends \code{ggplot2::element_rect()} by adding an \code{alpha} parameter for transparency effects
on \code{coord_3d()}'s foreground and background panels.
}
}

\examples{
# example code
p <- ggplot(sphere_points, aes(x, y, z)) +
  geom_hull_3d() +
  coord_3d(panels = "all") +
  theme(panel.background = element_rect(color = "black"),
          panel.border = element_rect(color = "black"),
          panel.foreground = element_rect(alpha = .3),
          panel.grid.foreground = element_line(color = "gray", linewidth = .25),
          axis.text = element_text(color = "darkblue"),
          axis.text.z = element_text(color = "darkred"),
          axis.title = element_text(margin = margin(t = 30)), # add padding
          axis.title.x = element_text(color = "magenta"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-identity-3d.R
\name{stat_identity_3d}
\alias{stat_identity_3d}
\title{3D-aware identity transformation}
\usage{
stat_identity_3d(
  mapping = NULL,
  data = NULL,
  geom = "point",
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  light = NULL,
  ...
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}.}

\item{data}{The data to be displayed in this layer.}

\item{geom}{The geometric object to use display the data.}

\item{position}{Position adjustment, defaults to "identity".}

\item{na.rm}{If \code{FALSE}, missing values are removed with a warning.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{...}{Other arguments passed on to geom layer.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
This stat performs identity transformation (passes data through unchanged) while
properly handling discrete scales and lighting specifications for 3D coordinate
systems. It also converts group values to hierarchical format to prevent
reordering withing groups during depth sorting.
}
\details{
This stat is primarily intended for use with 3D geoms that need discrete scale
or lighting support, following the same pattern as other ggcube stats.
}
\section{Computed variables}{

\itemize{
\item \code{x_raw}, \code{y_raw}, \code{z_raw}: Original values before discrete-to-numeric conversion
\item \code{group}: Converted to hierarchical format (e.g., "1__group", "2__group") for proper depth sorting.
}
}

\seealso{
\code{\link[=geom_point_3d]{geom_point_3d()}}, \code{\link[=geom_polygon_3d]{geom_polygon_3d()}} which use this stat for discrete scale support.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-path-3d.R
\name{geom_path_3d}
\alias{geom_path_3d}
\alias{stat_path_3d}
\title{3D paths connecting observations}
\usage{
geom_path_3d(
  mapping = NULL,
  data = NULL,
  stat = StatPath3D,
  position = "identity",
  ...,
  sort_method = "painter",
  scale_depth = TRUE,
  arrow = NULL,
  lineend = "butt",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_path_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomSegment3D,
  position = "identity",
  ...,
  sort_method = "painter",
  scale_depth = TRUE,
  arrow = NULL,
  lineend = "butt",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. Requires x, y, z
coordinates. Grouping aesthetics determine separate paths.}

\item{data}{The data to be displayed in this layer.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{"path_3d"}.}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed on to \code{\link[ggplot2:layer]{ggplot2::layer()}}.}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls depth-based scaling, drawing closer elements
larger or thicker and farther ones smaller or thinner. \code{TRUE} (the default)
applies full scaling and \code{FALSE} disables it; a number sets the strength
directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below 1
subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for panel and axis components.}

\item{arrow}{Specification for arrow heads, created by \code{\link[=arrow]{arrow()}}.}

\item{lineend}{Line end style, one of "round", "butt", "square".}

\item{na.rm}{If \code{FALSE}, missing values are removed with a warning.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{The geometric object used to display the data. Defaults to \code{"segment_3d"}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Connects observations in 3D space in the order they appear
in the data. It converts path data into individual segments for proper depth
sorting while maintaining the appearance of connected paths. Each path is
divided into segments that can be depth-sorted independently.
}
\section{Aesthetics}{

\code{geom_path_3d()} understands the following aesthetics:
\itemize{
\item \strong{x, y, z}: Coordinates (required)
\item \code{group}: Grouping variable to create separate paths
\item \code{colour}: Line color
\item \code{linewidth}: Line width (gets depth-scaled when \code{scale_depth = TRUE})
\item \code{linetype}: Line type
\item \code{alpha}: Transparency
}
}

\section{Computed variables for StatSegment3D}{

\itemize{
\item \code{x}, \code{y}, \code{z}: Start coordinates of each segment
\item \code{xend}, \code{yend}, \code{zend}: End coordinates of each segment
\item \code{group}: Hierarchical group identifier preserving original grouping
}
}

\section{Grouping}{

Multiple paths are created based on grouping aesthetics (group, colour, etc.).
Each group forms a separate path, and segments from different paths can be
interleaved during depth sorting for proper 3D rendering.
}

\examples{
library(ggplot2)

x <- seq(0, 20*pi, pi/16)
spiral <- data.frame(
  x = x,
  y = sin(x),
  z = cos(x))

# Basic path
ggplot(spiral, aes(x, y, z)) +
  geom_path_3d() +
  coord_3d()

# With aesthetic coloring
ggplot(spiral, aes(x, y, z, color = y)) +
  geom_path_3d(linewidth = 1, lineend = "round") +
  coord_3d() +
  scale_color_gradientn(colors = c("red", "purple", "blue"))

# With grouping
ggplot(spiral, aes(x, y, z, color = x > 30)) +
  geom_path_3d(linewidth = 1, lineend = "round") +
  coord_3d()

}
\seealso{
\code{\link[=geom_segment_3d]{geom_segment_3d()}} for individual segments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{mountain}
\alias{mountain}
\title{Volcano terrain data}
\format{
\subsection{\code{mountain}}{

A data frame with 1290 rows and 3 columns (x, y, and z)
}
}
\source{
derived from the \code{datasets::volcano} data set
}
\usage{
mountain
}
\description{
A terrain data set representing elevational variation over
an x-y grid.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale-z.R
\name{scale_z_continuous}
\alias{scale_z_continuous}
\title{Continuous z-axis scale for 3D plots}
\usage{
scale_z_continuous(
  name = waiver(),
  breaks = waiver(),
  minor_breaks = waiver(),
  n.breaks = NULL,
  labels = waiver(),
  limits = NULL,
  expand = waiver(),
  oob = scales::censor,
  na.value = NA_real_,
  transform = "identity",
  guide = "none",
  ...
)
}
\arguments{
\item{name}{The name of the scale, used as the axis title. Use \code{waiver()} for
the default, or \code{NULL} to omit the title.}

\item{breaks}{One of:
\itemize{
\item \code{NULL} for no breaks
\item \code{waiver()} for the default breaks
\item A numeric vector of positions
\item A function that takes the limits as input and returns breaks as output
}}

\item{minor_breaks}{One of:
\itemize{
\item \code{NULL} for no minor breaks
\item \code{waiver()} for the default minor breaks
\item A numeric vector of positions
\item A function that takes the limits as input and returns minor breaks as output
}}

\item{n.breaks}{An integer guiding the number of major breaks. The algorithm
may choose a slightly different number to ensure nice break labels.}

\item{labels}{One of:
\itemize{
\item \code{NULL} for no labels
\item \code{waiver()} for the default labels
\item A character vector giving labels (must be same length as breaks)
\item A function that takes the breaks as input and returns labels as output
}}

\item{limits}{A numeric vector of length two providing limits of the scale.
Use \code{NA} to refer to the existing minimum or maximum.}

\item{expand}{For position scales, a vector of range expansion constants used to add some
padding around the data to ensure that they are placed some distance away from the axes.}

\item{oob}{One of:
\itemize{
\item Function that handles limits outside the scale limits (out of bounds).
\item \code{scales::censor} for replacing out of bounds values with \code{NA}
\item \code{scales::squish} for squishing out of bounds values into range
}}

\item{na.value}{Missing values will be replaced with this value.}

\item{transform}{The name of a transformation object or the object itself. Default is "identity",
but works with standard transform options such as "log10", "sqrt", and "reverse",
detailed in the documentation for \code{\link[ggplot2:scale_x_continuous]{ggplot2::scale_x_continuous()}}.}

\item{guide}{A function used to create a guide or its name. Since z-axis guides are not
yet supported, this defaults to \code{"none"}.}

\item{...}{Other arguments passed on to \code{continuous_scale()}.}
}
\value{
A ggplot2 scale object for the z aesthetic.
}
\description{
\code{scale_z_continuous} creates a continuous scale for the z aesthetic in 3D plots.
It works similarly to \code{scale_x_continuous} and \code{scale_y_continuous}, providing
control over axis breaks, labels, limits, and transformations for the z dimension.
}
\examples{
# Custom breaks, labels, and limits
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
  geom_point() +
  scale_z_continuous(
    breaks = c(15, 17, 19, 21),
    labels = c("Fast", "Medium", "Slow", "Very Slow"),
    limits = c(10, NA)) +
  coord_3d()

# Works with standard scale transformations like "reverse", "log10", etc.
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
  geom_point() +
  scale_z_continuous(transform = "reverse") +
  coord_3d()

}
\seealso{
\code{\link{zlim}} for a shorthand way to set z-axis limits,
\code{\link{coord_3d}} for the 3D coordinate system

Other 3D scale functions:
\code{\link[=scale_z_discrete]{scale_z_discrete()}},
\code{\link[=zlim]{zlim()}}
}
\concept{3D scale functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-smooth-3d.R
\name{geom_smooth_3d}
\alias{geom_smooth_3d}
\alias{stat_smooth_3d}
\title{3D surface from smoothed conditional means}
\usage{
geom_smooth_3d(
  mapping = NULL,
  data = NULL,
  stat = StatSmooth3D,
  position = "identity",
  ...,
  method = "loess",
  formula = NULL,
  method.args = list(),
  xlim = NULL,
  ylim = NULL,
  n = NULL,
  grid = NULL,
  direction = NULL,
  trim = NULL,
  domain = c("bbox", "chull"),
  se = FALSE,
  level = 0.95,
  se_fill = NULL,
  se_colour = NULL,
  se_color = NULL,
  se_alpha = 0.5,
  se_linewidth = NULL,
  points = FALSE,
  point_colour = "black",
  point_color = NULL,
  point_fill = NA,
  point_size = 1.5,
  point_shape = 19,
  point_alpha = 1,
  point_stroke = 0.5,
  residuals = FALSE,
  residual_colour = "black",
  residual_color = NULL,
  residual_linewidth = 0.5,
  residual_linetype = 1,
  residual_alpha = 1,
  light = NULL,
  cull_backfaces = FALSE,
  sort_method = NULL,
  force_convex = TRUE,
  scale_depth = TRUE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_smooth_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomPolygon3D,
  position = "identity",
  ...,
  method = "loess",
  formula = NULL,
  method.args = list(),
  xlim = NULL,
  ylim = NULL,
  n = NULL,
  grid = NULL,
  direction = NULL,
  trim = NULL,
  domain = c("bbox", "chull"),
  se = FALSE,
  level = 0.95,
  se_fill = NULL,
  se_colour = NULL,
  se_color = NULL,
  se_alpha = 0.5,
  se_linewidth = NULL,
  points = FALSE,
  point_colour = "black",
  point_color = NULL,
  point_fill = NA,
  point_size = 1.5,
  point_shape = 19,
  point_alpha = 1,
  point_stroke = 0.5,
  residuals = FALSE,
  residual_colour = "black",
  residual_color = NULL,
  residual_linewidth = 0.5,
  residual_linetype = 1,
  residual_alpha = 1,
  light = NULL,
  cull_backfaces = FALSE,
  sort_method = NULL,
  force_convex = TRUE,
  scale_depth = TRUE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. This stat
requires \code{x}, \code{y}, and \code{z} aesthetics from the input data. By default, fill is
mapped to \code{after_stat(fitted)}.}

\item{data}{The data to be displayed in this layer. Must contain x, y, z columns.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{StatSmooth3D}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}

\item{method}{Smoothing method to use. Currently supported:
\itemize{
\item \code{"loess"} (default): Local polynomial regression
\item \code{"lm"}: Linear model
\item \code{"glm"}: Generalized linear model
\item \code{"gam"}: Generalized additive model (requires \code{mgcv} package)
}}

\item{formula}{Model formula. If \code{NULL} (default), uses method-appropriate defaults:
\code{z ~ x + y} for lm and glm, \code{z ~ s(x) + s(y)} for gam, auto for loess.}

\item{method.args}{List of additional arguments passed to the fitting function.
For loess, this might include \code{span} or \code{degree}. For lm, this might include \code{weights}.
For glm, this might include \code{family} (defaults to \code{gaussian()}). For gam, this might
include smoothing parameters or basis specifications.}

\item{xlim, ylim}{Numeric vectors of length 2 giving the range for prediction grid.
If \code{NULL} (default), uses the exact data range with no extrapolation.}

\item{grid, n, direction, trim}{Parameters determining the geometry, resolution, and
orientation of the surface grid. See \link{grid_generation} for details.}

\item{domain}{Character indicating the x-y domain over which to visualize the surface.
The default, \code{"bbox"}, shows predictions over the full rectangular bounding box of
the predictors.The alternative, \code{"chull"}, shows predictions only within the convex
hull of the input data, which prevents extrapolation into unoccupied corners of predictor space.}

\item{se}{Logical indicating whether to display confidence interval bands around
the smooth; if \code{TRUE}, these are rendered as additional surfaces; they inherit
aesthetics from the primary smooth layer unless otherwise specified.
Defaults to \code{FALSE}.}

\item{level}{Level of confidence interval to use (0.95 by default).}

\item{se_fill}{Fill colour for confidence interval bands. If \code{NULL}, inherits from
the main surface \code{fill} aesthetic.}

\item{se_colour, se_color}{Color for confidence interval band borders. If \code{NULL},
inherits from the main surface \code{color} aesthetic.}

\item{se_alpha}{Alpha transparency for confidence interval bands. Defaults to 0.5.}

\item{se_linewidth}{Line width for confidence interval band borders. If \code{NULL},
inherits from the main surface \code{linewidth} aesthetic.}

\item{points}{Logical indicating whether to overlay the original data points
on the fitted surface. Points are depth-sorted together with the surface
polygons for proper 3D rendering. Point styling is controlled via the
\verb{point_*} parameters; mapped aesthetics from the layer are not inherited
by annotation points. Default is \code{FALSE}.}

\item{point_colour, point_color}{Color for data points. Defaults to \code{"black"}.}

\item{point_fill}{Fill color for data points (only relevant for shapes 21-25).
Defaults to \code{NA} (transparent).}

\item{point_size}{Size of data points.}

\item{point_shape}{Shape of data points.}

\item{point_alpha}{Alpha transparency for data points.}

\item{point_stroke}{Stroke width for data points.}

\item{residuals}{Logical indicating whether to draw residual lines connecting
data points to the fitted surface. Default is \code{FALSE}. Residual styling is
controlled via the \verb{residual_*} parameters. Note: residual lines may render
incorrectly when combined with \code{se = TRUE}, as the lines intersect the
confidence interval surfaces and cannot be split at intersection points.}

\item{residual_colour, residual_color}{Color for residual lines.}

\item{residual_linewidth}{Line width for residual lines.}

\item{residual_linetype}{Line type for residual lines.}

\item{residual_alpha}{Alpha transparency for residual lines.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{The geometric object used to display the data. Defaults to \code{GeomPolygon3D}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
A 3D version of \code{ggplot2::geom_smooth()}.
Creates surfaces by fitting smoothing models to scattered (x,y,z) data points.
The fitted statistical model is evaluated on a regular grid and rendered as a
3D surface with optional standard error surfaces.
}
\section{Aesthetics}{

\code{stat_smooth_3d()} requires the following aesthetics from input data:
\itemize{
\item \strong{x}: X coordinate
\item \strong{y}: Y coordinate
\item \strong{z}: Z coordinate (response variable to be smoothed)
}
}

\section{Computed variables specific to StatSmooth3D}{

\itemize{
\item \code{level}: Type of surface ("fitted", "upper CI", or "lower CI" for confidence bands)
\item \code{fitted}: Smoothed predictions (same as \code{z} when \code{level == "fitted"})
\item \code{se}: Standard errors of the fitted values (available when \code{se = TRUE})
}
}

\section{Computed variables}{

The following computed variables are available via \code{after_stat()}:
\itemize{
\item \code{x}, \code{y}, \code{z}: Grid coordinates and function values
\item \code{normal_x}, \code{normal_y}, \code{normal_z}: Surface normal components
\item \code{slope}: Gradient magnitude from surface calculations
\item \code{aspect}: Direction of steepest slope from surface calculations
\item \code{dzdx}, \code{dzdy}: Partial derivatives from surface calculation
}
}

\examples{
# Generate scattered 3D data
set.seed(123)
d <- data.frame(
  x = runif(100, -1, 3),
  y = runif(100, -3, 3)
)
d$z <- abs(1 + d$x^2 - d$y^2 + rnorm(100, 0, 1))

# Base plot
p <- ggplot(d, aes(x, y, z)) +
  coord_3d(light = NULL) +
  scale_fill_viridis_c()

# Basic smooth surface with default loess model
p + geom_smooth_3d()

\donttest{

# Show data points
p + geom_smooth_3d(points = TRUE)

# Show data points with residual lines
p + geom_smooth_3d(points = TRUE, residuals = TRUE,
      point_color = "red", alpha = .8)

# Linear model surface with 90\% confidence intervals
p + geom_smooth_3d(aes(fill = after_stat(level)),
      method = "lm", color = "black", se = TRUE,
      level = 0.99, se_alpha = .7, n = 10) +
      scale_fill_manual(values = c("red", "darkorchid4", "steelblue"))

# Linear model surface with custom model formula
p + geom_smooth_3d(method = "lm", n = 10,
      formula = z ~ poly(x, 2) + poly(y, 2) + x:y)

# Loess with custom span parameter, and lighting effects
p + geom_smooth_3d(
      method = "loess", method.args = list(span = 0.3),
      fill = "steelblue", color = "white", n = 20,
      light = light(direction = c(0, -1, 0), color = FALSE))

# GLM with gamma family and log link
p + geom_smooth_3d(
      method = "glm", n = 10,
      method.args = list(family = Gamma(link = "log")),
      formula = z ~ poly(x, 2) + poly(y, 2))

# Visualize uncertainty with computed "standard error" variable
p + geom_smooth_3d(aes(fill = after_stat(se * 2))) +
  scale_fill_viridis_c()

# Extend surface beyond training data range (explicit extrapolation)
p + geom_smooth_3d(method = "lm", xlim = c(-5, 5), ylim = c(-5, 5))

# Clip surface to predictor convex hull
# to prevent extrapolation into corner areas
p + geom_smooth_3d(method = "lm", domain = "chull")

# Specify alternative grid geometry
p + geom_smooth_3d(grid = "right1", n = 30, direction = "y")

# Separate fits for data subgroups
ggplot(mtcars, aes(wt, mpg, qsec, fill = factor(cyl))) +
  geom_smooth_3d(method = "lm", alpha = .7,
    xlim = c(0, 5), ylim = c(0, 40)) + # specify shared domain
  coord_3d() + theme_light()

}

}
\seealso{
\code{\link[=stat_surface_3d]{stat_surface_3d()}} for surfaces from existing grid data,
\code{\link[=stat_function_3d]{stat_function_3d()}} for mathematical function surfaces.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale-z.R
\name{scale_z_discrete}
\alias{scale_z_discrete}
\title{Discrete z-axis scale for 3D plots}
\usage{
scale_z_discrete(
  name = waiver(),
  breaks = waiver(),
  labels = waiver(),
  limits = NULL,
  expand = waiver(),
  guide = "none",
  na.translate = TRUE,
  na.value = NA_real_,
  drop = TRUE,
  ...
)
}
\arguments{
\item{name}{The name of the scale, used as the axis title. Use \code{waiver()} for
the default, or \code{NULL} to omit the title.}

\item{breaks}{One of:
\itemize{
\item \code{NULL} for no breaks
\item \code{waiver()} for all breaks
\item A character vector of breaks
\item A function that takes the limits as input and returns breaks as output
}}

\item{labels}{One of:
\itemize{
\item \code{NULL} for no labels
\item \code{waiver()} for the default labels
\item A character vector giving labels (must be same length as breaks)
\item A function that takes the breaks as input and returns labels as output
}}

\item{limits}{A character vector that defines possible values of the scale and their order.}

\item{expand}{A numeric vector of length two giving multiplicative and additive expansion constants.
These constants ensure that data is placed some distance away from the axes. The defaults are
to expand by 0.6 units on each side for discrete scales.}

\item{guide}{A function used to create a guide or its name. Since z-axis guides are not
yet supported, this defaults to \code{"none"}.}

\item{na.translate}{Unlike continuous scales, discrete scales can easily show missing values,
and do so by default. If you want to remove missing values from a discrete scale, specify
\code{na.translate = FALSE}.}

\item{na.value}{If \code{na.translate = TRUE}, what aesthetic value should the missing values be
displayed as? Does not apply to position scales where \code{NA} is always placed at the far right.}

\item{drop}{Should unused factor levels be omitted from the scale? The default, \code{TRUE}, uses
the levels that appear in the data; \code{FALSE} uses all the levels in the factor.}

\item{...}{Other arguments passed on to \code{discrete_scale()}.}
}
\value{
A ggplot2 scale object for the z aesthetic.
}
\description{
\code{scale_z_discrete} creates a discrete scale for the z aesthetic in 3D plots.
It works with categorical/factor data, positioning each level at integer coordinates
(1, 2, 3, ...) in 3D space.
}
\examples{
library(ggplot2)

# Basic usage to control order, breaks, labels, expansion, etc.
ggplot(mpg, aes(displ, cty, drv, color = drv)) +
      geom_point() +
      scale_z_discrete(limits = c("f", "r", "4"), # change default order
                       breaks = c("f", "r", "4"),
                       labels = c("front", "rear", "4-wheel"),
                       expand = expansion(.5)) +
      coord_3d()

}
\seealso{
\code{\link{scale_z_continuous}} for continuous z-axis scaling,
\code{\link{coord_3d}} for the 3D coordinate system

Other 3D scale functions:
\code{\link[=scale_z_continuous]{scale_z_continuous()}},
\code{\link[=zlim]{zlim()}}
}
\concept{3D scale functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-bar-3d.R
\name{geom_bar_3d}
\alias{geom_bar_3d}
\alias{stat_bar_3d}
\title{3D bar chart with automatic counting or binning}
\usage{
geom_bar_3d(
  mapping = NULL,
  data = NULL,
  stat = StatBar3D,
  position = "identity",
  ...,
  bins = 10,
  binwidth = NULL,
  drop = TRUE,
  width = 1,
  faces = "all",
  light = NULL,
  cull_backfaces = TRUE,
  sort_method = NULL,
  scale_depth = TRUE,
  force_convex = FALSE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_bar_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomPolygon3D,
  position = "identity",
  ...,
  bins = 10,
  binwidth = NULL,
  drop = TRUE,
  width = 1,
  faces = "all",
  light = NULL,
  cull_backfaces = TRUE,
  sort_method = NULL,
  scale_depth = TRUE,
  force_convex = FALSE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}.}

\item{data}{The data to be displayed in this layer.}

\item{stat}{The statistical transformation to use. Defaults to \code{StatBar3D}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}

\item{bins}{Number of bins in each dimension. Either a single value (used for both x and y)
or a vector of length 2 giving \code{c(bins_x, bins_y)}. Default is 10. Ignored for discrete
variables.}

\item{binwidth}{Bin width in each dimension. Either a single value or a vector of length 2
giving \code{c(binwidth_x, binwidth_y)}. If provided, overrides \code{bins}. Ignored for discrete
variables.}

\item{drop}{If \code{TRUE} (the default), empty bins/combinations are not rendered. If \code{FALSE},
empty bins render as zero-height columns.}

\item{width}{Column width as a fraction of bin spacing. Either a single value (used for both
x and y) or a vector of length 2 giving \code{c(width_x, width_y)}. Default is 1.0
(columns touch). Use values less than 1 for gaps between columns.}

\item{faces}{Character vector specifying which faces to render. Options:
\itemize{
\item \code{"all"} (default): Render all 6 faces
\item \code{"none"}: Render no faces
\item Vector of face names: \code{c("zmax", "xmin", "ymax")}, etc.
}
Valid face names: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
Note that this setting acts jointly with backface culling, which removes faces whose
interior faces the viewer -- e.g., when \code{cull_backfaces = TRUE} and \code{faces = "all"}
(the default), only front faces are rendered.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{The geometric object used to display the data. Defaults to \code{GeomPolygon3D}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Creates 3D bar charts by automatically counting discrete data or binning continuous data.
This is the 3D analogue of \code{\link[ggplot2:geom_bar]{ggplot2::geom_bar()}} and \code{\link[ggplot2:geom_histogram]{ggplot2::geom_histogram()}}.
}
\details{
The stat automatically detects whether x and y are discrete or continuous:
\itemize{
\item \strong{Both discrete}: Counts occurrences of each (x, y) combination
\item \strong{Both continuous}: Performs 2D binning (like a 3D histogram)
\item \strong{Mixed}: Bins the continuous axis while preserving discrete groups
}

For pre-computed bar heights, use \code{\link[=geom_col_3d]{geom_col_3d()}} instead.
}
\section{Aesthetics}{

\code{stat_bar_3d()} requires the following aesthetics:
\itemize{
\item \strong{x}: X coordinate
\item \strong{y}: Y coordinate
}

And optionally understands:
\itemize{
\item \strong{weight}: Observation weights for counting
}
}

\section{Computed variables}{

These variables can be used with \code{after_stat()} to map to aesthetics:
\itemize{
\item \code{count}: Number of observations in each bin (default for z)
\item \code{proportion}: Count divided by total count (sums to 1)
\item \code{ncount}: Count scaled to maximum of 1
\item \code{density}: Count divided by (total count × bin area); integrates to 1 for continuous data
\item \code{ndensity}: Density scaled to maximum of 1
}
}

\examples{
# Discrete x and y: count combinations
d_discrete <- data.frame(
  x = sample(letters[1:4], 200, replace = TRUE),
  y = sample(LETTERS[1:3], 200, replace = TRUE)
)
ggplot(d_discrete, aes(x, y)) +
  geom_bar_3d() +
  coord_3d()

# Continuous x and y: 2D histogram
d_cont <- data.frame(x = rnorm(1000), y = rnorm(1000))
ggplot(d_cont, aes(x, y)) +
  geom_bar_3d() +
  coord_3d()

# Mixed: one discrete, one continuous
d_mixed <- data.frame(
  group = rep(c("A", "B", "C"), each = 100),
  value = c(rnorm(100, 2), rnorm(100, 1, 2), rnorm(100, 0))
)
ggplot(d_mixed, aes(x = group, y = value, fill = group)) +
  geom_bar_3d(bins = 20, width = c(.5, 1)) +
  coord_3d(scales = "fixed", ratio = c(1, 1, .1))

# Use density instead of count for z
ggplot(d_mixed,
      aes(x = group, y = value, z = after_stat(density))) +
  geom_bar_3d(bins = 20, width = c(.5, 1)) +
  coord_3d()

# Show empty bins with drop = FALSE
ggplot(d_cont, aes(x, y)) +
  geom_bar_3d(drop = FALSE) +
  coord_3d()

}
\seealso{
\code{\link[=geom_col_3d]{geom_col_3d()}} for pre-computed heights, \code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems,
\code{\link[=light]{light()}} for lighting specifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-ridgeline-3d.R
\name{geom_ridgeline_3d}
\alias{geom_ridgeline_3d}
\title{3D ridgeline plot from point grid}
\usage{
geom_ridgeline_3d(
  mapping = NULL,
  data = NULL,
  stat = "identity_3d",
  position = "identity",
  ...,
  direction = "x",
  base = NULL,
  cull_backfaces = FALSE,
  sort_method = "pairwise",
  scale_depth = TRUE,
  force_convex = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}.}

\item{data}{Point grid data with x, y, z coordinates.}

\item{stat}{Statistical transformation. Defaults to "identity".}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed to the layer.}

\item{direction}{Direction of ridges:
\describe{
\item{"x"}{One ridge per unique x value; ridge varies in y (default)}
\item{"y"}{One ridge per unique y value; ridge varies in x}
}}

\item{base}{Z-value for ridge polygon bottoms. If NULL, uses min(z).}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Renders a point grid as ridgeline polygons (Joy Division style). Each slice
along one axis becomes a filled polygon, creating a cross-sectional view
of the surface.
}
\section{Aesthetics}{

\describe{
\item{x, y, z}{Point coordinates (required)}
\item{fill}{Polygon fill color (default: "grey70")}
\item{colour}{Polygon border color (default: "black")}
\item{alpha}{Transparency}
\item{linewidth}{Border width}
\item{linetype}{Border type}
}
}

\examples{
# From point grid
grid_data <- expand.grid(x = seq(-3, 3, 0.5), y = seq(-3, 3, 0.1))
grid_data$z <- with(grid_data, dnorm(y) * dnorm(x) * 10)

ggplot(grid_data, aes(x, y, z = z)) +
  geom_ridgeline_3d() +
  coord_3d()

# With fill
ggplot(grid_data, aes(x, y, z = z, fill = x)) +
  geom_ridgeline_3d(colour = "white", linewidth = 0.3) +
  scale_fill_viridis_c() +
  coord_3d()

# Ridges in y direction
ggplot(grid_data, aes(x, y, z = z)) +
  geom_ridgeline_3d(direction = "y", fill = "steelblue") +
  coord_3d()

# With stat_function_3d
ggplot() +
  stat_function_3d(fun = function(x, y) sin(x) * cos(y),
                   xlim = c(-pi, pi), ylim = c(-pi, pi),
                   n = 30,
                   geom = "ridgeline_3d") +
  coord_3d()

}
\seealso{
\code{\link[=geom_surface_3d]{geom_surface_3d()}}, \code{\link[=stat_function_3d]{stat_function_3d()}}, \code{\link[=coord_3d]{coord_3d()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{polygon_params}
\alias{polygon_params}
\title{Polygon rendering parameters}
\arguments{
\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}
}
\description{
Parameters controlling depth sorting, backface culling, depth scaling, and
convexity enforcement for polygon-based 3D layers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/animate-3d.R
\name{anim_save_3d}
\alias{anim_save_3d}
\title{Save a 3D animation to a file}
\usage{
anim_save_3d(animation = NULL, filename)
}
\arguments{
\item{animation}{An animation object from \code{\link[=animate_3d]{animate_3d()}}, or \code{NULL}
to use the last rendered animation.}

\item{filename}{Output file path.}
}
\value{
Invisibly returns the output file path (a character string).
Called primarily for its side effect of copying the animation to
\code{filename}.
}
\description{
Saves an animation object produced by \code{\link[=animate_3d]{animate_3d()}} to a file.
If no animation is provided, saves the most recently rendered animation.
}
\examples{
\dontshow{if (requireNamespace("gifski", quietly = TRUE)) withAutoprint(\{ # examplesIf}
\donttest{
p <- ggplot() +
  geom_function_3d(
    fun = function(x, y) sin(x) * cos(y),
    xlim = c(-pi, pi), ylim = c(-pi, pi),
    fill = "steelblue", color = "steelblue") +
  coord_3d()

animate_3d(p, yaw = c(0, 360))
anim_save_3d(filename = file.path(tempdir(), "my_animation.gif"))
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggcube-package.R
\docType{package}
\name{ggcube}
\alias{ggcube-package}
\alias{ggcube}
\title{ggcube: Extend ggplot2 into 3D}
\description{
ggcube provides 3D geoms, stats, and coordinate systems for ggplot2.
Add \code{\link[=coord_3d]{coord_3d()}} to any plot with a z aesthetic to render it in 3D
with rotation, perspective, lighting, and depth sorting. Layer types
include surfaces, hulls, contours, ridgelines, paths, points, segments,
text, bars, and columns.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/matthewkling/ggcube}
  \item \url{https://matthewkling.github.io/ggcube/}
  \item Report bugs at \url{https://github.com/matthewkling/ggcube/issues}
}

}
\author{
\strong{Maintainer}: Matthew Kling \email{mattkling@berkeley.edu} [copyright holder]

Authors:
\itemize{
  \item Matthew Kling \email{mattkling@berkeley.edu} [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-point-3d.R
\name{geom_point_3d}
\alias{geom_point_3d}
\alias{stat_point_3d}
\title{3D scatter plot with 2D reference elements}
\usage{
geom_point_3d(
  mapping = NULL,
  data = NULL,
  stat = StatPoint3D,
  position = "identity",
  ...,
  na.rm = FALSE,
  sort_method = "painter",
  scale_depth = TRUE,
  raw_points = TRUE,
  ref_lines = FALSE,
  ref_points = FALSE,
  ref_faces = "zmin",
  ref_circle_radius = 1.5,
  ref_circle_vertices = 16,
  ref_line_color = NULL,
  ref_line_colour = NULL,
  ref_line_linewidth = 0.25,
  ref_line_linetype = NULL,
  ref_line_alpha = NULL,
  ref_point_color = NULL,
  ref_point_colour = NULL,
  ref_point_fill = NULL,
  ref_point_size = NULL,
  ref_point_alpha = NULL,
  ref_point_stroke = NULL,
  ref_point_shape = NULL,
  inherit.aes = TRUE,
  show.legend = NA
)

stat_point_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomPoint3D,
  position = "identity",
  ...,
  na.rm = FALSE,
  sort_method = "painter",
  scale_depth = TRUE,
  raw_points = TRUE,
  ref_lines = FALSE,
  ref_points = FALSE,
  ref_faces = "zmin",
  ref_circle_radius = 1.5,
  ref_circle_vertices = 16,
  ref_line_color = NULL,
  ref_line_colour = NULL,
  ref_line_linewidth = 0.25,
  ref_line_linetype = NULL,
  ref_line_alpha = NULL,
  ref_point_color = NULL,
  ref_point_colour = NULL,
  ref_point_fill = NULL,
  ref_point_size = NULL,
  ref_point_alpha = NULL,
  ref_point_stroke = NULL,
  ref_point_shape = NULL,
  inherit.aes = TRUE,
  show.legend = NA
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. In addition to
the standard point aesthetics, \code{geom_point_3d()} requires x, y, and z coordinates.}

\item{data}{The data to be displayed in this layer.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{"point_3d"}.}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed on to \code{\link[ggplot2:layer]{ggplot2::layer()}}.}

\item{na.rm}{If \code{FALSE}, missing values are removed with a warning.}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls depth-based scaling, drawing closer elements
larger or thicker and farther ones smaller or thinner. \code{TRUE} (the default)
applies full scaling and \code{FALSE} disables it; a number sets the strength
directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below 1
subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for panel and axis components.}

\item{raw_points}{Logical indicating whether to show the original 3D points.
Default is \code{TRUE}.}

\item{ref_lines}{Logical indicating whether to show reference lines projecting
from points to cube faces. Default is \code{FALSE}.}

\item{ref_points}{Type of reference points to create. Options:
\itemize{
\item \code{FALSE}: No reference points (default)
\item \code{TRUE or "circles"}: Circular reference points that project properly
\item \code{"points"}: Single-point references (renders faster and supports non-circular shapes,
but point shape is not 3D-transformed)
}}

\item{ref_faces}{Character vector specifying which cube faces to project to.
Valid face names are: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
Default is "zmin" (bottom face). Multiple faces can be specified.}

\item{ref_circle_radius}{Radius for circular reference points as percentage of standardized coordinate space.}

\item{ref_circle_vertices}{Number of vertices for circular reference points (higher = smoother).}

\item{ref_line_color, ref_line_colour, ref_line_linewidth, ref_line_linetype, ref_line_alpha}{Styling overrides for reference lines. When NULL, inherits from point aesthetics or uses defaults.
Both American (\code{ref_line_color}) and British (\code{ref_line_colour}) spellings are accepted.}

\item{ref_point_color, ref_point_colour, ref_point_fill, ref_point_size, ref_point_alpha, ref_point_stroke, ref_point_shape}{Styling overrides for reference points and circles. When NULL, inherits from raw point aesthetics
with shape-aware logic (complex shapes 21-25 vs simple shapes 0-20/characters).
Both American (\code{ref_point_color}) and British (\code{ref_point_colour}) spellings are accepted.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{geom}{The geometric object used to display the data. Defaults to \code{"point_3d"}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
\code{geom_point_3d()} creates scatter plots in 3D space with automatic depth-based
size scaling. Points closer to the viewer appear larger, while points farther
away appear smaller, creating realistic perspective effects. Optionally adds
reference lines and points projecting to cube faces, to illustrate point locations
in 2D.
}
\details{
StatPoint3D performs identity transformation (passes data through unchanged) while
properly handling discrete scales for 3D coordinate systems. It can optionally
generate reference lines and points projecting to cube faces.
}
\section{Aesthetics}{

\code{geom_point_3d()} supports all the same aesthetics as \code{\link[ggplot2:geom_point]{ggplot2::geom_point()}}, plus \code{z}:
\itemize{
\item \strong{x}: X coordinate (required)
\item \strong{y}: Y coordinate (required)
\item \strong{z}: Z coordinate (required)
\item \code{alpha}: Transparency
\item \code{colour}: Point border color
\item \code{fill}: Point fill color (for certain shapes)
\item \code{shape}: Point shape
\item \code{size}: Point size (gets depth-scaled when \code{scale_depth = TRUE})
\item \code{stroke}: Border width for shapes with borders
}
}

\section{StatPoint3D returns the following computed variables}{

\itemize{
\item \code{x_raw}, \code{y_raw}, \code{z_raw}: Original values before discrete-to-numeric conversion
\item \code{element_type}: Type of element ("raw_point", "ref_point", "ref_line", "ref_circle")
\item \code{segment_id}: ID linking the two endpoints of reference lines
\item \code{ref_face}: Which face reference elements project to
\item \code{ref_circle_radius}: Radius for circular reference points
\item \code{ref_circle_vertices}: Number of vertices for circular reference points
}
}

\section{Depth Scaling}{

The depth scaling uses an inverse relationship with distance, following the
mathematical relationship: \code{apparent_size = base_size * reference_distance / actual_distance}

This creates realistic perspective where:
\itemize{
\item Objects twice as far appear half as big
\item Objects twice as close appear twice as big
\item The center of the plot volume renders at exactly the user-specified size
}
}

\section{Point Rendering}{

ggcube uses shape-aware rendering for improved stroke behavior:
\itemize{
\item \strong{Simple shapes (0-20, characters):} Use \code{size} for fontsize, \code{stroke} for border width, no fill
\item \strong{Complex shapes (21-25):} Use ggplot2's approach: \code{size + 0.5*stroke} for fontsize to prevent gaps
\item Both size and stroke are depth-scaled when \code{scale_depth = TRUE}
\item All shapes preserve stroke depth scaling and parameter control
}
}

\section{Reference Features}{

Reference lines and points help visualize the 3D relationships by projecting
data points onto cube faces:
\itemize{
\item \strong{Reference lines}: Connect each 3D point to its projection on specified faces
\item \strong{Reference points}: Show the projected location on the faces
\item \strong{Reference circles}: Circular projections that appear as realistic 3D shadows
\item \strong{Depth sorting}: All elements (original points, reference lines, reference points)
are automatically depth-sorted for proper 3D rendering
}
}

\section{Aesthetic Inheritance}{

Reference elements intelligently inherit styling from raw points:
\itemize{
\item \strong{Shape-aware}: Complex shapes (21-25) with fill/colour are handled differently from simple shapes (0-20/characters)
\item \strong{Alpha detection}: If all points have alpha=1, uses 0.5 default for ref elements; otherwise inherits mapped alpha
\item \strong{Priority}: Explicit ref_* parameters > mapped aesthetics > smart defaults
}
}

\examples{
library(ggplot2)

# Basic 3D scatter plot with depth scaling
ggplot(expand.grid(x = 1:5, y = 1:5, z = 1:5),
       aes(x, y, z, fill = z)) +
  geom_point_3d(size = 10, shape = 21, color = "white", stroke = .1) +
  coord_3d(pitch = 40, roll = 5, yaw = 0, dist = 1.5) +
  scale_fill_viridis_c()

# Add circular reference points on 2D face panel
ggplot(mtcars, aes(mpg, wt, qsec)) +
  geom_point_3d(size = 3,
    ref_points = TRUE, ref_faces = c("ymax", "xmax")) +
  coord_3d()

}
\seealso{
\code{\link[=geom_segment_3d]{geom_segment_3d()}} for 3D segments, \code{\link[=geom_path_3d]{geom_path_3d()}} for 3D paths,
\code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guide_3d.R
\name{guide_3d}
\alias{guide_3d}
\alias{guide_colorbar_3d}
\alias{guide_legend_3d}
\title{Color guides showing lighting effects}
\usage{
guide_colorbar_3d(reverse_shade = FALSE, shade_range = c(-0.5, 0.5), ...)

guide_legend_3d(reverse_shade = FALSE, shade_range = c(-0.5, 0.5), ...)
}
\arguments{
\item{reverse_shade}{Logical. If TRUE, reverses the lighting gradient direction. By default,
shadows are placed on the left, or on the bottom for horizontal colorbars.}

\item{shade_range}{Length-2 numeric vector in the range -1 to 1, giving the limits
of the shading gradient. -1 is full shade, and 1 is full highlight. Default is \code{c(-.5, .5)}.}

\item{...}{Additional arguments passed to \code{guide_colorbar()} or \code{guide_legend()}.}
}
\value{
A guide object that displays shading effects
}
\description{
Creates color guides that show shading variation as gradients
within each color. Shows the full range of colors visible when
shading is enabled in 3D plots.
}
\details{
When fill and color aesthetics map to the same variable (e.g.,
\code{aes(fill = z, color = z)}), ggplot2 creates a shared scale with a
single guide. In this case, use \code{guides(fill = guide_*_3d())} to
apply shading, \strong{not} \code{guides(color = guide_*_3d())}, even if your layer
uses the color aesthetic. Only use the color guide when color and fill
map to different variables and you want separate guides for each.
}
\examples{
# continuous `colorbar` guide
ggplot(mountain, aes(x, y, z, fill = z)) +
   stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
   guides(fill = guide_colorbar_3d()) +
   scale_fill_gradientn(colors = c("tomato", "dodgerblue")) +
   coord_3d()

# discrete `legend` guide
ggplot(mountain, aes(x, y, z, fill = x > .5, group = 1)) +
   stat_surface_3d(light = light(mode = "hsl", direction = c(1, 0, 0))) +
   guides(fill = guide_legend_3d()) +
   coord_3d()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{surface_computed_vars}
\alias{surface_computed_vars}
\title{Computed variables for surfaces}
\description{
Computed variables for surfaces
}
\section{Computed variables}{

The following computed variables are available via \code{after_stat()}:
\itemize{
\item \code{x}, \code{y}, \code{z}: Grid coordinates and function values
\item \code{normal_x}, \code{normal_y}, \code{normal_z}: Surface normal components
\item \code{slope}: Gradient magnitude from surface calculations
\item \code{aspect}: Direction of steepest slope from surface calculations
\item \code{dzdx}, \code{dzdy}: Partial derivatives from surface calculation
}
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-zcontour-3d.R
\name{geom_contour_3d}
\alias{geom_contour_3d}
\alias{stat_contour_3d}
\title{Contours of a 3D surface}
\usage{
geom_contour_3d(
  mapping = NULL,
  data = NULL,
  stat = "surface_3d",
  position = "identity",
  ...,
  bins = 20,
  binwidth = NULL,
  breaks = NULL,
  cull_backfaces = FALSE,
  sort_method = "pairwise",
  scale_depth = TRUE,
  force_convex = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_contour_3d(
  mapping = NULL,
  data = NULL,
  geom = "contour_3d",
  position = "identity",
  ...,
  bins = 20,
  binwidth = NULL,
  breaks = NULL,
  cull_backfaces = FALSE,
  sort_method = "pairwise",
  scale_depth = TRUE,
  force_convex = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}.}

\item{data}{Point grid data with x, y, z coordinates.}

\item{stat}{Statistical transformation. Defaults to \code{"surface_3d"}.}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed to the layer.}

\item{bins}{Number of contour levels. Default is 20. Ignored if \code{breaks} or
\code{binwidth} is provided.}

\item{binwidth}{Width of each contour band. Overrides \code{bins} if provided.}

\item{breaks}{Numeric vector specifying exact contour break points. Overrides
both \code{bins} and \code{binwidth} if provided.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{Geometry function used to display the data. Defaults to \code{"surface_3d"}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Renders a surface as stacked horizontal contour bands. Each contour band is a
polygon placed at its corresponding z-level.
}
\details{
This geom takes point grid data (like that produced by \code{\link[=stat_surface_3d]{stat_surface_3d()}},
\code{\link[=stat_function_3d]{stat_function_3d()}}, \code{\link[=stat_smooth_3d]{stat_smooth_3d()}}, or \code{\link[=stat_density_3d]{stat_density_3d()}}) and converts
it to filled contour polygons using the \code{isoband} package.
}
\section{Aesthetics}{

\code{geom_contour_3d()} requires:
\describe{
\item{x, y, z}{Point coordinates forming a regular grid}
}

And understands these additional aesthetics:
\describe{
\item{fill}{Band fill color. For automatic coloring by elevation, use
\code{aes(fill = after_stat(z))}. Default is "grey60".}
\item{colour}{Band border color (default: "grey30")}
\item{alpha}{Transparency}
\item{linewidth}{Border width (default: 0.1)}
\item{linetype}{Border line type}
}
}

\section{Computed variables}{

Each contour band is placed at its corresponding z-level (the upper boundary
of the band). To color by elevation, use \code{aes(fill = after_stat(z))}.
}

\examples{
# Basic usage with volcano data
ggplot(mountain, aes(x, y, z)) +
      geom_contour_3d(color = "white", fill = "black") +
      coord_3d(light = "none", ratio = c(1.5, 2, 1))

# Map fill to elevation and customize number of levels
ggplot(mountain, aes(x, y, z, fill = after_stat(z))) +
      geom_contour_3d(bins = 12, color = "white") +
      scale_fill_viridis_c() +
      coord_3d(light = "none", ratio = c(1.5, 2, 1))

# Specify exact breaks
ggplot(mountain, aes(x, y, z, fill = after_stat(z))) +
  geom_contour_3d(breaks = seq(0, 200, by = 5)) +
  scale_fill_viridis_c() +
  coord_3d(light = "none")

# With stat_density_3d
ggplot(faithful, aes(eruptions, waiting)) +
  stat_density_3d(geom = "contour_3d",
    sort_method = "pairwise") +
  coord_3d()

# With stat_function_3d
ggplot() +
  stat_function_3d(
    fun = function(x, y) sin(x) * cos(y),
    xlim = c(-pi, pi), ylim = c(-pi, pi),
    geom = "contour_3d",
    bins = 50, color = "black"
  ) +
  scale_fill_viridis_c(option = "B") +
  coord_3d(light = "none")

}
\seealso{
\code{\link[=geom_surface_3d]{geom_surface_3d()}} for continuous surface rendering,
\code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}} for cross-sectional ridgeline rendering,
\code{\link[=stat_function_3d]{stat_function_3d()}} for mathematical function surfaces,
\code{\link[=stat_smooth_3d]{stat_smooth_3d()}} for fitted model surfaces,
\code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems,
\code{\link[ggplot2:geom_contour_filled]{ggplot2::geom_contour_filled()}} for the 2D equivalent.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{sphere_points}
\alias{sphere_points}
\title{Sphere surface points}
\format{
\subsection{\code{sphere_points}}{

A data frame with 200 rows and 3 columns (x, y, and z)
}
}
\usage{
sphere_points
}
\description{
A toy data set representing points on a sphere surface,
useful for testing hull functions
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-segment-3d.R
\name{geom_segment_3d}
\alias{geom_segment_3d}
\alias{stat_segment_3d}
\title{3D line segments}
\usage{
geom_segment_3d(
  mapping = NULL,
  data = NULL,
  stat = StatSegment3D,
  position = "identity",
  ...,
  sort_method = "painter",
  scale_depth = TRUE,
  arrow = NULL,
  lineend = "butt",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_segment_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomSegment3D,
  position = "identity",
  ...,
  sort_method = "painter",
  scale_depth = TRUE,
  arrow = NULL,
  lineend = "butt",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. Requires x, y, z
for start coordinates and xend, yend, zend for end coordinates.}

\item{data}{The data to be displayed in this layer.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{"segment_3d"}.}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed on to \code{\link[ggplot2:layer]{ggplot2::layer()}}.}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls depth-based scaling, drawing closer elements
larger or thicker and farther ones smaller or thinner. \code{TRUE} (the default)
applies full scaling and \code{FALSE} disables it; a number sets the strength
directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below 1
subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for panel and axis components.}

\item{arrow}{Specification for arrow heads, created by \code{\link[=arrow]{arrow()}}.}

\item{lineend}{Line end style, one of "round", "butt", "square".}

\item{na.rm}{If \code{FALSE}, missing values are removed with a warning.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{The geometric object used to display the data. Defaults to \code{"segment_3d"}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
\code{geom_segment_3d()} and \code{stat_segment_3d()} draw line segments in 3D space
with automatic depth-based linewidth scaling and proper depth sorting.
Each segment is defined by start coordinates (x, y, z) and end
coordinates (xend, yend, zend).
}
\section{Aesthetics}{

\code{geom_segment_3d()} understands the following aesthetics:
\itemize{
\item \strong{x, y, z}: Start coordinates (required)
\item \strong{xend, yend, zend}: End coordinates (required)
\item \code{colour}: Line color
\item \code{linewidth}: Line width (gets depth-scaled when \code{scale_depth = TRUE})
\item \code{linetype}: Line type
\item \code{alpha}: Transparency
}
}

\examples{
# Basic 3D segments
ggplot(sphere_points,
      aes(x, y, z, xend = 0, yend = 0, zend = 0)) +
  geom_segment_3d() +
  coord_3d()

# 3D vector field
data <- expand.grid(x = -1:2, y = -1:2, z = -1:2)
data2 <- data + seq(-.5, .5, length.out = length(as.matrix(data)))
data <- cbind(data, setNames(data2, c("x2", "y2", "z2")))
ggplot(data, aes(x, y, z,
      xend = x2, yend = y2, zend = z2, color = x)) +
  geom_segment_3d(arrow = arrow(length = unit(0.1, "inches"),
                  type = "closed", angle = 15),
                  linewidth = .5) +
  coord_3d()

}
\seealso{
\code{\link[=geom_path_3d]{geom_path_3d()}} for connected paths.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-text-3d.R
\name{text_outlines}
\alias{text_outlines}
\title{Convert text to polygon outlines}
\usage{
text_outlines(
  text,
  family = "sans",
  weight = "normal",
  italic = FALSE,
  size = 12,
  tolerance = 0.01,
  spacing = 0,
  hjust = 0.5,
  vjust = 0.5,
  width = NA,
  align = "left",
  lineheight = 1,
  internal_size = 144
)
}
\arguments{
\item{text}{Character string to convert}

\item{family}{Font family name (e.g., "Arial", "Helvetica", "sans")}

\item{weight}{Font weight ("normal", "bold", "thin", "light", etc.)}

\item{italic}{Logical; use italic variant}

\item{size}{Font size in points}

\item{tolerance}{Bezier curve tolerance; lower values give more detailed outlines}

\item{spacing}{Additional letter spacing in em units}

\item{hjust, vjust}{The justification of the textbox surrounding the text}

\item{width}{Maximum line width in inches for word wrapping; NA disables wrapping}

\item{align}{Text alignment: "left", "center", or "right"}

\item{lineheight}{Line height multiplier}

\item{internal_size}{Internal rendering size for kerning precision; generally
shouldn't need adjustment}
}
\value{
A data frame with columns: x, y, contour, glyph, letter, poly_id. X and y
are vertex coordinates, in point units.
}
\description{
Converts a text string into polygon vertices suitable for plotting with
ggplot2's geom_polygon. Uses systemfonts for font matching and glyph extraction.
}
\examples{
data <- text_outlines("Howdy, partner", family = "Arial")

ggplot2::ggplot(data, ggplot2::aes(x, y, group = letter, subgroup = poly_id)) +
  ggplot2::geom_polygon() +
  ggplot2::coord_fixed()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-text-3d.R
\name{camera_facing}
\alias{camera_facing}
\title{Compute camera-facing direction for 3D text}
\usage{
camera_facing(
  coord = NULL,
  pitch = 0,
  roll = -60,
  yaw = -30,
  dist = 2,
  mode = c("plane", "point")
)
}
\arguments{
\item{coord}{Optional objected created by \code{coord_3d}. If provided, coord
parameters are drawn from this and the \code{pitch}, \code{roll}, \code{yaw}, \code{dist} arguments
are ignored.}

\item{pitch, roll, yaw}{Rotation angles in degrees, matching the values
you'll use in \code{coord_3d()}.}

\item{dist}{Distance from viewer to center of the data cube, matching
the value you'll use in \code{coord_3d()}. Default is 2. Only used when
\code{mode = "point"}.}

\item{mode}{Character string specifying how labels should face the camera:
\itemize{
\item \code{"plane"} (default): All labels face parallel to the view plane,
like billboards. This is typically easier to read.
\item \code{"point"}: Each label faces toward the camera position. Labels
at the edges will angle inward, giving perspective-correct orientation.
}}
}
\value{
A \code{camera_facing_spec} object to pass to the \code{facing} parameter
of \code{geom_text_3d()}.
}
\description{
Creates a specification for camera-facing "billboard" text that will compute
the appropriate facing direction for each label position.
}
\details{
With \code{mode = "plane"}, all text labels face the same direction (parallel
to the viewing plane), similar to traditional billboard rendering. This
produces clean, readable labels.

With \code{mode = "point"}, each label faces directly toward the camera's
position in 3D space. Labels near the edges of the plot will angle inward,
which is geometrically correct but can look less clean.

Since \code{geom_text_3d()} computes vertex positions before the view is known,
you must specify the view parameters here and use the same values in
\code{coord_3d()}.
}
\examples{
\donttest{
# Parallel billboard text (default) - all labels face same direction
df <- expand.grid(x = c("H", "B"), y = c("a", "o", "u"), z = c("g", "t"))
df$label <- paste0(df$x, df$y, df$z)

ggplot(df, aes(x, y, z = z, label = label)) +
  geom_text_3d(method = "polygon",
               facing = camera_facing(pitch = 20, roll = -60, yaw = -30)) +
  coord_3d(pitch = 20, roll = -60, yaw = -30)

# Point-facing text - labels angle toward camera position
ggplot(df, aes(x, y, z = z, label = label)) +
  geom_text_3d(method = "polygon",
               facing = camera_facing(pitch = 20, roll = -60, yaw = -30,
                                      mode = "point")) +
  coord_3d(pitch = 20, roll = -60, yaw = -30)
}

}
\seealso{
\code{\link[=geom_text_3d]{geom_text_3d()}} for rendering 3D text labels
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aes.R
\name{aes}
\alias{aes}
\title{Aesthetic mapping with positional z support}
\usage{
aes(x, y, z, ...)
}
\arguments{
\item{x}{Variable to map to x aesthetic (required)}

\item{y}{Variable to map to y aesthetic (required)}

\item{z}{Variable to map to z aesthetic (optional)}

\item{...}{Other aesthetic mappings (color, size, etc.)}
}
\value{
An aesthetic mapping object, same as \code{ggplot2::aes()}
}
\description{
This function extends \code{ggplot2::aes()} to support positional mapping
to the z aesthetic. It maintains full backward compatibility with the original
\code{aes()} function while enabling the convenient \code{aes(x, y, z)} syntax
for 3D plots.
}
\details{
This function is a lightweight wrapper around \code{ggplot2::aes()} that:
\itemize{
\item Maintains full backward compatibility with existing 2D plots
\item Enables positional z mapping: \code{aes(x_var, y_var, z_var)}
\item Works with any geom that uses the z aesthetic (contour, raster, 3D plots)
\item Passes through all other aesthetics unchanged
}
}
\examples{
library(ggplot2)

# 2D plots work like regular ggplot2
ggplot(mtcars, aes(mpg, wt)) + geom_point()

# 3D plots can use positional syntax, or explicitly map to z
# same as: ggplot(mtcars, aes(mpg, wt, z = qsec)) + geom_point() + coord_3d()
ggplot(mtcars, aes(mpg, wt, qsec)) + geom_point() + coord_3d()

# Also works with non-ggcube z-aesthetic geoms
ggplot(mountain, aes(x, y, z)) + geom_contour()

}
\seealso{
\code{\link[ggplot2]{aes}} for the original aesthetic mapping function,
\code{\link{coord_3d}} for 3D coordinate systems
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-col-3d.R
\name{geom_col_3d}
\alias{geom_col_3d}
\alias{stat_col_3d}
\title{3D columns from grid data}
\usage{
geom_col_3d(
  mapping = NULL,
  data = NULL,
  stat = StatCol3D,
  position = "identity",
  ...,
  width = 1,
  faces = "all",
  zmin = NULL,
  light = NULL,
  cull_backfaces = TRUE,
  sort_method = NULL,
  scale_depth = TRUE,
  force_convex = FALSE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

stat_col_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomPolygon3D,
  position = "identity",
  ...,
  width = 1,
  faces = "all",
  zmin = NULL,
  light = NULL,
  cull_backfaces = TRUE,
  sort_method = NULL,
  scale_depth = TRUE,
  force_convex = FALSE,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}.}

\item{data}{The data to be displayed in this layer.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{StatCol3D}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}

\item{width}{Numeric value controlling box width as a fraction of grid spacing.
Default is 1.0 (volumes touch each other). Use 0.8 for small gaps, 1.2 for overlap.
Grid spacing is determined automatically using \code{ggplot2::resolution()}}

\item{faces}{Character vector specifying which faces to render. Options:
\itemize{
\item \code{"all"} (default): Render all 6 faces
\item \code{"none"}: Render no faces
\item Vector of face names: \code{c("zmax", "xmin", "ymax")}, etc.
}
Valid face names: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
Note that this setting acts jointly with backface culling, which removes faces whose
interior faces the viewer -- e.g., when \code{cull_backfaces = TRUE} and \code{faces = "all"}
(the default), only front faces are rendered.}

\item{zmin}{Base level for all columns. When provided as a parameter, overrides any
\code{zmin} aesthetic mapping. If \code{NULL} (the default), uses the \code{zmin}
aesthetic if mapped, otherwise defaults to 0.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{geom}{The geometric object used to display the data. Defaults to \code{GeomPolygon3D}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Creates 3D columns (rectangular prisms) from grid data in which x and y fall on a
regular grid. Works with both complete and sparse grid data. Each data point becomes
a rectangular 3D column extending from a base level to the data value.
}
\details{
This is analogous to \code{\link[ggplot2:geom_col]{ggplot2::geom_col()}} for 3D plots. For automatic counting or
binning, see \code{\link[=geom_bar_3d]{geom_bar_3d()}}.

Note that column geometries often require pairwise depth sorting for correct rendering.
This is the default for smaller data sets, but not for larger data sets due to compute speed;
in those cases you may wish to manually specify \code{sort_method = "pairwise"} for good results.
}
\section{Aesthetics}{

\code{stat_col_3d()} requires the following aesthetics:
\itemize{
\item \strong{x}: X coordinate (grid position)
\item \strong{y}: Y coordinate (grid position)
\item \strong{z}: Z coordinate (column top height)
}

And optionally understands:
\itemize{
\item \strong{zmin}: Base level for each column (can be overridden by the \code{zmin} parameter)
}
}

\section{Computed variables}{

\itemize{
\item \code{normal_x}, \code{normal_y}, \code{normal_z}: Face normal components
\item \code{col_id}: Sequential column number
\item \code{face_type}: Face name ("zmax", "xmin", etc.)
}
}

\examples{
# Basic 3D bar chart from regular grid
# (columns extend from z=0 by default)
d <- expand.grid(x = 1:5, y = 1:5)
d$z <- d$x + d$y + rnorm(25, 0, 0.5)
ggplot(d, aes(x, y, z)) +
  geom_col_3d() +
  coord_3d()

# Set uniform base level using `zmin` parameter
ggplot(d, aes(x, y, z)) +
  geom_col_3d(aes(fill = z), color = "white",
              zmin = 5) +
  coord_3d()

# Set variable base levels using `zmin` aesthetic
d$base_level <- runif(nrow(d), -5, 1)
ggplot(d, aes(x, y, z = z, zmin = base_level)) +
  geom_col_3d(color = "black") +
  coord_3d()

# Show only a subset of column faces
ggplot(d, aes(x, y, z)) +
  geom_col_3d(faces = c("zmax", "ymin"),
    cull_backfaces = FALSE,
    fill = "steelblue", color = "black") +
  coord_3d()

# With gaps between columns
ggplot(d, aes(x, y, z)) +
  geom_col_3d(color = "black", width = 0.6) +
  coord_3d()

}
\seealso{
\code{\link[=geom_bar_3d]{geom_bar_3d()}} for automatic counting/binning, \code{\link[=stat_surface_3d]{stat_surface_3d()}} for smooth
surface rendering, \code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems, \code{\link[=light]{light()}} for lighting
specifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/animate-3d.R
\name{renderers_3d}
\alias{renderers_3d}
\alias{gifski_renderer_3d}
\alias{av_renderer_3d}
\alias{file_renderer_3d}
\title{Animation renderers for animate_3d}
\usage{
gifski_renderer_3d(file = NULL, loop = TRUE)

av_renderer_3d(file = NULL, vfilter = "null", codec = NULL)

file_renderer_3d(dir = tempdir(), prefix = "ggcube_frame", overwrite = FALSE)
}
\arguments{
\item{file}{Output file path. If \code{NULL}, a temporary file is used.}

\item{loop}{Logical. Should the GIF loop? Default is \code{TRUE}.}

\item{vfilter}{A video filter string for ffmpeg (passed to \pkg{av}).}

\item{codec}{Video codec name. Default lets \pkg{av} choose.}

\item{dir}{Directory to copy frame files into. Defaults to a session
temporary directory (\code{tempdir()}). For animations you want to keep,
pass an explicit directory.}

\item{prefix}{Filename prefix for copied frame files.}

\item{overwrite}{Logical. Overwrite existing files? Default is \code{FALSE}.}
}
\value{
A function with signature \code{function(frames, fps, width, height)}
that produces the animation output.
}
\description{
Renderer functions that combine individual frame images into a final
animation file. These are factory functions that return the actual
rendering function.
}
\examples{
\dontshow{if (requireNamespace("gifski", quietly = TRUE) && requireNamespace("av", quietly = TRUE)) withAutoprint(\{ # examplesIf}
\donttest{
p <- ggplot() +
  geom_function_3d(
    fun = function(x, y) sin(x) * cos(y),
    xlim = c(-pi, pi), ylim = c(-pi, pi),
    fill = "steelblue", color = "steelblue") +
  coord_3d()

# GIF output (default)
animate_3d(p, yaw = c(0, 360), renderer = gifski_renderer_3d())

# Video output
animate_3d(p, yaw = c(0, 360), renderer = av_renderer_3d())

# Keep individual frame files in a chosen directory
animate_3d(p, yaw = c(0, 360),
    renderer = file_renderer_3d(dir = file.path(tempdir(), "my_frames")))
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{grid_params}
\alias{grid_params}
\title{Grid generation parameters}
\arguments{
\item{grid, n, direction, trim}{Parameters determining the geometry, resolution, and
orientation of the surface grid. See \link{grid_generation} for details.}
}
\description{
Grid generation parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme.R
\name{element_rect}
\alias{element_rect}
\title{Enhanced rectangle element with alpha support}
\usage{
element_rect(fill = NULL, alpha = NA, ...)
}
\arguments{
\item{fill}{Fill colour for the rectangle.}

\item{alpha}{Transparency level applied to the fill of foreground or
background panels in \code{coord_3d()}, ranging from 0 (completely
transparent) to 1 (completely opaque). For \code{panel.foreground}, the
alpha resolves in order: an explicit value here, then alpha inherited
from \code{panel.background}, then a default of 0.2. For
\code{panel.background}, an explicit value is used directly; otherwise
the panel is fully opaque. Has no effect on other theme elements or in
plots without a 3D coordinate system.}

\item{...}{Additional arguments passed to \code{\link[ggplot2]{element_rect}},
such as \code{colour}, \code{linewidth}, \code{linetype}, and
\code{inherit.blank}.}
}
\value{
A \code{ggplot2::element_rect} theme element with an attached
\code{"ggcube_alpha"} attribute.
}
\description{
This function extends \code{ggplot2::element_rect()} to support transparency
via an \code{alpha} parameter, which is attached to the returned element as
an R attribute (\code{"ggcube_alpha"}). The attribute survives ggplot2's
theme inheritance machinery, so the alpha value set here is available to
the renderer at draw time. The \code{alpha} parameter is consumed only by
\code{coord_3d()}'s panel rendering and has no effect on other theme
elements.
}
\examples{
# Basic 3D plot with semi-transparent foreground panels
ggplot(mountain, aes(x, y, z)) +
  stat_surface_3d(fill = "darkblue", color = "lightblue", linewidth = .1) +
  coord_3d(panels = c("background", "ymin")) +
  theme(panel.foreground = element_rect(alpha = 0.6))

# Completely transparent foreground panels
ggplot(mtcars, aes(mpg, wt, qsec)) +
  geom_point() +
  coord_3d(panels = "all") +
  theme(panel.border = element_rect(color = "black"),
        panel.foreground = element_rect(fill = "blue", alpha = 0))

}
\seealso{
\code{\link[ggplot2]{element_rect}} for the original function,
\code{\link{coord_3d}} for 3D coordinate systems that utilize foreground panels
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label-rendering.R
\name{makeContent.ggcube_axis_furniture}
\alias{makeContent.ggcube_axis_furniture}
\title{Resolve 3D axis labels and titles at draw time}
\usage{
\method{makeContent}{ggcube_axis_furniture}(x)
}
\arguments{
\item{x}{A \code{ggcube_axis_furniture} gTree.}
}
\value{
The gTree with its children populated.
}
\description{
Resolve 3D axis labels and titles at draw time
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/orbit-3d.R
\name{orbit_3d-shiny}
\alias{orbit_3d-shiny}
\alias{orbit3dOutput}
\alias{renderOrbit3d}
\title{Shiny bindings for orbit_3d}
\usage{
orbit3dOutput(outputId, width = "100\%", height = "400px")

renderOrbit3d(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{Output variable to read from.}

\item{width, height}{Must be a valid CSS unit (like \code{"100\%"}, \code{"400px"},
\code{"auto"}) or a number, which will be coerced to a string and have \code{"px"}
appended.}

\item{expr}{An expression that generates a orbit (a call to
\code{\link[=orbit_3d]{orbit_3d()}}).}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Is \code{expr} a quoted expression (with \code{quote()})? This is useful
if you want to save an expression in a variable.}
}
\description{
Output and render functions for using \code{\link[=orbit_3d]{orbit_3d()}} within Shiny
applications and interactive R Markdown documents.
}
\examples{
\dontshow{if (interactive() && requireNamespace("shiny", quietly = TRUE) && requireNamespace("base64enc", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(shiny)

p <- ggplot(mountain, aes(x, y, z, fill = z, color = z)) +
  geom_surface_3d(linewidth = .25) +
  coord_3d(light = light(anchor = "camera", direction = c(1, 1, 0)),
           ratio = c(1.5, 2, 1)) +
  scale_fill_viridis_c() + scale_color_viridis_c() +
  theme_void()

ui <- fluidPage(
  titlePanel("Interactive ggcube"),
  orbit3dOutput("fb", width = "500px", height = "500px")
)

server <- function(input, output, session) {
  output$fb <- renderOrbit3d({
    orbit_3d(p, yaw = c(360, 0), n = 12)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale-z.R
\name{zlim}
\alias{zlim}
\title{Set z-axis limits}
\usage{
zlim(min, max, ...)
}
\arguments{
\item{min, max}{The minimum and maximum values for the z-axis.}

\item{...}{Additional arguments passed to \code{scale_z_continuous()}.}
}
\value{
A ggplot2 scale object for the z aesthetic with specified limits.
}
\description{
This is a shorthand for \code{scale_z_continuous(limits = c(min, max))}.
It's a convenient way to set the z-axis limits without specifying other
scale parameters.
}
\examples{
library(ggplot2)

# Set z-axis limits
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
  geom_point() +
  zlim(15, 20) +
  coord_3d()

# Equivalent to:
ggplot(mtcars, aes(mpg, wt, z = qsec)) +
  geom_point() +
  scale_z_continuous(limits = c(15, 20)) +
  coord_3d()

}
\seealso{
\code{\link{scale_z_continuous}} for more control over z-axis scaling,
\code{\link{xlim}}, \code{\link{ylim}} for x and y axis limits

Other 3D scale functions:
\code{\link[=scale_z_continuous]{scale_z_continuous()}},
\code{\link[=scale_z_discrete]{scale_z_discrete()}}
}
\concept{3D scale functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord-3d.R
\name{coord_3d}
\alias{coord_3d}
\title{3D coordinate system}
\usage{
coord_3d(
  pitch = 0,
  roll = -60,
  yaw = -30,
  persp = TRUE,
  dist = 2,
  expand = TRUE,
  clip = "off",
  panels = "background",
  xlabels = "auto",
  ylabels = "auto",
  zlabels = "auto",
  title_position = c("auto", "center"),
  rotate_labels = TRUE,
  scales = "free",
  scale_depth = TRUE,
  ratio = c(1, 1, 1),
  zoom = 1,
  light = waiver(),
  ...
)
}
\arguments{
\item{roll, pitch, yaw}{Rotation around x, y, and z axes, respectively, in degrees.
Positive values rotate the near face of the plot "downward", "rightward", and clockwise, respectively.}

\item{persp}{Logical indicating whether to apply perspective projection. When \code{TRUE} (the default),
objects farther from the viewer appear smaller. When \code{FALSE}, produces an orthographic projection in which
lines that are parallel in 3D space render as parallel in the plot.}

\item{dist}{Distance from viewer to center of the data cube. Only used when \code{persp = TRUE}.
Larger values create less perspective distortion. Default is 2.
Values less than 1 are allowed but can be problematic for rendering.}

\item{expand}{Logical indicating whether to expand axis ranges beyond the data range,
similar to standard ggplot2 behavior. If \code{TRUE} (the default), expansion behavior
can be controlled using standard axis scaling functions, e.g.
\code{... + scale_x_continuous(expand = expansion(.5))}.}

\item{clip}{Character string indicating clipping behavior. Use \code{"off"} (the default, recommended
for some 3D plots) to allow drawing outside the plot panel.}

\item{panels}{Character vector specifying which panels to render, including one or more of the following:
\itemize{
\item \code{"background"} (the default), \code{"foreground"}: faces laying behind or in front of
the cube's interior volume, respectively. These panels vary depending on plot rotation.
\item \code{"xmin"}, \code{"ymax"}, etc.: names of specific cube faces.
\item \code{"all"}, \code{"none"}: display the full cube or remove all faces.
}
See \link{cube_theming} for details on panel styling, including transparency of foreground panels.}

\item{xlabels, ylabels, zlabels}{Character strings or length-2 character vectors specifying
axis label (text and title) placement. Labels are placed inline with grid lines for
the selected panel face. For each axis, there are four potential panels where labels
could be placed, and two potential edges for each panel. Labels can only be placed on
visible faces (see \code{panels} argument). Each parameter accepts:
\itemize{
\item \code{"auto"} (default): Automatic edge selection based on an algorithm
that prioritizes edges that are visible on the periphery of the plot and considers
several attributes of face geometry for better readability.
\item \code{c("face1", "face2")}: Manual edge specification using two adjacent
face names (e.g., \code{c("xmin", "ymin")} selects the edge shared by the
xmin and ymin faces). The \strong{first face} in the vector determines which
face the axis labels will be aligned with, while the second face
identifies which edge of this face gets labelled. Available face names are:
"xmin", "xmax", "ymin", "ymax", "zmin", "zmax".
}
See \link{cube_theming} for details on axis label styling.}

\item{title_position}{Character string controlling axis title placement. Currently only
affects titles for internal axes (not on the plot periphery); titles for axes on the
plot periphery are always centered along the axis edge.
\itemize{
\item \code{"auto"} (default): Internal axis titles are placed at the near
end of the axis, outside the plot area.
\item \code{"center"}: Internal axis titles are centered along the axis edge.
}}

\item{rotate_labels}{Logical indicating whether axis labels (text and titles) should automatically
rotate to align with the projected axis directions. When \code{FALSE}, uses theme
text and title angle settings.}

\item{scales}{Character string specifying aspect ratio behavior:
\itemize{
\item \code{"free"} (default): Each axis scales independently to fill cube space,
then \code{ratio} applies to standardized coordinates. This gives maximum
visual range for each dimension.
\item \code{"fixed"}: Maintains proportional relationships in raw data values,
as scaled by \code{ratio}. Similar to \code{coord_fixed()} but for 3D
(visual ratios match the labeled axis ranges).
}}

\item{scale_depth}{Controls how axis-related elements scale with viewing
distance under perspective projection, in which objects nearer the viewer are drawn
larger. Accepts:
\itemize{
\item \code{TRUE} (default) or \code{FALSE}: full depth scaling, or none.
\item A single number: applied to every element. \code{1} is the default
strength, \code{0} disables scaling, values below 1 subdue the effect and
values above 1 exaggerate it. Negative values are an error.
\item A named numeric vector, to set elements individually. Names are
\code{"grid"} (panel gridlines), \code{"border"} (cube edges), and
\code{"text"} (axis tick labels), after the theme elements they govern.
Unnamed elements keep their default.
}
Strength \code{s} maps a depth factor \code{d} to \code{d^s}, so the effect
scales geometrically and never produces a non-positive size. This affects axis-related
elements only; layers have their own \code{scale_depth} parameters. Axis
titles are never depth-scaled, since a single title has no series of sizes to
read as perspective. Has no effect when \code{persp = FALSE}, where all depth
factors are 1.}

\item{ratio}{Numeric vector of length 3 specifying relative axis lengths as
\code{c(x, y, z)}. Defaults to \code{c(1, 1, 1)} for equal proportions.
\itemize{
\item With \code{scales = "free"}: Ratios apply to scaled cube coordinates
\item With \code{scales = "fixed"}: Ratios apply to original data coordinates
}}

\item{zoom}{Numeric value controlling the framing of the plot. Values greater
than 1 zoom in (tighter framing, may crop edges), values less than 1
zoom out (more whitespace around the plot). Default is 1.}

\item{light}{Lighting specification for the plot. Usually left unset: add
\code{\link[=light]{light()}} to the plot instead, which is the recommended way to set plot-level
lighting and can appear anywhere in the plot expression. Supplying \code{light}
here as well as adding \code{\link[=light]{light()}} to the plot is an error. Use \code{NULL} to
disable lighting, or \code{light("none")} for an equivalent lighting object.}

\item{...}{Additional arguments reserved for internal use.}
}
\value{
A \code{Coord} object that can be added to a ggplot.
}
\description{
\code{coord_3d} is a 3D coordinate system that creates a 2D view of 3D data.
This is the essential core component of any plot made with \code{ggcube}.
It supports rotation, perspective projection, and options for controlling plot
aspect ratios, panel selection, axis label placement, and lighting.
}
\examples{
# base plot used in examples
p <- ggplot() +
  geom_function_3d(
    aes(fill = after_stat(z), color = after_stat(z)),
    fun = function(x, y) sin(x) * cos(y),
    xlim = c(-pi, pi), ylim = c(-pi, pi), n = 50) +
  light("direct", contrast = .7) +
  scale_fill_viridis_c() +
  scale_color_viridis_c() +
  theme(legend.position = "none")

# 3D plot with default coord settings
p + coord_3d()

\donttest{

# Use `pitch`, `roll`, `yaw` to control plot rotation ----------------------

# zero rotation gives view from x-y face
p + coord_3d(pitch = 0, roll = 0, yaw = 0)

# pitch rotates plot around y axis
p + coord_3d(pitch = 30, roll = 0, yaw = 0)

# roll rotates plot around x axis
p + coord_3d(pitch = 0, roll = 30, yaw = 0)

# yaw rotates plot around z axis
p + coord_3d(pitch = 0, roll = 0, yaw = 30)

# combine them to achieve arbitrary rotations
p + coord_3d(pitch = 20, roll = 40, yaw = 60)


# Use `persp` and `dist` to control perspective effects --------------------

# strong perspective effect as if seen from very close
p + coord_3d(dist = 1)

# weaker perspective effects as if seen from far away
p + coord_3d(dist = 3)

# orthographic projection (`dist = Inf` would be equivalent but it errors)
p + coord_3d(persp = FALSE)


# Use `scales` and `ratio` to control aspect ratio -------------------------

# The default "free" scales shown above give cube with maximum visual range.
# Use "fixed" scales to make figure match data scales, like coord_fixed.
p + coord_3d(scales = "fixed")

# Custom aspect ratios: make y twice as long visually
p + coord_3d(scales = "free", ratio = c(1, 2, 1))

# Combine behaviors: fix scales and make y twice as long
p + coord_3d(scales = "fixed", ratio = c(1, 2, 1))


# Use `panels` to select which cube faces to render ------------------------

p + coord_3d(panels = c("xmin", "xmax", "zmax"))

# foreground panels default to 20\% opaque (can be styled using `theme()`)
p + coord_3d(panels = "all")


# Use label params to control axis text placement and rotation -------------

p + coord_3d(xlabels = c("ymax", "zmax"),
             zlabels = c("xmax", "ymin"))

p + coord_3d(rotate_labels = FALSE)

}

}
\seealso{
\code{\link[=light]{light()}} for lighting specification, \link{cube_theming} for panel and
text styling, \link{polygon_params} for 3D-related parameters for polygon layers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/light.R
\name{light}
\alias{light}
\title{Lighting specification for 3D surface rendering}
\usage{
light(
  method = "diffuse",
  direction = c(-0.5, 0, 1),
  anchor = "scene",
  position = NULL,
  distance_falloff = FALSE,
  fill = TRUE,
  color = TRUE,
  mode = "hsv",
  contrast = 1,
  backface_scale = -1,
  backface_offset = 0
)
}
\arguments{
\item{method}{Character string specifying lighting model:
\itemize{
\item \code{"diffuse"}: The default. Atmospheric lighting with soft shadows (only surfaces pointing
directly away from light source are fully dark; base color occurs on surfaces perpendicular to light)
\item \code{"direct"}: Direct lighting with hard shadows (all surfaces angled beyond 90 degrees from
light source are fully dark; base color occurs on surfaces angled 45 degrees toward light)
\item \code{"rgb"}: Map surface orientation to three dimensional color space. Rather than the light model
darkening or brightening base colors as in the other models, this model generates a wholly new set of
colors. If this option is selected, parameters like \code{mode} and \code{contrast} are ignored, but
\code{fill}, \code{color}, and \code{direction} still apply.
\item \code{"none"}: Disable lighting entirely. All other parameters are ignored.
Equivalent to \code{light(fill = FALSE, color = FALSE)}, and to passing the
bare string \code{"none"} to a \code{light} argument.
}}

\item{direction}{Numeric vector of length 3 specifying direction in 3D space (x, y, z) that
light comes from for directional lighting. Depending on \code{anchor}, direction is relative to
either the plot axes or the viewing pane. The default is \code{c(-.5, 0, 1)}, giving
diagonal lighting from the upper right edge with default rotation and anchor. At least
one value must be non-zero. Values are automatically normalized, so magnitude doesn't
matter, only sign and relative magnitude. Direction is specified in visual space (relative
to the rendered cube's bounding box), not data space. This argument is ignored if
\code{position} is provided.}

\item{anchor}{Character string specifying the reference frame for light direction:
\itemize{
\item \code{"scene"} (default): Light \code{direction} is fixed relative to the data/scene.
Light rotates with the plot, so regardless of rotation, \code{direction = c(1, 0, 0)}
lights surfaces facing the "xmax" cube face, for example.
\item \code{"camera"}: Light \code{direction} is fixed relative to the camera/viewer.
Light direction stays in place regardless of rotation, so \code{direction = c(1, 0, 0)}
lights surfaces facing the right side of the plot, for example.
}
When all rotation parameters are zero, these options give the same result.}

\item{position}{Numeric vector of length 3 specifying light source position in
data coordinate space for positional lighting. When specified, each face gets
its own light direction calculated from the light position to the face center.
Mutually exclusive with \code{direction}. Default is NULL (use directional lighting).}

\item{distance_falloff}{Logical indicating whether to apply distance-based
intensity falloff for positional lighting using inverse square law
(intensity proportional to 1/distance²). Only used when \code{position} is specified.
Default is FALSE.}

\item{fill}{Logical indicating whether to apply lighting to fill colors. Default is TRUE.}

\item{color}{Logical indicating whether to apply lighting to border/line colors. Default is TRUE.}

\item{mode}{Character string specifying color lighting mode:
\itemize{
\item \code{"hsv"}: The default. Modifies \emph{value} component of \emph{HSV} color
(fades to bright colors at high end, black at low end).
\item \code{"hsl"}: Modifies \emph{lightness} component of \emph{HSL} color
(fades to white at high end, black at low end).
}
These two options give identical results for grayscale colors.}

\item{contrast}{Numeric value greater than zero controlling the intensity of lighting effects.
1.0 (the default) gives full black-to-white range. Values less than 1 give subtler effects, while
values greater than 1 give more dramatic effects.}

\item{backface_scale, backface_offset}{Numeric values that determine how "frontface" light values get
modified (scaled and then offset) to derive "backface" light values. A backface is the side of a
polygon that faces the underside of a surface or the inside of a volume. Frontface light values are
typically in the range \verb{[-1, 1]} (unless \code{contrast} is boosted). The default scale of -1 gives
backfaces highly contrasting lighting to frontfaces. To light backfaces the same as frontfaces,
set scale to 1. To uniformly darken (brighten) all backfaces, use a negative (positive) offset.}
}
\value{
A \code{lighting} object that can be added to a plot or passed to polygon-based 3D layers
or to \code{coord_3d()}.
}
\description{
Creates a lighting specification object for use with 3D polygon layers.
Lighting modifies the brightness of fill and/or base color aesthetics based on surface
orientation (i.e., it implements form shadows but not cast shadows).
Various options are available to control light qualities and light source location.
The result can be added directly to a plot with \code{+ light()} to set lighting for
the whole plot. It can also be passed to individual \verb{geom_*_3d()} layer functions,
which takes precedence over plot-level lighting.
}
\details{
Note that light-like effects can also be achieved in some stats by mapping color
aesthestics to computed variables such as \code{after_stat(dzdx)}; see \code{\link[=geom_surface_3d]{geom_surface_3d()}}
for examples.
}
\examples{
# base plot used in examples
p <- ggplot(sphere_points, aes(x, y, z)) +
  geom_hull_3d(fill = "#9e2602", color = "#5e1600") +
  coord_3d(scales = "fixed")

# Light qualities ----------------------------------------------------------

# default `"diffuse"` lighting
p

\donttest{

# use `"direct"` lighting to apply full shade to unlit surfaces
p + light(method = "direct")

# use `"hsl"` mode to fade highlights to white
p + light(mode = "hsl")

# adjust lighting intensity with `contrast`
p + light(mode = "hsl", contrast = 1.5)

# use `"rgb"` lighting to map face orientation to 3D color space
p + light(method = "rgb")


# Lighting targets ---------------------------------------------------------

# use `fill` and `color` to select which aesthetics get modified by light
p + light(fill = TRUE, color = FALSE)

# apply lighting to aesthetically mapped colors, with shaded guide to match
p + geom_hull_3d(aes(fill = x, color = x)) +
      scale_fill_viridis_c() +
      scale_color_viridis_c() +
      guides(fill = guide_colorbar_3d()) +
      light(mode = "hsl", contrast = .9)

# disable lighting entirely
p + light("none")

# if provided, layer-level lighting overrides plot-level lighting
p + light("direct") + # plot-level: affects original layer
  geom_hull_3d(aes(x = x + 2.5), fill = "#9e2602", color = "#5e1600",
      light = light("direct", mode = "hsl", direction = c(0, -1, 0)))


# Light sources ------------------------------------------------------------

# directional light from the xmin-ymin-zmin direction
# (`direction` is relative to rotated axes with default `anchor = "scene"`)
p + light(direction = c(-1, -1, -1))

# directional light from top-right corner of figure
# (`anchor = "camera"` makes `direction` fixed relative to the plot)
p + light(direction = c(1, 1, 0), anchor = "camera")

# positional light source within plot
ggplot(mountain, aes(x, y, z)) +
  stat_surface_3d(fill = "red", color = "red") +
  light(position = c(.5, .7, 95), distance_falloff = TRUE,
        mode = "hsl", contrast = .9) +
  coord_3d(ratio = c(1.5, 2, 1))


# Backface lighting --------------------------------------------------------

# backfaces get "opposite" lighting by default (`backface_scale = -1`)
p <- ggplot() +
  geom_function_3d(fun = function(x, y) x^2 + y^2,
    xlim = c(-3, 3), ylim = c(-3, 3),
    fill = "steelblue", color = "steelblue") +
    coord_3d(pitch = 0, roll = -70, yaw = 0)
p + light(mode = "hsl")

# use `backface_scale = 1` to light backfaces as if they're frontfaces
p + light(backface_scale = 1, mode = "hsl")

# use `backface_offset` to uniformly darken (or lighten) backfaces
p + light(backface_scale = 1, mode = "hsl", backface_offset = -.5)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-hull-3d.R
\name{geom_hull_3d}
\alias{geom_hull_3d}
\alias{stat_hull_3d}
\title{3D convex and alpha hulls}
\usage{
geom_hull_3d(
  mapping = NULL,
  data = NULL,
  stat = StatHull3D,
  position = "identity",
  ...,
  method = "convex",
  radius = NULL,
  singular = FALSE,
  light = NULL,
  cull_backfaces = TRUE,
  sort_method = NULL,
  scale_depth = TRUE,
  inherit.aes = TRUE,
  show.legend = TRUE
)

stat_hull_3d(
  mapping = NULL,
  data = NULL,
  geom = GeomPolygon3D,
  position = "identity",
  ...,
  method = "convex",
  radius = NULL,
  singular = FALSE,
  light = NULL,
  cull_backfaces = TRUE,
  sort_method = NULL,
  scale_depth = TRUE,
  inherit.aes = TRUE,
  show.legend = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. The required
aesthetics are \code{x}, \code{y}, and \code{z}. Additional aesthetics can use computed
variables with \code{\link[ggplot2:after_stat]{ggplot2::after_stat()}}.}

\item{data}{The data to be displayed in this layer.}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{StatHull3D}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}

\item{method}{Triangulation method. Either:
\itemize{
\item \code{"convex"}: Convex hull triangulation (default). Requires the \pkg{geometry} package.
\item \code{"alpha"}: Alpha shape triangulation (can capture non-convex topologies)
}}

\item{radius}{The "alpha" parameter when alpha method is used.
A face is included in the resulting alpha shape if it can be "exposed" by a sphere of this radius.
If NULL (the default), a simple heuristic based on the data scale is used to calculate a radius value.
Note that alpha shapes are quite sensitive to the coordinate scales of your data. See Details section.}

\item{singular}{Whether to include "singular" faces when \code{method = "alpha"}.
These are faces that do not bound any enclosed volume, appearing as isolated
sheets dangling from the surface, and are usually a sign that \code{radius} is
slightly too small. Because they enclose nothing, they have no well-defined
outward direction: their lighting is arbitrary and they are never culled.
Defaults to \code{FALSE}.}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{geom}{The geometric object used to display the data. Defaults to \code{GeomPolygon3D.}}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Turns 3D point clouds into surface hulls consisting of triangular polygons,
using either convex hull or alpha shape algorithms.
}
\section{Grouping}{

Hulls respect ggplot2 grouping aesthetics. To create separate hulls for different
subsets of your data, use \code{aes(group = category_variable)} or similar grouping aesthetics.
Each group will get its own independent hull.
}

\section{Alpha scale sensitivity}{

Alpha shape method is highly sensitive to coordinate scales. \code{radius} is expressed
in data units, so a value that works for data scaled 0-1 will likely fail for data
scaled 0-1000.
Guidelines for choosing radius:
\itemize{
\item Start at a small fraction of your data's overall extent and adjust based on results
\item For data with mixed scales (e.g., x: 0-1, y: 0-1000), consider rescaling your data first
\item Larger radius values → smoother, more connected surfaces
\item Smaller radius values → more detailed surfaces, but may fragment
\item If you get no triangles, try increasing radius by 10x
\item If surface fills unwanted holes, try decreasing radius by 10x
}
}

\section{Aesthetics}{

\code{stat_hull_3d()} requires the following aesthetics:
\itemize{
\item \strong{x}: X coordinate
\item \strong{y}: Y coordinate
\item \strong{z}: Z coordinate
}
}

\section{Computed variables}{

\itemize{
\item \code{normal_x}, \code{normal_y}, \code{normal_z}: Surface normal components
}
}

\examples{
\dontshow{if (requireNamespace("geometry", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# A solid torus: a shape whose hole a convex hull cannot represent
set.seed(1)
n <- 2000
theta <- runif(n, 0, 2 * pi)  # position around the ring
phi <- runif(n, 0, 2 * pi)    # position around the tube
rho <- sqrt(runif(n))         # distance from the tube center
torus <- data.frame(
  x = (3 + rho * cos(phi)) * cos(theta),
  y = (3 + rho * cos(phi)) * sin(theta),
  z = rho * sin(phi)
)

# Convex hull: the hole is bridged over
ggplot(torus, aes(x, y, z)) +
  geom_hull_3d(method = "convex", fill = "gray40") +
  coord_3d(scales = "fixed")

# Alpha shape: the hole is preserved
\donttest{
ggplot(torus, aes(x, y, z)) +
  geom_hull_3d(method = "alpha", radius = 0.6, fill = "gray40") +
  coord_3d(scales = "fixed")
}

# Use `cull_backfaces = FALSE` to render far side of hull
\donttest{
ggplot(torus, aes(x, y, z)) +
  geom_hull_3d( # default culling for comparison
    method = "alpha", radius = 0.6, light = NULL,
    fill = "steelblue", color = "darkred", linewidth = .5, alpha = .5) +
  geom_hull_3d( # culling disabled
    aes(x = x + 9), cull_backfaces = FALSE,
    method = "alpha", radius = 0.6, light = NULL,
    fill = "steelblue", color = "darkred", linewidth = .5, alpha = .5) +
  coord_3d(scales = "fixed")
}

# Use grouping to build separate hulls for data subsets
ggplot(iris, aes(Petal.Length, Sepal.Length, Sepal.Width,
                 color = Species, fill = Species)) +
      geom_hull_3d() +
      coord_3d(scales = "fixed")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems, \link{geom_polygon_3d} for the
default geometry with depth sorting, \code{\link[=light]{light()}} for lighting specifications.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{grid_generation}
\alias{grid_generation}
\title{Grid generation}
\description{
Parameters defining the geometry, resolution, and orientation of a regular grid
of surface tiles, as used by various ggcube layer functions including
\code{\link[=geom_smooth_3d]{geom_smooth_3d()}}, \code{\link[=geom_function_3d]{geom_function_3d()}}, \code{\link[=geom_density_3d]{geom_density_3d()}}, and
\code{\link[=geom_surface_3d]{geom_surface_3d()}}.
}
\section{Arguments}{

\describe{
\item{\code{grid}}{Character specifying tile geometry. Options:
\describe{
\item{\code{"rectangle"}}{Rectangular grid (the default).}
\item{\code{"right1"}}{Rectangular grid with each quad split into two right triangles
along the bottom-left to top-right diagonal.}
\item{\code{"right2"}}{Like \code{"right1"} but split along the bottom-right to top-left
diagonal.}
\item{\code{"equilateral"}}{Equilateral triangular lattice (tessellated via Delaunay
triangulation). Can prevent lighting artifacts where a surface curves past
parallel with the sight line.}
}
}
\item{\code{n}}{Either a single integer specifying grid resolution in both dimensions,
or a vector of length 2 specifying \code{c(nx, ny)} for separate x and y resolutions.
Default is \code{40}. Higher values create smoother surfaces but slower rendering.}
\item{\code{direction}}{Either \code{"x"} (the default) or \code{"y"}, specifying the orientation
of tile rows. Ignored for rectangular grids.}
\item{\code{trim}}{Logical. Only relevant for \code{grid = "equilateral"}. If \code{TRUE}
(default), trims edge vertices so that grid boundaries are straight lines. If
\code{FALSE}, preserves the full lattice, resulting in a grid with irregular edges.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{sort_params}
\alias{sort_params}
\title{Sorting parameters}
\arguments{
\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}
}
\description{
Sorting parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-function-3d.R
\name{stat_function_3d}
\alias{stat_function_3d}
\alias{geom_function_3d}
\title{3D surface from a function}
\usage{
stat_function_3d(
  mapping = NULL,
  data = ensure_nonempty_data,
  geom = "surface_3d",
  position = "identity",
  ...,
  fun = NULL,
  xlim = NULL,
  ylim = NULL,
  grid = "rectangle",
  n = 40,
  direction = "x",
  trim = TRUE,
  cull_backfaces = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

geom_function_3d(
  mapping = NULL,
  data = ensure_nonempty_data,
  stat = "function_3d",
  position = "identity",
  ...,
  fun = NULL,
  xlim = NULL,
  ylim = NULL,
  grid = "rectangle",
  n = 40,
  direction = "x",
  trim = TRUE,
  cull_backfaces = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. By default, fill
is mapped to \code{after_stat(z)}.}

\item{data}{Ignored; this stat generates its own data.}

\item{geom}{Geom to use for rendering. Defaults to \code{"surface_3d"} for mesh
surfaces. Use \code{"ridgeline_3d"} for ridgeline rendering.}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed to the layer.}

\item{fun}{Function to evaluate. Must accept (x, y) and return numeric z values.}

\item{xlim, ylim}{Length-2 numeric vectors giving the x and y ranges over which to
plot the function.}

\item{grid, n, direction, trim}{Parameters determining the geometry, resolution, and
orientation of the surface grid. See \link{grid_generation} for details.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{stat}{Statistical transformation to use on the data. Defaults to \code{"surface_3d"}.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Evaluates a function f(x,y) = z over a regular grid and renders the result
as a 3D surface or ridgeline plot.
}
\section{Computed variables}{

\describe{
\item{x, y, z}{Grid coordinates and function values}
\item{dzdx, dzdy}{Partial derivatives at each point}
\item{slope}{Gradient magnitude: sqrt(dzdx^2 + dzdy^2)}
\item{aspect}{Direction of steepest slope: atan2(dzdy, dzdx)}
}
}

\examples{
# Basic function surface
ggplot() +
  geom_function_3d(fun = function(x, y) sin(x) * cos(y),
                   xlim = c(-pi, pi), ylim = c(-pi, pi)) +
  coord_3d()

\donttest{

# Fill by slope
ggplot() +
  geom_function_3d(fun = function(x, y) x^2 + y^2,
                   xlim = c(-2, 2), ylim = c(-2, 2),
                   aes(fill = after_stat(slope)),
                   grid = "equilateral") +
  scale_fill_viridis_c() +
  coord_3d()

# As ridgelines
ggplot() +
  stat_function_3d(fun = function(x, y) dnorm(x) * dnorm(y) * 10,
                   xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), n = c(15, 30),
                   geom = "ridgeline_3d", base = 0, light = "none",
                   fill = "black", color = "white") +
  coord_3d()

}

}
\seealso{
\code{\link[=geom_surface_3d]{geom_surface_3d()}}, \code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}}, \code{\link[=coord_3d]{coord_3d()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-polygon-3d.R
\name{geom_polygon_3d}
\alias{geom_polygon_3d}
\title{3D polygon geometry with depth sorting}
\usage{
geom_polygon_3d(
  mapping = NULL,
  data = NULL,
  stat = StatIdentity3D,
  position = "identity",
  ...,
  rule = "evenodd",
  sort_method = "auto",
  scale_depth = TRUE,
  force_convex = FALSE,
  cull_backfaces = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}.}

\item{data}{The data to be displayed in this layer. Note that if you specify \code{light} or
\code{cull_backfaces}, behavior will depend on the "winding order" of polygon vertices, with
the counter-clockwise face considered the "front".}

\item{stat}{The statistical transformation to use on the data. Defaults to \code{"identity_3d"}.}

\item{position}{Position adjustment, defaults to "identity". To collapse the result
onto one 2D surface, use \code{position_on_face()}.}

\item{...}{Other arguments passed on to the layer function (typically GeomPolygon3D), such as
aesthetics like \code{colour}, \code{fill}, \code{linewidth}, \code{annotate = annotate_3d(...)}, etc.}

\item{rule}{Either \code{"evenodd"} or \code{"winding"}. If polygons with holes are
being drawn (using the \code{subgroup} aesthetic) this argument defines how the
hole coordinates are interpreted. See \code{\link[ggplot2:geom_polygon]{ggplot2::geom_polygon()}} for reference.}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
\code{geom_polygon_3d()} renders 3D polygons with proper depth sorting for realistic
3D surface visualization. It's designed to work with surface data
from \code{\link[=stat_hull_3d]{stat_hull_3d()}} and \code{\link[=stat_surface_3d]{stat_surface_3d()}}, as well as regular polygon data like maps.
}
\section{Aesthetics}{

\code{geom_polygon_3d()} requires:
\itemize{
\item \strong{x}: X coordinate
\item \strong{y}: Y coordinate
\item \strong{z}: Z coordinate (for depth sorting)
\item \strong{group}: Polygon grouping variable
}

And understands these additional aesthetics:
\itemize{
\item \code{fill}: Polygon fill color
\item \code{colour}: Border color
\item \code{linewidth}: Border line width
\item \code{linetype}: Border line type
\item \code{alpha}: Transparency
\item \code{subgroup}: Secondary grouping for polygons with holes
\item \code{order}: Vertex order within polygons (for proper polygon construction)
}
}

\examples{
# Typically used via stats such as stat_surface_3d() or stat_hull_3d()
ggplot(sphere_points, aes(x, y, z)) +
  stat_hull_3d(method = "convex", fill = "dodgerblue",
            light = light(fill = TRUE, mode = "hsl")) +
  coord_3d()

# Can be used directly with properly structured data
triangles <- data.frame(x = rep(c(3, 2, 1), 3),
                            y = rep(c(1, 3, 1), 3),
                            z = rep(1:3, each = 3),
                            shape = rep(letters[1:3], each = 3))
ggplot(triangles, aes(x, y, z, fill = shape)) +
  geom_polygon_3d(color = "black") +
  coord_3d()

# Use `sort_method` to choose between depth sorting algorithms
d <- data.frame(group = rep(letters[1:3], each = 4),
                x = c(1, 1, 2, 2,   1, 1, 3, 3,   2, 2, 3, 3),
                y = rep(c(1, 2, 2, 1), 3),
                z = rep(c(1, 1.5, 2), each = 4))
p <- ggplot(d, aes(x, y, z, group = group, fill = group)) +
      coord_3d(pitch = 50, roll = 20, yaw = 0,
               scales = "fixed", light = "none") +
      theme_light()

# fast, but rendering order is incorrect in this particular example
p + geom_polygon_3d(color = "black", linewidth = 1, alpha = .75,
      sort_method = "painter")

# correct rendering order (but slower for large data sets)
p + geom_polygon_3d(color = "black", linewidth = 1, alpha = .75,
      sort_method = "pairwise")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stat-surface-3d.R
\name{stat_surface_3d}
\alias{stat_surface_3d}
\alias{geom_surface_3d}
\title{3D surface from point data}
\usage{
stat_surface_3d(
  mapping = NULL,
  data = NULL,
  geom = "surface_3d",
  position = "identity",
  ...,
  cull_backfaces = FALSE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)

geom_surface_3d(
  mapping = NULL,
  data = NULL,
  stat = "surface_3d",
  position = "identity",
  ...,
  method = "auto",
  grid = NULL,
  cull_backfaces = FALSE,
  sort_method = "auto",
  scale_depth = TRUE,
  force_convex = TRUE,
  light = NULL,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[=aes]{aes()}}. Must include
\code{x}, \code{y}, and \code{z}. By default, fill is mapped to \code{after_stat(z)}.}

\item{data}{Data frame containing point coordinates.}

\item{geom}{Geom to use for rendering. Defaults to \code{"surface_3d"} for mesh
surfaces. Use \code{"ridgeline_3d"} for ridgeline rendering.}

\item{position}{Position adjustment, defaults to "identity".}

\item{...}{Other arguments passed to the layer.}

\item{cull_backfaces}{Logical indicating whether to remove back-facing polygons
from rendering. This is primarily for performance optimization but may be useful
for aesthetic reasons in some situations. Backfaces are determined using
screen-space winding order after 3D transformation. Defaults vary by geometry
type: FALSE for open surface-type geometries, TRUE for solid objects (hulls,
voxels, etc. where backfaces are generally hidden unless frontfaces are transparent
or explicitly disabled).}

\item{light}{A lighting specification object created by \code{\link[=light]{light()}}, \code{light("none")}
or the string \code{"none"} to disable lighting, or \code{NULL} to inherit plot-level
lighting. Set plot-level lighting by adding \code{\link[=light]{light()}} to the plot, and
layer-specific lighting via the \code{light} argument of \verb{geom_*_3d()} functions,
which takes precedence.}

\item{na.rm}{If \code{FALSE}, missing values are removed.}

\item{show.legend}{Logical indicating whether this layer should be included in legends.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics.}

\item{stat}{Statistical transformation to use on the data. Defaults to \code{"surface_3d"}.}

\item{method}{Tessellation method passed to \code{\link[=geom_surface_3d]{geom_surface_3d()}}:
"auto" (default), "grid", or "delaunay".}

\item{grid}{Tile geometry for regular grids: "rectangle" (default),
"right1", or "right2".}

\item{sort_method}{Depth sorting algorithm. See \link{sorting_methods} for details.}

\item{scale_depth}{Controls whether polygon linewidths are scaled to make closer lines
wider and farther lines narrower, based on the mean depth of a polygon. \code{TRUE}
(the default) applies full scaling and \code{FALSE} disables it; a number sets the
strength directly, where \code{1} matches \code{TRUE}, \code{0} matches \code{FALSE}, values below
1 subdue the effect and values above 1 exaggerate it. This affects the layer
only; use \code{coord_3d(scale_depth = )} for gridlines, cube borders, axis ticks,
and axis text.}

\item{force_convex}{Logical indicating whether to remove polygon vertices that are not part of the
convex hull. Default value varies by geom. Specifying TRUE can help reduce artifacts in surfaces
that have polygon tiles that wrap over a visible horizon. For prism-type geoms like columns and
voxels, FALSE is safe because polygons fill always be convex.}
}
\value{
A \code{Layer} object that can be added to a ggplot.
}
\description{
Takes user-provided (x, y, z) point data and prepares it for surface
rendering. If data form a regular grid, can render either a GeomSurface3D
of rectangular or right-triangular tiles, or a GeomRidgeline3D or GeomContour3D
set of surface slices; otherwise, renders irregular triangular tiles via Delaunay
trianuglation.
}
\details{
For regular grids, computes point-level gradients. Works with
both \code{\link[=geom_surface_3d]{geom_surface_3d()}} for mesh rendering and \code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}} for
ridgeline rendering.
}
\section{Computed variables}{

For regular grid data:
\describe{
\item{dzdx, dzdy}{Partial derivatives at each point}
\item{slope}{Gradient magnitude: sqrt(dzdx^2 + dzdy^2)}
\item{aspect}{Direction of steepest slope: atan2(dzdy, dzdx)}
}

For irregular data, gradient variables are NA.
}

\section{Grid detection}{

The stat automatically detects whether data forms a regular grid by checking
if \code{nrow(data) == length(unique(x)) * length(unique(y))}. Regular grids
get point-level gradient computation; irregular point clouds are passed
through for Delaunay tessellation by the geom.
}

\examples{
# Regular grid data ------------------------------------------

# simulated data and base plot for basic surface
d <- dplyr::mutate(tidyr::expand_grid(x = -10:10, y = -10:10),
                   z = sqrt(x^2 + y^2) / 1.5,
                   z = cos(z) - z)
p <- ggplot(d, aes(x, y, z)) +
      coord_3d(light = light(mode = "hsl", direction = c(1, 0, 0)))

# surface with 3d lighting
p + geom_surface_3d(fill = "steelblue", color = "steelblue", linewidth = .2)

\donttest{

# mesh wireframe (`fill = NA`) with aes line color
p + geom_surface_3d(aes(color = z), fill = NA,
                    linewidth = .5, light = "none") +
      scale_color_gradientn(colors = c("black", "blue", "red"))

# triangulated surface (can prevent lighting flaws)
p + geom_surface_3d(fill = "#9e2602", color = "black", grid = "right2")

# use after_stat to access computed surface-orientation variables
p + geom_surface_3d(aes(fill = after_stat(slope)), grid = "right2") +
      scale_fill_viridis_c() +
      guides(fill = guide_colorbar_3d())

# use `group` to plot data for multiple surfaces
d <- expand.grid(x = -5:5, y = -5:5)
d$z <- d$x^2 - d$y^2
d$g <- "a"
d2 <- d
d2$z <- d$z + 15
d2$g <- "b"
ggplot(rbind(d, d2), aes(x, y, z, group = g, fill = g)) +
      coord_3d(light = "none") +
      geom_surface_3d(color = "black", alpha = .5, light = NULL)

# terrain surface with topographic hillshade and elevational fill
ggplot(mountain, aes(x, y, z, fill = z, color = z)) +
      geom_surface_3d(light = light(direction = c(1, 0, .5),
                                    mode = "hsv", contrast = 1.5),
                      linewidth = .2) +
      coord_3d(ratio = c(1, 1.5, .75)) +
      theme_light() +
      scale_fill_gradientn(colors = c("darkgreen", "rosybrown4", "gray60")) +
      scale_color_gradientn(colors = c("darkgreen", "rosybrown4", "gray60")) +
      guides(fill = guide_colorbar_3d())

# stack of flat surfaces (e.g. a time series of raster maps)
d <- expand.grid(lon = 1:20, lat = 1:20, time = 1:6)
d$value <- sin(-d$lon * d$lat / 10) + d$time / 2
ggplot(d, aes(lon, lat, time, group = time,
              fill = value, color = value)) +
   geom_surface_3d(light = "none") +
   coord_3d() +
   scale_color_viridis_c() + scale_fill_viridis_c()

# stat_surface_3d with alternative geoms ----------------------------

# horizontal slices with geom_ridgeline_3d
ggplot(mountain, aes(x, y, z)) +
  stat_surface_3d(geom = "ridgeline_3d",
                  fill = "black", color = "white",
                  light = "none", linewidth = .1) +
      coord_3d(ratio = c(1, 1.5, .75), yaw = 45)

# elevation contours with geom_contour_3d
ggplot(mountain, aes(x, y, z, fill = z)) +
  stat_surface_3d(geom = "contour_3d", light = "none",
                  bins = 25, sort_method = "pairwise",
                  color = "black") +
      coord_3d(ratio = c(1, 1.5, .75), yaw = 45) +
      scale_fill_viridis_c(option = "B")


# Irregular point data ---------------------------------------

set.seed(42)
pts <- data.frame(x = runif(200, -2, 2), y = runif(200, -2, 2))
pts$z <- with(pts, sin(x) * cos(y))

ggplot(pts, aes(x, y, z = z, fill = z)) +
  stat_surface_3d(sort_method = "pairwise") +
  scale_fill_viridis_c() +
  coord_3d(light = "none")

}

}
\seealso{
\code{\link[=geom_surface_3d]{geom_surface_3d()}}, \code{\link[=geom_ridgeline_3d]{geom_ridgeline_3d()}}, \code{\link[=stat_function_3d]{stat_function_3d()}},
\code{\link[=coord_3d]{coord_3d()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shared-docs.R
\name{sorting_methods}
\alias{sorting_methods}
\title{Depth sorting method}
\description{
The \code{sort_method} argument determines the order in which elements are rendered.
This controls depth sorting for all geometry types within a layer, including
polygons, points, segments, and text. Default varies by geometry type.
}
\section{Options}{

\describe{
\item{\code{"painter"}}{Elements are sorted by the mean depth (distance from
viewer after rotation) of their vertices. This is fast, but can give incorrect
results when primitives overlap in screen space at different depths.}
\item{\code{"pairwise"}}{A more intensive sorting algorithm that compares every
pair of elements to determine occlusion order. Uses type-specific geometric
tests: polygon overlap detection for polygon-polygon pairs, point-in-polygon
tests for polygon-point pairs, line clipping for polygon-segment pairs, and
line intersection for segment-segment pairs. When elements are coplanar,
smaller primitives (points, segments) render on top of larger ones (polygons).
Slower but more accurate.}
\item{\code{"auto"}}{Uses pairwise if the data has fewer than 500 rows, and
painter otherwise.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/position-on-face.R
\name{position_on_face}
\alias{position_on_face}
\title{Position for projecting 2D and 3D layers onto cube faces}
\usage{
position_on_face(faces = "zmin", axes = NULL)
}
\arguments{
\item{faces}{Character string or vector specifying which cube face(s) to project onto.
Valid options are: "xmin", "xmax", "ymin", "ymax", "zmin", "zmax", "3D".
"3D" indicates the raw, non-flattened 3D position.
Multiple faces and "3D" are only supported for 3D layers (when \code{axes = NULL}).}

\item{axes}{For 2D layers only: Character vector of length 2 specifying which
3D dimensions the 2D layer's x and y aesthetics represent. For example,
\code{c("x", "z")} means the 2D x-axis maps to the 3D x-axis and the 2D y-axis
maps to the 3D z-axis. For 3D layers, use \code{NULL} (default) to preserve
the existing x,y,z mapping.}
}
\value{
A ggplot position object that can be supplied to the \code{position} argument
to layer functions.
}
\description{
Enables layers to be projected onto 2D faces of the 3D coordinate cube.
It can be used to flatten 3D ggcube layers onto a single cube face as a way of
visualizing them in 2D, or to add certain natively 2D ggplot2 layers like
\code{geom_density_2d()} or \code{geom_smooth()} to a cube face.
}
\details{
This position adjustment supports both 2D and 3D layers:

\strong{For 3D layers} (\code{axes = NULL}):
\itemize{
\item Data already has x,y,z coordinates in the correct order
\item Simply adds projection metadata for coord_3d to place the layer on the specified face(s)
\item The face coordinate will be overridden during coordinate transformation
\item Multiple faces are supported - the layer will be duplicated on each specified face.
All specified faces inherit aesthetics from the layer function; if you want different
parameters for different faces and on the primary 3D layer, add each as a separate layer call.
}

\strong{For 2D layers} (\code{axes = c("dim1", "dim2")}):
\itemize{
\item Renames the layer's x,y columns to match the specified 3D axes
\item Adds the missing third dimension
\item Adds projection metadata for coord_3d
\item Only single faces are supported for 2D layers
}

The actual projection happens during coordinate transformation in coord_3d.

\strong{Compatibility Note:} This position adjustment is not compatible with all 2D stats.
It works well with \code{stat_density_2d()} and other
stats that don't depend heavily on scale ranges during computation, but may cause
errors or rendering issues with \code{stat_density_2d_filled()} and similar stats that
generate polygons based on scale domains, as well as with layers like \code{stat_bin_2d()}
that return position variables other than x and y. For \code{geom_smooth()}, this position
works only if you set \code{se = FALSE} as shown in the example.
}
\examples{

# 3D point layer in raw 3D form, and projected onto 2D face
ggplot(sphere_points, aes(x, y, z)) +
  geom_point_3d(position = position_on_face("zmin"), color = "red") +
  geom_point_3d(color = "black") + # add this layer last so it appears on top
  coord_3d()

# 3D layer projected to multiple faces
set.seed(1)
d <- data.frame(x = round(rnorm(10)), y = round(rnorm(10)), z = round(rnorm(10)))
ggplot(d, aes(x, y, z)) +
  stat_voxel_3d(color = "black", fill = "steelblue",
    light = light(direction = c(1, 1, 0), mode = "hsv"),
    position = position_on_face(c("3D", "zmin", "xmax", "ymax"))) +
  coord_3d()

# 3D layer projected differently on individual faces
ggplot(sphere_points, aes(x, y, z)) +
  stat_hull_3d(position = position_on_face("zmin"), fill = "black") +
  geom_point_3d(position = position_on_face("ymax")) +
  geom_path_3d(position = position_on_face("xmax")) +
  stat_hull_3d(color = "black") +
  coord_3d()

# 2D density contour on a specific face
ggplot(iris, aes(Sepal.Length, Sepal.Width, Petal.Length, color = Species)) +
  stat_density_2d(position = position_on_face(faces = "zmin", axes = c("x", "y"))) +
  geom_point_3d() +
  coord_3d()

# Distinct 2D layers projected to different faces
ggplot(mtcars) +
  geom_smooth(aes(mpg, qsec), color = "red", alpha = .5, se = FALSE,
    position = position_on_face(faces = "ymax", axes = c("x", "z"))) +
  geom_density_2d(aes(mpg, wt), alpha = .5,
    position = position_on_face(faces = "zmin", axes = c("x", "y"))) +
  geom_path(aes(wt, qsec), color = "forestgreen", alpha = .5,
    position = position_on_face(faces = "xmax", axes = c("y", "z"))) +
  geom_point_3d(aes(mpg, wt, qsec)) +
    coord_3d() +
  theme_light()

}
\seealso{
\code{\link[=coord_3d]{coord_3d()}} for 3D coordinate systems, \code{\link[ggplot2:stat_density_2d]{ggplot2::stat_density_2d()}} and other
2D statistical transformations that can be projected onto faces.
}
